package jp.juggler.emoji;

import android.support.annotation.DrawableRes;
import android.support.annotation.NonNull;
import android.support.annotation.StringRes;
import android.util.SparseArray;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

@SuppressWarnings({ "WeakerAccess", "unused" })
public class EmojiMap201709 {
	
	public static final int CATEGORY_PEOPLE =0;
	public static final int CATEGORY_NATURE =1;
	public static final int CATEGORY_FOODS =2;
	public static final int CATEGORY_ACTIVITY =3;
	public static final int CATEGORY_PLACES =4;
	public static final int CATEGORY_OBJECTS =5;
	public static final int CATEGORY_SYMBOLS =6;
	public static final int CATEGORY_FLAGS =7;
	
	public static class Category {
		public final int category_id; // this is String resource id;
		public final ArrayList< String > emoji_list = new ArrayList<>();
		
		public Category(int category_id){
			this.category_id = category_id;
		}
	}
	
	public static class EmojiInfo {
		@DrawableRes public final int image_id;
		@NonNull public final String unified;
		
		public EmojiInfo( @DrawableRes int image_id, @NonNull String unified ){
			this.image_id = image_id;
			this.unified = unified;
		}
	}
	
	// 表示に使う。絵文字のユニコードシーケンスから画像リソースIDへのマップ
	public static final HashMap< String, Integer > sUTF16ToImageId = new HashMap<>();
	
	// 表示と投稿に使う。絵文字のショートコードから画像リソースIDとユニコードシーケンスへのマップ
	public static final HashMap< String, EmojiInfo > sShortNameToImageId = new HashMap<>();
	
	// 入力補完に使う。絵文字のショートコードのソートされたリスト
	public static final ArrayList< String > sShortNameList = new ArrayList<>();
	
	// ピッカーに使う。カテゴリのリスト
	public static final SparseArray< Category > sCategoryMap = new SparseArray<>();
	
	static{
		initAll();
		Collections.sort( sShortNameList );
	}
	
	private static void code( @DrawableRes int image_id, @NonNull String code ){
		// 素の数字とcopyright,registered, trademark は絵文字にしない
		int c = code.charAt( 0 );
		if( code.length() == 1 && c <= 0xae || c == 0x2122 ){
			return;
		}
		sUTF16ToImageId.put( code, image_id );
	}
	
	private static void name( @NonNull String name, @DrawableRes int image_id, @NonNull String unified ){
		sShortNameToImageId.put( name, new EmojiInfo( image_id, unified ) );
		sShortNameList.add( name );
	}
	
	private static void category( @StringRes int string_id, @NonNull String name ){
		Category c = sCategoryMap.get( string_id );
		if( c == null ){
			c = new Category( string_id );
			sCategoryMap.put( string_id, c );
		}
		c.emoji_list.add( name );
	}
	
	//////////////////////////////////////////////////////////////////////////
	// 以下は自動生成コード
	
	private static void init1(){
		code( R.drawable.emj_0023, "\u0023" );
		code( R.drawable.emj_0023_fe0f_20e3, "\u0023\u20e3" );
		code( R.drawable.emj_0023_fe0f_20e3, "\u0023\ufe0f\u20e3" );
		code( R.drawable.emj_0023_fe0f_20e3, "\ue210" );
		code( R.drawable.emj_0023_fe0f_20e3, "\ue6e0" );
		code( R.drawable.emj_0023_fe0f_20e3, "\ueb84" );
		code( R.drawable.emj_0023_fe0f_20e3, "\udbba\udc2c" );
		code( R.drawable.emj_002a, "\u002a" );
		code( R.drawable.emj_002a_fe0f_20e3, "\u002a\u20e3" );
		code( R.drawable.emj_002a_fe0f_20e3, "\u002a\ufe0f\u20e3" );
		code( R.drawable.emj_0030, "\u0030" );
		code( R.drawable.emj_0030_fe0f_20e3, "\u0030\u20e3" );
		code( R.drawable.emj_0030_fe0f_20e3, "\u0030\ufe0f\u20e3" );
		code( R.drawable.emj_0030_fe0f_20e3, "\ue225" );
		code( R.drawable.emj_0030_fe0f_20e3, "\ue5ac" );
		code( R.drawable.emj_0030_fe0f_20e3, "\ue6eb" );
		code( R.drawable.emj_0030_fe0f_20e3, "\udbba\udc37" );
		code( R.drawable.emj_0031, "\u0031" );
		code( R.drawable.emj_0031_fe0f_20e3, "\u0031\u20e3" );
		code( R.drawable.emj_0031_fe0f_20e3, "\u0031\ufe0f\u20e3" );
		code( R.drawable.emj_0031_fe0f_20e3, "\ue21c" );
		code( R.drawable.emj_0031_fe0f_20e3, "\ue522" );
		code( R.drawable.emj_0031_fe0f_20e3, "\ue6e2" );
		code( R.drawable.emj_0031_fe0f_20e3, "\udbba\udc2e" );
		code( R.drawable.emj_0032, "\u0032" );
		code( R.drawable.emj_0032_fe0f_20e3, "\u0032\u20e3" );
		code( R.drawable.emj_0032_fe0f_20e3, "\u0032\ufe0f\u20e3" );
		code( R.drawable.emj_0032_fe0f_20e3, "\ue21d" );
		code( R.drawable.emj_0032_fe0f_20e3, "\ue523" );
		code( R.drawable.emj_0032_fe0f_20e3, "\ue6e3" );
		code( R.drawable.emj_0032_fe0f_20e3, "\udbba\udc2f" );
		code( R.drawable.emj_0033, "\u0033" );
		code( R.drawable.emj_0033_fe0f_20e3, "\u0033\u20e3" );
		code( R.drawable.emj_0033_fe0f_20e3, "\u0033\ufe0f\u20e3" );
		code( R.drawable.emj_0033_fe0f_20e3, "\ue21e" );
		code( R.drawable.emj_0033_fe0f_20e3, "\ue524" );
		code( R.drawable.emj_0033_fe0f_20e3, "\ue6e4" );
		code( R.drawable.emj_0033_fe0f_20e3, "\udbba\udc30" );
		code( R.drawable.emj_0034, "\u0034" );
		code( R.drawable.emj_0034_fe0f_20e3, "\u0034\u20e3" );
		code( R.drawable.emj_0034_fe0f_20e3, "\u0034\ufe0f\u20e3" );
		code( R.drawable.emj_0034_fe0f_20e3, "\ue21f" );
		code( R.drawable.emj_0034_fe0f_20e3, "\ue525" );
		code( R.drawable.emj_0034_fe0f_20e3, "\ue6e5" );
		code( R.drawable.emj_0034_fe0f_20e3, "\udbba\udc31" );
		code( R.drawable.emj_0035, "\u0035" );
		code( R.drawable.emj_0035_fe0f_20e3, "\u0035\u20e3" );
		code( R.drawable.emj_0035_fe0f_20e3, "\u0035\ufe0f\u20e3" );
		code( R.drawable.emj_0035_fe0f_20e3, "\ue220" );
		code( R.drawable.emj_0035_fe0f_20e3, "\ue526" );
		code( R.drawable.emj_0035_fe0f_20e3, "\ue6e6" );
		code( R.drawable.emj_0035_fe0f_20e3, "\udbba\udc32" );
		code( R.drawable.emj_0036, "\u0036" );
		code( R.drawable.emj_0036_fe0f_20e3, "\u0036\u20e3" );
		code( R.drawable.emj_0036_fe0f_20e3, "\u0036\ufe0f\u20e3" );
		code( R.drawable.emj_0036_fe0f_20e3, "\ue221" );
		code( R.drawable.emj_0036_fe0f_20e3, "\ue527" );
		code( R.drawable.emj_0036_fe0f_20e3, "\ue6e7" );
		code( R.drawable.emj_0036_fe0f_20e3, "\udbba\udc33" );
		code( R.drawable.emj_0037, "\u0037" );
		code( R.drawable.emj_0037_fe0f_20e3, "\u0037\u20e3" );
		code( R.drawable.emj_0037_fe0f_20e3, "\u0037\ufe0f\u20e3" );
		code( R.drawable.emj_0037_fe0f_20e3, "\ue222" );
		code( R.drawable.emj_0037_fe0f_20e3, "\ue528" );
		code( R.drawable.emj_0037_fe0f_20e3, "\ue6e8" );
		code( R.drawable.emj_0037_fe0f_20e3, "\udbba\udc34" );
		code( R.drawable.emj_0038, "\u0038" );
		code( R.drawable.emj_0038_fe0f_20e3, "\u0038\u20e3" );
		code( R.drawable.emj_0038_fe0f_20e3, "\u0038\ufe0f\u20e3" );
		code( R.drawable.emj_0038_fe0f_20e3, "\ue223" );
		code( R.drawable.emj_0038_fe0f_20e3, "\ue529" );
		code( R.drawable.emj_0038_fe0f_20e3, "\ue6e9" );
		code( R.drawable.emj_0038_fe0f_20e3, "\udbba\udc35" );
		code( R.drawable.emj_0039, "\u0039" );
		code( R.drawable.emj_0039_fe0f_20e3, "\u0039\u20e3" );
		code( R.drawable.emj_0039_fe0f_20e3, "\u0039\ufe0f\u20e3" );
		code( R.drawable.emj_0039_fe0f_20e3, "\ue224" );
		code( R.drawable.emj_0039_fe0f_20e3, "\ue52a" );
		code( R.drawable.emj_0039_fe0f_20e3, "\ue6ea" );
		code( R.drawable.emj_0039_fe0f_20e3, "\udbba\udc36" );
		code( R.drawable.emj_00a9_fe0f, "\u00a9" );
		code( R.drawable.emj_00a9_fe0f, "\u00a9\ufe0f" );
		code( R.drawable.emj_00a9_fe0f, "\ue24e" );
		code( R.drawable.emj_00a9_fe0f, "\ue558" );
		code( R.drawable.emj_00a9_fe0f, "\ue731" );
		code( R.drawable.emj_00a9_fe0f, "\udbba\udf29" );
		code( R.drawable.emj_00ae_fe0f, "\u00ae" );
		code( R.drawable.emj_00ae_fe0f, "\u00ae\ufe0f" );
		code( R.drawable.emj_00ae_fe0f, "\ue24f" );
		code( R.drawable.emj_00ae_fe0f, "\ue559" );
		code( R.drawable.emj_00ae_fe0f, "\ue736" );
		code( R.drawable.emj_00ae_fe0f, "\udbba\udf2d" );
		code( R.drawable.emj_1f004, "\ud83c\udc04" );
		code( R.drawable.emj_1f004, "\ue12d" );
		code( R.drawable.emj_1f004, "\ue5d1" );
		code( R.drawable.emj_1f004, "\udbba\udc0b" );
		code( R.drawable.emj_1f0cf, "\ud83c\udccf" );
		code( R.drawable.emj_1f0cf, "\ueb6f" );
		code( R.drawable.emj_1f0cf, "\udbba\udc12" );
		code( R.drawable.emj_1f170_fe0f, "\ud83c\udd70" );
		code( R.drawable.emj_1f170_fe0f, "\ud83c\udd70\ufe0f" );
	}
	private static void init2(){
		code( R.drawable.emj_1f170_fe0f, "\ue532" );
		code( R.drawable.emj_1f170_fe0f, "\ueb26" );
		code( R.drawable.emj_1f170_fe0f, "\udbb9\udd0b" );
		code( R.drawable.emj_1f171_fe0f, "\ud83c\udd71" );
		code( R.drawable.emj_1f171_fe0f, "\ud83c\udd71\ufe0f" );
		code( R.drawable.emj_1f171_fe0f, "\ue533" );
		code( R.drawable.emj_1f171_fe0f, "\ueb27" );
		code( R.drawable.emj_1f171_fe0f, "\udbb9\udd0c" );
		code( R.drawable.emj_1f17e_fe0f, "\ud83c\udd7e" );
		code( R.drawable.emj_1f17e_fe0f, "\ud83c\udd7e\ufe0f" );
		code( R.drawable.emj_1f17e_fe0f, "\ue535" );
		code( R.drawable.emj_1f17e_fe0f, "\ueb28" );
		code( R.drawable.emj_1f17e_fe0f, "\udbb9\udd0e" );
		code( R.drawable.emj_1f17f_fe0f, "\ud83c\udd7f" );
		code( R.drawable.emj_1f17f_fe0f, "\ud83c\udd7f\ufe0f" );
		code( R.drawable.emj_1f17f_fe0f, "\ue14f" );
		code( R.drawable.emj_1f17f_fe0f, "\ue4a6" );
		code( R.drawable.emj_1f17f_fe0f, "\ue66c" );
		code( R.drawable.emj_1f17f_fe0f, "\udbb9\udff6" );
		code( R.drawable.emj_1f18e, "\ud83c\udd8e" );
		code( R.drawable.emj_1f18e, "\ue534" );
		code( R.drawable.emj_1f18e, "\ueb29" );
		code( R.drawable.emj_1f18e, "\udbb9\udd0d" );
		code( R.drawable.emj_1f191, "\ud83c\udd91" );
		code( R.drawable.emj_1f191, "\ue5ab" );
		code( R.drawable.emj_1f191, "\ue6db" );
		code( R.drawable.emj_1f191, "\udbba\udf84" );
		code( R.drawable.emj_1f192, "\ud83c\udd92" );
		code( R.drawable.emj_1f192, "\ue214" );
		code( R.drawable.emj_1f192, "\uea85" );
		code( R.drawable.emj_1f192, "\udbba\udf38" );
		code( R.drawable.emj_1f193, "\ud83c\udd93" );
		code( R.drawable.emj_1f193, "\ue578" );
		code( R.drawable.emj_1f193, "\ue6d7" );
		code( R.drawable.emj_1f193, "\udbba\udf21" );
		code( R.drawable.emj_1f194, "\ud83c\udd94" );
		code( R.drawable.emj_1f194, "\ue229" );
		code( R.drawable.emj_1f194, "\ue6d8" );
		code( R.drawable.emj_1f194, "\uea88" );
		code( R.drawable.emj_1f194, "\udbba\udf81" );
		code( R.drawable.emj_1f195, "\ud83c\udd95" );
		code( R.drawable.emj_1f195, "\ue212" );
		code( R.drawable.emj_1f195, "\ue5b5" );
		code( R.drawable.emj_1f195, "\ue6dd" );
		code( R.drawable.emj_1f195, "\udbba\udf36" );
		code( R.drawable.emj_1f196, "\ud83c\udd96" );
		code( R.drawable.emj_1f196, "\ue72f" );
		code( R.drawable.emj_1f196, "\udbba\udf28" );
		code( R.drawable.emj_1f197, "\ud83c\udd97" );
		code( R.drawable.emj_1f197, "\ue24d" );
		code( R.drawable.emj_1f197, "\ue5ad" );
		code( R.drawable.emj_1f197, "\ue70b" );
		code( R.drawable.emj_1f197, "\udbba\udf27" );
		code( R.drawable.emj_1f198, "\ud83c\udd98" );
		code( R.drawable.emj_1f198, "\ue4e8" );
		code( R.drawable.emj_1f198, "\udbba\udf4f" );
		code( R.drawable.emj_1f199, "\ud83c\udd99" );
		code( R.drawable.emj_1f199, "\ue213" );
		code( R.drawable.emj_1f199, "\ue50f" );
		code( R.drawable.emj_1f199, "\udbba\udf37" );
		code( R.drawable.emj_1f19a, "\ud83c\udd9a" );
		code( R.drawable.emj_1f19a, "\ue12e" );
		code( R.drawable.emj_1f19a, "\ue5d2" );
		code( R.drawable.emj_1f19a, "\udbba\udf32" );
		code( R.drawable.emj_1f1e6, "\ud83c\udde6" );
		code( R.drawable.emj_1f1e6_1f1e8, "\ud83c\udde6\ud83c\udde8" );
		code( R.drawable.emj_1f1e6_1f1e9, "\ud83c\udde6\ud83c\udde9" );
		code( R.drawable.emj_1f1e6_1f1ea, "\ud83c\udde6\ud83c\uddea" );
		code( R.drawable.emj_1f1e6_1f1eb, "\ud83c\udde6\ud83c\uddeb" );
		code( R.drawable.emj_1f1e6_1f1ec, "\ud83c\udde6\ud83c\uddec" );
		code( R.drawable.emj_1f1e6_1f1ee, "\ud83c\udde6\ud83c\uddee" );
		code( R.drawable.emj_1f1e6_1f1f1, "\ud83c\udde6\ud83c\uddf1" );
		code( R.drawable.emj_1f1e6_1f1f2, "\ud83c\udde6\ud83c\uddf2" );
		code( R.drawable.emj_1f1e6_1f1f4, "\ud83c\udde6\ud83c\uddf4" );
		code( R.drawable.emj_1f1e6_1f1f6, "\ud83c\udde6\ud83c\uddf6" );
		code( R.drawable.emj_1f1e6_1f1f7, "\ud83c\udde6\ud83c\uddf7" );
		code( R.drawable.emj_1f1e6_1f1f8, "\ud83c\udde6\ud83c\uddf8" );
		code( R.drawable.emj_1f1e6_1f1f9, "\ud83c\udde6\ud83c\uddf9" );
		code( R.drawable.emj_1f1e6_1f1fa, "\ud83c\udde6\ud83c\uddfa" );
		code( R.drawable.emj_1f1e6_1f1fc, "\ud83c\udde6\ud83c\uddfc" );
		code( R.drawable.emj_1f1e6_1f1fd, "\ud83c\udde6\ud83c\uddfd" );
		code( R.drawable.emj_1f1e6_1f1ff, "\ud83c\udde6\ud83c\uddff" );
		code( R.drawable.emj_1f1e7, "\ud83c\udde7" );
		code( R.drawable.emj_1f1e7_1f1e6, "\ud83c\udde7\ud83c\udde6" );
		code( R.drawable.emj_1f1e7_1f1e7, "\ud83c\udde7\ud83c\udde7" );
		code( R.drawable.emj_1f1e7_1f1e9, "\ud83c\udde7\ud83c\udde9" );
		code( R.drawable.emj_1f1e7_1f1ea, "\ud83c\udde7\ud83c\uddea" );
		code( R.drawable.emj_1f1e7_1f1eb, "\ud83c\udde7\ud83c\uddeb" );
		code( R.drawable.emj_1f1e7_1f1ec, "\ud83c\udde7\ud83c\uddec" );
		code( R.drawable.emj_1f1e7_1f1ed, "\ud83c\udde7\ud83c\udded" );
		code( R.drawable.emj_1f1e7_1f1ee, "\ud83c\udde7\ud83c\uddee" );
		code( R.drawable.emj_1f1e7_1f1ef, "\ud83c\udde7\ud83c\uddef" );
		code( R.drawable.emj_1f1e7_1f1f1, "\ud83c\udde7\ud83c\uddf1" );
		code( R.drawable.emj_1f1e7_1f1f2, "\ud83c\udde7\ud83c\uddf2" );
		code( R.drawable.emj_1f1e7_1f1f3, "\ud83c\udde7\ud83c\uddf3" );
		code( R.drawable.emj_1f1e7_1f1f4, "\ud83c\udde7\ud83c\uddf4" );
		code( R.drawable.emj_1f1e7_1f1f6, "\ud83c\udde7\ud83c\uddf6" );
		code( R.drawable.emj_1f1e7_1f1f7, "\ud83c\udde7\ud83c\uddf7" );
		code( R.drawable.emj_1f1e7_1f1f8, "\ud83c\udde7\ud83c\uddf8" );
		code( R.drawable.emj_1f1e7_1f1f9, "\ud83c\udde7\ud83c\uddf9" );
		code( R.drawable.emj_1f1e7_1f1fb, "\ud83c\udde7\ud83c\uddfb" );
	}
	private static void init3(){
		code( R.drawable.emj_1f1e7_1f1fc, "\ud83c\udde7\ud83c\uddfc" );
		code( R.drawable.emj_1f1e7_1f1fe, "\ud83c\udde7\ud83c\uddfe" );
		code( R.drawable.emj_1f1e7_1f1ff, "\ud83c\udde7\ud83c\uddff" );
		code( R.drawable.emj_1f1e8, "\ud83c\udde8" );
		code( R.drawable.emj_1f1e8_1f1e6, "\ud83c\udde8\ud83c\udde6" );
		code( R.drawable.emj_1f1e8_1f1e8, "\ud83c\udde8\ud83c\udde8" );
		code( R.drawable.emj_1f1e8_1f1e9, "\ud83c\udde8\ud83c\udde9" );
		code( R.drawable.emj_1f1e8_1f1eb, "\ud83c\udde8\ud83c\uddeb" );
		code( R.drawable.emj_1f1e8_1f1ec, "\ud83c\udde8\ud83c\uddec" );
		code( R.drawable.emj_1f1e8_1f1ed, "\ud83c\udde8\ud83c\udded" );
		code( R.drawable.emj_1f1e8_1f1ee, "\ud83c\udde8\ud83c\uddee" );
		code( R.drawable.emj_1f1e8_1f1f0, "\ud83c\udde8\ud83c\uddf0" );
		code( R.drawable.emj_1f1e8_1f1f1, "\ud83c\udde8\ud83c\uddf1" );
		code( R.drawable.emj_1f1e8_1f1f2, "\ud83c\udde8\ud83c\uddf2" );
		code( R.drawable.emj_1f1e8_1f1f3, "\ud83c\udde8\ud83c\uddf3" );
		code( R.drawable.emj_1f1e8_1f1f3, "\ue513" );
		code( R.drawable.emj_1f1e8_1f1f3, "\ueb11" );
		code( R.drawable.emj_1f1e8_1f1f3, "\udbb9\udced" );
		code( R.drawable.emj_1f1e8_1f1f4, "\ud83c\udde8\ud83c\uddf4" );
		code( R.drawable.emj_1f1e8_1f1f5, "\ud83c\udde8\ud83c\uddf5" );
		code( R.drawable.emj_1f1e8_1f1f7, "\ud83c\udde8\ud83c\uddf7" );
		code( R.drawable.emj_1f1e8_1f1fa, "\ud83c\udde8\ud83c\uddfa" );
		code( R.drawable.emj_1f1e8_1f1fb, "\ud83c\udde8\ud83c\uddfb" );
		code( R.drawable.emj_1f1e8_1f1fc, "\ud83c\udde8\ud83c\uddfc" );
		code( R.drawable.emj_1f1e8_1f1fd, "\ud83c\udde8\ud83c\uddfd" );
		code( R.drawable.emj_1f1e8_1f1fe, "\ud83c\udde8\ud83c\uddfe" );
		code( R.drawable.emj_1f1e8_1f1ff, "\ud83c\udde8\ud83c\uddff" );
		code( R.drawable.emj_1f1e9, "\ud83c\udde9" );
		code( R.drawable.emj_1f1e9_1f1ea, "\ud83c\udde9\ud83c\uddea" );
		code( R.drawable.emj_1f1e9_1f1ea, "\ue50e" );
		code( R.drawable.emj_1f1e9_1f1ea, "\ueb0e" );
		code( R.drawable.emj_1f1e9_1f1ea, "\udbb9\udce8" );
		code( R.drawable.emj_1f1e9_1f1ec, "\ud83c\udde9\ud83c\uddec" );
		code( R.drawable.emj_1f1e9_1f1ef, "\ud83c\udde9\ud83c\uddef" );
		code( R.drawable.emj_1f1e9_1f1f0, "\ud83c\udde9\ud83c\uddf0" );
		code( R.drawable.emj_1f1e9_1f1f2, "\ud83c\udde9\ud83c\uddf2" );
		code( R.drawable.emj_1f1e9_1f1f4, "\ud83c\udde9\ud83c\uddf4" );
		code( R.drawable.emj_1f1e9_1f1ff, "\ud83c\udde9\ud83c\uddff" );
		code( R.drawable.emj_1f1ea, "\ud83c\uddea" );
		code( R.drawable.emj_1f1ea_1f1e6, "\ud83c\uddea\ud83c\udde6" );
		code( R.drawable.emj_1f1ea_1f1e8, "\ud83c\uddea\ud83c\udde8" );
		code( R.drawable.emj_1f1ea_1f1ea, "\ud83c\uddea\ud83c\uddea" );
		code( R.drawable.emj_1f1ea_1f1ec, "\ud83c\uddea\ud83c\uddec" );
		code( R.drawable.emj_1f1ea_1f1ed, "\ud83c\uddea\ud83c\udded" );
		code( R.drawable.emj_1f1ea_1f1f7, "\ud83c\uddea\ud83c\uddf7" );
		code( R.drawable.emj_1f1ea_1f1f8, "\ud83c\uddea\ud83c\uddf8" );
		code( R.drawable.emj_1f1ea_1f1f8, "\ue511" );
		code( R.drawable.emj_1f1ea_1f1f8, "\ue5d5" );
		code( R.drawable.emj_1f1ea_1f1f8, "\udbb9\udceb" );
		code( R.drawable.emj_1f1ea_1f1f9, "\ud83c\uddea\ud83c\uddf9" );
		code( R.drawable.emj_1f1ea_1f1fa, "\ud83c\uddea\ud83c\uddfa" );
		code( R.drawable.emj_1f1eb, "\ud83c\uddeb" );
		code( R.drawable.emj_1f1eb_1f1ee, "\ud83c\uddeb\ud83c\uddee" );
		code( R.drawable.emj_1f1eb_1f1ef, "\ud83c\uddeb\ud83c\uddef" );
		code( R.drawable.emj_1f1eb_1f1f0, "\ud83c\uddeb\ud83c\uddf0" );
		code( R.drawable.emj_1f1eb_1f1f2, "\ud83c\uddeb\ud83c\uddf2" );
		code( R.drawable.emj_1f1eb_1f1f4, "\ud83c\uddeb\ud83c\uddf4" );
		code( R.drawable.emj_1f1eb_1f1f7, "\ud83c\uddeb\ud83c\uddf7" );
		code( R.drawable.emj_1f1eb_1f1f7, "\ue50d" );
		code( R.drawable.emj_1f1eb_1f1f7, "\ueafa" );
		code( R.drawable.emj_1f1eb_1f1f7, "\udbb9\udce7" );
		code( R.drawable.emj_1f1ec, "\ud83c\uddec" );
		code( R.drawable.emj_1f1ec_1f1e6, "\ud83c\uddec\ud83c\udde6" );
		code( R.drawable.emj_1f1ec_1f1e7, "\ud83c\uddec\ud83c\udde7" );
		code( R.drawable.emj_1f1ec_1f1e7, "\ue510" );
		code( R.drawable.emj_1f1ec_1f1e7, "\ueb10" );
		code( R.drawable.emj_1f1ec_1f1e7, "\udbb9\udcea" );
		code( R.drawable.emj_1f1ec_1f1e9, "\ud83c\uddec\ud83c\udde9" );
		code( R.drawable.emj_1f1ec_1f1ea, "\ud83c\uddec\ud83c\uddea" );
		code( R.drawable.emj_1f1ec_1f1eb, "\ud83c\uddec\ud83c\uddeb" );
		code( R.drawable.emj_1f1ec_1f1ec, "\ud83c\uddec\ud83c\uddec" );
		code( R.drawable.emj_1f1ec_1f1ed, "\ud83c\uddec\ud83c\udded" );
		code( R.drawable.emj_1f1ec_1f1ee, "\ud83c\uddec\ud83c\uddee" );
		code( R.drawable.emj_1f1ec_1f1f1, "\ud83c\uddec\ud83c\uddf1" );
		code( R.drawable.emj_1f1ec_1f1f2, "\ud83c\uddec\ud83c\uddf2" );
		code( R.drawable.emj_1f1ec_1f1f3, "\ud83c\uddec\ud83c\uddf3" );
		code( R.drawable.emj_1f1ec_1f1f5, "\ud83c\uddec\ud83c\uddf5" );
		code( R.drawable.emj_1f1ec_1f1f6, "\ud83c\uddec\ud83c\uddf6" );
		code( R.drawable.emj_1f1ec_1f1f7, "\ud83c\uddec\ud83c\uddf7" );
		code( R.drawable.emj_1f1ec_1f1f8, "\ud83c\uddec\ud83c\uddf8" );
		code( R.drawable.emj_1f1ec_1f1f9, "\ud83c\uddec\ud83c\uddf9" );
		code( R.drawable.emj_1f1ec_1f1fa, "\ud83c\uddec\ud83c\uddfa" );
		code( R.drawable.emj_1f1ec_1f1fc, "\ud83c\uddec\ud83c\uddfc" );
		code( R.drawable.emj_1f1ec_1f1fe, "\ud83c\uddec\ud83c\uddfe" );
		code( R.drawable.emj_1f1ed, "\ud83c\udded" );
		code( R.drawable.emj_1f1ed_1f1f0, "\ud83c\udded\ud83c\uddf0" );
		code( R.drawable.emj_1f1ed_1f1f2, "\ud83c\udded\ud83c\uddf2" );
		code( R.drawable.emj_1f1ed_1f1f3, "\ud83c\udded\ud83c\uddf3" );
		code( R.drawable.emj_1f1ed_1f1f7, "\ud83c\udded\ud83c\uddf7" );
		code( R.drawable.emj_1f1ed_1f1f9, "\ud83c\udded\ud83c\uddf9" );
		code( R.drawable.emj_1f1ed_1f1fa, "\ud83c\udded\ud83c\uddfa" );
		code( R.drawable.emj_1f1ee, "\ud83c\uddee" );
		code( R.drawable.emj_1f1ee_1f1e8, "\ud83c\uddee\ud83c\udde8" );
		code( R.drawable.emj_1f1ee_1f1e9, "\ud83c\uddee\ud83c\udde9" );
		code( R.drawable.emj_1f1ee_1f1ea, "\ud83c\uddee\ud83c\uddea" );
		code( R.drawable.emj_1f1ee_1f1f1, "\ud83c\uddee\ud83c\uddf1" );
		code( R.drawable.emj_1f1ee_1f1f2, "\ud83c\uddee\ud83c\uddf2" );
		code( R.drawable.emj_1f1ee_1f1f3, "\ud83c\uddee\ud83c\uddf3" );
		code( R.drawable.emj_1f1ee_1f1f4, "\ud83c\uddee\ud83c\uddf4" );
		code( R.drawable.emj_1f1ee_1f1f6, "\ud83c\uddee\ud83c\uddf6" );
		code( R.drawable.emj_1f1ee_1f1f7, "\ud83c\uddee\ud83c\uddf7" );
	}
	private static void init4(){
		code( R.drawable.emj_1f1ee_1f1f8, "\ud83c\uddee\ud83c\uddf8" );
		code( R.drawable.emj_1f1ee_1f1f9, "\ud83c\uddee\ud83c\uddf9" );
		code( R.drawable.emj_1f1ee_1f1f9, "\ue50f" );
		code( R.drawable.emj_1f1ee_1f1f9, "\ueb0f" );
		code( R.drawable.emj_1f1ee_1f1f9, "\udbb9\udce9" );
		code( R.drawable.emj_1f1ef, "\ud83c\uddef" );
		code( R.drawable.emj_1f1ef_1f1ea, "\ud83c\uddef\ud83c\uddea" );
		code( R.drawable.emj_1f1ef_1f1f2, "\ud83c\uddef\ud83c\uddf2" );
		code( R.drawable.emj_1f1ef_1f1f4, "\ud83c\uddef\ud83c\uddf4" );
		code( R.drawable.emj_1f1ef_1f1f5, "\ud83c\uddef\ud83c\uddf5" );
		code( R.drawable.emj_1f1ef_1f1f5, "\ue4cc" );
		code( R.drawable.emj_1f1ef_1f1f5, "\ue50b" );
		code( R.drawable.emj_1f1ef_1f1f5, "\udbb9\udce5" );
		code( R.drawable.emj_1f1f0, "\ud83c\uddf0" );
		code( R.drawable.emj_1f1f0_1f1ea, "\ud83c\uddf0\ud83c\uddea" );
		code( R.drawable.emj_1f1f0_1f1ec, "\ud83c\uddf0\ud83c\uddec" );
		code( R.drawable.emj_1f1f0_1f1ed, "\ud83c\uddf0\ud83c\udded" );
		code( R.drawable.emj_1f1f0_1f1ee, "\ud83c\uddf0\ud83c\uddee" );
		code( R.drawable.emj_1f1f0_1f1f2, "\ud83c\uddf0\ud83c\uddf2" );
		code( R.drawable.emj_1f1f0_1f1f3, "\ud83c\uddf0\ud83c\uddf3" );
		code( R.drawable.emj_1f1f0_1f1f5, "\ud83c\uddf0\ud83c\uddf5" );
		code( R.drawable.emj_1f1f0_1f1f7, "\ud83c\uddf0\ud83c\uddf7" );
		code( R.drawable.emj_1f1f0_1f1f7, "\ue514" );
		code( R.drawable.emj_1f1f0_1f1f7, "\ueb12" );
		code( R.drawable.emj_1f1f0_1f1f7, "\udbb9\udcee" );
		code( R.drawable.emj_1f1f0_1f1fc, "\ud83c\uddf0\ud83c\uddfc" );
		code( R.drawable.emj_1f1f0_1f1fe, "\ud83c\uddf0\ud83c\uddfe" );
		code( R.drawable.emj_1f1f0_1f1ff, "\ud83c\uddf0\ud83c\uddff" );
		code( R.drawable.emj_1f1f1, "\ud83c\uddf1" );
		code( R.drawable.emj_1f1f1_1f1e6, "\ud83c\uddf1\ud83c\udde6" );
		code( R.drawable.emj_1f1f1_1f1e7, "\ud83c\uddf1\ud83c\udde7" );
		code( R.drawable.emj_1f1f1_1f1e8, "\ud83c\uddf1\ud83c\udde8" );
		code( R.drawable.emj_1f1f1_1f1ee, "\ud83c\uddf1\ud83c\uddee" );
		code( R.drawable.emj_1f1f1_1f1f0, "\ud83c\uddf1\ud83c\uddf0" );
		code( R.drawable.emj_1f1f1_1f1f7, "\ud83c\uddf1\ud83c\uddf7" );
		code( R.drawable.emj_1f1f1_1f1f8, "\ud83c\uddf1\ud83c\uddf8" );
		code( R.drawable.emj_1f1f1_1f1f9, "\ud83c\uddf1\ud83c\uddf9" );
		code( R.drawable.emj_1f1f1_1f1fa, "\ud83c\uddf1\ud83c\uddfa" );
		code( R.drawable.emj_1f1f1_1f1fb, "\ud83c\uddf1\ud83c\uddfb" );
		code( R.drawable.emj_1f1f1_1f1fe, "\ud83c\uddf1\ud83c\uddfe" );
		code( R.drawable.emj_1f1f2, "\ud83c\uddf2" );
		code( R.drawable.emj_1f1f2_1f1e6, "\ud83c\uddf2\ud83c\udde6" );
		code( R.drawable.emj_1f1f2_1f1e8, "\ud83c\uddf2\ud83c\udde8" );
		code( R.drawable.emj_1f1f2_1f1e9, "\ud83c\uddf2\ud83c\udde9" );
		code( R.drawable.emj_1f1f2_1f1ea, "\ud83c\uddf2\ud83c\uddea" );
		code( R.drawable.emj_1f1f2_1f1eb, "\ud83c\uddf2\ud83c\uddeb" );
		code( R.drawable.emj_1f1f2_1f1ec, "\ud83c\uddf2\ud83c\uddec" );
		code( R.drawable.emj_1f1f2_1f1ed, "\ud83c\uddf2\ud83c\udded" );
		code( R.drawable.emj_1f1f2_1f1f0, "\ud83c\uddf2\ud83c\uddf0" );
		code( R.drawable.emj_1f1f2_1f1f1, "\ud83c\uddf2\ud83c\uddf1" );
		code( R.drawable.emj_1f1f2_1f1f2, "\ud83c\uddf2\ud83c\uddf2" );
		code( R.drawable.emj_1f1f2_1f1f3, "\ud83c\uddf2\ud83c\uddf3" );
		code( R.drawable.emj_1f1f2_1f1f4, "\ud83c\uddf2\ud83c\uddf4" );
		code( R.drawable.emj_1f1f2_1f1f5, "\ud83c\uddf2\ud83c\uddf5" );
		code( R.drawable.emj_1f1f2_1f1f6, "\ud83c\uddf2\ud83c\uddf6" );
		code( R.drawable.emj_1f1f2_1f1f7, "\ud83c\uddf2\ud83c\uddf7" );
		code( R.drawable.emj_1f1f2_1f1f8, "\ud83c\uddf2\ud83c\uddf8" );
		code( R.drawable.emj_1f1f2_1f1f9, "\ud83c\uddf2\ud83c\uddf9" );
		code( R.drawable.emj_1f1f2_1f1fa, "\ud83c\uddf2\ud83c\uddfa" );
		code( R.drawable.emj_1f1f2_1f1fb, "\ud83c\uddf2\ud83c\uddfb" );
		code( R.drawable.emj_1f1f2_1f1fc, "\ud83c\uddf2\ud83c\uddfc" );
		code( R.drawable.emj_1f1f2_1f1fd, "\ud83c\uddf2\ud83c\uddfd" );
		code( R.drawable.emj_1f1f2_1f1fe, "\ud83c\uddf2\ud83c\uddfe" );
		code( R.drawable.emj_1f1f2_1f1ff, "\ud83c\uddf2\ud83c\uddff" );
		code( R.drawable.emj_1f1f3, "\ud83c\uddf3" );
		code( R.drawable.emj_1f1f3_1f1e6, "\ud83c\uddf3\ud83c\udde6" );
		code( R.drawable.emj_1f1f3_1f1e8, "\ud83c\uddf3\ud83c\udde8" );
		code( R.drawable.emj_1f1f3_1f1ea, "\ud83c\uddf3\ud83c\uddea" );
		code( R.drawable.emj_1f1f3_1f1eb, "\ud83c\uddf3\ud83c\uddeb" );
		code( R.drawable.emj_1f1f3_1f1ec, "\ud83c\uddf3\ud83c\uddec" );
		code( R.drawable.emj_1f1f3_1f1ee, "\ud83c\uddf3\ud83c\uddee" );
		code( R.drawable.emj_1f1f3_1f1f1, "\ud83c\uddf3\ud83c\uddf1" );
		code( R.drawable.emj_1f1f3_1f1f4, "\ud83c\uddf3\ud83c\uddf4" );
		code( R.drawable.emj_1f1f3_1f1f5, "\ud83c\uddf3\ud83c\uddf5" );
		code( R.drawable.emj_1f1f3_1f1f7, "\ud83c\uddf3\ud83c\uddf7" );
		code( R.drawable.emj_1f1f3_1f1fa, "\ud83c\uddf3\ud83c\uddfa" );
		code( R.drawable.emj_1f1f3_1f1ff, "\ud83c\uddf3\ud83c\uddff" );
		code( R.drawable.emj_1f1f4, "\ud83c\uddf4" );
		code( R.drawable.emj_1f1f4_1f1f2, "\ud83c\uddf4\ud83c\uddf2" );
		code( R.drawable.emj_1f1f5, "\ud83c\uddf5" );
		code( R.drawable.emj_1f1f5_1f1e6, "\ud83c\uddf5\ud83c\udde6" );
		code( R.drawable.emj_1f1f5_1f1ea, "\ud83c\uddf5\ud83c\uddea" );
		code( R.drawable.emj_1f1f5_1f1eb, "\ud83c\uddf5\ud83c\uddeb" );
		code( R.drawable.emj_1f1f5_1f1ec, "\ud83c\uddf5\ud83c\uddec" );
		code( R.drawable.emj_1f1f5_1f1ed, "\ud83c\uddf5\ud83c\udded" );
		code( R.drawable.emj_1f1f5_1f1f0, "\ud83c\uddf5\ud83c\uddf0" );
		code( R.drawable.emj_1f1f5_1f1f1, "\ud83c\uddf5\ud83c\uddf1" );
		code( R.drawable.emj_1f1f5_1f1f2, "\ud83c\uddf5\ud83c\uddf2" );
		code( R.drawable.emj_1f1f5_1f1f3, "\ud83c\uddf5\ud83c\uddf3" );
		code( R.drawable.emj_1f1f5_1f1f7, "\ud83c\uddf5\ud83c\uddf7" );
		code( R.drawable.emj_1f1f5_1f1f8, "\ud83c\uddf5\ud83c\uddf8" );
		code( R.drawable.emj_1f1f5_1f1f9, "\ud83c\uddf5\ud83c\uddf9" );
		code( R.drawable.emj_1f1f5_1f1fc, "\ud83c\uddf5\ud83c\uddfc" );
		code( R.drawable.emj_1f1f5_1f1fe, "\ud83c\uddf5\ud83c\uddfe" );
		code( R.drawable.emj_1f1f6, "\ud83c\uddf6" );
		code( R.drawable.emj_1f1f6_1f1e6, "\ud83c\uddf6\ud83c\udde6" );
		code( R.drawable.emj_1f1f7, "\ud83c\uddf7" );
		code( R.drawable.emj_1f1f7_1f1ea, "\ud83c\uddf7\ud83c\uddea" );
		code( R.drawable.emj_1f1f7_1f1f4, "\ud83c\uddf7\ud83c\uddf4" );
		code( R.drawable.emj_1f1f7_1f1f8, "\ud83c\uddf7\ud83c\uddf8" );
		code( R.drawable.emj_1f1f7_1f1fa, "\ud83c\uddf7\ud83c\uddfa" );
	}
	private static void init5(){
		code( R.drawable.emj_1f1f7_1f1fa, "\ue512" );
		code( R.drawable.emj_1f1f7_1f1fa, "\ue5d6" );
		code( R.drawable.emj_1f1f7_1f1fa, "\udbb9\udcec" );
		code( R.drawable.emj_1f1f7_1f1fc, "\ud83c\uddf7\ud83c\uddfc" );
		code( R.drawable.emj_1f1f8, "\ud83c\uddf8" );
		code( R.drawable.emj_1f1f8_1f1e6, "\ud83c\uddf8\ud83c\udde6" );
		code( R.drawable.emj_1f1f8_1f1e7, "\ud83c\uddf8\ud83c\udde7" );
		code( R.drawable.emj_1f1f8_1f1e8, "\ud83c\uddf8\ud83c\udde8" );
		code( R.drawable.emj_1f1f8_1f1e9, "\ud83c\uddf8\ud83c\udde9" );
		code( R.drawable.emj_1f1f8_1f1ea, "\ud83c\uddf8\ud83c\uddea" );
		code( R.drawable.emj_1f1f8_1f1ec, "\ud83c\uddf8\ud83c\uddec" );
		code( R.drawable.emj_1f1f8_1f1ed, "\ud83c\uddf8\ud83c\udded" );
		code( R.drawable.emj_1f1f8_1f1ee, "\ud83c\uddf8\ud83c\uddee" );
		code( R.drawable.emj_1f1f8_1f1ef, "\ud83c\uddf8\ud83c\uddef" );
		code( R.drawable.emj_1f1f8_1f1f0, "\ud83c\uddf8\ud83c\uddf0" );
		code( R.drawable.emj_1f1f8_1f1f1, "\ud83c\uddf8\ud83c\uddf1" );
		code( R.drawable.emj_1f1f8_1f1f2, "\ud83c\uddf8\ud83c\uddf2" );
		code( R.drawable.emj_1f1f8_1f1f3, "\ud83c\uddf8\ud83c\uddf3" );
		code( R.drawable.emj_1f1f8_1f1f4, "\ud83c\uddf8\ud83c\uddf4" );
		code( R.drawable.emj_1f1f8_1f1f7, "\ud83c\uddf8\ud83c\uddf7" );
		code( R.drawable.emj_1f1f8_1f1f8, "\ud83c\uddf8\ud83c\uddf8" );
		code( R.drawable.emj_1f1f8_1f1f9, "\ud83c\uddf8\ud83c\uddf9" );
		code( R.drawable.emj_1f1f8_1f1fb, "\ud83c\uddf8\ud83c\uddfb" );
		code( R.drawable.emj_1f1f8_1f1fd, "\ud83c\uddf8\ud83c\uddfd" );
		code( R.drawable.emj_1f1f8_1f1fe, "\ud83c\uddf8\ud83c\uddfe" );
		code( R.drawable.emj_1f1f8_1f1ff, "\ud83c\uddf8\ud83c\uddff" );
		code( R.drawable.emj_1f1f9, "\ud83c\uddf9" );
		code( R.drawable.emj_1f1f9_1f1e6, "\ud83c\uddf9\ud83c\udde6" );
		code( R.drawable.emj_1f1f9_1f1e8, "\ud83c\uddf9\ud83c\udde8" );
		code( R.drawable.emj_1f1f9_1f1e9, "\ud83c\uddf9\ud83c\udde9" );
		code( R.drawable.emj_1f1f9_1f1eb, "\ud83c\uddf9\ud83c\uddeb" );
		code( R.drawable.emj_1f1f9_1f1ec, "\ud83c\uddf9\ud83c\uddec" );
		code( R.drawable.emj_1f1f9_1f1ed, "\ud83c\uddf9\ud83c\udded" );
		code( R.drawable.emj_1f1f9_1f1ef, "\ud83c\uddf9\ud83c\uddef" );
		code( R.drawable.emj_1f1f9_1f1f0, "\ud83c\uddf9\ud83c\uddf0" );
		code( R.drawable.emj_1f1f9_1f1f1, "\ud83c\uddf9\ud83c\uddf1" );
		code( R.drawable.emj_1f1f9_1f1f2, "\ud83c\uddf9\ud83c\uddf2" );
		code( R.drawable.emj_1f1f9_1f1f3, "\ud83c\uddf9\ud83c\uddf3" );
		code( R.drawable.emj_1f1f9_1f1f4, "\ud83c\uddf9\ud83c\uddf4" );
		code( R.drawable.emj_1f1f9_1f1f7, "\ud83c\uddf9\ud83c\uddf7" );
		code( R.drawable.emj_1f1f9_1f1f9, "\ud83c\uddf9\ud83c\uddf9" );
		code( R.drawable.emj_1f1f9_1f1fb, "\ud83c\uddf9\ud83c\uddfb" );
		code( R.drawable.emj_1f1f9_1f1fc, "\ud83c\uddf9\ud83c\uddfc" );
		code( R.drawable.emj_1f1f9_1f1ff, "\ud83c\uddf9\ud83c\uddff" );
		code( R.drawable.emj_1f1fa, "\ud83c\uddfa" );
		code( R.drawable.emj_1f1fa_1f1e6, "\ud83c\uddfa\ud83c\udde6" );
		code( R.drawable.emj_1f1fa_1f1ec, "\ud83c\uddfa\ud83c\uddec" );
		code( R.drawable.emj_1f1fa_1f1f2, "\ud83c\uddfa\ud83c\uddf2" );
		code( R.drawable.emj_1f1fa_1f1f3, "\ud83c\uddfa\ud83c\uddf3" );
		code( R.drawable.emj_1f1fa_1f1f8, "\ud83c\uddfa\ud83c\uddf8" );
		code( R.drawable.emj_1f1fa_1f1f8, "\ue50c" );
		code( R.drawable.emj_1f1fa_1f1f8, "\ue573" );
		code( R.drawable.emj_1f1fa_1f1f8, "\udbb9\udce6" );
		code( R.drawable.emj_1f1fa_1f1fe, "\ud83c\uddfa\ud83c\uddfe" );
		code( R.drawable.emj_1f1fa_1f1ff, "\ud83c\uddfa\ud83c\uddff" );
		code( R.drawable.emj_1f1fb, "\ud83c\uddfb" );
		code( R.drawable.emj_1f1fb_1f1e6, "\ud83c\uddfb\ud83c\udde6" );
		code( R.drawable.emj_1f1fb_1f1e8, "\ud83c\uddfb\ud83c\udde8" );
		code( R.drawable.emj_1f1fb_1f1ea, "\ud83c\uddfb\ud83c\uddea" );
		code( R.drawable.emj_1f1fb_1f1ec, "\ud83c\uddfb\ud83c\uddec" );
		code( R.drawable.emj_1f1fb_1f1ee, "\ud83c\uddfb\ud83c\uddee" );
		code( R.drawable.emj_1f1fb_1f1f3, "\ud83c\uddfb\ud83c\uddf3" );
		code( R.drawable.emj_1f1fb_1f1fa, "\ud83c\uddfb\ud83c\uddfa" );
		code( R.drawable.emj_1f1fc, "\ud83c\uddfc" );
		code( R.drawable.emj_1f1fc_1f1eb, "\ud83c\uddfc\ud83c\uddeb" );
		code( R.drawable.emj_1f1fc_1f1f8, "\ud83c\uddfc\ud83c\uddf8" );
		code( R.drawable.emj_1f1fd, "\ud83c\uddfd" );
		code( R.drawable.emj_1f1fd_1f1f0, "\ud83c\uddfd\ud83c\uddf0" );
		code( R.drawable.emj_1f1fe, "\ud83c\uddfe" );
		code( R.drawable.emj_1f1fe_1f1ea, "\ud83c\uddfe\ud83c\uddea" );
		code( R.drawable.emj_1f1fe_1f1f9, "\ud83c\uddfe\ud83c\uddf9" );
		code( R.drawable.emj_1f1ff, "\ud83c\uddff" );
		code( R.drawable.emj_1f1ff_1f1e6, "\ud83c\uddff\ud83c\udde6" );
		code( R.drawable.emj_1f1ff_1f1f2, "\ud83c\uddff\ud83c\uddf2" );
		code( R.drawable.emj_1f1ff_1f1fc, "\ud83c\uddff\ud83c\uddfc" );
		code( R.drawable.emj_1f201, "\ud83c\ude01" );
		code( R.drawable.emj_1f201, "\ue203" );
		code( R.drawable.emj_1f201, "\udbba\udf24" );
		code( R.drawable.emj_1f202_fe0f, "\ud83c\ude02" );
		code( R.drawable.emj_1f202_fe0f, "\ud83c\ude02\ufe0f" );
		code( R.drawable.emj_1f202_fe0f, "\ue228" );
		code( R.drawable.emj_1f202_fe0f, "\uea87" );
		code( R.drawable.emj_1f202_fe0f, "\udbba\udf3f" );
		code( R.drawable.emj_1f21a, "\ud83c\ude1a" );
		code( R.drawable.emj_1f21a, "\ue216" );
		code( R.drawable.emj_1f21a, "\udbba\udf3a" );
		code( R.drawable.emj_1f22f, "\ud83c\ude2f" );
		code( R.drawable.emj_1f22f, "\ue22c" );
		code( R.drawable.emj_1f22f, "\uea8b" );
		code( R.drawable.emj_1f22f, "\udbba\udf40" );
		code( R.drawable.emj_1f232, "\ud83c\ude32" );
		code( R.drawable.emj_1f232, "\ue738" );
		code( R.drawable.emj_1f232, "\udbba\udf2e" );
		code( R.drawable.emj_1f233, "\ud83c\ude33" );
		code( R.drawable.emj_1f233, "\ue22b" );
		code( R.drawable.emj_1f233, "\ue739" );
		code( R.drawable.emj_1f233, "\uea8a" );
		code( R.drawable.emj_1f233, "\udbba\udf2f" );
		code( R.drawable.emj_1f234, "\ud83c\ude34" );
		code( R.drawable.emj_1f234, "\ue73a" );
		code( R.drawable.emj_1f234, "\udbba\udf30" );
	}
	private static void init6(){
		code( R.drawable.emj_1f235, "\ud83c\ude35" );
		code( R.drawable.emj_1f235, "\ue22a" );
		code( R.drawable.emj_1f235, "\ue73b" );
		code( R.drawable.emj_1f235, "\uea89" );
		code( R.drawable.emj_1f235, "\udbba\udf31" );
		code( R.drawable.emj_1f236, "\ud83c\ude36" );
		code( R.drawable.emj_1f236, "\ue215" );
		code( R.drawable.emj_1f236, "\udbba\udf39" );
		code( R.drawable.emj_1f237_fe0f, "\ud83c\ude37" );
		code( R.drawable.emj_1f237_fe0f, "\ud83c\ude37\ufe0f" );
		code( R.drawable.emj_1f237_fe0f, "\ue217" );
		code( R.drawable.emj_1f237_fe0f, "\udbba\udf3b" );
		code( R.drawable.emj_1f238, "\ud83c\ude38" );
		code( R.drawable.emj_1f238, "\ue218" );
		code( R.drawable.emj_1f238, "\udbba\udf3c" );
		code( R.drawable.emj_1f239, "\ud83c\ude39" );
		code( R.drawable.emj_1f239, "\ue227" );
		code( R.drawable.emj_1f239, "\uea86" );
		code( R.drawable.emj_1f239, "\udbba\udf3e" );
		code( R.drawable.emj_1f23a, "\ud83c\ude3a" );
		code( R.drawable.emj_1f23a, "\ue22d" );
		code( R.drawable.emj_1f23a, "\uea8c" );
		code( R.drawable.emj_1f23a, "\udbba\udf41" );
		code( R.drawable.emj_1f250, "\ud83c\ude50" );
		code( R.drawable.emj_1f250, "\ue226" );
		code( R.drawable.emj_1f250, "\ue4f7" );
		code( R.drawable.emj_1f250, "\udbba\udf3d" );
		code( R.drawable.emj_1f251, "\ud83c\ude51" );
		code( R.drawable.emj_1f251, "\ueb01" );
		code( R.drawable.emj_1f251, "\udbba\udf50" );
		code( R.drawable.emj_1f300, "\ud83c\udf00" );
		code( R.drawable.emj_1f300, "\ue443" );
		code( R.drawable.emj_1f300, "\ue469" );
		code( R.drawable.emj_1f300, "\ue643" );
		code( R.drawable.emj_1f300, "\udbb8\udc05" );
		code( R.drawable.emj_1f301, "\ud83c\udf01" );
		code( R.drawable.emj_1f301, "\ue598" );
		code( R.drawable.emj_1f301, "\ue644" );
		code( R.drawable.emj_1f301, "\udbb8\udc06" );
		code( R.drawable.emj_1f302, "\ud83c\udf02" );
		code( R.drawable.emj_1f302, "\ue43c" );
		code( R.drawable.emj_1f302, "\ue645" );
		code( R.drawable.emj_1f302, "\ueae8" );
		code( R.drawable.emj_1f302, "\udbb8\udc07" );
		code( R.drawable.emj_1f303, "\ud83c\udf03" );
		code( R.drawable.emj_1f303, "\ue44b" );
		code( R.drawable.emj_1f303, "\ue6b3" );
		code( R.drawable.emj_1f303, "\ueaf1" );
		code( R.drawable.emj_1f303, "\udbb8\udc08" );
		code( R.drawable.emj_1f304, "\ud83c\udf04" );
		code( R.drawable.emj_1f304, "\ue04d" );
		code( R.drawable.emj_1f304, "\ue63e" );
		code( R.drawable.emj_1f304, "\udbb8\udc09" );
		code( R.drawable.emj_1f305, "\ud83c\udf05" );
		code( R.drawable.emj_1f305, "\ue449" );
		code( R.drawable.emj_1f305, "\ue63e" );
		code( R.drawable.emj_1f305, "\ueaf4" );
		code( R.drawable.emj_1f305, "\udbb8\udc0a" );
		code( R.drawable.emj_1f306, "\ud83c\udf06" );
		code( R.drawable.emj_1f306, "\ue146" );
		code( R.drawable.emj_1f306, "\ue5da" );
		code( R.drawable.emj_1f306, "\udbb8\udc0b" );
		code( R.drawable.emj_1f307, "\ud83c\udf07" );
		code( R.drawable.emj_1f307, "\ue44a" );
		code( R.drawable.emj_1f307, "\ue5da" );
		code( R.drawable.emj_1f307, "\ue63e" );
		code( R.drawable.emj_1f307, "\udbb8\udc0c" );
		code( R.drawable.emj_1f308, "\ud83c\udf08" );
		code( R.drawable.emj_1f308, "\ue44c" );
		code( R.drawable.emj_1f308, "\ueaf2" );
		code( R.drawable.emj_1f308, "\udbb8\udc0d" );
		code( R.drawable.emj_1f309, "\ud83c\udf09" );
		code( R.drawable.emj_1f309, "\ue4bf" );
		code( R.drawable.emj_1f309, "\udbb8\udc10" );
		code( R.drawable.emj_1f30a, "\ud83c\udf0a" );
		code( R.drawable.emj_1f30a, "\ue43e" );
		code( R.drawable.emj_1f30a, "\ue73f" );
		code( R.drawable.emj_1f30a, "\ueb7c" );
		code( R.drawable.emj_1f30a, "\udbb8\udc38" );
		code( R.drawable.emj_1f30b, "\ud83c\udf0b" );
		code( R.drawable.emj_1f30b, "\ueb53" );
		code( R.drawable.emj_1f30b, "\udbb8\udc3a" );
		code( R.drawable.emj_1f30c, "\ud83c\udf0c" );
		code( R.drawable.emj_1f30c, "\ueb5f" );
		code( R.drawable.emj_1f30c, "\udbb8\udc3b" );
		code( R.drawable.emj_1f30d, "\ud83c\udf0d" );
		code( R.drawable.emj_1f30e, "\ud83c\udf0e" );
		code( R.drawable.emj_1f30f, "\ud83c\udf0f" );
		code( R.drawable.emj_1f30f, "\ue5b3" );
		code( R.drawable.emj_1f30f, "\udbb8\udc39" );
		code( R.drawable.emj_1f310, "\ud83c\udf10" );
		code( R.drawable.emj_1f311, "\ud83c\udf11" );
		code( R.drawable.emj_1f311, "\ue5a8" );
		code( R.drawable.emj_1f311, "\ue69c" );
		code( R.drawable.emj_1f311, "\udbb8\udc11" );
		code( R.drawable.emj_1f312, "\ud83c\udf12" );
		code( R.drawable.emj_1f313, "\ud83c\udf13" );
		code( R.drawable.emj_1f313, "\ue5aa" );
		code( R.drawable.emj_1f313, "\ue69e" );
		code( R.drawable.emj_1f313, "\udbb8\udc13" );
		code( R.drawable.emj_1f314, "\ud83c\udf14" );
	}
	private static void init7(){
		code( R.drawable.emj_1f314, "\ue5a9" );
		code( R.drawable.emj_1f314, "\ue69d" );
		code( R.drawable.emj_1f314, "\udbb8\udc12" );
		code( R.drawable.emj_1f315, "\ud83c\udf15" );
		code( R.drawable.emj_1f315, "\ue6a0" );
		code( R.drawable.emj_1f315, "\udbb8\udc15" );
		code( R.drawable.emj_1f316, "\ud83c\udf16" );
		code( R.drawable.emj_1f317, "\ud83c\udf17" );
		code( R.drawable.emj_1f318, "\ud83c\udf18" );
		code( R.drawable.emj_1f319, "\ud83c\udf19" );
		code( R.drawable.emj_1f319, "\ue04c" );
		code( R.drawable.emj_1f319, "\ue486" );
		code( R.drawable.emj_1f319, "\ue69f" );
		code( R.drawable.emj_1f319, "\udbb8\udc14" );
		code( R.drawable.emj_1f31a, "\ud83c\udf1a" );
		code( R.drawable.emj_1f31b, "\ud83c\udf1b" );
		code( R.drawable.emj_1f31b, "\ue489" );
		code( R.drawable.emj_1f31b, "\udbb8\udc16" );
		code( R.drawable.emj_1f31c, "\ud83c\udf1c" );
		code( R.drawable.emj_1f31d, "\ud83c\udf1d" );
		code( R.drawable.emj_1f31e, "\ud83c\udf1e" );
		code( R.drawable.emj_1f31f, "\ud83c\udf1f" );
		code( R.drawable.emj_1f31f, "\ue335" );
		code( R.drawable.emj_1f31f, "\ue48b" );
		code( R.drawable.emj_1f31f, "\udbba\udf69" );
		code( R.drawable.emj_1f320, "\ud83c\udf20" );
		code( R.drawable.emj_1f320, "\ue468" );
		code( R.drawable.emj_1f320, "\udbba\udf6a" );
		code( R.drawable.emj_1f321_fe0f, "\ud83c\udf21" );
		code( R.drawable.emj_1f321_fe0f, "\ud83c\udf21\ufe0f" );
		code( R.drawable.emj_1f324_fe0f, "\ud83c\udf24" );
		code( R.drawable.emj_1f324_fe0f, "\ud83c\udf24\ufe0f" );
		code( R.drawable.emj_1f325_fe0f, "\ud83c\udf25" );
		code( R.drawable.emj_1f325_fe0f, "\ud83c\udf25\ufe0f" );
		code( R.drawable.emj_1f326_fe0f, "\ud83c\udf26" );
		code( R.drawable.emj_1f326_fe0f, "\ud83c\udf26\ufe0f" );
		code( R.drawable.emj_1f327_fe0f, "\ud83c\udf27" );
		code( R.drawable.emj_1f327_fe0f, "\ud83c\udf27\ufe0f" );
		code( R.drawable.emj_1f328_fe0f, "\ud83c\udf28" );
		code( R.drawable.emj_1f328_fe0f, "\ud83c\udf28\ufe0f" );
		code( R.drawable.emj_1f329_fe0f, "\ud83c\udf29" );
		code( R.drawable.emj_1f329_fe0f, "\ud83c\udf29\ufe0f" );
		code( R.drawable.emj_1f32a_fe0f, "\ud83c\udf2a" );
		code( R.drawable.emj_1f32a_fe0f, "\ud83c\udf2a\ufe0f" );
		code( R.drawable.emj_1f32b_fe0f, "\ud83c\udf2b" );
		code( R.drawable.emj_1f32b_fe0f, "\ud83c\udf2b\ufe0f" );
		code( R.drawable.emj_1f32c_fe0f, "\ud83c\udf2c" );
		code( R.drawable.emj_1f32c_fe0f, "\ud83c\udf2c\ufe0f" );
		code( R.drawable.emj_1f32d, "\ud83c\udf2d" );
		code( R.drawable.emj_1f32e, "\ud83c\udf2e" );
		code( R.drawable.emj_1f32f, "\ud83c\udf2f" );
		code( R.drawable.emj_1f330, "\ud83c\udf30" );
		code( R.drawable.emj_1f330, "\ueb38" );
		code( R.drawable.emj_1f330, "\udbb8\udc4c" );
		code( R.drawable.emj_1f331, "\ud83c\udf31" );
		code( R.drawable.emj_1f331, "\ue746" );
		code( R.drawable.emj_1f331, "\ueb7d" );
		code( R.drawable.emj_1f331, "\udbb8\udc3e" );
		code( R.drawable.emj_1f332, "\ud83c\udf32" );
		code( R.drawable.emj_1f333, "\ud83c\udf33" );
		code( R.drawable.emj_1f334, "\ud83c\udf34" );
		code( R.drawable.emj_1f334, "\ue307" );
		code( R.drawable.emj_1f334, "\ue4e2" );
		code( R.drawable.emj_1f334, "\udbb8\udc47" );
		code( R.drawable.emj_1f335, "\ud83c\udf35" );
		code( R.drawable.emj_1f335, "\ue308" );
		code( R.drawable.emj_1f335, "\uea96" );
		code( R.drawable.emj_1f335, "\udbb8\udc48" );
		code( R.drawable.emj_1f336_fe0f, "\ud83c\udf36" );
		code( R.drawable.emj_1f336_fe0f, "\ud83c\udf36\ufe0f" );
		code( R.drawable.emj_1f337, "\ud83c\udf37" );
		code( R.drawable.emj_1f337, "\ue304" );
		code( R.drawable.emj_1f337, "\ue4e4" );
		code( R.drawable.emj_1f337, "\ue743" );
		code( R.drawable.emj_1f337, "\udbb8\udc3d" );
		code( R.drawable.emj_1f338, "\ud83c\udf38" );
		code( R.drawable.emj_1f338, "\ue030" );
		code( R.drawable.emj_1f338, "\ue4ca" );
		code( R.drawable.emj_1f338, "\ue748" );
		code( R.drawable.emj_1f338, "\udbb8\udc40" );
		code( R.drawable.emj_1f339, "\ud83c\udf39" );
		code( R.drawable.emj_1f339, "\ue032" );
		code( R.drawable.emj_1f339, "\ue5ba" );
		code( R.drawable.emj_1f339, "\udbb8\udc41" );
		code( R.drawable.emj_1f33a, "\ud83c\udf3a" );
		code( R.drawable.emj_1f33a, "\ue303" );
		code( R.drawable.emj_1f33a, "\uea94" );
		code( R.drawable.emj_1f33a, "\udbb8\udc45" );
		code( R.drawable.emj_1f33b, "\ud83c\udf3b" );
		code( R.drawable.emj_1f33b, "\ue305" );
		code( R.drawable.emj_1f33b, "\ue4e3" );
		code( R.drawable.emj_1f33b, "\udbb8\udc46" );
		code( R.drawable.emj_1f33c, "\ud83c\udf3c" );
		code( R.drawable.emj_1f33c, "\ueb49" );
		code( R.drawable.emj_1f33c, "\udbb8\udc4d" );
		code( R.drawable.emj_1f33d, "\ud83c\udf3d" );
		code( R.drawable.emj_1f33d, "\ueb36" );
		code( R.drawable.emj_1f33d, "\udbb8\udc4a" );
		code( R.drawable.emj_1f33e, "\ud83c\udf3e" );
		code( R.drawable.emj_1f33e, "\ue444" );
		code( R.drawable.emj_1f33e, "\udbb8\udc49" );
	}
	private static void init8(){
		code( R.drawable.emj_1f33f, "\ud83c\udf3f" );
		code( R.drawable.emj_1f33f, "\ueb82" );
		code( R.drawable.emj_1f33f, "\udbb8\udc4e" );
		code( R.drawable.emj_1f340, "\ud83c\udf40" );
		code( R.drawable.emj_1f340, "\ue110" );
		code( R.drawable.emj_1f340, "\ue513" );
		code( R.drawable.emj_1f340, "\ue741" );
		code( R.drawable.emj_1f340, "\udbb8\udc3c" );
		code( R.drawable.emj_1f341, "\ud83c\udf41" );
		code( R.drawable.emj_1f341, "\ue118" );
		code( R.drawable.emj_1f341, "\ue4ce" );
		code( R.drawable.emj_1f341, "\ue747" );
		code( R.drawable.emj_1f341, "\udbb8\udc3f" );
		code( R.drawable.emj_1f342, "\ud83c\udf42" );
		code( R.drawable.emj_1f342, "\ue119" );
		code( R.drawable.emj_1f342, "\ue5cd" );
		code( R.drawable.emj_1f342, "\udbb8\udc42" );
		code( R.drawable.emj_1f343, "\ud83c\udf43" );
		code( R.drawable.emj_1f343, "\ue447" );
		code( R.drawable.emj_1f343, "\ue5cd" );
		code( R.drawable.emj_1f343, "\udbb8\udc43" );
		code( R.drawable.emj_1f344, "\ud83c\udf44" );
		code( R.drawable.emj_1f344, "\ueb37" );
		code( R.drawable.emj_1f344, "\udbb8\udc4b" );
		code( R.drawable.emj_1f345, "\ud83c\udf45" );
		code( R.drawable.emj_1f345, "\ue349" );
		code( R.drawable.emj_1f345, "\ueabb" );
		code( R.drawable.emj_1f345, "\udbb8\udc55" );
		code( R.drawable.emj_1f346, "\ud83c\udf46" );
		code( R.drawable.emj_1f346, "\ue34a" );
		code( R.drawable.emj_1f346, "\ueabc" );
		code( R.drawable.emj_1f346, "\udbb8\udc56" );
		code( R.drawable.emj_1f347, "\ud83c\udf47" );
		code( R.drawable.emj_1f347, "\ueb34" );
		code( R.drawable.emj_1f347, "\udbb8\udc59" );
		code( R.drawable.emj_1f348, "\ud83c\udf48" );
		code( R.drawable.emj_1f348, "\ueb32" );
		code( R.drawable.emj_1f348, "\udbb8\udc57" );
		code( R.drawable.emj_1f349, "\ud83c\udf49" );
		code( R.drawable.emj_1f349, "\ue348" );
		code( R.drawable.emj_1f349, "\ue4cd" );
		code( R.drawable.emj_1f349, "\udbb8\udc54" );
		code( R.drawable.emj_1f34a, "\ud83c\udf4a" );
		code( R.drawable.emj_1f34a, "\ue346" );
		code( R.drawable.emj_1f34a, "\ueaba" );
		code( R.drawable.emj_1f34a, "\udbb8\udc52" );
		code( R.drawable.emj_1f34b, "\ud83c\udf4b" );
		code( R.drawable.emj_1f34c, "\ud83c\udf4c" );
		code( R.drawable.emj_1f34c, "\ue744" );
		code( R.drawable.emj_1f34c, "\ueb35" );
		code( R.drawable.emj_1f34c, "\udbb8\udc50" );
		code( R.drawable.emj_1f34d, "\ud83c\udf4d" );
		code( R.drawable.emj_1f34d, "\ueb33" );
		code( R.drawable.emj_1f34d, "\udbb8\udc58" );
		code( R.drawable.emj_1f34e, "\ud83c\udf4e" );
		code( R.drawable.emj_1f34e, "\ue345" );
		code( R.drawable.emj_1f34e, "\ue745" );
		code( R.drawable.emj_1f34e, "\ueab9" );
		code( R.drawable.emj_1f34e, "\udbb8\udc51" );
		code( R.drawable.emj_1f34f, "\ud83c\udf4f" );
		code( R.drawable.emj_1f34f, "\ueb5a" );
		code( R.drawable.emj_1f34f, "\udbb8\udc5b" );
		code( R.drawable.emj_1f350, "\ud83c\udf50" );
		code( R.drawable.emj_1f351, "\ud83c\udf51" );
		code( R.drawable.emj_1f351, "\ueb39" );
		code( R.drawable.emj_1f351, "\udbb8\udc5a" );
		code( R.drawable.emj_1f352, "\ud83c\udf52" );
		code( R.drawable.emj_1f352, "\ue4d2" );
		code( R.drawable.emj_1f352, "\ue742" );
		code( R.drawable.emj_1f352, "\udbb8\udc4f" );
		code( R.drawable.emj_1f353, "\ud83c\udf53" );
		code( R.drawable.emj_1f353, "\ue347" );
		code( R.drawable.emj_1f353, "\ue4d4" );
		code( R.drawable.emj_1f353, "\udbb8\udc53" );
		code( R.drawable.emj_1f354, "\ud83c\udf54" );
		code( R.drawable.emj_1f354, "\ue120" );
		code( R.drawable.emj_1f354, "\ue4d6" );
		code( R.drawable.emj_1f354, "\ue673" );
		code( R.drawable.emj_1f354, "\udbba\udd60" );
		code( R.drawable.emj_1f355, "\ud83c\udf55" );
		code( R.drawable.emj_1f355, "\ueb3b" );
		code( R.drawable.emj_1f355, "\udbba\udd75" );
		code( R.drawable.emj_1f356, "\ud83c\udf56" );
		code( R.drawable.emj_1f356, "\ue4c4" );
		code( R.drawable.emj_1f356, "\udbba\udd72" );
		code( R.drawable.emj_1f357, "\ud83c\udf57" );
		code( R.drawable.emj_1f357, "\ueb3c" );
		code( R.drawable.emj_1f357, "\udbba\udd76" );
		code( R.drawable.emj_1f358, "\ud83c\udf58" );
		code( R.drawable.emj_1f358, "\ue33d" );
		code( R.drawable.emj_1f358, "\ueab3" );
		code( R.drawable.emj_1f358, "\udbba\udd69" );
		code( R.drawable.emj_1f359, "\ud83c\udf59" );
		code( R.drawable.emj_1f359, "\ue342" );
		code( R.drawable.emj_1f359, "\ue4d5" );
		code( R.drawable.emj_1f359, "\ue749" );
		code( R.drawable.emj_1f359, "\udbba\udd61" );
		code( R.drawable.emj_1f35a, "\ud83c\udf5a" );
		code( R.drawable.emj_1f35a, "\ue33e" );
		code( R.drawable.emj_1f35a, "\ueab4" );
		code( R.drawable.emj_1f35a, "\udbba\udd6a" );
	}
	private static void init9(){
		code( R.drawable.emj_1f35b, "\ud83c\udf5b" );
		code( R.drawable.emj_1f35b, "\ue341" );
		code( R.drawable.emj_1f35b, "\ueab6" );
		code( R.drawable.emj_1f35b, "\udbba\udd6c" );
		code( R.drawable.emj_1f35c, "\ud83c\udf5c" );
		code( R.drawable.emj_1f35c, "\ue340" );
		code( R.drawable.emj_1f35c, "\ue5b4" );
		code( R.drawable.emj_1f35c, "\ue74c" );
		code( R.drawable.emj_1f35c, "\udbba\udd63" );
		code( R.drawable.emj_1f35d, "\ud83c\udf5d" );
		code( R.drawable.emj_1f35d, "\ue33f" );
		code( R.drawable.emj_1f35d, "\ueab5" );
		code( R.drawable.emj_1f35d, "\udbba\udd6b" );
		code( R.drawable.emj_1f35e, "\ud83c\udf5e" );
		code( R.drawable.emj_1f35e, "\ue339" );
		code( R.drawable.emj_1f35e, "\ue74d" );
		code( R.drawable.emj_1f35e, "\ueaaf" );
		code( R.drawable.emj_1f35e, "\udbba\udd64" );
		code( R.drawable.emj_1f35f, "\ud83c\udf5f" );
		code( R.drawable.emj_1f35f, "\ue33b" );
		code( R.drawable.emj_1f35f, "\ueab1" );
		code( R.drawable.emj_1f35f, "\udbba\udd67" );
		code( R.drawable.emj_1f360, "\ud83c\udf60" );
		code( R.drawable.emj_1f360, "\ueb3a" );
		code( R.drawable.emj_1f360, "\udbba\udd74" );
		code( R.drawable.emj_1f361, "\ud83c\udf61" );
		code( R.drawable.emj_1f361, "\ue33c" );
		code( R.drawable.emj_1f361, "\ueab2" );
		code( R.drawable.emj_1f361, "\udbba\udd68" );
		code( R.drawable.emj_1f362, "\ud83c\udf62" );
		code( R.drawable.emj_1f362, "\ue343" );
		code( R.drawable.emj_1f362, "\ueab7" );
		code( R.drawable.emj_1f362, "\udbba\udd6d" );
		code( R.drawable.emj_1f363, "\ud83c\udf63" );
		code( R.drawable.emj_1f363, "\ue344" );
		code( R.drawable.emj_1f363, "\ueab8" );
		code( R.drawable.emj_1f363, "\udbba\udd6e" );
		code( R.drawable.emj_1f364, "\ud83c\udf64" );
		code( R.drawable.emj_1f364, "\ueb70" );
		code( R.drawable.emj_1f364, "\udbba\udd7f" );
		code( R.drawable.emj_1f365, "\ud83c\udf65" );
		code( R.drawable.emj_1f365, "\ue4ed" );
		code( R.drawable.emj_1f365, "\ue643" );
		code( R.drawable.emj_1f365, "\udbba\udd73" );
		code( R.drawable.emj_1f366, "\ud83c\udf66" );
		code( R.drawable.emj_1f366, "\ue33a" );
		code( R.drawable.emj_1f366, "\ueab0" );
		code( R.drawable.emj_1f366, "\udbba\udd66" );
		code( R.drawable.emj_1f367, "\ud83c\udf67" );
		code( R.drawable.emj_1f367, "\ue43f" );
		code( R.drawable.emj_1f367, "\ueaea" );
		code( R.drawable.emj_1f367, "\udbba\udd71" );
		code( R.drawable.emj_1f368, "\ud83c\udf68" );
		code( R.drawable.emj_1f368, "\ueb4a" );
		code( R.drawable.emj_1f368, "\udbba\udd77" );
		code( R.drawable.emj_1f369, "\ud83c\udf69" );
		code( R.drawable.emj_1f369, "\ueb4b" );
		code( R.drawable.emj_1f369, "\udbba\udd78" );
		code( R.drawable.emj_1f36a, "\ud83c\udf6a" );
		code( R.drawable.emj_1f36a, "\ueb4c" );
		code( R.drawable.emj_1f36a, "\udbba\udd79" );
		code( R.drawable.emj_1f36b, "\ud83c\udf6b" );
		code( R.drawable.emj_1f36b, "\ueb4d" );
		code( R.drawable.emj_1f36b, "\udbba\udd7a" );
		code( R.drawable.emj_1f36c, "\ud83c\udf6c" );
		code( R.drawable.emj_1f36c, "\ueb4e" );
		code( R.drawable.emj_1f36c, "\udbba\udd7b" );
		code( R.drawable.emj_1f36d, "\ud83c\udf6d" );
		code( R.drawable.emj_1f36d, "\ueb4f" );
		code( R.drawable.emj_1f36d, "\udbba\udd7c" );
		code( R.drawable.emj_1f36e, "\ud83c\udf6e" );
		code( R.drawable.emj_1f36e, "\ueb56" );
		code( R.drawable.emj_1f36e, "\udbba\udd7d" );
		code( R.drawable.emj_1f36f, "\ud83c\udf6f" );
		code( R.drawable.emj_1f36f, "\ueb59" );
		code( R.drawable.emj_1f36f, "\udbba\udd7e" );
		code( R.drawable.emj_1f370, "\ud83c\udf70" );
		code( R.drawable.emj_1f370, "\ue046" );
		code( R.drawable.emj_1f370, "\ue4d0" );
		code( R.drawable.emj_1f370, "\ue74a" );
		code( R.drawable.emj_1f370, "\udbba\udd62" );
		code( R.drawable.emj_1f371, "\ud83c\udf71" );
		code( R.drawable.emj_1f371, "\ue34c" );
		code( R.drawable.emj_1f371, "\ueabd" );
		code( R.drawable.emj_1f371, "\udbba\udd6f" );
		code( R.drawable.emj_1f372, "\ud83c\udf72" );
		code( R.drawable.emj_1f372, "\ue34d" );
		code( R.drawable.emj_1f372, "\ueabe" );
		code( R.drawable.emj_1f372, "\udbba\udd70" );
		code( R.drawable.emj_1f373, "\ud83c\udf73" );
		code( R.drawable.emj_1f373, "\ue147" );
		code( R.drawable.emj_1f373, "\ue4d1" );
		code( R.drawable.emj_1f373, "\udbba\udd65" );
		code( R.drawable.emj_1f374, "\ud83c\udf74" );
		code( R.drawable.emj_1f374, "\ue043" );
		code( R.drawable.emj_1f374, "\ue4ac" );
		code( R.drawable.emj_1f374, "\ue66f" );
		code( R.drawable.emj_1f374, "\udbba\udd80" );
		code( R.drawable.emj_1f375, "\ud83c\udf75" );
		code( R.drawable.emj_1f375, "\ue338" );
		code( R.drawable.emj_1f375, "\ue71e" );
	}
	private static void init10(){
		code( R.drawable.emj_1f375, "\ueaae" );
		code( R.drawable.emj_1f375, "\udbba\udd84" );
		code( R.drawable.emj_1f376, "\ud83c\udf76" );
		code( R.drawable.emj_1f376, "\ue30b" );
		code( R.drawable.emj_1f376, "\ue74b" );
		code( R.drawable.emj_1f376, "\uea97" );
		code( R.drawable.emj_1f376, "\udbba\udd85" );
		code( R.drawable.emj_1f377, "\ud83c\udf77" );
		code( R.drawable.emj_1f377, "\ue4c1" );
		code( R.drawable.emj_1f377, "\ue756" );
		code( R.drawable.emj_1f377, "\udbba\udd86" );
		code( R.drawable.emj_1f378, "\ud83c\udf78" );
		code( R.drawable.emj_1f378, "\ue044" );
		code( R.drawable.emj_1f378, "\ue4c2" );
		code( R.drawable.emj_1f378, "\ue671" );
		code( R.drawable.emj_1f378, "\udbba\udd82" );
		code( R.drawable.emj_1f379, "\ud83c\udf79" );
		code( R.drawable.emj_1f379, "\ue671" );
		code( R.drawable.emj_1f379, "\ueb3e" );
		code( R.drawable.emj_1f379, "\udbba\udd88" );
		code( R.drawable.emj_1f37a, "\ud83c\udf7a" );
		code( R.drawable.emj_1f37a, "\ue047" );
		code( R.drawable.emj_1f37a, "\ue4c3" );
		code( R.drawable.emj_1f37a, "\ue672" );
		code( R.drawable.emj_1f37a, "\udbba\udd83" );
		code( R.drawable.emj_1f37b, "\ud83c\udf7b" );
		code( R.drawable.emj_1f37b, "\ue30c" );
		code( R.drawable.emj_1f37b, "\uea98" );
		code( R.drawable.emj_1f37b, "\udbba\udd87" );
		code( R.drawable.emj_1f37c, "\ud83c\udf7c" );
		code( R.drawable.emj_1f37d_fe0f, "\ud83c\udf7d" );
		code( R.drawable.emj_1f37d_fe0f, "\ud83c\udf7d\ufe0f" );
		code( R.drawable.emj_1f37e, "\ud83c\udf7e" );
		code( R.drawable.emj_1f37f, "\ud83c\udf7f" );
		code( R.drawable.emj_1f380, "\ud83c\udf80" );
		code( R.drawable.emj_1f380, "\ue314" );
		code( R.drawable.emj_1f380, "\ue59f" );
		code( R.drawable.emj_1f380, "\ue684" );
		code( R.drawable.emj_1f380, "\udbb9\udd0f" );
		code( R.drawable.emj_1f381, "\ud83c\udf81" );
		code( R.drawable.emj_1f381, "\ue112" );
		code( R.drawable.emj_1f381, "\ue4cf" );
		code( R.drawable.emj_1f381, "\ue685" );
		code( R.drawable.emj_1f381, "\udbb9\udd10" );
		code( R.drawable.emj_1f382, "\ud83c\udf82" );
		code( R.drawable.emj_1f382, "\ue34b" );
		code( R.drawable.emj_1f382, "\ue5a0" );
		code( R.drawable.emj_1f382, "\ue686" );
		code( R.drawable.emj_1f382, "\udbb9\udd11" );
		code( R.drawable.emj_1f383, "\ud83c\udf83" );
		code( R.drawable.emj_1f383, "\ue445" );
		code( R.drawable.emj_1f383, "\ueaee" );
		code( R.drawable.emj_1f383, "\udbb9\udd1f" );
		code( R.drawable.emj_1f384, "\ud83c\udf84" );
		code( R.drawable.emj_1f384, "\ue033" );
		code( R.drawable.emj_1f384, "\ue4c9" );
		code( R.drawable.emj_1f384, "\ue6a4" );
		code( R.drawable.emj_1f384, "\udbb9\udd12" );
		code( R.drawable.emj_1f385, "\ud83c\udf85" );
		code( R.drawable.emj_1f385, "\ue448" );
		code( R.drawable.emj_1f385, "\ueaf0" );
		code( R.drawable.emj_1f385, "\udbb9\udd13" );
		code( R.drawable.emj_1f385_1f3fb, "\ud83c\udf85\ud83c\udffb" );
		code( R.drawable.emj_1f385_1f3fc, "\ud83c\udf85\ud83c\udffc" );
		code( R.drawable.emj_1f385_1f3fd, "\ud83c\udf85\ud83c\udffd" );
		code( R.drawable.emj_1f385_1f3fe, "\ud83c\udf85\ud83c\udffe" );
		code( R.drawable.emj_1f385_1f3ff, "\ud83c\udf85\ud83c\udfff" );
		code( R.drawable.emj_1f386, "\ud83c\udf86" );
		code( R.drawable.emj_1f386, "\ue117" );
		code( R.drawable.emj_1f386, "\ue5cc" );
		code( R.drawable.emj_1f386, "\udbb9\udd15" );
		code( R.drawable.emj_1f387, "\ud83c\udf87" );
		code( R.drawable.emj_1f387, "\ue440" );
		code( R.drawable.emj_1f387, "\ueaeb" );
		code( R.drawable.emj_1f387, "\udbb9\udd1d" );
		code( R.drawable.emj_1f388, "\ud83c\udf88" );
		code( R.drawable.emj_1f388, "\ue310" );
		code( R.drawable.emj_1f388, "\uea9b" );
		code( R.drawable.emj_1f388, "\udbb9\udd16" );
		code( R.drawable.emj_1f389, "\ud83c\udf89" );
		code( R.drawable.emj_1f389, "\ue312" );
		code( R.drawable.emj_1f389, "\uea9c" );
		code( R.drawable.emj_1f389, "\udbb9\udd17" );
		code( R.drawable.emj_1f38a, "\ud83c\udf8a" );
		code( R.drawable.emj_1f38a, "\ue46f" );
		code( R.drawable.emj_1f38a, "\udbb9\udd20" );
		code( R.drawable.emj_1f38b, "\ud83c\udf8b" );
		code( R.drawable.emj_1f38b, "\ueb3d" );
		code( R.drawable.emj_1f38b, "\udbb9\udd21" );
		code( R.drawable.emj_1f38c, "\ud83c\udf8c" );
		code( R.drawable.emj_1f38c, "\ue143" );
		code( R.drawable.emj_1f38c, "\ue5d9" );
		code( R.drawable.emj_1f38c, "\udbb9\udd14" );
		code( R.drawable.emj_1f38d, "\ud83c\udf8d" );
		code( R.drawable.emj_1f38d, "\ue436" );
		code( R.drawable.emj_1f38d, "\ueae3" );
		code( R.drawable.emj_1f38d, "\udbb9\udd18" );
		code( R.drawable.emj_1f38e, "\ud83c\udf8e" );
		code( R.drawable.emj_1f38e, "\ue438" );
		code( R.drawable.emj_1f38e, "\ueae4" );
		code( R.drawable.emj_1f38e, "\udbb9\udd19" );
	}
	private static void init11(){
		code( R.drawable.emj_1f38f, "\ud83c\udf8f" );
		code( R.drawable.emj_1f38f, "\ue43b" );
		code( R.drawable.emj_1f38f, "\ueae7" );
		code( R.drawable.emj_1f38f, "\udbb9\udd1c" );
		code( R.drawable.emj_1f390, "\ud83c\udf90" );
		code( R.drawable.emj_1f390, "\ue442" );
		code( R.drawable.emj_1f390, "\ueaed" );
		code( R.drawable.emj_1f390, "\udbb9\udd1e" );
		code( R.drawable.emj_1f391, "\ud83c\udf91" );
		code( R.drawable.emj_1f391, "\ue446" );
		code( R.drawable.emj_1f391, "\ueaef" );
		code( R.drawable.emj_1f391, "\udbb8\udc17" );
		code( R.drawable.emj_1f392, "\ud83c\udf92" );
		code( R.drawable.emj_1f392, "\ue43a" );
		code( R.drawable.emj_1f392, "\ueae6" );
		code( R.drawable.emj_1f392, "\udbb9\udd1b" );
		code( R.drawable.emj_1f393, "\ud83c\udf93" );
		code( R.drawable.emj_1f393, "\ue439" );
		code( R.drawable.emj_1f393, "\ueae5" );
		code( R.drawable.emj_1f393, "\udbb9\udd1a" );
		code( R.drawable.emj_1f396_fe0f, "\ud83c\udf96" );
		code( R.drawable.emj_1f396_fe0f, "\ud83c\udf96\ufe0f" );
		code( R.drawable.emj_1f397_fe0f, "\ud83c\udf97" );
		code( R.drawable.emj_1f397_fe0f, "\ud83c\udf97\ufe0f" );
		code( R.drawable.emj_1f399_fe0f, "\ud83c\udf99" );
		code( R.drawable.emj_1f399_fe0f, "\ud83c\udf99\ufe0f" );
		code( R.drawable.emj_1f39a_fe0f, "\ud83c\udf9a" );
		code( R.drawable.emj_1f39a_fe0f, "\ud83c\udf9a\ufe0f" );
		code( R.drawable.emj_1f39b_fe0f, "\ud83c\udf9b" );
		code( R.drawable.emj_1f39b_fe0f, "\ud83c\udf9b\ufe0f" );
		code( R.drawable.emj_1f39e_fe0f, "\ud83c\udf9e" );
		code( R.drawable.emj_1f39e_fe0f, "\ud83c\udf9e\ufe0f" );
		code( R.drawable.emj_1f39f_fe0f, "\ud83c\udf9f" );
		code( R.drawable.emj_1f39f_fe0f, "\ud83c\udf9f\ufe0f" );
		code( R.drawable.emj_1f3a0, "\ud83c\udfa0" );
		code( R.drawable.emj_1f3a0, "\ue679" );
		code( R.drawable.emj_1f3a0, "\udbb9\udffc" );
		code( R.drawable.emj_1f3a1, "\ud83c\udfa1" );
		code( R.drawable.emj_1f3a1, "\ue124" );
		code( R.drawable.emj_1f3a1, "\ue46d" );
		code( R.drawable.emj_1f3a1, "\udbb9\udffd" );
		code( R.drawable.emj_1f3a2, "\ud83c\udfa2" );
		code( R.drawable.emj_1f3a2, "\ue433" );
		code( R.drawable.emj_1f3a2, "\ueae2" );
		code( R.drawable.emj_1f3a2, "\udbb9\udffe" );
		code( R.drawable.emj_1f3a3, "\ud83c\udfa3" );
		code( R.drawable.emj_1f3a3, "\ueb42" );
		code( R.drawable.emj_1f3a3, "\udbb9\udfff" );
		code( R.drawable.emj_1f3a4, "\ud83c\udfa4" );
		code( R.drawable.emj_1f3a4, "\ue03c" );
		code( R.drawable.emj_1f3a4, "\ue503" );
		code( R.drawable.emj_1f3a4, "\ue676" );
		code( R.drawable.emj_1f3a4, "\udbba\udc00" );
		code( R.drawable.emj_1f3a5, "\ud83c\udfa5" );
		code( R.drawable.emj_1f3a5, "\ue03d" );
		code( R.drawable.emj_1f3a5, "\ue517" );
		code( R.drawable.emj_1f3a5, "\ue677" );
		code( R.drawable.emj_1f3a5, "\udbba\udc01" );
		code( R.drawable.emj_1f3a6, "\ud83c\udfa6" );
		code( R.drawable.emj_1f3a6, "\ue507" );
		code( R.drawable.emj_1f3a6, "\ue517" );
		code( R.drawable.emj_1f3a6, "\udbba\udc02" );
		code( R.drawable.emj_1f3a7, "\ud83c\udfa7" );
		code( R.drawable.emj_1f3a7, "\ue30a" );
		code( R.drawable.emj_1f3a7, "\ue508" );
		code( R.drawable.emj_1f3a7, "\ue67a" );
		code( R.drawable.emj_1f3a7, "\udbba\udc03" );
		code( R.drawable.emj_1f3a8, "\ud83c\udfa8" );
		code( R.drawable.emj_1f3a8, "\ue502" );
		code( R.drawable.emj_1f3a8, "\ue59c" );
		code( R.drawable.emj_1f3a8, "\ue67b" );
		code( R.drawable.emj_1f3a8, "\udbba\udc04" );
		code( R.drawable.emj_1f3a9, "\ud83c\udfa9" );
		code( R.drawable.emj_1f3a9, "\ue503" );
		code( R.drawable.emj_1f3a9, "\ue67c" );
		code( R.drawable.emj_1f3a9, "\ueaf5" );
		code( R.drawable.emj_1f3a9, "\udbba\udc05" );
		code( R.drawable.emj_1f3aa, "\ud83c\udfaa" );
		code( R.drawable.emj_1f3aa, "\ue59e" );
		code( R.drawable.emj_1f3aa, "\ue67d" );
		code( R.drawable.emj_1f3aa, "\udbba\udc06" );
		code( R.drawable.emj_1f3ab, "\ud83c\udfab" );
		code( R.drawable.emj_1f3ab, "\ue125" );
		code( R.drawable.emj_1f3ab, "\ue49e" );
		code( R.drawable.emj_1f3ab, "\ue67e" );
		code( R.drawable.emj_1f3ab, "\udbba\udc07" );
		code( R.drawable.emj_1f3ac, "\ud83c\udfac" );
		code( R.drawable.emj_1f3ac, "\ue324" );
		code( R.drawable.emj_1f3ac, "\ue4be" );
		code( R.drawable.emj_1f3ac, "\ue6ac" );
		code( R.drawable.emj_1f3ac, "\udbba\udc08" );
		code( R.drawable.emj_1f3ad, "\ud83c\udfad" );
		code( R.drawable.emj_1f3ad, "\ue59d" );
		code( R.drawable.emj_1f3ad, "\udbba\udc09" );
		code( R.drawable.emj_1f3ae, "\ud83c\udfae" );
		code( R.drawable.emj_1f3ae, "\ue4c6" );
		code( R.drawable.emj_1f3ae, "\ue68b" );
		code( R.drawable.emj_1f3ae, "\udbba\udc0a" );
		code( R.drawable.emj_1f3af, "\ud83c\udfaf" );
		code( R.drawable.emj_1f3af, "\ue130" );
		code( R.drawable.emj_1f3af, "\ue4c5" );
	}
	private static void init12(){
		code( R.drawable.emj_1f3af, "\udbba\udc0c" );
		code( R.drawable.emj_1f3b0, "\ud83c\udfb0" );
		code( R.drawable.emj_1f3b0, "\ue133" );
		code( R.drawable.emj_1f3b0, "\ue46e" );
		code( R.drawable.emj_1f3b0, "\udbba\udc0d" );
		code( R.drawable.emj_1f3b1, "\ud83c\udfb1" );
		code( R.drawable.emj_1f3b1, "\ue42c" );
		code( R.drawable.emj_1f3b1, "\ueadd" );
		code( R.drawable.emj_1f3b1, "\udbba\udc0e" );
		code( R.drawable.emj_1f3b2, "\ud83c\udfb2" );
		code( R.drawable.emj_1f3b2, "\ue4c8" );
		code( R.drawable.emj_1f3b2, "\udbba\udc0f" );
		code( R.drawable.emj_1f3b3, "\ud83c\udfb3" );
		code( R.drawable.emj_1f3b3, "\ueb43" );
		code( R.drawable.emj_1f3b3, "\udbba\udc10" );
		code( R.drawable.emj_1f3b4, "\ud83c\udfb4" );
		code( R.drawable.emj_1f3b4, "\ueb6e" );
		code( R.drawable.emj_1f3b4, "\udbba\udc11" );
		code( R.drawable.emj_1f3b5, "\ud83c\udfb5" );
		code( R.drawable.emj_1f3b5, "\ue03e" );
		code( R.drawable.emj_1f3b5, "\ue5be" );
		code( R.drawable.emj_1f3b5, "\ue6f6" );
		code( R.drawable.emj_1f3b5, "\udbba\udc13" );
		code( R.drawable.emj_1f3b6, "\ud83c\udfb6" );
		code( R.drawable.emj_1f3b6, "\ue326" );
		code( R.drawable.emj_1f3b6, "\ue505" );
		code( R.drawable.emj_1f3b6, "\ue6ff" );
		code( R.drawable.emj_1f3b6, "\udbba\udc14" );
		code( R.drawable.emj_1f3b7, "\ud83c\udfb7" );
		code( R.drawable.emj_1f3b7, "\ue040" );
		code( R.drawable.emj_1f3b7, "\udbba\udc15" );
		code( R.drawable.emj_1f3b8, "\ud83c\udfb8" );
		code( R.drawable.emj_1f3b8, "\ue041" );
		code( R.drawable.emj_1f3b8, "\ue506" );
		code( R.drawable.emj_1f3b8, "\udbba\udc16" );
		code( R.drawable.emj_1f3b9, "\ud83c\udfb9" );
		code( R.drawable.emj_1f3b9, "\ueb40" );
		code( R.drawable.emj_1f3b9, "\udbba\udc17" );
		code( R.drawable.emj_1f3ba, "\ud83c\udfba" );
		code( R.drawable.emj_1f3ba, "\ue042" );
		code( R.drawable.emj_1f3ba, "\ueadc" );
		code( R.drawable.emj_1f3ba, "\udbba\udc18" );
		code( R.drawable.emj_1f3bb, "\ud83c\udfbb" );
		code( R.drawable.emj_1f3bb, "\ue507" );
		code( R.drawable.emj_1f3bb, "\udbba\udc19" );
		code( R.drawable.emj_1f3bc, "\ud83c\udfbc" );
		code( R.drawable.emj_1f3bc, "\ueacc" );
		code( R.drawable.emj_1f3bc, "\udbba\udc1a" );
		code( R.drawable.emj_1f3bd, "\ud83c\udfbd" );
		code( R.drawable.emj_1f3bd, "\ue652" );
		code( R.drawable.emj_1f3bd, "\udbb9\udfd0" );
		code( R.drawable.emj_1f3be, "\ud83c\udfbe" );
		code( R.drawable.emj_1f3be, "\ue015" );
		code( R.drawable.emj_1f3be, "\ue4b7" );
		code( R.drawable.emj_1f3be, "\ue655" );
		code( R.drawable.emj_1f3be, "\udbb9\udfd3" );
		code( R.drawable.emj_1f3bf, "\ud83c\udfbf" );
		code( R.drawable.emj_1f3bf, "\ue013" );
		code( R.drawable.emj_1f3bf, "\ue657" );
		code( R.drawable.emj_1f3bf, "\ueaac" );
		code( R.drawable.emj_1f3bf, "\udbb9\udfd5" );
		code( R.drawable.emj_1f3c0, "\ud83c\udfc0" );
		code( R.drawable.emj_1f3c0, "\ue42a" );
		code( R.drawable.emj_1f3c0, "\ue59a" );
		code( R.drawable.emj_1f3c0, "\ue658" );
		code( R.drawable.emj_1f3c0, "\udbb9\udfd6" );
		code( R.drawable.emj_1f3c1, "\ud83c\udfc1" );
		code( R.drawable.emj_1f3c1, "\ue132" );
		code( R.drawable.emj_1f3c1, "\ue4b9" );
		code( R.drawable.emj_1f3c1, "\ue659" );
		code( R.drawable.emj_1f3c1, "\udbb9\udfd7" );
		code( R.drawable.emj_1f3c2, "\ud83c\udfc2" );
		code( R.drawable.emj_1f3c2, "\ue4b8" );
		code( R.drawable.emj_1f3c2, "\ue712" );
		code( R.drawable.emj_1f3c2, "\udbb9\udfd8" );
		code( R.drawable.emj_1f3c2_1f3fb, "\ud83c\udfc2\ud83c\udffb" );
		code( R.drawable.emj_1f3c2_1f3fc, "\ud83c\udfc2\ud83c\udffc" );
		code( R.drawable.emj_1f3c2_1f3fd, "\ud83c\udfc2\ud83c\udffd" );
		code( R.drawable.emj_1f3c2_1f3fe, "\ud83c\udfc2\ud83c\udffe" );
		code( R.drawable.emj_1f3c2_1f3ff, "\ud83c\udfc2\ud83c\udfff" );
		code( R.drawable.emj_1f3c3, "\ud83c\udfc3" );
		code( R.drawable.emj_1f3c3, "\ue115" );
		code( R.drawable.emj_1f3c3, "\ue46b" );
		code( R.drawable.emj_1f3c3, "\ue733" );
		code( R.drawable.emj_1f3c3, "\udbb9\udfd9" );
		code( R.drawable.emj_1f3c3_1f3fb, "\ud83c\udfc3\ud83c\udffb" );
		code( R.drawable.emj_1f3c3_1f3fb_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fb_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f3c3_1f3fb_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fb_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f3c3_1f3fc, "\ud83c\udfc3\ud83c\udffc" );
		code( R.drawable.emj_1f3c3_1f3fc_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fc_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f3c3_1f3fc_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fc_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f3c3_1f3fd, "\ud83c\udfc3\ud83c\udffd" );
		code( R.drawable.emj_1f3c3_1f3fd_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fd_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f3c3_1f3fd_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fd_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f3c3_1f3fe, "\ud83c\udfc3\ud83c\udffe" );
	}
	private static void init13(){
		code( R.drawable.emj_1f3c3_1f3fe_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fe_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f3c3_1f3fe_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3fe_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f3c3_1f3ff, "\ud83c\udfc3\ud83c\udfff" );
		code( R.drawable.emj_1f3c3_1f3ff_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3ff_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f3c3_1f3ff_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c3_1f3ff_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f3c3_200d_2640_fe0f, "\ud83c\udfc3\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c3_200d_2640_fe0f, "\ud83c\udfc3\u2640" );
		code( R.drawable.emj_1f3c3_200d_2642_fe0f, "\ud83c\udfc3\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c3_200d_2642_fe0f, "\ud83c\udfc3\u2642" );
		code( R.drawable.emj_1f3c4, "\ud83c\udfc4" );
		code( R.drawable.emj_1f3c4, "\ue017" );
		code( R.drawable.emj_1f3c4, "\ueb41" );
		code( R.drawable.emj_1f3c4, "\udbb9\udfda" );
		code( R.drawable.emj_1f3c4_1f3fb, "\ud83c\udfc4\ud83c\udffb" );
		code( R.drawable.emj_1f3c4_1f3fb_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fb_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f3c4_1f3fb_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fb_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f3c4_1f3fc, "\ud83c\udfc4\ud83c\udffc" );
		code( R.drawable.emj_1f3c4_1f3fc_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fc_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f3c4_1f3fc_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fc_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f3c4_1f3fd, "\ud83c\udfc4\ud83c\udffd" );
		code( R.drawable.emj_1f3c4_1f3fd_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fd_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f3c4_1f3fd_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fd_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f3c4_1f3fe, "\ud83c\udfc4\ud83c\udffe" );
		code( R.drawable.emj_1f3c4_1f3fe_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fe_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f3c4_1f3fe_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3fe_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f3c4_1f3ff, "\ud83c\udfc4\ud83c\udfff" );
		code( R.drawable.emj_1f3c4_1f3ff_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3ff_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f3c4_1f3ff_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c4_1f3ff_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f3c4_200d_2640_fe0f, "\ud83c\udfc4\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3c4_200d_2640_fe0f, "\ud83c\udfc4\u2640" );
		code( R.drawable.emj_1f3c4_200d_2642_fe0f, "\ud83c\udfc4\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3c4_200d_2642_fe0f, "\ud83c\udfc4\u2642" );
		code( R.drawable.emj_1f3c5, "\ud83c\udfc5" );
		code( R.drawable.emj_1f3c6, "\ud83c\udfc6" );
		code( R.drawable.emj_1f3c6, "\ue131" );
		code( R.drawable.emj_1f3c6, "\ue5d3" );
		code( R.drawable.emj_1f3c6, "\udbb9\udfdb" );
		code( R.drawable.emj_1f3c7, "\ud83c\udfc7" );
		code( R.drawable.emj_1f3c7_1f3fb, "\ud83c\udfc7\ud83c\udffb" );
		code( R.drawable.emj_1f3c7_1f3fc, "\ud83c\udfc7\ud83c\udffc" );
		code( R.drawable.emj_1f3c7_1f3fd, "\ud83c\udfc7\ud83c\udffd" );
		code( R.drawable.emj_1f3c7_1f3fe, "\ud83c\udfc7\ud83c\udffe" );
		code( R.drawable.emj_1f3c7_1f3ff, "\ud83c\udfc7\ud83c\udfff" );
		code( R.drawable.emj_1f3c8, "\ud83c\udfc8" );
		code( R.drawable.emj_1f3c8, "\ue42b" );
		code( R.drawable.emj_1f3c8, "\ue4bb" );
		code( R.drawable.emj_1f3c8, "\udbb9\udfdd" );
		code( R.drawable.emj_1f3c9, "\ud83c\udfc9" );
		code( R.drawable.emj_1f3ca, "\ud83c\udfca" );
		code( R.drawable.emj_1f3ca, "\ue42d" );
		code( R.drawable.emj_1f3ca, "\ueade" );
		code( R.drawable.emj_1f3ca, "\udbb9\udfde" );
		code( R.drawable.emj_1f3ca_1f3fb, "\ud83c\udfca\ud83c\udffb" );
		code( R.drawable.emj_1f3ca_1f3fb_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fb_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f3ca_1f3fb_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fb_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f3ca_1f3fc, "\ud83c\udfca\ud83c\udffc" );
		code( R.drawable.emj_1f3ca_1f3fc_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fc_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f3ca_1f3fc_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fc_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f3ca_1f3fd, "\ud83c\udfca\ud83c\udffd" );
		code( R.drawable.emj_1f3ca_1f3fd_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fd_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f3ca_1f3fd_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fd_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f3ca_1f3fe, "\ud83c\udfca\ud83c\udffe" );
		code( R.drawable.emj_1f3ca_1f3fe_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fe_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f3ca_1f3fe_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3fe_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f3ca_1f3ff, "\ud83c\udfca\ud83c\udfff" );
		code( R.drawable.emj_1f3ca_1f3ff_200d_2640_fe0f, "\ud83c\udfca\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3ff_200d_2640_fe0f, "\ud83c\udfca\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f3ca_1f3ff_200d_2642_fe0f, "\ud83c\udfca\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3ca_1f3ff_200d_2642_fe0f, "\ud83c\udfca\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f3ca_200d_2640_fe0f, "\ud83c\udfca\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3ca_200d_2640_fe0f, "\ud83c\udfca\u2640" );
		code( R.drawable.emj_1f3ca_200d_2642_fe0f, "\ud83c\udfca\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3ca_200d_2642_fe0f, "\ud83c\udfca\u2642" );
		code( R.drawable.emj_1f3cb_1f3fb, "\ud83c\udfcb\ud83c\udffb" );
		code( R.drawable.emj_1f3cb_1f3fb_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fb_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f3cb_1f3fb_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fb_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f3cb_1f3fc, "\ud83c\udfcb\ud83c\udffc" );
	}
	private static void init14(){
		code( R.drawable.emj_1f3cb_1f3fc_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fc_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f3cb_1f3fc_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fc_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f3cb_1f3fd, "\ud83c\udfcb\ud83c\udffd" );
		code( R.drawable.emj_1f3cb_1f3fd_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fd_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f3cb_1f3fd_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fd_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f3cb_1f3fe, "\ud83c\udfcb\ud83c\udffe" );
		code( R.drawable.emj_1f3cb_1f3fe_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fe_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f3cb_1f3fe_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3fe_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f3cb_1f3ff, "\ud83c\udfcb\ud83c\udfff" );
		code( R.drawable.emj_1f3cb_1f3ff_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3ff_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f3cb_1f3ff_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cb_1f3ff_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f3cb_fe0f, "\ud83c\udfcb" );
		code( R.drawable.emj_1f3cb_fe0f, "\ud83c\udfcb\ufe0f" );
		code( R.drawable.emj_1f3cb_fe0f_200d_2640_fe0f, "\ud83c\udfcb\u2640" );
		code( R.drawable.emj_1f3cb_fe0f_200d_2640_fe0f, "\ud83c\udfcb\ufe0f\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cb_fe0f_200d_2642_fe0f, "\ud83c\udfcb\u2642" );
		code( R.drawable.emj_1f3cb_fe0f_200d_2642_fe0f, "\ud83c\udfcb\ufe0f\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fb, "\ud83c\udfcc\ud83c\udffb" );
		code( R.drawable.emj_1f3cc_1f3fb_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fb_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f3cc_1f3fb_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fb_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f3cc_1f3fc, "\ud83c\udfcc\ud83c\udffc" );
		code( R.drawable.emj_1f3cc_1f3fc_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fc_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f3cc_1f3fc_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fc_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f3cc_1f3fd, "\ud83c\udfcc\ud83c\udffd" );
		code( R.drawable.emj_1f3cc_1f3fd_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fd_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f3cc_1f3fd_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fd_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f3cc_1f3fe, "\ud83c\udfcc\ud83c\udffe" );
		code( R.drawable.emj_1f3cc_1f3fe_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fe_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f3cc_1f3fe_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3fe_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f3cc_1f3ff, "\ud83c\udfcc\ud83c\udfff" );
		code( R.drawable.emj_1f3cc_1f3ff_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3ff_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f3cc_1f3ff_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cc_1f3ff_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f3cc_fe0f, "\ud83c\udfcc" );
		code( R.drawable.emj_1f3cc_fe0f, "\ud83c\udfcc\ufe0f" );
		code( R.drawable.emj_1f3cc_fe0f_200d_2640_fe0f, "\ud83c\udfcc\u2640" );
		code( R.drawable.emj_1f3cc_fe0f_200d_2640_fe0f, "\ud83c\udfcc\ufe0f\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f3cc_fe0f_200d_2642_fe0f, "\ud83c\udfcc\u2642" );
		code( R.drawable.emj_1f3cc_fe0f_200d_2642_fe0f, "\ud83c\udfcc\ufe0f\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f3cd_fe0f, "\ud83c\udfcd" );
		code( R.drawable.emj_1f3cd_fe0f, "\ud83c\udfcd\ufe0f" );
		code( R.drawable.emj_1f3ce_fe0f, "\ud83c\udfce" );
		code( R.drawable.emj_1f3ce_fe0f, "\ud83c\udfce\ufe0f" );
		code( R.drawable.emj_1f3cf, "\ud83c\udfcf" );
		code( R.drawable.emj_1f3d0, "\ud83c\udfd0" );
		code( R.drawable.emj_1f3d1, "\ud83c\udfd1" );
		code( R.drawable.emj_1f3d2, "\ud83c\udfd2" );
		code( R.drawable.emj_1f3d3, "\ud83c\udfd3" );
		code( R.drawable.emj_1f3d4_fe0f, "\ud83c\udfd4" );
		code( R.drawable.emj_1f3d4_fe0f, "\ud83c\udfd4\ufe0f" );
		code( R.drawable.emj_1f3d5_fe0f, "\ud83c\udfd5" );
		code( R.drawable.emj_1f3d5_fe0f, "\ud83c\udfd5\ufe0f" );
		code( R.drawable.emj_1f3d6_fe0f, "\ud83c\udfd6" );
		code( R.drawable.emj_1f3d6_fe0f, "\ud83c\udfd6\ufe0f" );
		code( R.drawable.emj_1f3d7_fe0f, "\ud83c\udfd7" );
		code( R.drawable.emj_1f3d7_fe0f, "\ud83c\udfd7\ufe0f" );
		code( R.drawable.emj_1f3d8_fe0f, "\ud83c\udfd8" );
		code( R.drawable.emj_1f3d8_fe0f, "\ud83c\udfd8\ufe0f" );
		code( R.drawable.emj_1f3d9_fe0f, "\ud83c\udfd9" );
		code( R.drawable.emj_1f3d9_fe0f, "\ud83c\udfd9\ufe0f" );
		code( R.drawable.emj_1f3da_fe0f, "\ud83c\udfda" );
		code( R.drawable.emj_1f3da_fe0f, "\ud83c\udfda\ufe0f" );
		code( R.drawable.emj_1f3db_fe0f, "\ud83c\udfdb" );
		code( R.drawable.emj_1f3db_fe0f, "\ud83c\udfdb\ufe0f" );
		code( R.drawable.emj_1f3dc_fe0f, "\ud83c\udfdc" );
		code( R.drawable.emj_1f3dc_fe0f, "\ud83c\udfdc\ufe0f" );
		code( R.drawable.emj_1f3dd_fe0f, "\ud83c\udfdd" );
		code( R.drawable.emj_1f3dd_fe0f, "\ud83c\udfdd\ufe0f" );
		code( R.drawable.emj_1f3de_fe0f, "\ud83c\udfde" );
		code( R.drawable.emj_1f3de_fe0f, "\ud83c\udfde\ufe0f" );
		code( R.drawable.emj_1f3df_fe0f, "\ud83c\udfdf" );
		code( R.drawable.emj_1f3df_fe0f, "\ud83c\udfdf\ufe0f" );
		code( R.drawable.emj_1f3e0, "\ud83c\udfe0" );
		code( R.drawable.emj_1f3e0, "\ue036" );
		code( R.drawable.emj_1f3e0, "\ue4ab" );
		code( R.drawable.emj_1f3e0, "\ue663" );
		code( R.drawable.emj_1f3e0, "\udbb9\udcb0" );
		code( R.drawable.emj_1f3e1, "\ud83c\udfe1" );
		code( R.drawable.emj_1f3e1, "\ue663" );
		code( R.drawable.emj_1f3e1, "\ueb09" );
		code( R.drawable.emj_1f3e1, "\udbb9\udcb1" );
		code( R.drawable.emj_1f3e2, "\ud83c\udfe2" );
		code( R.drawable.emj_1f3e2, "\ue038" );
		code( R.drawable.emj_1f3e2, "\ue4ad" );
	}
	private static void init15(){
		code( R.drawable.emj_1f3e2, "\ue664" );
		code( R.drawable.emj_1f3e2, "\udbb9\udcb2" );
		code( R.drawable.emj_1f3e3, "\ud83c\udfe3" );
		code( R.drawable.emj_1f3e3, "\ue153" );
		code( R.drawable.emj_1f3e3, "\ue5de" );
		code( R.drawable.emj_1f3e3, "\ue665" );
		code( R.drawable.emj_1f3e3, "\udbb9\udcb3" );
		code( R.drawable.emj_1f3e4, "\ud83c\udfe4" );
		code( R.drawable.emj_1f3e5, "\ud83c\udfe5" );
		code( R.drawable.emj_1f3e5, "\ue155" );
		code( R.drawable.emj_1f3e5, "\ue5df" );
		code( R.drawable.emj_1f3e5, "\ue666" );
		code( R.drawable.emj_1f3e5, "\udbb9\udcb4" );
		code( R.drawable.emj_1f3e6, "\ud83c\udfe6" );
		code( R.drawable.emj_1f3e6, "\ue14d" );
		code( R.drawable.emj_1f3e6, "\ue4aa" );
		code( R.drawable.emj_1f3e6, "\ue667" );
		code( R.drawable.emj_1f3e6, "\udbb9\udcb5" );
		code( R.drawable.emj_1f3e7, "\ud83c\udfe7" );
		code( R.drawable.emj_1f3e7, "\ue154" );
		code( R.drawable.emj_1f3e7, "\ue4a3" );
		code( R.drawable.emj_1f3e7, "\ue668" );
		code( R.drawable.emj_1f3e7, "\udbb9\udcb6" );
		code( R.drawable.emj_1f3e8, "\ud83c\udfe8" );
		code( R.drawable.emj_1f3e8, "\ue158" );
		code( R.drawable.emj_1f3e8, "\ue669" );
		code( R.drawable.emj_1f3e8, "\uea81" );
		code( R.drawable.emj_1f3e8, "\udbb9\udcb7" );
		code( R.drawable.emj_1f3e9, "\ud83c\udfe9" );
		code( R.drawable.emj_1f3e9, "\ue501" );
		code( R.drawable.emj_1f3e9, "\ue669\ue6ef" );
		code( R.drawable.emj_1f3e9, "\ueaf3" );
		code( R.drawable.emj_1f3e9, "\udbb9\udcb8" );
		code( R.drawable.emj_1f3ea, "\ud83c\udfea" );
		code( R.drawable.emj_1f3ea, "\ue156" );
		code( R.drawable.emj_1f3ea, "\ue4a4" );
		code( R.drawable.emj_1f3ea, "\ue66a" );
		code( R.drawable.emj_1f3ea, "\udbb9\udcb9" );
		code( R.drawable.emj_1f3eb, "\ud83c\udfeb" );
		code( R.drawable.emj_1f3eb, "\ue157" );
		code( R.drawable.emj_1f3eb, "\ue73e" );
		code( R.drawable.emj_1f3eb, "\uea80" );
		code( R.drawable.emj_1f3eb, "\udbb9\udcba" );
		code( R.drawable.emj_1f3ec, "\ud83c\udfec" );
		code( R.drawable.emj_1f3ec, "\ue504" );
		code( R.drawable.emj_1f3ec, "\ueaf6" );
		code( R.drawable.emj_1f3ec, "\udbb9\udcbd" );
		code( R.drawable.emj_1f3ed, "\ud83c\udfed" );
		code( R.drawable.emj_1f3ed, "\ue508" );
		code( R.drawable.emj_1f3ed, "\ueaf9" );
		code( R.drawable.emj_1f3ed, "\udbb9\udcc0" );
		code( R.drawable.emj_1f3ee, "\ud83c\udfee" );
		code( R.drawable.emj_1f3ee, "\ue4bd" );
		code( R.drawable.emj_1f3ee, "\udbb9\udcc2" );
		code( R.drawable.emj_1f3ef, "\ud83c\udfef" );
		code( R.drawable.emj_1f3ef, "\ue505" );
		code( R.drawable.emj_1f3ef, "\ueaf7" );
		code( R.drawable.emj_1f3ef, "\udbb9\udcbe" );
		code( R.drawable.emj_1f3f0, "\ud83c\udff0" );
		code( R.drawable.emj_1f3f0, "\ue506" );
		code( R.drawable.emj_1f3f0, "\ueaf8" );
		code( R.drawable.emj_1f3f0, "\udbb9\udcbf" );
		code( R.drawable.emj_1f3f3_fe0f, "\ud83c\udff3" );
		code( R.drawable.emj_1f3f3_fe0f, "\ud83c\udff3\ufe0f" );
		code( R.drawable.emj_1f3f3_fe0f_200d_1f308, "\ud83c\udff3\ud83c\udf08" );
		code( R.drawable.emj_1f3f3_fe0f_200d_1f308, "\ud83c\udff3\ufe0f\u200d\ud83c\udf08" );
		code( R.drawable.emj_1f3f4, "\ud83c\udff4" );
		code( R.drawable.emj_1f3f4_e0067_e0062_e0065_e006e_e0067_e007f, "\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc65\udb40\udc6e\udb40\udc67\udb40\udc7f" );
		code( R.drawable.emj_1f3f4_e0067_e0062_e0073_e0063_e0074_e007f, "\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc73\udb40\udc63\udb40\udc74\udb40\udc7f" );
		code( R.drawable.emj_1f3f4_e0067_e0062_e0077_e006c_e0073_e007f, "\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc77\udb40\udc6c\udb40\udc73\udb40\udc7f" );
		code( R.drawable.emj_1f3f5_fe0f, "\ud83c\udff5" );
		code( R.drawable.emj_1f3f5_fe0f, "\ud83c\udff5\ufe0f" );
		code( R.drawable.emj_1f3f7_fe0f, "\ud83c\udff7" );
		code( R.drawable.emj_1f3f7_fe0f, "\ud83c\udff7\ufe0f" );
		code( R.drawable.emj_1f3f8, "\ud83c\udff8" );
		code( R.drawable.emj_1f3f9, "\ud83c\udff9" );
		code( R.drawable.emj_1f3fa, "\ud83c\udffa" );
		code( R.drawable.emj_1f3fb, "\ud83c\udffb" );
		code( R.drawable.emj_1f3fc, "\ud83c\udffc" );
		code( R.drawable.emj_1f3fd, "\ud83c\udffd" );
		code( R.drawable.emj_1f3fe, "\ud83c\udffe" );
		code( R.drawable.emj_1f3ff, "\ud83c\udfff" );
		code( R.drawable.emj_1f400, "\ud83d\udc00" );
		code( R.drawable.emj_1f401, "\ud83d\udc01" );
		code( R.drawable.emj_1f402, "\ud83d\udc02" );
		code( R.drawable.emj_1f403, "\ud83d\udc03" );
		code( R.drawable.emj_1f404, "\ud83d\udc04" );
		code( R.drawable.emj_1f405, "\ud83d\udc05" );
		code( R.drawable.emj_1f406, "\ud83d\udc06" );
		code( R.drawable.emj_1f407, "\ud83d\udc07" );
		code( R.drawable.emj_1f408, "\ud83d\udc08" );
		code( R.drawable.emj_1f409, "\ud83d\udc09" );
		code( R.drawable.emj_1f40a, "\ud83d\udc0a" );
		code( R.drawable.emj_1f40b, "\ud83d\udc0b" );
		code( R.drawable.emj_1f40c, "\ud83d\udc0c" );
		code( R.drawable.emj_1f40c, "\ue74e" );
		code( R.drawable.emj_1f40c, "\ueb7e" );
		code( R.drawable.emj_1f40c, "\udbb8\uddb9" );
		code( R.drawable.emj_1f40d, "\ud83d\udc0d" );
		code( R.drawable.emj_1f40d, "\ue52d" );
		code( R.drawable.emj_1f40d, "\ueb22" );
	}
	private static void init16(){
		code( R.drawable.emj_1f40d, "\udbb8\uddd3" );
		code( R.drawable.emj_1f40e, "\ud83d\udc0e" );
		code( R.drawable.emj_1f40e, "\ue134" );
		code( R.drawable.emj_1f40e, "\ue4d8" );
		code( R.drawable.emj_1f40e, "\udbb9\udfdc" );
		code( R.drawable.emj_1f40f, "\ud83d\udc0f" );
		code( R.drawable.emj_1f410, "\ud83d\udc10" );
		code( R.drawable.emj_1f411, "\ud83d\udc11" );
		code( R.drawable.emj_1f411, "\ue48f" );
		code( R.drawable.emj_1f411, "\ue529" );
		code( R.drawable.emj_1f411, "\udbb8\uddcf" );
		code( R.drawable.emj_1f412, "\ud83d\udc12" );
		code( R.drawable.emj_1f412, "\ue4d9" );
		code( R.drawable.emj_1f412, "\ue528" );
		code( R.drawable.emj_1f412, "\udbb8\uddce" );
		code( R.drawable.emj_1f413, "\ud83d\udc13" );
		code( R.drawable.emj_1f414, "\ud83d\udc14" );
		code( R.drawable.emj_1f414, "\ue52e" );
		code( R.drawable.emj_1f414, "\ueb23" );
		code( R.drawable.emj_1f414, "\udbb8\uddd4" );
		code( R.drawable.emj_1f415, "\ud83d\udc15" );
		code( R.drawable.emj_1f416, "\ud83d\udc16" );
		code( R.drawable.emj_1f417, "\ud83d\udc17" );
		code( R.drawable.emj_1f417, "\ue52f" );
		code( R.drawable.emj_1f417, "\ueb24" );
		code( R.drawable.emj_1f417, "\udbb8\uddd5" );
		code( R.drawable.emj_1f418, "\ud83d\udc18" );
		code( R.drawable.emj_1f418, "\ue526" );
		code( R.drawable.emj_1f418, "\ueb1f" );
		code( R.drawable.emj_1f418, "\udbb8\uddcc" );
		code( R.drawable.emj_1f419, "\ud83d\udc19" );
		code( R.drawable.emj_1f419, "\ue10a" );
		code( R.drawable.emj_1f419, "\ue5c7" );
		code( R.drawable.emj_1f419, "\udbb8\uddc5" );
		code( R.drawable.emj_1f41a, "\ud83d\udc1a" );
		code( R.drawable.emj_1f41a, "\ue441" );
		code( R.drawable.emj_1f41a, "\ueaec" );
		code( R.drawable.emj_1f41a, "\udbb8\uddc6" );
		code( R.drawable.emj_1f41b, "\ud83d\udc1b" );
		code( R.drawable.emj_1f41b, "\ue525" );
		code( R.drawable.emj_1f41b, "\ueb1e" );
		code( R.drawable.emj_1f41b, "\udbb8\uddcb" );
		code( R.drawable.emj_1f41c, "\ud83d\udc1c" );
		code( R.drawable.emj_1f41c, "\ue4dd" );
		code( R.drawable.emj_1f41c, "\udbb8\uddda" );
		code( R.drawable.emj_1f41d, "\ud83d\udc1d" );
		code( R.drawable.emj_1f41d, "\ueb57" );
		code( R.drawable.emj_1f41d, "\udbb8\udde1" );
		code( R.drawable.emj_1f41e, "\ud83d\udc1e" );
		code( R.drawable.emj_1f41e, "\ueb58" );
		code( R.drawable.emj_1f41e, "\udbb8\udde2" );
		code( R.drawable.emj_1f41f, "\ud83d\udc1f" );
		code( R.drawable.emj_1f41f, "\ue019" );
		code( R.drawable.emj_1f41f, "\ue49a" );
		code( R.drawable.emj_1f41f, "\ue751" );
		code( R.drawable.emj_1f41f, "\udbb8\uddbd" );
		code( R.drawable.emj_1f420, "\ud83d\udc20" );
		code( R.drawable.emj_1f420, "\ue522" );
		code( R.drawable.emj_1f420, "\ueb1d" );
		code( R.drawable.emj_1f420, "\udbb8\uddc9" );
		code( R.drawable.emj_1f421, "\ud83d\udc21" );
		code( R.drawable.emj_1f421, "\ue4d3" );
		code( R.drawable.emj_1f421, "\udbb8\uddd9" );
		code( R.drawable.emj_1f422, "\ud83d\udc22" );
		code( R.drawable.emj_1f422, "\ue5d4" );
		code( R.drawable.emj_1f422, "\udbb8\udddc" );
		code( R.drawable.emj_1f423, "\ud83d\udc23" );
		code( R.drawable.emj_1f423, "\ue5db" );
		code( R.drawable.emj_1f423, "\udbb8\udddd" );
		code( R.drawable.emj_1f424, "\ud83d\udc24" );
		code( R.drawable.emj_1f424, "\ue4e0" );
		code( R.drawable.emj_1f424, "\ue523" );
		code( R.drawable.emj_1f424, "\ue74f" );
		code( R.drawable.emj_1f424, "\udbb8\uddba" );
		code( R.drawable.emj_1f425, "\ud83d\udc25" );
		code( R.drawable.emj_1f425, "\ueb76" );
		code( R.drawable.emj_1f425, "\udbb8\uddbb" );
		code( R.drawable.emj_1f426, "\ud83d\udc26" );
		code( R.drawable.emj_1f426, "\ue4e0" );
		code( R.drawable.emj_1f426, "\ue521" );
		code( R.drawable.emj_1f426, "\udbb8\uddc8" );
		code( R.drawable.emj_1f427, "\ud83d\udc27" );
		code( R.drawable.emj_1f427, "\ue055" );
		code( R.drawable.emj_1f427, "\ue4dc" );
		code( R.drawable.emj_1f427, "\ue750" );
		code( R.drawable.emj_1f427, "\udbb8\uddbc" );
		code( R.drawable.emj_1f428, "\ud83d\udc28" );
		code( R.drawable.emj_1f428, "\ue527" );
		code( R.drawable.emj_1f428, "\ueb20" );
		code( R.drawable.emj_1f428, "\udbb8\uddcd" );
		code( R.drawable.emj_1f429, "\ud83d\udc29" );
		code( R.drawable.emj_1f429, "\ue4df" );
		code( R.drawable.emj_1f429, "\udbb8\uddd8" );
		code( R.drawable.emj_1f42a, "\ud83d\udc2a" );
		code( R.drawable.emj_1f42b, "\ud83d\udc2b" );
		code( R.drawable.emj_1f42b, "\ue530" );
		code( R.drawable.emj_1f42b, "\ueb25" );
		code( R.drawable.emj_1f42b, "\udbb8\uddd6" );
		code( R.drawable.emj_1f42c, "\ud83d\udc2c" );
		code( R.drawable.emj_1f42c, "\ue520" );
		code( R.drawable.emj_1f42c, "\ueb1b" );
	}
	private static void init17(){
		code( R.drawable.emj_1f42c, "\udbb8\uddc7" );
		code( R.drawable.emj_1f42d, "\ud83d\udc2d" );
		code( R.drawable.emj_1f42d, "\ue053" );
		code( R.drawable.emj_1f42d, "\ue5c2" );
		code( R.drawable.emj_1f42d, "\udbb8\uddc2" );
		code( R.drawable.emj_1f42e, "\ud83d\udc2e" );
		code( R.drawable.emj_1f42e, "\ue52b" );
		code( R.drawable.emj_1f42e, "\ueb21" );
		code( R.drawable.emj_1f42e, "\udbb8\uddd1" );
		code( R.drawable.emj_1f42f, "\ud83d\udc2f" );
		code( R.drawable.emj_1f42f, "\ue050" );
		code( R.drawable.emj_1f42f, "\ue5c0" );
		code( R.drawable.emj_1f42f, "\udbb8\uddc0" );
		code( R.drawable.emj_1f430, "\ud83d\udc30" );
		code( R.drawable.emj_1f430, "\ue4d7" );
		code( R.drawable.emj_1f430, "\ue52c" );
		code( R.drawable.emj_1f430, "\udbb8\uddd2" );
		code( R.drawable.emj_1f431, "\ud83d\udc31" );
		code( R.drawable.emj_1f431, "\ue04f" );
		code( R.drawable.emj_1f431, "\ue4db" );
		code( R.drawable.emj_1f431, "\ue6a2" );
		code( R.drawable.emj_1f431, "\udbb8\uddb8" );
		code( R.drawable.emj_1f432, "\ud83d\udc32" );
		code( R.drawable.emj_1f432, "\ueb3f" );
		code( R.drawable.emj_1f432, "\udbb8\uddde" );
		code( R.drawable.emj_1f433, "\ud83d\udc33" );
		code( R.drawable.emj_1f433, "\ue054" );
		code( R.drawable.emj_1f433, "\ue470" );
		code( R.drawable.emj_1f433, "\udbb8\uddc3" );
		code( R.drawable.emj_1f434, "\ud83d\udc34" );
		code( R.drawable.emj_1f434, "\ue01a" );
		code( R.drawable.emj_1f434, "\ue4d8" );
		code( R.drawable.emj_1f434, "\ue754" );
		code( R.drawable.emj_1f434, "\udbb8\uddbe" );
		code( R.drawable.emj_1f435, "\ud83d\udc35" );
		code( R.drawable.emj_1f435, "\ue109" );
		code( R.drawable.emj_1f435, "\ue4d9" );
		code( R.drawable.emj_1f435, "\udbb8\uddc4" );
		code( R.drawable.emj_1f436, "\ud83d\udc36" );
		code( R.drawable.emj_1f436, "\ue052" );
		code( R.drawable.emj_1f436, "\ue4e1" );
		code( R.drawable.emj_1f436, "\ue6a1" );
		code( R.drawable.emj_1f436, "\udbb8\uddb7" );
		code( R.drawable.emj_1f437, "\ud83d\udc37" );
		code( R.drawable.emj_1f437, "\ue10b" );
		code( R.drawable.emj_1f437, "\ue4de" );
		code( R.drawable.emj_1f437, "\ue755" );
		code( R.drawable.emj_1f437, "\udbb8\uddbf" );
		code( R.drawable.emj_1f438, "\ud83d\udc38" );
		code( R.drawable.emj_1f438, "\ue4da" );
		code( R.drawable.emj_1f438, "\ue531" );
		code( R.drawable.emj_1f438, "\udbb8\uddd7" );
		code( R.drawable.emj_1f439, "\ud83d\udc39" );
		code( R.drawable.emj_1f439, "\ue524" );
		code( R.drawable.emj_1f439, "\udbb8\uddca" );
		code( R.drawable.emj_1f43a, "\ud83d\udc3a" );
		code( R.drawable.emj_1f43a, "\ue4e1" );
		code( R.drawable.emj_1f43a, "\ue52a" );
		code( R.drawable.emj_1f43a, "\udbb8\uddd0" );
		code( R.drawable.emj_1f43b, "\ud83d\udc3b" );
		code( R.drawable.emj_1f43b, "\ue051" );
		code( R.drawable.emj_1f43b, "\ue5c1" );
		code( R.drawable.emj_1f43b, "\udbb8\uddc1" );
		code( R.drawable.emj_1f43c, "\ud83d\udc3c" );
		code( R.drawable.emj_1f43c, "\ueb46" );
		code( R.drawable.emj_1f43c, "\udbb8\udddf" );
		code( R.drawable.emj_1f43d, "\ud83d\udc3d" );
		code( R.drawable.emj_1f43d, "\ueb48" );
		code( R.drawable.emj_1f43d, "\udbb8\udde0" );
		code( R.drawable.emj_1f43e, "\ud83d\udc3e" );
		code( R.drawable.emj_1f43e, "\ue4ee" );
		code( R.drawable.emj_1f43e, "\udbb8\udddb" );
		code( R.drawable.emj_1f43f_fe0f, "\ud83d\udc3f" );
		code( R.drawable.emj_1f43f_fe0f, "\ud83d\udc3f\ufe0f" );
		code( R.drawable.emj_1f440, "\ud83d\udc40" );
		code( R.drawable.emj_1f440, "\ue419" );
		code( R.drawable.emj_1f440, "\ue5a4" );
		code( R.drawable.emj_1f440, "\ue691" );
		code( R.drawable.emj_1f440, "\udbb8\udd90" );
		code( R.drawable.emj_1f441_fe0f, "\ud83d\udc41" );
		code( R.drawable.emj_1f441_fe0f, "\ud83d\udc41\ufe0f" );
		code( R.drawable.emj_1f441_fe0f_200d_1f5e8_fe0f, "\ud83d\udc41\ud83d\udde8" );
		code( R.drawable.emj_1f441_fe0f_200d_1f5e8_fe0f, "\ud83d\udc41\ufe0f\u200d\ud83d\udde8\ufe0f" );
		code( R.drawable.emj_1f442, "\ud83d\udc42" );
		code( R.drawable.emj_1f442, "\ue41b" );
		code( R.drawable.emj_1f442, "\ue5a5" );
		code( R.drawable.emj_1f442, "\ue692" );
		code( R.drawable.emj_1f442, "\udbb8\udd91" );
		code( R.drawable.emj_1f442_1f3fb, "\ud83d\udc42\ud83c\udffb" );
		code( R.drawable.emj_1f442_1f3fc, "\ud83d\udc42\ud83c\udffc" );
		code( R.drawable.emj_1f442_1f3fd, "\ud83d\udc42\ud83c\udffd" );
		code( R.drawable.emj_1f442_1f3fe, "\ud83d\udc42\ud83c\udffe" );
		code( R.drawable.emj_1f442_1f3ff, "\ud83d\udc42\ud83c\udfff" );
		code( R.drawable.emj_1f443, "\ud83d\udc43" );
		code( R.drawable.emj_1f443, "\ue41a" );
		code( R.drawable.emj_1f443, "\uead0" );
		code( R.drawable.emj_1f443, "\udbb8\udd92" );
		code( R.drawable.emj_1f443_1f3fb, "\ud83d\udc43\ud83c\udffb" );
		code( R.drawable.emj_1f443_1f3fc, "\ud83d\udc43\ud83c\udffc" );
		code( R.drawable.emj_1f443_1f3fd, "\ud83d\udc43\ud83c\udffd" );
		code( R.drawable.emj_1f443_1f3fe, "\ud83d\udc43\ud83c\udffe" );
	}
	private static void init18(){
		code( R.drawable.emj_1f443_1f3ff, "\ud83d\udc43\ud83c\udfff" );
		code( R.drawable.emj_1f444, "\ud83d\udc44" );
		code( R.drawable.emj_1f444, "\ue41c" );
		code( R.drawable.emj_1f444, "\uead1" );
		code( R.drawable.emj_1f444, "\udbb8\udd93" );
		code( R.drawable.emj_1f445, "\ud83d\udc45" );
		code( R.drawable.emj_1f445, "\ueb47" );
		code( R.drawable.emj_1f445, "\udbb8\udd94" );
		code( R.drawable.emj_1f446, "\ud83d\udc46" );
		code( R.drawable.emj_1f446, "\ue22e" );
		code( R.drawable.emj_1f446, "\uea8d" );
		code( R.drawable.emj_1f446, "\udbba\udf99" );
		code( R.drawable.emj_1f446_1f3fb, "\ud83d\udc46\ud83c\udffb" );
		code( R.drawable.emj_1f446_1f3fc, "\ud83d\udc46\ud83c\udffc" );
		code( R.drawable.emj_1f446_1f3fd, "\ud83d\udc46\ud83c\udffd" );
		code( R.drawable.emj_1f446_1f3fe, "\ud83d\udc46\ud83c\udffe" );
		code( R.drawable.emj_1f446_1f3ff, "\ud83d\udc46\ud83c\udfff" );
		code( R.drawable.emj_1f447, "\ud83d\udc47" );
		code( R.drawable.emj_1f447, "\ue22f" );
		code( R.drawable.emj_1f447, "\uea8e" );
		code( R.drawable.emj_1f447, "\udbba\udf9a" );
		code( R.drawable.emj_1f447_1f3fb, "\ud83d\udc47\ud83c\udffb" );
		code( R.drawable.emj_1f447_1f3fc, "\ud83d\udc47\ud83c\udffc" );
		code( R.drawable.emj_1f447_1f3fd, "\ud83d\udc47\ud83c\udffd" );
		code( R.drawable.emj_1f447_1f3fe, "\ud83d\udc47\ud83c\udffe" );
		code( R.drawable.emj_1f447_1f3ff, "\ud83d\udc47\ud83c\udfff" );
		code( R.drawable.emj_1f448, "\ud83d\udc48" );
		code( R.drawable.emj_1f448, "\ue230" );
		code( R.drawable.emj_1f448, "\ue4ff" );
		code( R.drawable.emj_1f448, "\udbba\udf9b" );
		code( R.drawable.emj_1f448_1f3fb, "\ud83d\udc48\ud83c\udffb" );
		code( R.drawable.emj_1f448_1f3fc, "\ud83d\udc48\ud83c\udffc" );
		code( R.drawable.emj_1f448_1f3fd, "\ud83d\udc48\ud83c\udffd" );
		code( R.drawable.emj_1f448_1f3fe, "\ud83d\udc48\ud83c\udffe" );
		code( R.drawable.emj_1f448_1f3ff, "\ud83d\udc48\ud83c\udfff" );
		code( R.drawable.emj_1f449, "\ud83d\udc49" );
		code( R.drawable.emj_1f449, "\ue231" );
		code( R.drawable.emj_1f449, "\ue500" );
		code( R.drawable.emj_1f449, "\udbba\udf9c" );
		code( R.drawable.emj_1f449_1f3fb, "\ud83d\udc49\ud83c\udffb" );
		code( R.drawable.emj_1f449_1f3fc, "\ud83d\udc49\ud83c\udffc" );
		code( R.drawable.emj_1f449_1f3fd, "\ud83d\udc49\ud83c\udffd" );
		code( R.drawable.emj_1f449_1f3fe, "\ud83d\udc49\ud83c\udffe" );
		code( R.drawable.emj_1f449_1f3ff, "\ud83d\udc49\ud83c\udfff" );
		code( R.drawable.emj_1f44a, "\ud83d\udc4a" );
		code( R.drawable.emj_1f44a, "\ue00d" );
		code( R.drawable.emj_1f44a, "\ue4f3" );
		code( R.drawable.emj_1f44a, "\ue6fd" );
		code( R.drawable.emj_1f44a, "\udbba\udf96" );
		code( R.drawable.emj_1f44a_1f3fb, "\ud83d\udc4a\ud83c\udffb" );
		code( R.drawable.emj_1f44a_1f3fc, "\ud83d\udc4a\ud83c\udffc" );
		code( R.drawable.emj_1f44a_1f3fd, "\ud83d\udc4a\ud83c\udffd" );
		code( R.drawable.emj_1f44a_1f3fe, "\ud83d\udc4a\ud83c\udffe" );
		code( R.drawable.emj_1f44a_1f3ff, "\ud83d\udc4a\ud83c\udfff" );
		code( R.drawable.emj_1f44b, "\ud83d\udc4b" );
		code( R.drawable.emj_1f44b, "\ue41e" );
		code( R.drawable.emj_1f44b, "\uead6" );
		code( R.drawable.emj_1f44b, "\udbba\udf9d" );
		code( R.drawable.emj_1f44b_1f3fb, "\ud83d\udc4b\ud83c\udffb" );
		code( R.drawable.emj_1f44b_1f3fc, "\ud83d\udc4b\ud83c\udffc" );
		code( R.drawable.emj_1f44b_1f3fd, "\ud83d\udc4b\ud83c\udffd" );
		code( R.drawable.emj_1f44b_1f3fe, "\ud83d\udc4b\ud83c\udffe" );
		code( R.drawable.emj_1f44b_1f3ff, "\ud83d\udc4b\ud83c\udfff" );
		code( R.drawable.emj_1f44c, "\ud83d\udc4c" );
		code( R.drawable.emj_1f44c, "\ue420" );
		code( R.drawable.emj_1f44c, "\uead4" );
		code( R.drawable.emj_1f44c, "\udbba\udf9f" );
		code( R.drawable.emj_1f44c_1f3fb, "\ud83d\udc4c\ud83c\udffb" );
		code( R.drawable.emj_1f44c_1f3fc, "\ud83d\udc4c\ud83c\udffc" );
		code( R.drawable.emj_1f44c_1f3fd, "\ud83d\udc4c\ud83c\udffd" );
		code( R.drawable.emj_1f44c_1f3fe, "\ud83d\udc4c\ud83c\udffe" );
		code( R.drawable.emj_1f44c_1f3ff, "\ud83d\udc4c\ud83c\udfff" );
		code( R.drawable.emj_1f44d, "\ud83d\udc4d" );
		code( R.drawable.emj_1f44d, "\ue00e" );
		code( R.drawable.emj_1f44d, "\ue4f9" );
		code( R.drawable.emj_1f44d, "\ue727" );
		code( R.drawable.emj_1f44d, "\udbba\udf97" );
		code( R.drawable.emj_1f44d_1f3fb, "\ud83d\udc4d\ud83c\udffb" );
		code( R.drawable.emj_1f44d_1f3fc, "\ud83d\udc4d\ud83c\udffc" );
		code( R.drawable.emj_1f44d_1f3fd, "\ud83d\udc4d\ud83c\udffd" );
		code( R.drawable.emj_1f44d_1f3fe, "\ud83d\udc4d\ud83c\udffe" );
		code( R.drawable.emj_1f44d_1f3ff, "\ud83d\udc4d\ud83c\udfff" );
		code( R.drawable.emj_1f44e, "\ud83d\udc4e" );
		code( R.drawable.emj_1f44e, "\ue421" );
		code( R.drawable.emj_1f44e, "\uead5" );
		code( R.drawable.emj_1f44e, "\udbba\udfa0" );
		code( R.drawable.emj_1f44e_1f3fb, "\ud83d\udc4e\ud83c\udffb" );
		code( R.drawable.emj_1f44e_1f3fc, "\ud83d\udc4e\ud83c\udffc" );
		code( R.drawable.emj_1f44e_1f3fd, "\ud83d\udc4e\ud83c\udffd" );
		code( R.drawable.emj_1f44e_1f3fe, "\ud83d\udc4e\ud83c\udffe" );
		code( R.drawable.emj_1f44e_1f3ff, "\ud83d\udc4e\ud83c\udfff" );
		code( R.drawable.emj_1f44f, "\ud83d\udc4f" );
		code( R.drawable.emj_1f44f, "\ue41f" );
		code( R.drawable.emj_1f44f, "\uead3" );
		code( R.drawable.emj_1f44f, "\udbba\udf9e" );
		code( R.drawable.emj_1f44f_1f3fb, "\ud83d\udc4f\ud83c\udffb" );
		code( R.drawable.emj_1f44f_1f3fc, "\ud83d\udc4f\ud83c\udffc" );
		code( R.drawable.emj_1f44f_1f3fd, "\ud83d\udc4f\ud83c\udffd" );
		code( R.drawable.emj_1f44f_1f3fe, "\ud83d\udc4f\ud83c\udffe" );
		code( R.drawable.emj_1f44f_1f3ff, "\ud83d\udc4f\ud83c\udfff" );
		code( R.drawable.emj_1f450, "\ud83d\udc50" );
	}
	private static void init19(){
		code( R.drawable.emj_1f450, "\ue422" );
		code( R.drawable.emj_1f450, "\udbba\udfa1" );
		code( R.drawable.emj_1f450_1f3fb, "\ud83d\udc50\ud83c\udffb" );
		code( R.drawable.emj_1f450_1f3fc, "\ud83d\udc50\ud83c\udffc" );
		code( R.drawable.emj_1f450_1f3fd, "\ud83d\udc50\ud83c\udffd" );
		code( R.drawable.emj_1f450_1f3fe, "\ud83d\udc50\ud83c\udffe" );
		code( R.drawable.emj_1f450_1f3ff, "\ud83d\udc50\ud83c\udfff" );
		code( R.drawable.emj_1f451, "\ud83d\udc51" );
		code( R.drawable.emj_1f451, "\ue10e" );
		code( R.drawable.emj_1f451, "\ue5c9" );
		code( R.drawable.emj_1f451, "\ue71a" );
		code( R.drawable.emj_1f451, "\udbb9\udcd1" );
		code( R.drawable.emj_1f452, "\ud83d\udc52" );
		code( R.drawable.emj_1f452, "\ue318" );
		code( R.drawable.emj_1f452, "\uea9e" );
		code( R.drawable.emj_1f452, "\udbb9\udcd4" );
		code( R.drawable.emj_1f453, "\ud83d\udc53" );
		code( R.drawable.emj_1f453, "\ue4fe" );
		code( R.drawable.emj_1f453, "\ue69a" );
		code( R.drawable.emj_1f453, "\udbb9\udcce" );
		code( R.drawable.emj_1f454, "\ud83d\udc54" );
		code( R.drawable.emj_1f454, "\ue302" );
		code( R.drawable.emj_1f454, "\uea93" );
		code( R.drawable.emj_1f454, "\udbb9\udcd3" );
		code( R.drawable.emj_1f455, "\ud83d\udc55" );
		code( R.drawable.emj_1f455, "\ue006" );
		code( R.drawable.emj_1f455, "\ue5b6" );
		code( R.drawable.emj_1f455, "\ue70e" );
		code( R.drawable.emj_1f455, "\udbb9\udccf" );
		code( R.drawable.emj_1f456, "\ud83d\udc56" );
		code( R.drawable.emj_1f456, "\ue711" );
		code( R.drawable.emj_1f456, "\ueb77" );
		code( R.drawable.emj_1f456, "\udbb9\udcd0" );
		code( R.drawable.emj_1f457, "\ud83d\udc57" );
		code( R.drawable.emj_1f457, "\ue319" );
		code( R.drawable.emj_1f457, "\ueb6b" );
		code( R.drawable.emj_1f457, "\udbb9\udcd5" );
		code( R.drawable.emj_1f458, "\ud83d\udc58" );
		code( R.drawable.emj_1f458, "\ue321" );
		code( R.drawable.emj_1f458, "\ueaa3" );
		code( R.drawable.emj_1f458, "\udbb9\udcd9" );
		code( R.drawable.emj_1f459, "\ud83d\udc59" );
		code( R.drawable.emj_1f459, "\ue322" );
		code( R.drawable.emj_1f459, "\ueaa4" );
		code( R.drawable.emj_1f459, "\udbb9\udcda" );
		code( R.drawable.emj_1f45a, "\ud83d\udc5a" );
		code( R.drawable.emj_1f45a, "\ue50d" );
		code( R.drawable.emj_1f45a, "\udbb9\udcdb" );
		code( R.drawable.emj_1f45b, "\ud83d\udc5b" );
		code( R.drawable.emj_1f45b, "\ue504" );
		code( R.drawable.emj_1f45b, "\ue70f" );
		code( R.drawable.emj_1f45b, "\udbb9\udcdc" );
		code( R.drawable.emj_1f45c, "\ud83d\udc5c" );
		code( R.drawable.emj_1f45c, "\ue323" );
		code( R.drawable.emj_1f45c, "\ue49c" );
		code( R.drawable.emj_1f45c, "\ue682" );
		code( R.drawable.emj_1f45c, "\udbb9\udcf0" );
		code( R.drawable.emj_1f45d, "\ud83d\udc5d" );
		code( R.drawable.emj_1f45d, "\ue6ad" );
		code( R.drawable.emj_1f45d, "\udbb9\udcf1" );
		code( R.drawable.emj_1f45e, "\ud83d\udc5e" );
		code( R.drawable.emj_1f45e, "\ue5b7" );
		code( R.drawable.emj_1f45e, "\udbb9\udccc" );
		code( R.drawable.emj_1f45f, "\ud83d\udc5f" );
		code( R.drawable.emj_1f45f, "\ue007" );
		code( R.drawable.emj_1f45f, "\ue699" );
		code( R.drawable.emj_1f45f, "\ueb2b" );
		code( R.drawable.emj_1f45f, "\udbb9\udccd" );
		code( R.drawable.emj_1f460, "\ud83d\udc60" );
		code( R.drawable.emj_1f460, "\ue13e" );
		code( R.drawable.emj_1f460, "\ue51a" );
		code( R.drawable.emj_1f460, "\ue674" );
		code( R.drawable.emj_1f460, "\udbb9\udcd6" );
		code( R.drawable.emj_1f461, "\ud83d\udc61" );
		code( R.drawable.emj_1f461, "\ue31a" );
		code( R.drawable.emj_1f461, "\ue51a" );
		code( R.drawable.emj_1f461, "\udbb9\udcd7" );
		code( R.drawable.emj_1f462, "\ud83d\udc62" );
		code( R.drawable.emj_1f462, "\ue31b" );
		code( R.drawable.emj_1f462, "\uea9f" );
		code( R.drawable.emj_1f462, "\udbb9\udcd8" );
		code( R.drawable.emj_1f463, "\ud83d\udc63" );
		code( R.drawable.emj_1f463, "\ue536" );
		code( R.drawable.emj_1f463, "\ue698" );
		code( R.drawable.emj_1f463, "\ueb2a" );
		code( R.drawable.emj_1f463, "\udbb9\udd53" );
		code( R.drawable.emj_1f464, "\ud83d\udc64" );
		code( R.drawable.emj_1f464, "\ue6b1" );
		code( R.drawable.emj_1f464, "\udbb8\udd9a" );
		code( R.drawable.emj_1f465, "\ud83d\udc65" );
		code( R.drawable.emj_1f466, "\ud83d\udc66" );
		code( R.drawable.emj_1f466, "\ue001" );
		code( R.drawable.emj_1f466, "\ue4fc" );
		code( R.drawable.emj_1f466, "\udbb8\udd9b" );
		code( R.drawable.emj_1f466_1f3fb, "\ud83d\udc66\ud83c\udffb" );
		code( R.drawable.emj_1f466_1f3fc, "\ud83d\udc66\ud83c\udffc" );
		code( R.drawable.emj_1f466_1f3fd, "\ud83d\udc66\ud83c\udffd" );
		code( R.drawable.emj_1f466_1f3fe, "\ud83d\udc66\ud83c\udffe" );
		code( R.drawable.emj_1f466_1f3ff, "\ud83d\udc66\ud83c\udfff" );
		code( R.drawable.emj_1f467, "\ud83d\udc67" );
		code( R.drawable.emj_1f467, "\ue002" );
	}
	private static void init20(){
		code( R.drawable.emj_1f467, "\ue4fa" );
		code( R.drawable.emj_1f467, "\udbb8\udd9c" );
		code( R.drawable.emj_1f467_1f3fb, "\ud83d\udc67\ud83c\udffb" );
		code( R.drawable.emj_1f467_1f3fc, "\ud83d\udc67\ud83c\udffc" );
		code( R.drawable.emj_1f467_1f3fd, "\ud83d\udc67\ud83c\udffd" );
		code( R.drawable.emj_1f467_1f3fe, "\ud83d\udc67\ud83c\udffe" );
		code( R.drawable.emj_1f467_1f3ff, "\ud83d\udc67\ud83c\udfff" );
		code( R.drawable.emj_1f468, "\ud83d\udc68" );
		code( R.drawable.emj_1f468, "\ue004" );
		code( R.drawable.emj_1f468, "\ue4fc" );
		code( R.drawable.emj_1f468, "\udbb8\udd9d" );
		code( R.drawable.emj_1f468_1f3fb, "\ud83d\udc68\ud83c\udffb" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f33e, "\ud83d\udc68\ud83c\udffb\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f33e, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f373, "\ud83d\udc68\ud83c\udffb\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f373, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f393, "\ud83d\udc68\ud83c\udffb\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f393, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3a4, "\ud83d\udc68\ud83c\udffb\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3a4, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3a8, "\ud83d\udc68\ud83c\udffb\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3a8, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3eb, "\ud83d\udc68\ud83c\udffb\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3eb, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3ed, "\ud83d\udc68\ud83c\udffb\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f3ed, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f4bb, "\ud83d\udc68\ud83c\udffb\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f4bb, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f4bc, "\ud83d\udc68\ud83c\udffb\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f4bc, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f527, "\ud83d\udc68\ud83c\udffb\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f527, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f52c, "\ud83d\udc68\ud83c\udffb\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f52c, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f680, "\ud83d\udc68\ud83c\udffb\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f680, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f692, "\ud83d\udc68\ud83c\udffb\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fb_200d_1f692, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fb_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f468_1f3fb_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffb\u2695" );
		code( R.drawable.emj_1f468_1f3fb_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f468_1f3fb_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffb\u2696" );
		code( R.drawable.emj_1f468_1f3fb_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f468_1f3fb_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffb\u2708" );
		code( R.drawable.emj_1f468_1f3fc, "\ud83d\udc68\ud83c\udffc" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f33e, "\ud83d\udc68\ud83c\udffc\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f33e, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f373, "\ud83d\udc68\ud83c\udffc\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f373, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f393, "\ud83d\udc68\ud83c\udffc\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f393, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3a4, "\ud83d\udc68\ud83c\udffc\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3a4, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3a8, "\ud83d\udc68\ud83c\udffc\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3a8, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3eb, "\ud83d\udc68\ud83c\udffc\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3eb, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3ed, "\ud83d\udc68\ud83c\udffc\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f3ed, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f4bb, "\ud83d\udc68\ud83c\udffc\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f4bb, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f4bc, "\ud83d\udc68\ud83c\udffc\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f4bc, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f527, "\ud83d\udc68\ud83c\udffc\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f527, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f52c, "\ud83d\udc68\ud83c\udffc\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f52c, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f680, "\ud83d\udc68\ud83c\udffc\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f680, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f692, "\ud83d\udc68\ud83c\udffc\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fc_200d_1f692, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fc_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f468_1f3fc_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffc\u2695" );
		code( R.drawable.emj_1f468_1f3fc_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f468_1f3fc_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffc\u2696" );
		code( R.drawable.emj_1f468_1f3fc_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f468_1f3fc_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffc\u2708" );
		code( R.drawable.emj_1f468_1f3fd, "\ud83d\udc68\ud83c\udffd" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f33e, "\ud83d\udc68\ud83c\udffd\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f33e, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f373, "\ud83d\udc68\ud83c\udffd\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f373, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f393, "\ud83d\udc68\ud83c\udffd\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f393, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3a4, "\ud83d\udc68\ud83c\udffd\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3a4, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3a8, "\ud83d\udc68\ud83c\udffd\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3a8, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3eb, "\ud83d\udc68\ud83c\udffd\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3eb, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3ed, "\ud83d\udc68\ud83c\udffd\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f3ed, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f4bb, "\ud83d\udc68\ud83c\udffd\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f4bb, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f4bc, "\ud83d\udc68\ud83c\udffd\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f4bc, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f527, "\ud83d\udc68\ud83c\udffd\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f527, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f52c, "\ud83d\udc68\ud83c\udffd\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f52c, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f680, "\ud83d\udc68\ud83c\udffd\ud83d\ude80" );
	}
	private static void init21(){
		code( R.drawable.emj_1f468_1f3fd_200d_1f680, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f692, "\ud83d\udc68\ud83c\udffd\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fd_200d_1f692, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fd_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f468_1f3fd_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffd\u2695" );
		code( R.drawable.emj_1f468_1f3fd_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f468_1f3fd_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffd\u2696" );
		code( R.drawable.emj_1f468_1f3fd_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f468_1f3fd_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffd\u2708" );
		code( R.drawable.emj_1f468_1f3fe, "\ud83d\udc68\ud83c\udffe" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f33e, "\ud83d\udc68\ud83c\udffe\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f33e, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f373, "\ud83d\udc68\ud83c\udffe\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f373, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f393, "\ud83d\udc68\ud83c\udffe\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f393, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3a4, "\ud83d\udc68\ud83c\udffe\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3a4, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3a8, "\ud83d\udc68\ud83c\udffe\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3a8, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3eb, "\ud83d\udc68\ud83c\udffe\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3eb, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3ed, "\ud83d\udc68\ud83c\udffe\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f3ed, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f4bb, "\ud83d\udc68\ud83c\udffe\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f4bb, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f4bc, "\ud83d\udc68\ud83c\udffe\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f4bc, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f527, "\ud83d\udc68\ud83c\udffe\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f527, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f52c, "\ud83d\udc68\ud83c\udffe\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f52c, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f680, "\ud83d\udc68\ud83c\udffe\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f680, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f692, "\ud83d\udc68\ud83c\udffe\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fe_200d_1f692, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3fe_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f468_1f3fe_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffe\u2695" );
		code( R.drawable.emj_1f468_1f3fe_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f468_1f3fe_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffe\u2696" );
		code( R.drawable.emj_1f468_1f3fe_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f468_1f3fe_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffe\u2708" );
		code( R.drawable.emj_1f468_1f3ff, "\ud83d\udc68\ud83c\udfff" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f33e, "\ud83d\udc68\ud83c\udfff\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f33e, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f373, "\ud83d\udc68\ud83c\udfff\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f373, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f393, "\ud83d\udc68\ud83c\udfff\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f393, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3a4, "\ud83d\udc68\ud83c\udfff\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3a4, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3a8, "\ud83d\udc68\ud83c\udfff\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3a8, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3eb, "\ud83d\udc68\ud83c\udfff\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3eb, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3ed, "\ud83d\udc68\ud83c\udfff\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f3ed, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f4bb, "\ud83d\udc68\ud83c\udfff\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f4bb, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f4bc, "\ud83d\udc68\ud83c\udfff\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f4bc, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f527, "\ud83d\udc68\ud83c\udfff\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f527, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f52c, "\ud83d\udc68\ud83c\udfff\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f52c, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f680, "\ud83d\udc68\ud83c\udfff\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f680, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f692, "\ud83d\udc68\ud83c\udfff\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3ff_200d_1f692, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f468_1f3ff_200d_2695_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f468_1f3ff_200d_2695_fe0f, "\ud83d\udc68\ud83c\udfff\u2695" );
		code( R.drawable.emj_1f468_1f3ff_200d_2696_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f468_1f3ff_200d_2696_fe0f, "\ud83d\udc68\ud83c\udfff\u2696" );
		code( R.drawable.emj_1f468_1f3ff_200d_2708_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f468_1f3ff_200d_2708_fe0f, "\ud83d\udc68\ud83c\udfff\u2708" );
		code( R.drawable.emj_1f468_200d_1f33e, "\ud83d\udc68\ud83c\udf3e" );
		code( R.drawable.emj_1f468_200d_1f33e, "\ud83d\udc68\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f468_200d_1f373, "\ud83d\udc68\ud83c\udf73" );
		code( R.drawable.emj_1f468_200d_1f373, "\ud83d\udc68\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f468_200d_1f393, "\ud83d\udc68\ud83c\udf93" );
		code( R.drawable.emj_1f468_200d_1f393, "\ud83d\udc68\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f468_200d_1f3a4, "\ud83d\udc68\ud83c\udfa4" );
		code( R.drawable.emj_1f468_200d_1f3a4, "\ud83d\udc68\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f468_200d_1f3a8, "\ud83d\udc68\ud83c\udfa8" );
		code( R.drawable.emj_1f468_200d_1f3a8, "\ud83d\udc68\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f468_200d_1f3eb, "\ud83d\udc68\ud83c\udfeb" );
		code( R.drawable.emj_1f468_200d_1f3eb, "\ud83d\udc68\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f468_200d_1f3ed, "\ud83d\udc68\ud83c\udfed" );
		code( R.drawable.emj_1f468_200d_1f3ed, "\ud83d\udc68\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f468_200d_1f466, "\ud83d\udc68\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\ud83d\udc66\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f467, "\ud83d\udc68\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\ud83d\udc67\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\ud83d\udc67\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f466, "\ud83d\udc68\ud83d\udc68\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc66" );
	}
	private static void init22(){
		code( R.drawable.emj_1f468_200d_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\ud83d\udc68\ud83d\udc66\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f467, "\ud83d\udc68\ud83d\udc68\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\ud83d\udc68\ud83d\udc67\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\ud83d\udc68\ud83d\udc67\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f466, "\ud83d\udc68\ud83d\udc69\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc68\ud83d\udc69\ud83d\udc66\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f467, "\ud83d\udc68\ud83d\udc69\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc68\ud83d\udc69\ud83d\udc67\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc68\ud83d\udc69\ud83d\udc67\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f468_200d_1f4bb, "\ud83d\udc68\ud83d\udcbb" );
		code( R.drawable.emj_1f468_200d_1f4bb, "\ud83d\udc68\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f468_200d_1f4bc, "\ud83d\udc68\ud83d\udcbc" );
		code( R.drawable.emj_1f468_200d_1f4bc, "\ud83d\udc68\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f468_200d_1f527, "\ud83d\udc68\ud83d\udd27" );
		code( R.drawable.emj_1f468_200d_1f527, "\ud83d\udc68\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f468_200d_1f52c, "\ud83d\udc68\ud83d\udd2c" );
		code( R.drawable.emj_1f468_200d_1f52c, "\ud83d\udc68\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f468_200d_1f680, "\ud83d\udc68\ud83d\ude80" );
		code( R.drawable.emj_1f468_200d_1f680, "\ud83d\udc68\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f468_200d_1f692, "\ud83d\udc68\ud83d\ude92" );
		code( R.drawable.emj_1f468_200d_1f692, "\ud83d\udc68\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f468_200d_2695_fe0f, "\ud83d\udc68\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f468_200d_2695_fe0f, "\ud83d\udc68\u2695" );
		code( R.drawable.emj_1f468_200d_2696_fe0f, "\ud83d\udc68\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f468_200d_2696_fe0f, "\ud83d\udc68\u2696" );
		code( R.drawable.emj_1f468_200d_2708_fe0f, "\ud83d\udc68\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f468_200d_2708_fe0f, "\ud83d\udc68\u2708" );
		code( R.drawable.emj_1f468_200d_2764_fe0f_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc68" );
		code( R.drawable.emj_1f468_200d_2764_fe0f_200d_1f468, "\ud83d\udc68\u2764\ud83d\udc68" );
		code( R.drawable.emj_1f468_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68" );
		code( R.drawable.emj_1f468_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc68\u2764\ud83d\udc8b\ud83d\udc68" );
		code( R.drawable.emj_1f469, "\ud83d\udc69" );
		code( R.drawable.emj_1f469, "\ue005" );
		code( R.drawable.emj_1f469, "\ue4fa" );
		code( R.drawable.emj_1f469, "\udbb8\udd9e" );
		code( R.drawable.emj_1f469_1f3fb, "\ud83d\udc69\ud83c\udffb" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f33e, "\ud83d\udc69\ud83c\udffb\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f33e, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f373, "\ud83d\udc69\ud83c\udffb\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f373, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f393, "\ud83d\udc69\ud83c\udffb\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f393, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3a4, "\ud83d\udc69\ud83c\udffb\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3a4, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3a8, "\ud83d\udc69\ud83c\udffb\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3a8, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3eb, "\ud83d\udc69\ud83c\udffb\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3eb, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3ed, "\ud83d\udc69\ud83c\udffb\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f3ed, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f4bb, "\ud83d\udc69\ud83c\udffb\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f4bb, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f4bc, "\ud83d\udc69\ud83c\udffb\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f4bc, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f527, "\ud83d\udc69\ud83c\udffb\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f527, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f52c, "\ud83d\udc69\ud83c\udffb\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f52c, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f680, "\ud83d\udc69\ud83c\udffb\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f680, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f692, "\ud83d\udc69\ud83c\udffb\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fb_200d_1f692, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fb_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f469_1f3fb_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffb\u2695" );
		code( R.drawable.emj_1f469_1f3fb_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f469_1f3fb_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffb\u2696" );
		code( R.drawable.emj_1f469_1f3fb_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f469_1f3fb_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffb\u2708" );
		code( R.drawable.emj_1f469_1f3fc, "\ud83d\udc69\ud83c\udffc" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f33e, "\ud83d\udc69\ud83c\udffc\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f33e, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f373, "\ud83d\udc69\ud83c\udffc\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f373, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f393, "\ud83d\udc69\ud83c\udffc\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f393, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3a4, "\ud83d\udc69\ud83c\udffc\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3a4, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3a8, "\ud83d\udc69\ud83c\udffc\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3a8, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3eb, "\ud83d\udc69\ud83c\udffc\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3eb, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3ed, "\ud83d\udc69\ud83c\udffc\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f3ed, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f4bb, "\ud83d\udc69\ud83c\udffc\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f4bb, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f4bc, "\ud83d\udc69\ud83c\udffc\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f4bc, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f527, "\ud83d\udc69\ud83c\udffc\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f527, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f52c, "\ud83d\udc69\ud83c\udffc\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f52c, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f680, "\ud83d\udc69\ud83c\udffc\ud83d\ude80" );
	}
	private static void init23(){
		code( R.drawable.emj_1f469_1f3fc_200d_1f680, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f692, "\ud83d\udc69\ud83c\udffc\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fc_200d_1f692, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fc_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f469_1f3fc_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffc\u2695" );
		code( R.drawable.emj_1f469_1f3fc_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f469_1f3fc_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffc\u2696" );
		code( R.drawable.emj_1f469_1f3fc_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f469_1f3fc_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffc\u2708" );
		code( R.drawable.emj_1f469_1f3fd, "\ud83d\udc69\ud83c\udffd" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f33e, "\ud83d\udc69\ud83c\udffd\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f33e, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f373, "\ud83d\udc69\ud83c\udffd\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f373, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f393, "\ud83d\udc69\ud83c\udffd\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f393, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3a4, "\ud83d\udc69\ud83c\udffd\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3a4, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3a8, "\ud83d\udc69\ud83c\udffd\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3a8, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3eb, "\ud83d\udc69\ud83c\udffd\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3eb, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3ed, "\ud83d\udc69\ud83c\udffd\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f3ed, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f4bb, "\ud83d\udc69\ud83c\udffd\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f4bb, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f4bc, "\ud83d\udc69\ud83c\udffd\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f4bc, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f527, "\ud83d\udc69\ud83c\udffd\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f527, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f52c, "\ud83d\udc69\ud83c\udffd\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f52c, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f680, "\ud83d\udc69\ud83c\udffd\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f680, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f692, "\ud83d\udc69\ud83c\udffd\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fd_200d_1f692, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fd_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f469_1f3fd_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffd\u2695" );
		code( R.drawable.emj_1f469_1f3fd_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f469_1f3fd_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffd\u2696" );
		code( R.drawable.emj_1f469_1f3fd_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f469_1f3fd_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffd\u2708" );
		code( R.drawable.emj_1f469_1f3fe, "\ud83d\udc69\ud83c\udffe" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f33e, "\ud83d\udc69\ud83c\udffe\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f33e, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f373, "\ud83d\udc69\ud83c\udffe\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f373, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f393, "\ud83d\udc69\ud83c\udffe\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f393, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3a4, "\ud83d\udc69\ud83c\udffe\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3a4, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3a8, "\ud83d\udc69\ud83c\udffe\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3a8, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3eb, "\ud83d\udc69\ud83c\udffe\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3eb, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3ed, "\ud83d\udc69\ud83c\udffe\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f3ed, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f4bb, "\ud83d\udc69\ud83c\udffe\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f4bb, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f4bc, "\ud83d\udc69\ud83c\udffe\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f4bc, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f527, "\ud83d\udc69\ud83c\udffe\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f527, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f52c, "\ud83d\udc69\ud83c\udffe\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f52c, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f680, "\ud83d\udc69\ud83c\udffe\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f680, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f692, "\ud83d\udc69\ud83c\udffe\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fe_200d_1f692, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3fe_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f469_1f3fe_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffe\u2695" );
		code( R.drawable.emj_1f469_1f3fe_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f469_1f3fe_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffe\u2696" );
		code( R.drawable.emj_1f469_1f3fe_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f469_1f3fe_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffe\u2708" );
		code( R.drawable.emj_1f469_1f3ff, "\ud83d\udc69\ud83c\udfff" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f33e, "\ud83d\udc69\ud83c\udfff\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f33e, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f373, "\ud83d\udc69\ud83c\udfff\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f373, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f393, "\ud83d\udc69\ud83c\udfff\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f393, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3a4, "\ud83d\udc69\ud83c\udfff\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3a4, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3a8, "\ud83d\udc69\ud83c\udfff\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3a8, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3eb, "\ud83d\udc69\ud83c\udfff\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3eb, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3ed, "\ud83d\udc69\ud83c\udfff\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f3ed, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f4bb, "\ud83d\udc69\ud83c\udfff\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f4bb, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f4bc, "\ud83d\udc69\ud83c\udfff\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f4bc, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f527, "\ud83d\udc69\ud83c\udfff\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f527, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f52c, "\ud83d\udc69\ud83c\udfff\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f52c, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f680, "\ud83d\udc69\ud83c\udfff\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f680, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f469_1f3ff_200d_1f692, "\ud83d\udc69\ud83c\udfff\ud83d\ude92" );
	}
	private static void init24(){
		code( R.drawable.emj_1f469_1f3ff_200d_1f692, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f469_1f3ff_200d_2695_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f469_1f3ff_200d_2695_fe0f, "\ud83d\udc69\ud83c\udfff\u2695" );
		code( R.drawable.emj_1f469_1f3ff_200d_2696_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f469_1f3ff_200d_2696_fe0f, "\ud83d\udc69\ud83c\udfff\u2696" );
		code( R.drawable.emj_1f469_1f3ff_200d_2708_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f469_1f3ff_200d_2708_fe0f, "\ud83d\udc69\ud83c\udfff\u2708" );
		code( R.drawable.emj_1f469_200d_1f33e, "\ud83d\udc69\ud83c\udf3e" );
		code( R.drawable.emj_1f469_200d_1f33e, "\ud83d\udc69\u200d\ud83c\udf3e" );
		code( R.drawable.emj_1f469_200d_1f373, "\ud83d\udc69\ud83c\udf73" );
		code( R.drawable.emj_1f469_200d_1f373, "\ud83d\udc69\u200d\ud83c\udf73" );
		code( R.drawable.emj_1f469_200d_1f393, "\ud83d\udc69\ud83c\udf93" );
		code( R.drawable.emj_1f469_200d_1f393, "\ud83d\udc69\u200d\ud83c\udf93" );
		code( R.drawable.emj_1f469_200d_1f3a4, "\ud83d\udc69\ud83c\udfa4" );
		code( R.drawable.emj_1f469_200d_1f3a4, "\ud83d\udc69\u200d\ud83c\udfa4" );
		code( R.drawable.emj_1f469_200d_1f3a8, "\ud83d\udc69\ud83c\udfa8" );
		code( R.drawable.emj_1f469_200d_1f3a8, "\ud83d\udc69\u200d\ud83c\udfa8" );
		code( R.drawable.emj_1f469_200d_1f3eb, "\ud83d\udc69\ud83c\udfeb" );
		code( R.drawable.emj_1f469_200d_1f3eb, "\ud83d\udc69\u200d\ud83c\udfeb" );
		code( R.drawable.emj_1f469_200d_1f3ed, "\ud83d\udc69\ud83c\udfed" );
		code( R.drawable.emj_1f469_200d_1f3ed, "\ud83d\udc69\u200d\ud83c\udfed" );
		code( R.drawable.emj_1f469_200d_1f466, "\ud83d\udc69\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\ud83d\udc66\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f467, "\ud83d\udc69\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\ud83d\udc67\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\ud83d\udc67\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f466, "\ud83d\udc69\ud83d\udc69\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\ud83d\udc69\ud83d\udc66\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f467, "\ud83d\udc69\ud83d\udc69\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\ud83d\udc69\ud83d\udc67\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\ud83d\udc69\ud83d\udc67\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		code( R.drawable.emj_1f469_200d_1f4bb, "\ud83d\udc69\ud83d\udcbb" );
		code( R.drawable.emj_1f469_200d_1f4bb, "\ud83d\udc69\u200d\ud83d\udcbb" );
		code( R.drawable.emj_1f469_200d_1f4bc, "\ud83d\udc69\ud83d\udcbc" );
		code( R.drawable.emj_1f469_200d_1f4bc, "\ud83d\udc69\u200d\ud83d\udcbc" );
		code( R.drawable.emj_1f469_200d_1f527, "\ud83d\udc69\ud83d\udd27" );
		code( R.drawable.emj_1f469_200d_1f527, "\ud83d\udc69\u200d\ud83d\udd27" );
		code( R.drawable.emj_1f469_200d_1f52c, "\ud83d\udc69\ud83d\udd2c" );
		code( R.drawable.emj_1f469_200d_1f52c, "\ud83d\udc69\u200d\ud83d\udd2c" );
		code( R.drawable.emj_1f469_200d_1f680, "\ud83d\udc69\ud83d\ude80" );
		code( R.drawable.emj_1f469_200d_1f680, "\ud83d\udc69\u200d\ud83d\ude80" );
		code( R.drawable.emj_1f469_200d_1f692, "\ud83d\udc69\ud83d\ude92" );
		code( R.drawable.emj_1f469_200d_1f692, "\ud83d\udc69\u200d\ud83d\ude92" );
		code( R.drawable.emj_1f469_200d_2695_fe0f, "\ud83d\udc69\u200d\u2695\ufe0f" );
		code( R.drawable.emj_1f469_200d_2695_fe0f, "\ud83d\udc69\u2695" );
		code( R.drawable.emj_1f469_200d_2696_fe0f, "\ud83d\udc69\u200d\u2696\ufe0f" );
		code( R.drawable.emj_1f469_200d_2696_fe0f, "\ud83d\udc69\u2696" );
		code( R.drawable.emj_1f469_200d_2708_fe0f, "\ud83d\udc69\u200d\u2708\ufe0f" );
		code( R.drawable.emj_1f469_200d_2708_fe0f, "\ud83d\udc69\u2708" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f468, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc68" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f468, "\ud83d\udc69\u2764\ud83d\udc68" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc69" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f469, "\ud83d\udc69\u2764\ud83d\udc69" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc69\u2764\ud83d\udc8b\ud83d\udc68" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69" );
		code( R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f469, "\ud83d\udc69\u2764\ud83d\udc8b\ud83d\udc69" );
		code( R.drawable.emj_1f46a, "\ud83d\udc6a" );
		code( R.drawable.emj_1f46a, "\ue501" );
		code( R.drawable.emj_1f46a, "\udbb8\udd9f" );
		code( R.drawable.emj_1f46b, "\ud83d\udc6b" );
		code( R.drawable.emj_1f46b, "\ue428" );
		code( R.drawable.emj_1f46b, "\udbb8\udda0" );
		code( R.drawable.emj_1f46c, "\ud83d\udc6c" );
		code( R.drawable.emj_1f46d, "\ud83d\udc6d" );
		code( R.drawable.emj_1f46e, "\ud83d\udc6e" );
		code( R.drawable.emj_1f46e, "\ue152" );
		code( R.drawable.emj_1f46e, "\ue5dd" );
		code( R.drawable.emj_1f46e, "\udbb8\udda1" );
		code( R.drawable.emj_1f46e_1f3fb, "\ud83d\udc6e\ud83c\udffb" );
		code( R.drawable.emj_1f46e_1f3fb_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fb_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f46e_1f3fb_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fb_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f46e_1f3fc, "\ud83d\udc6e\ud83c\udffc" );
		code( R.drawable.emj_1f46e_1f3fc_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fc_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f46e_1f3fc_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fc_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f46e_1f3fd, "\ud83d\udc6e\ud83c\udffd" );
		code( R.drawable.emj_1f46e_1f3fd_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fd_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f46e_1f3fd_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fd_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f46e_1f3fe, "\ud83d\udc6e\ud83c\udffe" );
		code( R.drawable.emj_1f46e_1f3fe_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fe_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f46e_1f3fe_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f46e_1f3fe_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f46e_1f3ff, "\ud83d\udc6e\ud83c\udfff" );
		code( R.drawable.emj_1f46e_1f3ff_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2640\ufe0f" );
	}
	private static void init25(){
		code( R.drawable.emj_1f46e_1f3ff_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f46e_1f3ff_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f46e_1f3ff_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f46e_200d_2640_fe0f, "\ud83d\udc6e\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f46e_200d_2640_fe0f, "\ud83d\udc6e\u2640" );
		code( R.drawable.emj_1f46e_200d_2642_fe0f, "\ud83d\udc6e\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f46e_200d_2642_fe0f, "\ud83d\udc6e\u2642" );
		code( R.drawable.emj_1f46f, "\ud83d\udc6f" );
		code( R.drawable.emj_1f46f, "\ue429" );
		code( R.drawable.emj_1f46f, "\ueadb" );
		code( R.drawable.emj_1f46f, "\udbb8\udda2" );
		code( R.drawable.emj_1f46f_200d_2640_fe0f, "\ud83d\udc6f\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f46f_200d_2640_fe0f, "\ud83d\udc6f\u2640" );
		code( R.drawable.emj_1f46f_200d_2642_fe0f, "\ud83d\udc6f\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f46f_200d_2642_fe0f, "\ud83d\udc6f\u2642" );
		code( R.drawable.emj_1f470, "\ud83d\udc70" );
		code( R.drawable.emj_1f470, "\ueae9" );
		code( R.drawable.emj_1f470, "\udbb8\udda3" );
		code( R.drawable.emj_1f470_1f3fb, "\ud83d\udc70\ud83c\udffb" );
		code( R.drawable.emj_1f470_1f3fc, "\ud83d\udc70\ud83c\udffc" );
		code( R.drawable.emj_1f470_1f3fd, "\ud83d\udc70\ud83c\udffd" );
		code( R.drawable.emj_1f470_1f3fe, "\ud83d\udc70\ud83c\udffe" );
		code( R.drawable.emj_1f470_1f3ff, "\ud83d\udc70\ud83c\udfff" );
		code( R.drawable.emj_1f471, "\ud83d\udc71" );
		code( R.drawable.emj_1f471, "\ue515" );
		code( R.drawable.emj_1f471, "\ueb13" );
		code( R.drawable.emj_1f471, "\udbb8\udda4" );
		code( R.drawable.emj_1f471_1f3fb, "\ud83d\udc71\ud83c\udffb" );
		code( R.drawable.emj_1f471_1f3fb_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f471_1f3fb_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f471_1f3fb_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f471_1f3fb_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f471_1f3fc, "\ud83d\udc71\ud83c\udffc" );
		code( R.drawable.emj_1f471_1f3fc_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f471_1f3fc_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f471_1f3fc_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f471_1f3fc_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f471_1f3fd, "\ud83d\udc71\ud83c\udffd" );
		code( R.drawable.emj_1f471_1f3fd_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f471_1f3fd_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f471_1f3fd_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f471_1f3fd_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f471_1f3fe, "\ud83d\udc71\ud83c\udffe" );
		code( R.drawable.emj_1f471_1f3fe_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f471_1f3fe_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f471_1f3fe_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f471_1f3fe_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f471_1f3ff, "\ud83d\udc71\ud83c\udfff" );
		code( R.drawable.emj_1f471_1f3ff_200d_2640_fe0f, "\ud83d\udc71\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f471_1f3ff_200d_2640_fe0f, "\ud83d\udc71\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f471_1f3ff_200d_2642_fe0f, "\ud83d\udc71\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f471_1f3ff_200d_2642_fe0f, "\ud83d\udc71\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f471_200d_2640_fe0f, "\ud83d\udc71\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f471_200d_2640_fe0f, "\ud83d\udc71\u2640" );
		code( R.drawable.emj_1f471_200d_2642_fe0f, "\ud83d\udc71\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f471_200d_2642_fe0f, "\ud83d\udc71\u2642" );
		code( R.drawable.emj_1f472, "\ud83d\udc72" );
		code( R.drawable.emj_1f472, "\ue516" );
		code( R.drawable.emj_1f472, "\ueb14" );
		code( R.drawable.emj_1f472, "\udbb8\udda5" );
		code( R.drawable.emj_1f472_1f3fb, "\ud83d\udc72\ud83c\udffb" );
		code( R.drawable.emj_1f472_1f3fc, "\ud83d\udc72\ud83c\udffc" );
		code( R.drawable.emj_1f472_1f3fd, "\ud83d\udc72\ud83c\udffd" );
		code( R.drawable.emj_1f472_1f3fe, "\ud83d\udc72\ud83c\udffe" );
		code( R.drawable.emj_1f472_1f3ff, "\ud83d\udc72\ud83c\udfff" );
		code( R.drawable.emj_1f473, "\ud83d\udc73" );
		code( R.drawable.emj_1f473, "\ue517" );
		code( R.drawable.emj_1f473, "\ueb15" );
		code( R.drawable.emj_1f473, "\udbb8\udda6" );
		code( R.drawable.emj_1f473_1f3fb, "\ud83d\udc73\ud83c\udffb" );
		code( R.drawable.emj_1f473_1f3fb_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f473_1f3fb_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f473_1f3fb_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f473_1f3fb_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f473_1f3fc, "\ud83d\udc73\ud83c\udffc" );
		code( R.drawable.emj_1f473_1f3fc_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f473_1f3fc_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f473_1f3fc_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f473_1f3fc_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f473_1f3fd, "\ud83d\udc73\ud83c\udffd" );
		code( R.drawable.emj_1f473_1f3fd_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f473_1f3fd_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f473_1f3fd_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f473_1f3fd_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f473_1f3fe, "\ud83d\udc73\ud83c\udffe" );
		code( R.drawable.emj_1f473_1f3fe_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f473_1f3fe_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f473_1f3fe_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f473_1f3fe_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f473_1f3ff, "\ud83d\udc73\ud83c\udfff" );
		code( R.drawable.emj_1f473_1f3ff_200d_2640_fe0f, "\ud83d\udc73\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f473_1f3ff_200d_2640_fe0f, "\ud83d\udc73\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f473_1f3ff_200d_2642_fe0f, "\ud83d\udc73\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f473_1f3ff_200d_2642_fe0f, "\ud83d\udc73\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f473_200d_2640_fe0f, "\ud83d\udc73\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f473_200d_2640_fe0f, "\ud83d\udc73\u2640" );
		code( R.drawable.emj_1f473_200d_2642_fe0f, "\ud83d\udc73\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f473_200d_2642_fe0f, "\ud83d\udc73\u2642" );
		code( R.drawable.emj_1f474, "\ud83d\udc74" );
		code( R.drawable.emj_1f474, "\ue518" );
		code( R.drawable.emj_1f474, "\ueb16" );
	}
	private static void init26(){
		code( R.drawable.emj_1f474, "\udbb8\udda7" );
		code( R.drawable.emj_1f474_1f3fb, "\ud83d\udc74\ud83c\udffb" );
		code( R.drawable.emj_1f474_1f3fc, "\ud83d\udc74\ud83c\udffc" );
		code( R.drawable.emj_1f474_1f3fd, "\ud83d\udc74\ud83c\udffd" );
		code( R.drawable.emj_1f474_1f3fe, "\ud83d\udc74\ud83c\udffe" );
		code( R.drawable.emj_1f474_1f3ff, "\ud83d\udc74\ud83c\udfff" );
		code( R.drawable.emj_1f475, "\ud83d\udc75" );
		code( R.drawable.emj_1f475, "\ue519" );
		code( R.drawable.emj_1f475, "\ueb17" );
		code( R.drawable.emj_1f475, "\udbb8\udda8" );
		code( R.drawable.emj_1f475_1f3fb, "\ud83d\udc75\ud83c\udffb" );
		code( R.drawable.emj_1f475_1f3fc, "\ud83d\udc75\ud83c\udffc" );
		code( R.drawable.emj_1f475_1f3fd, "\ud83d\udc75\ud83c\udffd" );
		code( R.drawable.emj_1f475_1f3fe, "\ud83d\udc75\ud83c\udffe" );
		code( R.drawable.emj_1f475_1f3ff, "\ud83d\udc75\ud83c\udfff" );
		code( R.drawable.emj_1f476, "\ud83d\udc76" );
		code( R.drawable.emj_1f476, "\ue51a" );
		code( R.drawable.emj_1f476, "\ueb18" );
		code( R.drawable.emj_1f476, "\udbb8\udda9" );
		code( R.drawable.emj_1f476_1f3fb, "\ud83d\udc76\ud83c\udffb" );
		code( R.drawable.emj_1f476_1f3fc, "\ud83d\udc76\ud83c\udffc" );
		code( R.drawable.emj_1f476_1f3fd, "\ud83d\udc76\ud83c\udffd" );
		code( R.drawable.emj_1f476_1f3fe, "\ud83d\udc76\ud83c\udffe" );
		code( R.drawable.emj_1f476_1f3ff, "\ud83d\udc76\ud83c\udfff" );
		code( R.drawable.emj_1f477, "\ud83d\udc77" );
		code( R.drawable.emj_1f477, "\ue51b" );
		code( R.drawable.emj_1f477, "\ueb19" );
		code( R.drawable.emj_1f477, "\udbb8\uddaa" );
		code( R.drawable.emj_1f477_1f3fb, "\ud83d\udc77\ud83c\udffb" );
		code( R.drawable.emj_1f477_1f3fb_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f477_1f3fb_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f477_1f3fb_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f477_1f3fb_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f477_1f3fc, "\ud83d\udc77\ud83c\udffc" );
		code( R.drawable.emj_1f477_1f3fc_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f477_1f3fc_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f477_1f3fc_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f477_1f3fc_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f477_1f3fd, "\ud83d\udc77\ud83c\udffd" );
		code( R.drawable.emj_1f477_1f3fd_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f477_1f3fd_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f477_1f3fd_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f477_1f3fd_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f477_1f3fe, "\ud83d\udc77\ud83c\udffe" );
		code( R.drawable.emj_1f477_1f3fe_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f477_1f3fe_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f477_1f3fe_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f477_1f3fe_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f477_1f3ff, "\ud83d\udc77\ud83c\udfff" );
		code( R.drawable.emj_1f477_1f3ff_200d_2640_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f477_1f3ff_200d_2640_fe0f, "\ud83d\udc77\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f477_1f3ff_200d_2642_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f477_1f3ff_200d_2642_fe0f, "\ud83d\udc77\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f477_200d_2640_fe0f, "\ud83d\udc77\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f477_200d_2640_fe0f, "\ud83d\udc77\u2640" );
		code( R.drawable.emj_1f477_200d_2642_fe0f, "\ud83d\udc77\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f477_200d_2642_fe0f, "\ud83d\udc77\u2642" );
		code( R.drawable.emj_1f478, "\ud83d\udc78" );
		code( R.drawable.emj_1f478, "\ue51c" );
		code( R.drawable.emj_1f478, "\ueb1a" );
		code( R.drawable.emj_1f478, "\udbb8\uddab" );
		code( R.drawable.emj_1f478_1f3fb, "\ud83d\udc78\ud83c\udffb" );
		code( R.drawable.emj_1f478_1f3fc, "\ud83d\udc78\ud83c\udffc" );
		code( R.drawable.emj_1f478_1f3fd, "\ud83d\udc78\ud83c\udffd" );
		code( R.drawable.emj_1f478_1f3fe, "\ud83d\udc78\ud83c\udffe" );
		code( R.drawable.emj_1f478_1f3ff, "\ud83d\udc78\ud83c\udfff" );
		code( R.drawable.emj_1f479, "\ud83d\udc79" );
		code( R.drawable.emj_1f479, "\ueb44" );
		code( R.drawable.emj_1f479, "\udbb8\uddac" );
		code( R.drawable.emj_1f47a, "\ud83d\udc7a" );
		code( R.drawable.emj_1f47a, "\ueb45" );
		code( R.drawable.emj_1f47a, "\udbb8\uddad" );
		code( R.drawable.emj_1f47b, "\ud83d\udc7b" );
		code( R.drawable.emj_1f47b, "\ue11b" );
		code( R.drawable.emj_1f47b, "\ue4cb" );
		code( R.drawable.emj_1f47b, "\udbb8\uddae" );
		code( R.drawable.emj_1f47c, "\ud83d\udc7c" );
		code( R.drawable.emj_1f47c, "\ue04e" );
		code( R.drawable.emj_1f47c, "\ue5bf" );
		code( R.drawable.emj_1f47c, "\udbb8\uddaf" );
		code( R.drawable.emj_1f47c_1f3fb, "\ud83d\udc7c\ud83c\udffb" );
		code( R.drawable.emj_1f47c_1f3fc, "\ud83d\udc7c\ud83c\udffc" );
		code( R.drawable.emj_1f47c_1f3fd, "\ud83d\udc7c\ud83c\udffd" );
		code( R.drawable.emj_1f47c_1f3fe, "\ud83d\udc7c\ud83c\udffe" );
		code( R.drawable.emj_1f47c_1f3ff, "\ud83d\udc7c\ud83c\udfff" );
		code( R.drawable.emj_1f47d, "\ud83d\udc7d" );
		code( R.drawable.emj_1f47d, "\ue10c" );
		code( R.drawable.emj_1f47d, "\ue50e" );
		code( R.drawable.emj_1f47d, "\udbb8\uddb0" );
		code( R.drawable.emj_1f47e, "\ud83d\udc7e" );
		code( R.drawable.emj_1f47e, "\ue12b" );
		code( R.drawable.emj_1f47e, "\ue4ec" );
		code( R.drawable.emj_1f47e, "\udbb8\uddb1" );
		code( R.drawable.emj_1f47f, "\ud83d\udc7f" );
		code( R.drawable.emj_1f47f, "\ue11a" );
		code( R.drawable.emj_1f47f, "\ue4ef" );
		code( R.drawable.emj_1f47f, "\udbb8\uddb2" );
		code( R.drawable.emj_1f480, "\ud83d\udc80" );
		code( R.drawable.emj_1f480, "\ue11c" );
		code( R.drawable.emj_1f480, "\ue4f8" );
		code( R.drawable.emj_1f480, "\udbb8\uddb3" );
	}
	private static void init27(){
		code( R.drawable.emj_1f481, "\ud83d\udc81" );
		code( R.drawable.emj_1f481, "\ue253" );
		code( R.drawable.emj_1f481, "\udbb8\uddb4" );
		code( R.drawable.emj_1f481_1f3fb, "\ud83d\udc81\ud83c\udffb" );
		code( R.drawable.emj_1f481_1f3fb_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f481_1f3fb_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f481_1f3fb_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f481_1f3fb_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f481_1f3fc, "\ud83d\udc81\ud83c\udffc" );
		code( R.drawable.emj_1f481_1f3fc_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f481_1f3fc_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f481_1f3fc_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f481_1f3fc_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f481_1f3fd, "\ud83d\udc81\ud83c\udffd" );
		code( R.drawable.emj_1f481_1f3fd_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f481_1f3fd_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f481_1f3fd_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f481_1f3fd_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f481_1f3fe, "\ud83d\udc81\ud83c\udffe" );
		code( R.drawable.emj_1f481_1f3fe_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f481_1f3fe_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f481_1f3fe_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f481_1f3fe_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f481_1f3ff, "\ud83d\udc81\ud83c\udfff" );
		code( R.drawable.emj_1f481_1f3ff_200d_2640_fe0f, "\ud83d\udc81\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f481_1f3ff_200d_2640_fe0f, "\ud83d\udc81\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f481_1f3ff_200d_2642_fe0f, "\ud83d\udc81\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f481_1f3ff_200d_2642_fe0f, "\ud83d\udc81\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f481_200d_2640_fe0f, "\ud83d\udc81\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f481_200d_2640_fe0f, "\ud83d\udc81\u2640" );
		code( R.drawable.emj_1f481_200d_2642_fe0f, "\ud83d\udc81\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f481_200d_2642_fe0f, "\ud83d\udc81\u2642" );
		code( R.drawable.emj_1f482, "\ud83d\udc82" );
		code( R.drawable.emj_1f482, "\ue51e" );
		code( R.drawable.emj_1f482, "\udbb8\uddb5" );
		code( R.drawable.emj_1f482_1f3fb, "\ud83d\udc82\ud83c\udffb" );
		code( R.drawable.emj_1f482_1f3fb_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f482_1f3fb_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f482_1f3fb_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f482_1f3fb_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f482_1f3fc, "\ud83d\udc82\ud83c\udffc" );
		code( R.drawable.emj_1f482_1f3fc_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f482_1f3fc_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f482_1f3fc_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f482_1f3fc_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f482_1f3fd, "\ud83d\udc82\ud83c\udffd" );
		code( R.drawable.emj_1f482_1f3fd_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f482_1f3fd_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f482_1f3fd_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f482_1f3fd_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f482_1f3fe, "\ud83d\udc82\ud83c\udffe" );
		code( R.drawable.emj_1f482_1f3fe_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f482_1f3fe_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f482_1f3fe_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f482_1f3fe_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f482_1f3ff, "\ud83d\udc82\ud83c\udfff" );
		code( R.drawable.emj_1f482_1f3ff_200d_2640_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f482_1f3ff_200d_2640_fe0f, "\ud83d\udc82\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f482_1f3ff_200d_2642_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f482_1f3ff_200d_2642_fe0f, "\ud83d\udc82\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f482_200d_2640_fe0f, "\ud83d\udc82\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f482_200d_2640_fe0f, "\ud83d\udc82\u2640" );
		code( R.drawable.emj_1f482_200d_2642_fe0f, "\ud83d\udc82\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f482_200d_2642_fe0f, "\ud83d\udc82\u2642" );
		code( R.drawable.emj_1f483, "\ud83d\udc83" );
		code( R.drawable.emj_1f483, "\ue51f" );
		code( R.drawable.emj_1f483, "\ueb1c" );
		code( R.drawable.emj_1f483, "\udbb8\uddb6" );
		code( R.drawable.emj_1f483_1f3fb, "\ud83d\udc83\ud83c\udffb" );
		code( R.drawable.emj_1f483_1f3fc, "\ud83d\udc83\ud83c\udffc" );
		code( R.drawable.emj_1f483_1f3fd, "\ud83d\udc83\ud83c\udffd" );
		code( R.drawable.emj_1f483_1f3fe, "\ud83d\udc83\ud83c\udffe" );
		code( R.drawable.emj_1f483_1f3ff, "\ud83d\udc83\ud83c\udfff" );
		code( R.drawable.emj_1f484, "\ud83d\udc84" );
		code( R.drawable.emj_1f484, "\ue31c" );
		code( R.drawable.emj_1f484, "\ue509" );
		code( R.drawable.emj_1f484, "\ue710" );
		code( R.drawable.emj_1f484, "\udbb8\udd95" );
		code( R.drawable.emj_1f485, "\ud83d\udc85" );
		code( R.drawable.emj_1f485, "\ue31d" );
		code( R.drawable.emj_1f485, "\ueaa0" );
		code( R.drawable.emj_1f485, "\udbb8\udd96" );
		code( R.drawable.emj_1f485_1f3fb, "\ud83d\udc85\ud83c\udffb" );
		code( R.drawable.emj_1f485_1f3fc, "\ud83d\udc85\ud83c\udffc" );
		code( R.drawable.emj_1f485_1f3fd, "\ud83d\udc85\ud83c\udffd" );
		code( R.drawable.emj_1f485_1f3fe, "\ud83d\udc85\ud83c\udffe" );
		code( R.drawable.emj_1f485_1f3ff, "\ud83d\udc85\ud83c\udfff" );
		code( R.drawable.emj_1f486, "\ud83d\udc86" );
		code( R.drawable.emj_1f486, "\ue31e" );
		code( R.drawable.emj_1f486, "\ue50b" );
		code( R.drawable.emj_1f486, "\udbb8\udd97" );
		code( R.drawable.emj_1f486_1f3fb, "\ud83d\udc86\ud83c\udffb" );
		code( R.drawable.emj_1f486_1f3fb_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f486_1f3fb_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f486_1f3fb_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f486_1f3fb_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f486_1f3fc, "\ud83d\udc86\ud83c\udffc" );
		code( R.drawable.emj_1f486_1f3fc_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f486_1f3fc_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f486_1f3fc_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f486_1f3fc_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffc\u2642" );
	}
	private static void init28(){
		code( R.drawable.emj_1f486_1f3fd, "\ud83d\udc86\ud83c\udffd" );
		code( R.drawable.emj_1f486_1f3fd_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f486_1f3fd_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f486_1f3fd_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f486_1f3fd_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f486_1f3fe, "\ud83d\udc86\ud83c\udffe" );
		code( R.drawable.emj_1f486_1f3fe_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f486_1f3fe_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f486_1f3fe_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f486_1f3fe_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f486_1f3ff, "\ud83d\udc86\ud83c\udfff" );
		code( R.drawable.emj_1f486_1f3ff_200d_2640_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f486_1f3ff_200d_2640_fe0f, "\ud83d\udc86\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f486_1f3ff_200d_2642_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f486_1f3ff_200d_2642_fe0f, "\ud83d\udc86\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f486_200d_2640_fe0f, "\ud83d\udc86\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f486_200d_2640_fe0f, "\ud83d\udc86\u2640" );
		code( R.drawable.emj_1f486_200d_2642_fe0f, "\ud83d\udc86\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f486_200d_2642_fe0f, "\ud83d\udc86\u2642" );
		code( R.drawable.emj_1f487, "\ud83d\udc87" );
		code( R.drawable.emj_1f487, "\ue31f" );
		code( R.drawable.emj_1f487, "\ueaa1" );
		code( R.drawable.emj_1f487, "\udbb8\udd98" );
		code( R.drawable.emj_1f487_1f3fb, "\ud83d\udc87\ud83c\udffb" );
		code( R.drawable.emj_1f487_1f3fb_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f487_1f3fb_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f487_1f3fb_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f487_1f3fb_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f487_1f3fc, "\ud83d\udc87\ud83c\udffc" );
		code( R.drawable.emj_1f487_1f3fc_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f487_1f3fc_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f487_1f3fc_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f487_1f3fc_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f487_1f3fd, "\ud83d\udc87\ud83c\udffd" );
		code( R.drawable.emj_1f487_1f3fd_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f487_1f3fd_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f487_1f3fd_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f487_1f3fd_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f487_1f3fe, "\ud83d\udc87\ud83c\udffe" );
		code( R.drawable.emj_1f487_1f3fe_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f487_1f3fe_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f487_1f3fe_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f487_1f3fe_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f487_1f3ff, "\ud83d\udc87\ud83c\udfff" );
		code( R.drawable.emj_1f487_1f3ff_200d_2640_fe0f, "\ud83d\udc87\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f487_1f3ff_200d_2640_fe0f, "\ud83d\udc87\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f487_1f3ff_200d_2642_fe0f, "\ud83d\udc87\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f487_1f3ff_200d_2642_fe0f, "\ud83d\udc87\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f487_200d_2640_fe0f, "\ud83d\udc87\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f487_200d_2640_fe0f, "\ud83d\udc87\u2640" );
		code( R.drawable.emj_1f487_200d_2642_fe0f, "\ud83d\udc87\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f487_200d_2642_fe0f, "\ud83d\udc87\u2642" );
		code( R.drawable.emj_1f488, "\ud83d\udc88" );
		code( R.drawable.emj_1f488, "\ue320" );
		code( R.drawable.emj_1f488, "\ueaa2" );
		code( R.drawable.emj_1f488, "\udbb8\udd99" );
		code( R.drawable.emj_1f489, "\ud83d\udc89" );
		code( R.drawable.emj_1f489, "\ue13b" );
		code( R.drawable.emj_1f489, "\ue510" );
		code( R.drawable.emj_1f489, "\udbb9\udd09" );
		code( R.drawable.emj_1f48a, "\ud83d\udc8a" );
		code( R.drawable.emj_1f48a, "\ue30f" );
		code( R.drawable.emj_1f48a, "\uea9a" );
		code( R.drawable.emj_1f48a, "\udbb9\udd0a" );
		code( R.drawable.emj_1f48b, "\ud83d\udc8b" );
		code( R.drawable.emj_1f48b, "\ue003" );
		code( R.drawable.emj_1f48b, "\ue4eb" );
		code( R.drawable.emj_1f48b, "\ue6f9" );
		code( R.drawable.emj_1f48b, "\udbba\udc23" );
		code( R.drawable.emj_1f48c, "\ud83d\udc8c" );
		code( R.drawable.emj_1f48c, "\ue717" );
		code( R.drawable.emj_1f48c, "\ueb78" );
		code( R.drawable.emj_1f48c, "\udbba\udc24" );
		code( R.drawable.emj_1f48d, "\ud83d\udc8d" );
		code( R.drawable.emj_1f48d, "\ue034" );
		code( R.drawable.emj_1f48d, "\ue514" );
		code( R.drawable.emj_1f48d, "\ue71b" );
		code( R.drawable.emj_1f48d, "\udbba\udc25" );
		code( R.drawable.emj_1f48e, "\ud83d\udc8e" );
		code( R.drawable.emj_1f48e, "\ue035" );
		code( R.drawable.emj_1f48e, "\ue514" );
		code( R.drawable.emj_1f48e, "\udbba\udc26" );
		code( R.drawable.emj_1f48f, "\ud83d\udc8f" );
		code( R.drawable.emj_1f48f, "\ue111" );
		code( R.drawable.emj_1f48f, "\ue5ca" );
		code( R.drawable.emj_1f48f, "\udbba\udc27" );
		code( R.drawable.emj_1f490, "\ud83d\udc90" );
		code( R.drawable.emj_1f490, "\ue306" );
		code( R.drawable.emj_1f490, "\uea95" );
		code( R.drawable.emj_1f490, "\udbba\udc28" );
		code( R.drawable.emj_1f491, "\ud83d\udc91" );
		code( R.drawable.emj_1f491, "\ue425" );
		code( R.drawable.emj_1f491, "\ueada" );
		code( R.drawable.emj_1f491, "\udbba\udc29" );
		code( R.drawable.emj_1f492, "\ud83d\udc92" );
		code( R.drawable.emj_1f492, "\ue43d" );
		code( R.drawable.emj_1f492, "\ue5bb" );
		code( R.drawable.emj_1f492, "\udbba\udc2a" );
		code( R.drawable.emj_1f493, "\ud83d\udc93" );
		code( R.drawable.emj_1f493, "\ue327" );
		code( R.drawable.emj_1f493, "\ueb75" );
	}
	private static void init29(){
		code( R.drawable.emj_1f493, "\udbba\udf0d" );
		code( R.drawable.emj_1f494, "\ud83d\udc94" );
		code( R.drawable.emj_1f494, "\ue023" );
		code( R.drawable.emj_1f494, "\ue477" );
		code( R.drawable.emj_1f494, "\ue6ee" );
		code( R.drawable.emj_1f494, "\udbba\udf0e" );
		code( R.drawable.emj_1f495, "\ud83d\udc95" );
		code( R.drawable.emj_1f495, "\ue478" );
		code( R.drawable.emj_1f495, "\ue6ef" );
		code( R.drawable.emj_1f495, "\udbba\udf0f" );
		code( R.drawable.emj_1f496, "\ud83d\udc96" );
		code( R.drawable.emj_1f496, "\ue6ec" );
		code( R.drawable.emj_1f496, "\ueaa6" );
		code( R.drawable.emj_1f496, "\udbba\udf10" );
		code( R.drawable.emj_1f497, "\ud83d\udc97" );
		code( R.drawable.emj_1f497, "\ue328" );
		code( R.drawable.emj_1f497, "\udbba\udf11" );
		code( R.drawable.emj_1f498, "\ud83d\udc98" );
		code( R.drawable.emj_1f498, "\ue329" );
		code( R.drawable.emj_1f498, "\ue4ea" );
		code( R.drawable.emj_1f498, "\udbba\udf12" );
		code( R.drawable.emj_1f499, "\ud83d\udc99" );
		code( R.drawable.emj_1f499, "\ue32a" );
		code( R.drawable.emj_1f499, "\ueaa7" );
		code( R.drawable.emj_1f499, "\udbba\udf13" );
		code( R.drawable.emj_1f49a, "\ud83d\udc9a" );
		code( R.drawable.emj_1f49a, "\ue32b" );
		code( R.drawable.emj_1f49a, "\ueaa8" );
		code( R.drawable.emj_1f49a, "\udbba\udf14" );
		code( R.drawable.emj_1f49b, "\ud83d\udc9b" );
		code( R.drawable.emj_1f49b, "\ue32c" );
		code( R.drawable.emj_1f49b, "\ueaa9" );
		code( R.drawable.emj_1f49b, "\udbba\udf15" );
		code( R.drawable.emj_1f49c, "\ud83d\udc9c" );
		code( R.drawable.emj_1f49c, "\ue32d" );
		code( R.drawable.emj_1f49c, "\ueaaa" );
		code( R.drawable.emj_1f49c, "\udbba\udf16" );
		code( R.drawable.emj_1f49d, "\ud83d\udc9d" );
		code( R.drawable.emj_1f49d, "\ue437" );
		code( R.drawable.emj_1f49d, "\ueb54" );
		code( R.drawable.emj_1f49d, "\udbba\udf17" );
		code( R.drawable.emj_1f49e, "\ud83d\udc9e" );
		code( R.drawable.emj_1f49e, "\ue5af" );
		code( R.drawable.emj_1f49e, "\ue6ed" );
		code( R.drawable.emj_1f49e, "\udbba\udf18" );
		code( R.drawable.emj_1f49f, "\ud83d\udc9f" );
		code( R.drawable.emj_1f49f, "\ue204" );
		code( R.drawable.emj_1f49f, "\ue595" );
		code( R.drawable.emj_1f49f, "\udbba\udf19" );
		code( R.drawable.emj_1f4a0, "\ud83d\udca0" );
		code( R.drawable.emj_1f4a0, "\udbba\udf55" );
		code( R.drawable.emj_1f4a1, "\ud83d\udca1" );
		code( R.drawable.emj_1f4a1, "\ue10f" );
		code( R.drawable.emj_1f4a1, "\ue476" );
		code( R.drawable.emj_1f4a1, "\ue6fb" );
		code( R.drawable.emj_1f4a1, "\udbba\udf56" );
		code( R.drawable.emj_1f4a2, "\ud83d\udca2" );
		code( R.drawable.emj_1f4a2, "\ue334" );
		code( R.drawable.emj_1f4a2, "\ue4e5" );
		code( R.drawable.emj_1f4a2, "\ue6fc" );
		code( R.drawable.emj_1f4a2, "\udbba\udf57" );
		code( R.drawable.emj_1f4a3, "\ud83d\udca3" );
		code( R.drawable.emj_1f4a3, "\ue311" );
		code( R.drawable.emj_1f4a3, "\ue47a" );
		code( R.drawable.emj_1f4a3, "\ue6fe" );
		code( R.drawable.emj_1f4a3, "\udbba\udf58" );
		code( R.drawable.emj_1f4a4, "\ud83d\udca4" );
		code( R.drawable.emj_1f4a4, "\ue13c" );
		code( R.drawable.emj_1f4a4, "\ue475" );
		code( R.drawable.emj_1f4a4, "\ue701" );
		code( R.drawable.emj_1f4a4, "\udbba\udf59" );
		code( R.drawable.emj_1f4a5, "\ud83d\udca5" );
		code( R.drawable.emj_1f4a5, "\ue5b0" );
		code( R.drawable.emj_1f4a5, "\ue705" );
		code( R.drawable.emj_1f4a5, "\udbba\udf5a" );
		code( R.drawable.emj_1f4a6, "\ud83d\udca6" );
		code( R.drawable.emj_1f4a6, "\ue331" );
		code( R.drawable.emj_1f4a6, "\ue5b1" );
		code( R.drawable.emj_1f4a6, "\ue706" );
		code( R.drawable.emj_1f4a6, "\udbba\udf5b" );
		code( R.drawable.emj_1f4a7, "\ud83d\udca7" );
		code( R.drawable.emj_1f4a7, "\ue4e6" );
		code( R.drawable.emj_1f4a7, "\ue707" );
		code( R.drawable.emj_1f4a7, "\udbba\udf5c" );
		code( R.drawable.emj_1f4a8, "\ud83d\udca8" );
		code( R.drawable.emj_1f4a8, "\ue330" );
		code( R.drawable.emj_1f4a8, "\ue4f4" );
		code( R.drawable.emj_1f4a8, "\ue708" );
		code( R.drawable.emj_1f4a8, "\udbba\udf5d" );
		code( R.drawable.emj_1f4a9, "\ud83d\udca9" );
		code( R.drawable.emj_1f4a9, "\ue05a" );
		code( R.drawable.emj_1f4a9, "\ue4f5" );
		code( R.drawable.emj_1f4a9, "\udbb9\udcf4" );
		code( R.drawable.emj_1f4aa, "\ud83d\udcaa" );
		code( R.drawable.emj_1f4aa, "\ue14c" );
		code( R.drawable.emj_1f4aa, "\ue4e9" );
		code( R.drawable.emj_1f4aa, "\udbba\udf5e" );
		code( R.drawable.emj_1f4aa_1f3fb, "\ud83d\udcaa\ud83c\udffb" );
		code( R.drawable.emj_1f4aa_1f3fc, "\ud83d\udcaa\ud83c\udffc" );
		code( R.drawable.emj_1f4aa_1f3fd, "\ud83d\udcaa\ud83c\udffd" );
		code( R.drawable.emj_1f4aa_1f3fe, "\ud83d\udcaa\ud83c\udffe" );
	}
	private static void init30(){
		code( R.drawable.emj_1f4aa_1f3ff, "\ud83d\udcaa\ud83c\udfff" );
		code( R.drawable.emj_1f4ab, "\ud83d\udcab" );
		code( R.drawable.emj_1f4ab, "\ueb5c" );
		code( R.drawable.emj_1f4ab, "\udbba\udf5f" );
		code( R.drawable.emj_1f4ac, "\ud83d\udcac" );
		code( R.drawable.emj_1f4ac, "\ue4fd" );
		code( R.drawable.emj_1f4ac, "\udbb9\udd32" );
		code( R.drawable.emj_1f4ad, "\ud83d\udcad" );
		code( R.drawable.emj_1f4ae, "\ud83d\udcae" );
		code( R.drawable.emj_1f4ae, "\ue4f0" );
		code( R.drawable.emj_1f4ae, "\udbba\udf7a" );
		code( R.drawable.emj_1f4af, "\ud83d\udcaf" );
		code( R.drawable.emj_1f4af, "\ue4f2" );
		code( R.drawable.emj_1f4af, "\udbba\udf7b" );
		code( R.drawable.emj_1f4b0, "\ud83d\udcb0" );
		code( R.drawable.emj_1f4b0, "\ue12f" );
		code( R.drawable.emj_1f4b0, "\ue4c7" );
		code( R.drawable.emj_1f4b0, "\ue715" );
		code( R.drawable.emj_1f4b0, "\udbb9\udcdd" );
		code( R.drawable.emj_1f4b1, "\ud83d\udcb1" );
		code( R.drawable.emj_1f4b1, "\ue149" );
		code( R.drawable.emj_1f4b1, "\udbb9\udcde" );
		code( R.drawable.emj_1f4b2, "\ud83d\udcb2" );
		code( R.drawable.emj_1f4b2, "\ue579" );
		code( R.drawable.emj_1f4b2, "\udbb9\udce0" );
		code( R.drawable.emj_1f4b3, "\ud83d\udcb3" );
		code( R.drawable.emj_1f4b3, "\ue57c" );
		code( R.drawable.emj_1f4b3, "\udbb9\udce1" );
		code( R.drawable.emj_1f4b4, "\ud83d\udcb4" );
		code( R.drawable.emj_1f4b4, "\ue57d" );
		code( R.drawable.emj_1f4b4, "\ue6d6" );
		code( R.drawable.emj_1f4b4, "\udbb9\udce2" );
		code( R.drawable.emj_1f4b5, "\ud83d\udcb5" );
		code( R.drawable.emj_1f4b5, "\ue585" );
		code( R.drawable.emj_1f4b5, "\udbb9\udce3" );
		code( R.drawable.emj_1f4b6, "\ud83d\udcb6" );
		code( R.drawable.emj_1f4b7, "\ud83d\udcb7" );
		code( R.drawable.emj_1f4b8, "\ud83d\udcb8" );
		code( R.drawable.emj_1f4b8, "\ueb5b" );
		code( R.drawable.emj_1f4b8, "\udbb9\udce4" );
		code( R.drawable.emj_1f4b9, "\ud83d\udcb9" );
		code( R.drawable.emj_1f4b9, "\ue14a" );
		code( R.drawable.emj_1f4b9, "\ue5dc" );
		code( R.drawable.emj_1f4b9, "\udbb9\udcdf" );
		code( R.drawable.emj_1f4ba, "\ud83d\udcba" );
		code( R.drawable.emj_1f4ba, "\ue11f" );
		code( R.drawable.emj_1f4ba, "\ue6b2" );
		code( R.drawable.emj_1f4ba, "\udbb9\udd37" );
		code( R.drawable.emj_1f4bb, "\ud83d\udcbb" );
		code( R.drawable.emj_1f4bb, "\ue00c" );
		code( R.drawable.emj_1f4bb, "\ue5b8" );
		code( R.drawable.emj_1f4bb, "\ue716" );
		code( R.drawable.emj_1f4bb, "\udbb9\udd38" );
		code( R.drawable.emj_1f4bc, "\ud83d\udcbc" );
		code( R.drawable.emj_1f4bc, "\ue11e" );
		code( R.drawable.emj_1f4bc, "\ue5ce" );
		code( R.drawable.emj_1f4bc, "\udbb9\udd3b" );
		code( R.drawable.emj_1f4bd, "\ud83d\udcbd" );
		code( R.drawable.emj_1f4bd, "\ue316" );
		code( R.drawable.emj_1f4bd, "\ue582" );
		code( R.drawable.emj_1f4bd, "\udbb9\udd3c" );
		code( R.drawable.emj_1f4be, "\ud83d\udcbe" );
		code( R.drawable.emj_1f4be, "\ue562" );
		code( R.drawable.emj_1f4be, "\udbb9\udd3d" );
		code( R.drawable.emj_1f4bf, "\ud83d\udcbf" );
		code( R.drawable.emj_1f4bf, "\ue126" );
		code( R.drawable.emj_1f4bf, "\ue50c" );
		code( R.drawable.emj_1f4bf, "\ue68c" );
		code( R.drawable.emj_1f4bf, "\udbba\udc1d" );
		code( R.drawable.emj_1f4c0, "\ud83d\udcc0" );
		code( R.drawable.emj_1f4c0, "\ue127" );
		code( R.drawable.emj_1f4c0, "\ue50c" );
		code( R.drawable.emj_1f4c0, "\udbba\udc1e" );
		code( R.drawable.emj_1f4c1, "\ud83d\udcc1" );
		code( R.drawable.emj_1f4c1, "\ue58f" );
		code( R.drawable.emj_1f4c1, "\udbb9\udd43" );
		code( R.drawable.emj_1f4c2, "\ud83d\udcc2" );
		code( R.drawable.emj_1f4c2, "\ue590" );
		code( R.drawable.emj_1f4c2, "\udbb9\udd44" );
		code( R.drawable.emj_1f4c3, "\ud83d\udcc3" );
		code( R.drawable.emj_1f4c3, "\ue561" );
		code( R.drawable.emj_1f4c3, "\udbb9\udd40" );
		code( R.drawable.emj_1f4c4, "\ud83d\udcc4" );
		code( R.drawable.emj_1f4c4, "\ue569" );
		code( R.drawable.emj_1f4c4, "\ue689" );
		code( R.drawable.emj_1f4c4, "\udbb9\udd41" );
		code( R.drawable.emj_1f4c5, "\ud83d\udcc5" );
		code( R.drawable.emj_1f4c5, "\ue563" );
		code( R.drawable.emj_1f4c5, "\udbb9\udd42" );
		code( R.drawable.emj_1f4c6, "\ud83d\udcc6" );
		code( R.drawable.emj_1f4c6, "\ue56a" );
		code( R.drawable.emj_1f4c6, "\udbb9\udd49" );
		code( R.drawable.emj_1f4c7, "\ud83d\udcc7" );
		code( R.drawable.emj_1f4c7, "\ue56c" );
		code( R.drawable.emj_1f4c7, "\udbb9\udd4d" );
		code( R.drawable.emj_1f4c8, "\ud83d\udcc8" );
		code( R.drawable.emj_1f4c8, "\ue575" );
		code( R.drawable.emj_1f4c8, "\udbb9\udd4b" );
		code( R.drawable.emj_1f4c9, "\ud83d\udcc9" );
		code( R.drawable.emj_1f4c9, "\ue576" );
		code( R.drawable.emj_1f4c9, "\udbb9\udd4c" );
	}
	private static void init31(){
		code( R.drawable.emj_1f4ca, "\ud83d\udcca" );
		code( R.drawable.emj_1f4ca, "\ue574" );
		code( R.drawable.emj_1f4ca, "\udbb9\udd4a" );
		code( R.drawable.emj_1f4cb, "\ud83d\udccb" );
		code( R.drawable.emj_1f4cb, "\ue564" );
		code( R.drawable.emj_1f4cb, "\udbb9\udd48" );
		code( R.drawable.emj_1f4cc, "\ud83d\udccc" );
		code( R.drawable.emj_1f4cc, "\ue56d" );
		code( R.drawable.emj_1f4cc, "\udbb9\udd4e" );
		code( R.drawable.emj_1f4cd, "\ud83d\udccd" );
		code( R.drawable.emj_1f4cd, "\ue560" );
		code( R.drawable.emj_1f4cd, "\udbb9\udd3f" );
		code( R.drawable.emj_1f4ce, "\ud83d\udcce" );
		code( R.drawable.emj_1f4ce, "\ue4a0" );
		code( R.drawable.emj_1f4ce, "\ue730" );
		code( R.drawable.emj_1f4ce, "\udbb9\udd3a" );
		code( R.drawable.emj_1f4cf, "\ud83d\udccf" );
		code( R.drawable.emj_1f4cf, "\ue570" );
		code( R.drawable.emj_1f4cf, "\udbb9\udd50" );
		code( R.drawable.emj_1f4d0, "\ud83d\udcd0" );
		code( R.drawable.emj_1f4d0, "\ue4a2" );
		code( R.drawable.emj_1f4d0, "\udbb9\udd51" );
		code( R.drawable.emj_1f4d1, "\ud83d\udcd1" );
		code( R.drawable.emj_1f4d1, "\ueb0b" );
		code( R.drawable.emj_1f4d1, "\udbb9\udd52" );
		code( R.drawable.emj_1f4d2, "\ud83d\udcd2" );
		code( R.drawable.emj_1f4d2, "\ue56e" );
		code( R.drawable.emj_1f4d2, "\udbb9\udd4f" );
		code( R.drawable.emj_1f4d3, "\ud83d\udcd3" );
		code( R.drawable.emj_1f4d3, "\ue56b" );
		code( R.drawable.emj_1f4d3, "\udbb9\udd45" );
		code( R.drawable.emj_1f4d4, "\ud83d\udcd4" );
		code( R.drawable.emj_1f4d4, "\ue49d" );
		code( R.drawable.emj_1f4d4, "\udbb9\udd47" );
		code( R.drawable.emj_1f4d5, "\ud83d\udcd5" );
		code( R.drawable.emj_1f4d5, "\ue568" );
		code( R.drawable.emj_1f4d5, "\udbb9\udd02" );
		code( R.drawable.emj_1f4d6, "\ud83d\udcd6" );
		code( R.drawable.emj_1f4d6, "\ue148" );
		code( R.drawable.emj_1f4d6, "\ue49f" );
		code( R.drawable.emj_1f4d6, "\udbb9\udd46" );
		code( R.drawable.emj_1f4d7, "\ud83d\udcd7" );
		code( R.drawable.emj_1f4d7, "\ue565" );
		code( R.drawable.emj_1f4d7, "\udbb9\udcff" );
		code( R.drawable.emj_1f4d8, "\ud83d\udcd8" );
		code( R.drawable.emj_1f4d8, "\ue566" );
		code( R.drawable.emj_1f4d8, "\udbb9\udd00" );
		code( R.drawable.emj_1f4d9, "\ud83d\udcd9" );
		code( R.drawable.emj_1f4d9, "\ue567" );
		code( R.drawable.emj_1f4d9, "\ue683" );
		code( R.drawable.emj_1f4d9, "\udbb9\udd01" );
		code( R.drawable.emj_1f4da, "\ud83d\udcda" );
		code( R.drawable.emj_1f4da, "\ue56f" );
		code( R.drawable.emj_1f4da, "\udbb9\udd03" );
		code( R.drawable.emj_1f4db, "\ud83d\udcdb" );
		code( R.drawable.emj_1f4db, "\ue51d" );
		code( R.drawable.emj_1f4db, "\udbb9\udd04" );
		code( R.drawable.emj_1f4dc, "\ud83d\udcdc" );
		code( R.drawable.emj_1f4dc, "\ue55f" );
		code( R.drawable.emj_1f4dc, "\udbb9\udcfd" );
		code( R.drawable.emj_1f4dd, "\ud83d\udcdd" );
		code( R.drawable.emj_1f4dd, "\ue301" );
		code( R.drawable.emj_1f4dd, "\uea92" );
		code( R.drawable.emj_1f4dd, "\udbb9\udd27" );
		code( R.drawable.emj_1f4de, "\ud83d\udcde" );
		code( R.drawable.emj_1f4de, "\ue51e" );
		code( R.drawable.emj_1f4de, "\udbb9\udd24" );
		code( R.drawable.emj_1f4df, "\ud83d\udcdf" );
		code( R.drawable.emj_1f4df, "\ue59b" );
		code( R.drawable.emj_1f4df, "\ue65a" );
		code( R.drawable.emj_1f4df, "\udbb9\udd22" );
		code( R.drawable.emj_1f4e0, "\ud83d\udce0" );
		code( R.drawable.emj_1f4e0, "\ue00b" );
		code( R.drawable.emj_1f4e0, "\ue520" );
		code( R.drawable.emj_1f4e0, "\ue6d0" );
		code( R.drawable.emj_1f4e0, "\udbb9\udd28" );
		code( R.drawable.emj_1f4e1, "\ud83d\udce1" );
		code( R.drawable.emj_1f4e1, "\ue14b" );
		code( R.drawable.emj_1f4e1, "\ue4a8" );
		code( R.drawable.emj_1f4e1, "\udbb9\udd31" );
		code( R.drawable.emj_1f4e2, "\ud83d\udce2" );
		code( R.drawable.emj_1f4e2, "\ue142" );
		code( R.drawable.emj_1f4e2, "\ue511" );
		code( R.drawable.emj_1f4e2, "\udbb9\udd2f" );
		code( R.drawable.emj_1f4e3, "\ud83d\udce3" );
		code( R.drawable.emj_1f4e3, "\ue317" );
		code( R.drawable.emj_1f4e3, "\ue511" );
		code( R.drawable.emj_1f4e3, "\udbb9\udd30" );
		code( R.drawable.emj_1f4e4, "\ud83d\udce4" );
		code( R.drawable.emj_1f4e4, "\ue592" );
		code( R.drawable.emj_1f4e4, "\udbb9\udd33" );
		code( R.drawable.emj_1f4e5, "\ud83d\udce5" );
		code( R.drawable.emj_1f4e5, "\ue593" );
		code( R.drawable.emj_1f4e5, "\udbb9\udd34" );
		code( R.drawable.emj_1f4e6, "\ud83d\udce6" );
		code( R.drawable.emj_1f4e6, "\ue51f" );
		code( R.drawable.emj_1f4e6, "\udbb9\udd35" );
		code( R.drawable.emj_1f4e7, "\ud83d\udce7" );
		code( R.drawable.emj_1f4e7, "\ue6d3" );
		code( R.drawable.emj_1f4e7, "\ueb71" );
		code( R.drawable.emj_1f4e7, "\udbba\udf92" );
	}
	private static void init32(){
		code( R.drawable.emj_1f4e8, "\ud83d\udce8" );
		code( R.drawable.emj_1f4e8, "\ue591" );
		code( R.drawable.emj_1f4e8, "\udbb9\udd2a" );
		code( R.drawable.emj_1f4e9, "\ud83d\udce9" );
		code( R.drawable.emj_1f4e9, "\ue103" );
		code( R.drawable.emj_1f4e9, "\ue6cf" );
		code( R.drawable.emj_1f4e9, "\ueb62" );
		code( R.drawable.emj_1f4e9, "\udbb9\udd2b" );
		code( R.drawable.emj_1f4ea, "\ud83d\udcea" );
		code( R.drawable.emj_1f4ea, "\ue51b" );
		code( R.drawable.emj_1f4ea, "\ue665" );
		code( R.drawable.emj_1f4ea, "\udbb9\udd2c" );
		code( R.drawable.emj_1f4eb, "\ud83d\udceb" );
		code( R.drawable.emj_1f4eb, "\ue101" );
		code( R.drawable.emj_1f4eb, "\ue665" );
		code( R.drawable.emj_1f4eb, "\ueb0a" );
		code( R.drawable.emj_1f4eb, "\udbb9\udd2d" );
		code( R.drawable.emj_1f4ec, "\ud83d\udcec" );
		code( R.drawable.emj_1f4ed, "\ud83d\udced" );
		code( R.drawable.emj_1f4ee, "\ud83d\udcee" );
		code( R.drawable.emj_1f4ee, "\ue102" );
		code( R.drawable.emj_1f4ee, "\ue51b" );
		code( R.drawable.emj_1f4ee, "\ue665" );
		code( R.drawable.emj_1f4ee, "\udbb9\udd2e" );
		code( R.drawable.emj_1f4ef, "\ud83d\udcef" );
		code( R.drawable.emj_1f4f0, "\ud83d\udcf0" );
		code( R.drawable.emj_1f4f0, "\ue58b" );
		code( R.drawable.emj_1f4f0, "\udbba\udc22" );
		code( R.drawable.emj_1f4f1, "\ud83d\udcf1" );
		code( R.drawable.emj_1f4f1, "\ue00a" );
		code( R.drawable.emj_1f4f1, "\ue588" );
		code( R.drawable.emj_1f4f1, "\ue688" );
		code( R.drawable.emj_1f4f1, "\udbb9\udd25" );
		code( R.drawable.emj_1f4f2, "\ud83d\udcf2" );
		code( R.drawable.emj_1f4f2, "\ue104" );
		code( R.drawable.emj_1f4f2, "\ue6ce" );
		code( R.drawable.emj_1f4f2, "\ueb08" );
		code( R.drawable.emj_1f4f2, "\udbb9\udd26" );
		code( R.drawable.emj_1f4f3, "\ud83d\udcf3" );
		code( R.drawable.emj_1f4f3, "\ue250" );
		code( R.drawable.emj_1f4f3, "\uea90" );
		code( R.drawable.emj_1f4f3, "\udbba\udc39" );
		code( R.drawable.emj_1f4f4, "\ud83d\udcf4" );
		code( R.drawable.emj_1f4f4, "\ue251" );
		code( R.drawable.emj_1f4f4, "\uea91" );
		code( R.drawable.emj_1f4f4, "\udbba\udc3a" );
		code( R.drawable.emj_1f4f5, "\ud83d\udcf5" );
		code( R.drawable.emj_1f4f6, "\ud83d\udcf6" );
		code( R.drawable.emj_1f4f6, "\ue20b" );
		code( R.drawable.emj_1f4f6, "\uea84" );
		code( R.drawable.emj_1f4f6, "\udbba\udc38" );
		code( R.drawable.emj_1f4f7, "\ud83d\udcf7" );
		code( R.drawable.emj_1f4f7, "\ue008" );
		code( R.drawable.emj_1f4f7, "\ue515" );
		code( R.drawable.emj_1f4f7, "\ue681" );
		code( R.drawable.emj_1f4f7, "\udbb9\udcef" );
		code( R.drawable.emj_1f4f8, "\ud83d\udcf8" );
		code( R.drawable.emj_1f4f9, "\ud83d\udcf9" );
		code( R.drawable.emj_1f4f9, "\ue57e" );
		code( R.drawable.emj_1f4f9, "\udbb9\udcf9" );
		code( R.drawable.emj_1f4fa, "\ud83d\udcfa" );
		code( R.drawable.emj_1f4fa, "\ue12a" );
		code( R.drawable.emj_1f4fa, "\ue502" );
		code( R.drawable.emj_1f4fa, "\ue68a" );
		code( R.drawable.emj_1f4fa, "\udbba\udc1c" );
		code( R.drawable.emj_1f4fb, "\ud83d\udcfb" );
		code( R.drawable.emj_1f4fb, "\ue128" );
		code( R.drawable.emj_1f4fb, "\ue5b9" );
		code( R.drawable.emj_1f4fb, "\udbba\udc1f" );
		code( R.drawable.emj_1f4fc, "\ud83d\udcfc" );
		code( R.drawable.emj_1f4fc, "\ue129" );
		code( R.drawable.emj_1f4fc, "\ue580" );
		code( R.drawable.emj_1f4fc, "\udbba\udc20" );
		code( R.drawable.emj_1f4fd_fe0f, "\ud83d\udcfd" );
		code( R.drawable.emj_1f4fd_fe0f, "\ud83d\udcfd\ufe0f" );
		code( R.drawable.emj_1f4ff, "\ud83d\udcff" );
		code( R.drawable.emj_1f500, "\ud83d\udd00" );
		code( R.drawable.emj_1f501, "\ud83d\udd01" );
		code( R.drawable.emj_1f502, "\ud83d\udd02" );
		code( R.drawable.emj_1f503, "\ud83d\udd03" );
		code( R.drawable.emj_1f503, "\ueb0d" );
		code( R.drawable.emj_1f503, "\udbba\udf91" );
		code( R.drawable.emj_1f504, "\ud83d\udd04" );
		code( R.drawable.emj_1f505, "\ud83d\udd05" );
		code( R.drawable.emj_1f506, "\ud83d\udd06" );
		code( R.drawable.emj_1f507, "\ud83d\udd07" );
		code( R.drawable.emj_1f508, "\ud83d\udd08" );
		code( R.drawable.emj_1f509, "\ud83d\udd09" );
		code( R.drawable.emj_1f50a, "\ud83d\udd0a" );
		code( R.drawable.emj_1f50a, "\ue141" );
		code( R.drawable.emj_1f50a, "\ue511" );
		code( R.drawable.emj_1f50a, "\udbba\udc21" );
		code( R.drawable.emj_1f50b, "\ud83d\udd0b" );
		code( R.drawable.emj_1f50b, "\ue584" );
		code( R.drawable.emj_1f50b, "\udbb9\udcfc" );
		code( R.drawable.emj_1f50c, "\ud83d\udd0c" );
		code( R.drawable.emj_1f50c, "\ue589" );
		code( R.drawable.emj_1f50c, "\udbb9\udcfe" );
		code( R.drawable.emj_1f50d, "\ud83d\udd0d" );
		code( R.drawable.emj_1f50d, "\ue114" );
		code( R.drawable.emj_1f50d, "\ue518" );
	}
	private static void init33(){
		code( R.drawable.emj_1f50d, "\ue6dc" );
		code( R.drawable.emj_1f50d, "\udbba\udf85" );
		code( R.drawable.emj_1f50e, "\ud83d\udd0e" );
		code( R.drawable.emj_1f50e, "\ueb05" );
		code( R.drawable.emj_1f50e, "\udbba\udf8d" );
		code( R.drawable.emj_1f50f, "\ud83d\udd0f" );
		code( R.drawable.emj_1f50f, "\ueb0c" );
		code( R.drawable.emj_1f50f, "\udbba\udf90" );
		code( R.drawable.emj_1f510, "\ud83d\udd10" );
		code( R.drawable.emj_1f510, "\ueafc" );
		code( R.drawable.emj_1f510, "\udbba\udf8a" );
		code( R.drawable.emj_1f511, "\ud83d\udd11" );
		code( R.drawable.emj_1f511, "\ue03f" );
		code( R.drawable.emj_1f511, "\ue519" );
		code( R.drawable.emj_1f511, "\ue6d9" );
		code( R.drawable.emj_1f511, "\udbba\udf82" );
		code( R.drawable.emj_1f512, "\ud83d\udd12" );
		code( R.drawable.emj_1f512, "\ue144" );
		code( R.drawable.emj_1f512, "\ue51c" );
		code( R.drawable.emj_1f512, "\udbba\udf86" );
		code( R.drawable.emj_1f513, "\ud83d\udd13" );
		code( R.drawable.emj_1f513, "\ue145" );
		code( R.drawable.emj_1f513, "\ue51c" );
		code( R.drawable.emj_1f513, "\udbba\udf87" );
		code( R.drawable.emj_1f514, "\ud83d\udd14" );
		code( R.drawable.emj_1f514, "\ue325" );
		code( R.drawable.emj_1f514, "\ue512" );
		code( R.drawable.emj_1f514, "\ue713" );
		code( R.drawable.emj_1f514, "\udbb9\udcf2" );
		code( R.drawable.emj_1f515, "\ud83d\udd15" );
		code( R.drawable.emj_1f516, "\ud83d\udd16" );
		code( R.drawable.emj_1f516, "\ueb07" );
		code( R.drawable.emj_1f516, "\udbba\udf8f" );
		code( R.drawable.emj_1f517, "\ud83d\udd17" );
		code( R.drawable.emj_1f517, "\ue58a" );
		code( R.drawable.emj_1f517, "\udbba\udf4b" );
		code( R.drawable.emj_1f518, "\ud83d\udd18" );
		code( R.drawable.emj_1f518, "\ueb04" );
		code( R.drawable.emj_1f518, "\udbba\udf8c" );
		code( R.drawable.emj_1f519, "\ud83d\udd19" );
		code( R.drawable.emj_1f519, "\ueb06" );
		code( R.drawable.emj_1f519, "\udbba\udf8e" );
		code( R.drawable.emj_1f51a, "\ud83d\udd1a" );
		code( R.drawable.emj_1f51a, "\ue6b9" );
		code( R.drawable.emj_1f51a, "\udbb8\udc1a" );
		code( R.drawable.emj_1f51b, "\ud83d\udd1b" );
		code( R.drawable.emj_1f51b, "\ue6b8" );
		code( R.drawable.emj_1f51b, "\udbb8\udc19" );
		code( R.drawable.emj_1f51c, "\ud83d\udd1c" );
		code( R.drawable.emj_1f51c, "\ue6b7" );
		code( R.drawable.emj_1f51c, "\udbb8\udc18" );
		code( R.drawable.emj_1f51d, "\ud83d\udd1d" );
		code( R.drawable.emj_1f51d, "\ue24c" );
		code( R.drawable.emj_1f51d, "\udbba\udf42" );
		code( R.drawable.emj_1f51e, "\ud83d\udd1e" );
		code( R.drawable.emj_1f51e, "\ue207" );
		code( R.drawable.emj_1f51e, "\uea83" );
		code( R.drawable.emj_1f51e, "\udbba\udf25" );
		code( R.drawable.emj_1f51f, "\ud83d\udd1f" );
		code( R.drawable.emj_1f51f, "\ue52b" );
		code( R.drawable.emj_1f51f, "\udbba\udc3b" );
		code( R.drawable.emj_1f520, "\ud83d\udd20" );
		code( R.drawable.emj_1f520, "\ueafd" );
		code( R.drawable.emj_1f520, "\udbba\udf7c" );
		code( R.drawable.emj_1f521, "\ud83d\udd21" );
		code( R.drawable.emj_1f521, "\ueafe" );
		code( R.drawable.emj_1f521, "\udbba\udf7d" );
		code( R.drawable.emj_1f522, "\ud83d\udd22" );
		code( R.drawable.emj_1f522, "\ueaff" );
		code( R.drawable.emj_1f522, "\udbba\udf7e" );
		code( R.drawable.emj_1f523, "\ud83d\udd23" );
		code( R.drawable.emj_1f523, "\ueb00" );
		code( R.drawable.emj_1f523, "\udbba\udf7f" );
		code( R.drawable.emj_1f524, "\ud83d\udd24" );
		code( R.drawable.emj_1f524, "\ueb55" );
		code( R.drawable.emj_1f524, "\udbba\udf80" );
		code( R.drawable.emj_1f525, "\ud83d\udd25" );
		code( R.drawable.emj_1f525, "\ue11d" );
		code( R.drawable.emj_1f525, "\ue47b" );
		code( R.drawable.emj_1f525, "\udbb9\udcf6" );
		code( R.drawable.emj_1f526, "\ud83d\udd26" );
		code( R.drawable.emj_1f526, "\ue583" );
		code( R.drawable.emj_1f526, "\udbb9\udcfb" );
		code( R.drawable.emj_1f527, "\ud83d\udd27" );
		code( R.drawable.emj_1f527, "\ue587" );
		code( R.drawable.emj_1f527, "\ue718" );
		code( R.drawable.emj_1f527, "\udbb9\udcc9" );
		code( R.drawable.emj_1f528, "\ud83d\udd28" );
		code( R.drawable.emj_1f528, "\ue116" );
		code( R.drawable.emj_1f528, "\ue5cb" );
		code( R.drawable.emj_1f528, "\udbb9\udcca" );
		code( R.drawable.emj_1f529, "\ud83d\udd29" );
		code( R.drawable.emj_1f529, "\ue581" );
		code( R.drawable.emj_1f529, "\udbb9\udccb" );
		code( R.drawable.emj_1f52a, "\ud83d\udd2a" );
		code( R.drawable.emj_1f52a, "\ue57f" );
		code( R.drawable.emj_1f52a, "\udbb9\udcfa" );
		code( R.drawable.emj_1f52b, "\ud83d\udd2b" );
		code( R.drawable.emj_1f52b, "\ue113" );
		code( R.drawable.emj_1f52b, "\ue50a" );
		code( R.drawable.emj_1f52b, "\udbb9\udcf5" );
	}
	private static void init34(){
		code( R.drawable.emj_1f52c, "\ud83d\udd2c" );
		code( R.drawable.emj_1f52d, "\ud83d\udd2d" );
		code( R.drawable.emj_1f52e, "\ud83d\udd2e" );
		code( R.drawable.emj_1f52e, "\uea8f" );
		code( R.drawable.emj_1f52e, "\udbb9\udcf7" );
		code( R.drawable.emj_1f52f, "\ud83d\udd2f" );
		code( R.drawable.emj_1f52f, "\ue23e" );
		code( R.drawable.emj_1f52f, "\udbb9\udcf8" );
		code( R.drawable.emj_1f530, "\ud83d\udd30" );
		code( R.drawable.emj_1f530, "\ue209" );
		code( R.drawable.emj_1f530, "\ue480" );
		code( R.drawable.emj_1f530, "\udbb8\udc44" );
		code( R.drawable.emj_1f531, "\ud83d\udd31" );
		code( R.drawable.emj_1f531, "\ue031" );
		code( R.drawable.emj_1f531, "\ue5c9" );
		code( R.drawable.emj_1f531, "\udbb9\udcd2" );
		code( R.drawable.emj_1f532, "\ud83d\udd32" );
		code( R.drawable.emj_1f532, "\ue21a" );
		code( R.drawable.emj_1f532, "\ue54b" );
		code( R.drawable.emj_1f533, "\ud83d\udd33" );
		code( R.drawable.emj_1f533, "\ue21b" );
		code( R.drawable.emj_1f533, "\ue54b" );
		code( R.drawable.emj_1f533, "\udbba\udf67" );
		code( R.drawable.emj_1f534, "\ud83d\udd34" );
		code( R.drawable.emj_1f534, "\ue219" );
		code( R.drawable.emj_1f534, "\ue54a" );
		code( R.drawable.emj_1f534, "\udbba\udf63" );
		code( R.drawable.emj_1f535, "\ud83d\udd35" );
		code( R.drawable.emj_1f535, "\ue54b" );
		code( R.drawable.emj_1f535, "\udbba\udf64" );
		code( R.drawable.emj_1f536, "\ud83d\udd36" );
		code( R.drawable.emj_1f536, "\ue546" );
		code( R.drawable.emj_1f536, "\udbba\udf73" );
		code( R.drawable.emj_1f537, "\ud83d\udd37" );
		code( R.drawable.emj_1f537, "\ue547" );
		code( R.drawable.emj_1f537, "\udbba\udf74" );
		code( R.drawable.emj_1f538, "\ud83d\udd38" );
		code( R.drawable.emj_1f538, "\ue536" );
		code( R.drawable.emj_1f538, "\udbba\udf75" );
		code( R.drawable.emj_1f539, "\ud83d\udd39" );
		code( R.drawable.emj_1f539, "\ue537" );
		code( R.drawable.emj_1f539, "\udbba\udf76" );
		code( R.drawable.emj_1f53a, "\ud83d\udd3a" );
		code( R.drawable.emj_1f53a, "\ue55a" );
		code( R.drawable.emj_1f53a, "\udbba\udf78" );
		code( R.drawable.emj_1f53b, "\ud83d\udd3b" );
		code( R.drawable.emj_1f53b, "\ue55b" );
		code( R.drawable.emj_1f53b, "\udbba\udf79" );
		code( R.drawable.emj_1f53c, "\ud83d\udd3c" );
		code( R.drawable.emj_1f53c, "\ue543" );
		code( R.drawable.emj_1f53c, "\udbba\udf01" );
		code( R.drawable.emj_1f53d, "\ud83d\udd3d" );
		code( R.drawable.emj_1f53d, "\ue542" );
		code( R.drawable.emj_1f53d, "\udbba\udf00" );
		code( R.drawable.emj_1f549_fe0f, "\ud83d\udd49" );
		code( R.drawable.emj_1f549_fe0f, "\ud83d\udd49\ufe0f" );
		code( R.drawable.emj_1f54a_fe0f, "\ud83d\udd4a" );
		code( R.drawable.emj_1f54a_fe0f, "\ud83d\udd4a\ufe0f" );
		code( R.drawable.emj_1f54b, "\ud83d\udd4b" );
		code( R.drawable.emj_1f54c, "\ud83d\udd4c" );
		code( R.drawable.emj_1f54d, "\ud83d\udd4d" );
		code( R.drawable.emj_1f54e, "\ud83d\udd4e" );
		code( R.drawable.emj_1f550, "\ud83d\udd50" );
		code( R.drawable.emj_1f550, "\ue024" );
		code( R.drawable.emj_1f550, "\ue594" );
		code( R.drawable.emj_1f550, "\udbb8\udc1e" );
		code( R.drawable.emj_1f551, "\ud83d\udd51" );
		code( R.drawable.emj_1f551, "\ue025" );
		code( R.drawable.emj_1f551, "\ue594" );
		code( R.drawable.emj_1f551, "\udbb8\udc1f" );
		code( R.drawable.emj_1f552, "\ud83d\udd52" );
		code( R.drawable.emj_1f552, "\ue026" );
		code( R.drawable.emj_1f552, "\ue594" );
		code( R.drawable.emj_1f552, "\udbb8\udc20" );
		code( R.drawable.emj_1f553, "\ud83d\udd53" );
		code( R.drawable.emj_1f553, "\ue027" );
		code( R.drawable.emj_1f553, "\ue594" );
		code( R.drawable.emj_1f553, "\udbb8\udc21" );
		code( R.drawable.emj_1f554, "\ud83d\udd54" );
		code( R.drawable.emj_1f554, "\ue028" );
		code( R.drawable.emj_1f554, "\ue594" );
		code( R.drawable.emj_1f554, "\udbb8\udc22" );
		code( R.drawable.emj_1f555, "\ud83d\udd55" );
		code( R.drawable.emj_1f555, "\ue029" );
		code( R.drawable.emj_1f555, "\ue594" );
		code( R.drawable.emj_1f555, "\udbb8\udc23" );
		code( R.drawable.emj_1f556, "\ud83d\udd56" );
		code( R.drawable.emj_1f556, "\ue02a" );
		code( R.drawable.emj_1f556, "\ue594" );
		code( R.drawable.emj_1f556, "\udbb8\udc24" );
		code( R.drawable.emj_1f557, "\ud83d\udd57" );
		code( R.drawable.emj_1f557, "\ue02b" );
		code( R.drawable.emj_1f557, "\ue594" );
		code( R.drawable.emj_1f557, "\udbb8\udc25" );
		code( R.drawable.emj_1f558, "\ud83d\udd58" );
		code( R.drawable.emj_1f558, "\ue02c" );
		code( R.drawable.emj_1f558, "\ue594" );
		code( R.drawable.emj_1f558, "\udbb8\udc26" );
		code( R.drawable.emj_1f559, "\ud83d\udd59" );
		code( R.drawable.emj_1f559, "\ue02d" );
		code( R.drawable.emj_1f559, "\ue594" );
	}
	private static void init35(){
		code( R.drawable.emj_1f559, "\udbb8\udc27" );
		code( R.drawable.emj_1f55a, "\ud83d\udd5a" );
		code( R.drawable.emj_1f55a, "\ue02e" );
		code( R.drawable.emj_1f55a, "\ue594" );
		code( R.drawable.emj_1f55a, "\udbb8\udc28" );
		code( R.drawable.emj_1f55b, "\ud83d\udd5b" );
		code( R.drawable.emj_1f55b, "\ue02f" );
		code( R.drawable.emj_1f55b, "\ue594" );
		code( R.drawable.emj_1f55b, "\udbb8\udc29" );
		code( R.drawable.emj_1f55c, "\ud83d\udd5c" );
		code( R.drawable.emj_1f55d, "\ud83d\udd5d" );
		code( R.drawable.emj_1f55e, "\ud83d\udd5e" );
		code( R.drawable.emj_1f55f, "\ud83d\udd5f" );
		code( R.drawable.emj_1f560, "\ud83d\udd60" );
		code( R.drawable.emj_1f561, "\ud83d\udd61" );
		code( R.drawable.emj_1f562, "\ud83d\udd62" );
		code( R.drawable.emj_1f563, "\ud83d\udd63" );
		code( R.drawable.emj_1f564, "\ud83d\udd64" );
		code( R.drawable.emj_1f565, "\ud83d\udd65" );
		code( R.drawable.emj_1f566, "\ud83d\udd66" );
		code( R.drawable.emj_1f567, "\ud83d\udd67" );
		code( R.drawable.emj_1f56f_fe0f, "\ud83d\udd6f" );
		code( R.drawable.emj_1f56f_fe0f, "\ud83d\udd6f\ufe0f" );
		code( R.drawable.emj_1f570_fe0f, "\ud83d\udd70" );
		code( R.drawable.emj_1f570_fe0f, "\ud83d\udd70\ufe0f" );
		code( R.drawable.emj_1f573_fe0f, "\ud83d\udd73" );
		code( R.drawable.emj_1f573_fe0f, "\ud83d\udd73\ufe0f" );
		code( R.drawable.emj_1f574_1f3fb, "\ud83d\udd74\ud83c\udffb" );
		code( R.drawable.emj_1f574_1f3fc, "\ud83d\udd74\ud83c\udffc" );
		code( R.drawable.emj_1f574_1f3fd, "\ud83d\udd74\ud83c\udffd" );
		code( R.drawable.emj_1f574_1f3fe, "\ud83d\udd74\ud83c\udffe" );
		code( R.drawable.emj_1f574_1f3ff, "\ud83d\udd74\ud83c\udfff" );
		code( R.drawable.emj_1f574_fe0f, "\ud83d\udd74" );
		code( R.drawable.emj_1f574_fe0f, "\ud83d\udd74\ufe0f" );
		code( R.drawable.emj_1f575_1f3fb, "\ud83d\udd75\ud83c\udffb" );
		code( R.drawable.emj_1f575_1f3fb_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f575_1f3fb_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f575_1f3fb_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f575_1f3fb_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f575_1f3fc, "\ud83d\udd75\ud83c\udffc" );
		code( R.drawable.emj_1f575_1f3fc_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f575_1f3fc_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f575_1f3fc_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f575_1f3fc_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f575_1f3fd, "\ud83d\udd75\ud83c\udffd" );
		code( R.drawable.emj_1f575_1f3fd_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f575_1f3fd_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f575_1f3fd_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f575_1f3fd_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f575_1f3fe, "\ud83d\udd75\ud83c\udffe" );
		code( R.drawable.emj_1f575_1f3fe_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f575_1f3fe_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f575_1f3fe_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f575_1f3fe_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f575_1f3ff, "\ud83d\udd75\ud83c\udfff" );
		code( R.drawable.emj_1f575_1f3ff_200d_2640_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f575_1f3ff_200d_2640_fe0f, "\ud83d\udd75\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f575_1f3ff_200d_2642_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f575_1f3ff_200d_2642_fe0f, "\ud83d\udd75\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f575_fe0f, "\ud83d\udd75" );
		code( R.drawable.emj_1f575_fe0f, "\ud83d\udd75\ufe0f" );
		code( R.drawable.emj_1f575_fe0f_200d_2640_fe0f, "\ud83d\udd75\u2640" );
		code( R.drawable.emj_1f575_fe0f_200d_2640_fe0f, "\ud83d\udd75\ufe0f\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f575_fe0f_200d_2642_fe0f, "\ud83d\udd75\u2642" );
		code( R.drawable.emj_1f575_fe0f_200d_2642_fe0f, "\ud83d\udd75\ufe0f\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f576_fe0f, "\ud83d\udd76" );
		code( R.drawable.emj_1f576_fe0f, "\ud83d\udd76\ufe0f" );
		code( R.drawable.emj_1f577_fe0f, "\ud83d\udd77" );
		code( R.drawable.emj_1f577_fe0f, "\ud83d\udd77\ufe0f" );
		code( R.drawable.emj_1f578_fe0f, "\ud83d\udd78" );
		code( R.drawable.emj_1f578_fe0f, "\ud83d\udd78\ufe0f" );
		code( R.drawable.emj_1f579_fe0f, "\ud83d\udd79" );
		code( R.drawable.emj_1f579_fe0f, "\ud83d\udd79\ufe0f" );
		code( R.drawable.emj_1f57a, "\ud83d\udd7a" );
		code( R.drawable.emj_1f57a_1f3fb, "\ud83d\udd7a\ud83c\udffb" );
		code( R.drawable.emj_1f57a_1f3fc, "\ud83d\udd7a\ud83c\udffc" );
		code( R.drawable.emj_1f57a_1f3fd, "\ud83d\udd7a\ud83c\udffd" );
		code( R.drawable.emj_1f57a_1f3fe, "\ud83d\udd7a\ud83c\udffe" );
		code( R.drawable.emj_1f57a_1f3ff, "\ud83d\udd7a\ud83c\udfff" );
		code( R.drawable.emj_1f587_fe0f, "\ud83d\udd87" );
		code( R.drawable.emj_1f587_fe0f, "\ud83d\udd87\ufe0f" );
		code( R.drawable.emj_1f58a_fe0f, "\ud83d\udd8a" );
		code( R.drawable.emj_1f58a_fe0f, "\ud83d\udd8a\ufe0f" );
		code( R.drawable.emj_1f58b_fe0f, "\ud83d\udd8b" );
		code( R.drawable.emj_1f58b_fe0f, "\ud83d\udd8b\ufe0f" );
		code( R.drawable.emj_1f58c_fe0f, "\ud83d\udd8c" );
		code( R.drawable.emj_1f58c_fe0f, "\ud83d\udd8c\ufe0f" );
		code( R.drawable.emj_1f58d_fe0f, "\ud83d\udd8d" );
		code( R.drawable.emj_1f58d_fe0f, "\ud83d\udd8d\ufe0f" );
		code( R.drawable.emj_1f590_1f3fb, "\ud83d\udd90\ud83c\udffb" );
		code( R.drawable.emj_1f590_1f3fc, "\ud83d\udd90\ud83c\udffc" );
		code( R.drawable.emj_1f590_1f3fd, "\ud83d\udd90\ud83c\udffd" );
		code( R.drawable.emj_1f590_1f3fe, "\ud83d\udd90\ud83c\udffe" );
		code( R.drawable.emj_1f590_1f3ff, "\ud83d\udd90\ud83c\udfff" );
		code( R.drawable.emj_1f590_fe0f, "\ud83d\udd90" );
		code( R.drawable.emj_1f590_fe0f, "\ud83d\udd90\ufe0f" );
		code( R.drawable.emj_1f595, "\ud83d\udd95" );
		code( R.drawable.emj_1f595_1f3fb, "\ud83d\udd95\ud83c\udffb" );
		code( R.drawable.emj_1f595_1f3fc, "\ud83d\udd95\ud83c\udffc" );
		code( R.drawable.emj_1f595_1f3fd, "\ud83d\udd95\ud83c\udffd" );
		code( R.drawable.emj_1f595_1f3fe, "\ud83d\udd95\ud83c\udffe" );
	}
	private static void init36(){
		code( R.drawable.emj_1f595_1f3ff, "\ud83d\udd95\ud83c\udfff" );
		code( R.drawable.emj_1f596, "\ud83d\udd96" );
		code( R.drawable.emj_1f596_1f3fb, "\ud83d\udd96\ud83c\udffb" );
		code( R.drawable.emj_1f596_1f3fc, "\ud83d\udd96\ud83c\udffc" );
		code( R.drawable.emj_1f596_1f3fd, "\ud83d\udd96\ud83c\udffd" );
		code( R.drawable.emj_1f596_1f3fe, "\ud83d\udd96\ud83c\udffe" );
		code( R.drawable.emj_1f596_1f3ff, "\ud83d\udd96\ud83c\udfff" );
		code( R.drawable.emj_1f5a4, "\ud83d\udda4" );
		code( R.drawable.emj_1f5a5_fe0f, "\ud83d\udda5" );
		code( R.drawable.emj_1f5a5_fe0f, "\ud83d\udda5\ufe0f" );
		code( R.drawable.emj_1f5a8_fe0f, "\ud83d\udda8" );
		code( R.drawable.emj_1f5a8_fe0f, "\ud83d\udda8\ufe0f" );
		code( R.drawable.emj_1f5b1_fe0f, "\ud83d\uddb1" );
		code( R.drawable.emj_1f5b1_fe0f, "\ud83d\uddb1\ufe0f" );
		code( R.drawable.emj_1f5b2_fe0f, "\ud83d\uddb2" );
		code( R.drawable.emj_1f5b2_fe0f, "\ud83d\uddb2\ufe0f" );
		code( R.drawable.emj_1f5bc_fe0f, "\ud83d\uddbc" );
		code( R.drawable.emj_1f5bc_fe0f, "\ud83d\uddbc\ufe0f" );
		code( R.drawable.emj_1f5c2_fe0f, "\ud83d\uddc2" );
		code( R.drawable.emj_1f5c2_fe0f, "\ud83d\uddc2\ufe0f" );
		code( R.drawable.emj_1f5c3_fe0f, "\ud83d\uddc3" );
		code( R.drawable.emj_1f5c3_fe0f, "\ud83d\uddc3\ufe0f" );
		code( R.drawable.emj_1f5c4_fe0f, "\ud83d\uddc4" );
		code( R.drawable.emj_1f5c4_fe0f, "\ud83d\uddc4\ufe0f" );
		code( R.drawable.emj_1f5d1_fe0f, "\ud83d\uddd1" );
		code( R.drawable.emj_1f5d1_fe0f, "\ud83d\uddd1\ufe0f" );
		code( R.drawable.emj_1f5d2_fe0f, "\ud83d\uddd2" );
		code( R.drawable.emj_1f5d2_fe0f, "\ud83d\uddd2\ufe0f" );
		code( R.drawable.emj_1f5d3_fe0f, "\ud83d\uddd3" );
		code( R.drawable.emj_1f5d3_fe0f, "\ud83d\uddd3\ufe0f" );
		code( R.drawable.emj_1f5dc_fe0f, "\ud83d\udddc" );
		code( R.drawable.emj_1f5dc_fe0f, "\ud83d\udddc\ufe0f" );
		code( R.drawable.emj_1f5dd_fe0f, "\ud83d\udddd" );
		code( R.drawable.emj_1f5dd_fe0f, "\ud83d\udddd\ufe0f" );
		code( R.drawable.emj_1f5de_fe0f, "\ud83d\uddde" );
		code( R.drawable.emj_1f5de_fe0f, "\ud83d\uddde\ufe0f" );
		code( R.drawable.emj_1f5e1_fe0f, "\ud83d\udde1" );
		code( R.drawable.emj_1f5e1_fe0f, "\ud83d\udde1\ufe0f" );
		code( R.drawable.emj_1f5e3_fe0f, "\ud83d\udde3" );
		code( R.drawable.emj_1f5e3_fe0f, "\ud83d\udde3\ufe0f" );
		code( R.drawable.emj_1f5e8_fe0f, "\ud83d\udde8" );
		code( R.drawable.emj_1f5e8_fe0f, "\ud83d\udde8\ufe0f" );
		code( R.drawable.emj_1f5ef_fe0f, "\ud83d\uddef" );
		code( R.drawable.emj_1f5ef_fe0f, "\ud83d\uddef\ufe0f" );
		code( R.drawable.emj_1f5f3_fe0f, "\ud83d\uddf3" );
		code( R.drawable.emj_1f5f3_fe0f, "\ud83d\uddf3\ufe0f" );
		code( R.drawable.emj_1f5fa_fe0f, "\ud83d\uddfa" );
		code( R.drawable.emj_1f5fa_fe0f, "\ud83d\uddfa\ufe0f" );
		code( R.drawable.emj_1f5fb, "\ud83d\uddfb" );
		code( R.drawable.emj_1f5fb, "\ue03b" );
		code( R.drawable.emj_1f5fb, "\ue5bd" );
		code( R.drawable.emj_1f5fb, "\ue740" );
		code( R.drawable.emj_1f5fb, "\udbb9\udcc3" );
		code( R.drawable.emj_1f5fc, "\ud83d\uddfc" );
		code( R.drawable.emj_1f5fc, "\ue4c0" );
		code( R.drawable.emj_1f5fc, "\ue509" );
		code( R.drawable.emj_1f5fc, "\udbb9\udcc4" );
		code( R.drawable.emj_1f5fd, "\ud83d\uddfd" );
		code( R.drawable.emj_1f5fd, "\ue51d" );
		code( R.drawable.emj_1f5fd, "\udbb9\udcc6" );
		code( R.drawable.emj_1f5fe, "\ud83d\uddfe" );
		code( R.drawable.emj_1f5fe, "\ue572" );
		code( R.drawable.emj_1f5fe, "\udbb9\udcc7" );
		code( R.drawable.emj_1f5ff, "\ud83d\uddff" );
		code( R.drawable.emj_1f5ff, "\ueb6c" );
		code( R.drawable.emj_1f5ff, "\udbb9\udcc8" );
		code( R.drawable.emj_1f600, "\ud83d\ude00" );
		code( R.drawable.emj_1f601, "\ud83d\ude01" );
		code( R.drawable.emj_1f601, "\ue404" );
		code( R.drawable.emj_1f601, "\ue753" );
		code( R.drawable.emj_1f601, "\ueb80" );
		code( R.drawable.emj_1f601, "\udbb8\udf33" );
		code( R.drawable.emj_1f602, "\ud83d\ude02" );
		code( R.drawable.emj_1f602, "\ue412" );
		code( R.drawable.emj_1f602, "\ueb64" );
		code( R.drawable.emj_1f602, "\udbb8\udf34" );
		code( R.drawable.emj_1f603, "\ud83d\ude03" );
		code( R.drawable.emj_1f603, "\ue057" );
		code( R.drawable.emj_1f603, "\ue471" );
		code( R.drawable.emj_1f603, "\udbb8\udf30" );
		code( R.drawable.emj_1f604, "\ud83d\ude04" );
		code( R.drawable.emj_1f604, "\ue415" );
		code( R.drawable.emj_1f604, "\ue471" );
		code( R.drawable.emj_1f604, "\ue6f0" );
		code( R.drawable.emj_1f604, "\udbb8\udf38" );
		code( R.drawable.emj_1f605, "\ud83d\ude05" );
		code( R.drawable.emj_1f605, "\ue471\ue5b1" );
		code( R.drawable.emj_1f605, "\ue722" );
		code( R.drawable.emj_1f605, "\udbb8\udf31" );
		code( R.drawable.emj_1f606, "\ud83d\ude06" );
		code( R.drawable.emj_1f606, "\ue72a" );
		code( R.drawable.emj_1f606, "\udbb8\udf32" );
		code( R.drawable.emj_1f607, "\ud83d\ude07" );
		code( R.drawable.emj_1f608, "\ud83d\ude08" );
		code( R.drawable.emj_1f609, "\ud83d\ude09" );
		code( R.drawable.emj_1f609, "\ue405" );
		code( R.drawable.emj_1f609, "\ue5c3" );
		code( R.drawable.emj_1f609, "\ue729" );
		code( R.drawable.emj_1f609, "\udbb8\udf47" );
		code( R.drawable.emj_1f60a, "\ud83d\ude0a" );
		code( R.drawable.emj_1f60a, "\ue056" );
	}
	private static void init37(){
		code( R.drawable.emj_1f60a, "\ueacd" );
		code( R.drawable.emj_1f60a, "\udbb8\udf35" );
		code( R.drawable.emj_1f60b, "\ud83d\ude0b" );
		code( R.drawable.emj_1f60b, "\ue752" );
		code( R.drawable.emj_1f60b, "\udbb8\udf2b" );
		code( R.drawable.emj_1f60c, "\ud83d\ude0c" );
		code( R.drawable.emj_1f60c, "\ue40a" );
		code( R.drawable.emj_1f60c, "\ue721" );
		code( R.drawable.emj_1f60c, "\ueac5" );
		code( R.drawable.emj_1f60c, "\udbb8\udf3e" );
		code( R.drawable.emj_1f60d, "\ud83d\ude0d" );
		code( R.drawable.emj_1f60d, "\ue106" );
		code( R.drawable.emj_1f60d, "\ue5c4" );
		code( R.drawable.emj_1f60d, "\ue726" );
		code( R.drawable.emj_1f60d, "\udbb8\udf27" );
		code( R.drawable.emj_1f60e, "\ud83d\ude0e" );
		code( R.drawable.emj_1f60f, "\ud83d\ude0f" );
		code( R.drawable.emj_1f60f, "\ue402" );
		code( R.drawable.emj_1f60f, "\ue72c" );
		code( R.drawable.emj_1f60f, "\ueabf" );
		code( R.drawable.emj_1f60f, "\udbb8\udf43" );
		code( R.drawable.emj_1f610, "\ud83d\ude10" );
		code( R.drawable.emj_1f611, "\ud83d\ude11" );
		code( R.drawable.emj_1f612, "\ud83d\ude12" );
		code( R.drawable.emj_1f612, "\ue40e" );
		code( R.drawable.emj_1f612, "\ue725" );
		code( R.drawable.emj_1f612, "\ueac9" );
		code( R.drawable.emj_1f612, "\udbb8\udf26" );
		code( R.drawable.emj_1f613, "\ud83d\ude13" );
		code( R.drawable.emj_1f613, "\ue108" );
		code( R.drawable.emj_1f613, "\ue5c6" );
		code( R.drawable.emj_1f613, "\ue723" );
		code( R.drawable.emj_1f613, "\udbb8\udf44" );
		code( R.drawable.emj_1f614, "\ud83d\ude14" );
		code( R.drawable.emj_1f614, "\ue403" );
		code( R.drawable.emj_1f614, "\ue720" );
		code( R.drawable.emj_1f614, "\udbb8\udf40" );
		code( R.drawable.emj_1f615, "\ud83d\ude15" );
		code( R.drawable.emj_1f616, "\ud83d\ude16" );
		code( R.drawable.emj_1f616, "\ue407" );
		code( R.drawable.emj_1f616, "\ue6f3" );
		code( R.drawable.emj_1f616, "\ueac3" );
		code( R.drawable.emj_1f616, "\udbb8\udf3f" );
		code( R.drawable.emj_1f617, "\ud83d\ude17" );
		code( R.drawable.emj_1f618, "\ud83d\ude18" );
		code( R.drawable.emj_1f618, "\ue418" );
		code( R.drawable.emj_1f618, "\ueacf" );
		code( R.drawable.emj_1f618, "\udbb8\udf2c" );
		code( R.drawable.emj_1f619, "\ud83d\ude19" );
		code( R.drawable.emj_1f61a, "\ud83d\ude1a" );
		code( R.drawable.emj_1f61a, "\ue417" );
		code( R.drawable.emj_1f61a, "\ueace" );
		code( R.drawable.emj_1f61a, "\udbb8\udf2d" );
		code( R.drawable.emj_1f61b, "\ud83d\ude1b" );
		code( R.drawable.emj_1f61c, "\ud83d\ude1c" );
		code( R.drawable.emj_1f61c, "\ue105" );
		code( R.drawable.emj_1f61c, "\ue4e7" );
		code( R.drawable.emj_1f61c, "\ue728" );
		code( R.drawable.emj_1f61c, "\udbb8\udf29" );
		code( R.drawable.emj_1f61d, "\ud83d\ude1d" );
		code( R.drawable.emj_1f61d, "\ue409" );
		code( R.drawable.emj_1f61d, "\ue4e7" );
		code( R.drawable.emj_1f61d, "\udbb8\udf2a" );
		code( R.drawable.emj_1f61e, "\ud83d\ude1e" );
		code( R.drawable.emj_1f61e, "\ue058" );
		code( R.drawable.emj_1f61e, "\ue6f2" );
		code( R.drawable.emj_1f61e, "\ueac0" );
		code( R.drawable.emj_1f61e, "\udbb8\udf23" );
		code( R.drawable.emj_1f61f, "\ud83d\ude1f" );
		code( R.drawable.emj_1f620, "\ud83d\ude20" );
		code( R.drawable.emj_1f620, "\ue059" );
		code( R.drawable.emj_1f620, "\ue472" );
		code( R.drawable.emj_1f620, "\ue6f1" );
		code( R.drawable.emj_1f620, "\udbb8\udf20" );
		code( R.drawable.emj_1f621, "\ud83d\ude21" );
		code( R.drawable.emj_1f621, "\ue416" );
		code( R.drawable.emj_1f621, "\ue724" );
		code( R.drawable.emj_1f621, "\ueb5d" );
		code( R.drawable.emj_1f621, "\udbb8\udf3d" );
		code( R.drawable.emj_1f622, "\ud83d\ude22" );
		code( R.drawable.emj_1f622, "\ue413" );
		code( R.drawable.emj_1f622, "\ue72e" );
		code( R.drawable.emj_1f622, "\ueb69" );
		code( R.drawable.emj_1f622, "\udbb8\udf39" );
		code( R.drawable.emj_1f623, "\ud83d\ude23" );
		code( R.drawable.emj_1f623, "\ue406" );
		code( R.drawable.emj_1f623, "\ue72b" );
		code( R.drawable.emj_1f623, "\ueac2" );
		code( R.drawable.emj_1f623, "\udbb8\udf3c" );
		code( R.drawable.emj_1f624, "\ud83d\ude24" );
		code( R.drawable.emj_1f624, "\ueac1" );
		code( R.drawable.emj_1f624, "\udbb8\udf28" );
		code( R.drawable.emj_1f625, "\ud83d\ude25" );
		code( R.drawable.emj_1f625, "\ue401" );
		code( R.drawable.emj_1f625, "\ue5c6" );
		code( R.drawable.emj_1f625, "\udbb8\udf45" );
		code( R.drawable.emj_1f626, "\ud83d\ude26" );
		code( R.drawable.emj_1f627, "\ud83d\ude27" );
		code( R.drawable.emj_1f628, "\ud83d\ude28" );
		code( R.drawable.emj_1f628, "\ue40b" );
		code( R.drawable.emj_1f628, "\ueac6" );
	}
	private static void init38(){
		code( R.drawable.emj_1f628, "\udbb8\udf3b" );
		code( R.drawable.emj_1f629, "\ud83d\ude29" );
		code( R.drawable.emj_1f629, "\ueb67" );
		code( R.drawable.emj_1f629, "\udbb8\udf21" );
		code( R.drawable.emj_1f62a, "\ud83d\ude2a" );
		code( R.drawable.emj_1f62a, "\ue408" );
		code( R.drawable.emj_1f62a, "\ueac4" );
		code( R.drawable.emj_1f62a, "\udbb8\udf42" );
		code( R.drawable.emj_1f62b, "\ud83d\ude2b" );
		code( R.drawable.emj_1f62b, "\ue474" );
		code( R.drawable.emj_1f62b, "\udbb8\udf46" );
		code( R.drawable.emj_1f62c, "\ud83d\ude2c" );
		code( R.drawable.emj_1f62d, "\ud83d\ude2d" );
		code( R.drawable.emj_1f62d, "\ue411" );
		code( R.drawable.emj_1f62d, "\ue473" );
		code( R.drawable.emj_1f62d, "\ue72d" );
		code( R.drawable.emj_1f62d, "\udbb8\udf3a" );
		code( R.drawable.emj_1f62e, "\ud83d\ude2e" );
		code( R.drawable.emj_1f62f, "\ud83d\ude2f" );
		code( R.drawable.emj_1f630, "\ud83d\ude30" );
		code( R.drawable.emj_1f630, "\ue40f" );
		code( R.drawable.emj_1f630, "\ueacb" );
		code( R.drawable.emj_1f630, "\udbb8\udf25" );
		code( R.drawable.emj_1f631, "\ud83d\ude31" );
		code( R.drawable.emj_1f631, "\ue107" );
		code( R.drawable.emj_1f631, "\ue5c5" );
		code( R.drawable.emj_1f631, "\ue757" );
		code( R.drawable.emj_1f631, "\udbb8\udf41" );
		code( R.drawable.emj_1f632, "\ud83d\ude32" );
		code( R.drawable.emj_1f632, "\ue410" );
		code( R.drawable.emj_1f632, "\ueaca" );
		code( R.drawable.emj_1f632, "\udbb8\udf22" );
		code( R.drawable.emj_1f633, "\ud83d\ude33" );
		code( R.drawable.emj_1f633, "\ue40d" );
		code( R.drawable.emj_1f633, "\ueac8" );
		code( R.drawable.emj_1f633, "\udbb8\udf2f" );
		code( R.drawable.emj_1f634, "\ud83d\ude34" );
		code( R.drawable.emj_1f635, "\ud83d\ude35" );
		code( R.drawable.emj_1f635, "\ue5ae" );
		code( R.drawable.emj_1f635, "\ue6f4" );
		code( R.drawable.emj_1f635, "\udbb8\udf24" );
		code( R.drawable.emj_1f636, "\ud83d\ude36" );
		code( R.drawable.emj_1f637, "\ud83d\ude37" );
		code( R.drawable.emj_1f637, "\ue40c" );
		code( R.drawable.emj_1f637, "\ueac7" );
		code( R.drawable.emj_1f637, "\udbb8\udf2e" );
		code( R.drawable.emj_1f638, "\ud83d\ude38" );
		code( R.drawable.emj_1f638, "\ueb7f" );
		code( R.drawable.emj_1f638, "\udbb8\udf49" );
		code( R.drawable.emj_1f639, "\ud83d\ude39" );
		code( R.drawable.emj_1f639, "\ueb63" );
		code( R.drawable.emj_1f639, "\udbb8\udf4a" );
		code( R.drawable.emj_1f63a, "\ud83d\ude3a" );
		code( R.drawable.emj_1f63a, "\ueb61" );
		code( R.drawable.emj_1f63a, "\udbb8\udf48" );
		code( R.drawable.emj_1f63b, "\ud83d\ude3b" );
		code( R.drawable.emj_1f63b, "\ueb65" );
		code( R.drawable.emj_1f63b, "\udbb8\udf4c" );
		code( R.drawable.emj_1f63c, "\ud83d\ude3c" );
		code( R.drawable.emj_1f63c, "\ueb6a" );
		code( R.drawable.emj_1f63c, "\udbb8\udf4f" );
		code( R.drawable.emj_1f63d, "\ud83d\ude3d" );
		code( R.drawable.emj_1f63d, "\ueb60" );
		code( R.drawable.emj_1f63d, "\udbb8\udf4b" );
		code( R.drawable.emj_1f63e, "\ud83d\ude3e" );
		code( R.drawable.emj_1f63e, "\ueb5e" );
		code( R.drawable.emj_1f63e, "\udbb8\udf4e" );
		code( R.drawable.emj_1f63f, "\ud83d\ude3f" );
		code( R.drawable.emj_1f63f, "\ueb68" );
		code( R.drawable.emj_1f63f, "\udbb8\udf4d" );
		code( R.drawable.emj_1f640, "\ud83d\ude40" );
		code( R.drawable.emj_1f640, "\ueb66" );
		code( R.drawable.emj_1f640, "\udbb8\udf50" );
		code( R.drawable.emj_1f641, "\ud83d\ude41" );
		code( R.drawable.emj_1f642, "\ud83d\ude42" );
		code( R.drawable.emj_1f643, "\ud83d\ude43" );
		code( R.drawable.emj_1f644, "\ud83d\ude44" );
		code( R.drawable.emj_1f645, "\ud83d\ude45" );
		code( R.drawable.emj_1f645, "\ue423" );
		code( R.drawable.emj_1f645, "\uead7" );
		code( R.drawable.emj_1f645, "\udbb8\udf51" );
		code( R.drawable.emj_1f645_1f3fb, "\ud83d\ude45\ud83c\udffb" );
		code( R.drawable.emj_1f645_1f3fb_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f645_1f3fb_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f645_1f3fb_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f645_1f3fb_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f645_1f3fc, "\ud83d\ude45\ud83c\udffc" );
		code( R.drawable.emj_1f645_1f3fc_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f645_1f3fc_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f645_1f3fc_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f645_1f3fc_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f645_1f3fd, "\ud83d\ude45\ud83c\udffd" );
		code( R.drawable.emj_1f645_1f3fd_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f645_1f3fd_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f645_1f3fd_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f645_1f3fd_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f645_1f3fe, "\ud83d\ude45\ud83c\udffe" );
		code( R.drawable.emj_1f645_1f3fe_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f645_1f3fe_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f645_1f3fe_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f645_1f3fe_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffe\u2642" );
	}
	private static void init39(){
		code( R.drawable.emj_1f645_1f3ff, "\ud83d\ude45\ud83c\udfff" );
		code( R.drawable.emj_1f645_1f3ff_200d_2640_fe0f, "\ud83d\ude45\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f645_1f3ff_200d_2640_fe0f, "\ud83d\ude45\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f645_1f3ff_200d_2642_fe0f, "\ud83d\ude45\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f645_1f3ff_200d_2642_fe0f, "\ud83d\ude45\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f645_200d_2640_fe0f, "\ud83d\ude45\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f645_200d_2640_fe0f, "\ud83d\ude45\u2640" );
		code( R.drawable.emj_1f645_200d_2642_fe0f, "\ud83d\ude45\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f645_200d_2642_fe0f, "\ud83d\ude45\u2642" );
		code( R.drawable.emj_1f646, "\ud83d\ude46" );
		code( R.drawable.emj_1f646, "\ue424" );
		code( R.drawable.emj_1f646, "\uead8" );
		code( R.drawable.emj_1f646, "\udbb8\udf52" );
		code( R.drawable.emj_1f646_1f3fb, "\ud83d\ude46\ud83c\udffb" );
		code( R.drawable.emj_1f646_1f3fb_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f646_1f3fb_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f646_1f3fb_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f646_1f3fb_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f646_1f3fc, "\ud83d\ude46\ud83c\udffc" );
		code( R.drawable.emj_1f646_1f3fc_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f646_1f3fc_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f646_1f3fc_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f646_1f3fc_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f646_1f3fd, "\ud83d\ude46\ud83c\udffd" );
		code( R.drawable.emj_1f646_1f3fd_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f646_1f3fd_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f646_1f3fd_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f646_1f3fd_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f646_1f3fe, "\ud83d\ude46\ud83c\udffe" );
		code( R.drawable.emj_1f646_1f3fe_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f646_1f3fe_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f646_1f3fe_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f646_1f3fe_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f646_1f3ff, "\ud83d\ude46\ud83c\udfff" );
		code( R.drawable.emj_1f646_1f3ff_200d_2640_fe0f, "\ud83d\ude46\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f646_1f3ff_200d_2640_fe0f, "\ud83d\ude46\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f646_1f3ff_200d_2642_fe0f, "\ud83d\ude46\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f646_1f3ff_200d_2642_fe0f, "\ud83d\ude46\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f646_200d_2640_fe0f, "\ud83d\ude46\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f646_200d_2640_fe0f, "\ud83d\ude46\u2640" );
		code( R.drawable.emj_1f646_200d_2642_fe0f, "\ud83d\ude46\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f646_200d_2642_fe0f, "\ud83d\ude46\u2642" );
		code( R.drawable.emj_1f647, "\ud83d\ude47" );
		code( R.drawable.emj_1f647, "\ue426" );
		code( R.drawable.emj_1f647, "\uead9" );
		code( R.drawable.emj_1f647, "\udbb8\udf53" );
		code( R.drawable.emj_1f647_1f3fb, "\ud83d\ude47\ud83c\udffb" );
		code( R.drawable.emj_1f647_1f3fb_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f647_1f3fb_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f647_1f3fb_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f647_1f3fb_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f647_1f3fc, "\ud83d\ude47\ud83c\udffc" );
		code( R.drawable.emj_1f647_1f3fc_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f647_1f3fc_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f647_1f3fc_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f647_1f3fc_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f647_1f3fd, "\ud83d\ude47\ud83c\udffd" );
		code( R.drawable.emj_1f647_1f3fd_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f647_1f3fd_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f647_1f3fd_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f647_1f3fd_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f647_1f3fe, "\ud83d\ude47\ud83c\udffe" );
		code( R.drawable.emj_1f647_1f3fe_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f647_1f3fe_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f647_1f3fe_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f647_1f3fe_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f647_1f3ff, "\ud83d\ude47\ud83c\udfff" );
		code( R.drawable.emj_1f647_1f3ff_200d_2640_fe0f, "\ud83d\ude47\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f647_1f3ff_200d_2640_fe0f, "\ud83d\ude47\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f647_1f3ff_200d_2642_fe0f, "\ud83d\ude47\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f647_1f3ff_200d_2642_fe0f, "\ud83d\ude47\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f647_200d_2640_fe0f, "\ud83d\ude47\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f647_200d_2640_fe0f, "\ud83d\ude47\u2640" );
		code( R.drawable.emj_1f647_200d_2642_fe0f, "\ud83d\ude47\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f647_200d_2642_fe0f, "\ud83d\ude47\u2642" );
		code( R.drawable.emj_1f648, "\ud83d\ude48" );
		code( R.drawable.emj_1f648, "\ueb50" );
		code( R.drawable.emj_1f648, "\udbb8\udf54" );
		code( R.drawable.emj_1f649, "\ud83d\ude49" );
		code( R.drawable.emj_1f649, "\ueb52" );
		code( R.drawable.emj_1f649, "\udbb8\udf56" );
		code( R.drawable.emj_1f64a, "\ud83d\ude4a" );
		code( R.drawable.emj_1f64a, "\ueb51" );
		code( R.drawable.emj_1f64a, "\udbb8\udf55" );
		code( R.drawable.emj_1f64b, "\ud83d\ude4b" );
		code( R.drawable.emj_1f64b, "\ueb85" );
		code( R.drawable.emj_1f64b, "\udbb8\udf57" );
		code( R.drawable.emj_1f64b_1f3fb, "\ud83d\ude4b\ud83c\udffb" );
		code( R.drawable.emj_1f64b_1f3fb_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64b_1f3fb_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f64b_1f3fb_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64b_1f3fb_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f64b_1f3fc, "\ud83d\ude4b\ud83c\udffc" );
		code( R.drawable.emj_1f64b_1f3fc_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64b_1f3fc_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f64b_1f3fc_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64b_1f3fc_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f64b_1f3fd, "\ud83d\ude4b\ud83c\udffd" );
		code( R.drawable.emj_1f64b_1f3fd_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64b_1f3fd_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f64b_1f3fd_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffd\u200d\u2642\ufe0f" );
	}
	private static void init40(){
		code( R.drawable.emj_1f64b_1f3fd_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f64b_1f3fe, "\ud83d\ude4b\ud83c\udffe" );
		code( R.drawable.emj_1f64b_1f3fe_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64b_1f3fe_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f64b_1f3fe_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64b_1f3fe_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f64b_1f3ff, "\ud83d\ude4b\ud83c\udfff" );
		code( R.drawable.emj_1f64b_1f3ff_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64b_1f3ff_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f64b_1f3ff_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64b_1f3ff_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f64b_200d_2640_fe0f, "\ud83d\ude4b\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64b_200d_2640_fe0f, "\ud83d\ude4b\u2640" );
		code( R.drawable.emj_1f64b_200d_2642_fe0f, "\ud83d\ude4b\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64b_200d_2642_fe0f, "\ud83d\ude4b\u2642" );
		code( R.drawable.emj_1f64c, "\ud83d\ude4c" );
		code( R.drawable.emj_1f64c, "\ue427" );
		code( R.drawable.emj_1f64c, "\ueb86" );
		code( R.drawable.emj_1f64c, "\udbb8\udf58" );
		code( R.drawable.emj_1f64c_1f3fb, "\ud83d\ude4c\ud83c\udffb" );
		code( R.drawable.emj_1f64c_1f3fc, "\ud83d\ude4c\ud83c\udffc" );
		code( R.drawable.emj_1f64c_1f3fd, "\ud83d\ude4c\ud83c\udffd" );
		code( R.drawable.emj_1f64c_1f3fe, "\ud83d\ude4c\ud83c\udffe" );
		code( R.drawable.emj_1f64c_1f3ff, "\ud83d\ude4c\ud83c\udfff" );
		code( R.drawable.emj_1f64d, "\ud83d\ude4d" );
		code( R.drawable.emj_1f64d, "\ueb87" );
		code( R.drawable.emj_1f64d, "\udbb8\udf59" );
		code( R.drawable.emj_1f64d_1f3fb, "\ud83d\ude4d\ud83c\udffb" );
		code( R.drawable.emj_1f64d_1f3fb_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fb_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f64d_1f3fb_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fb_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f64d_1f3fc, "\ud83d\ude4d\ud83c\udffc" );
		code( R.drawable.emj_1f64d_1f3fc_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fc_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f64d_1f3fc_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fc_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f64d_1f3fd, "\ud83d\ude4d\ud83c\udffd" );
		code( R.drawable.emj_1f64d_1f3fd_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fd_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f64d_1f3fd_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fd_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f64d_1f3fe, "\ud83d\ude4d\ud83c\udffe" );
		code( R.drawable.emj_1f64d_1f3fe_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fe_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f64d_1f3fe_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64d_1f3fe_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f64d_1f3ff, "\ud83d\ude4d\ud83c\udfff" );
		code( R.drawable.emj_1f64d_1f3ff_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64d_1f3ff_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f64d_1f3ff_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64d_1f3ff_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f64d_200d_2640_fe0f, "\ud83d\ude4d\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64d_200d_2640_fe0f, "\ud83d\ude4d\u2640" );
		code( R.drawable.emj_1f64d_200d_2642_fe0f, "\ud83d\ude4d\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64d_200d_2642_fe0f, "\ud83d\ude4d\u2642" );
		code( R.drawable.emj_1f64e, "\ud83d\ude4e" );
		code( R.drawable.emj_1f64e, "\ueb88" );
		code( R.drawable.emj_1f64e, "\udbb8\udf5a" );
		code( R.drawable.emj_1f64e_1f3fb, "\ud83d\ude4e\ud83c\udffb" );
		code( R.drawable.emj_1f64e_1f3fb_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fb_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f64e_1f3fb_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fb_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f64e_1f3fc, "\ud83d\ude4e\ud83c\udffc" );
		code( R.drawable.emj_1f64e_1f3fc_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fc_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f64e_1f3fc_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fc_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f64e_1f3fd, "\ud83d\ude4e\ud83c\udffd" );
		code( R.drawable.emj_1f64e_1f3fd_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fd_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f64e_1f3fd_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fd_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f64e_1f3fe, "\ud83d\ude4e\ud83c\udffe" );
		code( R.drawable.emj_1f64e_1f3fe_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fe_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f64e_1f3fe_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64e_1f3fe_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f64e_1f3ff, "\ud83d\ude4e\ud83c\udfff" );
		code( R.drawable.emj_1f64e_1f3ff_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64e_1f3ff_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f64e_1f3ff_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64e_1f3ff_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f64e_200d_2640_fe0f, "\ud83d\ude4e\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f64e_200d_2640_fe0f, "\ud83d\ude4e\u2640" );
		code( R.drawable.emj_1f64e_200d_2642_fe0f, "\ud83d\ude4e\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f64e_200d_2642_fe0f, "\ud83d\ude4e\u2642" );
		code( R.drawable.emj_1f64f, "\ud83d\ude4f" );
		code( R.drawable.emj_1f64f, "\ue41d" );
		code( R.drawable.emj_1f64f, "\uead2" );
		code( R.drawable.emj_1f64f, "\udbb8\udf5b" );
		code( R.drawable.emj_1f64f_1f3fb, "\ud83d\ude4f\ud83c\udffb" );
		code( R.drawable.emj_1f64f_1f3fc, "\ud83d\ude4f\ud83c\udffc" );
		code( R.drawable.emj_1f64f_1f3fd, "\ud83d\ude4f\ud83c\udffd" );
		code( R.drawable.emj_1f64f_1f3fe, "\ud83d\ude4f\ud83c\udffe" );
		code( R.drawable.emj_1f64f_1f3ff, "\ud83d\ude4f\ud83c\udfff" );
		code( R.drawable.emj_1f680, "\ud83d\ude80" );
		code( R.drawable.emj_1f680, "\ue10d" );
		code( R.drawable.emj_1f680, "\ue5c8" );
		code( R.drawable.emj_1f680, "\udbb9\udfed" );
	}
	private static void init41(){
		code( R.drawable.emj_1f681, "\ud83d\ude81" );
		code( R.drawable.emj_1f682, "\ud83d\ude82" );
		code( R.drawable.emj_1f683, "\ud83d\ude83" );
		code( R.drawable.emj_1f683, "\ue01e" );
		code( R.drawable.emj_1f683, "\ue4b5" );
		code( R.drawable.emj_1f683, "\ue65b" );
		code( R.drawable.emj_1f683, "\udbb9\udfdf" );
		code( R.drawable.emj_1f684, "\ud83d\ude84" );
		code( R.drawable.emj_1f684, "\ue435" );
		code( R.drawable.emj_1f684, "\ue4b0" );
		code( R.drawable.emj_1f684, "\ue65d" );
		code( R.drawable.emj_1f684, "\udbb9\udfe2" );
		code( R.drawable.emj_1f685, "\ud83d\ude85" );
		code( R.drawable.emj_1f685, "\ue01f" );
		code( R.drawable.emj_1f685, "\ue4b0" );
		code( R.drawable.emj_1f685, "\ue65d" );
		code( R.drawable.emj_1f685, "\udbb9\udfe3" );
		code( R.drawable.emj_1f686, "\ud83d\ude86" );
		code( R.drawable.emj_1f687, "\ud83d\ude87" );
		code( R.drawable.emj_1f687, "\ue434" );
		code( R.drawable.emj_1f687, "\ue5bc" );
		code( R.drawable.emj_1f687, "\ue65c" );
		code( R.drawable.emj_1f687, "\udbb9\udfe0" );
		code( R.drawable.emj_1f688, "\ud83d\ude88" );
		code( R.drawable.emj_1f689, "\ud83d\ude89" );
		code( R.drawable.emj_1f689, "\ue039" );
		code( R.drawable.emj_1f689, "\ueb6d" );
		code( R.drawable.emj_1f689, "\udbb9\udfec" );
		code( R.drawable.emj_1f68a, "\ud83d\ude8a" );
		code( R.drawable.emj_1f68b, "\ud83d\ude8b" );
		code( R.drawable.emj_1f68c, "\ud83d\ude8c" );
		code( R.drawable.emj_1f68c, "\ue159" );
		code( R.drawable.emj_1f68c, "\ue4af" );
		code( R.drawable.emj_1f68c, "\ue660" );
		code( R.drawable.emj_1f68c, "\udbb9\udfe6" );
		code( R.drawable.emj_1f68d, "\ud83d\ude8d" );
		code( R.drawable.emj_1f68e, "\ud83d\ude8e" );
		code( R.drawable.emj_1f68f, "\ud83d\ude8f" );
		code( R.drawable.emj_1f68f, "\ue150" );
		code( R.drawable.emj_1f68f, "\ue4a7" );
		code( R.drawable.emj_1f68f, "\udbb9\udfe7" );
		code( R.drawable.emj_1f690, "\ud83d\ude90" );
		code( R.drawable.emj_1f691, "\ud83d\ude91" );
		code( R.drawable.emj_1f691, "\ue431" );
		code( R.drawable.emj_1f691, "\ueae0" );
		code( R.drawable.emj_1f691, "\udbb9\udff3" );
		code( R.drawable.emj_1f692, "\ud83d\ude92" );
		code( R.drawable.emj_1f692, "\ue430" );
		code( R.drawable.emj_1f692, "\ueadf" );
		code( R.drawable.emj_1f692, "\udbb9\udff2" );
		code( R.drawable.emj_1f693, "\ud83d\ude93" );
		code( R.drawable.emj_1f693, "\ue432" );
		code( R.drawable.emj_1f693, "\ueae1" );
		code( R.drawable.emj_1f693, "\udbb9\udff4" );
		code( R.drawable.emj_1f694, "\ud83d\ude94" );
		code( R.drawable.emj_1f695, "\ud83d\ude95" );
		code( R.drawable.emj_1f695, "\ue15a" );
		code( R.drawable.emj_1f695, "\ue4b1" );
		code( R.drawable.emj_1f695, "\ue65e" );
		code( R.drawable.emj_1f695, "\udbb9\udfef" );
		code( R.drawable.emj_1f696, "\ud83d\ude96" );
		code( R.drawable.emj_1f697, "\ud83d\ude97" );
		code( R.drawable.emj_1f697, "\ue01b" );
		code( R.drawable.emj_1f697, "\ue4b1" );
		code( R.drawable.emj_1f697, "\ue65e" );
		code( R.drawable.emj_1f697, "\udbb9\udfe4" );
		code( R.drawable.emj_1f698, "\ud83d\ude98" );
		code( R.drawable.emj_1f699, "\ud83d\ude99" );
		code( R.drawable.emj_1f699, "\ue42e" );
		code( R.drawable.emj_1f699, "\ue4b1" );
		code( R.drawable.emj_1f699, "\ue65f" );
		code( R.drawable.emj_1f699, "\udbb9\udfe5" );
		code( R.drawable.emj_1f69a, "\ud83d\ude9a" );
		code( R.drawable.emj_1f69a, "\ue42f" );
		code( R.drawable.emj_1f69a, "\ue4b2" );
		code( R.drawable.emj_1f69a, "\udbb9\udff1" );
		code( R.drawable.emj_1f69b, "\ud83d\ude9b" );
		code( R.drawable.emj_1f69c, "\ud83d\ude9c" );
		code( R.drawable.emj_1f69d, "\ud83d\ude9d" );
		code( R.drawable.emj_1f69e, "\ud83d\ude9e" );
		code( R.drawable.emj_1f69f, "\ud83d\ude9f" );
		code( R.drawable.emj_1f6a0, "\ud83d\udea0" );
		code( R.drawable.emj_1f6a1, "\ud83d\udea1" );
		code( R.drawable.emj_1f6a2, "\ud83d\udea2" );
		code( R.drawable.emj_1f6a2, "\ue202" );
		code( R.drawable.emj_1f6a2, "\ue661" );
		code( R.drawable.emj_1f6a2, "\uea82" );
		code( R.drawable.emj_1f6a2, "\udbb9\udfe8" );
		code( R.drawable.emj_1f6a3, "\ud83d\udea3" );
		code( R.drawable.emj_1f6a3_1f3fb, "\ud83d\udea3\ud83c\udffb" );
		code( R.drawable.emj_1f6a3_1f3fb_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3fb_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f6a3_1f3fb_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3fb_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f6a3_1f3fc, "\ud83d\udea3\ud83c\udffc" );
		code( R.drawable.emj_1f6a3_1f3fc_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3fc_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f6a3_1f3fc_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3fc_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f6a3_1f3fd, "\ud83d\udea3\ud83c\udffd" );
		code( R.drawable.emj_1f6a3_1f3fd_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffd\u200d\u2640\ufe0f" );
	}
	private static void init42(){
		code( R.drawable.emj_1f6a3_1f3fd_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f6a3_1f3fd_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3fd_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f6a3_1f3fe, "\ud83d\udea3\ud83c\udffe" );
		code( R.drawable.emj_1f6a3_1f3fe_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3fe_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f6a3_1f3fe_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3fe_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f6a3_1f3ff, "\ud83d\udea3\ud83c\udfff" );
		code( R.drawable.emj_1f6a3_1f3ff_200d_2640_fe0f, "\ud83d\udea3\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3ff_200d_2640_fe0f, "\ud83d\udea3\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f6a3_1f3ff_200d_2642_fe0f, "\ud83d\udea3\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6a3_1f3ff_200d_2642_fe0f, "\ud83d\udea3\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f6a3_200d_2640_fe0f, "\ud83d\udea3\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6a3_200d_2640_fe0f, "\ud83d\udea3\u2640" );
		code( R.drawable.emj_1f6a3_200d_2642_fe0f, "\ud83d\udea3\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6a3_200d_2642_fe0f, "\ud83d\udea3\u2642" );
		code( R.drawable.emj_1f6a4, "\ud83d\udea4" );
		code( R.drawable.emj_1f6a4, "\ue135" );
		code( R.drawable.emj_1f6a4, "\ue4b4" );
		code( R.drawable.emj_1f6a4, "\udbb9\udfee" );
		code( R.drawable.emj_1f6a5, "\ud83d\udea5" );
		code( R.drawable.emj_1f6a5, "\ue14e" );
		code( R.drawable.emj_1f6a5, "\ue46a" );
		code( R.drawable.emj_1f6a5, "\ue66d" );
		code( R.drawable.emj_1f6a5, "\udbb9\udff7" );
		code( R.drawable.emj_1f6a6, "\ud83d\udea6" );
		code( R.drawable.emj_1f6a7, "\ud83d\udea7" );
		code( R.drawable.emj_1f6a7, "\ue137" );
		code( R.drawable.emj_1f6a7, "\ue5d7" );
		code( R.drawable.emj_1f6a7, "\udbb9\udff8" );
		code( R.drawable.emj_1f6a8, "\ud83d\udea8" );
		code( R.drawable.emj_1f6a8, "\ueb73" );
		code( R.drawable.emj_1f6a8, "\udbb9\udff9" );
		code( R.drawable.emj_1f6a9, "\ud83d\udea9" );
		code( R.drawable.emj_1f6a9, "\ue6de" );
		code( R.drawable.emj_1f6a9, "\ueb2c" );
		code( R.drawable.emj_1f6a9, "\udbba\udf22" );
		code( R.drawable.emj_1f6aa, "\ud83d\udeaa" );
		code( R.drawable.emj_1f6aa, "\ue714" );
		code( R.drawable.emj_1f6aa, "\udbb9\udcf3" );
		code( R.drawable.emj_1f6ab, "\ud83d\udeab" );
		code( R.drawable.emj_1f6ab, "\ue541" );
		code( R.drawable.emj_1f6ab, "\udbba\udf48" );
		code( R.drawable.emj_1f6ac, "\ud83d\udeac" );
		code( R.drawable.emj_1f6ac, "\ue30e" );
		code( R.drawable.emj_1f6ac, "\ue47d" );
		code( R.drawable.emj_1f6ac, "\ue67f" );
		code( R.drawable.emj_1f6ac, "\udbba\udf1e" );
		code( R.drawable.emj_1f6ad, "\ud83d\udead" );
		code( R.drawable.emj_1f6ad, "\ue208" );
		code( R.drawable.emj_1f6ad, "\ue47e" );
		code( R.drawable.emj_1f6ad, "\ue680" );
		code( R.drawable.emj_1f6ad, "\udbba\udf1f" );
		code( R.drawable.emj_1f6ae, "\ud83d\udeae" );
		code( R.drawable.emj_1f6af, "\ud83d\udeaf" );
		code( R.drawable.emj_1f6b0, "\ud83d\udeb0" );
		code( R.drawable.emj_1f6b1, "\ud83d\udeb1" );
		code( R.drawable.emj_1f6b2, "\ud83d\udeb2" );
		code( R.drawable.emj_1f6b2, "\ue136" );
		code( R.drawable.emj_1f6b2, "\ue4ae" );
		code( R.drawable.emj_1f6b2, "\ue71d" );
		code( R.drawable.emj_1f6b2, "\udbb9\udfeb" );
		code( R.drawable.emj_1f6b3, "\ud83d\udeb3" );
		code( R.drawable.emj_1f6b4, "\ud83d\udeb4" );
		code( R.drawable.emj_1f6b4_1f3fb, "\ud83d\udeb4\ud83c\udffb" );
		code( R.drawable.emj_1f6b4_1f3fb_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fb_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f6b4_1f3fb_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fb_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f6b4_1f3fc, "\ud83d\udeb4\ud83c\udffc" );
		code( R.drawable.emj_1f6b4_1f3fc_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fc_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f6b4_1f3fc_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fc_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f6b4_1f3fd, "\ud83d\udeb4\ud83c\udffd" );
		code( R.drawable.emj_1f6b4_1f3fd_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fd_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f6b4_1f3fd_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fd_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f6b4_1f3fe, "\ud83d\udeb4\ud83c\udffe" );
		code( R.drawable.emj_1f6b4_1f3fe_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fe_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f6b4_1f3fe_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3fe_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f6b4_1f3ff, "\ud83d\udeb4\ud83c\udfff" );
		code( R.drawable.emj_1f6b4_1f3ff_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3ff_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f6b4_1f3ff_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b4_1f3ff_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f6b4_200d_2640_fe0f, "\ud83d\udeb4\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b4_200d_2640_fe0f, "\ud83d\udeb4\u2640" );
		code( R.drawable.emj_1f6b4_200d_2642_fe0f, "\ud83d\udeb4\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b4_200d_2642_fe0f, "\ud83d\udeb4\u2642" );
		code( R.drawable.emj_1f6b5, "\ud83d\udeb5" );
		code( R.drawable.emj_1f6b5_1f3fb, "\ud83d\udeb5\ud83c\udffb" );
		code( R.drawable.emj_1f6b5_1f3fb_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fb_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f6b5_1f3fb_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fb_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f6b5_1f3fc, "\ud83d\udeb5\ud83c\udffc" );
	}
	private static void init43(){
		code( R.drawable.emj_1f6b5_1f3fc_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fc_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f6b5_1f3fc_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fc_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f6b5_1f3fd, "\ud83d\udeb5\ud83c\udffd" );
		code( R.drawable.emj_1f6b5_1f3fd_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fd_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f6b5_1f3fd_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fd_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f6b5_1f3fe, "\ud83d\udeb5\ud83c\udffe" );
		code( R.drawable.emj_1f6b5_1f3fe_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fe_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f6b5_1f3fe_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3fe_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f6b5_1f3ff, "\ud83d\udeb5\ud83c\udfff" );
		code( R.drawable.emj_1f6b5_1f3ff_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3ff_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f6b5_1f3ff_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b5_1f3ff_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f6b5_200d_2640_fe0f, "\ud83d\udeb5\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b5_200d_2640_fe0f, "\ud83d\udeb5\u2640" );
		code( R.drawable.emj_1f6b5_200d_2642_fe0f, "\ud83d\udeb5\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b5_200d_2642_fe0f, "\ud83d\udeb5\u2642" );
		code( R.drawable.emj_1f6b6, "\ud83d\udeb6" );
		code( R.drawable.emj_1f6b6, "\ue201" );
		code( R.drawable.emj_1f6b6, "\ueb72" );
		code( R.drawable.emj_1f6b6, "\udbb9\udff0" );
		code( R.drawable.emj_1f6b6_1f3fb, "\ud83d\udeb6\ud83c\udffb" );
		code( R.drawable.emj_1f6b6_1f3fb_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fb_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f6b6_1f3fb_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fb_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f6b6_1f3fc, "\ud83d\udeb6\ud83c\udffc" );
		code( R.drawable.emj_1f6b6_1f3fc_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fc_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f6b6_1f3fc_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fc_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f6b6_1f3fd, "\ud83d\udeb6\ud83c\udffd" );
		code( R.drawable.emj_1f6b6_1f3fd_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fd_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f6b6_1f3fd_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fd_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f6b6_1f3fe, "\ud83d\udeb6\ud83c\udffe" );
		code( R.drawable.emj_1f6b6_1f3fe_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fe_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f6b6_1f3fe_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3fe_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f6b6_1f3ff, "\ud83d\udeb6\ud83c\udfff" );
		code( R.drawable.emj_1f6b6_1f3ff_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3ff_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f6b6_1f3ff_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b6_1f3ff_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f6b6_200d_2640_fe0f, "\ud83d\udeb6\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f6b6_200d_2640_fe0f, "\ud83d\udeb6\u2640" );
		code( R.drawable.emj_1f6b6_200d_2642_fe0f, "\ud83d\udeb6\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f6b6_200d_2642_fe0f, "\ud83d\udeb6\u2642" );
		code( R.drawable.emj_1f6b7, "\ud83d\udeb7" );
		code( R.drawable.emj_1f6b8, "\ud83d\udeb8" );
		code( R.drawable.emj_1f6b9, "\ud83d\udeb9" );
		code( R.drawable.emj_1f6b9, "\ue138" );
		code( R.drawable.emj_1f6b9, "\udbba\udf33" );
		code( R.drawable.emj_1f6ba, "\ud83d\udeba" );
		code( R.drawable.emj_1f6ba, "\ue139" );
		code( R.drawable.emj_1f6ba, "\udbba\udf34" );
		code( R.drawable.emj_1f6bb, "\ud83d\udebb" );
		code( R.drawable.emj_1f6bb, "\ue151" );
		code( R.drawable.emj_1f6bb, "\ue4a5" );
		code( R.drawable.emj_1f6bb, "\ue66e" );
		code( R.drawable.emj_1f6bb, "\udbb9\udd06" );
		code( R.drawable.emj_1f6bc, "\ud83d\udebc" );
		code( R.drawable.emj_1f6bc, "\ue13a" );
		code( R.drawable.emj_1f6bc, "\udbba\udf35" );
		code( R.drawable.emj_1f6bd, "\ud83d\udebd" );
		code( R.drawable.emj_1f6bd, "\ue140" );
		code( R.drawable.emj_1f6bd, "\ue4a5" );
		code( R.drawable.emj_1f6bd, "\ue66e" );
		code( R.drawable.emj_1f6bd, "\udbb9\udd07" );
		code( R.drawable.emj_1f6be, "\ud83d\udebe" );
		code( R.drawable.emj_1f6be, "\ue309" );
		code( R.drawable.emj_1f6be, "\ue4a5" );
		code( R.drawable.emj_1f6be, "\ue66e" );
		code( R.drawable.emj_1f6be, "\udbb9\udd08" );
		code( R.drawable.emj_1f6bf, "\ud83d\udebf" );
		code( R.drawable.emj_1f6c0, "\ud83d\udec0" );
		code( R.drawable.emj_1f6c0, "\ue13f" );
		code( R.drawable.emj_1f6c0, "\ue5d8" );
		code( R.drawable.emj_1f6c0, "\udbb9\udd05" );
		code( R.drawable.emj_1f6c0_1f3fb, "\ud83d\udec0\ud83c\udffb" );
		code( R.drawable.emj_1f6c0_1f3fc, "\ud83d\udec0\ud83c\udffc" );
		code( R.drawable.emj_1f6c0_1f3fd, "\ud83d\udec0\ud83c\udffd" );
		code( R.drawable.emj_1f6c0_1f3fe, "\ud83d\udec0\ud83c\udffe" );
		code( R.drawable.emj_1f6c0_1f3ff, "\ud83d\udec0\ud83c\udfff" );
		code( R.drawable.emj_1f6c1, "\ud83d\udec1" );
		code( R.drawable.emj_1f6c2, "\ud83d\udec2" );
		code( R.drawable.emj_1f6c3, "\ud83d\udec3" );
		code( R.drawable.emj_1f6c4, "\ud83d\udec4" );
		code( R.drawable.emj_1f6c5, "\ud83d\udec5" );
		code( R.drawable.emj_1f6cb_fe0f, "\ud83d\udecb" );
		code( R.drawable.emj_1f6cb_fe0f, "\ud83d\udecb\ufe0f" );
		code( R.drawable.emj_1f6cc, "\ud83d\udecc" );
		code( R.drawable.emj_1f6cc_1f3fb, "\ud83d\udecc\ud83c\udffb" );
	}
	private static void init44(){
		code( R.drawable.emj_1f6cc_1f3fc, "\ud83d\udecc\ud83c\udffc" );
		code( R.drawable.emj_1f6cc_1f3fd, "\ud83d\udecc\ud83c\udffd" );
		code( R.drawable.emj_1f6cc_1f3fe, "\ud83d\udecc\ud83c\udffe" );
		code( R.drawable.emj_1f6cc_1f3ff, "\ud83d\udecc\ud83c\udfff" );
		code( R.drawable.emj_1f6cd_fe0f, "\ud83d\udecd" );
		code( R.drawable.emj_1f6cd_fe0f, "\ud83d\udecd\ufe0f" );
		code( R.drawable.emj_1f6ce_fe0f, "\ud83d\udece" );
		code( R.drawable.emj_1f6ce_fe0f, "\ud83d\udece\ufe0f" );
		code( R.drawable.emj_1f6cf_fe0f, "\ud83d\udecf" );
		code( R.drawable.emj_1f6cf_fe0f, "\ud83d\udecf\ufe0f" );
		code( R.drawable.emj_1f6d0, "\ud83d\uded0" );
		code( R.drawable.emj_1f6d1, "\ud83d\uded1" );
		code( R.drawable.emj_1f6d2, "\ud83d\uded2" );
		code( R.drawable.emj_1f6e0_fe0f, "\ud83d\udee0" );
		code( R.drawable.emj_1f6e0_fe0f, "\ud83d\udee0\ufe0f" );
		code( R.drawable.emj_1f6e1_fe0f, "\ud83d\udee1" );
		code( R.drawable.emj_1f6e1_fe0f, "\ud83d\udee1\ufe0f" );
		code( R.drawable.emj_1f6e2_fe0f, "\ud83d\udee2" );
		code( R.drawable.emj_1f6e2_fe0f, "\ud83d\udee2\ufe0f" );
		code( R.drawable.emj_1f6e3_fe0f, "\ud83d\udee3" );
		code( R.drawable.emj_1f6e3_fe0f, "\ud83d\udee3\ufe0f" );
		code( R.drawable.emj_1f6e4_fe0f, "\ud83d\udee4" );
		code( R.drawable.emj_1f6e4_fe0f, "\ud83d\udee4\ufe0f" );
		code( R.drawable.emj_1f6e5_fe0f, "\ud83d\udee5" );
		code( R.drawable.emj_1f6e5_fe0f, "\ud83d\udee5\ufe0f" );
		code( R.drawable.emj_1f6e9_fe0f, "\ud83d\udee9" );
		code( R.drawable.emj_1f6e9_fe0f, "\ud83d\udee9\ufe0f" );
		code( R.drawable.emj_1f6eb, "\ud83d\udeeb" );
		code( R.drawable.emj_1f6ec, "\ud83d\udeec" );
		code( R.drawable.emj_1f6f0_fe0f, "\ud83d\udef0" );
		code( R.drawable.emj_1f6f0_fe0f, "\ud83d\udef0\ufe0f" );
		code( R.drawable.emj_1f6f3_fe0f, "\ud83d\udef3" );
		code( R.drawable.emj_1f6f3_fe0f, "\ud83d\udef3\ufe0f" );
		code( R.drawable.emj_1f6f4, "\ud83d\udef4" );
		code( R.drawable.emj_1f6f5, "\ud83d\udef5" );
		code( R.drawable.emj_1f6f6, "\ud83d\udef6" );
		code( R.drawable.emj_1f6f7, "\ud83d\udef7" );
		code( R.drawable.emj_1f6f8, "\ud83d\udef8" );
		code( R.drawable.emj_1f910, "\ud83e\udd10" );
		code( R.drawable.emj_1f911, "\ud83e\udd11" );
		code( R.drawable.emj_1f912, "\ud83e\udd12" );
		code( R.drawable.emj_1f913, "\ud83e\udd13" );
		code( R.drawable.emj_1f914, "\ud83e\udd14" );
		code( R.drawable.emj_1f915, "\ud83e\udd15" );
		code( R.drawable.emj_1f916, "\ud83e\udd16" );
		code( R.drawable.emj_1f917, "\ud83e\udd17" );
		code( R.drawable.emj_1f918, "\ud83e\udd18" );
		code( R.drawable.emj_1f918_1f3fb, "\ud83e\udd18\ud83c\udffb" );
		code( R.drawable.emj_1f918_1f3fc, "\ud83e\udd18\ud83c\udffc" );
		code( R.drawable.emj_1f918_1f3fd, "\ud83e\udd18\ud83c\udffd" );
		code( R.drawable.emj_1f918_1f3fe, "\ud83e\udd18\ud83c\udffe" );
		code( R.drawable.emj_1f918_1f3ff, "\ud83e\udd18\ud83c\udfff" );
		code( R.drawable.emj_1f919, "\ud83e\udd19" );
		code( R.drawable.emj_1f919_1f3fb, "\ud83e\udd19\ud83c\udffb" );
		code( R.drawable.emj_1f919_1f3fc, "\ud83e\udd19\ud83c\udffc" );
		code( R.drawable.emj_1f919_1f3fd, "\ud83e\udd19\ud83c\udffd" );
		code( R.drawable.emj_1f919_1f3fe, "\ud83e\udd19\ud83c\udffe" );
		code( R.drawable.emj_1f919_1f3ff, "\ud83e\udd19\ud83c\udfff" );
		code( R.drawable.emj_1f91a, "\ud83e\udd1a" );
		code( R.drawable.emj_1f91a_1f3fb, "\ud83e\udd1a\ud83c\udffb" );
		code( R.drawable.emj_1f91a_1f3fc, "\ud83e\udd1a\ud83c\udffc" );
		code( R.drawable.emj_1f91a_1f3fd, "\ud83e\udd1a\ud83c\udffd" );
		code( R.drawable.emj_1f91a_1f3fe, "\ud83e\udd1a\ud83c\udffe" );
		code( R.drawable.emj_1f91a_1f3ff, "\ud83e\udd1a\ud83c\udfff" );
		code( R.drawable.emj_1f91b, "\ud83e\udd1b" );
		code( R.drawable.emj_1f91b_1f3fb, "\ud83e\udd1b\ud83c\udffb" );
		code( R.drawable.emj_1f91b_1f3fc, "\ud83e\udd1b\ud83c\udffc" );
		code( R.drawable.emj_1f91b_1f3fd, "\ud83e\udd1b\ud83c\udffd" );
		code( R.drawable.emj_1f91b_1f3fe, "\ud83e\udd1b\ud83c\udffe" );
		code( R.drawable.emj_1f91b_1f3ff, "\ud83e\udd1b\ud83c\udfff" );
		code( R.drawable.emj_1f91c, "\ud83e\udd1c" );
		code( R.drawable.emj_1f91c_1f3fb, "\ud83e\udd1c\ud83c\udffb" );
		code( R.drawable.emj_1f91c_1f3fc, "\ud83e\udd1c\ud83c\udffc" );
		code( R.drawable.emj_1f91c_1f3fd, "\ud83e\udd1c\ud83c\udffd" );
		code( R.drawable.emj_1f91c_1f3fe, "\ud83e\udd1c\ud83c\udffe" );
		code( R.drawable.emj_1f91c_1f3ff, "\ud83e\udd1c\ud83c\udfff" );
		code( R.drawable.emj_1f91d, "\ud83e\udd1d" );
		code( R.drawable.emj_1f91e, "\ud83e\udd1e" );
		code( R.drawable.emj_1f91e_1f3fb, "\ud83e\udd1e\ud83c\udffb" );
		code( R.drawable.emj_1f91e_1f3fc, "\ud83e\udd1e\ud83c\udffc" );
		code( R.drawable.emj_1f91e_1f3fd, "\ud83e\udd1e\ud83c\udffd" );
		code( R.drawable.emj_1f91e_1f3fe, "\ud83e\udd1e\ud83c\udffe" );
		code( R.drawable.emj_1f91e_1f3ff, "\ud83e\udd1e\ud83c\udfff" );
		code( R.drawable.emj_1f91f, "\ud83e\udd1f" );
		code( R.drawable.emj_1f91f_1f3fb, "\ud83e\udd1f\ud83c\udffb" );
		code( R.drawable.emj_1f91f_1f3fc, "\ud83e\udd1f\ud83c\udffc" );
		code( R.drawable.emj_1f91f_1f3fd, "\ud83e\udd1f\ud83c\udffd" );
		code( R.drawable.emj_1f91f_1f3fe, "\ud83e\udd1f\ud83c\udffe" );
		code( R.drawable.emj_1f91f_1f3ff, "\ud83e\udd1f\ud83c\udfff" );
		code( R.drawable.emj_1f920, "\ud83e\udd20" );
		code( R.drawable.emj_1f921, "\ud83e\udd21" );
		code( R.drawable.emj_1f922, "\ud83e\udd22" );
		code( R.drawable.emj_1f923, "\ud83e\udd23" );
		code( R.drawable.emj_1f924, "\ud83e\udd24" );
		code( R.drawable.emj_1f925, "\ud83e\udd25" );
		code( R.drawable.emj_1f926, "\ud83e\udd26" );
		code( R.drawable.emj_1f926_1f3fb, "\ud83e\udd26\ud83c\udffb" );
		code( R.drawable.emj_1f926_1f3fb_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f926_1f3fb_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f926_1f3fb_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f926_1f3fb_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffb\u2642" );
	}
	private static void init45(){
		code( R.drawable.emj_1f926_1f3fc, "\ud83e\udd26\ud83c\udffc" );
		code( R.drawable.emj_1f926_1f3fc_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f926_1f3fc_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f926_1f3fc_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f926_1f3fc_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f926_1f3fd, "\ud83e\udd26\ud83c\udffd" );
		code( R.drawable.emj_1f926_1f3fd_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f926_1f3fd_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f926_1f3fd_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f926_1f3fd_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f926_1f3fe, "\ud83e\udd26\ud83c\udffe" );
		code( R.drawable.emj_1f926_1f3fe_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f926_1f3fe_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f926_1f3fe_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f926_1f3fe_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f926_1f3ff, "\ud83e\udd26\ud83c\udfff" );
		code( R.drawable.emj_1f926_1f3ff_200d_2640_fe0f, "\ud83e\udd26\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f926_1f3ff_200d_2640_fe0f, "\ud83e\udd26\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f926_1f3ff_200d_2642_fe0f, "\ud83e\udd26\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f926_1f3ff_200d_2642_fe0f, "\ud83e\udd26\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f926_200d_2640_fe0f, "\ud83e\udd26\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f926_200d_2640_fe0f, "\ud83e\udd26\u2640" );
		code( R.drawable.emj_1f926_200d_2642_fe0f, "\ud83e\udd26\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f926_200d_2642_fe0f, "\ud83e\udd26\u2642" );
		code( R.drawable.emj_1f927, "\ud83e\udd27" );
		code( R.drawable.emj_1f928, "\ud83e\udd28" );
		code( R.drawable.emj_1f929, "\ud83e\udd29" );
		code( R.drawable.emj_1f92a, "\ud83e\udd2a" );
		code( R.drawable.emj_1f92b, "\ud83e\udd2b" );
		code( R.drawable.emj_1f92c, "\ud83e\udd2c" );
		code( R.drawable.emj_1f92d, "\ud83e\udd2d" );
		code( R.drawable.emj_1f92e, "\ud83e\udd2e" );
		code( R.drawable.emj_1f92f, "\ud83e\udd2f" );
		code( R.drawable.emj_1f930, "\ud83e\udd30" );
		code( R.drawable.emj_1f930_1f3fb, "\ud83e\udd30\ud83c\udffb" );
		code( R.drawable.emj_1f930_1f3fc, "\ud83e\udd30\ud83c\udffc" );
		code( R.drawable.emj_1f930_1f3fd, "\ud83e\udd30\ud83c\udffd" );
		code( R.drawable.emj_1f930_1f3fe, "\ud83e\udd30\ud83c\udffe" );
		code( R.drawable.emj_1f930_1f3ff, "\ud83e\udd30\ud83c\udfff" );
		code( R.drawable.emj_1f931, "\ud83e\udd31" );
		code( R.drawable.emj_1f931_1f3fb, "\ud83e\udd31\ud83c\udffb" );
		code( R.drawable.emj_1f931_1f3fc, "\ud83e\udd31\ud83c\udffc" );
		code( R.drawable.emj_1f931_1f3fd, "\ud83e\udd31\ud83c\udffd" );
		code( R.drawable.emj_1f931_1f3fe, "\ud83e\udd31\ud83c\udffe" );
		code( R.drawable.emj_1f931_1f3ff, "\ud83e\udd31\ud83c\udfff" );
		code( R.drawable.emj_1f932, "\ud83e\udd32" );
		code( R.drawable.emj_1f932_1f3fb, "\ud83e\udd32\ud83c\udffb" );
		code( R.drawable.emj_1f932_1f3fc, "\ud83e\udd32\ud83c\udffc" );
		code( R.drawable.emj_1f932_1f3fd, "\ud83e\udd32\ud83c\udffd" );
		code( R.drawable.emj_1f932_1f3fe, "\ud83e\udd32\ud83c\udffe" );
		code( R.drawable.emj_1f932_1f3ff, "\ud83e\udd32\ud83c\udfff" );
		code( R.drawable.emj_1f933, "\ud83e\udd33" );
		code( R.drawable.emj_1f933_1f3fb, "\ud83e\udd33\ud83c\udffb" );
		code( R.drawable.emj_1f933_1f3fc, "\ud83e\udd33\ud83c\udffc" );
		code( R.drawable.emj_1f933_1f3fd, "\ud83e\udd33\ud83c\udffd" );
		code( R.drawable.emj_1f933_1f3fe, "\ud83e\udd33\ud83c\udffe" );
		code( R.drawable.emj_1f933_1f3ff, "\ud83e\udd33\ud83c\udfff" );
		code( R.drawable.emj_1f934, "\ud83e\udd34" );
		code( R.drawable.emj_1f934_1f3fb, "\ud83e\udd34\ud83c\udffb" );
		code( R.drawable.emj_1f934_1f3fc, "\ud83e\udd34\ud83c\udffc" );
		code( R.drawable.emj_1f934_1f3fd, "\ud83e\udd34\ud83c\udffd" );
		code( R.drawable.emj_1f934_1f3fe, "\ud83e\udd34\ud83c\udffe" );
		code( R.drawable.emj_1f934_1f3ff, "\ud83e\udd34\ud83c\udfff" );
		code( R.drawable.emj_1f935, "\ud83e\udd35" );
		code( R.drawable.emj_1f935_1f3fb, "\ud83e\udd35\ud83c\udffb" );
		code( R.drawable.emj_1f935_1f3fc, "\ud83e\udd35\ud83c\udffc" );
		code( R.drawable.emj_1f935_1f3fd, "\ud83e\udd35\ud83c\udffd" );
		code( R.drawable.emj_1f935_1f3fe, "\ud83e\udd35\ud83c\udffe" );
		code( R.drawable.emj_1f935_1f3ff, "\ud83e\udd35\ud83c\udfff" );
		code( R.drawable.emj_1f936, "\ud83e\udd36" );
		code( R.drawable.emj_1f936_1f3fb, "\ud83e\udd36\ud83c\udffb" );
		code( R.drawable.emj_1f936_1f3fc, "\ud83e\udd36\ud83c\udffc" );
		code( R.drawable.emj_1f936_1f3fd, "\ud83e\udd36\ud83c\udffd" );
		code( R.drawable.emj_1f936_1f3fe, "\ud83e\udd36\ud83c\udffe" );
		code( R.drawable.emj_1f936_1f3ff, "\ud83e\udd36\ud83c\udfff" );
		code( R.drawable.emj_1f937, "\ud83e\udd37" );
		code( R.drawable.emj_1f937_1f3fb, "\ud83e\udd37\ud83c\udffb" );
		code( R.drawable.emj_1f937_1f3fb_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f937_1f3fb_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f937_1f3fb_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f937_1f3fb_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f937_1f3fc, "\ud83e\udd37\ud83c\udffc" );
		code( R.drawable.emj_1f937_1f3fc_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f937_1f3fc_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f937_1f3fc_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f937_1f3fc_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f937_1f3fd, "\ud83e\udd37\ud83c\udffd" );
		code( R.drawable.emj_1f937_1f3fd_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f937_1f3fd_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f937_1f3fd_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f937_1f3fd_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f937_1f3fe, "\ud83e\udd37\ud83c\udffe" );
		code( R.drawable.emj_1f937_1f3fe_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f937_1f3fe_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f937_1f3fe_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f937_1f3fe_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f937_1f3ff, "\ud83e\udd37\ud83c\udfff" );
		code( R.drawable.emj_1f937_1f3ff_200d_2640_fe0f, "\ud83e\udd37\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f937_1f3ff_200d_2640_fe0f, "\ud83e\udd37\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f937_1f3ff_200d_2642_fe0f, "\ud83e\udd37\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f937_1f3ff_200d_2642_fe0f, "\ud83e\udd37\ud83c\udfff\u2642" );
	}
	private static void init46(){
		code( R.drawable.emj_1f937_200d_2640_fe0f, "\ud83e\udd37\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f937_200d_2640_fe0f, "\ud83e\udd37\u2640" );
		code( R.drawable.emj_1f937_200d_2642_fe0f, "\ud83e\udd37\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f937_200d_2642_fe0f, "\ud83e\udd37\u2642" );
		code( R.drawable.emj_1f938, "\ud83e\udd38" );
		code( R.drawable.emj_1f938_1f3fb, "\ud83e\udd38\ud83c\udffb" );
		code( R.drawable.emj_1f938_1f3fb_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f938_1f3fb_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f938_1f3fb_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f938_1f3fb_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f938_1f3fc, "\ud83e\udd38\ud83c\udffc" );
		code( R.drawable.emj_1f938_1f3fc_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f938_1f3fc_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f938_1f3fc_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f938_1f3fc_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f938_1f3fd, "\ud83e\udd38\ud83c\udffd" );
		code( R.drawable.emj_1f938_1f3fd_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f938_1f3fd_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f938_1f3fd_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f938_1f3fd_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f938_1f3fe, "\ud83e\udd38\ud83c\udffe" );
		code( R.drawable.emj_1f938_1f3fe_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f938_1f3fe_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f938_1f3fe_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f938_1f3fe_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f938_1f3ff, "\ud83e\udd38\ud83c\udfff" );
		code( R.drawable.emj_1f938_1f3ff_200d_2640_fe0f, "\ud83e\udd38\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f938_1f3ff_200d_2640_fe0f, "\ud83e\udd38\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f938_1f3ff_200d_2642_fe0f, "\ud83e\udd38\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f938_1f3ff_200d_2642_fe0f, "\ud83e\udd38\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f938_200d_2640_fe0f, "\ud83e\udd38\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f938_200d_2640_fe0f, "\ud83e\udd38\u2640" );
		code( R.drawable.emj_1f938_200d_2642_fe0f, "\ud83e\udd38\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f938_200d_2642_fe0f, "\ud83e\udd38\u2642" );
		code( R.drawable.emj_1f939, "\ud83e\udd39" );
		code( R.drawable.emj_1f939_1f3fb, "\ud83e\udd39\ud83c\udffb" );
		code( R.drawable.emj_1f939_1f3fb_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f939_1f3fb_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f939_1f3fb_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f939_1f3fb_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f939_1f3fc, "\ud83e\udd39\ud83c\udffc" );
		code( R.drawable.emj_1f939_1f3fc_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f939_1f3fc_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f939_1f3fc_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f939_1f3fc_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f939_1f3fd, "\ud83e\udd39\ud83c\udffd" );
		code( R.drawable.emj_1f939_1f3fd_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f939_1f3fd_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f939_1f3fd_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f939_1f3fd_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f939_1f3fe, "\ud83e\udd39\ud83c\udffe" );
		code( R.drawable.emj_1f939_1f3fe_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f939_1f3fe_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f939_1f3fe_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f939_1f3fe_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f939_1f3ff, "\ud83e\udd39\ud83c\udfff" );
		code( R.drawable.emj_1f939_1f3ff_200d_2640_fe0f, "\ud83e\udd39\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f939_1f3ff_200d_2640_fe0f, "\ud83e\udd39\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f939_1f3ff_200d_2642_fe0f, "\ud83e\udd39\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f939_1f3ff_200d_2642_fe0f, "\ud83e\udd39\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f939_200d_2640_fe0f, "\ud83e\udd39\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f939_200d_2640_fe0f, "\ud83e\udd39\u2640" );
		code( R.drawable.emj_1f939_200d_2642_fe0f, "\ud83e\udd39\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f939_200d_2642_fe0f, "\ud83e\udd39\u2642" );
		code( R.drawable.emj_1f93a, "\ud83e\udd3a" );
		code( R.drawable.emj_1f93c, "\ud83e\udd3c" );
		code( R.drawable.emj_1f93c_200d_2640_fe0f, "\ud83e\udd3c\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93c_200d_2640_fe0f, "\ud83e\udd3c\u2640" );
		code( R.drawable.emj_1f93c_200d_2642_fe0f, "\ud83e\udd3c\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93c_200d_2642_fe0f, "\ud83e\udd3c\u2642" );
		code( R.drawable.emj_1f93d, "\ud83e\udd3d" );
		code( R.drawable.emj_1f93d_1f3fb, "\ud83e\udd3d\ud83c\udffb" );
		code( R.drawable.emj_1f93d_1f3fb_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fb_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f93d_1f3fb_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fb_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f93d_1f3fc, "\ud83e\udd3d\ud83c\udffc" );
		code( R.drawable.emj_1f93d_1f3fc_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fc_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f93d_1f3fc_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fc_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f93d_1f3fd, "\ud83e\udd3d\ud83c\udffd" );
		code( R.drawable.emj_1f93d_1f3fd_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fd_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f93d_1f3fd_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fd_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f93d_1f3fe, "\ud83e\udd3d\ud83c\udffe" );
		code( R.drawable.emj_1f93d_1f3fe_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fe_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f93d_1f3fe_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93d_1f3fe_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f93d_1f3ff, "\ud83e\udd3d\ud83c\udfff" );
		code( R.drawable.emj_1f93d_1f3ff_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93d_1f3ff_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f93d_1f3ff_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93d_1f3ff_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f93d_200d_2640_fe0f, "\ud83e\udd3d\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93d_200d_2640_fe0f, "\ud83e\udd3d\u2640" );
		code( R.drawable.emj_1f93d_200d_2642_fe0f, "\ud83e\udd3d\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93d_200d_2642_fe0f, "\ud83e\udd3d\u2642" );
		code( R.drawable.emj_1f93e, "\ud83e\udd3e" );
	}
	private static void init47(){
		code( R.drawable.emj_1f93e_1f3fb, "\ud83e\udd3e\ud83c\udffb" );
		code( R.drawable.emj_1f93e_1f3fb_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fb_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffb\u2640" );
		code( R.drawable.emj_1f93e_1f3fb_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fb_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffb\u2642" );
		code( R.drawable.emj_1f93e_1f3fc, "\ud83e\udd3e\ud83c\udffc" );
		code( R.drawable.emj_1f93e_1f3fc_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fc_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffc\u2640" );
		code( R.drawable.emj_1f93e_1f3fc_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fc_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffc\u2642" );
		code( R.drawable.emj_1f93e_1f3fd, "\ud83e\udd3e\ud83c\udffd" );
		code( R.drawable.emj_1f93e_1f3fd_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fd_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffd\u2640" );
		code( R.drawable.emj_1f93e_1f3fd_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fd_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffd\u2642" );
		code( R.drawable.emj_1f93e_1f3fe, "\ud83e\udd3e\ud83c\udffe" );
		code( R.drawable.emj_1f93e_1f3fe_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fe_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffe\u2640" );
		code( R.drawable.emj_1f93e_1f3fe_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93e_1f3fe_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffe\u2642" );
		code( R.drawable.emj_1f93e_1f3ff, "\ud83e\udd3e\ud83c\udfff" );
		code( R.drawable.emj_1f93e_1f3ff_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93e_1f3ff_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udfff\u2640" );
		code( R.drawable.emj_1f93e_1f3ff_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93e_1f3ff_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udfff\u2642" );
		code( R.drawable.emj_1f93e_200d_2640_fe0f, "\ud83e\udd3e\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f93e_200d_2640_fe0f, "\ud83e\udd3e\u2640" );
		code( R.drawable.emj_1f93e_200d_2642_fe0f, "\ud83e\udd3e\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f93e_200d_2642_fe0f, "\ud83e\udd3e\u2642" );
		code( R.drawable.emj_1f940, "\ud83e\udd40" );
		code( R.drawable.emj_1f941, "\ud83e\udd41" );
		code( R.drawable.emj_1f942, "\ud83e\udd42" );
		code( R.drawable.emj_1f943, "\ud83e\udd43" );
		code( R.drawable.emj_1f944, "\ud83e\udd44" );
		code( R.drawable.emj_1f945, "\ud83e\udd45" );
		code( R.drawable.emj_1f947, "\ud83e\udd47" );
		code( R.drawable.emj_1f948, "\ud83e\udd48" );
		code( R.drawable.emj_1f949, "\ud83e\udd49" );
		code( R.drawable.emj_1f94a, "\ud83e\udd4a" );
		code( R.drawable.emj_1f94b, "\ud83e\udd4b" );
		code( R.drawable.emj_1f94c, "\ud83e\udd4c" );
		code( R.drawable.emj_1f950, "\ud83e\udd50" );
		code( R.drawable.emj_1f951, "\ud83e\udd51" );
		code( R.drawable.emj_1f952, "\ud83e\udd52" );
		code( R.drawable.emj_1f953, "\ud83e\udd53" );
		code( R.drawable.emj_1f954, "\ud83e\udd54" );
		code( R.drawable.emj_1f955, "\ud83e\udd55" );
		code( R.drawable.emj_1f956, "\ud83e\udd56" );
		code( R.drawable.emj_1f957, "\ud83e\udd57" );
		code( R.drawable.emj_1f958, "\ud83e\udd58" );
		code( R.drawable.emj_1f959, "\ud83e\udd59" );
		code( R.drawable.emj_1f95a, "\ud83e\udd5a" );
		code( R.drawable.emj_1f95b, "\ud83e\udd5b" );
		code( R.drawable.emj_1f95c, "\ud83e\udd5c" );
		code( R.drawable.emj_1f95d, "\ud83e\udd5d" );
		code( R.drawable.emj_1f95e, "\ud83e\udd5e" );
		code( R.drawable.emj_1f95f, "\ud83e\udd5f" );
		code( R.drawable.emj_1f960, "\ud83e\udd60" );
		code( R.drawable.emj_1f961, "\ud83e\udd61" );
		code( R.drawable.emj_1f962, "\ud83e\udd62" );
		code( R.drawable.emj_1f963, "\ud83e\udd63" );
		code( R.drawable.emj_1f964, "\ud83e\udd64" );
		code( R.drawable.emj_1f965, "\ud83e\udd65" );
		code( R.drawable.emj_1f966, "\ud83e\udd66" );
		code( R.drawable.emj_1f967, "\ud83e\udd67" );
		code( R.drawable.emj_1f968, "\ud83e\udd68" );
		code( R.drawable.emj_1f969, "\ud83e\udd69" );
		code( R.drawable.emj_1f96a, "\ud83e\udd6a" );
		code( R.drawable.emj_1f96b, "\ud83e\udd6b" );
		code( R.drawable.emj_1f980, "\ud83e\udd80" );
		code( R.drawable.emj_1f981, "\ud83e\udd81" );
		code( R.drawable.emj_1f982, "\ud83e\udd82" );
		code( R.drawable.emj_1f983, "\ud83e\udd83" );
		code( R.drawable.emj_1f984, "\ud83e\udd84" );
		code( R.drawable.emj_1f985, "\ud83e\udd85" );
		code( R.drawable.emj_1f986, "\ud83e\udd86" );
		code( R.drawable.emj_1f987, "\ud83e\udd87" );
		code( R.drawable.emj_1f988, "\ud83e\udd88" );
		code( R.drawable.emj_1f989, "\ud83e\udd89" );
		code( R.drawable.emj_1f98a, "\ud83e\udd8a" );
		code( R.drawable.emj_1f98b, "\ud83e\udd8b" );
		code( R.drawable.emj_1f98c, "\ud83e\udd8c" );
		code( R.drawable.emj_1f98d, "\ud83e\udd8d" );
		code( R.drawable.emj_1f98e, "\ud83e\udd8e" );
		code( R.drawable.emj_1f98f, "\ud83e\udd8f" );
		code( R.drawable.emj_1f990, "\ud83e\udd90" );
		code( R.drawable.emj_1f991, "\ud83e\udd91" );
		code( R.drawable.emj_1f992, "\ud83e\udd92" );
		code( R.drawable.emj_1f993, "\ud83e\udd93" );
		code( R.drawable.emj_1f994, "\ud83e\udd94" );
		code( R.drawable.emj_1f995, "\ud83e\udd95" );
		code( R.drawable.emj_1f996, "\ud83e\udd96" );
		code( R.drawable.emj_1f997, "\ud83e\udd97" );
		code( R.drawable.emj_1f9c0, "\ud83e\uddc0" );
		code( R.drawable.emj_1f9d0, "\ud83e\uddd0" );
		code( R.drawable.emj_1f9d1, "\ud83e\uddd1" );
		code( R.drawable.emj_1f9d1_1f3fb, "\ud83e\uddd1\ud83c\udffb" );
		code( R.drawable.emj_1f9d1_1f3fc, "\ud83e\uddd1\ud83c\udffc" );
		code( R.drawable.emj_1f9d1_1f3fd, "\ud83e\uddd1\ud83c\udffd" );
		code( R.drawable.emj_1f9d1_1f3fe, "\ud83e\uddd1\ud83c\udffe" );
		code( R.drawable.emj_1f9d1_1f3ff, "\ud83e\uddd1\ud83c\udfff" );
	}
	private static void init48(){
		code( R.drawable.emj_1f9d2, "\ud83e\uddd2" );
		code( R.drawable.emj_1f9d2_1f3fb, "\ud83e\uddd2\ud83c\udffb" );
		code( R.drawable.emj_1f9d2_1f3fc, "\ud83e\uddd2\ud83c\udffc" );
		code( R.drawable.emj_1f9d2_1f3fd, "\ud83e\uddd2\ud83c\udffd" );
		code( R.drawable.emj_1f9d2_1f3fe, "\ud83e\uddd2\ud83c\udffe" );
		code( R.drawable.emj_1f9d2_1f3ff, "\ud83e\uddd2\ud83c\udfff" );
		code( R.drawable.emj_1f9d3, "\ud83e\uddd3" );
		code( R.drawable.emj_1f9d3_1f3fb, "\ud83e\uddd3\ud83c\udffb" );
		code( R.drawable.emj_1f9d3_1f3fc, "\ud83e\uddd3\ud83c\udffc" );
		code( R.drawable.emj_1f9d3_1f3fd, "\ud83e\uddd3\ud83c\udffd" );
		code( R.drawable.emj_1f9d3_1f3fe, "\ud83e\uddd3\ud83c\udffe" );
		code( R.drawable.emj_1f9d3_1f3ff, "\ud83e\uddd3\ud83c\udfff" );
		code( R.drawable.emj_1f9d4, "\ud83e\uddd4" );
		code( R.drawable.emj_1f9d4_1f3fb, "\ud83e\uddd4\ud83c\udffb" );
		code( R.drawable.emj_1f9d4_1f3fc, "\ud83e\uddd4\ud83c\udffc" );
		code( R.drawable.emj_1f9d4_1f3fd, "\ud83e\uddd4\ud83c\udffd" );
		code( R.drawable.emj_1f9d4_1f3fe, "\ud83e\uddd4\ud83c\udffe" );
		code( R.drawable.emj_1f9d4_1f3ff, "\ud83e\uddd4\ud83c\udfff" );
		code( R.drawable.emj_1f9d5, "\ud83e\uddd5" );
		code( R.drawable.emj_1f9d5_1f3fb, "\ud83e\uddd5\ud83c\udffb" );
		code( R.drawable.emj_1f9d5_1f3fc, "\ud83e\uddd5\ud83c\udffc" );
		code( R.drawable.emj_1f9d5_1f3fd, "\ud83e\uddd5\ud83c\udffd" );
		code( R.drawable.emj_1f9d5_1f3fe, "\ud83e\uddd5\ud83c\udffe" );
		code( R.drawable.emj_1f9d5_1f3ff, "\ud83e\uddd5\ud83c\udfff" );
		code( R.drawable.emj_1f9d6, "\ud83e\uddd6" );
		code( R.drawable.emj_1f9d6_1f3fb, "\ud83e\uddd6\ud83c\udffb" );
		code( R.drawable.emj_1f9d6_1f3fb_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3fb_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3fc, "\ud83e\uddd6\ud83c\udffc" );
		code( R.drawable.emj_1f9d6_1f3fc_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3fc_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3fd, "\ud83e\uddd6\ud83c\udffd" );
		code( R.drawable.emj_1f9d6_1f3fd_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3fd_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3fe, "\ud83e\uddd6\ud83c\udffe" );
		code( R.drawable.emj_1f9d6_1f3fe_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3fe_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3ff, "\ud83e\uddd6\ud83c\udfff" );
		code( R.drawable.emj_1f9d6_1f3ff_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d6_1f3ff_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d6_200d_2640_fe0f, "\ud83e\uddd6\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d6_200d_2642_fe0f, "\ud83e\uddd6\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d7, "\ud83e\uddd7" );
		code( R.drawable.emj_1f9d7_1f3fb, "\ud83e\uddd7\ud83c\udffb" );
		code( R.drawable.emj_1f9d7_1f3fb_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3fb_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3fc, "\ud83e\uddd7\ud83c\udffc" );
		code( R.drawable.emj_1f9d7_1f3fc_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3fc_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3fd, "\ud83e\uddd7\ud83c\udffd" );
		code( R.drawable.emj_1f9d7_1f3fd_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3fd_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3fe, "\ud83e\uddd7\ud83c\udffe" );
		code( R.drawable.emj_1f9d7_1f3fe_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3fe_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3ff, "\ud83e\uddd7\ud83c\udfff" );
		code( R.drawable.emj_1f9d7_1f3ff_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d7_1f3ff_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d7_200d_2640_fe0f, "\ud83e\uddd7\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d7_200d_2642_fe0f, "\ud83e\uddd7\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d8, "\ud83e\uddd8" );
		code( R.drawable.emj_1f9d8_1f3fb, "\ud83e\uddd8\ud83c\udffb" );
		code( R.drawable.emj_1f9d8_1f3fb_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3fb_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3fc, "\ud83e\uddd8\ud83c\udffc" );
		code( R.drawable.emj_1f9d8_1f3fc_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3fc_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3fd, "\ud83e\uddd8\ud83c\udffd" );
		code( R.drawable.emj_1f9d8_1f3fd_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3fd_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3fe, "\ud83e\uddd8\ud83c\udffe" );
		code( R.drawable.emj_1f9d8_1f3fe_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3fe_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3ff, "\ud83e\uddd8\ud83c\udfff" );
		code( R.drawable.emj_1f9d8_1f3ff_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d8_1f3ff_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d8_200d_2640_fe0f, "\ud83e\uddd8\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d8_200d_2642_fe0f, "\ud83e\uddd8\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d9, "\ud83e\uddd9" );
		code( R.drawable.emj_1f9d9_1f3fb, "\ud83e\uddd9\ud83c\udffb" );
		code( R.drawable.emj_1f9d9_1f3fb_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3fb_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3fc, "\ud83e\uddd9\ud83c\udffc" );
		code( R.drawable.emj_1f9d9_1f3fc_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3fc_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3fd, "\ud83e\uddd9\ud83c\udffd" );
		code( R.drawable.emj_1f9d9_1f3fd_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3fd_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3fe, "\ud83e\uddd9\ud83c\udffe" );
		code( R.drawable.emj_1f9d9_1f3fe_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3fe_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3ff, "\ud83e\uddd9\ud83c\udfff" );
		code( R.drawable.emj_1f9d9_1f3ff_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d9_1f3ff_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9d9_200d_2640_fe0f, "\ud83e\uddd9\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9d9_200d_2642_fe0f, "\ud83e\uddd9\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9da, "\ud83e\uddda" );
		code( R.drawable.emj_1f9da_1f3fb, "\ud83e\uddda\ud83c\udffb" );
		code( R.drawable.emj_1f9da_1f3fb_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9da_1f3fb_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9da_1f3fc, "\ud83e\uddda\ud83c\udffc" );
	}
	private static void init49(){
		code( R.drawable.emj_1f9da_1f3fc_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9da_1f3fc_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9da_1f3fd, "\ud83e\uddda\ud83c\udffd" );
		code( R.drawable.emj_1f9da_1f3fd_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9da_1f3fd_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9da_1f3fe, "\ud83e\uddda\ud83c\udffe" );
		code( R.drawable.emj_1f9da_1f3fe_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9da_1f3fe_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9da_1f3ff, "\ud83e\uddda\ud83c\udfff" );
		code( R.drawable.emj_1f9da_1f3ff_200d_2640_fe0f, "\ud83e\uddda\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9da_1f3ff_200d_2642_fe0f, "\ud83e\uddda\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9da_200d_2640_fe0f, "\ud83e\uddda\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9da_200d_2642_fe0f, "\ud83e\uddda\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9db, "\ud83e\udddb" );
		code( R.drawable.emj_1f9db_1f3fb, "\ud83e\udddb\ud83c\udffb" );
		code( R.drawable.emj_1f9db_1f3fb_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9db_1f3fb_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9db_1f3fc, "\ud83e\udddb\ud83c\udffc" );
		code( R.drawable.emj_1f9db_1f3fc_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9db_1f3fc_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9db_1f3fd, "\ud83e\udddb\ud83c\udffd" );
		code( R.drawable.emj_1f9db_1f3fd_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9db_1f3fd_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9db_1f3fe, "\ud83e\udddb\ud83c\udffe" );
		code( R.drawable.emj_1f9db_1f3fe_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9db_1f3fe_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9db_1f3ff, "\ud83e\udddb\ud83c\udfff" );
		code( R.drawable.emj_1f9db_1f3ff_200d_2640_fe0f, "\ud83e\udddb\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9db_1f3ff_200d_2642_fe0f, "\ud83e\udddb\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9db_200d_2640_fe0f, "\ud83e\udddb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9db_200d_2642_fe0f, "\ud83e\udddb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dc, "\ud83e\udddc" );
		code( R.drawable.emj_1f9dc_1f3fb, "\ud83e\udddc\ud83c\udffb" );
		code( R.drawable.emj_1f9dc_1f3fb_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3fb_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3fc, "\ud83e\udddc\ud83c\udffc" );
		code( R.drawable.emj_1f9dc_1f3fc_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3fc_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3fd, "\ud83e\udddc\ud83c\udffd" );
		code( R.drawable.emj_1f9dc_1f3fd_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3fd_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3fe, "\ud83e\udddc\ud83c\udffe" );
		code( R.drawable.emj_1f9dc_1f3fe_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3fe_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3ff, "\ud83e\udddc\ud83c\udfff" );
		code( R.drawable.emj_1f9dc_1f3ff_200d_2640_fe0f, "\ud83e\udddc\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dc_1f3ff_200d_2642_fe0f, "\ud83e\udddc\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dc_200d_2640_fe0f, "\ud83e\udddc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dc_200d_2642_fe0f, "\ud83e\udddc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dd, "\ud83e\udddd" );
		code( R.drawable.emj_1f9dd_1f3fb, "\ud83e\udddd\ud83c\udffb" );
		code( R.drawable.emj_1f9dd_1f3fb_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3fb_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3fc, "\ud83e\udddd\ud83c\udffc" );
		code( R.drawable.emj_1f9dd_1f3fc_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3fc_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3fd, "\ud83e\udddd\ud83c\udffd" );
		code( R.drawable.emj_1f9dd_1f3fd_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3fd_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3fe, "\ud83e\udddd\ud83c\udffe" );
		code( R.drawable.emj_1f9dd_1f3fe_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3fe_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3ff, "\ud83e\udddd\ud83c\udfff" );
		code( R.drawable.emj_1f9dd_1f3ff_200d_2640_fe0f, "\ud83e\udddd\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dd_1f3ff_200d_2642_fe0f, "\ud83e\udddd\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9dd_200d_2640_fe0f, "\ud83e\udddd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9dd_200d_2642_fe0f, "\ud83e\udddd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9de, "\ud83e\uddde" );
		code( R.drawable.emj_1f9de_200d_2640_fe0f, "\ud83e\uddde\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9de_200d_2642_fe0f, "\ud83e\uddde\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9df, "\ud83e\udddf" );
		code( R.drawable.emj_1f9df_200d_2640_fe0f, "\ud83e\udddf\u200d\u2640\ufe0f" );
		code( R.drawable.emj_1f9df_200d_2642_fe0f, "\ud83e\udddf\u200d\u2642\ufe0f" );
		code( R.drawable.emj_1f9e0, "\ud83e\udde0" );
		code( R.drawable.emj_1f9e1, "\ud83e\udde1" );
		code( R.drawable.emj_1f9e2, "\ud83e\udde2" );
		code( R.drawable.emj_1f9e3, "\ud83e\udde3" );
		code( R.drawable.emj_1f9e4, "\ud83e\udde4" );
		code( R.drawable.emj_1f9e5, "\ud83e\udde5" );
		code( R.drawable.emj_1f9e6, "\ud83e\udde6" );
		code( R.drawable.emj_203c_fe0f, "\u203c" );
		code( R.drawable.emj_203c_fe0f, "\u203c\ufe0f" );
		code( R.drawable.emj_203c_fe0f, "\ue704" );
		code( R.drawable.emj_203c_fe0f, "\ueb30" );
		code( R.drawable.emj_203c_fe0f, "\udbba\udf06" );
		code( R.drawable.emj_2049_fe0f, "\u2049" );
		code( R.drawable.emj_2049_fe0f, "\u2049\ufe0f" );
		code( R.drawable.emj_2049_fe0f, "\ue703" );
		code( R.drawable.emj_2049_fe0f, "\ueb2f" );
		code( R.drawable.emj_2049_fe0f, "\udbba\udf05" );
		code( R.drawable.emj_2122_fe0f, "\u2122" );
		code( R.drawable.emj_2122_fe0f, "\u2122\ufe0f" );
		code( R.drawable.emj_2122_fe0f, "\ue537" );
		code( R.drawable.emj_2122_fe0f, "\ue54e" );
		code( R.drawable.emj_2122_fe0f, "\ue732" );
		code( R.drawable.emj_2122_fe0f, "\udbba\udf2a" );
		code( R.drawable.emj_2139_fe0f, "\u2139" );
		code( R.drawable.emj_2139_fe0f, "\u2139\ufe0f" );
		code( R.drawable.emj_2139_fe0f, "\ue533" );
		code( R.drawable.emj_2139_fe0f, "\udbba\udf47" );
		code( R.drawable.emj_2194_fe0f, "\u2194" );
	}
	private static void init50(){
		code( R.drawable.emj_2194_fe0f, "\u2194\ufe0f" );
		code( R.drawable.emj_2194_fe0f, "\ue73c" );
		code( R.drawable.emj_2194_fe0f, "\ueb7a" );
		code( R.drawable.emj_2194_fe0f, "\udbba\udef6" );
		code( R.drawable.emj_2195_fe0f, "\u2195" );
		code( R.drawable.emj_2195_fe0f, "\u2195\ufe0f" );
		code( R.drawable.emj_2195_fe0f, "\ue73d" );
		code( R.drawable.emj_2195_fe0f, "\ueb7b" );
		code( R.drawable.emj_2195_fe0f, "\udbba\udef7" );
		code( R.drawable.emj_2196_fe0f, "\u2196" );
		code( R.drawable.emj_2196_fe0f, "\u2196\ufe0f" );
		code( R.drawable.emj_2196_fe0f, "\ue237" );
		code( R.drawable.emj_2196_fe0f, "\ue54c" );
		code( R.drawable.emj_2196_fe0f, "\ue697" );
		code( R.drawable.emj_2196_fe0f, "\udbba\udef2" );
		code( R.drawable.emj_2197_fe0f, "\u2197" );
		code( R.drawable.emj_2197_fe0f, "\u2197\ufe0f" );
		code( R.drawable.emj_2197_fe0f, "\ue236" );
		code( R.drawable.emj_2197_fe0f, "\ue555" );
		code( R.drawable.emj_2197_fe0f, "\ue678" );
		code( R.drawable.emj_2197_fe0f, "\udbba\udef0" );
		code( R.drawable.emj_2198_fe0f, "\u2198" );
		code( R.drawable.emj_2198_fe0f, "\u2198\ufe0f" );
		code( R.drawable.emj_2198_fe0f, "\ue238" );
		code( R.drawable.emj_2198_fe0f, "\ue54d" );
		code( R.drawable.emj_2198_fe0f, "\ue696" );
		code( R.drawable.emj_2198_fe0f, "\udbba\udef1" );
		code( R.drawable.emj_2199_fe0f, "\u2199" );
		code( R.drawable.emj_2199_fe0f, "\u2199\ufe0f" );
		code( R.drawable.emj_2199_fe0f, "\ue239" );
		code( R.drawable.emj_2199_fe0f, "\ue556" );
		code( R.drawable.emj_2199_fe0f, "\ue6a5" );
		code( R.drawable.emj_2199_fe0f, "\udbba\udef3" );
		code( R.drawable.emj_21a9_fe0f, "\u21a9" );
		code( R.drawable.emj_21a9_fe0f, "\u21a9\ufe0f" );
		code( R.drawable.emj_21a9_fe0f, "\ue55d" );
		code( R.drawable.emj_21a9_fe0f, "\ue6da" );
		code( R.drawable.emj_21a9_fe0f, "\udbba\udf83" );
		code( R.drawable.emj_21aa_fe0f, "\u21aa" );
		code( R.drawable.emj_21aa_fe0f, "\u21aa\ufe0f" );
		code( R.drawable.emj_21aa_fe0f, "\ue55c" );
		code( R.drawable.emj_21aa_fe0f, "\udbba\udf88" );
		code( R.drawable.emj_231a, "\u231a" );
		code( R.drawable.emj_231a, "\ue57a" );
		code( R.drawable.emj_231a, "\ue71f" );
		code( R.drawable.emj_231a, "\udbb8\udc1d" );
		code( R.drawable.emj_231b, "\u231b" );
		code( R.drawable.emj_231b, "\ue57b" );
		code( R.drawable.emj_231b, "\ue71c" );
		code( R.drawable.emj_231b, "\udbb8\udc1c" );
		code( R.drawable.emj_2328_fe0f, "\u2328" );
		code( R.drawable.emj_2328_fe0f, "\u2328\ufe0f" );
		code( R.drawable.emj_23cf_fe0f, "\u23cf" );
		code( R.drawable.emj_23cf_fe0f, "\u23cf\ufe0f" );
		code( R.drawable.emj_23e9, "\u23e9" );
		code( R.drawable.emj_23e9, "\ue23c" );
		code( R.drawable.emj_23e9, "\ue530" );
		code( R.drawable.emj_23e9, "\udbba\udefe" );
		code( R.drawable.emj_23ea, "\u23ea" );
		code( R.drawable.emj_23ea, "\ue23d" );
		code( R.drawable.emj_23ea, "\ue52f" );
		code( R.drawable.emj_23ea, "\udbba\udeff" );
		code( R.drawable.emj_23eb, "\u23eb" );
		code( R.drawable.emj_23eb, "\ue545" );
		code( R.drawable.emj_23eb, "\udbba\udf03" );
		code( R.drawable.emj_23ec, "\u23ec" );
		code( R.drawable.emj_23ec, "\ue544" );
		code( R.drawable.emj_23ec, "\udbba\udf02" );
		code( R.drawable.emj_23ed_fe0f, "\u23ed" );
		code( R.drawable.emj_23ed_fe0f, "\u23ed\ufe0f" );
		code( R.drawable.emj_23ee_fe0f, "\u23ee" );
		code( R.drawable.emj_23ee_fe0f, "\u23ee\ufe0f" );
		code( R.drawable.emj_23ef_fe0f, "\u23ef" );
		code( R.drawable.emj_23ef_fe0f, "\u23ef\ufe0f" );
		code( R.drawable.emj_23f0, "\u23f0" );
		code( R.drawable.emj_23f0, "\ue594" );
		code( R.drawable.emj_23f0, "\ue6ba" );
		code( R.drawable.emj_23f0, "\udbb8\udc2a" );
		code( R.drawable.emj_23f1_fe0f, "\u23f1" );
		code( R.drawable.emj_23f1_fe0f, "\u23f1\ufe0f" );
		code( R.drawable.emj_23f2_fe0f, "\u23f2" );
		code( R.drawable.emj_23f2_fe0f, "\u23f2\ufe0f" );
		code( R.drawable.emj_23f3, "\u23f3" );
		code( R.drawable.emj_23f3, "\ue47c" );
		code( R.drawable.emj_23f3, "\udbb8\udc1b" );
		code( R.drawable.emj_23f8_fe0f, "\u23f8" );
		code( R.drawable.emj_23f8_fe0f, "\u23f8\ufe0f" );
		code( R.drawable.emj_23f9_fe0f, "\u23f9" );
		code( R.drawable.emj_23f9_fe0f, "\u23f9\ufe0f" );
		code( R.drawable.emj_23fa_fe0f, "\u23fa" );
		code( R.drawable.emj_23fa_fe0f, "\u23fa\ufe0f" );
		code( R.drawable.emj_24c2_fe0f, "\u24c2" );
		code( R.drawable.emj_24c2_fe0f, "\u24c2\ufe0f" );
		code( R.drawable.emj_24c2_fe0f, "\ue5bc" );
		code( R.drawable.emj_24c2_fe0f, "\ue65c" );
		code( R.drawable.emj_24c2_fe0f, "\udbb9\udfe1" );
		code( R.drawable.emj_25aa_fe0f, "\u25aa" );
		code( R.drawable.emj_25aa_fe0f, "\u25aa\ufe0f" );
		code( R.drawable.emj_25aa_fe0f, "\ue532" );
		code( R.drawable.emj_25aa_fe0f, "\udbba\udf6e" );
		code( R.drawable.emj_25ab_fe0f, "\u25ab" );
	}
	private static void init51(){
		code( R.drawable.emj_25ab_fe0f, "\u25ab\ufe0f" );
		code( R.drawable.emj_25ab_fe0f, "\ue531" );
		code( R.drawable.emj_25ab_fe0f, "\udbba\udf6d" );
		code( R.drawable.emj_25b6_fe0f, "\u25b6" );
		code( R.drawable.emj_25b6_fe0f, "\u25b6\ufe0f" );
		code( R.drawable.emj_25b6_fe0f, "\ue23a" );
		code( R.drawable.emj_25b6_fe0f, "\ue52e" );
		code( R.drawable.emj_25b6_fe0f, "\udbba\udefc" );
		code( R.drawable.emj_25c0_fe0f, "\u25c0" );
		code( R.drawable.emj_25c0_fe0f, "\u25c0\ufe0f" );
		code( R.drawable.emj_25c0_fe0f, "\ue23b" );
		code( R.drawable.emj_25c0_fe0f, "\ue52d" );
		code( R.drawable.emj_25c0_fe0f, "\udbba\udefd" );
		code( R.drawable.emj_25fb_fe0f, "\u25fb" );
		code( R.drawable.emj_25fb_fe0f, "\u25fb\ufe0f" );
		code( R.drawable.emj_25fb_fe0f, "\ue538" );
		code( R.drawable.emj_25fb_fe0f, "\udbba\udf71" );
		code( R.drawable.emj_25fc_fe0f, "\u25fc" );
		code( R.drawable.emj_25fc_fe0f, "\u25fc\ufe0f" );
		code( R.drawable.emj_25fc_fe0f, "\ue539" );
		code( R.drawable.emj_25fc_fe0f, "\udbba\udf72" );
		code( R.drawable.emj_25fd, "\u25fd" );
		code( R.drawable.emj_25fd, "\ue534" );
		code( R.drawable.emj_25fd, "\udbba\udf6f" );
		code( R.drawable.emj_25fe, "\u25fe" );
		code( R.drawable.emj_25fe, "\ue535" );
		code( R.drawable.emj_25fe, "\udbba\udf70" );
		code( R.drawable.emj_2600_fe0f, "\u2600" );
		code( R.drawable.emj_2600_fe0f, "\u2600\ufe0f" );
		code( R.drawable.emj_2600_fe0f, "\ue04a" );
		code( R.drawable.emj_2600_fe0f, "\ue488" );
		code( R.drawable.emj_2600_fe0f, "\ue63e" );
		code( R.drawable.emj_2600_fe0f, "\udbb8\udc00" );
		code( R.drawable.emj_2601_fe0f, "\u2601" );
		code( R.drawable.emj_2601_fe0f, "\u2601\ufe0f" );
		code( R.drawable.emj_2601_fe0f, "\ue049" );
		code( R.drawable.emj_2601_fe0f, "\ue48d" );
		code( R.drawable.emj_2601_fe0f, "\ue63f" );
		code( R.drawable.emj_2601_fe0f, "\udbb8\udc01" );
		code( R.drawable.emj_2602_fe0f, "\u2602" );
		code( R.drawable.emj_2602_fe0f, "\u2602\ufe0f" );
		code( R.drawable.emj_2603_fe0f, "\u2603" );
		code( R.drawable.emj_2603_fe0f, "\u2603\ufe0f" );
		code( R.drawable.emj_2604_fe0f, "\u2604" );
		code( R.drawable.emj_2604_fe0f, "\u2604\ufe0f" );
		code( R.drawable.emj_260e_fe0f, "\u260e" );
		code( R.drawable.emj_260e_fe0f, "\u260e\ufe0f" );
		code( R.drawable.emj_260e_fe0f, "\ue009" );
		code( R.drawable.emj_260e_fe0f, "\ue596" );
		code( R.drawable.emj_260e_fe0f, "\ue687" );
		code( R.drawable.emj_260e_fe0f, "\udbb9\udd23" );
		code( R.drawable.emj_2611_fe0f, "\u2611" );
		code( R.drawable.emj_2611_fe0f, "\u2611\ufe0f" );
		code( R.drawable.emj_2611_fe0f, "\ueb02" );
		code( R.drawable.emj_2611_fe0f, "\udbba\udf8b" );
		code( R.drawable.emj_2614, "\u2614" );
		code( R.drawable.emj_2614, "\ue04b" );
		code( R.drawable.emj_2614, "\ue48c" );
		code( R.drawable.emj_2614, "\ue640" );
		code( R.drawable.emj_2614, "\udbb8\udc02" );
		code( R.drawable.emj_2615, "\u2615" );
		code( R.drawable.emj_2615, "\ue045" );
		code( R.drawable.emj_2615, "\ue597" );
		code( R.drawable.emj_2615, "\ue670" );
		code( R.drawable.emj_2615, "\udbba\udd81" );
		code( R.drawable.emj_2618_fe0f, "\u2618" );
		code( R.drawable.emj_2618_fe0f, "\u2618\ufe0f" );
		code( R.drawable.emj_261d_1f3fb, "\u261d\ud83c\udffb" );
		code( R.drawable.emj_261d_1f3fc, "\u261d\ud83c\udffc" );
		code( R.drawable.emj_261d_1f3fd, "\u261d\ud83c\udffd" );
		code( R.drawable.emj_261d_1f3fe, "\u261d\ud83c\udffe" );
		code( R.drawable.emj_261d_1f3ff, "\u261d\ud83c\udfff" );
		code( R.drawable.emj_261d_fe0f, "\u261d" );
		code( R.drawable.emj_261d_fe0f, "\u261d\ufe0f" );
		code( R.drawable.emj_261d_fe0f, "\ue00f" );
		code( R.drawable.emj_261d_fe0f, "\ue4f6" );
		code( R.drawable.emj_261d_fe0f, "\udbba\udf98" );
		code( R.drawable.emj_2620_fe0f, "\u2620" );
		code( R.drawable.emj_2620_fe0f, "\u2620\ufe0f" );
		code( R.drawable.emj_2622_fe0f, "\u2622" );
		code( R.drawable.emj_2622_fe0f, "\u2622\ufe0f" );
		code( R.drawable.emj_2623_fe0f, "\u2623" );
		code( R.drawable.emj_2623_fe0f, "\u2623\ufe0f" );
		code( R.drawable.emj_2626_fe0f, "\u2626" );
		code( R.drawable.emj_2626_fe0f, "\u2626\ufe0f" );
		code( R.drawable.emj_262a_fe0f, "\u262a" );
		code( R.drawable.emj_262a_fe0f, "\u262a\ufe0f" );
		code( R.drawable.emj_262e_fe0f, "\u262e" );
		code( R.drawable.emj_262e_fe0f, "\u262e\ufe0f" );
		code( R.drawable.emj_262f_fe0f, "\u262f" );
		code( R.drawable.emj_262f_fe0f, "\u262f\ufe0f" );
		code( R.drawable.emj_2638_fe0f, "\u2638" );
		code( R.drawable.emj_2638_fe0f, "\u2638\ufe0f" );
		code( R.drawable.emj_2639_fe0f, "\u2639" );
		code( R.drawable.emj_2639_fe0f, "\u2639\ufe0f" );
		code( R.drawable.emj_263a_fe0f, "\u263a" );
		code( R.drawable.emj_263a_fe0f, "\u263a\ufe0f" );
		code( R.drawable.emj_263a_fe0f, "\ue414" );
		code( R.drawable.emj_263a_fe0f, "\ue4fb" );
		code( R.drawable.emj_263a_fe0f, "\udbb8\udf36" );
		code( R.drawable.emj_2640_fe0f, "\u2640" );
	}
	private static void init52(){
		code( R.drawable.emj_2640_fe0f, "\u2640\ufe0f" );
		code( R.drawable.emj_2642_fe0f, "\u2642" );
		code( R.drawable.emj_2642_fe0f, "\u2642\ufe0f" );
		code( R.drawable.emj_2648, "\u2648" );
		code( R.drawable.emj_2648, "\ue23f" );
		code( R.drawable.emj_2648, "\ue48f" );
		code( R.drawable.emj_2648, "\ue646" );
		code( R.drawable.emj_2648, "\udbb8\udc2b" );
		code( R.drawable.emj_2649, "\u2649" );
		code( R.drawable.emj_2649, "\ue240" );
		code( R.drawable.emj_2649, "\ue490" );
		code( R.drawable.emj_2649, "\ue647" );
		code( R.drawable.emj_2649, "\udbb8\udc2c" );
		code( R.drawable.emj_264a, "\u264a" );
		code( R.drawable.emj_264a, "\ue241" );
		code( R.drawable.emj_264a, "\ue491" );
		code( R.drawable.emj_264a, "\ue648" );
		code( R.drawable.emj_264a, "\udbb8\udc2d" );
		code( R.drawable.emj_264b, "\u264b" );
		code( R.drawable.emj_264b, "\ue242" );
		code( R.drawable.emj_264b, "\ue492" );
		code( R.drawable.emj_264b, "\ue649" );
		code( R.drawable.emj_264b, "\udbb8\udc2e" );
		code( R.drawable.emj_264c, "\u264c" );
		code( R.drawable.emj_264c, "\ue243" );
		code( R.drawable.emj_264c, "\ue493" );
		code( R.drawable.emj_264c, "\ue64a" );
		code( R.drawable.emj_264c, "\udbb8\udc2f" );
		code( R.drawable.emj_264d, "\u264d" );
		code( R.drawable.emj_264d, "\ue244" );
		code( R.drawable.emj_264d, "\ue494" );
		code( R.drawable.emj_264d, "\ue64b" );
		code( R.drawable.emj_264d, "\udbb8\udc30" );
		code( R.drawable.emj_264e, "\u264e" );
		code( R.drawable.emj_264e, "\ue245" );
		code( R.drawable.emj_264e, "\ue495" );
		code( R.drawable.emj_264e, "\ue64c" );
		code( R.drawable.emj_264e, "\udbb8\udc31" );
		code( R.drawable.emj_264f, "\u264f" );
		code( R.drawable.emj_264f, "\ue246" );
		code( R.drawable.emj_264f, "\ue496" );
		code( R.drawable.emj_264f, "\ue64d" );
		code( R.drawable.emj_264f, "\udbb8\udc32" );
		code( R.drawable.emj_2650, "\u2650" );
		code( R.drawable.emj_2650, "\ue247" );
		code( R.drawable.emj_2650, "\ue497" );
		code( R.drawable.emj_2650, "\ue64e" );
		code( R.drawable.emj_2650, "\udbb8\udc33" );
		code( R.drawable.emj_2651, "\u2651" );
		code( R.drawable.emj_2651, "\ue248" );
		code( R.drawable.emj_2651, "\ue498" );
		code( R.drawable.emj_2651, "\ue64f" );
		code( R.drawable.emj_2651, "\udbb8\udc34" );
		code( R.drawable.emj_2652, "\u2652" );
		code( R.drawable.emj_2652, "\ue249" );
		code( R.drawable.emj_2652, "\ue499" );
		code( R.drawable.emj_2652, "\ue650" );
		code( R.drawable.emj_2652, "\udbb8\udc35" );
		code( R.drawable.emj_2653, "\u2653" );
		code( R.drawable.emj_2653, "\ue24a" );
		code( R.drawable.emj_2653, "\ue49a" );
		code( R.drawable.emj_2653, "\ue651" );
		code( R.drawable.emj_2653, "\udbb8\udc36" );
		code( R.drawable.emj_2660_fe0f, "\u2660" );
		code( R.drawable.emj_2660_fe0f, "\u2660\ufe0f" );
		code( R.drawable.emj_2660_fe0f, "\ue20e" );
		code( R.drawable.emj_2660_fe0f, "\ue5a1" );
		code( R.drawable.emj_2660_fe0f, "\ue68e" );
		code( R.drawable.emj_2660_fe0f, "\udbba\udf1b" );
		code( R.drawable.emj_2663_fe0f, "\u2663" );
		code( R.drawable.emj_2663_fe0f, "\u2663\ufe0f" );
		code( R.drawable.emj_2663_fe0f, "\ue20f" );
		code( R.drawable.emj_2663_fe0f, "\ue5a3" );
		code( R.drawable.emj_2663_fe0f, "\ue690" );
		code( R.drawable.emj_2663_fe0f, "\udbba\udf1d" );
		code( R.drawable.emj_2665_fe0f, "\u2665" );
		code( R.drawable.emj_2665_fe0f, "\u2665\ufe0f" );
		code( R.drawable.emj_2665_fe0f, "\ue20c" );
		code( R.drawable.emj_2665_fe0f, "\ue68d" );
		code( R.drawable.emj_2665_fe0f, "\ueaa5" );
		code( R.drawable.emj_2665_fe0f, "\udbba\udf1a" );
		code( R.drawable.emj_2666_fe0f, "\u2666" );
		code( R.drawable.emj_2666_fe0f, "\u2666\ufe0f" );
		code( R.drawable.emj_2666_fe0f, "\ue20d" );
		code( R.drawable.emj_2666_fe0f, "\ue5a2" );
		code( R.drawable.emj_2666_fe0f, "\ue68f" );
		code( R.drawable.emj_2666_fe0f, "\udbba\udf1c" );
		code( R.drawable.emj_2668_fe0f, "\u2668" );
		code( R.drawable.emj_2668_fe0f, "\u2668\ufe0f" );
		code( R.drawable.emj_2668_fe0f, "\ue123" );
		code( R.drawable.emj_2668_fe0f, "\ue4bc" );
		code( R.drawable.emj_2668_fe0f, "\ue6f7" );
		code( R.drawable.emj_2668_fe0f, "\udbb9\udffa" );
		code( R.drawable.emj_267b_fe0f, "\u267b" );
		code( R.drawable.emj_267b_fe0f, "\u267b\ufe0f" );
		code( R.drawable.emj_267b_fe0f, "\ue735" );
		code( R.drawable.emj_267b_fe0f, "\ueb79" );
		code( R.drawable.emj_267b_fe0f, "\udbba\udf2c" );
		code( R.drawable.emj_267f, "\u267f" );
		code( R.drawable.emj_267f, "\ue20a" );
		code( R.drawable.emj_267f, "\ue47f" );
	}
	private static void init53(){
		code( R.drawable.emj_267f, "\ue69b" );
		code( R.drawable.emj_267f, "\udbba\udf20" );
		code( R.drawable.emj_2692_fe0f, "\u2692" );
		code( R.drawable.emj_2692_fe0f, "\u2692\ufe0f" );
		code( R.drawable.emj_2693, "\u2693" );
		code( R.drawable.emj_2693, "\ue4a9" );
		code( R.drawable.emj_2693, "\ue661" );
		code( R.drawable.emj_2693, "\udbb9\udcc1" );
		code( R.drawable.emj_2694_fe0f, "\u2694" );
		code( R.drawable.emj_2694_fe0f, "\u2694\ufe0f" );
		code( R.drawable.emj_2695_fe0f, "\u2695" );
		code( R.drawable.emj_2695_fe0f, "\u2695\ufe0f" );
		code( R.drawable.emj_2696_fe0f, "\u2696" );
		code( R.drawable.emj_2696_fe0f, "\u2696\ufe0f" );
		code( R.drawable.emj_2697_fe0f, "\u2697" );
		code( R.drawable.emj_2697_fe0f, "\u2697\ufe0f" );
		code( R.drawable.emj_2699_fe0f, "\u2699" );
		code( R.drawable.emj_2699_fe0f, "\u2699\ufe0f" );
		code( R.drawable.emj_269b_fe0f, "\u269b" );
		code( R.drawable.emj_269b_fe0f, "\u269b\ufe0f" );
		code( R.drawable.emj_269c_fe0f, "\u269c" );
		code( R.drawable.emj_269c_fe0f, "\u269c\ufe0f" );
		code( R.drawable.emj_26a0_fe0f, "\u26a0" );
		code( R.drawable.emj_26a0_fe0f, "\u26a0\ufe0f" );
		code( R.drawable.emj_26a0_fe0f, "\ue252" );
		code( R.drawable.emj_26a0_fe0f, "\ue481" );
		code( R.drawable.emj_26a0_fe0f, "\ue737" );
		code( R.drawable.emj_26a0_fe0f, "\udbba\udf23" );
		code( R.drawable.emj_26a1, "\u26a1" );
		code( R.drawable.emj_26a1, "\ue13d" );
		code( R.drawable.emj_26a1, "\ue487" );
		code( R.drawable.emj_26a1, "\ue642" );
		code( R.drawable.emj_26a1, "\udbb8\udc04" );
		code( R.drawable.emj_26aa, "\u26aa" );
		code( R.drawable.emj_26aa, "\ue53a" );
		code( R.drawable.emj_26aa, "\udbba\udf65" );
		code( R.drawable.emj_26ab, "\u26ab" );
		code( R.drawable.emj_26ab, "\ue53b" );
		code( R.drawable.emj_26ab, "\udbba\udf66" );
		code( R.drawable.emj_26b0_fe0f, "\u26b0" );
		code( R.drawable.emj_26b0_fe0f, "\u26b0\ufe0f" );
		code( R.drawable.emj_26b1_fe0f, "\u26b1" );
		code( R.drawable.emj_26b1_fe0f, "\u26b1\ufe0f" );
		code( R.drawable.emj_26bd, "\u26bd" );
		code( R.drawable.emj_26bd, "\ue018" );
		code( R.drawable.emj_26bd, "\ue4b6" );
		code( R.drawable.emj_26bd, "\ue656" );
		code( R.drawable.emj_26bd, "\udbb9\udfd4" );
		code( R.drawable.emj_26be, "\u26be" );
		code( R.drawable.emj_26be, "\ue016" );
		code( R.drawable.emj_26be, "\ue4ba" );
		code( R.drawable.emj_26be, "\ue653" );
		code( R.drawable.emj_26be, "\udbb9\udfd1" );
		code( R.drawable.emj_26c4, "\u26c4" );
		code( R.drawable.emj_26c4, "\ue048" );
		code( R.drawable.emj_26c4, "\ue485" );
		code( R.drawable.emj_26c4, "\ue641" );
		code( R.drawable.emj_26c4, "\udbb8\udc03" );
		code( R.drawable.emj_26c5, "\u26c5" );
		code( R.drawable.emj_26c5, "\ue48e" );
		code( R.drawable.emj_26c5, "\ue63e\ue63f" );
		code( R.drawable.emj_26c5, "\udbb8\udc0f" );
		code( R.drawable.emj_26c8_fe0f, "\u26c8" );
		code( R.drawable.emj_26c8_fe0f, "\u26c8\ufe0f" );
		code( R.drawable.emj_26ce, "\u26ce" );
		code( R.drawable.emj_26ce, "\ue24b" );
		code( R.drawable.emj_26ce, "\ue49b" );
		code( R.drawable.emj_26ce, "\udbb8\udc37" );
		code( R.drawable.emj_26cf_fe0f, "\u26cf" );
		code( R.drawable.emj_26cf_fe0f, "\u26cf\ufe0f" );
		code( R.drawable.emj_26d1_fe0f, "\u26d1" );
		code( R.drawable.emj_26d1_fe0f, "\u26d1\ufe0f" );
		code( R.drawable.emj_26d3_fe0f, "\u26d3" );
		code( R.drawable.emj_26d3_fe0f, "\u26d3\ufe0f" );
		code( R.drawable.emj_26d4, "\u26d4" );
		code( R.drawable.emj_26d4, "\ue484" );
		code( R.drawable.emj_26d4, "\udbba\udf26" );
		code( R.drawable.emj_26e9_fe0f, "\u26e9" );
		code( R.drawable.emj_26e9_fe0f, "\u26e9\ufe0f" );
		code( R.drawable.emj_26ea, "\u26ea" );
		code( R.drawable.emj_26ea, "\ue037" );
		code( R.drawable.emj_26ea, "\ue5bb" );
		code( R.drawable.emj_26ea, "\udbb9\udcbb" );
		code( R.drawable.emj_26f0_fe0f, "\u26f0" );
		code( R.drawable.emj_26f0_fe0f, "\u26f0\ufe0f" );
		code( R.drawable.emj_26f1_fe0f, "\u26f1" );
		code( R.drawable.emj_26f1_fe0f, "\u26f1\ufe0f" );
		code( R.drawable.emj_26f2, "\u26f2" );
		code( R.drawable.emj_26f2, "\ue121" );
		code( R.drawable.emj_26f2, "\ue5cf" );
		code( R.drawable.emj_26f2, "\udbb9\udcbc" );
		code( R.drawable.emj_26f3, "\u26f3" );
		code( R.drawable.emj_26f3, "\ue014" );
		code( R.drawable.emj_26f3, "\ue599" );
		code( R.drawable.emj_26f3, "\ue654" );
		code( R.drawable.emj_26f3, "\udbb9\udfd2" );
		code( R.drawable.emj_26f4_fe0f, "\u26f4" );
		code( R.drawable.emj_26f4_fe0f, "\u26f4\ufe0f" );
		code( R.drawable.emj_26f5, "\u26f5" );
		code( R.drawable.emj_26f5, "\ue01c" );
		code( R.drawable.emj_26f5, "\ue4b4" );
	}
	private static void init54(){
		code( R.drawable.emj_26f5, "\ue6a3" );
		code( R.drawable.emj_26f5, "\udbb9\udfea" );
		code( R.drawable.emj_26f7_fe0f, "\u26f7" );
		code( R.drawable.emj_26f7_fe0f, "\u26f7\ufe0f" );
		code( R.drawable.emj_26f8_fe0f, "\u26f8" );
		code( R.drawable.emj_26f8_fe0f, "\u26f8\ufe0f" );
		code( R.drawable.emj_26f9_1f3fb, "\u26f9\ud83c\udffb" );
		code( R.drawable.emj_26f9_1f3fb_200d_2640_fe0f, "\u26f9\ud83c\udffb\u200d\u2640\ufe0f" );
		code( R.drawable.emj_26f9_1f3fb_200d_2640_fe0f, "\u26f9\ud83c\udffb\u2640" );
		code( R.drawable.emj_26f9_1f3fb_200d_2642_fe0f, "\u26f9\ud83c\udffb\u200d\u2642\ufe0f" );
		code( R.drawable.emj_26f9_1f3fb_200d_2642_fe0f, "\u26f9\ud83c\udffb\u2642" );
		code( R.drawable.emj_26f9_1f3fc, "\u26f9\ud83c\udffc" );
		code( R.drawable.emj_26f9_1f3fc_200d_2640_fe0f, "\u26f9\ud83c\udffc\u200d\u2640\ufe0f" );
		code( R.drawable.emj_26f9_1f3fc_200d_2640_fe0f, "\u26f9\ud83c\udffc\u2640" );
		code( R.drawable.emj_26f9_1f3fc_200d_2642_fe0f, "\u26f9\ud83c\udffc\u200d\u2642\ufe0f" );
		code( R.drawable.emj_26f9_1f3fc_200d_2642_fe0f, "\u26f9\ud83c\udffc\u2642" );
		code( R.drawable.emj_26f9_1f3fd, "\u26f9\ud83c\udffd" );
		code( R.drawable.emj_26f9_1f3fd_200d_2640_fe0f, "\u26f9\ud83c\udffd\u200d\u2640\ufe0f" );
		code( R.drawable.emj_26f9_1f3fd_200d_2640_fe0f, "\u26f9\ud83c\udffd\u2640" );
		code( R.drawable.emj_26f9_1f3fd_200d_2642_fe0f, "\u26f9\ud83c\udffd\u200d\u2642\ufe0f" );
		code( R.drawable.emj_26f9_1f3fd_200d_2642_fe0f, "\u26f9\ud83c\udffd\u2642" );
		code( R.drawable.emj_26f9_1f3fe, "\u26f9\ud83c\udffe" );
		code( R.drawable.emj_26f9_1f3fe_200d_2640_fe0f, "\u26f9\ud83c\udffe\u200d\u2640\ufe0f" );
		code( R.drawable.emj_26f9_1f3fe_200d_2640_fe0f, "\u26f9\ud83c\udffe\u2640" );
		code( R.drawable.emj_26f9_1f3fe_200d_2642_fe0f, "\u26f9\ud83c\udffe\u200d\u2642\ufe0f" );
		code( R.drawable.emj_26f9_1f3fe_200d_2642_fe0f, "\u26f9\ud83c\udffe\u2642" );
		code( R.drawable.emj_26f9_1f3ff, "\u26f9\ud83c\udfff" );
		code( R.drawable.emj_26f9_1f3ff_200d_2640_fe0f, "\u26f9\ud83c\udfff\u200d\u2640\ufe0f" );
		code( R.drawable.emj_26f9_1f3ff_200d_2640_fe0f, "\u26f9\ud83c\udfff\u2640" );
		code( R.drawable.emj_26f9_1f3ff_200d_2642_fe0f, "\u26f9\ud83c\udfff\u200d\u2642\ufe0f" );
		code( R.drawable.emj_26f9_1f3ff_200d_2642_fe0f, "\u26f9\ud83c\udfff\u2642" );
		code( R.drawable.emj_26f9_fe0f, "\u26f9" );
		code( R.drawable.emj_26f9_fe0f, "\u26f9\ufe0f" );
		code( R.drawable.emj_26f9_fe0f_200d_2640_fe0f, "\u26f9\u2640" );
		code( R.drawable.emj_26f9_fe0f_200d_2640_fe0f, "\u26f9\ufe0f\u200d\u2640\ufe0f" );
		code( R.drawable.emj_26f9_fe0f_200d_2642_fe0f, "\u26f9\u2642" );
		code( R.drawable.emj_26f9_fe0f_200d_2642_fe0f, "\u26f9\ufe0f\u200d\u2642\ufe0f" );
		code( R.drawable.emj_26fa, "\u26fa" );
		code( R.drawable.emj_26fa, "\ue122" );
		code( R.drawable.emj_26fa, "\ue5d0" );
		code( R.drawable.emj_26fa, "\udbb9\udffb" );
		code( R.drawable.emj_26fd, "\u26fd" );
		code( R.drawable.emj_26fd, "\ue03a" );
		code( R.drawable.emj_26fd, "\ue571" );
		code( R.drawable.emj_26fd, "\ue66b" );
		code( R.drawable.emj_26fd, "\udbb9\udff5" );
		code( R.drawable.emj_2702_fe0f, "\u2702" );
		code( R.drawable.emj_2702_fe0f, "\u2702\ufe0f" );
		code( R.drawable.emj_2702_fe0f, "\ue313" );
		code( R.drawable.emj_2702_fe0f, "\ue516" );
		code( R.drawable.emj_2702_fe0f, "\ue675" );
		code( R.drawable.emj_2702_fe0f, "\udbb9\udd3e" );
		code( R.drawable.emj_2705, "\u2705" );
		code( R.drawable.emj_2705, "\ue55e" );
		code( R.drawable.emj_2705, "\udbba\udf4a" );
		code( R.drawable.emj_2708_fe0f, "\u2708" );
		code( R.drawable.emj_2708_fe0f, "\u2708\ufe0f" );
		code( R.drawable.emj_2708_fe0f, "\ue01d" );
		code( R.drawable.emj_2708_fe0f, "\ue4b3" );
		code( R.drawable.emj_2708_fe0f, "\ue662" );
		code( R.drawable.emj_2708_fe0f, "\udbb9\udfe9" );
		code( R.drawable.emj_2709_fe0f, "\u2709" );
		code( R.drawable.emj_2709_fe0f, "\u2709\ufe0f" );
		code( R.drawable.emj_2709_fe0f, "\ue521" );
		code( R.drawable.emj_2709_fe0f, "\udbb9\udd29" );
		code( R.drawable.emj_270a, "\u270a" );
		code( R.drawable.emj_270a, "\ue010" );
		code( R.drawable.emj_270a, "\ue693" );
		code( R.drawable.emj_270a, "\ueb83" );
		code( R.drawable.emj_270a, "\udbba\udf93" );
		code( R.drawable.emj_270a_1f3fb, "\u270a\ud83c\udffb" );
		code( R.drawable.emj_270a_1f3fc, "\u270a\ud83c\udffc" );
		code( R.drawable.emj_270a_1f3fd, "\u270a\ud83c\udffd" );
		code( R.drawable.emj_270a_1f3fe, "\u270a\ud83c\udffe" );
		code( R.drawable.emj_270a_1f3ff, "\u270a\ud83c\udfff" );
		code( R.drawable.emj_270b, "\u270b" );
		code( R.drawable.emj_270b, "\ue012" );
		code( R.drawable.emj_270b, "\ue5a7" );
		code( R.drawable.emj_270b, "\ue695" );
		code( R.drawable.emj_270b, "\udbba\udf95" );
		code( R.drawable.emj_270b_1f3fb, "\u270b\ud83c\udffb" );
		code( R.drawable.emj_270b_1f3fc, "\u270b\ud83c\udffc" );
		code( R.drawable.emj_270b_1f3fd, "\u270b\ud83c\udffd" );
		code( R.drawable.emj_270b_1f3fe, "\u270b\ud83c\udffe" );
		code( R.drawable.emj_270b_1f3ff, "\u270b\ud83c\udfff" );
		code( R.drawable.emj_270c_1f3fb, "\u270c\ud83c\udffb" );
		code( R.drawable.emj_270c_1f3fc, "\u270c\ud83c\udffc" );
		code( R.drawable.emj_270c_1f3fd, "\u270c\ud83c\udffd" );
		code( R.drawable.emj_270c_1f3fe, "\u270c\ud83c\udffe" );
		code( R.drawable.emj_270c_1f3ff, "\u270c\ud83c\udfff" );
		code( R.drawable.emj_270c_fe0f, "\u270c" );
		code( R.drawable.emj_270c_fe0f, "\u270c\ufe0f" );
		code( R.drawable.emj_270c_fe0f, "\ue011" );
		code( R.drawable.emj_270c_fe0f, "\ue5a6" );
		code( R.drawable.emj_270c_fe0f, "\ue694" );
		code( R.drawable.emj_270c_fe0f, "\udbba\udf94" );
		code( R.drawable.emj_270d_1f3fb, "\u270d\ud83c\udffb" );
		code( R.drawable.emj_270d_1f3fc, "\u270d\ud83c\udffc" );
		code( R.drawable.emj_270d_1f3fd, "\u270d\ud83c\udffd" );
		code( R.drawable.emj_270d_1f3fe, "\u270d\ud83c\udffe" );
		code( R.drawable.emj_270d_1f3ff, "\u270d\ud83c\udfff" );
	}
	private static void init55(){
		code( R.drawable.emj_270d_fe0f, "\u270d" );
		code( R.drawable.emj_270d_fe0f, "\u270d\ufe0f" );
		code( R.drawable.emj_270f_fe0f, "\u270f" );
		code( R.drawable.emj_270f_fe0f, "\u270f\ufe0f" );
		code( R.drawable.emj_270f_fe0f, "\ue4a1" );
		code( R.drawable.emj_270f_fe0f, "\ue719" );
		code( R.drawable.emj_270f_fe0f, "\udbb9\udd39" );
		code( R.drawable.emj_2712_fe0f, "\u2712" );
		code( R.drawable.emj_2712_fe0f, "\u2712\ufe0f" );
		code( R.drawable.emj_2712_fe0f, "\ue6ae" );
		code( R.drawable.emj_2712_fe0f, "\ueb03" );
		code( R.drawable.emj_2712_fe0f, "\udbb9\udd36" );
		code( R.drawable.emj_2714_fe0f, "\u2714" );
		code( R.drawable.emj_2714_fe0f, "\u2714\ufe0f" );
		code( R.drawable.emj_2714_fe0f, "\ue557" );
		code( R.drawable.emj_2714_fe0f, "\udbba\udf49" );
		code( R.drawable.emj_2716_fe0f, "\u2716" );
		code( R.drawable.emj_2716_fe0f, "\u2716\ufe0f" );
		code( R.drawable.emj_2716_fe0f, "\ue54f" );
		code( R.drawable.emj_2716_fe0f, "\udbba\udf53" );
		code( R.drawable.emj_271d_fe0f, "\u271d" );
		code( R.drawable.emj_271d_fe0f, "\u271d\ufe0f" );
		code( R.drawable.emj_2721_fe0f, "\u2721" );
		code( R.drawable.emj_2721_fe0f, "\u2721\ufe0f" );
		code( R.drawable.emj_2728, "\u2728" );
		code( R.drawable.emj_2728, "\ue32e" );
		code( R.drawable.emj_2728, "\ue6fa" );
		code( R.drawable.emj_2728, "\ueaab" );
		code( R.drawable.emj_2728, "\udbba\udf60" );
		code( R.drawable.emj_2733_fe0f, "\u2733" );
		code( R.drawable.emj_2733_fe0f, "\u2733\ufe0f" );
		code( R.drawable.emj_2733_fe0f, "\ue206" );
		code( R.drawable.emj_2733_fe0f, "\ue53e" );
		code( R.drawable.emj_2733_fe0f, "\udbba\udf62" );
		code( R.drawable.emj_2734_fe0f, "\u2734" );
		code( R.drawable.emj_2734_fe0f, "\u2734\ufe0f" );
		code( R.drawable.emj_2734_fe0f, "\ue205" );
		code( R.drawable.emj_2734_fe0f, "\ue479" );
		code( R.drawable.emj_2734_fe0f, "\ue6f8" );
		code( R.drawable.emj_2734_fe0f, "\udbba\udf61" );
		code( R.drawable.emj_2744_fe0f, "\u2744" );
		code( R.drawable.emj_2744_fe0f, "\u2744\ufe0f" );
		code( R.drawable.emj_2744_fe0f, "\ue48a" );
		code( R.drawable.emj_2744_fe0f, "\udbb8\udc0e" );
		code( R.drawable.emj_2747_fe0f, "\u2747" );
		code( R.drawable.emj_2747_fe0f, "\u2747\ufe0f" );
		code( R.drawable.emj_2747_fe0f, "\ue46c" );
		code( R.drawable.emj_2747_fe0f, "\udbba\udf77" );
		code( R.drawable.emj_274c, "\u274c" );
		code( R.drawable.emj_274c, "\ue333" );
		code( R.drawable.emj_274c, "\ue550" );
		code( R.drawable.emj_274c, "\udbba\udf45" );
		code( R.drawable.emj_274e, "\u274e" );
		code( R.drawable.emj_274e, "\ue551" );
		code( R.drawable.emj_274e, "\udbba\udf46" );
		code( R.drawable.emj_2753, "\u2753" );
		code( R.drawable.emj_2753, "\ue020" );
		code( R.drawable.emj_2753, "\ue483" );
		code( R.drawable.emj_2753, "\udbba\udf09" );
		code( R.drawable.emj_2754, "\u2754" );
		code( R.drawable.emj_2754, "\ue336" );
		code( R.drawable.emj_2754, "\ue483" );
		code( R.drawable.emj_2754, "\udbba\udf0a" );
		code( R.drawable.emj_2755, "\u2755" );
		code( R.drawable.emj_2755, "\ue337" );
		code( R.drawable.emj_2755, "\ue482" );
		code( R.drawable.emj_2755, "\udbba\udf0b" );
		code( R.drawable.emj_2757, "\u2757" );
		code( R.drawable.emj_2757, "\ue021" );
		code( R.drawable.emj_2757, "\ue482" );
		code( R.drawable.emj_2757, "\ue702" );
		code( R.drawable.emj_2757, "\udbba\udf04" );
		code( R.drawable.emj_2763_fe0f, "\u2763" );
		code( R.drawable.emj_2763_fe0f, "\u2763\ufe0f" );
		code( R.drawable.emj_2764_fe0f, "\u2764" );
		code( R.drawable.emj_2764_fe0f, "\u2764\ufe0f" );
		code( R.drawable.emj_2764_fe0f, "\ue022" );
		code( R.drawable.emj_2764_fe0f, "\ue595" );
		code( R.drawable.emj_2764_fe0f, "\udbba\udf0c" );
		code( R.drawable.emj_2795, "\u2795" );
		code( R.drawable.emj_2795, "\ue53c" );
		code( R.drawable.emj_2795, "\udbba\udf51" );
		code( R.drawable.emj_2796, "\u2796" );
		code( R.drawable.emj_2796, "\ue53d" );
		code( R.drawable.emj_2796, "\udbba\udf52" );
		code( R.drawable.emj_2797, "\u2797" );
		code( R.drawable.emj_2797, "\ue554" );
		code( R.drawable.emj_2797, "\udbba\udf54" );
		code( R.drawable.emj_27a1_fe0f, "\u27a1" );
		code( R.drawable.emj_27a1_fe0f, "\u27a1\ufe0f" );
		code( R.drawable.emj_27a1_fe0f, "\ue234" );
		code( R.drawable.emj_27a1_fe0f, "\ue552" );
		code( R.drawable.emj_27a1_fe0f, "\udbba\udefa" );
		code( R.drawable.emj_27b0, "\u27b0" );
		code( R.drawable.emj_27b0, "\ue70a" );
		code( R.drawable.emj_27b0, "\ueb31" );
		code( R.drawable.emj_27b0, "\udbba\udf08" );
		code( R.drawable.emj_27bf, "\u27bf" );
		code( R.drawable.emj_27bf, "\ue211" );
		code( R.drawable.emj_27bf, "\ue6df" );
		code( R.drawable.emj_27bf, "\udbba\udc2b" );
	}
	private static void init56(){
		code( R.drawable.emj_2934_fe0f, "\u2934" );
		code( R.drawable.emj_2934_fe0f, "\u2934\ufe0f" );
		code( R.drawable.emj_2934_fe0f, "\ue6f5" );
		code( R.drawable.emj_2934_fe0f, "\ueb2d" );
		code( R.drawable.emj_2934_fe0f, "\udbba\udef4" );
		code( R.drawable.emj_2935_fe0f, "\u2935" );
		code( R.drawable.emj_2935_fe0f, "\u2935\ufe0f" );
		code( R.drawable.emj_2935_fe0f, "\ue700" );
		code( R.drawable.emj_2935_fe0f, "\ueb2e" );
		code( R.drawable.emj_2935_fe0f, "\udbba\udef5" );
		code( R.drawable.emj_2b05_fe0f, "\u2b05" );
		code( R.drawable.emj_2b05_fe0f, "\u2b05\ufe0f" );
		code( R.drawable.emj_2b05_fe0f, "\ue235" );
		code( R.drawable.emj_2b05_fe0f, "\ue553" );
		code( R.drawable.emj_2b05_fe0f, "\udbba\udefb" );
		code( R.drawable.emj_2b06_fe0f, "\u2b06" );
		code( R.drawable.emj_2b06_fe0f, "\u2b06\ufe0f" );
		code( R.drawable.emj_2b06_fe0f, "\ue232" );
		code( R.drawable.emj_2b06_fe0f, "\ue53f" );
		code( R.drawable.emj_2b06_fe0f, "\udbba\udef8" );
		code( R.drawable.emj_2b07_fe0f, "\u2b07" );
		code( R.drawable.emj_2b07_fe0f, "\u2b07\ufe0f" );
		code( R.drawable.emj_2b07_fe0f, "\ue233" );
		code( R.drawable.emj_2b07_fe0f, "\ue540" );
		code( R.drawable.emj_2b07_fe0f, "\udbba\udef9" );
		code( R.drawable.emj_2b1b, "\u2b1b" );
		code( R.drawable.emj_2b1b, "\ue549" );
		code( R.drawable.emj_2b1b, "\udbba\udf6c" );
		code( R.drawable.emj_2b1c, "\u2b1c" );
		code( R.drawable.emj_2b1c, "\ue548" );
		code( R.drawable.emj_2b1c, "\udbba\udf6b" );
		code( R.drawable.emj_2b50, "\u2b50" );
		code( R.drawable.emj_2b50, "\ue32f" );
		code( R.drawable.emj_2b50, "\ue48b" );
		code( R.drawable.emj_2b50, "\udbba\udf68" );
		code( R.drawable.emj_2b55, "\u2b55" );
		code( R.drawable.emj_2b55, "\ue332" );
		code( R.drawable.emj_2b55, "\ueaad" );
		code( R.drawable.emj_2b55, "\udbba\udf44" );
		code( R.drawable.emj_3030_fe0f, "\u3030" );
		code( R.drawable.emj_3030_fe0f, "\u3030\ufe0f" );
		code( R.drawable.emj_3030_fe0f, "\ue709" );
		code( R.drawable.emj_3030_fe0f, "\udbba\udf07" );
		code( R.drawable.emj_303d_fe0f, "\u303d" );
		code( R.drawable.emj_303d_fe0f, "\u303d\ufe0f" );
		code( R.drawable.emj_303d_fe0f, "\ue12c" );
		code( R.drawable.emj_303d_fe0f, "\udbba\udc1b" );
		code( R.drawable.emj_3297_fe0f, "\u3297" );
		code( R.drawable.emj_3297_fe0f, "\u3297\ufe0f" );
		code( R.drawable.emj_3297_fe0f, "\ue30d" );
		code( R.drawable.emj_3297_fe0f, "\uea99" );
		code( R.drawable.emj_3297_fe0f, "\udbba\udf43" );
		code( R.drawable.emj_3299_fe0f, "\u3299" );
		code( R.drawable.emj_3299_fe0f, "\u3299\ufe0f" );
		code( R.drawable.emj_3299_fe0f, "\ue315" );
		code( R.drawable.emj_3299_fe0f, "\ue4f1" );
		code( R.drawable.emj_3299_fe0f, "\ue734" );
		code( R.drawable.emj_3299_fe0f, "\udbba\udf2b" );
		name( "+1", R.drawable.emj_1f44d, "\ud83d\udc4d" );
		name( "+1_dark_skin_tone", R.drawable.emj_1f44d_1f3ff, "\ud83d\udc4d\ud83c\udfff" );
		name( "+1_light_skin_tone", R.drawable.emj_1f44d_1f3fb, "\ud83d\udc4d\ud83c\udffb" );
		name( "+1_medium_dark_skin_tone", R.drawable.emj_1f44d_1f3fe, "\ud83d\udc4d\ud83c\udffe" );
		name( "+1_medium_light_skin_tone", R.drawable.emj_1f44d_1f3fc, "\ud83d\udc4d\ud83c\udffc" );
		name( "+1_medium_skin_tone", R.drawable.emj_1f44d_1f3fd, "\ud83d\udc4d\ud83c\udffd" );
		name( "+1_tone1", R.drawable.emj_1f44d_1f3fb, "\ud83d\udc4d\ud83c\udffb" );
		name( "+1_tone2", R.drawable.emj_1f44d_1f3fc, "\ud83d\udc4d\ud83c\udffc" );
		name( "+1_tone3", R.drawable.emj_1f44d_1f3fd, "\ud83d\udc4d\ud83c\udffd" );
		name( "+1_tone4", R.drawable.emj_1f44d_1f3fe, "\ud83d\udc4d\ud83c\udffe" );
		name( "+1_tone5", R.drawable.emj_1f44d_1f3ff, "\ud83d\udc4d\ud83c\udfff" );
		name( "100", R.drawable.emj_1f4af, "\ud83d\udcaf" );
		name( "1234", R.drawable.emj_1f522, "\ud83d\udd22" );
		name( "8ball", R.drawable.emj_1f3b1, "\ud83c\udfb1" );
		name( "_1", R.drawable.emj_1f44e, "\ud83d\udc4e" );
		name( "_1_dark_skin_tone", R.drawable.emj_1f44e_1f3ff, "\ud83d\udc4e\ud83c\udfff" );
		name( "_1_light_skin_tone", R.drawable.emj_1f44e_1f3fb, "\ud83d\udc4e\ud83c\udffb" );
		name( "_1_medium_dark_skin_tone", R.drawable.emj_1f44e_1f3fe, "\ud83d\udc4e\ud83c\udffe" );
		name( "_1_medium_light_skin_tone", R.drawable.emj_1f44e_1f3fc, "\ud83d\udc4e\ud83c\udffc" );
		name( "_1_medium_skin_tone", R.drawable.emj_1f44e_1f3fd, "\ud83d\udc4e\ud83c\udffd" );
		name( "_1_tone1", R.drawable.emj_1f44e_1f3fb, "\ud83d\udc4e\ud83c\udffb" );
		name( "_1_tone2", R.drawable.emj_1f44e_1f3fc, "\ud83d\udc4e\ud83c\udffc" );
		name( "_1_tone3", R.drawable.emj_1f44e_1f3fd, "\ud83d\udc4e\ud83c\udffd" );
		name( "_1_tone4", R.drawable.emj_1f44e_1f3fe, "\ud83d\udc4e\ud83c\udffe" );
		name( "_1_tone5", R.drawable.emj_1f44e_1f3ff, "\ud83d\udc4e\ud83c\udfff" );
		name( "a", R.drawable.emj_1f170_fe0f, "\ud83c\udd70\ufe0f" );
		name( "ab", R.drawable.emj_1f18e, "\ud83c\udd8e" );
		name( "abc", R.drawable.emj_1f524, "\ud83d\udd24" );
		name( "abcd", R.drawable.emj_1f521, "\ud83d\udd21" );
		name( "ac", R.drawable.emj_1f1e6_1f1e8, "\ud83c\udde6\ud83c\udde8" );
		name( "accept", R.drawable.emj_1f251, "\ud83c\ude51" );
		name( "ad", R.drawable.emj_1f1e6_1f1e9, "\ud83c\udde6\ud83c\udde9" );
		name( "admission_tickets", R.drawable.emj_1f39f_fe0f, "\ud83c\udf9f\ufe0f" );
		name( "adult", R.drawable.emj_1f9d1, "\ud83e\uddd1" );
		name( "adult_dark_skin_tone", R.drawable.emj_1f9d1_1f3ff, "\ud83e\uddd1\ud83c\udfff" );
		name( "adult_light_skin_tone", R.drawable.emj_1f9d1_1f3fb, "\ud83e\uddd1\ud83c\udffb" );
		name( "adult_medium_dark_skin_tone", R.drawable.emj_1f9d1_1f3fe, "\ud83e\uddd1\ud83c\udffe" );
		name( "adult_medium_light_skin_tone", R.drawable.emj_1f9d1_1f3fc, "\ud83e\uddd1\ud83c\udffc" );
		name( "adult_medium_skin_tone", R.drawable.emj_1f9d1_1f3fd, "\ud83e\uddd1\ud83c\udffd" );
		name( "adult_tone1", R.drawable.emj_1f9d1_1f3fb, "\ud83e\uddd1\ud83c\udffb" );
		name( "adult_tone2", R.drawable.emj_1f9d1_1f3fc, "\ud83e\uddd1\ud83c\udffc" );
		name( "adult_tone3", R.drawable.emj_1f9d1_1f3fd, "\ud83e\uddd1\ud83c\udffd" );
		name( "adult_tone4", R.drawable.emj_1f9d1_1f3fe, "\ud83e\uddd1\ud83c\udffe" );
	}
	private static void init57(){
		name( "adult_tone5", R.drawable.emj_1f9d1_1f3ff, "\ud83e\uddd1\ud83c\udfff" );
		name( "ae", R.drawable.emj_1f1e6_1f1ea, "\ud83c\udde6\ud83c\uddea" );
		name( "aerial_tramway", R.drawable.emj_1f6a1, "\ud83d\udea1" );
		name( "af", R.drawable.emj_1f1e6_1f1eb, "\ud83c\udde6\ud83c\uddeb" );
		name( "ag", R.drawable.emj_1f1e6_1f1ec, "\ud83c\udde6\ud83c\uddec" );
		name( "ai", R.drawable.emj_1f1e6_1f1ee, "\ud83c\udde6\ud83c\uddee" );
		name( "airplane", R.drawable.emj_2708_fe0f, "\u2708\ufe0f" );
		name( "airplane_arriving", R.drawable.emj_1f6ec, "\ud83d\udeec" );
		name( "airplane_departure", R.drawable.emj_1f6eb, "\ud83d\udeeb" );
		name( "airplane_small", R.drawable.emj_1f6e9_fe0f, "\ud83d\udee9\ufe0f" );
		name( "al", R.drawable.emj_1f1e6_1f1f1, "\ud83c\udde6\ud83c\uddf1" );
		name( "alarm_clock", R.drawable.emj_23f0, "\u23f0" );
		name( "alembic", R.drawable.emj_2697_fe0f, "\u2697\ufe0f" );
		name( "alien", R.drawable.emj_1f47d, "\ud83d\udc7d" );
		name( "am", R.drawable.emj_1f1e6_1f1f2, "\ud83c\udde6\ud83c\uddf2" );
		name( "ambulance", R.drawable.emj_1f691, "\ud83d\ude91" );
		name( "amphora", R.drawable.emj_1f3fa, "\ud83c\udffa" );
		name( "anchor", R.drawable.emj_2693, "\u2693" );
		name( "angel", R.drawable.emj_1f47c, "\ud83d\udc7c" );
		name( "angel_dark_skin_tone", R.drawable.emj_1f47c_1f3ff, "\ud83d\udc7c\ud83c\udfff" );
		name( "angel_light_skin_tone", R.drawable.emj_1f47c_1f3fb, "\ud83d\udc7c\ud83c\udffb" );
		name( "angel_medium_dark_skin_tone", R.drawable.emj_1f47c_1f3fe, "\ud83d\udc7c\ud83c\udffe" );
		name( "angel_medium_light_skin_tone", R.drawable.emj_1f47c_1f3fc, "\ud83d\udc7c\ud83c\udffc" );
		name( "angel_medium_skin_tone", R.drawable.emj_1f47c_1f3fd, "\ud83d\udc7c\ud83c\udffd" );
		name( "angel_tone1", R.drawable.emj_1f47c_1f3fb, "\ud83d\udc7c\ud83c\udffb" );
		name( "angel_tone2", R.drawable.emj_1f47c_1f3fc, "\ud83d\udc7c\ud83c\udffc" );
		name( "angel_tone3", R.drawable.emj_1f47c_1f3fd, "\ud83d\udc7c\ud83c\udffd" );
		name( "angel_tone4", R.drawable.emj_1f47c_1f3fe, "\ud83d\udc7c\ud83c\udffe" );
		name( "angel_tone5", R.drawable.emj_1f47c_1f3ff, "\ud83d\udc7c\ud83c\udfff" );
		name( "anger", R.drawable.emj_1f4a2, "\ud83d\udca2" );
		name( "anger_right", R.drawable.emj_1f5ef_fe0f, "\ud83d\uddef\ufe0f" );
		name( "angry", R.drawable.emj_1f620, "\ud83d\ude20" );
		name( "anguished", R.drawable.emj_1f627, "\ud83d\ude27" );
		name( "ant", R.drawable.emj_1f41c, "\ud83d\udc1c" );
		name( "ao", R.drawable.emj_1f1e6_1f1f4, "\ud83c\udde6\ud83c\uddf4" );
		name( "apple", R.drawable.emj_1f34e, "\ud83c\udf4e" );
		name( "aq", R.drawable.emj_1f1e6_1f1f6, "\ud83c\udde6\ud83c\uddf6" );
		name( "aquarius", R.drawable.emj_2652, "\u2652" );
		name( "ar", R.drawable.emj_1f1e6_1f1f7, "\ud83c\udde6\ud83c\uddf7" );
		name( "archery", R.drawable.emj_1f3f9, "\ud83c\udff9" );
		name( "aries", R.drawable.emj_2648, "\u2648" );
		name( "arrow_backward", R.drawable.emj_25c0_fe0f, "\u25c0\ufe0f" );
		name( "arrow_double_down", R.drawable.emj_23ec, "\u23ec" );
		name( "arrow_double_up", R.drawable.emj_23eb, "\u23eb" );
		name( "arrow_down", R.drawable.emj_2b07_fe0f, "\u2b07\ufe0f" );
		name( "arrow_down_small", R.drawable.emj_1f53d, "\ud83d\udd3d" );
		name( "arrow_forward", R.drawable.emj_25b6_fe0f, "\u25b6\ufe0f" );
		name( "arrow_heading_down", R.drawable.emj_2935_fe0f, "\u2935\ufe0f" );
		name( "arrow_heading_up", R.drawable.emj_2934_fe0f, "\u2934\ufe0f" );
		name( "arrow_left", R.drawable.emj_2b05_fe0f, "\u2b05\ufe0f" );
		name( "arrow_lower_left", R.drawable.emj_2199_fe0f, "\u2199\ufe0f" );
		name( "arrow_lower_right", R.drawable.emj_2198_fe0f, "\u2198\ufe0f" );
		name( "arrow_right", R.drawable.emj_27a1_fe0f, "\u27a1\ufe0f" );
		name( "arrow_right_hook", R.drawable.emj_21aa_fe0f, "\u21aa\ufe0f" );
		name( "arrow_up", R.drawable.emj_2b06_fe0f, "\u2b06\ufe0f" );
		name( "arrow_up_down", R.drawable.emj_2195_fe0f, "\u2195\ufe0f" );
		name( "arrow_up_small", R.drawable.emj_1f53c, "\ud83d\udd3c" );
		name( "arrow_upper_left", R.drawable.emj_2196_fe0f, "\u2196\ufe0f" );
		name( "arrow_upper_right", R.drawable.emj_2197_fe0f, "\u2197\ufe0f" );
		name( "arrows_clockwise", R.drawable.emj_1f503, "\ud83d\udd03" );
		name( "arrows_counterclockwise", R.drawable.emj_1f504, "\ud83d\udd04" );
		name( "art", R.drawable.emj_1f3a8, "\ud83c\udfa8" );
		name( "articulated_lorry", R.drawable.emj_1f69b, "\ud83d\ude9b" );
		name( "as", R.drawable.emj_1f1e6_1f1f8, "\ud83c\udde6\ud83c\uddf8" );
		name( "asterisk", R.drawable.emj_002a_fe0f_20e3, "\u002a\ufe0f\u20e3" );
		name( "asterisk_symbol", R.drawable.emj_002a, "\u002a" );
		name( "astonished", R.drawable.emj_1f632, "\ud83d\ude32" );
		name( "at", R.drawable.emj_1f1e6_1f1f9, "\ud83c\udde6\ud83c\uddf9" );
		name( "athletic_shoe", R.drawable.emj_1f45f, "\ud83d\udc5f" );
		name( "atm", R.drawable.emj_1f3e7, "\ud83c\udfe7" );
		name( "atom", R.drawable.emj_269b_fe0f, "\u269b\ufe0f" );
		name( "atom_symbol", R.drawable.emj_269b_fe0f, "\u269b\ufe0f" );
		name( "au", R.drawable.emj_1f1e6_1f1fa, "\ud83c\udde6\ud83c\uddfa" );
		name( "avocado", R.drawable.emj_1f951, "\ud83e\udd51" );
		name( "aw", R.drawable.emj_1f1e6_1f1fc, "\ud83c\udde6\ud83c\uddfc" );
		name( "ax", R.drawable.emj_1f1e6_1f1fd, "\ud83c\udde6\ud83c\uddfd" );
		name( "az", R.drawable.emj_1f1e6_1f1ff, "\ud83c\udde6\ud83c\uddff" );
		name( "b", R.drawable.emj_1f171_fe0f, "\ud83c\udd71\ufe0f" );
		name( "ba", R.drawable.emj_1f1e7_1f1e6, "\ud83c\udde7\ud83c\udde6" );
		name( "baby", R.drawable.emj_1f476, "\ud83d\udc76" );
		name( "baby_bottle", R.drawable.emj_1f37c, "\ud83c\udf7c" );
		name( "baby_chick", R.drawable.emj_1f424, "\ud83d\udc24" );
		name( "baby_dark_skin_tone", R.drawable.emj_1f476_1f3ff, "\ud83d\udc76\ud83c\udfff" );
		name( "baby_light_skin_tone", R.drawable.emj_1f476_1f3fb, "\ud83d\udc76\ud83c\udffb" );
		name( "baby_medium_dark_skin_tone", R.drawable.emj_1f476_1f3fe, "\ud83d\udc76\ud83c\udffe" );
		name( "baby_medium_light_skin_tone", R.drawable.emj_1f476_1f3fc, "\ud83d\udc76\ud83c\udffc" );
		name( "baby_medium_skin_tone", R.drawable.emj_1f476_1f3fd, "\ud83d\udc76\ud83c\udffd" );
		name( "baby_symbol", R.drawable.emj_1f6bc, "\ud83d\udebc" );
		name( "baby_tone1", R.drawable.emj_1f476_1f3fb, "\ud83d\udc76\ud83c\udffb" );
		name( "baby_tone2", R.drawable.emj_1f476_1f3fc, "\ud83d\udc76\ud83c\udffc" );
		name( "baby_tone3", R.drawable.emj_1f476_1f3fd, "\ud83d\udc76\ud83c\udffd" );
		name( "baby_tone4", R.drawable.emj_1f476_1f3fe, "\ud83d\udc76\ud83c\udffe" );
		name( "baby_tone5", R.drawable.emj_1f476_1f3ff, "\ud83d\udc76\ud83c\udfff" );
		name( "back", R.drawable.emj_1f519, "\ud83d\udd19" );
		name( "back_of_hand", R.drawable.emj_1f91a, "\ud83e\udd1a" );
		name( "back_of_hand_tone1", R.drawable.emj_1f91a_1f3fb, "\ud83e\udd1a\ud83c\udffb" );
		name( "back_of_hand_tone2", R.drawable.emj_1f91a_1f3fc, "\ud83e\udd1a\ud83c\udffc" );
		name( "back_of_hand_tone3", R.drawable.emj_1f91a_1f3fd, "\ud83e\udd1a\ud83c\udffd" );
		name( "back_of_hand_tone4", R.drawable.emj_1f91a_1f3fe, "\ud83e\udd1a\ud83c\udffe" );
		name( "back_of_hand_tone5", R.drawable.emj_1f91a_1f3ff, "\ud83e\udd1a\ud83c\udfff" );
		name( "bacon", R.drawable.emj_1f953, "\ud83e\udd53" );
	}
	private static void init58(){
		name( "badminton", R.drawable.emj_1f3f8, "\ud83c\udff8" );
		name( "badminton_racquet_and_shuttlecock", R.drawable.emj_1f3f8, "\ud83c\udff8" );
		name( "baggage_claim", R.drawable.emj_1f6c4, "\ud83d\udec4" );
		name( "baguette_bread", R.drawable.emj_1f956, "\ud83e\udd56" );
		name( "balloon", R.drawable.emj_1f388, "\ud83c\udf88" );
		name( "ballot_box", R.drawable.emj_1f5f3_fe0f, "\ud83d\uddf3\ufe0f" );
		name( "ballot_box_with_ballot", R.drawable.emj_1f5f3_fe0f, "\ud83d\uddf3\ufe0f" );
		name( "ballot_box_with_check", R.drawable.emj_2611_fe0f, "\u2611\ufe0f" );
		name( "bamboo", R.drawable.emj_1f38d, "\ud83c\udf8d" );
		name( "banana", R.drawable.emj_1f34c, "\ud83c\udf4c" );
		name( "bangbang", R.drawable.emj_203c_fe0f, "\u203c\ufe0f" );
		name( "bank", R.drawable.emj_1f3e6, "\ud83c\udfe6" );
		name( "bar_chart", R.drawable.emj_1f4ca, "\ud83d\udcca" );
		name( "barber", R.drawable.emj_1f488, "\ud83d\udc88" );
		name( "barely_sunny", R.drawable.emj_1f325_fe0f, "\ud83c\udf25\ufe0f" );
		name( "baseball", R.drawable.emj_26be, "\u26be" );
		name( "basketball", R.drawable.emj_1f3c0, "\ud83c\udfc0" );
		name( "basketball_player", R.drawable.emj_26f9_fe0f, "\u26f9\ufe0f" );
		name( "basketball_player_tone1", R.drawable.emj_26f9_1f3fb, "\u26f9\ud83c\udffb" );
		name( "basketball_player_tone2", R.drawable.emj_26f9_1f3fc, "\u26f9\ud83c\udffc" );
		name( "basketball_player_tone3", R.drawable.emj_26f9_1f3fd, "\u26f9\ud83c\udffd" );
		name( "basketball_player_tone4", R.drawable.emj_26f9_1f3fe, "\u26f9\ud83c\udffe" );
		name( "basketball_player_tone5", R.drawable.emj_26f9_1f3ff, "\u26f9\ud83c\udfff" );
		name( "bat", R.drawable.emj_1f987, "\ud83e\udd87" );
		name( "bath", R.drawable.emj_1f6c0, "\ud83d\udec0" );
		name( "bath_dark_skin_tone", R.drawable.emj_1f6c0_1f3ff, "\ud83d\udec0\ud83c\udfff" );
		name( "bath_light_skin_tone", R.drawable.emj_1f6c0_1f3fb, "\ud83d\udec0\ud83c\udffb" );
		name( "bath_medium_dark_skin_tone", R.drawable.emj_1f6c0_1f3fe, "\ud83d\udec0\ud83c\udffe" );
		name( "bath_medium_light_skin_tone", R.drawable.emj_1f6c0_1f3fc, "\ud83d\udec0\ud83c\udffc" );
		name( "bath_medium_skin_tone", R.drawable.emj_1f6c0_1f3fd, "\ud83d\udec0\ud83c\udffd" );
		name( "bath_tone1", R.drawable.emj_1f6c0_1f3fb, "\ud83d\udec0\ud83c\udffb" );
		name( "bath_tone2", R.drawable.emj_1f6c0_1f3fc, "\ud83d\udec0\ud83c\udffc" );
		name( "bath_tone3", R.drawable.emj_1f6c0_1f3fd, "\ud83d\udec0\ud83c\udffd" );
		name( "bath_tone4", R.drawable.emj_1f6c0_1f3fe, "\ud83d\udec0\ud83c\udffe" );
		name( "bath_tone5", R.drawable.emj_1f6c0_1f3ff, "\ud83d\udec0\ud83c\udfff" );
		name( "bathtub", R.drawable.emj_1f6c1, "\ud83d\udec1" );
		name( "battery", R.drawable.emj_1f50b, "\ud83d\udd0b" );
		name( "bb", R.drawable.emj_1f1e7_1f1e7, "\ud83c\udde7\ud83c\udde7" );
		name( "bd", R.drawable.emj_1f1e7_1f1e9, "\ud83c\udde7\ud83c\udde9" );
		name( "be", R.drawable.emj_1f1e7_1f1ea, "\ud83c\udde7\ud83c\uddea" );
		name( "beach", R.drawable.emj_1f3d6_fe0f, "\ud83c\udfd6\ufe0f" );
		name( "beach_umbrella", R.drawable.emj_26f1_fe0f, "\u26f1\ufe0f" );
		name( "beach_with_umbrella", R.drawable.emj_1f3d6_fe0f, "\ud83c\udfd6\ufe0f" );
		name( "bear", R.drawable.emj_1f43b, "\ud83d\udc3b" );
		name( "bearded_person", R.drawable.emj_1f9d4, "\ud83e\uddd4" );
		name( "bearded_person_dark_skin_tone", R.drawable.emj_1f9d4_1f3ff, "\ud83e\uddd4\ud83c\udfff" );
		name( "bearded_person_light_skin_tone", R.drawable.emj_1f9d4_1f3fb, "\ud83e\uddd4\ud83c\udffb" );
		name( "bearded_person_medium_dark_skin_tone", R.drawable.emj_1f9d4_1f3fe, "\ud83e\uddd4\ud83c\udffe" );
		name( "bearded_person_medium_light_skin_tone", R.drawable.emj_1f9d4_1f3fc, "\ud83e\uddd4\ud83c\udffc" );
		name( "bearded_person_medium_skin_tone", R.drawable.emj_1f9d4_1f3fd, "\ud83e\uddd4\ud83c\udffd" );
		name( "bearded_person_tone1", R.drawable.emj_1f9d4_1f3fb, "\ud83e\uddd4\ud83c\udffb" );
		name( "bearded_person_tone2", R.drawable.emj_1f9d4_1f3fc, "\ud83e\uddd4\ud83c\udffc" );
		name( "bearded_person_tone3", R.drawable.emj_1f9d4_1f3fd, "\ud83e\uddd4\ud83c\udffd" );
		name( "bearded_person_tone4", R.drawable.emj_1f9d4_1f3fe, "\ud83e\uddd4\ud83c\udffe" );
		name( "bearded_person_tone5", R.drawable.emj_1f9d4_1f3ff, "\ud83e\uddd4\ud83c\udfff" );
		name( "bed", R.drawable.emj_1f6cf_fe0f, "\ud83d\udecf\ufe0f" );
		name( "bee", R.drawable.emj_1f41d, "\ud83d\udc1d" );
		name( "beer", R.drawable.emj_1f37a, "\ud83c\udf7a" );
		name( "beers", R.drawable.emj_1f37b, "\ud83c\udf7b" );
		name( "beetle", R.drawable.emj_1f41e, "\ud83d\udc1e" );
		name( "beginner", R.drawable.emj_1f530, "\ud83d\udd30" );
		name( "bell", R.drawable.emj_1f514, "\ud83d\udd14" );
		name( "bellhop", R.drawable.emj_1f6ce_fe0f, "\ud83d\udece\ufe0f" );
		name( "bellhop_bell", R.drawable.emj_1f6ce_fe0f, "\ud83d\udece\ufe0f" );
		name( "bento", R.drawable.emj_1f371, "\ud83c\udf71" );
		name( "bf", R.drawable.emj_1f1e7_1f1eb, "\ud83c\udde7\ud83c\uddeb" );
		name( "bg", R.drawable.emj_1f1e7_1f1ec, "\ud83c\udde7\ud83c\uddec" );
		name( "bh", R.drawable.emj_1f1e7_1f1ed, "\ud83c\udde7\ud83c\udded" );
		name( "bi", R.drawable.emj_1f1e7_1f1ee, "\ud83c\udde7\ud83c\uddee" );
		name( "bicyclist", R.drawable.emj_1f6b4, "\ud83d\udeb4" );
		name( "bicyclist_dark_skin_tone", R.drawable.emj_1f6b4_1f3ff, "\ud83d\udeb4\ud83c\udfff" );
		name( "bicyclist_light_skin_tone", R.drawable.emj_1f6b4_1f3fb, "\ud83d\udeb4\ud83c\udffb" );
		name( "bicyclist_medium_dark_skin_tone", R.drawable.emj_1f6b4_1f3fe, "\ud83d\udeb4\ud83c\udffe" );
		name( "bicyclist_medium_light_skin_tone", R.drawable.emj_1f6b4_1f3fc, "\ud83d\udeb4\ud83c\udffc" );
		name( "bicyclist_medium_skin_tone", R.drawable.emj_1f6b4_1f3fd, "\ud83d\udeb4\ud83c\udffd" );
		name( "bicyclist_tone1", R.drawable.emj_1f6b4_1f3fb, "\ud83d\udeb4\ud83c\udffb" );
		name( "bicyclist_tone2", R.drawable.emj_1f6b4_1f3fc, "\ud83d\udeb4\ud83c\udffc" );
		name( "bicyclist_tone3", R.drawable.emj_1f6b4_1f3fd, "\ud83d\udeb4\ud83c\udffd" );
		name( "bicyclist_tone4", R.drawable.emj_1f6b4_1f3fe, "\ud83d\udeb4\ud83c\udffe" );
		name( "bicyclist_tone5", R.drawable.emj_1f6b4_1f3ff, "\ud83d\udeb4\ud83c\udfff" );
		name( "bike", R.drawable.emj_1f6b2, "\ud83d\udeb2" );
		name( "bikini", R.drawable.emj_1f459, "\ud83d\udc59" );
		name( "billed_cap", R.drawable.emj_1f9e2, "\ud83e\udde2" );
		name( "biohazard", R.drawable.emj_2623_fe0f, "\u2623\ufe0f" );
		name( "biohazard_sign", R.drawable.emj_2623_fe0f, "\u2623\ufe0f" );
		name( "bird", R.drawable.emj_1f426, "\ud83d\udc26" );
		name( "birthday", R.drawable.emj_1f382, "\ud83c\udf82" );
		name( "bj", R.drawable.emj_1f1e7_1f1ef, "\ud83c\udde7\ud83c\uddef" );
		name( "bl", R.drawable.emj_1f1e7_1f1f1, "\ud83c\udde7\ud83c\uddf1" );
		name( "black_circle", R.drawable.emj_26ab, "\u26ab" );
		name( "black_circle_for_record", R.drawable.emj_23fa_fe0f, "\u23fa\ufe0f" );
		name( "black_heart", R.drawable.emj_1f5a4, "\ud83d\udda4" );
		name( "black_joker", R.drawable.emj_1f0cf, "\ud83c\udccf" );
		name( "black_large_square", R.drawable.emj_2b1b, "\u2b1b" );
		name( "black_left_pointing_double_triangle_with_vertical_bar", R.drawable.emj_23ee_fe0f, "\u23ee\ufe0f" );
		name( "black_medium_small_square", R.drawable.emj_25fe, "\u25fe" );
		name( "black_medium_square", R.drawable.emj_25fc_fe0f, "\u25fc\ufe0f" );
		name( "black_nib", R.drawable.emj_2712_fe0f, "\u2712\ufe0f" );
		name( "black_right_pointing_double_triangle_with_vertical_bar", R.drawable.emj_23ed_fe0f, "\u23ed\ufe0f" );
		name( "black_right_pointing_triangle_with_double_vertical_bar", R.drawable.emj_23ef_fe0f, "\u23ef\ufe0f" );
		name( "black_small_square", R.drawable.emj_25aa_fe0f, "\u25aa\ufe0f" );
	}
	private static void init59(){
		name( "black_square_button", R.drawable.emj_1f532, "\ud83d\udd32" );
		name( "black_square_for_stop", R.drawable.emj_23f9_fe0f, "\u23f9\ufe0f" );
		name( "blond_haired_man", R.drawable.emj_1f471_200d_2642_fe0f, "\ud83d\udc71\u200d\u2642\ufe0f" );
		name( "blond_haired_man_dark_skin_tone", R.drawable.emj_1f471_1f3ff_200d_2642_fe0f, "\ud83d\udc71\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "blond_haired_man_light_skin_tone", R.drawable.emj_1f471_1f3fb_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "blond_haired_man_medium_dark_skin_tone", R.drawable.emj_1f471_1f3fe_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "blond_haired_man_medium_light_skin_tone", R.drawable.emj_1f471_1f3fc_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "blond_haired_man_medium_skin_tone", R.drawable.emj_1f471_1f3fd_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "blond_haired_man_tone1", R.drawable.emj_1f471_1f3fb_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "blond_haired_man_tone2", R.drawable.emj_1f471_1f3fc_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "blond_haired_man_tone3", R.drawable.emj_1f471_1f3fd_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "blond_haired_man_tone4", R.drawable.emj_1f471_1f3fe_200d_2642_fe0f, "\ud83d\udc71\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "blond_haired_man_tone5", R.drawable.emj_1f471_1f3ff_200d_2642_fe0f, "\ud83d\udc71\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "blond_haired_woman", R.drawable.emj_1f471_200d_2640_fe0f, "\ud83d\udc71\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_dark_skin_tone", R.drawable.emj_1f471_1f3ff_200d_2640_fe0f, "\ud83d\udc71\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_light_skin_tone", R.drawable.emj_1f471_1f3fb_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_medium_dark_skin_tone", R.drawable.emj_1f471_1f3fe_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_medium_light_skin_tone", R.drawable.emj_1f471_1f3fc_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_medium_skin_tone", R.drawable.emj_1f471_1f3fd_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_tone1", R.drawable.emj_1f471_1f3fb_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_tone2", R.drawable.emj_1f471_1f3fc_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_tone3", R.drawable.emj_1f471_1f3fd_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_tone4", R.drawable.emj_1f471_1f3fe_200d_2640_fe0f, "\ud83d\udc71\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "blond_haired_woman_tone5", R.drawable.emj_1f471_1f3ff_200d_2640_fe0f, "\ud83d\udc71\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "blossom", R.drawable.emj_1f33c, "\ud83c\udf3c" );
		name( "blowfish", R.drawable.emj_1f421, "\ud83d\udc21" );
		name( "blue_book", R.drawable.emj_1f4d8, "\ud83d\udcd8" );
		name( "blue_car", R.drawable.emj_1f699, "\ud83d\ude99" );
		name( "blue_circle", R.drawable.emj_1f535, "\ud83d\udd35" );
		name( "blue_heart", R.drawable.emj_1f499, "\ud83d\udc99" );
		name( "blush", R.drawable.emj_1f60a, "\ud83d\ude0a" );
		name( "bm", R.drawable.emj_1f1e7_1f1f2, "\ud83c\udde7\ud83c\uddf2" );
		name( "bn", R.drawable.emj_1f1e7_1f1f3, "\ud83c\udde7\ud83c\uddf3" );
		name( "bo", R.drawable.emj_1f1e7_1f1f4, "\ud83c\udde7\ud83c\uddf4" );
		name( "boar", R.drawable.emj_1f417, "\ud83d\udc17" );
		name( "boat", R.drawable.emj_26f5, "\u26f5" );
		name( "bomb", R.drawable.emj_1f4a3, "\ud83d\udca3" );
		name( "book", R.drawable.emj_1f4d6, "\ud83d\udcd6" );
		name( "bookmark", R.drawable.emj_1f516, "\ud83d\udd16" );
		name( "bookmark_tabs", R.drawable.emj_1f4d1, "\ud83d\udcd1" );
		name( "books", R.drawable.emj_1f4da, "\ud83d\udcda" );
		name( "boom", R.drawable.emj_1f4a5, "\ud83d\udca5" );
		name( "boot", R.drawable.emj_1f462, "\ud83d\udc62" );
		name( "bottle_with_popping_cork", R.drawable.emj_1f37e, "\ud83c\udf7e" );
		name( "bouquet", R.drawable.emj_1f490, "\ud83d\udc90" );
		name( "bow", R.drawable.emj_1f647, "\ud83d\ude47" );
		name( "bow_and_arrow", R.drawable.emj_1f3f9, "\ud83c\udff9" );
		name( "bow_dark_skin_tone", R.drawable.emj_1f647_1f3ff, "\ud83d\ude47\ud83c\udfff" );
		name( "bow_light_skin_tone", R.drawable.emj_1f647_1f3fb, "\ud83d\ude47\ud83c\udffb" );
		name( "bow_medium_dark_skin_tone", R.drawable.emj_1f647_1f3fe, "\ud83d\ude47\ud83c\udffe" );
		name( "bow_medium_light_skin_tone", R.drawable.emj_1f647_1f3fc, "\ud83d\ude47\ud83c\udffc" );
		name( "bow_medium_skin_tone", R.drawable.emj_1f647_1f3fd, "\ud83d\ude47\ud83c\udffd" );
		name( "bow_tone1", R.drawable.emj_1f647_1f3fb, "\ud83d\ude47\ud83c\udffb" );
		name( "bow_tone2", R.drawable.emj_1f647_1f3fc, "\ud83d\ude47\ud83c\udffc" );
		name( "bow_tone3", R.drawable.emj_1f647_1f3fd, "\ud83d\ude47\ud83c\udffd" );
		name( "bow_tone4", R.drawable.emj_1f647_1f3fe, "\ud83d\ude47\ud83c\udffe" );
		name( "bow_tone5", R.drawable.emj_1f647_1f3ff, "\ud83d\ude47\ud83c\udfff" );
		name( "bowl_with_spoon", R.drawable.emj_1f963, "\ud83e\udd63" );
		name( "bowling", R.drawable.emj_1f3b3, "\ud83c\udfb3" );
		name( "boxing_glove", R.drawable.emj_1f94a, "\ud83e\udd4a" );
		name( "boxing_gloves", R.drawable.emj_1f94a, "\ud83e\udd4a" );
		name( "boy", R.drawable.emj_1f466, "\ud83d\udc66" );
		name( "boy_dark_skin_tone", R.drawable.emj_1f466_1f3ff, "\ud83d\udc66\ud83c\udfff" );
		name( "boy_light_skin_tone", R.drawable.emj_1f466_1f3fb, "\ud83d\udc66\ud83c\udffb" );
		name( "boy_medium_dark_skin_tone", R.drawable.emj_1f466_1f3fe, "\ud83d\udc66\ud83c\udffe" );
		name( "boy_medium_light_skin_tone", R.drawable.emj_1f466_1f3fc, "\ud83d\udc66\ud83c\udffc" );
		name( "boy_medium_skin_tone", R.drawable.emj_1f466_1f3fd, "\ud83d\udc66\ud83c\udffd" );
		name( "boy_tone1", R.drawable.emj_1f466_1f3fb, "\ud83d\udc66\ud83c\udffb" );
		name( "boy_tone2", R.drawable.emj_1f466_1f3fc, "\ud83d\udc66\ud83c\udffc" );
		name( "boy_tone3", R.drawable.emj_1f466_1f3fd, "\ud83d\udc66\ud83c\udffd" );
		name( "boy_tone4", R.drawable.emj_1f466_1f3fe, "\ud83d\udc66\ud83c\udffe" );
		name( "boy_tone5", R.drawable.emj_1f466_1f3ff, "\ud83d\udc66\ud83c\udfff" );
		name( "bq", R.drawable.emj_1f1e7_1f1f6, "\ud83c\udde7\ud83c\uddf6" );
		name( "br", R.drawable.emj_1f1e7_1f1f7, "\ud83c\udde7\ud83c\uddf7" );
		name( "brain", R.drawable.emj_1f9e0, "\ud83e\udde0" );
		name( "bread", R.drawable.emj_1f35e, "\ud83c\udf5e" );
		name( "breast_feeding", R.drawable.emj_1f931, "\ud83e\udd31" );
		name( "breast_feeding_dark_skin_tone", R.drawable.emj_1f931_1f3ff, "\ud83e\udd31\ud83c\udfff" );
		name( "breast_feeding_light_skin_tone", R.drawable.emj_1f931_1f3fb, "\ud83e\udd31\ud83c\udffb" );
		name( "breast_feeding_medium_dark_skin_tone", R.drawable.emj_1f931_1f3fe, "\ud83e\udd31\ud83c\udffe" );
		name( "breast_feeding_medium_light_skin_tone", R.drawable.emj_1f931_1f3fc, "\ud83e\udd31\ud83c\udffc" );
		name( "breast_feeding_medium_skin_tone", R.drawable.emj_1f931_1f3fd, "\ud83e\udd31\ud83c\udffd" );
		name( "breast_feeding_tone1", R.drawable.emj_1f931_1f3fb, "\ud83e\udd31\ud83c\udffb" );
		name( "breast_feeding_tone2", R.drawable.emj_1f931_1f3fc, "\ud83e\udd31\ud83c\udffc" );
		name( "breast_feeding_tone3", R.drawable.emj_1f931_1f3fd, "\ud83e\udd31\ud83c\udffd" );
		name( "breast_feeding_tone4", R.drawable.emj_1f931_1f3fe, "\ud83e\udd31\ud83c\udffe" );
		name( "breast_feeding_tone5", R.drawable.emj_1f931_1f3ff, "\ud83e\udd31\ud83c\udfff" );
		name( "bride_with_veil", R.drawable.emj_1f470, "\ud83d\udc70" );
		name( "bride_with_veil_dark_skin_tone", R.drawable.emj_1f470_1f3ff, "\ud83d\udc70\ud83c\udfff" );
		name( "bride_with_veil_light_skin_tone", R.drawable.emj_1f470_1f3fb, "\ud83d\udc70\ud83c\udffb" );
		name( "bride_with_veil_medium_dark_skin_tone", R.drawable.emj_1f470_1f3fe, "\ud83d\udc70\ud83c\udffe" );
		name( "bride_with_veil_medium_light_skin_tone", R.drawable.emj_1f470_1f3fc, "\ud83d\udc70\ud83c\udffc" );
		name( "bride_with_veil_medium_skin_tone", R.drawable.emj_1f470_1f3fd, "\ud83d\udc70\ud83c\udffd" );
		name( "bride_with_veil_tone1", R.drawable.emj_1f470_1f3fb, "\ud83d\udc70\ud83c\udffb" );
		name( "bride_with_veil_tone2", R.drawable.emj_1f470_1f3fc, "\ud83d\udc70\ud83c\udffc" );
		name( "bride_with_veil_tone3", R.drawable.emj_1f470_1f3fd, "\ud83d\udc70\ud83c\udffd" );
		name( "bride_with_veil_tone4", R.drawable.emj_1f470_1f3fe, "\ud83d\udc70\ud83c\udffe" );
		name( "bride_with_veil_tone5", R.drawable.emj_1f470_1f3ff, "\ud83d\udc70\ud83c\udfff" );
		name( "bridge_at_night", R.drawable.emj_1f309, "\ud83c\udf09" );
		name( "briefcase", R.drawable.emj_1f4bc, "\ud83d\udcbc" );
		name( "broccoli", R.drawable.emj_1f966, "\ud83e\udd66" );
	}
	private static void init60(){
		name( "broken_heart", R.drawable.emj_1f494, "\ud83d\udc94" );
		name( "bs", R.drawable.emj_1f1e7_1f1f8, "\ud83c\udde7\ud83c\uddf8" );
		name( "bt", R.drawable.emj_1f1e7_1f1f9, "\ud83c\udde7\ud83c\uddf9" );
		name( "bug", R.drawable.emj_1f41b, "\ud83d\udc1b" );
		name( "building_construction", R.drawable.emj_1f3d7_fe0f, "\ud83c\udfd7\ufe0f" );
		name( "bulb", R.drawable.emj_1f4a1, "\ud83d\udca1" );
		name( "bullettrain_front", R.drawable.emj_1f685, "\ud83d\ude85" );
		name( "bullettrain_side", R.drawable.emj_1f684, "\ud83d\ude84" );
		name( "burrito", R.drawable.emj_1f32f, "\ud83c\udf2f" );
		name( "bus", R.drawable.emj_1f68c, "\ud83d\ude8c" );
		name( "busstop", R.drawable.emj_1f68f, "\ud83d\ude8f" );
		name( "bust_in_silhouette", R.drawable.emj_1f464, "\ud83d\udc64" );
		name( "busts_in_silhouette", R.drawable.emj_1f465, "\ud83d\udc65" );
		name( "butterfly", R.drawable.emj_1f98b, "\ud83e\udd8b" );
		name( "bv", R.drawable.emj_1f1e7_1f1fb, "\ud83c\udde7\ud83c\uddfb" );
		name( "bw", R.drawable.emj_1f1e7_1f1fc, "\ud83c\udde7\ud83c\uddfc" );
		name( "by", R.drawable.emj_1f1e7_1f1fe, "\ud83c\udde7\ud83c\uddfe" );
		name( "bz", R.drawable.emj_1f1e7_1f1ff, "\ud83c\udde7\ud83c\uddff" );
		name( "ca", R.drawable.emj_1f1e8_1f1e6, "\ud83c\udde8\ud83c\udde6" );
		name( "cactus", R.drawable.emj_1f335, "\ud83c\udf35" );
		name( "cake", R.drawable.emj_1f370, "\ud83c\udf70" );
		name( "calendar", R.drawable.emj_1f4c6, "\ud83d\udcc6" );
		name( "calendar_spiral", R.drawable.emj_1f5d3_fe0f, "\ud83d\uddd3\ufe0f" );
		name( "call_me", R.drawable.emj_1f919, "\ud83e\udd19" );
		name( "call_me_hand", R.drawable.emj_1f919, "\ud83e\udd19" );
		name( "call_me_hand_dark_skin_tone", R.drawable.emj_1f919_1f3ff, "\ud83e\udd19\ud83c\udfff" );
		name( "call_me_hand_light_skin_tone", R.drawable.emj_1f919_1f3fb, "\ud83e\udd19\ud83c\udffb" );
		name( "call_me_hand_medium_dark_skin_tone", R.drawable.emj_1f919_1f3fe, "\ud83e\udd19\ud83c\udffe" );
		name( "call_me_hand_medium_light_skin_tone", R.drawable.emj_1f919_1f3fc, "\ud83e\udd19\ud83c\udffc" );
		name( "call_me_hand_medium_skin_tone", R.drawable.emj_1f919_1f3fd, "\ud83e\udd19\ud83c\udffd" );
		name( "call_me_hand_tone1", R.drawable.emj_1f919_1f3fb, "\ud83e\udd19\ud83c\udffb" );
		name( "call_me_hand_tone2", R.drawable.emj_1f919_1f3fc, "\ud83e\udd19\ud83c\udffc" );
		name( "call_me_hand_tone3", R.drawable.emj_1f919_1f3fd, "\ud83e\udd19\ud83c\udffd" );
		name( "call_me_hand_tone4", R.drawable.emj_1f919_1f3fe, "\ud83e\udd19\ud83c\udffe" );
		name( "call_me_hand_tone5", R.drawable.emj_1f919_1f3ff, "\ud83e\udd19\ud83c\udfff" );
		name( "call_me_tone1", R.drawable.emj_1f919_1f3fb, "\ud83e\udd19\ud83c\udffb" );
		name( "call_me_tone2", R.drawable.emj_1f919_1f3fc, "\ud83e\udd19\ud83c\udffc" );
		name( "call_me_tone3", R.drawable.emj_1f919_1f3fd, "\ud83e\udd19\ud83c\udffd" );
		name( "call_me_tone4", R.drawable.emj_1f919_1f3fe, "\ud83e\udd19\ud83c\udffe" );
		name( "call_me_tone5", R.drawable.emj_1f919_1f3ff, "\ud83e\udd19\ud83c\udfff" );
		name( "calling", R.drawable.emj_1f4f2, "\ud83d\udcf2" );
		name( "camel", R.drawable.emj_1f42b, "\ud83d\udc2b" );
		name( "camera", R.drawable.emj_1f4f7, "\ud83d\udcf7" );
		name( "camera_with_flash", R.drawable.emj_1f4f8, "\ud83d\udcf8" );
		name( "camping", R.drawable.emj_1f3d5_fe0f, "\ud83c\udfd5\ufe0f" );
		name( "cancer", R.drawable.emj_264b, "\u264b" );
		name( "candle", R.drawable.emj_1f56f_fe0f, "\ud83d\udd6f\ufe0f" );
		name( "candy", R.drawable.emj_1f36c, "\ud83c\udf6c" );
		name( "canned_food", R.drawable.emj_1f96b, "\ud83e\udd6b" );
		name( "canoe", R.drawable.emj_1f6f6, "\ud83d\udef6" );
		name( "capital_abcd", R.drawable.emj_1f520, "\ud83d\udd20" );
		name( "capricorn", R.drawable.emj_2651, "\u2651" );
		name( "car", R.drawable.emj_1f697, "\ud83d\ude97" );
		name( "card_box", R.drawable.emj_1f5c3_fe0f, "\ud83d\uddc3\ufe0f" );
		name( "card_file_box", R.drawable.emj_1f5c3_fe0f, "\ud83d\uddc3\ufe0f" );
		name( "card_index", R.drawable.emj_1f4c7, "\ud83d\udcc7" );
		name( "card_index_dividers", R.drawable.emj_1f5c2_fe0f, "\ud83d\uddc2\ufe0f" );
		name( "carousel_horse", R.drawable.emj_1f3a0, "\ud83c\udfa0" );
		name( "carrot", R.drawable.emj_1f955, "\ud83e\udd55" );
		name( "cartwheel", R.drawable.emj_1f938, "\ud83e\udd38" );
		name( "cartwheel_tone1", R.drawable.emj_1f938_1f3fb, "\ud83e\udd38\ud83c\udffb" );
		name( "cartwheel_tone2", R.drawable.emj_1f938_1f3fc, "\ud83e\udd38\ud83c\udffc" );
		name( "cartwheel_tone3", R.drawable.emj_1f938_1f3fd, "\ud83e\udd38\ud83c\udffd" );
		name( "cartwheel_tone4", R.drawable.emj_1f938_1f3fe, "\ud83e\udd38\ud83c\udffe" );
		name( "cartwheel_tone5", R.drawable.emj_1f938_1f3ff, "\ud83e\udd38\ud83c\udfff" );
		name( "cat", R.drawable.emj_1f431, "\ud83d\udc31" );
		name( "cat2", R.drawable.emj_1f408, "\ud83d\udc08" );
		name( "cc", R.drawable.emj_1f1e8_1f1e8, "\ud83c\udde8\ud83c\udde8" );
		name( "cd", R.drawable.emj_1f4bf, "\ud83d\udcbf" );
		name( "cf", R.drawable.emj_1f1e8_1f1eb, "\ud83c\udde8\ud83c\uddeb" );
		name( "cg", R.drawable.emj_1f1e8_1f1ec, "\ud83c\udde8\ud83c\uddec" );
		name( "ch", R.drawable.emj_1f1e8_1f1ed, "\ud83c\udde8\ud83c\udded" );
		name( "chains", R.drawable.emj_26d3_fe0f, "\u26d3\ufe0f" );
		name( "champagne", R.drawable.emj_1f37e, "\ud83c\udf7e" );
		name( "champagne_glass", R.drawable.emj_1f942, "\ud83e\udd42" );
		name( "chart", R.drawable.emj_1f4b9, "\ud83d\udcb9" );
		name( "chart_with_downwards_trend", R.drawable.emj_1f4c9, "\ud83d\udcc9" );
		name( "chart_with_upwards_trend", R.drawable.emj_1f4c8, "\ud83d\udcc8" );
		name( "checkered_flag", R.drawable.emj_1f3c1, "\ud83c\udfc1" );
		name( "cheese", R.drawable.emj_1f9c0, "\ud83e\uddc0" );
		name( "cheese_wedge", R.drawable.emj_1f9c0, "\ud83e\uddc0" );
		name( "cherries", R.drawable.emj_1f352, "\ud83c\udf52" );
		name( "cherry_blossom", R.drawable.emj_1f338, "\ud83c\udf38" );
		name( "chestnut", R.drawable.emj_1f330, "\ud83c\udf30" );
		name( "chicken", R.drawable.emj_1f414, "\ud83d\udc14" );
		name( "child", R.drawable.emj_1f9d2, "\ud83e\uddd2" );
		name( "child_dark_skin_tone", R.drawable.emj_1f9d2_1f3ff, "\ud83e\uddd2\ud83c\udfff" );
		name( "child_light_skin_tone", R.drawable.emj_1f9d2_1f3fb, "\ud83e\uddd2\ud83c\udffb" );
		name( "child_medium_dark_skin_tone", R.drawable.emj_1f9d2_1f3fe, "\ud83e\uddd2\ud83c\udffe" );
		name( "child_medium_light_skin_tone", R.drawable.emj_1f9d2_1f3fc, "\ud83e\uddd2\ud83c\udffc" );
		name( "child_medium_skin_tone", R.drawable.emj_1f9d2_1f3fd, "\ud83e\uddd2\ud83c\udffd" );
		name( "child_tone1", R.drawable.emj_1f9d2_1f3fb, "\ud83e\uddd2\ud83c\udffb" );
		name( "child_tone2", R.drawable.emj_1f9d2_1f3fc, "\ud83e\uddd2\ud83c\udffc" );
		name( "child_tone3", R.drawable.emj_1f9d2_1f3fd, "\ud83e\uddd2\ud83c\udffd" );
		name( "child_tone4", R.drawable.emj_1f9d2_1f3fe, "\ud83e\uddd2\ud83c\udffe" );
		name( "child_tone5", R.drawable.emj_1f9d2_1f3ff, "\ud83e\uddd2\ud83c\udfff" );
		name( "children_crossing", R.drawable.emj_1f6b8, "\ud83d\udeb8" );
		name( "chile", R.drawable.emj_1f1e8_1f1f1, "\ud83c\udde8\ud83c\uddf1" );
		name( "chipmunk", R.drawable.emj_1f43f_fe0f, "\ud83d\udc3f\ufe0f" );
		name( "chocolate_bar", R.drawable.emj_1f36b, "\ud83c\udf6b" );
		name( "chopsticks", R.drawable.emj_1f962, "\ud83e\udd62" );
	}
	private static void init61(){
		name( "christmas_tree", R.drawable.emj_1f384, "\ud83c\udf84" );
		name( "church", R.drawable.emj_26ea, "\u26ea" );
		name( "ci", R.drawable.emj_1f1e8_1f1ee, "\ud83c\udde8\ud83c\uddee" );
		name( "cinema", R.drawable.emj_1f3a6, "\ud83c\udfa6" );
		name( "circus_tent", R.drawable.emj_1f3aa, "\ud83c\udfaa" );
		name( "city_dusk", R.drawable.emj_1f306, "\ud83c\udf06" );
		name( "city_sunrise", R.drawable.emj_1f307, "\ud83c\udf07" );
		name( "city_sunset", R.drawable.emj_1f307, "\ud83c\udf07" );
		name( "cityscape", R.drawable.emj_1f3d9_fe0f, "\ud83c\udfd9\ufe0f" );
		name( "ck", R.drawable.emj_1f1e8_1f1f0, "\ud83c\udde8\ud83c\uddf0" );
		name( "cl", R.drawable.emj_1f191, "\ud83c\udd91" );
		name( "clap", R.drawable.emj_1f44f, "\ud83d\udc4f" );
		name( "clap_dark_skin_tone", R.drawable.emj_1f44f_1f3ff, "\ud83d\udc4f\ud83c\udfff" );
		name( "clap_light_skin_tone", R.drawable.emj_1f44f_1f3fb, "\ud83d\udc4f\ud83c\udffb" );
		name( "clap_medium_dark_skin_tone", R.drawable.emj_1f44f_1f3fe, "\ud83d\udc4f\ud83c\udffe" );
		name( "clap_medium_light_skin_tone", R.drawable.emj_1f44f_1f3fc, "\ud83d\udc4f\ud83c\udffc" );
		name( "clap_medium_skin_tone", R.drawable.emj_1f44f_1f3fd, "\ud83d\udc4f\ud83c\udffd" );
		name( "clap_tone1", R.drawable.emj_1f44f_1f3fb, "\ud83d\udc4f\ud83c\udffb" );
		name( "clap_tone2", R.drawable.emj_1f44f_1f3fc, "\ud83d\udc4f\ud83c\udffc" );
		name( "clap_tone3", R.drawable.emj_1f44f_1f3fd, "\ud83d\udc4f\ud83c\udffd" );
		name( "clap_tone4", R.drawable.emj_1f44f_1f3fe, "\ud83d\udc4f\ud83c\udffe" );
		name( "clap_tone5", R.drawable.emj_1f44f_1f3ff, "\ud83d\udc4f\ud83c\udfff" );
		name( "clapper", R.drawable.emj_1f3ac, "\ud83c\udfac" );
		name( "classical_building", R.drawable.emj_1f3db_fe0f, "\ud83c\udfdb\ufe0f" );
		name( "clinking_glass", R.drawable.emj_1f942, "\ud83e\udd42" );
		name( "clinking_glasses", R.drawable.emj_1f942, "\ud83e\udd42" );
		name( "clipboard", R.drawable.emj_1f4cb, "\ud83d\udccb" );
		name( "clock", R.drawable.emj_1f570_fe0f, "\ud83d\udd70\ufe0f" );
		name( "clock1", R.drawable.emj_1f550, "\ud83d\udd50" );
		name( "clock10", R.drawable.emj_1f559, "\ud83d\udd59" );
		name( "clock1030", R.drawable.emj_1f565, "\ud83d\udd65" );
		name( "clock11", R.drawable.emj_1f55a, "\ud83d\udd5a" );
		name( "clock1130", R.drawable.emj_1f566, "\ud83d\udd66" );
		name( "clock12", R.drawable.emj_1f55b, "\ud83d\udd5b" );
		name( "clock1230", R.drawable.emj_1f567, "\ud83d\udd67" );
		name( "clock130", R.drawable.emj_1f55c, "\ud83d\udd5c" );
		name( "clock2", R.drawable.emj_1f551, "\ud83d\udd51" );
		name( "clock230", R.drawable.emj_1f55d, "\ud83d\udd5d" );
		name( "clock3", R.drawable.emj_1f552, "\ud83d\udd52" );
		name( "clock330", R.drawable.emj_1f55e, "\ud83d\udd5e" );
		name( "clock4", R.drawable.emj_1f553, "\ud83d\udd53" );
		name( "clock430", R.drawable.emj_1f55f, "\ud83d\udd5f" );
		name( "clock5", R.drawable.emj_1f554, "\ud83d\udd54" );
		name( "clock530", R.drawable.emj_1f560, "\ud83d\udd60" );
		name( "clock6", R.drawable.emj_1f555, "\ud83d\udd55" );
		name( "clock630", R.drawable.emj_1f561, "\ud83d\udd61" );
		name( "clock7", R.drawable.emj_1f556, "\ud83d\udd56" );
		name( "clock730", R.drawable.emj_1f562, "\ud83d\udd62" );
		name( "clock8", R.drawable.emj_1f557, "\ud83d\udd57" );
		name( "clock830", R.drawable.emj_1f563, "\ud83d\udd63" );
		name( "clock9", R.drawable.emj_1f558, "\ud83d\udd58" );
		name( "clock930", R.drawable.emj_1f564, "\ud83d\udd64" );
		name( "closed_book", R.drawable.emj_1f4d5, "\ud83d\udcd5" );
		name( "closed_lock_with_key", R.drawable.emj_1f510, "\ud83d\udd10" );
		name( "closed_umbrella", R.drawable.emj_1f302, "\ud83c\udf02" );
		name( "cloud", R.drawable.emj_2601_fe0f, "\u2601\ufe0f" );
		name( "cloud_lightning", R.drawable.emj_1f329_fe0f, "\ud83c\udf29\ufe0f" );
		name( "cloud_rain", R.drawable.emj_1f327_fe0f, "\ud83c\udf27\ufe0f" );
		name( "cloud_snow", R.drawable.emj_1f328_fe0f, "\ud83c\udf28\ufe0f" );
		name( "cloud_tornado", R.drawable.emj_1f32a_fe0f, "\ud83c\udf2a\ufe0f" );
		name( "cloud_with_lightning", R.drawable.emj_1f329_fe0f, "\ud83c\udf29\ufe0f" );
		name( "cloud_with_rain", R.drawable.emj_1f327_fe0f, "\ud83c\udf27\ufe0f" );
		name( "cloud_with_snow", R.drawable.emj_1f328_fe0f, "\ud83c\udf28\ufe0f" );
		name( "cloud_with_tornado", R.drawable.emj_1f32a_fe0f, "\ud83c\udf2a\ufe0f" );
		name( "clown", R.drawable.emj_1f921, "\ud83e\udd21" );
		name( "clown_face", R.drawable.emj_1f921, "\ud83e\udd21" );
		name( "clubs", R.drawable.emj_2663_fe0f, "\u2663\ufe0f" );
		name( "cm", R.drawable.emj_1f1e8_1f1f2, "\ud83c\udde8\ud83c\uddf2" );
		name( "cn", R.drawable.emj_1f1e8_1f1f3, "\ud83c\udde8\ud83c\uddf3" );
		name( "co", R.drawable.emj_1f1e8_1f1f4, "\ud83c\udde8\ud83c\uddf4" );
		name( "coat", R.drawable.emj_1f9e5, "\ud83e\udde5" );
		name( "cocktail", R.drawable.emj_1f378, "\ud83c\udf78" );
		name( "coconut", R.drawable.emj_1f965, "\ud83e\udd65" );
		name( "coffee", R.drawable.emj_2615, "\u2615" );
		name( "coffin", R.drawable.emj_26b0_fe0f, "\u26b0\ufe0f" );
		name( "cold_sweat", R.drawable.emj_1f630, "\ud83d\ude30" );
		name( "collision", R.drawable.emj_1f4a5, "\ud83d\udca5" );
		name( "comet", R.drawable.emj_2604_fe0f, "\u2604\ufe0f" );
		name( "compression", R.drawable.emj_1f5dc_fe0f, "\ud83d\udddc\ufe0f" );
		name( "computer", R.drawable.emj_1f4bb, "\ud83d\udcbb" );
		name( "confetti_ball", R.drawable.emj_1f38a, "\ud83c\udf8a" );
		name( "confounded", R.drawable.emj_1f616, "\ud83d\ude16" );
		name( "confused", R.drawable.emj_1f615, "\ud83d\ude15" );
		name( "congo", R.drawable.emj_1f1e8_1f1e9, "\ud83c\udde8\ud83c\udde9" );
		name( "congratulations", R.drawable.emj_3297_fe0f, "\u3297\ufe0f" );
		name( "construction", R.drawable.emj_1f6a7, "\ud83d\udea7" );
		name( "construction_site", R.drawable.emj_1f3d7_fe0f, "\ud83c\udfd7\ufe0f" );
		name( "construction_worker", R.drawable.emj_1f477, "\ud83d\udc77" );
		name( "construction_worker_dark_skin_tone", R.drawable.emj_1f477_1f3ff, "\ud83d\udc77\ud83c\udfff" );
		name( "construction_worker_light_skin_tone", R.drawable.emj_1f477_1f3fb, "\ud83d\udc77\ud83c\udffb" );
		name( "construction_worker_medium_dark_skin_tone", R.drawable.emj_1f477_1f3fe, "\ud83d\udc77\ud83c\udffe" );
		name( "construction_worker_medium_light_skin_tone", R.drawable.emj_1f477_1f3fc, "\ud83d\udc77\ud83c\udffc" );
		name( "construction_worker_medium_skin_tone", R.drawable.emj_1f477_1f3fd, "\ud83d\udc77\ud83c\udffd" );
		name( "construction_worker_tone1", R.drawable.emj_1f477_1f3fb, "\ud83d\udc77\ud83c\udffb" );
		name( "construction_worker_tone2", R.drawable.emj_1f477_1f3fc, "\ud83d\udc77\ud83c\udffc" );
		name( "construction_worker_tone3", R.drawable.emj_1f477_1f3fd, "\ud83d\udc77\ud83c\udffd" );
		name( "construction_worker_tone4", R.drawable.emj_1f477_1f3fe, "\ud83d\udc77\ud83c\udffe" );
		name( "construction_worker_tone5", R.drawable.emj_1f477_1f3ff, "\ud83d\udc77\ud83c\udfff" );
		name( "control_knobs", R.drawable.emj_1f39b_fe0f, "\ud83c\udf9b\ufe0f" );
		name( "convenience_store", R.drawable.emj_1f3ea, "\ud83c\udfea" );
		name( "cookie", R.drawable.emj_1f36a, "\ud83c\udf6a" );
	}
	private static void init62(){
		name( "cooking", R.drawable.emj_1f373, "\ud83c\udf73" );
		name( "cool", R.drawable.emj_1f192, "\ud83c\udd92" );
		name( "cop", R.drawable.emj_1f46e, "\ud83d\udc6e" );
		name( "cop_dark_skin_tone", R.drawable.emj_1f46e_1f3ff, "\ud83d\udc6e\ud83c\udfff" );
		name( "cop_light_skin_tone", R.drawable.emj_1f46e_1f3fb, "\ud83d\udc6e\ud83c\udffb" );
		name( "cop_medium_dark_skin_tone", R.drawable.emj_1f46e_1f3fe, "\ud83d\udc6e\ud83c\udffe" );
		name( "cop_medium_light_skin_tone", R.drawable.emj_1f46e_1f3fc, "\ud83d\udc6e\ud83c\udffc" );
		name( "cop_medium_skin_tone", R.drawable.emj_1f46e_1f3fd, "\ud83d\udc6e\ud83c\udffd" );
		name( "cop_tone1", R.drawable.emj_1f46e_1f3fb, "\ud83d\udc6e\ud83c\udffb" );
		name( "cop_tone2", R.drawable.emj_1f46e_1f3fc, "\ud83d\udc6e\ud83c\udffc" );
		name( "cop_tone3", R.drawable.emj_1f46e_1f3fd, "\ud83d\udc6e\ud83c\udffd" );
		name( "cop_tone4", R.drawable.emj_1f46e_1f3fe, "\ud83d\udc6e\ud83c\udffe" );
		name( "cop_tone5", R.drawable.emj_1f46e_1f3ff, "\ud83d\udc6e\ud83c\udfff" );
		name( "copyright", R.drawable.emj_00a9_fe0f, "\u00a9\ufe0f" );
		name( "corn", R.drawable.emj_1f33d, "\ud83c\udf3d" );
		name( "couch", R.drawable.emj_1f6cb_fe0f, "\ud83d\udecb\ufe0f" );
		name( "couch_and_lamp", R.drawable.emj_1f6cb_fe0f, "\ud83d\udecb\ufe0f" );
		name( "couple", R.drawable.emj_1f46b, "\ud83d\udc6b" );
		name( "couple_mm", R.drawable.emj_1f468_200d_2764_fe0f_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc68" );
		name( "couple_with_heart", R.drawable.emj_1f491, "\ud83d\udc91" );
		name( "couple_with_heart_mm", R.drawable.emj_1f468_200d_2764_fe0f_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc68" );
		name( "couple_with_heart_woman_man", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f468, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc68" );
		name( "couple_with_heart_ww", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc69" );
		name( "couple_ww", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc69" );
		name( "couplekiss", R.drawable.emj_1f48f, "\ud83d\udc8f" );
		name( "couplekiss_mm", R.drawable.emj_1f468_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68" );
		name( "couplekiss_ww", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69" );
		name( "cow", R.drawable.emj_1f42e, "\ud83d\udc2e" );
		name( "cow2", R.drawable.emj_1f404, "\ud83d\udc04" );
		name( "cowboy", R.drawable.emj_1f920, "\ud83e\udd20" );
		name( "cp", R.drawable.emj_1f1e8_1f1f5, "\ud83c\udde8\ud83c\uddf5" );
		name( "cr", R.drawable.emj_1f1e8_1f1f7, "\ud83c\udde8\ud83c\uddf7" );
		name( "crab", R.drawable.emj_1f980, "\ud83e\udd80" );
		name( "crayon", R.drawable.emj_1f58d_fe0f, "\ud83d\udd8d\ufe0f" );
		name( "credit_card", R.drawable.emj_1f4b3, "\ud83d\udcb3" );
		name( "crescent_moon", R.drawable.emj_1f319, "\ud83c\udf19" );
		name( "cricket", R.drawable.emj_1f997, "\ud83e\udd97" );
		name( "cricket_bat_and_ball", R.drawable.emj_1f3cf, "\ud83c\udfcf" );
		name( "cricket_bat_ball", R.drawable.emj_1f3cf, "\ud83c\udfcf" );
		name( "crocodile", R.drawable.emj_1f40a, "\ud83d\udc0a" );
		name( "croissant", R.drawable.emj_1f950, "\ud83e\udd50" );
		name( "cross", R.drawable.emj_271d_fe0f, "\u271d\ufe0f" );
		name( "crossed_fingers", R.drawable.emj_1f91e, "\ud83e\udd1e" );
		name( "crossed_fingers_dark_skin_tone", R.drawable.emj_1f91e_1f3ff, "\ud83e\udd1e\ud83c\udfff" );
		name( "crossed_fingers_light_skin_tone", R.drawable.emj_1f91e_1f3fb, "\ud83e\udd1e\ud83c\udffb" );
		name( "crossed_fingers_medium_dark_skin_tone", R.drawable.emj_1f91e_1f3fe, "\ud83e\udd1e\ud83c\udffe" );
		name( "crossed_fingers_medium_light_skin_tone", R.drawable.emj_1f91e_1f3fc, "\ud83e\udd1e\ud83c\udffc" );
		name( "crossed_fingers_medium_skin_tone", R.drawable.emj_1f91e_1f3fd, "\ud83e\udd1e\ud83c\udffd" );
		name( "crossed_fingers_tone1", R.drawable.emj_1f91e_1f3fb, "\ud83e\udd1e\ud83c\udffb" );
		name( "crossed_fingers_tone2", R.drawable.emj_1f91e_1f3fc, "\ud83e\udd1e\ud83c\udffc" );
		name( "crossed_fingers_tone3", R.drawable.emj_1f91e_1f3fd, "\ud83e\udd1e\ud83c\udffd" );
		name( "crossed_fingers_tone4", R.drawable.emj_1f91e_1f3fe, "\ud83e\udd1e\ud83c\udffe" );
		name( "crossed_fingers_tone5", R.drawable.emj_1f91e_1f3ff, "\ud83e\udd1e\ud83c\udfff" );
		name( "crossed_flags", R.drawable.emj_1f38c, "\ud83c\udf8c" );
		name( "crossed_swords", R.drawable.emj_2694_fe0f, "\u2694\ufe0f" );
		name( "crown", R.drawable.emj_1f451, "\ud83d\udc51" );
		name( "cruise_ship", R.drawable.emj_1f6f3_fe0f, "\ud83d\udef3\ufe0f" );
		name( "cry", R.drawable.emj_1f622, "\ud83d\ude22" );
		name( "crying_cat_face", R.drawable.emj_1f63f, "\ud83d\ude3f" );
		name( "crystal_ball", R.drawable.emj_1f52e, "\ud83d\udd2e" );
		name( "cu", R.drawable.emj_1f1e8_1f1fa, "\ud83c\udde8\ud83c\uddfa" );
		name( "cucumber", R.drawable.emj_1f952, "\ud83e\udd52" );
		name( "cup_with_straw", R.drawable.emj_1f964, "\ud83e\udd64" );
		name( "cupid", R.drawable.emj_1f498, "\ud83d\udc98" );
		name( "curling_stone", R.drawable.emj_1f94c, "\ud83e\udd4c" );
		name( "curly_loop", R.drawable.emj_27b0, "\u27b0" );
		name( "currency_exchange", R.drawable.emj_1f4b1, "\ud83d\udcb1" );
		name( "curry", R.drawable.emj_1f35b, "\ud83c\udf5b" );
		name( "custard", R.drawable.emj_1f36e, "\ud83c\udf6e" );
		name( "customs", R.drawable.emj_1f6c3, "\ud83d\udec3" );
		name( "cut_of_meat", R.drawable.emj_1f969, "\ud83e\udd69" );
		name( "cv", R.drawable.emj_1f1e8_1f1fb, "\ud83c\udde8\ud83c\uddfb" );
		name( "cw", R.drawable.emj_1f1e8_1f1fc, "\ud83c\udde8\ud83c\uddfc" );
		name( "cx", R.drawable.emj_1f1e8_1f1fd, "\ud83c\udde8\ud83c\uddfd" );
		name( "cy", R.drawable.emj_1f1e8_1f1fe, "\ud83c\udde8\ud83c\uddfe" );
		name( "cyclone", R.drawable.emj_1f300, "\ud83c\udf00" );
		name( "cz", R.drawable.emj_1f1e8_1f1ff, "\ud83c\udde8\ud83c\uddff" );
		name( "dagger", R.drawable.emj_1f5e1_fe0f, "\ud83d\udde1\ufe0f" );
		name( "dagger_knife", R.drawable.emj_1f5e1_fe0f, "\ud83d\udde1\ufe0f" );
		name( "dancer", R.drawable.emj_1f483, "\ud83d\udc83" );
		name( "dancer_dark_skin_tone", R.drawable.emj_1f483_1f3ff, "\ud83d\udc83\ud83c\udfff" );
		name( "dancer_light_skin_tone", R.drawable.emj_1f483_1f3fb, "\ud83d\udc83\ud83c\udffb" );
		name( "dancer_medium_dark_skin_tone", R.drawable.emj_1f483_1f3fe, "\ud83d\udc83\ud83c\udffe" );
		name( "dancer_medium_light_skin_tone", R.drawable.emj_1f483_1f3fc, "\ud83d\udc83\ud83c\udffc" );
		name( "dancer_medium_skin_tone", R.drawable.emj_1f483_1f3fd, "\ud83d\udc83\ud83c\udffd" );
		name( "dancer_tone1", R.drawable.emj_1f483_1f3fb, "\ud83d\udc83\ud83c\udffb" );
		name( "dancer_tone2", R.drawable.emj_1f483_1f3fc, "\ud83d\udc83\ud83c\udffc" );
		name( "dancer_tone3", R.drawable.emj_1f483_1f3fd, "\ud83d\udc83\ud83c\udffd" );
		name( "dancer_tone4", R.drawable.emj_1f483_1f3fe, "\ud83d\udc83\ud83c\udffe" );
		name( "dancer_tone5", R.drawable.emj_1f483_1f3ff, "\ud83d\udc83\ud83c\udfff" );
		name( "dancers", R.drawable.emj_1f46f, "\ud83d\udc6f" );
		name( "dango", R.drawable.emj_1f361, "\ud83c\udf61" );
		name( "dark_sunglasses", R.drawable.emj_1f576_fe0f, "\ud83d\udd76\ufe0f" );
		name( "dart", R.drawable.emj_1f3af, "\ud83c\udfaf" );
		name( "dash", R.drawable.emj_1f4a8, "\ud83d\udca8" );
		name( "date", R.drawable.emj_1f4c5, "\ud83d\udcc5" );
		name( "de", R.drawable.emj_1f1e9_1f1ea, "\ud83c\udde9\ud83c\uddea" );
		name( "deciduous_tree", R.drawable.emj_1f333, "\ud83c\udf33" );
		name( "deer", R.drawable.emj_1f98c, "\ud83e\udd8c" );
		name( "department_store", R.drawable.emj_1f3ec, "\ud83c\udfec" );
		name( "derelict_house_building", R.drawable.emj_1f3da_fe0f, "\ud83c\udfda\ufe0f" );
	}
	private static void init63(){
		name( "desert", R.drawable.emj_1f3dc_fe0f, "\ud83c\udfdc\ufe0f" );
		name( "desert_island", R.drawable.emj_1f3dd_fe0f, "\ud83c\udfdd\ufe0f" );
		name( "desktop", R.drawable.emj_1f5a5_fe0f, "\ud83d\udda5\ufe0f" );
		name( "desktop_computer", R.drawable.emj_1f5a5_fe0f, "\ud83d\udda5\ufe0f" );
		name( "dg", R.drawable.emj_1f1e9_1f1ec, "\ud83c\udde9\ud83c\uddec" );
		name( "diamond_shape_with_a_dot_inside", R.drawable.emj_1f4a0, "\ud83d\udca0" );
		name( "diamonds", R.drawable.emj_2666_fe0f, "\u2666\ufe0f" );
		name( "digit_eight", R.drawable.emj_0038, "\u0038" );
		name( "digit_five", R.drawable.emj_0035, "\u0035" );
		name( "digit_four", R.drawable.emj_0034, "\u0034" );
		name( "digit_nine", R.drawable.emj_0039, "\u0039" );
		name( "digit_one", R.drawable.emj_0031, "\u0031" );
		name( "digit_seven", R.drawable.emj_0037, "\u0037" );
		name( "digit_six", R.drawable.emj_0036, "\u0036" );
		name( "digit_three", R.drawable.emj_0033, "\u0033" );
		name( "digit_two", R.drawable.emj_0032, "\u0032" );
		name( "digit_zero", R.drawable.emj_0030, "\u0030" );
		name( "disappointed", R.drawable.emj_1f61e, "\ud83d\ude1e" );
		name( "disappointed_relieved", R.drawable.emj_1f625, "\ud83d\ude25" );
		name( "dividers", R.drawable.emj_1f5c2_fe0f, "\ud83d\uddc2\ufe0f" );
		name( "dizzy", R.drawable.emj_1f4ab, "\ud83d\udcab" );
		name( "dizzy_face", R.drawable.emj_1f635, "\ud83d\ude35" );
		name( "dj", R.drawable.emj_1f1e9_1f1ef, "\ud83c\udde9\ud83c\uddef" );
		name( "dk", R.drawable.emj_1f1e9_1f1f0, "\ud83c\udde9\ud83c\uddf0" );
		name( "dm", R.drawable.emj_1f1e9_1f1f2, "\ud83c\udde9\ud83c\uddf2" );
		name( "do", R.drawable.emj_1f1e9_1f1f4, "\ud83c\udde9\ud83c\uddf4" );
		name( "do_not_litter", R.drawable.emj_1f6af, "\ud83d\udeaf" );
		name( "dog", R.drawable.emj_1f436, "\ud83d\udc36" );
		name( "dog2", R.drawable.emj_1f415, "\ud83d\udc15" );
		name( "dollar", R.drawable.emj_1f4b5, "\ud83d\udcb5" );
		name( "dolls", R.drawable.emj_1f38e, "\ud83c\udf8e" );
		name( "dolphin", R.drawable.emj_1f42c, "\ud83d\udc2c" );
		name( "door", R.drawable.emj_1f6aa, "\ud83d\udeaa" );
		name( "double_vertical_bar", R.drawable.emj_23f8_fe0f, "\u23f8\ufe0f" );
		name( "doughnut", R.drawable.emj_1f369, "\ud83c\udf69" );
		name( "dove", R.drawable.emj_1f54a_fe0f, "\ud83d\udd4a\ufe0f" );
		name( "dove_of_peace", R.drawable.emj_1f54a_fe0f, "\ud83d\udd4a\ufe0f" );
		name( "dragon", R.drawable.emj_1f409, "\ud83d\udc09" );
		name( "dragon_face", R.drawable.emj_1f432, "\ud83d\udc32" );
		name( "dress", R.drawable.emj_1f457, "\ud83d\udc57" );
		name( "dromedary_camel", R.drawable.emj_1f42a, "\ud83d\udc2a" );
		name( "drool", R.drawable.emj_1f924, "\ud83e\udd24" );
		name( "drooling_face", R.drawable.emj_1f924, "\ud83e\udd24" );
		name( "droplet", R.drawable.emj_1f4a7, "\ud83d\udca7" );
		name( "drum", R.drawable.emj_1f941, "\ud83e\udd41" );
		name( "drum_with_drumsticks", R.drawable.emj_1f941, "\ud83e\udd41" );
		name( "duck", R.drawable.emj_1f986, "\ud83e\udd86" );
		name( "dumpling", R.drawable.emj_1f95f, "\ud83e\udd5f" );
		name( "dvd", R.drawable.emj_1f4c0, "\ud83d\udcc0" );
		name( "dz", R.drawable.emj_1f1e9_1f1ff, "\ud83c\udde9\ud83c\uddff" );
		name( "e_mail", R.drawable.emj_1f4e7, "\ud83d\udce7" );
		name( "ea", R.drawable.emj_1f1ea_1f1e6, "\ud83c\uddea\ud83c\udde6" );
		name( "eagle", R.drawable.emj_1f985, "\ud83e\udd85" );
		name( "ear", R.drawable.emj_1f442, "\ud83d\udc42" );
		name( "ear_dark_skin_tone", R.drawable.emj_1f442_1f3ff, "\ud83d\udc42\ud83c\udfff" );
		name( "ear_light_skin_tone", R.drawable.emj_1f442_1f3fb, "\ud83d\udc42\ud83c\udffb" );
		name( "ear_medium_dark_skin_tone", R.drawable.emj_1f442_1f3fe, "\ud83d\udc42\ud83c\udffe" );
		name( "ear_medium_light_skin_tone", R.drawable.emj_1f442_1f3fc, "\ud83d\udc42\ud83c\udffc" );
		name( "ear_medium_skin_tone", R.drawable.emj_1f442_1f3fd, "\ud83d\udc42\ud83c\udffd" );
		name( "ear_of_rice", R.drawable.emj_1f33e, "\ud83c\udf3e" );
		name( "ear_tone1", R.drawable.emj_1f442_1f3fb, "\ud83d\udc42\ud83c\udffb" );
		name( "ear_tone2", R.drawable.emj_1f442_1f3fc, "\ud83d\udc42\ud83c\udffc" );
		name( "ear_tone3", R.drawable.emj_1f442_1f3fd, "\ud83d\udc42\ud83c\udffd" );
		name( "ear_tone4", R.drawable.emj_1f442_1f3fe, "\ud83d\udc42\ud83c\udffe" );
		name( "ear_tone5", R.drawable.emj_1f442_1f3ff, "\ud83d\udc42\ud83c\udfff" );
		name( "earth_africa", R.drawable.emj_1f30d, "\ud83c\udf0d" );
		name( "earth_americas", R.drawable.emj_1f30e, "\ud83c\udf0e" );
		name( "earth_asia", R.drawable.emj_1f30f, "\ud83c\udf0f" );
		name( "ec", R.drawable.emj_1f1ea_1f1e8, "\ud83c\uddea\ud83c\udde8" );
		name( "ee", R.drawable.emj_1f1ea_1f1ea, "\ud83c\uddea\ud83c\uddea" );
		name( "eg", R.drawable.emj_1f1ea_1f1ec, "\ud83c\uddea\ud83c\uddec" );
		name( "egg", R.drawable.emj_1f95a, "\ud83e\udd5a" );
		name( "eggplant", R.drawable.emj_1f346, "\ud83c\udf46" );
		name( "eh", R.drawable.emj_1f1ea_1f1ed, "\ud83c\uddea\ud83c\udded" );
		name( "eight", R.drawable.emj_0038_fe0f_20e3, "\u0038\ufe0f\u20e3" );
		name( "eight_pointed_black_star", R.drawable.emj_2734_fe0f, "\u2734\ufe0f" );
		name( "eight_spoked_asterisk", R.drawable.emj_2733_fe0f, "\u2733\ufe0f" );
		name( "eject", R.drawable.emj_23cf_fe0f, "\u23cf\ufe0f" );
		name( "eject_symbol", R.drawable.emj_23cf_fe0f, "\u23cf\ufe0f" );
		name( "electric_plug", R.drawable.emj_1f50c, "\ud83d\udd0c" );
		name( "elephant", R.drawable.emj_1f418, "\ud83d\udc18" );
		name( "elf", R.drawable.emj_1f9dd, "\ud83e\udddd" );
		name( "elf_dark_skin_tone", R.drawable.emj_1f9dd_1f3ff, "\ud83e\udddd\ud83c\udfff" );
		name( "elf_light_skin_tone", R.drawable.emj_1f9dd_1f3fb, "\ud83e\udddd\ud83c\udffb" );
		name( "elf_medium_dark_skin_tone", R.drawable.emj_1f9dd_1f3fe, "\ud83e\udddd\ud83c\udffe" );
		name( "elf_medium_light_skin_tone", R.drawable.emj_1f9dd_1f3fc, "\ud83e\udddd\ud83c\udffc" );
		name( "elf_medium_skin_tone", R.drawable.emj_1f9dd_1f3fd, "\ud83e\udddd\ud83c\udffd" );
		name( "elf_tone1", R.drawable.emj_1f9dd_1f3fb, "\ud83e\udddd\ud83c\udffb" );
		name( "elf_tone2", R.drawable.emj_1f9dd_1f3fc, "\ud83e\udddd\ud83c\udffc" );
		name( "elf_tone3", R.drawable.emj_1f9dd_1f3fd, "\ud83e\udddd\ud83c\udffd" );
		name( "elf_tone4", R.drawable.emj_1f9dd_1f3fe, "\ud83e\udddd\ud83c\udffe" );
		name( "elf_tone5", R.drawable.emj_1f9dd_1f3ff, "\ud83e\udddd\ud83c\udfff" );
		name( "email", R.drawable.emj_1f4e7, "\ud83d\udce7" );
		name( "end", R.drawable.emj_1f51a, "\ud83d\udd1a" );
		name( "envelope", R.drawable.emj_2709_fe0f, "\u2709\ufe0f" );
		name( "envelope_with_arrow", R.drawable.emj_1f4e9, "\ud83d\udce9" );
		name( "er", R.drawable.emj_1f1ea_1f1f7, "\ud83c\uddea\ud83c\uddf7" );
		name( "es", R.drawable.emj_1f1ea_1f1f8, "\ud83c\uddea\ud83c\uddf8" );
		name( "et", R.drawable.emj_1f1ea_1f1f9, "\ud83c\uddea\ud83c\uddf9" );
		name( "eu", R.drawable.emj_1f1ea_1f1fa, "\ud83c\uddea\ud83c\uddfa" );
		name( "euro", R.drawable.emj_1f4b6, "\ud83d\udcb6" );
	}
	private static void init64(){
		name( "european_castle", R.drawable.emj_1f3f0, "\ud83c\udff0" );
		name( "european_post_office", R.drawable.emj_1f3e4, "\ud83c\udfe4" );
		name( "evergreen_tree", R.drawable.emj_1f332, "\ud83c\udf32" );
		name( "exclamation", R.drawable.emj_2757, "\u2757" );
		name( "expecting_woman", R.drawable.emj_1f930, "\ud83e\udd30" );
		name( "expecting_woman_tone1", R.drawable.emj_1f930_1f3fb, "\ud83e\udd30\ud83c\udffb" );
		name( "expecting_woman_tone2", R.drawable.emj_1f930_1f3fc, "\ud83e\udd30\ud83c\udffc" );
		name( "expecting_woman_tone3", R.drawable.emj_1f930_1f3fd, "\ud83e\udd30\ud83c\udffd" );
		name( "expecting_woman_tone4", R.drawable.emj_1f930_1f3fe, "\ud83e\udd30\ud83c\udffe" );
		name( "expecting_woman_tone5", R.drawable.emj_1f930_1f3ff, "\ud83e\udd30\ud83c\udfff" );
		name( "exploding_head", R.drawable.emj_1f92f, "\ud83e\udd2f" );
		name( "expressionless", R.drawable.emj_1f611, "\ud83d\ude11" );
		name( "eye", R.drawable.emj_1f441_fe0f, "\ud83d\udc41\ufe0f" );
		name( "eye_in_speech_bubble", R.drawable.emj_1f441_fe0f_200d_1f5e8_fe0f, "\ud83d\udc41\ufe0f\u200d\ud83d\udde8\ufe0f" );
		name( "eyeglasses", R.drawable.emj_1f453, "\ud83d\udc53" );
		name( "eyes", R.drawable.emj_1f440, "\ud83d\udc40" );
		name( "face_palm", R.drawable.emj_1f926, "\ud83e\udd26" );
		name( "face_palm_dark_skin_tone", R.drawable.emj_1f926_1f3ff, "\ud83e\udd26\ud83c\udfff" );
		name( "face_palm_light_skin_tone", R.drawable.emj_1f926_1f3fb, "\ud83e\udd26\ud83c\udffb" );
		name( "face_palm_medium_dark_skin_tone", R.drawable.emj_1f926_1f3fe, "\ud83e\udd26\ud83c\udffe" );
		name( "face_palm_medium_light_skin_tone", R.drawable.emj_1f926_1f3fc, "\ud83e\udd26\ud83c\udffc" );
		name( "face_palm_medium_skin_tone", R.drawable.emj_1f926_1f3fd, "\ud83e\udd26\ud83c\udffd" );
		name( "face_palm_tone1", R.drawable.emj_1f926_1f3fb, "\ud83e\udd26\ud83c\udffb" );
		name( "face_palm_tone2", R.drawable.emj_1f926_1f3fc, "\ud83e\udd26\ud83c\udffc" );
		name( "face_palm_tone3", R.drawable.emj_1f926_1f3fd, "\ud83e\udd26\ud83c\udffd" );
		name( "face_palm_tone4", R.drawable.emj_1f926_1f3fe, "\ud83e\udd26\ud83c\udffe" );
		name( "face_palm_tone5", R.drawable.emj_1f926_1f3ff, "\ud83e\udd26\ud83c\udfff" );
		name( "face_vomiting", R.drawable.emj_1f92e, "\ud83e\udd2e" );
		name( "face_with_cowboy_hat", R.drawable.emj_1f920, "\ud83e\udd20" );
		name( "face_with_finger_covering_closed_lips", R.drawable.emj_1f92b, "\ud83e\udd2b" );
		name( "face_with_hand_over_mouth", R.drawable.emj_1f92d, "\ud83e\udd2d" );
		name( "face_with_head_bandage", R.drawable.emj_1f915, "\ud83e\udd15" );
		name( "face_with_monocle", R.drawable.emj_1f9d0, "\ud83e\uddd0" );
		name( "face_with_one_eyebrow_raised", R.drawable.emj_1f928, "\ud83e\udd28" );
		name( "face_with_open_mouth_vomiting", R.drawable.emj_1f92e, "\ud83e\udd2e" );
		name( "face_with_raised_eyebrow", R.drawable.emj_1f928, "\ud83e\udd28" );
		name( "face_with_rolling_eyes", R.drawable.emj_1f644, "\ud83d\ude44" );
		name( "face_with_symbols_on_mouth", R.drawable.emj_1f92c, "\ud83e\udd2c" );
		name( "face_with_thermometer", R.drawable.emj_1f912, "\ud83e\udd12" );
		name( "facepalm", R.drawable.emj_1f926, "\ud83e\udd26" );
		name( "facepalm_tone1", R.drawable.emj_1f926_1f3fb, "\ud83e\udd26\ud83c\udffb" );
		name( "facepalm_tone2", R.drawable.emj_1f926_1f3fc, "\ud83e\udd26\ud83c\udffc" );
		name( "facepalm_tone3", R.drawable.emj_1f926_1f3fd, "\ud83e\udd26\ud83c\udffd" );
		name( "facepalm_tone4", R.drawable.emj_1f926_1f3fe, "\ud83e\udd26\ud83c\udffe" );
		name( "facepalm_tone5", R.drawable.emj_1f926_1f3ff, "\ud83e\udd26\ud83c\udfff" );
		name( "facepunch", R.drawable.emj_1f44a, "\ud83d\udc4a" );
		name( "facepunch_dark_skin_tone", R.drawable.emj_1f44a_1f3ff, "\ud83d\udc4a\ud83c\udfff" );
		name( "facepunch_light_skin_tone", R.drawable.emj_1f44a_1f3fb, "\ud83d\udc4a\ud83c\udffb" );
		name( "facepunch_medium_dark_skin_tone", R.drawable.emj_1f44a_1f3fe, "\ud83d\udc4a\ud83c\udffe" );
		name( "facepunch_medium_light_skin_tone", R.drawable.emj_1f44a_1f3fc, "\ud83d\udc4a\ud83c\udffc" );
		name( "facepunch_medium_skin_tone", R.drawable.emj_1f44a_1f3fd, "\ud83d\udc4a\ud83c\udffd" );
		name( "facepunch_tone1", R.drawable.emj_1f44a_1f3fb, "\ud83d\udc4a\ud83c\udffb" );
		name( "facepunch_tone2", R.drawable.emj_1f44a_1f3fc, "\ud83d\udc4a\ud83c\udffc" );
		name( "facepunch_tone3", R.drawable.emj_1f44a_1f3fd, "\ud83d\udc4a\ud83c\udffd" );
		name( "facepunch_tone4", R.drawable.emj_1f44a_1f3fe, "\ud83d\udc4a\ud83c\udffe" );
		name( "facepunch_tone5", R.drawable.emj_1f44a_1f3ff, "\ud83d\udc4a\ud83c\udfff" );
		name( "factory", R.drawable.emj_1f3ed, "\ud83c\udfed" );
		name( "fairy", R.drawable.emj_1f9da, "\ud83e\uddda" );
		name( "fairy_dark_skin_tone", R.drawable.emj_1f9da_1f3ff, "\ud83e\uddda\ud83c\udfff" );
		name( "fairy_light_skin_tone", R.drawable.emj_1f9da_1f3fb, "\ud83e\uddda\ud83c\udffb" );
		name( "fairy_medium_dark_skin_tone", R.drawable.emj_1f9da_1f3fe, "\ud83e\uddda\ud83c\udffe" );
		name( "fairy_medium_light_skin_tone", R.drawable.emj_1f9da_1f3fc, "\ud83e\uddda\ud83c\udffc" );
		name( "fairy_medium_skin_tone", R.drawable.emj_1f9da_1f3fd, "\ud83e\uddda\ud83c\udffd" );
		name( "fairy_tone1", R.drawable.emj_1f9da_1f3fb, "\ud83e\uddda\ud83c\udffb" );
		name( "fairy_tone2", R.drawable.emj_1f9da_1f3fc, "\ud83e\uddda\ud83c\udffc" );
		name( "fairy_tone3", R.drawable.emj_1f9da_1f3fd, "\ud83e\uddda\ud83c\udffd" );
		name( "fairy_tone4", R.drawable.emj_1f9da_1f3fe, "\ud83e\uddda\ud83c\udffe" );
		name( "fairy_tone5", R.drawable.emj_1f9da_1f3ff, "\ud83e\uddda\ud83c\udfff" );
		name( "fallen_leaf", R.drawable.emj_1f342, "\ud83c\udf42" );
		name( "family", R.drawable.emj_1f46a, "\ud83d\udc6a" );
		name( "family_man_boy", R.drawable.emj_1f468_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc66" );
		name( "family_man_boy_boy", R.drawable.emj_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "family_man_girl", R.drawable.emj_1f468_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc67" );
		name( "family_man_girl_boy", R.drawable.emj_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "family_man_girl_girl", R.drawable.emj_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "family_man_woman_boy", R.drawable.emj_1f468_200d_1f469_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc66" );
		name( "family_mmb", R.drawable.emj_1f468_200d_1f468_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc66" );
		name( "family_mmbb", R.drawable.emj_1f468_200d_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "family_mmg", R.drawable.emj_1f468_200d_1f468_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67" );
		name( "family_mmgb", R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "family_mmgg", R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "family_mwbb", R.drawable.emj_1f468_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "family_mwg", R.drawable.emj_1f468_200d_1f469_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67" );
		name( "family_mwgb", R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "family_mwgg", R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "family_woman_boy", R.drawable.emj_1f469_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc66" );
		name( "family_woman_boy_boy", R.drawable.emj_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "family_woman_girl", R.drawable.emj_1f469_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc67" );
		name( "family_woman_girl_boy", R.drawable.emj_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "family_woman_girl_girl", R.drawable.emj_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "family_wwb", R.drawable.emj_1f469_200d_1f469_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc66" );
		name( "family_wwbb", R.drawable.emj_1f469_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "family_wwg", R.drawable.emj_1f469_200d_1f469_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67" );
		name( "family_wwgb", R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "family_wwgg", R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "fast_forward", R.drawable.emj_23e9, "\u23e9" );
		name( "fax", R.drawable.emj_1f4e0, "\ud83d\udce0" );
		name( "fearful", R.drawable.emj_1f628, "\ud83d\ude28" );
		name( "feet", R.drawable.emj_1f43e, "\ud83d\udc3e" );
		name( "female_artist", R.drawable.emj_1f469_200d_1f3a8, "\ud83d\udc69\u200d\ud83c\udfa8" );
		name( "female_artist_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3a8, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa8" );
	}
	private static void init65(){
		name( "female_artist_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3a8, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "female_artist_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3a8, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "female_artist_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3a8, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "female_artist_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3a8, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "female_artist_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3a8, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "female_artist_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3a8, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "female_artist_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3a8, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "female_artist_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3a8, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "female_artist_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3a8, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa8" );
		name( "female_astronaut", R.drawable.emj_1f469_200d_1f680, "\ud83d\udc69\u200d\ud83d\ude80" );
		name( "female_astronaut_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f680, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude80" );
		name( "female_astronaut_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f680, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude80" );
		name( "female_astronaut_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f680, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude80" );
		name( "female_astronaut_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f680, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude80" );
		name( "female_astronaut_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f680, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude80" );
		name( "female_astronaut_tone1", R.drawable.emj_1f469_1f3fb_200d_1f680, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude80" );
		name( "female_astronaut_tone2", R.drawable.emj_1f469_1f3fc_200d_1f680, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude80" );
		name( "female_astronaut_tone3", R.drawable.emj_1f469_1f3fd_200d_1f680, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude80" );
		name( "female_astronaut_tone4", R.drawable.emj_1f469_1f3fe_200d_1f680, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude80" );
		name( "female_astronaut_tone5", R.drawable.emj_1f469_1f3ff_200d_1f680, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude80" );
		name( "female_construction_worker", R.drawable.emj_1f477_200d_2640_fe0f, "\ud83d\udc77\u200d\u2640\ufe0f" );
		name( "female_construction_worker_dark_skin_tone", R.drawable.emj_1f477_1f3ff_200d_2640_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_construction_worker_light_skin_tone", R.drawable.emj_1f477_1f3fb_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_construction_worker_medium_dark_skin_tone", R.drawable.emj_1f477_1f3fe_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_construction_worker_medium_light_skin_tone", R.drawable.emj_1f477_1f3fc_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_construction_worker_medium_skin_tone", R.drawable.emj_1f477_1f3fd_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_construction_worker_tone1", R.drawable.emj_1f477_1f3fb_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_construction_worker_tone2", R.drawable.emj_1f477_1f3fc_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_construction_worker_tone3", R.drawable.emj_1f477_1f3fd_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_construction_worker_tone4", R.drawable.emj_1f477_1f3fe_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_construction_worker_tone5", R.drawable.emj_1f477_1f3ff_200d_2640_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_cook", R.drawable.emj_1f469_200d_1f373, "\ud83d\udc69\u200d\ud83c\udf73" );
		name( "female_cook_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f373, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf73" );
		name( "female_cook_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f373, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf73" );
		name( "female_cook_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f373, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf73" );
		name( "female_cook_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f373, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf73" );
		name( "female_cook_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f373, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf73" );
		name( "female_cook_tone1", R.drawable.emj_1f469_1f3fb_200d_1f373, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf73" );
		name( "female_cook_tone2", R.drawable.emj_1f469_1f3fc_200d_1f373, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf73" );
		name( "female_cook_tone3", R.drawable.emj_1f469_1f3fd_200d_1f373, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf73" );
		name( "female_cook_tone4", R.drawable.emj_1f469_1f3fe_200d_1f373, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf73" );
		name( "female_cook_tone5", R.drawable.emj_1f469_1f3ff_200d_1f373, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf73" );
		name( "female_detective", R.drawable.emj_1f575_fe0f_200d_2640_fe0f, "\ud83d\udd75\ufe0f\u200d\u2640\ufe0f" );
		name( "female_detective_dark_skin_tone", R.drawable.emj_1f575_1f3ff_200d_2640_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_detective_light_skin_tone", R.drawable.emj_1f575_1f3fb_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_detective_medium_dark_skin_tone", R.drawable.emj_1f575_1f3fe_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_detective_medium_light_skin_tone", R.drawable.emj_1f575_1f3fc_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_detective_medium_skin_tone", R.drawable.emj_1f575_1f3fd_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_detective_tone1", R.drawable.emj_1f575_1f3fb_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_detective_tone2", R.drawable.emj_1f575_1f3fc_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_detective_tone3", R.drawable.emj_1f575_1f3fd_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_detective_tone4", R.drawable.emj_1f575_1f3fe_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_detective_tone5", R.drawable.emj_1f575_1f3ff_200d_2640_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_doctor", R.drawable.emj_1f469_200d_2695_fe0f, "\ud83d\udc69\u200d\u2695\ufe0f" );
		name( "female_doctor_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_2695_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "female_doctor_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "female_doctor_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "female_doctor_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2695\ufe0f" );
		name( "female_doctor_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "female_doctor_tone1", R.drawable.emj_1f469_1f3fb_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "female_doctor_tone2", R.drawable.emj_1f469_1f3fc_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2695\ufe0f" );
		name( "female_doctor_tone3", R.drawable.emj_1f469_1f3fd_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "female_doctor_tone4", R.drawable.emj_1f469_1f3fe_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "female_doctor_tone5", R.drawable.emj_1f469_1f3ff_200d_2695_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "female_elf", R.drawable.emj_1f9dd_200d_2640_fe0f, "\ud83e\udddd\u200d\u2640\ufe0f" );
		name( "female_elf_dark_skin_tone", R.drawable.emj_1f9dd_1f3ff_200d_2640_fe0f, "\ud83e\udddd\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_elf_light_skin_tone", R.drawable.emj_1f9dd_1f3fb_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_elf_medium_dark_skin_tone", R.drawable.emj_1f9dd_1f3fe_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_elf_medium_light_skin_tone", R.drawable.emj_1f9dd_1f3fc_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_elf_medium_skin_tone", R.drawable.emj_1f9dd_1f3fd_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_elf_tone1", R.drawable.emj_1f9dd_1f3fb_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_elf_tone2", R.drawable.emj_1f9dd_1f3fc_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_elf_tone3", R.drawable.emj_1f9dd_1f3fd_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_elf_tone4", R.drawable.emj_1f9dd_1f3fe_200d_2640_fe0f, "\ud83e\udddd\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_elf_tone5", R.drawable.emj_1f9dd_1f3ff_200d_2640_fe0f, "\ud83e\udddd\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_factory_worker", R.drawable.emj_1f469_200d_1f3ed, "\ud83d\udc69\u200d\ud83c\udfed" );
		name( "female_factory_worker_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3ed, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfed" );
		name( "female_factory_worker_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3ed, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfed" );
		name( "female_factory_worker_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3ed, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfed" );
		name( "female_factory_worker_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3ed, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfed" );
		name( "female_factory_worker_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3ed, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfed" );
		name( "female_factory_worker_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3ed, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfed" );
		name( "female_factory_worker_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3ed, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfed" );
		name( "female_factory_worker_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3ed, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfed" );
		name( "female_factory_worker_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3ed, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfed" );
		name( "female_factory_worker_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3ed, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfed" );
		name( "female_fairy", R.drawable.emj_1f9da_200d_2640_fe0f, "\ud83e\uddda\u200d\u2640\ufe0f" );
		name( "female_fairy_dark_skin_tone", R.drawable.emj_1f9da_1f3ff_200d_2640_fe0f, "\ud83e\uddda\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_fairy_light_skin_tone", R.drawable.emj_1f9da_1f3fb_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_fairy_medium_dark_skin_tone", R.drawable.emj_1f9da_1f3fe_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_fairy_medium_light_skin_tone", R.drawable.emj_1f9da_1f3fc_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_fairy_medium_skin_tone", R.drawable.emj_1f9da_1f3fd_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_fairy_tone1", R.drawable.emj_1f9da_1f3fb_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_fairy_tone2", R.drawable.emj_1f9da_1f3fc_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_fairy_tone3", R.drawable.emj_1f9da_1f3fd_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_fairy_tone4", R.drawable.emj_1f9da_1f3fe_200d_2640_fe0f, "\ud83e\uddda\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_fairy_tone5", R.drawable.emj_1f9da_1f3ff_200d_2640_fe0f, "\ud83e\uddda\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_farmer", R.drawable.emj_1f469_200d_1f33e, "\ud83d\udc69\u200d\ud83c\udf3e" );
		name( "female_farmer_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f33e, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "female_farmer_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f33e, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "female_farmer_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f33e, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf3e" );
	}
	private static void init66(){
		name( "female_farmer_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f33e, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "female_farmer_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f33e, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "female_farmer_tone1", R.drawable.emj_1f469_1f3fb_200d_1f33e, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "female_farmer_tone2", R.drawable.emj_1f469_1f3fc_200d_1f33e, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "female_farmer_tone3", R.drawable.emj_1f469_1f3fd_200d_1f33e, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "female_farmer_tone4", R.drawable.emj_1f469_1f3fe_200d_1f33e, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf3e" );
		name( "female_farmer_tone5", R.drawable.emj_1f469_1f3ff_200d_1f33e, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "female_firefighter", R.drawable.emj_1f469_200d_1f692, "\ud83d\udc69\u200d\ud83d\ude92" );
		name( "female_firefighter_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f692, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude92" );
		name( "female_firefighter_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f692, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude92" );
		name( "female_firefighter_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f692, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude92" );
		name( "female_firefighter_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f692, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude92" );
		name( "female_firefighter_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f692, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude92" );
		name( "female_firefighter_tone1", R.drawable.emj_1f469_1f3fb_200d_1f692, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude92" );
		name( "female_firefighter_tone2", R.drawable.emj_1f469_1f3fc_200d_1f692, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude92" );
		name( "female_firefighter_tone3", R.drawable.emj_1f469_1f3fd_200d_1f692, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude92" );
		name( "female_firefighter_tone4", R.drawable.emj_1f469_1f3fe_200d_1f692, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude92" );
		name( "female_firefighter_tone5", R.drawable.emj_1f469_1f3ff_200d_1f692, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude92" );
		name( "female_genie", R.drawable.emj_1f9de_200d_2640_fe0f, "\ud83e\uddde\u200d\u2640\ufe0f" );
		name( "female_guard", R.drawable.emj_1f482_200d_2640_fe0f, "\ud83d\udc82\u200d\u2640\ufe0f" );
		name( "female_guard_dark_skin_tone", R.drawable.emj_1f482_1f3ff_200d_2640_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_guard_light_skin_tone", R.drawable.emj_1f482_1f3fb_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_guard_medium_dark_skin_tone", R.drawable.emj_1f482_1f3fe_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_guard_medium_light_skin_tone", R.drawable.emj_1f482_1f3fc_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_guard_medium_skin_tone", R.drawable.emj_1f482_1f3fd_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_guard_tone1", R.drawable.emj_1f482_1f3fb_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_guard_tone2", R.drawable.emj_1f482_1f3fc_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_guard_tone3", R.drawable.emj_1f482_1f3fd_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_guard_tone4", R.drawable.emj_1f482_1f3fe_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_guard_tone5", R.drawable.emj_1f482_1f3ff_200d_2640_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_judge", R.drawable.emj_1f469_200d_2696_fe0f, "\ud83d\udc69\u200d\u2696\ufe0f" );
		name( "female_judge_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_2696_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "female_judge_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "female_judge_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "female_judge_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "female_judge_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "female_judge_tone1", R.drawable.emj_1f469_1f3fb_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "female_judge_tone2", R.drawable.emj_1f469_1f3fc_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "female_judge_tone3", R.drawable.emj_1f469_1f3fd_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "female_judge_tone4", R.drawable.emj_1f469_1f3fe_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "female_judge_tone5", R.drawable.emj_1f469_1f3ff_200d_2696_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "female_mage", R.drawable.emj_1f9d9_200d_2640_fe0f, "\ud83e\uddd9\u200d\u2640\ufe0f" );
		name( "female_mage_dark_skin_tone", R.drawable.emj_1f9d9_1f3ff_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_mage_light_skin_tone", R.drawable.emj_1f9d9_1f3fb_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_mage_medium_dark_skin_tone", R.drawable.emj_1f9d9_1f3fe_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_mage_medium_light_skin_tone", R.drawable.emj_1f9d9_1f3fc_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_mage_medium_skin_tone", R.drawable.emj_1f9d9_1f3fd_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_mage_tone1", R.drawable.emj_1f9d9_1f3fb_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_mage_tone2", R.drawable.emj_1f9d9_1f3fc_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_mage_tone3", R.drawable.emj_1f9d9_1f3fd_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_mage_tone4", R.drawable.emj_1f9d9_1f3fe_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_mage_tone5", R.drawable.emj_1f9d9_1f3ff_200d_2640_fe0f, "\ud83e\uddd9\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_mechanic", R.drawable.emj_1f469_200d_1f527, "\ud83d\udc69\u200d\ud83d\udd27" );
		name( "female_mechanic_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f527, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd27" );
		name( "female_mechanic_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f527, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd27" );
		name( "female_mechanic_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f527, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd27" );
		name( "female_mechanic_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f527, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd27" );
		name( "female_mechanic_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f527, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd27" );
		name( "female_mechanic_tone1", R.drawable.emj_1f469_1f3fb_200d_1f527, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd27" );
		name( "female_mechanic_tone2", R.drawable.emj_1f469_1f3fc_200d_1f527, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd27" );
		name( "female_mechanic_tone3", R.drawable.emj_1f469_1f3fd_200d_1f527, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd27" );
		name( "female_mechanic_tone4", R.drawable.emj_1f469_1f3fe_200d_1f527, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd27" );
		name( "female_mechanic_tone5", R.drawable.emj_1f469_1f3ff_200d_1f527, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd27" );
		name( "female_office_worker", R.drawable.emj_1f469_200d_1f4bc, "\ud83d\udc69\u200d\ud83d\udcbc" );
		name( "female_office_worker_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f4bc, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "female_office_worker_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f4bc, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "female_office_worker_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f4bc, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "female_office_worker_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f4bc, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "female_office_worker_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f4bc, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "female_office_worker_tone1", R.drawable.emj_1f469_1f3fb_200d_1f4bc, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "female_office_worker_tone2", R.drawable.emj_1f469_1f3fc_200d_1f4bc, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "female_office_worker_tone3", R.drawable.emj_1f469_1f3fd_200d_1f4bc, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "female_office_worker_tone4", R.drawable.emj_1f469_1f3fe_200d_1f4bc, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "female_office_worker_tone5", R.drawable.emj_1f469_1f3ff_200d_1f4bc, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "female_pilot", R.drawable.emj_1f469_200d_2708_fe0f, "\ud83d\udc69\u200d\u2708\ufe0f" );
		name( "female_pilot_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_2708_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "female_pilot_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "female_pilot_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "female_pilot_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "female_pilot_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2708\ufe0f" );
		name( "female_pilot_tone1", R.drawable.emj_1f469_1f3fb_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "female_pilot_tone2", R.drawable.emj_1f469_1f3fc_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "female_pilot_tone3", R.drawable.emj_1f469_1f3fd_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2708\ufe0f" );
		name( "female_pilot_tone4", R.drawable.emj_1f469_1f3fe_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "female_pilot_tone5", R.drawable.emj_1f469_1f3ff_200d_2708_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "female_police_officer", R.drawable.emj_1f46e_200d_2640_fe0f, "\ud83d\udc6e\u200d\u2640\ufe0f" );
		name( "female_police_officer_dark_skin_tone", R.drawable.emj_1f46e_1f3ff_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_police_officer_light_skin_tone", R.drawable.emj_1f46e_1f3fb_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_police_officer_medium_dark_skin_tone", R.drawable.emj_1f46e_1f3fe_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_police_officer_medium_light_skin_tone", R.drawable.emj_1f46e_1f3fc_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_police_officer_medium_skin_tone", R.drawable.emj_1f46e_1f3fd_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_police_officer_tone1", R.drawable.emj_1f46e_1f3fb_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_police_officer_tone2", R.drawable.emj_1f46e_1f3fc_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_police_officer_tone3", R.drawable.emj_1f46e_1f3fd_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_police_officer_tone4", R.drawable.emj_1f46e_1f3fe_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_police_officer_tone5", R.drawable.emj_1f46e_1f3ff_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_scientist", R.drawable.emj_1f469_200d_1f52c, "\ud83d\udc69\u200d\ud83d\udd2c" );
		name( "female_scientist_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f52c, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "female_scientist_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f52c, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "female_scientist_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f52c, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "female_scientist_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f52c, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd2c" );
	}
	private static void init67(){
		name( "female_scientist_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f52c, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "female_scientist_tone1", R.drawable.emj_1f469_1f3fb_200d_1f52c, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "female_scientist_tone2", R.drawable.emj_1f469_1f3fc_200d_1f52c, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd2c" );
		name( "female_scientist_tone3", R.drawable.emj_1f469_1f3fd_200d_1f52c, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "female_scientist_tone4", R.drawable.emj_1f469_1f3fe_200d_1f52c, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "female_scientist_tone5", R.drawable.emj_1f469_1f3ff_200d_1f52c, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "female_sign", R.drawable.emj_2640_fe0f, "\u2640\ufe0f" );
		name( "female_singer", R.drawable.emj_1f469_200d_1f3a4, "\ud83d\udc69\u200d\ud83c\udfa4" );
		name( "female_singer_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3a4, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "female_singer_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3a4, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "female_singer_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3a4, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "female_singer_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3a4, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "female_singer_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3a4, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "female_singer_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3a4, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "female_singer_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3a4, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "female_singer_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3a4, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "female_singer_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3a4, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "female_singer_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3a4, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "female_student", R.drawable.emj_1f469_200d_1f393, "\ud83d\udc69\u200d\ud83c\udf93" );
		name( "female_student_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f393, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf93" );
		name( "female_student_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f393, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf93" );
		name( "female_student_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f393, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf93" );
		name( "female_student_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f393, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf93" );
		name( "female_student_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f393, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf93" );
		name( "female_student_tone1", R.drawable.emj_1f469_1f3fb_200d_1f393, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf93" );
		name( "female_student_tone2", R.drawable.emj_1f469_1f3fc_200d_1f393, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf93" );
		name( "female_student_tone3", R.drawable.emj_1f469_1f3fd_200d_1f393, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf93" );
		name( "female_student_tone4", R.drawable.emj_1f469_1f3fe_200d_1f393, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf93" );
		name( "female_student_tone5", R.drawable.emj_1f469_1f3ff_200d_1f393, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf93" );
		name( "female_teacher", R.drawable.emj_1f469_200d_1f3eb, "\ud83d\udc69\u200d\ud83c\udfeb" );
		name( "female_teacher_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3eb, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "female_teacher_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3eb, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "female_teacher_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3eb, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "female_teacher_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3eb, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "female_teacher_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3eb, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "female_teacher_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3eb, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "female_teacher_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3eb, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "female_teacher_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3eb, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "female_teacher_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3eb, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "female_teacher_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3eb, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "female_technologist", R.drawable.emj_1f469_200d_1f4bb, "\ud83d\udc69\u200d\ud83d\udcbb" );
		name( "female_technologist_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f4bb, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "female_technologist_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f4bb, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "female_technologist_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f4bb, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "female_technologist_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f4bb, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "female_technologist_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f4bb, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "female_technologist_tone1", R.drawable.emj_1f469_1f3fb_200d_1f4bb, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "female_technologist_tone2", R.drawable.emj_1f469_1f3fc_200d_1f4bb, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "female_technologist_tone3", R.drawable.emj_1f469_1f3fd_200d_1f4bb, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "female_technologist_tone4", R.drawable.emj_1f469_1f3fe_200d_1f4bb, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "female_technologist_tone5", R.drawable.emj_1f469_1f3ff_200d_1f4bb, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "female_vampire", R.drawable.emj_1f9db_200d_2640_fe0f, "\ud83e\udddb\u200d\u2640\ufe0f" );
		name( "female_vampire_dark_skin_tone", R.drawable.emj_1f9db_1f3ff_200d_2640_fe0f, "\ud83e\udddb\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_vampire_light_skin_tone", R.drawable.emj_1f9db_1f3fb_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_vampire_medium_dark_skin_tone", R.drawable.emj_1f9db_1f3fe_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_vampire_medium_light_skin_tone", R.drawable.emj_1f9db_1f3fc_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_vampire_medium_skin_tone", R.drawable.emj_1f9db_1f3fd_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_vampire_tone1", R.drawable.emj_1f9db_1f3fb_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "female_vampire_tone2", R.drawable.emj_1f9db_1f3fc_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "female_vampire_tone3", R.drawable.emj_1f9db_1f3fd_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "female_vampire_tone4", R.drawable.emj_1f9db_1f3fe_200d_2640_fe0f, "\ud83e\udddb\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "female_vampire_tone5", R.drawable.emj_1f9db_1f3ff_200d_2640_fe0f, "\ud83e\udddb\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "female_zombie", R.drawable.emj_1f9df_200d_2640_fe0f, "\ud83e\udddf\u200d\u2640\ufe0f" );
		name( "fencer", R.drawable.emj_1f93a, "\ud83e\udd3a" );
		name( "fencing", R.drawable.emj_1f93a, "\ud83e\udd3a" );
		name( "ferris_wheel", R.drawable.emj_1f3a1, "\ud83c\udfa1" );
		name( "ferry", R.drawable.emj_26f4_fe0f, "\u26f4\ufe0f" );
		name( "fi", R.drawable.emj_1f1eb_1f1ee, "\ud83c\uddeb\ud83c\uddee" );
		name( "field_hockey", R.drawable.emj_1f3d1, "\ud83c\udfd1" );
		name( "field_hockey_stick_and_ball", R.drawable.emj_1f3d1, "\ud83c\udfd1" );
		name( "file_cabinet", R.drawable.emj_1f5c4_fe0f, "\ud83d\uddc4\ufe0f" );
		name( "file_folder", R.drawable.emj_1f4c1, "\ud83d\udcc1" );
		name( "film_frames", R.drawable.emj_1f39e_fe0f, "\ud83c\udf9e\ufe0f" );
		name( "film_projector", R.drawable.emj_1f4fd_fe0f, "\ud83d\udcfd\ufe0f" );
		name( "fingers_crossed", R.drawable.emj_1f91e, "\ud83e\udd1e" );
		name( "fingers_crossed_tone1", R.drawable.emj_1f91e_1f3fb, "\ud83e\udd1e\ud83c\udffb" );
		name( "fingers_crossed_tone2", R.drawable.emj_1f91e_1f3fc, "\ud83e\udd1e\ud83c\udffc" );
		name( "fingers_crossed_tone3", R.drawable.emj_1f91e_1f3fd, "\ud83e\udd1e\ud83c\udffd" );
		name( "fingers_crossed_tone4", R.drawable.emj_1f91e_1f3fe, "\ud83e\udd1e\ud83c\udffe" );
		name( "fingers_crossed_tone5", R.drawable.emj_1f91e_1f3ff, "\ud83e\udd1e\ud83c\udfff" );
		name( "fire", R.drawable.emj_1f525, "\ud83d\udd25" );
		name( "fire_engine", R.drawable.emj_1f692, "\ud83d\ude92" );
		name( "fireworks", R.drawable.emj_1f386, "\ud83c\udf86" );
		name( "first_place", R.drawable.emj_1f947, "\ud83e\udd47" );
		name( "first_place_medal", R.drawable.emj_1f947, "\ud83e\udd47" );
		name( "first_quarter_moon", R.drawable.emj_1f313, "\ud83c\udf13" );
		name( "first_quarter_moon_with_face", R.drawable.emj_1f31b, "\ud83c\udf1b" );
		name( "fish", R.drawable.emj_1f41f, "\ud83d\udc1f" );
		name( "fish_cake", R.drawable.emj_1f365, "\ud83c\udf65" );
		name( "fishing_pole_and_fish", R.drawable.emj_1f3a3, "\ud83c\udfa3" );
		name( "fist", R.drawable.emj_270a, "\u270a" );
		name( "fist_dark_skin_tone", R.drawable.emj_270a_1f3ff, "\u270a\ud83c\udfff" );
		name( "fist_light_skin_tone", R.drawable.emj_270a_1f3fb, "\u270a\ud83c\udffb" );
		name( "fist_medium_dark_skin_tone", R.drawable.emj_270a_1f3fe, "\u270a\ud83c\udffe" );
		name( "fist_medium_light_skin_tone", R.drawable.emj_270a_1f3fc, "\u270a\ud83c\udffc" );
		name( "fist_medium_skin_tone", R.drawable.emj_270a_1f3fd, "\u270a\ud83c\udffd" );
		name( "fist_tone1", R.drawable.emj_270a_1f3fb, "\u270a\ud83c\udffb" );
		name( "fist_tone2", R.drawable.emj_270a_1f3fc, "\u270a\ud83c\udffc" );
		name( "fist_tone3", R.drawable.emj_270a_1f3fd, "\u270a\ud83c\udffd" );
		name( "fist_tone4", R.drawable.emj_270a_1f3fe, "\u270a\ud83c\udffe" );
		name( "fist_tone5", R.drawable.emj_270a_1f3ff, "\u270a\ud83c\udfff" );
	}
	private static void init68(){
		name( "five", R.drawable.emj_0035_fe0f_20e3, "\u0035\ufe0f\u20e3" );
		name( "fj", R.drawable.emj_1f1eb_1f1ef, "\ud83c\uddeb\ud83c\uddef" );
		name( "fk", R.drawable.emj_1f1eb_1f1f0, "\ud83c\uddeb\ud83c\uddf0" );
		name( "flag_ac", R.drawable.emj_1f1e6_1f1e8, "\ud83c\udde6\ud83c\udde8" );
		name( "flag_ad", R.drawable.emj_1f1e6_1f1e9, "\ud83c\udde6\ud83c\udde9" );
		name( "flag_ae", R.drawable.emj_1f1e6_1f1ea, "\ud83c\udde6\ud83c\uddea" );
		name( "flag_af", R.drawable.emj_1f1e6_1f1eb, "\ud83c\udde6\ud83c\uddeb" );
		name( "flag_ag", R.drawable.emj_1f1e6_1f1ec, "\ud83c\udde6\ud83c\uddec" );
		name( "flag_ai", R.drawable.emj_1f1e6_1f1ee, "\ud83c\udde6\ud83c\uddee" );
		name( "flag_al", R.drawable.emj_1f1e6_1f1f1, "\ud83c\udde6\ud83c\uddf1" );
		name( "flag_am", R.drawable.emj_1f1e6_1f1f2, "\ud83c\udde6\ud83c\uddf2" );
		name( "flag_ao", R.drawable.emj_1f1e6_1f1f4, "\ud83c\udde6\ud83c\uddf4" );
		name( "flag_aq", R.drawable.emj_1f1e6_1f1f6, "\ud83c\udde6\ud83c\uddf6" );
		name( "flag_ar", R.drawable.emj_1f1e6_1f1f7, "\ud83c\udde6\ud83c\uddf7" );
		name( "flag_as", R.drawable.emj_1f1e6_1f1f8, "\ud83c\udde6\ud83c\uddf8" );
		name( "flag_at", R.drawable.emj_1f1e6_1f1f9, "\ud83c\udde6\ud83c\uddf9" );
		name( "flag_au", R.drawable.emj_1f1e6_1f1fa, "\ud83c\udde6\ud83c\uddfa" );
		name( "flag_aw", R.drawable.emj_1f1e6_1f1fc, "\ud83c\udde6\ud83c\uddfc" );
		name( "flag_ax", R.drawable.emj_1f1e6_1f1fd, "\ud83c\udde6\ud83c\uddfd" );
		name( "flag_az", R.drawable.emj_1f1e6_1f1ff, "\ud83c\udde6\ud83c\uddff" );
		name( "flag_ba", R.drawable.emj_1f1e7_1f1e6, "\ud83c\udde7\ud83c\udde6" );
		name( "flag_bb", R.drawable.emj_1f1e7_1f1e7, "\ud83c\udde7\ud83c\udde7" );
		name( "flag_bd", R.drawable.emj_1f1e7_1f1e9, "\ud83c\udde7\ud83c\udde9" );
		name( "flag_be", R.drawable.emj_1f1e7_1f1ea, "\ud83c\udde7\ud83c\uddea" );
		name( "flag_bf", R.drawable.emj_1f1e7_1f1eb, "\ud83c\udde7\ud83c\uddeb" );
		name( "flag_bg", R.drawable.emj_1f1e7_1f1ec, "\ud83c\udde7\ud83c\uddec" );
		name( "flag_bh", R.drawable.emj_1f1e7_1f1ed, "\ud83c\udde7\ud83c\udded" );
		name( "flag_bi", R.drawable.emj_1f1e7_1f1ee, "\ud83c\udde7\ud83c\uddee" );
		name( "flag_bj", R.drawable.emj_1f1e7_1f1ef, "\ud83c\udde7\ud83c\uddef" );
		name( "flag_bl", R.drawable.emj_1f1e7_1f1f1, "\ud83c\udde7\ud83c\uddf1" );
		name( "flag_black", R.drawable.emj_1f3f4, "\ud83c\udff4" );
		name( "flag_bm", R.drawable.emj_1f1e7_1f1f2, "\ud83c\udde7\ud83c\uddf2" );
		name( "flag_bn", R.drawable.emj_1f1e7_1f1f3, "\ud83c\udde7\ud83c\uddf3" );
		name( "flag_bo", R.drawable.emj_1f1e7_1f1f4, "\ud83c\udde7\ud83c\uddf4" );
		name( "flag_bq", R.drawable.emj_1f1e7_1f1f6, "\ud83c\udde7\ud83c\uddf6" );
		name( "flag_br", R.drawable.emj_1f1e7_1f1f7, "\ud83c\udde7\ud83c\uddf7" );
		name( "flag_bs", R.drawable.emj_1f1e7_1f1f8, "\ud83c\udde7\ud83c\uddf8" );
		name( "flag_bt", R.drawable.emj_1f1e7_1f1f9, "\ud83c\udde7\ud83c\uddf9" );
		name( "flag_bv", R.drawable.emj_1f1e7_1f1fb, "\ud83c\udde7\ud83c\uddfb" );
		name( "flag_bw", R.drawable.emj_1f1e7_1f1fc, "\ud83c\udde7\ud83c\uddfc" );
		name( "flag_by", R.drawable.emj_1f1e7_1f1fe, "\ud83c\udde7\ud83c\uddfe" );
		name( "flag_bz", R.drawable.emj_1f1e7_1f1ff, "\ud83c\udde7\ud83c\uddff" );
		name( "flag_ca", R.drawable.emj_1f1e8_1f1e6, "\ud83c\udde8\ud83c\udde6" );
		name( "flag_cc", R.drawable.emj_1f1e8_1f1e8, "\ud83c\udde8\ud83c\udde8" );
		name( "flag_cd", R.drawable.emj_1f1e8_1f1e9, "\ud83c\udde8\ud83c\udde9" );
		name( "flag_cf", R.drawable.emj_1f1e8_1f1eb, "\ud83c\udde8\ud83c\uddeb" );
		name( "flag_cg", R.drawable.emj_1f1e8_1f1ec, "\ud83c\udde8\ud83c\uddec" );
		name( "flag_ch", R.drawable.emj_1f1e8_1f1ed, "\ud83c\udde8\ud83c\udded" );
		name( "flag_ci", R.drawable.emj_1f1e8_1f1ee, "\ud83c\udde8\ud83c\uddee" );
		name( "flag_ck", R.drawable.emj_1f1e8_1f1f0, "\ud83c\udde8\ud83c\uddf0" );
		name( "flag_cl", R.drawable.emj_1f1e8_1f1f1, "\ud83c\udde8\ud83c\uddf1" );
		name( "flag_cm", R.drawable.emj_1f1e8_1f1f2, "\ud83c\udde8\ud83c\uddf2" );
		name( "flag_cn", R.drawable.emj_1f1e8_1f1f3, "\ud83c\udde8\ud83c\uddf3" );
		name( "flag_co", R.drawable.emj_1f1e8_1f1f4, "\ud83c\udde8\ud83c\uddf4" );
		name( "flag_cp", R.drawable.emj_1f1e8_1f1f5, "\ud83c\udde8\ud83c\uddf5" );
		name( "flag_cr", R.drawable.emj_1f1e8_1f1f7, "\ud83c\udde8\ud83c\uddf7" );
		name( "flag_cu", R.drawable.emj_1f1e8_1f1fa, "\ud83c\udde8\ud83c\uddfa" );
		name( "flag_cv", R.drawable.emj_1f1e8_1f1fb, "\ud83c\udde8\ud83c\uddfb" );
		name( "flag_cw", R.drawable.emj_1f1e8_1f1fc, "\ud83c\udde8\ud83c\uddfc" );
		name( "flag_cx", R.drawable.emj_1f1e8_1f1fd, "\ud83c\udde8\ud83c\uddfd" );
		name( "flag_cy", R.drawable.emj_1f1e8_1f1fe, "\ud83c\udde8\ud83c\uddfe" );
		name( "flag_cz", R.drawable.emj_1f1e8_1f1ff, "\ud83c\udde8\ud83c\uddff" );
		name( "flag_de", R.drawable.emj_1f1e9_1f1ea, "\ud83c\udde9\ud83c\uddea" );
		name( "flag_dg", R.drawable.emj_1f1e9_1f1ec, "\ud83c\udde9\ud83c\uddec" );
		name( "flag_dj", R.drawable.emj_1f1e9_1f1ef, "\ud83c\udde9\ud83c\uddef" );
		name( "flag_dk", R.drawable.emj_1f1e9_1f1f0, "\ud83c\udde9\ud83c\uddf0" );
		name( "flag_dm", R.drawable.emj_1f1e9_1f1f2, "\ud83c\udde9\ud83c\uddf2" );
		name( "flag_do", R.drawable.emj_1f1e9_1f1f4, "\ud83c\udde9\ud83c\uddf4" );
		name( "flag_dz", R.drawable.emj_1f1e9_1f1ff, "\ud83c\udde9\ud83c\uddff" );
		name( "flag_ea", R.drawable.emj_1f1ea_1f1e6, "\ud83c\uddea\ud83c\udde6" );
		name( "flag_ec", R.drawable.emj_1f1ea_1f1e8, "\ud83c\uddea\ud83c\udde8" );
		name( "flag_ee", R.drawable.emj_1f1ea_1f1ea, "\ud83c\uddea\ud83c\uddea" );
		name( "flag_eg", R.drawable.emj_1f1ea_1f1ec, "\ud83c\uddea\ud83c\uddec" );
		name( "flag_eh", R.drawable.emj_1f1ea_1f1ed, "\ud83c\uddea\ud83c\udded" );
		name( "flag_england", R.drawable.emj_1f3f4_e0067_e0062_e0065_e006e_e0067_e007f, "\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc65\udb40\udc6e\udb40\udc67\udb40\udc7f" );
		name( "flag_er", R.drawable.emj_1f1ea_1f1f7, "\ud83c\uddea\ud83c\uddf7" );
		name( "flag_es", R.drawable.emj_1f1ea_1f1f8, "\ud83c\uddea\ud83c\uddf8" );
		name( "flag_et", R.drawable.emj_1f1ea_1f1f9, "\ud83c\uddea\ud83c\uddf9" );
		name( "flag_eu", R.drawable.emj_1f1ea_1f1fa, "\ud83c\uddea\ud83c\uddfa" );
		name( "flag_fi", R.drawable.emj_1f1eb_1f1ee, "\ud83c\uddeb\ud83c\uddee" );
		name( "flag_fj", R.drawable.emj_1f1eb_1f1ef, "\ud83c\uddeb\ud83c\uddef" );
		name( "flag_fk", R.drawable.emj_1f1eb_1f1f0, "\ud83c\uddeb\ud83c\uddf0" );
		name( "flag_fm", R.drawable.emj_1f1eb_1f1f2, "\ud83c\uddeb\ud83c\uddf2" );
		name( "flag_fo", R.drawable.emj_1f1eb_1f1f4, "\ud83c\uddeb\ud83c\uddf4" );
		name( "flag_fr", R.drawable.emj_1f1eb_1f1f7, "\ud83c\uddeb\ud83c\uddf7" );
		name( "flag_ga", R.drawable.emj_1f1ec_1f1e6, "\ud83c\uddec\ud83c\udde6" );
		name( "flag_gb", R.drawable.emj_1f1ec_1f1e7, "\ud83c\uddec\ud83c\udde7" );
		name( "flag_gd", R.drawable.emj_1f1ec_1f1e9, "\ud83c\uddec\ud83c\udde9" );
		name( "flag_ge", R.drawable.emj_1f1ec_1f1ea, "\ud83c\uddec\ud83c\uddea" );
		name( "flag_gf", R.drawable.emj_1f1ec_1f1eb, "\ud83c\uddec\ud83c\uddeb" );
		name( "flag_gg", R.drawable.emj_1f1ec_1f1ec, "\ud83c\uddec\ud83c\uddec" );
		name( "flag_gh", R.drawable.emj_1f1ec_1f1ed, "\ud83c\uddec\ud83c\udded" );
		name( "flag_gi", R.drawable.emj_1f1ec_1f1ee, "\ud83c\uddec\ud83c\uddee" );
		name( "flag_gl", R.drawable.emj_1f1ec_1f1f1, "\ud83c\uddec\ud83c\uddf1" );
		name( "flag_gm", R.drawable.emj_1f1ec_1f1f2, "\ud83c\uddec\ud83c\uddf2" );
		name( "flag_gn", R.drawable.emj_1f1ec_1f1f3, "\ud83c\uddec\ud83c\uddf3" );
		name( "flag_gp", R.drawable.emj_1f1ec_1f1f5, "\ud83c\uddec\ud83c\uddf5" );
		name( "flag_gq", R.drawable.emj_1f1ec_1f1f6, "\ud83c\uddec\ud83c\uddf6" );
		name( "flag_gr", R.drawable.emj_1f1ec_1f1f7, "\ud83c\uddec\ud83c\uddf7" );
		name( "flag_gs", R.drawable.emj_1f1ec_1f1f8, "\ud83c\uddec\ud83c\uddf8" );
		name( "flag_gt", R.drawable.emj_1f1ec_1f1f9, "\ud83c\uddec\ud83c\uddf9" );
	}
	private static void init69(){
		name( "flag_gu", R.drawable.emj_1f1ec_1f1fa, "\ud83c\uddec\ud83c\uddfa" );
		name( "flag_gw", R.drawable.emj_1f1ec_1f1fc, "\ud83c\uddec\ud83c\uddfc" );
		name( "flag_gy", R.drawable.emj_1f1ec_1f1fe, "\ud83c\uddec\ud83c\uddfe" );
		name( "flag_hk", R.drawable.emj_1f1ed_1f1f0, "\ud83c\udded\ud83c\uddf0" );
		name( "flag_hm", R.drawable.emj_1f1ed_1f1f2, "\ud83c\udded\ud83c\uddf2" );
		name( "flag_hn", R.drawable.emj_1f1ed_1f1f3, "\ud83c\udded\ud83c\uddf3" );
		name( "flag_hr", R.drawable.emj_1f1ed_1f1f7, "\ud83c\udded\ud83c\uddf7" );
		name( "flag_ht", R.drawable.emj_1f1ed_1f1f9, "\ud83c\udded\ud83c\uddf9" );
		name( "flag_hu", R.drawable.emj_1f1ed_1f1fa, "\ud83c\udded\ud83c\uddfa" );
		name( "flag_ic", R.drawable.emj_1f1ee_1f1e8, "\ud83c\uddee\ud83c\udde8" );
		name( "flag_id", R.drawable.emj_1f1ee_1f1e9, "\ud83c\uddee\ud83c\udde9" );
		name( "flag_ie", R.drawable.emj_1f1ee_1f1ea, "\ud83c\uddee\ud83c\uddea" );
		name( "flag_il", R.drawable.emj_1f1ee_1f1f1, "\ud83c\uddee\ud83c\uddf1" );
		name( "flag_im", R.drawable.emj_1f1ee_1f1f2, "\ud83c\uddee\ud83c\uddf2" );
		name( "flag_in", R.drawable.emj_1f1ee_1f1f3, "\ud83c\uddee\ud83c\uddf3" );
		name( "flag_io", R.drawable.emj_1f1ee_1f1f4, "\ud83c\uddee\ud83c\uddf4" );
		name( "flag_iq", R.drawable.emj_1f1ee_1f1f6, "\ud83c\uddee\ud83c\uddf6" );
		name( "flag_ir", R.drawable.emj_1f1ee_1f1f7, "\ud83c\uddee\ud83c\uddf7" );
		name( "flag_is", R.drawable.emj_1f1ee_1f1f8, "\ud83c\uddee\ud83c\uddf8" );
		name( "flag_it", R.drawable.emj_1f1ee_1f1f9, "\ud83c\uddee\ud83c\uddf9" );
		name( "flag_je", R.drawable.emj_1f1ef_1f1ea, "\ud83c\uddef\ud83c\uddea" );
		name( "flag_jm", R.drawable.emj_1f1ef_1f1f2, "\ud83c\uddef\ud83c\uddf2" );
		name( "flag_jo", R.drawable.emj_1f1ef_1f1f4, "\ud83c\uddef\ud83c\uddf4" );
		name( "flag_jp", R.drawable.emj_1f1ef_1f1f5, "\ud83c\uddef\ud83c\uddf5" );
		name( "flag_ke", R.drawable.emj_1f1f0_1f1ea, "\ud83c\uddf0\ud83c\uddea" );
		name( "flag_kg", R.drawable.emj_1f1f0_1f1ec, "\ud83c\uddf0\ud83c\uddec" );
		name( "flag_kh", R.drawable.emj_1f1f0_1f1ed, "\ud83c\uddf0\ud83c\udded" );
		name( "flag_ki", R.drawable.emj_1f1f0_1f1ee, "\ud83c\uddf0\ud83c\uddee" );
		name( "flag_km", R.drawable.emj_1f1f0_1f1f2, "\ud83c\uddf0\ud83c\uddf2" );
		name( "flag_kn", R.drawable.emj_1f1f0_1f1f3, "\ud83c\uddf0\ud83c\uddf3" );
		name( "flag_kp", R.drawable.emj_1f1f0_1f1f5, "\ud83c\uddf0\ud83c\uddf5" );
		name( "flag_kr", R.drawable.emj_1f1f0_1f1f7, "\ud83c\uddf0\ud83c\uddf7" );
		name( "flag_kw", R.drawable.emj_1f1f0_1f1fc, "\ud83c\uddf0\ud83c\uddfc" );
		name( "flag_ky", R.drawable.emj_1f1f0_1f1fe, "\ud83c\uddf0\ud83c\uddfe" );
		name( "flag_kz", R.drawable.emj_1f1f0_1f1ff, "\ud83c\uddf0\ud83c\uddff" );
		name( "flag_la", R.drawable.emj_1f1f1_1f1e6, "\ud83c\uddf1\ud83c\udde6" );
		name( "flag_lb", R.drawable.emj_1f1f1_1f1e7, "\ud83c\uddf1\ud83c\udde7" );
		name( "flag_lc", R.drawable.emj_1f1f1_1f1e8, "\ud83c\uddf1\ud83c\udde8" );
		name( "flag_li", R.drawable.emj_1f1f1_1f1ee, "\ud83c\uddf1\ud83c\uddee" );
		name( "flag_lk", R.drawable.emj_1f1f1_1f1f0, "\ud83c\uddf1\ud83c\uddf0" );
		name( "flag_lr", R.drawable.emj_1f1f1_1f1f7, "\ud83c\uddf1\ud83c\uddf7" );
		name( "flag_ls", R.drawable.emj_1f1f1_1f1f8, "\ud83c\uddf1\ud83c\uddf8" );
		name( "flag_lt", R.drawable.emj_1f1f1_1f1f9, "\ud83c\uddf1\ud83c\uddf9" );
		name( "flag_lu", R.drawable.emj_1f1f1_1f1fa, "\ud83c\uddf1\ud83c\uddfa" );
		name( "flag_lv", R.drawable.emj_1f1f1_1f1fb, "\ud83c\uddf1\ud83c\uddfb" );
		name( "flag_ly", R.drawable.emj_1f1f1_1f1fe, "\ud83c\uddf1\ud83c\uddfe" );
		name( "flag_ma", R.drawable.emj_1f1f2_1f1e6, "\ud83c\uddf2\ud83c\udde6" );
		name( "flag_mc", R.drawable.emj_1f1f2_1f1e8, "\ud83c\uddf2\ud83c\udde8" );
		name( "flag_md", R.drawable.emj_1f1f2_1f1e9, "\ud83c\uddf2\ud83c\udde9" );
		name( "flag_me", R.drawable.emj_1f1f2_1f1ea, "\ud83c\uddf2\ud83c\uddea" );
		name( "flag_mf", R.drawable.emj_1f1f2_1f1eb, "\ud83c\uddf2\ud83c\uddeb" );
		name( "flag_mg", R.drawable.emj_1f1f2_1f1ec, "\ud83c\uddf2\ud83c\uddec" );
		name( "flag_mh", R.drawable.emj_1f1f2_1f1ed, "\ud83c\uddf2\ud83c\udded" );
		name( "flag_mk", R.drawable.emj_1f1f2_1f1f0, "\ud83c\uddf2\ud83c\uddf0" );
		name( "flag_ml", R.drawable.emj_1f1f2_1f1f1, "\ud83c\uddf2\ud83c\uddf1" );
		name( "flag_mm", R.drawable.emj_1f1f2_1f1f2, "\ud83c\uddf2\ud83c\uddf2" );
		name( "flag_mn", R.drawable.emj_1f1f2_1f1f3, "\ud83c\uddf2\ud83c\uddf3" );
		name( "flag_mo", R.drawable.emj_1f1f2_1f1f4, "\ud83c\uddf2\ud83c\uddf4" );
		name( "flag_mp", R.drawable.emj_1f1f2_1f1f5, "\ud83c\uddf2\ud83c\uddf5" );
		name( "flag_mq", R.drawable.emj_1f1f2_1f1f6, "\ud83c\uddf2\ud83c\uddf6" );
		name( "flag_mr", R.drawable.emj_1f1f2_1f1f7, "\ud83c\uddf2\ud83c\uddf7" );
		name( "flag_ms", R.drawable.emj_1f1f2_1f1f8, "\ud83c\uddf2\ud83c\uddf8" );
		name( "flag_mt", R.drawable.emj_1f1f2_1f1f9, "\ud83c\uddf2\ud83c\uddf9" );
		name( "flag_mu", R.drawable.emj_1f1f2_1f1fa, "\ud83c\uddf2\ud83c\uddfa" );
		name( "flag_mv", R.drawable.emj_1f1f2_1f1fb, "\ud83c\uddf2\ud83c\uddfb" );
		name( "flag_mw", R.drawable.emj_1f1f2_1f1fc, "\ud83c\uddf2\ud83c\uddfc" );
		name( "flag_mx", R.drawable.emj_1f1f2_1f1fd, "\ud83c\uddf2\ud83c\uddfd" );
		name( "flag_my", R.drawable.emj_1f1f2_1f1fe, "\ud83c\uddf2\ud83c\uddfe" );
		name( "flag_mz", R.drawable.emj_1f1f2_1f1ff, "\ud83c\uddf2\ud83c\uddff" );
		name( "flag_na", R.drawable.emj_1f1f3_1f1e6, "\ud83c\uddf3\ud83c\udde6" );
		name( "flag_nc", R.drawable.emj_1f1f3_1f1e8, "\ud83c\uddf3\ud83c\udde8" );
		name( "flag_ne", R.drawable.emj_1f1f3_1f1ea, "\ud83c\uddf3\ud83c\uddea" );
		name( "flag_nf", R.drawable.emj_1f1f3_1f1eb, "\ud83c\uddf3\ud83c\uddeb" );
		name( "flag_ng", R.drawable.emj_1f1f3_1f1ec, "\ud83c\uddf3\ud83c\uddec" );
		name( "flag_ni", R.drawable.emj_1f1f3_1f1ee, "\ud83c\uddf3\ud83c\uddee" );
		name( "flag_nl", R.drawable.emj_1f1f3_1f1f1, "\ud83c\uddf3\ud83c\uddf1" );
		name( "flag_no", R.drawable.emj_1f1f3_1f1f4, "\ud83c\uddf3\ud83c\uddf4" );
		name( "flag_np", R.drawable.emj_1f1f3_1f1f5, "\ud83c\uddf3\ud83c\uddf5" );
		name( "flag_nr", R.drawable.emj_1f1f3_1f1f7, "\ud83c\uddf3\ud83c\uddf7" );
		name( "flag_nu", R.drawable.emj_1f1f3_1f1fa, "\ud83c\uddf3\ud83c\uddfa" );
		name( "flag_nz", R.drawable.emj_1f1f3_1f1ff, "\ud83c\uddf3\ud83c\uddff" );
		name( "flag_om", R.drawable.emj_1f1f4_1f1f2, "\ud83c\uddf4\ud83c\uddf2" );
		name( "flag_pa", R.drawable.emj_1f1f5_1f1e6, "\ud83c\uddf5\ud83c\udde6" );
		name( "flag_pe", R.drawable.emj_1f1f5_1f1ea, "\ud83c\uddf5\ud83c\uddea" );
		name( "flag_pf", R.drawable.emj_1f1f5_1f1eb, "\ud83c\uddf5\ud83c\uddeb" );
		name( "flag_pg", R.drawable.emj_1f1f5_1f1ec, "\ud83c\uddf5\ud83c\uddec" );
		name( "flag_ph", R.drawable.emj_1f1f5_1f1ed, "\ud83c\uddf5\ud83c\udded" );
		name( "flag_pk", R.drawable.emj_1f1f5_1f1f0, "\ud83c\uddf5\ud83c\uddf0" );
		name( "flag_pl", R.drawable.emj_1f1f5_1f1f1, "\ud83c\uddf5\ud83c\uddf1" );
		name( "flag_pm", R.drawable.emj_1f1f5_1f1f2, "\ud83c\uddf5\ud83c\uddf2" );
		name( "flag_pn", R.drawable.emj_1f1f5_1f1f3, "\ud83c\uddf5\ud83c\uddf3" );
		name( "flag_pr", R.drawable.emj_1f1f5_1f1f7, "\ud83c\uddf5\ud83c\uddf7" );
		name( "flag_ps", R.drawable.emj_1f1f5_1f1f8, "\ud83c\uddf5\ud83c\uddf8" );
		name( "flag_pt", R.drawable.emj_1f1f5_1f1f9, "\ud83c\uddf5\ud83c\uddf9" );
		name( "flag_pw", R.drawable.emj_1f1f5_1f1fc, "\ud83c\uddf5\ud83c\uddfc" );
		name( "flag_py", R.drawable.emj_1f1f5_1f1fe, "\ud83c\uddf5\ud83c\uddfe" );
		name( "flag_qa", R.drawable.emj_1f1f6_1f1e6, "\ud83c\uddf6\ud83c\udde6" );
		name( "flag_re", R.drawable.emj_1f1f7_1f1ea, "\ud83c\uddf7\ud83c\uddea" );
		name( "flag_ro", R.drawable.emj_1f1f7_1f1f4, "\ud83c\uddf7\ud83c\uddf4" );
		name( "flag_rs", R.drawable.emj_1f1f7_1f1f8, "\ud83c\uddf7\ud83c\uddf8" );
		name( "flag_ru", R.drawable.emj_1f1f7_1f1fa, "\ud83c\uddf7\ud83c\uddfa" );
	}
	private static void init70(){
		name( "flag_rw", R.drawable.emj_1f1f7_1f1fc, "\ud83c\uddf7\ud83c\uddfc" );
		name( "flag_sa", R.drawable.emj_1f1f8_1f1e6, "\ud83c\uddf8\ud83c\udde6" );
		name( "flag_sb", R.drawable.emj_1f1f8_1f1e7, "\ud83c\uddf8\ud83c\udde7" );
		name( "flag_sc", R.drawable.emj_1f1f8_1f1e8, "\ud83c\uddf8\ud83c\udde8" );
		name( "flag_scotland", R.drawable.emj_1f3f4_e0067_e0062_e0073_e0063_e0074_e007f, "\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc73\udb40\udc63\udb40\udc74\udb40\udc7f" );
		name( "flag_sd", R.drawable.emj_1f1f8_1f1e9, "\ud83c\uddf8\ud83c\udde9" );
		name( "flag_se", R.drawable.emj_1f1f8_1f1ea, "\ud83c\uddf8\ud83c\uddea" );
		name( "flag_sg", R.drawable.emj_1f1f8_1f1ec, "\ud83c\uddf8\ud83c\uddec" );
		name( "flag_sh", R.drawable.emj_1f1f8_1f1ed, "\ud83c\uddf8\ud83c\udded" );
		name( "flag_si", R.drawable.emj_1f1f8_1f1ee, "\ud83c\uddf8\ud83c\uddee" );
		name( "flag_sj", R.drawable.emj_1f1f8_1f1ef, "\ud83c\uddf8\ud83c\uddef" );
		name( "flag_sk", R.drawable.emj_1f1f8_1f1f0, "\ud83c\uddf8\ud83c\uddf0" );
		name( "flag_sl", R.drawable.emj_1f1f8_1f1f1, "\ud83c\uddf8\ud83c\uddf1" );
		name( "flag_sm", R.drawable.emj_1f1f8_1f1f2, "\ud83c\uddf8\ud83c\uddf2" );
		name( "flag_sn", R.drawable.emj_1f1f8_1f1f3, "\ud83c\uddf8\ud83c\uddf3" );
		name( "flag_so", R.drawable.emj_1f1f8_1f1f4, "\ud83c\uddf8\ud83c\uddf4" );
		name( "flag_sr", R.drawable.emj_1f1f8_1f1f7, "\ud83c\uddf8\ud83c\uddf7" );
		name( "flag_ss", R.drawable.emj_1f1f8_1f1f8, "\ud83c\uddf8\ud83c\uddf8" );
		name( "flag_st", R.drawable.emj_1f1f8_1f1f9, "\ud83c\uddf8\ud83c\uddf9" );
		name( "flag_sv", R.drawable.emj_1f1f8_1f1fb, "\ud83c\uddf8\ud83c\uddfb" );
		name( "flag_sx", R.drawable.emj_1f1f8_1f1fd, "\ud83c\uddf8\ud83c\uddfd" );
		name( "flag_sy", R.drawable.emj_1f1f8_1f1fe, "\ud83c\uddf8\ud83c\uddfe" );
		name( "flag_sz", R.drawable.emj_1f1f8_1f1ff, "\ud83c\uddf8\ud83c\uddff" );
		name( "flag_ta", R.drawable.emj_1f1f9_1f1e6, "\ud83c\uddf9\ud83c\udde6" );
		name( "flag_tc", R.drawable.emj_1f1f9_1f1e8, "\ud83c\uddf9\ud83c\udde8" );
		name( "flag_td", R.drawable.emj_1f1f9_1f1e9, "\ud83c\uddf9\ud83c\udde9" );
		name( "flag_tf", R.drawable.emj_1f1f9_1f1eb, "\ud83c\uddf9\ud83c\uddeb" );
		name( "flag_tg", R.drawable.emj_1f1f9_1f1ec, "\ud83c\uddf9\ud83c\uddec" );
		name( "flag_th", R.drawable.emj_1f1f9_1f1ed, "\ud83c\uddf9\ud83c\udded" );
		name( "flag_tj", R.drawable.emj_1f1f9_1f1ef, "\ud83c\uddf9\ud83c\uddef" );
		name( "flag_tk", R.drawable.emj_1f1f9_1f1f0, "\ud83c\uddf9\ud83c\uddf0" );
		name( "flag_tl", R.drawable.emj_1f1f9_1f1f1, "\ud83c\uddf9\ud83c\uddf1" );
		name( "flag_tm", R.drawable.emj_1f1f9_1f1f2, "\ud83c\uddf9\ud83c\uddf2" );
		name( "flag_tn", R.drawable.emj_1f1f9_1f1f3, "\ud83c\uddf9\ud83c\uddf3" );
		name( "flag_to", R.drawable.emj_1f1f9_1f1f4, "\ud83c\uddf9\ud83c\uddf4" );
		name( "flag_tr", R.drawable.emj_1f1f9_1f1f7, "\ud83c\uddf9\ud83c\uddf7" );
		name( "flag_tt", R.drawable.emj_1f1f9_1f1f9, "\ud83c\uddf9\ud83c\uddf9" );
		name( "flag_tv", R.drawable.emj_1f1f9_1f1fb, "\ud83c\uddf9\ud83c\uddfb" );
		name( "flag_tw", R.drawable.emj_1f1f9_1f1fc, "\ud83c\uddf9\ud83c\uddfc" );
		name( "flag_tz", R.drawable.emj_1f1f9_1f1ff, "\ud83c\uddf9\ud83c\uddff" );
		name( "flag_ua", R.drawable.emj_1f1fa_1f1e6, "\ud83c\uddfa\ud83c\udde6" );
		name( "flag_ug", R.drawable.emj_1f1fa_1f1ec, "\ud83c\uddfa\ud83c\uddec" );
		name( "flag_um", R.drawable.emj_1f1fa_1f1f2, "\ud83c\uddfa\ud83c\uddf2" );
		name( "flag_un", R.drawable.emj_1f1fa_1f1f3, "\ud83c\uddfa\ud83c\uddf3" );
		name( "flag_us", R.drawable.emj_1f1fa_1f1f8, "\ud83c\uddfa\ud83c\uddf8" );
		name( "flag_uy", R.drawable.emj_1f1fa_1f1fe, "\ud83c\uddfa\ud83c\uddfe" );
		name( "flag_uz", R.drawable.emj_1f1fa_1f1ff, "\ud83c\uddfa\ud83c\uddff" );
		name( "flag_va", R.drawable.emj_1f1fb_1f1e6, "\ud83c\uddfb\ud83c\udde6" );
		name( "flag_vc", R.drawable.emj_1f1fb_1f1e8, "\ud83c\uddfb\ud83c\udde8" );
		name( "flag_ve", R.drawable.emj_1f1fb_1f1ea, "\ud83c\uddfb\ud83c\uddea" );
		name( "flag_vg", R.drawable.emj_1f1fb_1f1ec, "\ud83c\uddfb\ud83c\uddec" );
		name( "flag_vi", R.drawable.emj_1f1fb_1f1ee, "\ud83c\uddfb\ud83c\uddee" );
		name( "flag_vn", R.drawable.emj_1f1fb_1f1f3, "\ud83c\uddfb\ud83c\uddf3" );
		name( "flag_vu", R.drawable.emj_1f1fb_1f1fa, "\ud83c\uddfb\ud83c\uddfa" );
		name( "flag_wales", R.drawable.emj_1f3f4_e0067_e0062_e0077_e006c_e0073_e007f, "\ud83c\udff4\udb40\udc67\udb40\udc62\udb40\udc77\udb40\udc6c\udb40\udc73\udb40\udc7f" );
		name( "flag_wf", R.drawable.emj_1f1fc_1f1eb, "\ud83c\uddfc\ud83c\uddeb" );
		name( "flag_white", R.drawable.emj_1f3f3_fe0f, "\ud83c\udff3\ufe0f" );
		name( "flag_ws", R.drawable.emj_1f1fc_1f1f8, "\ud83c\uddfc\ud83c\uddf8" );
		name( "flag_xk", R.drawable.emj_1f1fd_1f1f0, "\ud83c\uddfd\ud83c\uddf0" );
		name( "flag_ye", R.drawable.emj_1f1fe_1f1ea, "\ud83c\uddfe\ud83c\uddea" );
		name( "flag_yt", R.drawable.emj_1f1fe_1f1f9, "\ud83c\uddfe\ud83c\uddf9" );
		name( "flag_za", R.drawable.emj_1f1ff_1f1e6, "\ud83c\uddff\ud83c\udde6" );
		name( "flag_zm", R.drawable.emj_1f1ff_1f1f2, "\ud83c\uddff\ud83c\uddf2" );
		name( "flag_zw", R.drawable.emj_1f1ff_1f1fc, "\ud83c\uddff\ud83c\uddfc" );
		name( "flags", R.drawable.emj_1f38f, "\ud83c\udf8f" );
		name( "flame", R.drawable.emj_1f525, "\ud83d\udd25" );
		name( "flan", R.drawable.emj_1f36e, "\ud83c\udf6e" );
		name( "flashlight", R.drawable.emj_1f526, "\ud83d\udd26" );
		name( "fleur_de_lis", R.drawable.emj_269c_fe0f, "\u269c\ufe0f" );
		name( "flipper", R.drawable.emj_1f42c, "\ud83d\udc2c" );
		name( "floppy_disk", R.drawable.emj_1f4be, "\ud83d\udcbe" );
		name( "flower_playing_cards", R.drawable.emj_1f3b4, "\ud83c\udfb4" );
		name( "flushed", R.drawable.emj_1f633, "\ud83d\ude33" );
		name( "flying_saucer", R.drawable.emj_1f6f8, "\ud83d\udef8" );
		name( "fm", R.drawable.emj_1f1eb_1f1f2, "\ud83c\uddeb\ud83c\uddf2" );
		name( "fo", R.drawable.emj_1f1eb_1f1f4, "\ud83c\uddeb\ud83c\uddf4" );
		name( "fog", R.drawable.emj_1f32b_fe0f, "\ud83c\udf2b\ufe0f" );
		name( "foggy", R.drawable.emj_1f301, "\ud83c\udf01" );
		name( "football", R.drawable.emj_1f3c8, "\ud83c\udfc8" );
		name( "footprints", R.drawable.emj_1f463, "\ud83d\udc63" );
		name( "fork_and_knife", R.drawable.emj_1f374, "\ud83c\udf74" );
		name( "fork_and_knife_with_plate", R.drawable.emj_1f37d_fe0f, "\ud83c\udf7d\ufe0f" );
		name( "fork_knife_plate", R.drawable.emj_1f37d_fe0f, "\ud83c\udf7d\ufe0f" );
		name( "fortune_cookie", R.drawable.emj_1f960, "\ud83e\udd60" );
		name( "fountain", R.drawable.emj_26f2, "\u26f2" );
		name( "four", R.drawable.emj_0034_fe0f_20e3, "\u0034\ufe0f\u20e3" );
		name( "four_leaf_clover", R.drawable.emj_1f340, "\ud83c\udf40" );
		name( "fox", R.drawable.emj_1f98a, "\ud83e\udd8a" );
		name( "fox_face", R.drawable.emj_1f98a, "\ud83e\udd8a" );
		name( "fr", R.drawable.emj_1f1eb_1f1f7, "\ud83c\uddeb\ud83c\uddf7" );
		name( "frame_photo", R.drawable.emj_1f5bc_fe0f, "\ud83d\uddbc\ufe0f" );
		name( "frame_with_picture", R.drawable.emj_1f5bc_fe0f, "\ud83d\uddbc\ufe0f" );
		name( "free", R.drawable.emj_1f193, "\ud83c\udd93" );
		name( "french_bread", R.drawable.emj_1f956, "\ud83e\udd56" );
		name( "fried_egg", R.drawable.emj_1f373, "\ud83c\udf73" );
		name( "fried_shrimp", R.drawable.emj_1f364, "\ud83c\udf64" );
		name( "fries", R.drawable.emj_1f35f, "\ud83c\udf5f" );
		name( "frog", R.drawable.emj_1f438, "\ud83d\udc38" );
		name( "frowning", R.drawable.emj_1f626, "\ud83d\ude26" );
		name( "frowning2", R.drawable.emj_2639_fe0f, "\u2639\ufe0f" );
		name( "fuelpump", R.drawable.emj_26fd, "\u26fd" );
	}
	private static void init71(){
		name( "full_moon", R.drawable.emj_1f315, "\ud83c\udf15" );
		name( "full_moon_with_face", R.drawable.emj_1f31d, "\ud83c\udf1d" );
		name( "funeral_urn", R.drawable.emj_26b1_fe0f, "\u26b1\ufe0f" );
		name( "ga", R.drawable.emj_1f1ec_1f1e6, "\ud83c\uddec\ud83c\udde6" );
		name( "game_die", R.drawable.emj_1f3b2, "\ud83c\udfb2" );
		name( "gay_pride_flag", R.drawable.emj_1f3f3_fe0f_200d_1f308, "\ud83c\udff3\ufe0f\u200d\ud83c\udf08" );
		name( "gb", R.drawable.emj_1f1ec_1f1e7, "\ud83c\uddec\ud83c\udde7" );
		name( "gd", R.drawable.emj_1f1ec_1f1e9, "\ud83c\uddec\ud83c\udde9" );
		name( "ge", R.drawable.emj_1f1ec_1f1ea, "\ud83c\uddec\ud83c\uddea" );
		name( "gear", R.drawable.emj_2699_fe0f, "\u2699\ufe0f" );
		name( "gem", R.drawable.emj_1f48e, "\ud83d\udc8e" );
		name( "gemini", R.drawable.emj_264a, "\u264a" );
		name( "genie", R.drawable.emj_1f9de, "\ud83e\uddde" );
		name( "gf", R.drawable.emj_1f1ec_1f1eb, "\ud83c\uddec\ud83c\uddeb" );
		name( "gg", R.drawable.emj_1f1ec_1f1ec, "\ud83c\uddec\ud83c\uddec" );
		name( "gh", R.drawable.emj_1f1ec_1f1ed, "\ud83c\uddec\ud83c\udded" );
		name( "ghost", R.drawable.emj_1f47b, "\ud83d\udc7b" );
		name( "gi", R.drawable.emj_1f1ec_1f1ee, "\ud83c\uddec\ud83c\uddee" );
		name( "gift", R.drawable.emj_1f381, "\ud83c\udf81" );
		name( "gift_heart", R.drawable.emj_1f49d, "\ud83d\udc9d" );
		name( "giraffe_face", R.drawable.emj_1f992, "\ud83e\udd92" );
		name( "girl", R.drawable.emj_1f467, "\ud83d\udc67" );
		name( "girl_dark_skin_tone", R.drawable.emj_1f467_1f3ff, "\ud83d\udc67\ud83c\udfff" );
		name( "girl_light_skin_tone", R.drawable.emj_1f467_1f3fb, "\ud83d\udc67\ud83c\udffb" );
		name( "girl_medium_dark_skin_tone", R.drawable.emj_1f467_1f3fe, "\ud83d\udc67\ud83c\udffe" );
		name( "girl_medium_light_skin_tone", R.drawable.emj_1f467_1f3fc, "\ud83d\udc67\ud83c\udffc" );
		name( "girl_medium_skin_tone", R.drawable.emj_1f467_1f3fd, "\ud83d\udc67\ud83c\udffd" );
		name( "girl_tone1", R.drawable.emj_1f467_1f3fb, "\ud83d\udc67\ud83c\udffb" );
		name( "girl_tone2", R.drawable.emj_1f467_1f3fc, "\ud83d\udc67\ud83c\udffc" );
		name( "girl_tone3", R.drawable.emj_1f467_1f3fd, "\ud83d\udc67\ud83c\udffd" );
		name( "girl_tone4", R.drawable.emj_1f467_1f3fe, "\ud83d\udc67\ud83c\udffe" );
		name( "girl_tone5", R.drawable.emj_1f467_1f3ff, "\ud83d\udc67\ud83c\udfff" );
		name( "gl", R.drawable.emj_1f1ec_1f1f1, "\ud83c\uddec\ud83c\uddf1" );
		name( "glass_of_milk", R.drawable.emj_1f95b, "\ud83e\udd5b" );
		name( "globe_with_meridians", R.drawable.emj_1f310, "\ud83c\udf10" );
		name( "gloves", R.drawable.emj_1f9e4, "\ud83e\udde4" );
		name( "gm", R.drawable.emj_1f1ec_1f1f2, "\ud83c\uddec\ud83c\uddf2" );
		name( "gn", R.drawable.emj_1f1ec_1f1f3, "\ud83c\uddec\ud83c\uddf3" );
		name( "goal", R.drawable.emj_1f945, "\ud83e\udd45" );
		name( "goal_net", R.drawable.emj_1f945, "\ud83e\udd45" );
		name( "goat", R.drawable.emj_1f410, "\ud83d\udc10" );
		name( "golf", R.drawable.emj_26f3, "\u26f3" );
		name( "golfer", R.drawable.emj_1f3cc_fe0f, "\ud83c\udfcc\ufe0f" );
		name( "golfer_dark_skin_tone", R.drawable.emj_1f3cc_1f3ff, "\ud83c\udfcc\ud83c\udfff" );
		name( "golfer_light_skin_tone", R.drawable.emj_1f3cc_1f3fb, "\ud83c\udfcc\ud83c\udffb" );
		name( "golfer_medium_dark_skin_tone", R.drawable.emj_1f3cc_1f3fe, "\ud83c\udfcc\ud83c\udffe" );
		name( "golfer_medium_light_skin_tone", R.drawable.emj_1f3cc_1f3fc, "\ud83c\udfcc\ud83c\udffc" );
		name( "golfer_medium_skin_tone", R.drawable.emj_1f3cc_1f3fd, "\ud83c\udfcc\ud83c\udffd" );
		name( "golfer_tone1", R.drawable.emj_1f3cc_1f3fb, "\ud83c\udfcc\ud83c\udffb" );
		name( "golfer_tone2", R.drawable.emj_1f3cc_1f3fc, "\ud83c\udfcc\ud83c\udffc" );
		name( "golfer_tone3", R.drawable.emj_1f3cc_1f3fd, "\ud83c\udfcc\ud83c\udffd" );
		name( "golfer_tone4", R.drawable.emj_1f3cc_1f3fe, "\ud83c\udfcc\ud83c\udffe" );
		name( "golfer_tone5", R.drawable.emj_1f3cc_1f3ff, "\ud83c\udfcc\ud83c\udfff" );
		name( "gorilla", R.drawable.emj_1f98d, "\ud83e\udd8d" );
		name( "gp", R.drawable.emj_1f1ec_1f1f5, "\ud83c\uddec\ud83c\uddf5" );
		name( "gq", R.drawable.emj_1f1ec_1f1f6, "\ud83c\uddec\ud83c\uddf6" );
		name( "gr", R.drawable.emj_1f1ec_1f1f7, "\ud83c\uddec\ud83c\uddf7" );
		name( "grandma", R.drawable.emj_1f475, "\ud83d\udc75" );
		name( "grandma_tone1", R.drawable.emj_1f475_1f3fb, "\ud83d\udc75\ud83c\udffb" );
		name( "grandma_tone2", R.drawable.emj_1f475_1f3fc, "\ud83d\udc75\ud83c\udffc" );
		name( "grandma_tone3", R.drawable.emj_1f475_1f3fd, "\ud83d\udc75\ud83c\udffd" );
		name( "grandma_tone4", R.drawable.emj_1f475_1f3fe, "\ud83d\udc75\ud83c\udffe" );
		name( "grandma_tone5", R.drawable.emj_1f475_1f3ff, "\ud83d\udc75\ud83c\udfff" );
		name( "grapes", R.drawable.emj_1f347, "\ud83c\udf47" );
		name( "green_apple", R.drawable.emj_1f34f, "\ud83c\udf4f" );
		name( "green_book", R.drawable.emj_1f4d7, "\ud83d\udcd7" );
		name( "green_heart", R.drawable.emj_1f49a, "\ud83d\udc9a" );
		name( "green_salad", R.drawable.emj_1f957, "\ud83e\udd57" );
		name( "grey_exclamation", R.drawable.emj_2755, "\u2755" );
		name( "grey_question", R.drawable.emj_2754, "\u2754" );
		name( "grimacing", R.drawable.emj_1f62c, "\ud83d\ude2c" );
		name( "grin", R.drawable.emj_1f601, "\ud83d\ude01" );
		name( "grinning", R.drawable.emj_1f600, "\ud83d\ude00" );
		name( "grinning_face_with_one_large_and_one_small_eye", R.drawable.emj_1f92a, "\ud83e\udd2a" );
		name( "grinning_face_with_star_eyes", R.drawable.emj_1f929, "\ud83e\udd29" );
		name( "gs", R.drawable.emj_1f1ec_1f1f8, "\ud83c\uddec\ud83c\uddf8" );
		name( "gt", R.drawable.emj_1f1ec_1f1f9, "\ud83c\uddec\ud83c\uddf9" );
		name( "gu", R.drawable.emj_1f1ec_1f1fa, "\ud83c\uddec\ud83c\uddfa" );
		name( "guardsman", R.drawable.emj_1f482, "\ud83d\udc82" );
		name( "guardsman_dark_skin_tone", R.drawable.emj_1f482_1f3ff, "\ud83d\udc82\ud83c\udfff" );
		name( "guardsman_light_skin_tone", R.drawable.emj_1f482_1f3fb, "\ud83d\udc82\ud83c\udffb" );
		name( "guardsman_medium_dark_skin_tone", R.drawable.emj_1f482_1f3fe, "\ud83d\udc82\ud83c\udffe" );
		name( "guardsman_medium_light_skin_tone", R.drawable.emj_1f482_1f3fc, "\ud83d\udc82\ud83c\udffc" );
		name( "guardsman_medium_skin_tone", R.drawable.emj_1f482_1f3fd, "\ud83d\udc82\ud83c\udffd" );
		name( "guardsman_tone1", R.drawable.emj_1f482_1f3fb, "\ud83d\udc82\ud83c\udffb" );
		name( "guardsman_tone2", R.drawable.emj_1f482_1f3fc, "\ud83d\udc82\ud83c\udffc" );
		name( "guardsman_tone3", R.drawable.emj_1f482_1f3fd, "\ud83d\udc82\ud83c\udffd" );
		name( "guardsman_tone4", R.drawable.emj_1f482_1f3fe, "\ud83d\udc82\ud83c\udffe" );
		name( "guardsman_tone5", R.drawable.emj_1f482_1f3ff, "\ud83d\udc82\ud83c\udfff" );
		name( "guitar", R.drawable.emj_1f3b8, "\ud83c\udfb8" );
		name( "gun", R.drawable.emj_1f52b, "\ud83d\udd2b" );
		name( "gw", R.drawable.emj_1f1ec_1f1fc, "\ud83c\uddec\ud83c\uddfc" );
		name( "gy", R.drawable.emj_1f1ec_1f1fe, "\ud83c\uddec\ud83c\uddfe" );
		name( "haircut", R.drawable.emj_1f487, "\ud83d\udc87" );
		name( "haircut_dark_skin_tone", R.drawable.emj_1f487_1f3ff, "\ud83d\udc87\ud83c\udfff" );
		name( "haircut_light_skin_tone", R.drawable.emj_1f487_1f3fb, "\ud83d\udc87\ud83c\udffb" );
		name( "haircut_medium_dark_skin_tone", R.drawable.emj_1f487_1f3fe, "\ud83d\udc87\ud83c\udffe" );
		name( "haircut_medium_light_skin_tone", R.drawable.emj_1f487_1f3fc, "\ud83d\udc87\ud83c\udffc" );
		name( "haircut_medium_skin_tone", R.drawable.emj_1f487_1f3fd, "\ud83d\udc87\ud83c\udffd" );
		name( "haircut_tone1", R.drawable.emj_1f487_1f3fb, "\ud83d\udc87\ud83c\udffb" );
		name( "haircut_tone2", R.drawable.emj_1f487_1f3fc, "\ud83d\udc87\ud83c\udffc" );
	}
	private static void init72(){
		name( "haircut_tone3", R.drawable.emj_1f487_1f3fd, "\ud83d\udc87\ud83c\udffd" );
		name( "haircut_tone4", R.drawable.emj_1f487_1f3fe, "\ud83d\udc87\ud83c\udffe" );
		name( "haircut_tone5", R.drawable.emj_1f487_1f3ff, "\ud83d\udc87\ud83c\udfff" );
		name( "hamburger", R.drawable.emj_1f354, "\ud83c\udf54" );
		name( "hammer", R.drawable.emj_1f528, "\ud83d\udd28" );
		name( "hammer_and_pick", R.drawable.emj_2692_fe0f, "\u2692\ufe0f" );
		name( "hammer_and_wrench", R.drawable.emj_1f6e0_fe0f, "\ud83d\udee0\ufe0f" );
		name( "hammer_pick", R.drawable.emj_2692_fe0f, "\u2692\ufe0f" );
		name( "hamster", R.drawable.emj_1f439, "\ud83d\udc39" );
		name( "hand", R.drawable.emj_270b, "\u270b" );
		name( "hand_dark_skin_tone", R.drawable.emj_270b_1f3ff, "\u270b\ud83c\udfff" );
		name( "hand_light_skin_tone", R.drawable.emj_270b_1f3fb, "\u270b\ud83c\udffb" );
		name( "hand_medium_dark_skin_tone", R.drawable.emj_270b_1f3fe, "\u270b\ud83c\udffe" );
		name( "hand_medium_light_skin_tone", R.drawable.emj_270b_1f3fc, "\u270b\ud83c\udffc" );
		name( "hand_medium_skin_tone", R.drawable.emj_270b_1f3fd, "\u270b\ud83c\udffd" );
		name( "hand_splayed", R.drawable.emj_1f590_fe0f, "\ud83d\udd90\ufe0f" );
		name( "hand_splayed_tone1", R.drawable.emj_1f590_1f3fb, "\ud83d\udd90\ud83c\udffb" );
		name( "hand_splayed_tone2", R.drawable.emj_1f590_1f3fc, "\ud83d\udd90\ud83c\udffc" );
		name( "hand_splayed_tone3", R.drawable.emj_1f590_1f3fd, "\ud83d\udd90\ud83c\udffd" );
		name( "hand_splayed_tone4", R.drawable.emj_1f590_1f3fe, "\ud83d\udd90\ud83c\udffe" );
		name( "hand_splayed_tone5", R.drawable.emj_1f590_1f3ff, "\ud83d\udd90\ud83c\udfff" );
		name( "hand_tone1", R.drawable.emj_270b_1f3fb, "\u270b\ud83c\udffb" );
		name( "hand_tone2", R.drawable.emj_270b_1f3fc, "\u270b\ud83c\udffc" );
		name( "hand_tone3", R.drawable.emj_270b_1f3fd, "\u270b\ud83c\udffd" );
		name( "hand_tone4", R.drawable.emj_270b_1f3fe, "\u270b\ud83c\udffe" );
		name( "hand_tone5", R.drawable.emj_270b_1f3ff, "\u270b\ud83c\udfff" );
		name( "hand_with_index_and_middle_finger_crossed", R.drawable.emj_1f91e, "\ud83e\udd1e" );
		name( "hand_with_index_and_middle_fingers_crossed", R.drawable.emj_1f91e, "\ud83e\udd1e" );
		name( "hand_with_index_and_middle_fingers_crossed_dark_skin_tone", R.drawable.emj_1f91e_1f3ff, "\ud83e\udd1e\ud83c\udfff" );
		name( "hand_with_index_and_middle_fingers_crossed_light_skin_tone", R.drawable.emj_1f91e_1f3fb, "\ud83e\udd1e\ud83c\udffb" );
		name( "hand_with_index_and_middle_fingers_crossed_medium_dark_skin_tone", R.drawable.emj_1f91e_1f3fe, "\ud83e\udd1e\ud83c\udffe" );
		name( "hand_with_index_and_middle_fingers_crossed_medium_light_skin_tone", R.drawable.emj_1f91e_1f3fc, "\ud83e\udd1e\ud83c\udffc" );
		name( "hand_with_index_and_middle_fingers_crossed_medium_skin_tone", R.drawable.emj_1f91e_1f3fd, "\ud83e\udd1e\ud83c\udffd" );
		name( "hand_with_index_and_middle_fingers_crossed_tone1", R.drawable.emj_1f91e_1f3fb, "\ud83e\udd1e\ud83c\udffb" );
		name( "hand_with_index_and_middle_fingers_crossed_tone2", R.drawable.emj_1f91e_1f3fc, "\ud83e\udd1e\ud83c\udffc" );
		name( "hand_with_index_and_middle_fingers_crossed_tone3", R.drawable.emj_1f91e_1f3fd, "\ud83e\udd1e\ud83c\udffd" );
		name( "hand_with_index_and_middle_fingers_crossed_tone4", R.drawable.emj_1f91e_1f3fe, "\ud83e\udd1e\ud83c\udffe" );
		name( "hand_with_index_and_middle_fingers_crossed_tone5", R.drawable.emj_1f91e_1f3ff, "\ud83e\udd1e\ud83c\udfff" );
		name( "handbag", R.drawable.emj_1f45c, "\ud83d\udc5c" );
		name( "handball", R.drawable.emj_1f93e, "\ud83e\udd3e" );
		name( "handball_dark_skin_tone", R.drawable.emj_1f93e_1f3ff, "\ud83e\udd3e\ud83c\udfff" );
		name( "handball_light_skin_tone", R.drawable.emj_1f93e_1f3fb, "\ud83e\udd3e\ud83c\udffb" );
		name( "handball_medium_dark_skin_tone", R.drawable.emj_1f93e_1f3fe, "\ud83e\udd3e\ud83c\udffe" );
		name( "handball_medium_light_skin_tone", R.drawable.emj_1f93e_1f3fc, "\ud83e\udd3e\ud83c\udffc" );
		name( "handball_medium_skin_tone", R.drawable.emj_1f93e_1f3fd, "\ud83e\udd3e\ud83c\udffd" );
		name( "handball_tone1", R.drawable.emj_1f93e_1f3fb, "\ud83e\udd3e\ud83c\udffb" );
		name( "handball_tone2", R.drawable.emj_1f93e_1f3fc, "\ud83e\udd3e\ud83c\udffc" );
		name( "handball_tone3", R.drawable.emj_1f93e_1f3fd, "\ud83e\udd3e\ud83c\udffd" );
		name( "handball_tone4", R.drawable.emj_1f93e_1f3fe, "\ud83e\udd3e\ud83c\udffe" );
		name( "handball_tone5", R.drawable.emj_1f93e_1f3ff, "\ud83e\udd3e\ud83c\udfff" );
		name( "handshake", R.drawable.emj_1f91d, "\ud83e\udd1d" );
		name( "hankey", R.drawable.emj_1f4a9, "\ud83d\udca9" );
		name( "hash", R.drawable.emj_0023_fe0f_20e3, "\u0023\ufe0f\u20e3" );
		name( "hatched_chick", R.drawable.emj_1f425, "\ud83d\udc25" );
		name( "hatching_chick", R.drawable.emj_1f423, "\ud83d\udc23" );
		name( "head_bandage", R.drawable.emj_1f915, "\ud83e\udd15" );
		name( "headphones", R.drawable.emj_1f3a7, "\ud83c\udfa7" );
		name( "hear_no_evil", R.drawable.emj_1f649, "\ud83d\ude49" );
		name( "heart", R.drawable.emj_2764_fe0f, "\u2764\ufe0f" );
		name( "heart_decoration", R.drawable.emj_1f49f, "\ud83d\udc9f" );
		name( "heart_exclamation", R.drawable.emj_2763_fe0f, "\u2763\ufe0f" );
		name( "heart_eyes", R.drawable.emj_1f60d, "\ud83d\ude0d" );
		name( "heart_eyes_cat", R.drawable.emj_1f63b, "\ud83d\ude3b" );
		name( "heartbeat", R.drawable.emj_1f493, "\ud83d\udc93" );
		name( "heartpulse", R.drawable.emj_1f497, "\ud83d\udc97" );
		name( "hearts", R.drawable.emj_2665_fe0f, "\u2665\ufe0f" );
		name( "heavy_check_mark", R.drawable.emj_2714_fe0f, "\u2714\ufe0f" );
		name( "heavy_division_sign", R.drawable.emj_2797, "\u2797" );
		name( "heavy_dollar_sign", R.drawable.emj_1f4b2, "\ud83d\udcb2" );
		name( "heavy_exclamation_mark", R.drawable.emj_2757, "\u2757" );
		name( "heavy_heart_exclamation_mark_ornament", R.drawable.emj_2763_fe0f, "\u2763\ufe0f" );
		name( "heavy_minus_sign", R.drawable.emj_2796, "\u2796" );
		name( "heavy_multiplication_x", R.drawable.emj_2716_fe0f, "\u2716\ufe0f" );
		name( "heavy_plus_sign", R.drawable.emj_2795, "\u2795" );
		name( "hedgehog", R.drawable.emj_1f994, "\ud83e\udd94" );
		name( "helicopter", R.drawable.emj_1f681, "\ud83d\ude81" );
		name( "helmet_with_cross", R.drawable.emj_26d1_fe0f, "\u26d1\ufe0f" );
		name( "helmet_with_white_cross", R.drawable.emj_26d1_fe0f, "\u26d1\ufe0f" );
		name( "herb", R.drawable.emj_1f33f, "\ud83c\udf3f" );
		name( "hibiscus", R.drawable.emj_1f33a, "\ud83c\udf3a" );
		name( "high_brightness", R.drawable.emj_1f506, "\ud83d\udd06" );
		name( "high_heel", R.drawable.emj_1f460, "\ud83d\udc60" );
		name( "hk", R.drawable.emj_1f1ed_1f1f0, "\ud83c\udded\ud83c\uddf0" );
		name( "hm", R.drawable.emj_1f1ed_1f1f2, "\ud83c\udded\ud83c\uddf2" );
		name( "hn", R.drawable.emj_1f1ed_1f1f3, "\ud83c\udded\ud83c\uddf3" );
		name( "hocho", R.drawable.emj_1f52a, "\ud83d\udd2a" );
		name( "hockey", R.drawable.emj_1f3d2, "\ud83c\udfd2" );
		name( "hole", R.drawable.emj_1f573_fe0f, "\ud83d\udd73\ufe0f" );
		name( "homes", R.drawable.emj_1f3d8_fe0f, "\ud83c\udfd8\ufe0f" );
		name( "honey_pot", R.drawable.emj_1f36f, "\ud83c\udf6f" );
		name( "honeybee", R.drawable.emj_1f41d, "\ud83d\udc1d" );
		name( "horse", R.drawable.emj_1f434, "\ud83d\udc34" );
		name( "horse_racing", R.drawable.emj_1f3c7, "\ud83c\udfc7" );
		name( "horse_racing_dark_skin_tone", R.drawable.emj_1f3c7_1f3ff, "\ud83c\udfc7\ud83c\udfff" );
		name( "horse_racing_light_skin_tone", R.drawable.emj_1f3c7_1f3fb, "\ud83c\udfc7\ud83c\udffb" );
		name( "horse_racing_medium_dark_skin_tone", R.drawable.emj_1f3c7_1f3fe, "\ud83c\udfc7\ud83c\udffe" );
		name( "horse_racing_medium_light_skin_tone", R.drawable.emj_1f3c7_1f3fc, "\ud83c\udfc7\ud83c\udffc" );
		name( "horse_racing_medium_skin_tone", R.drawable.emj_1f3c7_1f3fd, "\ud83c\udfc7\ud83c\udffd" );
		name( "horse_racing_tone1", R.drawable.emj_1f3c7_1f3fb, "\ud83c\udfc7\ud83c\udffb" );
		name( "horse_racing_tone2", R.drawable.emj_1f3c7_1f3fc, "\ud83c\udfc7\ud83c\udffc" );
		name( "horse_racing_tone3", R.drawable.emj_1f3c7_1f3fd, "\ud83c\udfc7\ud83c\udffd" );
	}
	private static void init73(){
		name( "horse_racing_tone4", R.drawable.emj_1f3c7_1f3fe, "\ud83c\udfc7\ud83c\udffe" );
		name( "horse_racing_tone5", R.drawable.emj_1f3c7_1f3ff, "\ud83c\udfc7\ud83c\udfff" );
		name( "hospital", R.drawable.emj_1f3e5, "\ud83c\udfe5" );
		name( "hot_dog", R.drawable.emj_1f32d, "\ud83c\udf2d" );
		name( "hot_pepper", R.drawable.emj_1f336_fe0f, "\ud83c\udf36\ufe0f" );
		name( "hotdog", R.drawable.emj_1f32d, "\ud83c\udf2d" );
		name( "hotel", R.drawable.emj_1f3e8, "\ud83c\udfe8" );
		name( "hotsprings", R.drawable.emj_2668_fe0f, "\u2668\ufe0f" );
		name( "hourglass", R.drawable.emj_231b, "\u231b" );
		name( "hourglass_flowing_sand", R.drawable.emj_23f3, "\u23f3" );
		name( "house", R.drawable.emj_1f3e0, "\ud83c\udfe0" );
		name( "house_abandoned", R.drawable.emj_1f3da_fe0f, "\ud83c\udfda\ufe0f" );
		name( "house_buildings", R.drawable.emj_1f3d8_fe0f, "\ud83c\udfd8\ufe0f" );
		name( "house_with_garden", R.drawable.emj_1f3e1, "\ud83c\udfe1" );
		name( "hr", R.drawable.emj_1f1ed_1f1f7, "\ud83c\udded\ud83c\uddf7" );
		name( "ht", R.drawable.emj_1f1ed_1f1f9, "\ud83c\udded\ud83c\uddf9" );
		name( "hu", R.drawable.emj_1f1ed_1f1fa, "\ud83c\udded\ud83c\uddfa" );
		name( "hugging", R.drawable.emj_1f917, "\ud83e\udd17" );
		name( "hugging_face", R.drawable.emj_1f917, "\ud83e\udd17" );
		name( "hushed", R.drawable.emj_1f62f, "\ud83d\ude2f" );
		name( "i_love_you_hand_sign", R.drawable.emj_1f91f, "\ud83e\udd1f" );
		name( "i_love_you_hand_sign_dark_skin_tone", R.drawable.emj_1f91f_1f3ff, "\ud83e\udd1f\ud83c\udfff" );
		name( "i_love_you_hand_sign_light_skin_tone", R.drawable.emj_1f91f_1f3fb, "\ud83e\udd1f\ud83c\udffb" );
		name( "i_love_you_hand_sign_medium_dark_skin_tone", R.drawable.emj_1f91f_1f3fe, "\ud83e\udd1f\ud83c\udffe" );
		name( "i_love_you_hand_sign_medium_light_skin_tone", R.drawable.emj_1f91f_1f3fc, "\ud83e\udd1f\ud83c\udffc" );
		name( "i_love_you_hand_sign_medium_skin_tone", R.drawable.emj_1f91f_1f3fd, "\ud83e\udd1f\ud83c\udffd" );
		name( "i_love_you_hand_sign_tone1", R.drawable.emj_1f91f_1f3fb, "\ud83e\udd1f\ud83c\udffb" );
		name( "i_love_you_hand_sign_tone2", R.drawable.emj_1f91f_1f3fc, "\ud83e\udd1f\ud83c\udffc" );
		name( "i_love_you_hand_sign_tone3", R.drawable.emj_1f91f_1f3fd, "\ud83e\udd1f\ud83c\udffd" );
		name( "i_love_you_hand_sign_tone4", R.drawable.emj_1f91f_1f3fe, "\ud83e\udd1f\ud83c\udffe" );
		name( "i_love_you_hand_sign_tone5", R.drawable.emj_1f91f_1f3ff, "\ud83e\udd1f\ud83c\udfff" );
		name( "ic", R.drawable.emj_1f1ee_1f1e8, "\ud83c\uddee\ud83c\udde8" );
		name( "ice_cream", R.drawable.emj_1f368, "\ud83c\udf68" );
		name( "ice_hockey_stick_and_puck", R.drawable.emj_1f3d2, "\ud83c\udfd2" );
		name( "ice_skate", R.drawable.emj_26f8_fe0f, "\u26f8\ufe0f" );
		name( "icecream", R.drawable.emj_1f366, "\ud83c\udf66" );
		name( "id", R.drawable.emj_1f194, "\ud83c\udd94" );
		name( "ideograph_advantage", R.drawable.emj_1f250, "\ud83c\ude50" );
		name( "ie", R.drawable.emj_1f1ee_1f1ea, "\ud83c\uddee\ud83c\uddea" );
		name( "il", R.drawable.emj_1f1ee_1f1f1, "\ud83c\uddee\ud83c\uddf1" );
		name( "im", R.drawable.emj_1f1ee_1f1f2, "\ud83c\uddee\ud83c\uddf2" );
		name( "imp", R.drawable.emj_1f47f, "\ud83d\udc7f" );
		name( "in", R.drawable.emj_1f1ee_1f1f3, "\ud83c\uddee\ud83c\uddf3" );
		name( "inbox_tray", R.drawable.emj_1f4e5, "\ud83d\udce5" );
		name( "incoming_envelope", R.drawable.emj_1f4e8, "\ud83d\udce8" );
		name( "indonesia", R.drawable.emj_1f1ee_1f1e9, "\ud83c\uddee\ud83c\udde9" );
		name( "information_desk_person", R.drawable.emj_1f481, "\ud83d\udc81" );
		name( "information_desk_person_dark_skin_tone", R.drawable.emj_1f481_1f3ff, "\ud83d\udc81\ud83c\udfff" );
		name( "information_desk_person_light_skin_tone", R.drawable.emj_1f481_1f3fb, "\ud83d\udc81\ud83c\udffb" );
		name( "information_desk_person_medium_dark_skin_tone", R.drawable.emj_1f481_1f3fe, "\ud83d\udc81\ud83c\udffe" );
		name( "information_desk_person_medium_light_skin_tone", R.drawable.emj_1f481_1f3fc, "\ud83d\udc81\ud83c\udffc" );
		name( "information_desk_person_medium_skin_tone", R.drawable.emj_1f481_1f3fd, "\ud83d\udc81\ud83c\udffd" );
		name( "information_desk_person_tone1", R.drawable.emj_1f481_1f3fb, "\ud83d\udc81\ud83c\udffb" );
		name( "information_desk_person_tone2", R.drawable.emj_1f481_1f3fc, "\ud83d\udc81\ud83c\udffc" );
		name( "information_desk_person_tone3", R.drawable.emj_1f481_1f3fd, "\ud83d\udc81\ud83c\udffd" );
		name( "information_desk_person_tone4", R.drawable.emj_1f481_1f3fe, "\ud83d\udc81\ud83c\udffe" );
		name( "information_desk_person_tone5", R.drawable.emj_1f481_1f3ff, "\ud83d\udc81\ud83c\udfff" );
		name( "information_source", R.drawable.emj_2139_fe0f, "\u2139\ufe0f" );
		name( "innocent", R.drawable.emj_1f607, "\ud83d\ude07" );
		name( "interrobang", R.drawable.emj_2049_fe0f, "\u2049\ufe0f" );
		name( "io", R.drawable.emj_1f1ee_1f1f4, "\ud83c\uddee\ud83c\uddf4" );
		name( "iphone", R.drawable.emj_1f4f1, "\ud83d\udcf1" );
		name( "iq", R.drawable.emj_1f1ee_1f1f6, "\ud83c\uddee\ud83c\uddf6" );
		name( "ir", R.drawable.emj_1f1ee_1f1f7, "\ud83c\uddee\ud83c\uddf7" );
		name( "is", R.drawable.emj_1f1ee_1f1f8, "\ud83c\uddee\ud83c\uddf8" );
		name( "island", R.drawable.emj_1f3dd_fe0f, "\ud83c\udfdd\ufe0f" );
		name( "it", R.drawable.emj_1f1ee_1f1f9, "\ud83c\uddee\ud83c\uddf9" );
		name( "izakaya_lantern", R.drawable.emj_1f3ee, "\ud83c\udfee" );
		name( "jack_o_lantern", R.drawable.emj_1f383, "\ud83c\udf83" );
		name( "japan", R.drawable.emj_1f5fe, "\ud83d\uddfe" );
		name( "japanese_castle", R.drawable.emj_1f3ef, "\ud83c\udfef" );
		name( "japanese_goblin", R.drawable.emj_1f47a, "\ud83d\udc7a" );
		name( "japanese_ogre", R.drawable.emj_1f479, "\ud83d\udc79" );
		name( "je", R.drawable.emj_1f1ef_1f1ea, "\ud83c\uddef\ud83c\uddea" );
		name( "jeans", R.drawable.emj_1f456, "\ud83d\udc56" );
		name( "jm", R.drawable.emj_1f1ef_1f1f2, "\ud83c\uddef\ud83c\uddf2" );
		name( "jo", R.drawable.emj_1f1ef_1f1f4, "\ud83c\uddef\ud83c\uddf4" );
		name( "joy", R.drawable.emj_1f602, "\ud83d\ude02" );
		name( "joy_cat", R.drawable.emj_1f639, "\ud83d\ude39" );
		name( "joystick", R.drawable.emj_1f579_fe0f, "\ud83d\udd79\ufe0f" );
		name( "jp", R.drawable.emj_1f1ef_1f1f5, "\ud83c\uddef\ud83c\uddf5" );
		name( "juggler", R.drawable.emj_1f939, "\ud83e\udd39" );
		name( "juggler_tone1", R.drawable.emj_1f939_1f3fb, "\ud83e\udd39\ud83c\udffb" );
		name( "juggler_tone2", R.drawable.emj_1f939_1f3fc, "\ud83e\udd39\ud83c\udffc" );
		name( "juggler_tone3", R.drawable.emj_1f939_1f3fd, "\ud83e\udd39\ud83c\udffd" );
		name( "juggler_tone4", R.drawable.emj_1f939_1f3fe, "\ud83e\udd39\ud83c\udffe" );
		name( "juggler_tone5", R.drawable.emj_1f939_1f3ff, "\ud83e\udd39\ud83c\udfff" );
		name( "juggling", R.drawable.emj_1f939, "\ud83e\udd39" );
		name( "juggling_dark_skin_tone", R.drawable.emj_1f939_1f3ff, "\ud83e\udd39\ud83c\udfff" );
		name( "juggling_light_skin_tone", R.drawable.emj_1f939_1f3fb, "\ud83e\udd39\ud83c\udffb" );
		name( "juggling_medium_dark_skin_tone", R.drawable.emj_1f939_1f3fe, "\ud83e\udd39\ud83c\udffe" );
		name( "juggling_medium_light_skin_tone", R.drawable.emj_1f939_1f3fc, "\ud83e\udd39\ud83c\udffc" );
		name( "juggling_medium_skin_tone", R.drawable.emj_1f939_1f3fd, "\ud83e\udd39\ud83c\udffd" );
		name( "juggling_tone1", R.drawable.emj_1f939_1f3fb, "\ud83e\udd39\ud83c\udffb" );
		name( "juggling_tone2", R.drawable.emj_1f939_1f3fc, "\ud83e\udd39\ud83c\udffc" );
		name( "juggling_tone3", R.drawable.emj_1f939_1f3fd, "\ud83e\udd39\ud83c\udffd" );
		name( "juggling_tone4", R.drawable.emj_1f939_1f3fe, "\ud83e\udd39\ud83c\udffe" );
		name( "juggling_tone5", R.drawable.emj_1f939_1f3ff, "\ud83e\udd39\ud83c\udfff" );
		name( "kaaba", R.drawable.emj_1f54b, "\ud83d\udd4b" );
		name( "karate_uniform", R.drawable.emj_1f94b, "\ud83e\udd4b" );
		name( "kayak", R.drawable.emj_1f6f6, "\ud83d\udef6" );
	}
	private static void init74(){
		name( "ke", R.drawable.emj_1f1f0_1f1ea, "\ud83c\uddf0\ud83c\uddea" );
		name( "key", R.drawable.emj_1f511, "\ud83d\udd11" );
		name( "key2", R.drawable.emj_1f5dd_fe0f, "\ud83d\udddd\ufe0f" );
		name( "keyboard", R.drawable.emj_2328_fe0f, "\u2328\ufe0f" );
		name( "keycap_asterisk", R.drawable.emj_002a_fe0f_20e3, "\u002a\ufe0f\u20e3" );
		name( "keycap_star", R.drawable.emj_002a_fe0f_20e3, "\u002a\ufe0f\u20e3" );
		name( "keycap_ten", R.drawable.emj_1f51f, "\ud83d\udd1f" );
		name( "kg", R.drawable.emj_1f1f0_1f1ec, "\ud83c\uddf0\ud83c\uddec" );
		name( "kh", R.drawable.emj_1f1f0_1f1ed, "\ud83c\uddf0\ud83c\udded" );
		name( "ki", R.drawable.emj_1f1f0_1f1ee, "\ud83c\uddf0\ud83c\uddee" );
		name( "kimono", R.drawable.emj_1f458, "\ud83d\udc58" );
		name( "kiss", R.drawable.emj_1f48b, "\ud83d\udc8b" );
		name( "kiss_mm", R.drawable.emj_1f468_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68" );
		name( "kiss_woman_man", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68" );
		name( "kiss_ww", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69" );
		name( "kissing", R.drawable.emj_1f617, "\ud83d\ude17" );
		name( "kissing_cat", R.drawable.emj_1f63d, "\ud83d\ude3d" );
		name( "kissing_closed_eyes", R.drawable.emj_1f61a, "\ud83d\ude1a" );
		name( "kissing_heart", R.drawable.emj_1f618, "\ud83d\ude18" );
		name( "kissing_smiling_eyes", R.drawable.emj_1f619, "\ud83d\ude19" );
		name( "kiwi", R.drawable.emj_1f95d, "\ud83e\udd5d" );
		name( "kiwifruit", R.drawable.emj_1f95d, "\ud83e\udd5d" );
		name( "km", R.drawable.emj_1f1f0_1f1f2, "\ud83c\uddf0\ud83c\uddf2" );
		name( "kn", R.drawable.emj_1f1f0_1f1f3, "\ud83c\uddf0\ud83c\uddf3" );
		name( "knife", R.drawable.emj_1f52a, "\ud83d\udd2a" );
		name( "knife_fork_plate", R.drawable.emj_1f37d_fe0f, "\ud83c\udf7d\ufe0f" );
		name( "koala", R.drawable.emj_1f428, "\ud83d\udc28" );
		name( "koko", R.drawable.emj_1f201, "\ud83c\ude01" );
		name( "kp", R.drawable.emj_1f1f0_1f1f5, "\ud83c\uddf0\ud83c\uddf5" );
		name( "kr", R.drawable.emj_1f1f0_1f1f7, "\ud83c\uddf0\ud83c\uddf7" );
		name( "kw", R.drawable.emj_1f1f0_1f1fc, "\ud83c\uddf0\ud83c\uddfc" );
		name( "ky", R.drawable.emj_1f1f0_1f1fe, "\ud83c\uddf0\ud83c\uddfe" );
		name( "kz", R.drawable.emj_1f1f0_1f1ff, "\ud83c\uddf0\ud83c\uddff" );
		name( "la", R.drawable.emj_1f1f1_1f1e6, "\ud83c\uddf1\ud83c\udde6" );
		name( "label", R.drawable.emj_1f3f7_fe0f, "\ud83c\udff7\ufe0f" );
		name( "lantern", R.drawable.emj_1f3ee, "\ud83c\udfee" );
		name( "large_blue_circle", R.drawable.emj_1f535, "\ud83d\udd35" );
		name( "large_blue_diamond", R.drawable.emj_1f537, "\ud83d\udd37" );
		name( "large_orange_diamond", R.drawable.emj_1f536, "\ud83d\udd36" );
		name( "last_quarter_moon", R.drawable.emj_1f317, "\ud83c\udf17" );
		name( "last_quarter_moon_with_face", R.drawable.emj_1f31c, "\ud83c\udf1c" );
		name( "latin_cross", R.drawable.emj_271d_fe0f, "\u271d\ufe0f" );
		name( "laughing", R.drawable.emj_1f606, "\ud83d\ude06" );
		name( "lb", R.drawable.emj_1f1f1_1f1e7, "\ud83c\uddf1\ud83c\udde7" );
		name( "lc", R.drawable.emj_1f1f1_1f1e8, "\ud83c\uddf1\ud83c\udde8" );
		name( "leaves", R.drawable.emj_1f343, "\ud83c\udf43" );
		name( "ledger", R.drawable.emj_1f4d2, "\ud83d\udcd2" );
		name( "left_facing_fist", R.drawable.emj_1f91b, "\ud83e\udd1b" );
		name( "left_facing_fist_dark_skin_tone", R.drawable.emj_1f91b_1f3ff, "\ud83e\udd1b\ud83c\udfff" );
		name( "left_facing_fist_light_skin_tone", R.drawable.emj_1f91b_1f3fb, "\ud83e\udd1b\ud83c\udffb" );
		name( "left_facing_fist_medium_dark_skin_tone", R.drawable.emj_1f91b_1f3fe, "\ud83e\udd1b\ud83c\udffe" );
		name( "left_facing_fist_medium_light_skin_tone", R.drawable.emj_1f91b_1f3fc, "\ud83e\udd1b\ud83c\udffc" );
		name( "left_facing_fist_medium_skin_tone", R.drawable.emj_1f91b_1f3fd, "\ud83e\udd1b\ud83c\udffd" );
		name( "left_facing_fist_tone1", R.drawable.emj_1f91b_1f3fb, "\ud83e\udd1b\ud83c\udffb" );
		name( "left_facing_fist_tone2", R.drawable.emj_1f91b_1f3fc, "\ud83e\udd1b\ud83c\udffc" );
		name( "left_facing_fist_tone3", R.drawable.emj_1f91b_1f3fd, "\ud83e\udd1b\ud83c\udffd" );
		name( "left_facing_fist_tone4", R.drawable.emj_1f91b_1f3fe, "\ud83e\udd1b\ud83c\udffe" );
		name( "left_facing_fist_tone5", R.drawable.emj_1f91b_1f3ff, "\ud83e\udd1b\ud83c\udfff" );
		name( "left_fist", R.drawable.emj_1f91b, "\ud83e\udd1b" );
		name( "left_fist_tone1", R.drawable.emj_1f91b_1f3fb, "\ud83e\udd1b\ud83c\udffb" );
		name( "left_fist_tone2", R.drawable.emj_1f91b_1f3fc, "\ud83e\udd1b\ud83c\udffc" );
		name( "left_fist_tone3", R.drawable.emj_1f91b_1f3fd, "\ud83e\udd1b\ud83c\udffd" );
		name( "left_fist_tone4", R.drawable.emj_1f91b_1f3fe, "\ud83e\udd1b\ud83c\udffe" );
		name( "left_fist_tone5", R.drawable.emj_1f91b_1f3ff, "\ud83e\udd1b\ud83c\udfff" );
		name( "left_luggage", R.drawable.emj_1f6c5, "\ud83d\udec5" );
		name( "left_right_arrow", R.drawable.emj_2194_fe0f, "\u2194\ufe0f" );
		name( "left_speech_bubble", R.drawable.emj_1f5e8_fe0f, "\ud83d\udde8\ufe0f" );
		name( "leftwards_arrow_with_hook", R.drawable.emj_21a9_fe0f, "\u21a9\ufe0f" );
		name( "lemon", R.drawable.emj_1f34b, "\ud83c\udf4b" );
		name( "leo", R.drawable.emj_264c, "\u264c" );
		name( "leopard", R.drawable.emj_1f406, "\ud83d\udc06" );
		name( "level_slider", R.drawable.emj_1f39a_fe0f, "\ud83c\udf9a\ufe0f" );
		name( "levitate", R.drawable.emj_1f574_fe0f, "\ud83d\udd74\ufe0f" );
		name( "li", R.drawable.emj_1f1f1_1f1ee, "\ud83c\uddf1\ud83c\uddee" );
		name( "liar", R.drawable.emj_1f925, "\ud83e\udd25" );
		name( "libra", R.drawable.emj_264e, "\u264e" );
		name( "lifter", R.drawable.emj_1f3cb_fe0f, "\ud83c\udfcb\ufe0f" );
		name( "lifter_tone1", R.drawable.emj_1f3cb_1f3fb, "\ud83c\udfcb\ud83c\udffb" );
		name( "lifter_tone2", R.drawable.emj_1f3cb_1f3fc, "\ud83c\udfcb\ud83c\udffc" );
		name( "lifter_tone3", R.drawable.emj_1f3cb_1f3fd, "\ud83c\udfcb\ud83c\udffd" );
		name( "lifter_tone4", R.drawable.emj_1f3cb_1f3fe, "\ud83c\udfcb\ud83c\udffe" );
		name( "lifter_tone5", R.drawable.emj_1f3cb_1f3ff, "\ud83c\udfcb\ud83c\udfff" );
		name( "light_rail", R.drawable.emj_1f688, "\ud83d\ude88" );
		name( "lightning", R.drawable.emj_1f329_fe0f, "\ud83c\udf29\ufe0f" );
		name( "lightning_cloud", R.drawable.emj_1f329_fe0f, "\ud83c\udf29\ufe0f" );
		name( "link", R.drawable.emj_1f517, "\ud83d\udd17" );
		name( "linked_paperclips", R.drawable.emj_1f587_fe0f, "\ud83d\udd87\ufe0f" );
		name( "lion", R.drawable.emj_1f981, "\ud83e\udd81" );
		name( "lion_face", R.drawable.emj_1f981, "\ud83e\udd81" );
		name( "lips", R.drawable.emj_1f444, "\ud83d\udc44" );
		name( "lipstick", R.drawable.emj_1f484, "\ud83d\udc84" );
		name( "lizard", R.drawable.emj_1f98e, "\ud83e\udd8e" );
		name( "lk", R.drawable.emj_1f1f1_1f1f0, "\ud83c\uddf1\ud83c\uddf0" );
		name( "lock", R.drawable.emj_1f512, "\ud83d\udd12" );
		name( "lock_with_ink_pen", R.drawable.emj_1f50f, "\ud83d\udd0f" );
		name( "lollipop", R.drawable.emj_1f36d, "\ud83c\udf6d" );
		name( "loop", R.drawable.emj_27bf, "\u27bf" );
		name( "loud_sound", R.drawable.emj_1f50a, "\ud83d\udd0a" );
		name( "loudspeaker", R.drawable.emj_1f4e2, "\ud83d\udce2" );
		name( "love_hotel", R.drawable.emj_1f3e9, "\ud83c\udfe9" );
		name( "love_letter", R.drawable.emj_1f48c, "\ud83d\udc8c" );
	}
	private static void init75(){
		name( "low_brightness", R.drawable.emj_1f505, "\ud83d\udd05" );
		name( "lower_left_ballpoint_pen", R.drawable.emj_1f58a_fe0f, "\ud83d\udd8a\ufe0f" );
		name( "lower_left_crayon", R.drawable.emj_1f58d_fe0f, "\ud83d\udd8d\ufe0f" );
		name( "lower_left_fountain_pen", R.drawable.emj_1f58b_fe0f, "\ud83d\udd8b\ufe0f" );
		name( "lower_left_paintbrush", R.drawable.emj_1f58c_fe0f, "\ud83d\udd8c\ufe0f" );
		name( "lr", R.drawable.emj_1f1f1_1f1f7, "\ud83c\uddf1\ud83c\uddf7" );
		name( "ls", R.drawable.emj_1f1f1_1f1f8, "\ud83c\uddf1\ud83c\uddf8" );
		name( "lt", R.drawable.emj_1f1f1_1f1f9, "\ud83c\uddf1\ud83c\uddf9" );
		name( "lu", R.drawable.emj_1f1f1_1f1fa, "\ud83c\uddf1\ud83c\uddfa" );
		name( "lv", R.drawable.emj_1f1f1_1f1fb, "\ud83c\uddf1\ud83c\uddfb" );
		name( "ly", R.drawable.emj_1f1f1_1f1fe, "\ud83c\uddf1\ud83c\uddfe" );
		name( "lying_face", R.drawable.emj_1f925, "\ud83e\udd25" );
		name( "m", R.drawable.emj_24c2_fe0f, "\u24c2\ufe0f" );
		name( "ma", R.drawable.emj_1f1f2_1f1e6, "\ud83c\uddf2\ud83c\udde6" );
		name( "mag", R.drawable.emj_1f50d, "\ud83d\udd0d" );
		name( "mag_right", R.drawable.emj_1f50e, "\ud83d\udd0e" );
		name( "mage", R.drawable.emj_1f9d9, "\ud83e\uddd9" );
		name( "mage_dark_skin_tone", R.drawable.emj_1f9d9_1f3ff, "\ud83e\uddd9\ud83c\udfff" );
		name( "mage_light_skin_tone", R.drawable.emj_1f9d9_1f3fb, "\ud83e\uddd9\ud83c\udffb" );
		name( "mage_medium_dark_skin_tone", R.drawable.emj_1f9d9_1f3fe, "\ud83e\uddd9\ud83c\udffe" );
		name( "mage_medium_light_skin_tone", R.drawable.emj_1f9d9_1f3fc, "\ud83e\uddd9\ud83c\udffc" );
		name( "mage_medium_skin_tone", R.drawable.emj_1f9d9_1f3fd, "\ud83e\uddd9\ud83c\udffd" );
		name( "mage_tone1", R.drawable.emj_1f9d9_1f3fb, "\ud83e\uddd9\ud83c\udffb" );
		name( "mage_tone2", R.drawable.emj_1f9d9_1f3fc, "\ud83e\uddd9\ud83c\udffc" );
		name( "mage_tone3", R.drawable.emj_1f9d9_1f3fd, "\ud83e\uddd9\ud83c\udffd" );
		name( "mage_tone4", R.drawable.emj_1f9d9_1f3fe, "\ud83e\uddd9\ud83c\udffe" );
		name( "mage_tone5", R.drawable.emj_1f9d9_1f3ff, "\ud83e\uddd9\ud83c\udfff" );
		name( "mahjong", R.drawable.emj_1f004, "\ud83c\udc04" );
		name( "mailbox", R.drawable.emj_1f4eb, "\ud83d\udceb" );
		name( "mailbox_closed", R.drawable.emj_1f4ea, "\ud83d\udcea" );
		name( "mailbox_with_mail", R.drawable.emj_1f4ec, "\ud83d\udcec" );
		name( "mailbox_with_no_mail", R.drawable.emj_1f4ed, "\ud83d\udced" );
		name( "male_artist", R.drawable.emj_1f468_200d_1f3a8, "\ud83d\udc68\u200d\ud83c\udfa8" );
		name( "male_artist_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3a8, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa8" );
		name( "male_artist_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3a8, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "male_artist_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3a8, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "male_artist_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3a8, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "male_artist_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3a8, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "male_artist_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3a8, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "male_artist_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3a8, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "male_artist_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3a8, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "male_artist_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3a8, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "male_artist_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3a8, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa8" );
		name( "male_astronaut", R.drawable.emj_1f468_200d_1f680, "\ud83d\udc68\u200d\ud83d\ude80" );
		name( "male_astronaut_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f680, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude80" );
		name( "male_astronaut_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f680, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude80" );
		name( "male_astronaut_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f680, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude80" );
		name( "male_astronaut_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f680, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude80" );
		name( "male_astronaut_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f680, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude80" );
		name( "male_astronaut_tone1", R.drawable.emj_1f468_1f3fb_200d_1f680, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude80" );
		name( "male_astronaut_tone2", R.drawable.emj_1f468_1f3fc_200d_1f680, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude80" );
		name( "male_astronaut_tone3", R.drawable.emj_1f468_1f3fd_200d_1f680, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude80" );
		name( "male_astronaut_tone4", R.drawable.emj_1f468_1f3fe_200d_1f680, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude80" );
		name( "male_astronaut_tone5", R.drawable.emj_1f468_1f3ff_200d_1f680, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude80" );
		name( "male_construction_worker", R.drawable.emj_1f477_200d_2642_fe0f, "\ud83d\udc77\u200d\u2642\ufe0f" );
		name( "male_construction_worker_dark_skin_tone", R.drawable.emj_1f477_1f3ff_200d_2642_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_construction_worker_light_skin_tone", R.drawable.emj_1f477_1f3fb_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_construction_worker_medium_dark_skin_tone", R.drawable.emj_1f477_1f3fe_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_construction_worker_medium_light_skin_tone", R.drawable.emj_1f477_1f3fc_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_construction_worker_medium_skin_tone", R.drawable.emj_1f477_1f3fd_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_construction_worker_tone1", R.drawable.emj_1f477_1f3fb_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_construction_worker_tone2", R.drawable.emj_1f477_1f3fc_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_construction_worker_tone3", R.drawable.emj_1f477_1f3fd_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_construction_worker_tone4", R.drawable.emj_1f477_1f3fe_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_construction_worker_tone5", R.drawable.emj_1f477_1f3ff_200d_2642_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_cook", R.drawable.emj_1f468_200d_1f373, "\ud83d\udc68\u200d\ud83c\udf73" );
		name( "male_cook_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f373, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf73" );
		name( "male_cook_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f373, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf73" );
		name( "male_cook_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f373, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf73" );
		name( "male_cook_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f373, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf73" );
		name( "male_cook_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f373, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf73" );
		name( "male_cook_tone1", R.drawable.emj_1f468_1f3fb_200d_1f373, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf73" );
		name( "male_cook_tone2", R.drawable.emj_1f468_1f3fc_200d_1f373, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf73" );
		name( "male_cook_tone3", R.drawable.emj_1f468_1f3fd_200d_1f373, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf73" );
		name( "male_cook_tone4", R.drawable.emj_1f468_1f3fe_200d_1f373, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf73" );
		name( "male_cook_tone5", R.drawable.emj_1f468_1f3ff_200d_1f373, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf73" );
		name( "male_dancer", R.drawable.emj_1f57a, "\ud83d\udd7a" );
		name( "male_dancer_tone1", R.drawable.emj_1f57a_1f3fb, "\ud83d\udd7a\ud83c\udffb" );
		name( "male_dancer_tone2", R.drawable.emj_1f57a_1f3fc, "\ud83d\udd7a\ud83c\udffc" );
		name( "male_dancer_tone3", R.drawable.emj_1f57a_1f3fd, "\ud83d\udd7a\ud83c\udffd" );
		name( "male_dancer_tone4", R.drawable.emj_1f57a_1f3fe, "\ud83d\udd7a\ud83c\udffe" );
		name( "male_dancer_tone5", R.drawable.emj_1f57a_1f3ff, "\ud83d\udd7a\ud83c\udfff" );
		name( "male_detective", R.drawable.emj_1f575_fe0f_200d_2642_fe0f, "\ud83d\udd75\ufe0f\u200d\u2642\ufe0f" );
		name( "male_detective_dark_skin_tone", R.drawable.emj_1f575_1f3ff_200d_2642_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_detective_light_skin_tone", R.drawable.emj_1f575_1f3fb_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_detective_medium_dark_skin_tone", R.drawable.emj_1f575_1f3fe_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_detective_medium_light_skin_tone", R.drawable.emj_1f575_1f3fc_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_detective_medium_skin_tone", R.drawable.emj_1f575_1f3fd_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_detective_tone1", R.drawable.emj_1f575_1f3fb_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_detective_tone2", R.drawable.emj_1f575_1f3fc_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_detective_tone3", R.drawable.emj_1f575_1f3fd_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_detective_tone4", R.drawable.emj_1f575_1f3fe_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_detective_tone5", R.drawable.emj_1f575_1f3ff_200d_2642_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_doctor", R.drawable.emj_1f468_200d_2695_fe0f, "\ud83d\udc68\u200d\u2695\ufe0f" );
		name( "male_doctor_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_2695_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "male_doctor_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "male_doctor_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "male_doctor_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2695\ufe0f" );
		name( "male_doctor_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "male_doctor_tone1", R.drawable.emj_1f468_1f3fb_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "male_doctor_tone2", R.drawable.emj_1f468_1f3fc_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2695\ufe0f" );
	}
	private static void init76(){
		name( "male_doctor_tone3", R.drawable.emj_1f468_1f3fd_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "male_doctor_tone4", R.drawable.emj_1f468_1f3fe_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "male_doctor_tone5", R.drawable.emj_1f468_1f3ff_200d_2695_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "male_elf", R.drawable.emj_1f9dd_200d_2642_fe0f, "\ud83e\udddd\u200d\u2642\ufe0f" );
		name( "male_elf_dark_skin_tone", R.drawable.emj_1f9dd_1f3ff_200d_2642_fe0f, "\ud83e\udddd\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_elf_light_skin_tone", R.drawable.emj_1f9dd_1f3fb_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_elf_medium_dark_skin_tone", R.drawable.emj_1f9dd_1f3fe_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_elf_medium_light_skin_tone", R.drawable.emj_1f9dd_1f3fc_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_elf_medium_skin_tone", R.drawable.emj_1f9dd_1f3fd_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_elf_tone1", R.drawable.emj_1f9dd_1f3fb_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_elf_tone2", R.drawable.emj_1f9dd_1f3fc_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_elf_tone3", R.drawable.emj_1f9dd_1f3fd_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_elf_tone4", R.drawable.emj_1f9dd_1f3fe_200d_2642_fe0f, "\ud83e\udddd\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_elf_tone5", R.drawable.emj_1f9dd_1f3ff_200d_2642_fe0f, "\ud83e\udddd\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_factory_worker", R.drawable.emj_1f468_200d_1f3ed, "\ud83d\udc68\u200d\ud83c\udfed" );
		name( "male_factory_worker_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3ed, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfed" );
		name( "male_factory_worker_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3ed, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfed" );
		name( "male_factory_worker_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3ed, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfed" );
		name( "male_factory_worker_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3ed, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfed" );
		name( "male_factory_worker_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3ed, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfed" );
		name( "male_factory_worker_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3ed, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfed" );
		name( "male_factory_worker_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3ed, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfed" );
		name( "male_factory_worker_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3ed, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfed" );
		name( "male_factory_worker_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3ed, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfed" );
		name( "male_factory_worker_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3ed, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfed" );
		name( "male_fairy", R.drawable.emj_1f9da_200d_2642_fe0f, "\ud83e\uddda\u200d\u2642\ufe0f" );
		name( "male_fairy_dark_skin_tone", R.drawable.emj_1f9da_1f3ff_200d_2642_fe0f, "\ud83e\uddda\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_fairy_light_skin_tone", R.drawable.emj_1f9da_1f3fb_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_fairy_medium_dark_skin_tone", R.drawable.emj_1f9da_1f3fe_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_fairy_medium_light_skin_tone", R.drawable.emj_1f9da_1f3fc_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_fairy_medium_skin_tone", R.drawable.emj_1f9da_1f3fd_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_fairy_tone1", R.drawable.emj_1f9da_1f3fb_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_fairy_tone2", R.drawable.emj_1f9da_1f3fc_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_fairy_tone3", R.drawable.emj_1f9da_1f3fd_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_fairy_tone4", R.drawable.emj_1f9da_1f3fe_200d_2642_fe0f, "\ud83e\uddda\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_fairy_tone5", R.drawable.emj_1f9da_1f3ff_200d_2642_fe0f, "\ud83e\uddda\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_farmer", R.drawable.emj_1f468_200d_1f33e, "\ud83d\udc68\u200d\ud83c\udf3e" );
		name( "male_farmer_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f33e, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "male_farmer_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f33e, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "male_farmer_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f33e, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf3e" );
		name( "male_farmer_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f33e, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "male_farmer_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f33e, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "male_farmer_tone1", R.drawable.emj_1f468_1f3fb_200d_1f33e, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "male_farmer_tone2", R.drawable.emj_1f468_1f3fc_200d_1f33e, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "male_farmer_tone3", R.drawable.emj_1f468_1f3fd_200d_1f33e, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "male_farmer_tone4", R.drawable.emj_1f468_1f3fe_200d_1f33e, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf3e" );
		name( "male_farmer_tone5", R.drawable.emj_1f468_1f3ff_200d_1f33e, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "male_firefighter", R.drawable.emj_1f468_200d_1f692, "\ud83d\udc68\u200d\ud83d\ude92" );
		name( "male_firefighter_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f692, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude92" );
		name( "male_firefighter_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f692, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude92" );
		name( "male_firefighter_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f692, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude92" );
		name( "male_firefighter_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f692, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude92" );
		name( "male_firefighter_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f692, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude92" );
		name( "male_firefighter_tone1", R.drawable.emj_1f468_1f3fb_200d_1f692, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude92" );
		name( "male_firefighter_tone2", R.drawable.emj_1f468_1f3fc_200d_1f692, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude92" );
		name( "male_firefighter_tone3", R.drawable.emj_1f468_1f3fd_200d_1f692, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude92" );
		name( "male_firefighter_tone4", R.drawable.emj_1f468_1f3fe_200d_1f692, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude92" );
		name( "male_firefighter_tone5", R.drawable.emj_1f468_1f3ff_200d_1f692, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude92" );
		name( "male_genie", R.drawable.emj_1f9de_200d_2642_fe0f, "\ud83e\uddde\u200d\u2642\ufe0f" );
		name( "male_guard", R.drawable.emj_1f482_200d_2642_fe0f, "\ud83d\udc82\u200d\u2642\ufe0f" );
		name( "male_guard_dark_skin_tone", R.drawable.emj_1f482_1f3ff_200d_2642_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_guard_light_skin_tone", R.drawable.emj_1f482_1f3fb_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_guard_medium_dark_skin_tone", R.drawable.emj_1f482_1f3fe_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_guard_medium_light_skin_tone", R.drawable.emj_1f482_1f3fc_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_guard_medium_skin_tone", R.drawable.emj_1f482_1f3fd_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_guard_tone1", R.drawable.emj_1f482_1f3fb_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_guard_tone2", R.drawable.emj_1f482_1f3fc_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_guard_tone3", R.drawable.emj_1f482_1f3fd_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_guard_tone4", R.drawable.emj_1f482_1f3fe_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_guard_tone5", R.drawable.emj_1f482_1f3ff_200d_2642_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_judge", R.drawable.emj_1f468_200d_2696_fe0f, "\ud83d\udc68\u200d\u2696\ufe0f" );
		name( "male_judge_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_2696_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "male_judge_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "male_judge_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "male_judge_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "male_judge_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "male_judge_tone1", R.drawable.emj_1f468_1f3fb_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "male_judge_tone2", R.drawable.emj_1f468_1f3fc_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "male_judge_tone3", R.drawable.emj_1f468_1f3fd_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "male_judge_tone4", R.drawable.emj_1f468_1f3fe_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "male_judge_tone5", R.drawable.emj_1f468_1f3ff_200d_2696_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "male_mage", R.drawable.emj_1f9d9_200d_2642_fe0f, "\ud83e\uddd9\u200d\u2642\ufe0f" );
		name( "male_mage_dark_skin_tone", R.drawable.emj_1f9d9_1f3ff_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_mage_light_skin_tone", R.drawable.emj_1f9d9_1f3fb_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_mage_medium_dark_skin_tone", R.drawable.emj_1f9d9_1f3fe_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_mage_medium_light_skin_tone", R.drawable.emj_1f9d9_1f3fc_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_mage_medium_skin_tone", R.drawable.emj_1f9d9_1f3fd_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_mage_tone1", R.drawable.emj_1f9d9_1f3fb_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_mage_tone2", R.drawable.emj_1f9d9_1f3fc_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_mage_tone3", R.drawable.emj_1f9d9_1f3fd_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_mage_tone4", R.drawable.emj_1f9d9_1f3fe_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_mage_tone5", R.drawable.emj_1f9d9_1f3ff_200d_2642_fe0f, "\ud83e\uddd9\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_mechanic", R.drawable.emj_1f468_200d_1f527, "\ud83d\udc68\u200d\ud83d\udd27" );
		name( "male_mechanic_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f527, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd27" );
		name( "male_mechanic_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f527, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd27" );
		name( "male_mechanic_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f527, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd27" );
		name( "male_mechanic_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f527, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd27" );
		name( "male_mechanic_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f527, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd27" );
		name( "male_mechanic_tone1", R.drawable.emj_1f468_1f3fb_200d_1f527, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd27" );
		name( "male_mechanic_tone2", R.drawable.emj_1f468_1f3fc_200d_1f527, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd27" );
		name( "male_mechanic_tone3", R.drawable.emj_1f468_1f3fd_200d_1f527, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd27" );
	}
	private static void init77(){
		name( "male_mechanic_tone4", R.drawable.emj_1f468_1f3fe_200d_1f527, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd27" );
		name( "male_mechanic_tone5", R.drawable.emj_1f468_1f3ff_200d_1f527, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd27" );
		name( "male_office_worker", R.drawable.emj_1f468_200d_1f4bc, "\ud83d\udc68\u200d\ud83d\udcbc" );
		name( "male_office_worker_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f4bc, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "male_office_worker_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f4bc, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "male_office_worker_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f4bc, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "male_office_worker_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f4bc, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "male_office_worker_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f4bc, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "male_office_worker_tone1", R.drawable.emj_1f468_1f3fb_200d_1f4bc, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "male_office_worker_tone2", R.drawable.emj_1f468_1f3fc_200d_1f4bc, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "male_office_worker_tone3", R.drawable.emj_1f468_1f3fd_200d_1f4bc, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "male_office_worker_tone4", R.drawable.emj_1f468_1f3fe_200d_1f4bc, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "male_office_worker_tone5", R.drawable.emj_1f468_1f3ff_200d_1f4bc, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "male_pilot", R.drawable.emj_1f468_200d_2708_fe0f, "\ud83d\udc68\u200d\u2708\ufe0f" );
		name( "male_pilot_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_2708_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "male_pilot_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "male_pilot_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "male_pilot_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "male_pilot_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2708\ufe0f" );
		name( "male_pilot_tone1", R.drawable.emj_1f468_1f3fb_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "male_pilot_tone2", R.drawable.emj_1f468_1f3fc_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "male_pilot_tone3", R.drawable.emj_1f468_1f3fd_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2708\ufe0f" );
		name( "male_pilot_tone4", R.drawable.emj_1f468_1f3fe_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "male_pilot_tone5", R.drawable.emj_1f468_1f3ff_200d_2708_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "male_police_officer", R.drawable.emj_1f46e_200d_2642_fe0f, "\ud83d\udc6e\u200d\u2642\ufe0f" );
		name( "male_police_officer_dark_skin_tone", R.drawable.emj_1f46e_1f3ff_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_police_officer_light_skin_tone", R.drawable.emj_1f46e_1f3fb_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_police_officer_medium_dark_skin_tone", R.drawable.emj_1f46e_1f3fe_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_police_officer_medium_light_skin_tone", R.drawable.emj_1f46e_1f3fc_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_police_officer_medium_skin_tone", R.drawable.emj_1f46e_1f3fd_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_police_officer_tone1", R.drawable.emj_1f46e_1f3fb_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_police_officer_tone2", R.drawable.emj_1f46e_1f3fc_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_police_officer_tone3", R.drawable.emj_1f46e_1f3fd_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_police_officer_tone4", R.drawable.emj_1f46e_1f3fe_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_police_officer_tone5", R.drawable.emj_1f46e_1f3ff_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_scientist", R.drawable.emj_1f468_200d_1f52c, "\ud83d\udc68\u200d\ud83d\udd2c" );
		name( "male_scientist_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f52c, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "male_scientist_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f52c, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "male_scientist_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f52c, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "male_scientist_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f52c, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd2c" );
		name( "male_scientist_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f52c, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "male_scientist_tone1", R.drawable.emj_1f468_1f3fb_200d_1f52c, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "male_scientist_tone2", R.drawable.emj_1f468_1f3fc_200d_1f52c, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd2c" );
		name( "male_scientist_tone3", R.drawable.emj_1f468_1f3fd_200d_1f52c, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "male_scientist_tone4", R.drawable.emj_1f468_1f3fe_200d_1f52c, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "male_scientist_tone5", R.drawable.emj_1f468_1f3ff_200d_1f52c, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "male_sign", R.drawable.emj_2642_fe0f, "\u2642\ufe0f" );
		name( "male_singer", R.drawable.emj_1f468_200d_1f3a4, "\ud83d\udc68\u200d\ud83c\udfa4" );
		name( "male_singer_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3a4, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "male_singer_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3a4, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "male_singer_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3a4, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "male_singer_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3a4, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "male_singer_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3a4, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "male_singer_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3a4, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "male_singer_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3a4, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "male_singer_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3a4, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "male_singer_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3a4, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "male_singer_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3a4, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "male_student", R.drawable.emj_1f468_200d_1f393, "\ud83d\udc68\u200d\ud83c\udf93" );
		name( "male_student_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f393, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf93" );
		name( "male_student_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f393, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf93" );
		name( "male_student_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f393, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf93" );
		name( "male_student_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f393, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf93" );
		name( "male_student_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f393, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf93" );
		name( "male_student_tone1", R.drawable.emj_1f468_1f3fb_200d_1f393, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf93" );
		name( "male_student_tone2", R.drawable.emj_1f468_1f3fc_200d_1f393, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf93" );
		name( "male_student_tone3", R.drawable.emj_1f468_1f3fd_200d_1f393, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf93" );
		name( "male_student_tone4", R.drawable.emj_1f468_1f3fe_200d_1f393, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf93" );
		name( "male_student_tone5", R.drawable.emj_1f468_1f3ff_200d_1f393, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf93" );
		name( "male_teacher", R.drawable.emj_1f468_200d_1f3eb, "\ud83d\udc68\u200d\ud83c\udfeb" );
		name( "male_teacher_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3eb, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "male_teacher_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3eb, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "male_teacher_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3eb, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "male_teacher_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3eb, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "male_teacher_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3eb, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "male_teacher_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3eb, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "male_teacher_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3eb, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "male_teacher_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3eb, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "male_teacher_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3eb, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "male_teacher_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3eb, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "male_technologist", R.drawable.emj_1f468_200d_1f4bb, "\ud83d\udc68\u200d\ud83d\udcbb" );
		name( "male_technologist_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f4bb, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "male_technologist_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f4bb, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "male_technologist_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f4bb, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "male_technologist_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f4bb, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "male_technologist_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f4bb, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "male_technologist_tone1", R.drawable.emj_1f468_1f3fb_200d_1f4bb, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "male_technologist_tone2", R.drawable.emj_1f468_1f3fc_200d_1f4bb, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "male_technologist_tone3", R.drawable.emj_1f468_1f3fd_200d_1f4bb, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "male_technologist_tone4", R.drawable.emj_1f468_1f3fe_200d_1f4bb, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "male_technologist_tone5", R.drawable.emj_1f468_1f3ff_200d_1f4bb, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "male_vampire", R.drawable.emj_1f9db_200d_2642_fe0f, "\ud83e\udddb\u200d\u2642\ufe0f" );
		name( "male_vampire_dark_skin_tone", R.drawable.emj_1f9db_1f3ff_200d_2642_fe0f, "\ud83e\udddb\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_vampire_light_skin_tone", R.drawable.emj_1f9db_1f3fb_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_vampire_medium_dark_skin_tone", R.drawable.emj_1f9db_1f3fe_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "male_vampire_medium_light_skin_tone", R.drawable.emj_1f9db_1f3fc_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_vampire_medium_skin_tone", R.drawable.emj_1f9db_1f3fd_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_vampire_tone1", R.drawable.emj_1f9db_1f3fb_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "male_vampire_tone2", R.drawable.emj_1f9db_1f3fc_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "male_vampire_tone3", R.drawable.emj_1f9db_1f3fd_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "male_vampire_tone4", R.drawable.emj_1f9db_1f3fe_200d_2642_fe0f, "\ud83e\udddb\ud83c\udffe\u200d\u2642\ufe0f" );
	}
	private static void init78(){
		name( "male_vampire_tone5", R.drawable.emj_1f9db_1f3ff_200d_2642_fe0f, "\ud83e\udddb\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "male_zombie", R.drawable.emj_1f9df_200d_2642_fe0f, "\ud83e\udddf\u200d\u2642\ufe0f" );
		name( "man", R.drawable.emj_1f468, "\ud83d\udc68" );
		name( "man_and_woman_holding_hands", R.drawable.emj_1f46b, "\ud83d\udc6b" );
		name( "man_artist", R.drawable.emj_1f468_200d_1f3a8, "\ud83d\udc68\u200d\ud83c\udfa8" );
		name( "man_artist_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3a8, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa8" );
		name( "man_artist_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3a8, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "man_artist_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3a8, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "man_artist_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3a8, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "man_artist_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3a8, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "man_artist_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3a8, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "man_artist_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3a8, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "man_artist_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3a8, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "man_artist_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3a8, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "man_artist_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3a8, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa8" );
		name( "man_astronaut", R.drawable.emj_1f468_200d_1f680, "\ud83d\udc68\u200d\ud83d\ude80" );
		name( "man_astronaut_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f680, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude80" );
		name( "man_astronaut_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f680, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude80" );
		name( "man_astronaut_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f680, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude80" );
		name( "man_astronaut_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f680, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude80" );
		name( "man_astronaut_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f680, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude80" );
		name( "man_astronaut_tone1", R.drawable.emj_1f468_1f3fb_200d_1f680, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude80" );
		name( "man_astronaut_tone2", R.drawable.emj_1f468_1f3fc_200d_1f680, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude80" );
		name( "man_astronaut_tone3", R.drawable.emj_1f468_1f3fd_200d_1f680, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude80" );
		name( "man_astronaut_tone4", R.drawable.emj_1f468_1f3fe_200d_1f680, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude80" );
		name( "man_astronaut_tone5", R.drawable.emj_1f468_1f3ff_200d_1f680, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude80" );
		name( "man_biking", R.drawable.emj_1f6b4_200d_2642_fe0f, "\ud83d\udeb4\u200d\u2642\ufe0f" );
		name( "man_biking_dark_skin_tone", R.drawable.emj_1f6b4_1f3ff_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_biking_light_skin_tone", R.drawable.emj_1f6b4_1f3fb_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_biking_medium_dark_skin_tone", R.drawable.emj_1f6b4_1f3fe_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_biking_medium_light_skin_tone", R.drawable.emj_1f6b4_1f3fc_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_biking_medium_skin_tone", R.drawable.emj_1f6b4_1f3fd_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_biking_tone1", R.drawable.emj_1f6b4_1f3fb_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_biking_tone2", R.drawable.emj_1f6b4_1f3fc_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_biking_tone3", R.drawable.emj_1f6b4_1f3fd_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_biking_tone4", R.drawable.emj_1f6b4_1f3fe_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_biking_tone5", R.drawable.emj_1f6b4_1f3ff_200d_2642_fe0f, "\ud83d\udeb4\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball", R.drawable.emj_26f9_fe0f_200d_2642_fe0f, "\u26f9\ufe0f\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_dark_skin_tone", R.drawable.emj_26f9_1f3ff_200d_2642_fe0f, "\u26f9\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_light_skin_tone", R.drawable.emj_26f9_1f3fb_200d_2642_fe0f, "\u26f9\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_medium_dark_skin_tone", R.drawable.emj_26f9_1f3fe_200d_2642_fe0f, "\u26f9\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_medium_light_skin_tone", R.drawable.emj_26f9_1f3fc_200d_2642_fe0f, "\u26f9\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_medium_skin_tone", R.drawable.emj_26f9_1f3fd_200d_2642_fe0f, "\u26f9\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_tone1", R.drawable.emj_26f9_1f3fb_200d_2642_fe0f, "\u26f9\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_tone2", R.drawable.emj_26f9_1f3fc_200d_2642_fe0f, "\u26f9\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_tone3", R.drawable.emj_26f9_1f3fd_200d_2642_fe0f, "\u26f9\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_tone4", R.drawable.emj_26f9_1f3fe_200d_2642_fe0f, "\u26f9\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_bouncing_ball_tone5", R.drawable.emj_26f9_1f3ff_200d_2642_fe0f, "\u26f9\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_bowing", R.drawable.emj_1f647_200d_2642_fe0f, "\ud83d\ude47\u200d\u2642\ufe0f" );
		name( "man_bowing_dark_skin_tone", R.drawable.emj_1f647_1f3ff_200d_2642_fe0f, "\ud83d\ude47\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_bowing_light_skin_tone", R.drawable.emj_1f647_1f3fb_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_bowing_medium_dark_skin_tone", R.drawable.emj_1f647_1f3fe_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_bowing_medium_light_skin_tone", R.drawable.emj_1f647_1f3fc_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_bowing_medium_skin_tone", R.drawable.emj_1f647_1f3fd_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_bowing_tone1", R.drawable.emj_1f647_1f3fb_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_bowing_tone2", R.drawable.emj_1f647_1f3fc_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_bowing_tone3", R.drawable.emj_1f647_1f3fd_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_bowing_tone4", R.drawable.emj_1f647_1f3fe_200d_2642_fe0f, "\ud83d\ude47\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_bowing_tone5", R.drawable.emj_1f647_1f3ff_200d_2642_fe0f, "\ud83d\ude47\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_boy", R.drawable.emj_1f468_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc66" );
		name( "man_boy_boy", R.drawable.emj_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "man_cartwheeling", R.drawable.emj_1f938_200d_2642_fe0f, "\ud83e\udd38\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_dark_skin_tone", R.drawable.emj_1f938_1f3ff_200d_2642_fe0f, "\ud83e\udd38\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_light_skin_tone", R.drawable.emj_1f938_1f3fb_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_medium_dark_skin_tone", R.drawable.emj_1f938_1f3fe_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_medium_light_skin_tone", R.drawable.emj_1f938_1f3fc_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_medium_skin_tone", R.drawable.emj_1f938_1f3fd_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_tone1", R.drawable.emj_1f938_1f3fb_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_tone2", R.drawable.emj_1f938_1f3fc_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_tone3", R.drawable.emj_1f938_1f3fd_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_tone4", R.drawable.emj_1f938_1f3fe_200d_2642_fe0f, "\ud83e\udd38\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_cartwheeling_tone5", R.drawable.emj_1f938_1f3ff_200d_2642_fe0f, "\ud83e\udd38\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_climbing", R.drawable.emj_1f9d7_200d_2642_fe0f, "\ud83e\uddd7\u200d\u2642\ufe0f" );
		name( "man_climbing_dark_skin_tone", R.drawable.emj_1f9d7_1f3ff_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_climbing_light_skin_tone", R.drawable.emj_1f9d7_1f3fb_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_climbing_medium_dark_skin_tone", R.drawable.emj_1f9d7_1f3fe_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_climbing_medium_light_skin_tone", R.drawable.emj_1f9d7_1f3fc_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_climbing_medium_skin_tone", R.drawable.emj_1f9d7_1f3fd_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_climbing_tone1", R.drawable.emj_1f9d7_1f3fb_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_climbing_tone2", R.drawable.emj_1f9d7_1f3fc_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_climbing_tone3", R.drawable.emj_1f9d7_1f3fd_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_climbing_tone4", R.drawable.emj_1f9d7_1f3fe_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_climbing_tone5", R.drawable.emj_1f9d7_1f3ff_200d_2642_fe0f, "\ud83e\uddd7\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_construction_worker", R.drawable.emj_1f477_200d_2642_fe0f, "\ud83d\udc77\u200d\u2642\ufe0f" );
		name( "man_construction_worker_dark_skin_tone", R.drawable.emj_1f477_1f3ff_200d_2642_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_construction_worker_light_skin_tone", R.drawable.emj_1f477_1f3fb_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_construction_worker_medium_dark_skin_tone", R.drawable.emj_1f477_1f3fe_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_construction_worker_medium_light_skin_tone", R.drawable.emj_1f477_1f3fc_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_construction_worker_medium_skin_tone", R.drawable.emj_1f477_1f3fd_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_construction_worker_tone1", R.drawable.emj_1f477_1f3fb_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_construction_worker_tone2", R.drawable.emj_1f477_1f3fc_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_construction_worker_tone3", R.drawable.emj_1f477_1f3fd_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_construction_worker_tone4", R.drawable.emj_1f477_1f3fe_200d_2642_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_construction_worker_tone5", R.drawable.emj_1f477_1f3ff_200d_2642_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_cook", R.drawable.emj_1f468_200d_1f373, "\ud83d\udc68\u200d\ud83c\udf73" );
		name( "man_cook_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f373, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf73" );
		name( "man_cook_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f373, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf73" );
		name( "man_cook_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f373, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf73" );
		name( "man_cook_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f373, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf73" );
		name( "man_cook_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f373, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf73" );
		name( "man_cook_tone1", R.drawable.emj_1f468_1f3fb_200d_1f373, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf73" );
	}
	private static void init79(){
		name( "man_cook_tone2", R.drawable.emj_1f468_1f3fc_200d_1f373, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf73" );
		name( "man_cook_tone3", R.drawable.emj_1f468_1f3fd_200d_1f373, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf73" );
		name( "man_cook_tone4", R.drawable.emj_1f468_1f3fe_200d_1f373, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf73" );
		name( "man_cook_tone5", R.drawable.emj_1f468_1f3ff_200d_1f373, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf73" );
		name( "man_dancing", R.drawable.emj_1f57a, "\ud83d\udd7a" );
		name( "man_dancing_dark_skin_tone", R.drawable.emj_1f57a_1f3ff, "\ud83d\udd7a\ud83c\udfff" );
		name( "man_dancing_light_skin_tone", R.drawable.emj_1f57a_1f3fb, "\ud83d\udd7a\ud83c\udffb" );
		name( "man_dancing_medium_dark_skin_tone", R.drawable.emj_1f57a_1f3fe, "\ud83d\udd7a\ud83c\udffe" );
		name( "man_dancing_medium_light_skin_tone", R.drawable.emj_1f57a_1f3fc, "\ud83d\udd7a\ud83c\udffc" );
		name( "man_dancing_medium_skin_tone", R.drawable.emj_1f57a_1f3fd, "\ud83d\udd7a\ud83c\udffd" );
		name( "man_dancing_tone1", R.drawable.emj_1f57a_1f3fb, "\ud83d\udd7a\ud83c\udffb" );
		name( "man_dancing_tone2", R.drawable.emj_1f57a_1f3fc, "\ud83d\udd7a\ud83c\udffc" );
		name( "man_dancing_tone3", R.drawable.emj_1f57a_1f3fd, "\ud83d\udd7a\ud83c\udffd" );
		name( "man_dancing_tone4", R.drawable.emj_1f57a_1f3fe, "\ud83d\udd7a\ud83c\udffe" );
		name( "man_dancing_tone5", R.drawable.emj_1f57a_1f3ff, "\ud83d\udd7a\ud83c\udfff" );
		name( "man_dark_skin_tone", R.drawable.emj_1f468_1f3ff, "\ud83d\udc68\ud83c\udfff" );
		name( "man_detective", R.drawable.emj_1f575_fe0f_200d_2642_fe0f, "\ud83d\udd75\ufe0f\u200d\u2642\ufe0f" );
		name( "man_detective_dark_skin_tone", R.drawable.emj_1f575_1f3ff_200d_2642_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_detective_light_skin_tone", R.drawable.emj_1f575_1f3fb_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_detective_medium_dark_skin_tone", R.drawable.emj_1f575_1f3fe_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_detective_medium_light_skin_tone", R.drawable.emj_1f575_1f3fc_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_detective_medium_skin_tone", R.drawable.emj_1f575_1f3fd_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_detective_tone1", R.drawable.emj_1f575_1f3fb_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_detective_tone2", R.drawable.emj_1f575_1f3fc_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_detective_tone3", R.drawable.emj_1f575_1f3fd_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_detective_tone4", R.drawable.emj_1f575_1f3fe_200d_2642_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_detective_tone5", R.drawable.emj_1f575_1f3ff_200d_2642_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_facepalming", R.drawable.emj_1f926_200d_2642_fe0f, "\ud83e\udd26\u200d\u2642\ufe0f" );
		name( "man_facepalming_dark_skin_tone", R.drawable.emj_1f926_1f3ff_200d_2642_fe0f, "\ud83e\udd26\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_facepalming_light_skin_tone", R.drawable.emj_1f926_1f3fb_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_facepalming_medium_dark_skin_tone", R.drawable.emj_1f926_1f3fe_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_facepalming_medium_light_skin_tone", R.drawable.emj_1f926_1f3fc_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_facepalming_medium_skin_tone", R.drawable.emj_1f926_1f3fd_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_facepalming_tone1", R.drawable.emj_1f926_1f3fb_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_facepalming_tone2", R.drawable.emj_1f926_1f3fc_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_facepalming_tone3", R.drawable.emj_1f926_1f3fd_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_facepalming_tone4", R.drawable.emj_1f926_1f3fe_200d_2642_fe0f, "\ud83e\udd26\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_facepalming_tone5", R.drawable.emj_1f926_1f3ff_200d_2642_fe0f, "\ud83e\udd26\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_factory_worker", R.drawable.emj_1f468_200d_1f3ed, "\ud83d\udc68\u200d\ud83c\udfed" );
		name( "man_factory_worker_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3ed, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfed" );
		name( "man_factory_worker_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3ed, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfed" );
		name( "man_factory_worker_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3ed, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfed" );
		name( "man_factory_worker_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3ed, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfed" );
		name( "man_factory_worker_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3ed, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfed" );
		name( "man_factory_worker_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3ed, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfed" );
		name( "man_factory_worker_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3ed, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfed" );
		name( "man_factory_worker_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3ed, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfed" );
		name( "man_factory_worker_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3ed, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfed" );
		name( "man_factory_worker_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3ed, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfed" );
		name( "man_farmer", R.drawable.emj_1f468_200d_1f33e, "\ud83d\udc68\u200d\ud83c\udf3e" );
		name( "man_farmer_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f33e, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "man_farmer_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f33e, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "man_farmer_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f33e, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf3e" );
		name( "man_farmer_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f33e, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "man_farmer_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f33e, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "man_farmer_tone1", R.drawable.emj_1f468_1f3fb_200d_1f33e, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "man_farmer_tone2", R.drawable.emj_1f468_1f3fc_200d_1f33e, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "man_farmer_tone3", R.drawable.emj_1f468_1f3fd_200d_1f33e, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "man_farmer_tone4", R.drawable.emj_1f468_1f3fe_200d_1f33e, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf3e" );
		name( "man_farmer_tone5", R.drawable.emj_1f468_1f3ff_200d_1f33e, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "man_firefighter", R.drawable.emj_1f468_200d_1f692, "\ud83d\udc68\u200d\ud83d\ude92" );
		name( "man_firefighter_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f692, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude92" );
		name( "man_firefighter_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f692, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude92" );
		name( "man_firefighter_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f692, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude92" );
		name( "man_firefighter_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f692, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude92" );
		name( "man_firefighter_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f692, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude92" );
		name( "man_firefighter_tone1", R.drawable.emj_1f468_1f3fb_200d_1f692, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\ude92" );
		name( "man_firefighter_tone2", R.drawable.emj_1f468_1f3fc_200d_1f692, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\ude92" );
		name( "man_firefighter_tone3", R.drawable.emj_1f468_1f3fd_200d_1f692, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\ude92" );
		name( "man_firefighter_tone4", R.drawable.emj_1f468_1f3fe_200d_1f692, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\ude92" );
		name( "man_firefighter_tone5", R.drawable.emj_1f468_1f3ff_200d_1f692, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\ude92" );
		name( "man_frowning", R.drawable.emj_1f64d_200d_2642_fe0f, "\ud83d\ude4d\u200d\u2642\ufe0f" );
		name( "man_frowning_dark_skin_tone", R.drawable.emj_1f64d_1f3ff_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_frowning_light_skin_tone", R.drawable.emj_1f64d_1f3fb_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_frowning_medium_dark_skin_tone", R.drawable.emj_1f64d_1f3fe_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_frowning_medium_light_skin_tone", R.drawable.emj_1f64d_1f3fc_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_frowning_medium_skin_tone", R.drawable.emj_1f64d_1f3fd_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_frowning_tone1", R.drawable.emj_1f64d_1f3fb_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_frowning_tone2", R.drawable.emj_1f64d_1f3fc_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_frowning_tone3", R.drawable.emj_1f64d_1f3fd_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_frowning_tone4", R.drawable.emj_1f64d_1f3fe_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_frowning_tone5", R.drawable.emj_1f64d_1f3ff_200d_2642_fe0f, "\ud83d\ude4d\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_gesturing_no", R.drawable.emj_1f645_200d_2642_fe0f, "\ud83d\ude45\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_dark_skin_tone", R.drawable.emj_1f645_1f3ff_200d_2642_fe0f, "\ud83d\ude45\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_light_skin_tone", R.drawable.emj_1f645_1f3fb_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_medium_dark_skin_tone", R.drawable.emj_1f645_1f3fe_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_medium_light_skin_tone", R.drawable.emj_1f645_1f3fc_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_medium_skin_tone", R.drawable.emj_1f645_1f3fd_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_tone1", R.drawable.emj_1f645_1f3fb_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_tone2", R.drawable.emj_1f645_1f3fc_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_tone3", R.drawable.emj_1f645_1f3fd_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_tone4", R.drawable.emj_1f645_1f3fe_200d_2642_fe0f, "\ud83d\ude45\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_gesturing_no_tone5", R.drawable.emj_1f645_1f3ff_200d_2642_fe0f, "\ud83d\ude45\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok", R.drawable.emj_1f646_200d_2642_fe0f, "\ud83d\ude46\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_dark_skin_tone", R.drawable.emj_1f646_1f3ff_200d_2642_fe0f, "\ud83d\ude46\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_light_skin_tone", R.drawable.emj_1f646_1f3fb_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_medium_dark_skin_tone", R.drawable.emj_1f646_1f3fe_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_medium_light_skin_tone", R.drawable.emj_1f646_1f3fc_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_medium_skin_tone", R.drawable.emj_1f646_1f3fd_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_tone1", R.drawable.emj_1f646_1f3fb_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_tone2", R.drawable.emj_1f646_1f3fc_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffc\u200d\u2642\ufe0f" );
	}
	private static void init80(){
		name( "man_gesturing_ok_tone3", R.drawable.emj_1f646_1f3fd_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_tone4", R.drawable.emj_1f646_1f3fe_200d_2642_fe0f, "\ud83d\ude46\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_gesturing_ok_tone5", R.drawable.emj_1f646_1f3ff_200d_2642_fe0f, "\ud83d\ude46\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage", R.drawable.emj_1f486_200d_2642_fe0f, "\ud83d\udc86\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_dark_skin_tone", R.drawable.emj_1f486_1f3ff_200d_2642_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_light_skin_tone", R.drawable.emj_1f486_1f3fb_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_medium_dark_skin_tone", R.drawable.emj_1f486_1f3fe_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_medium_light_skin_tone", R.drawable.emj_1f486_1f3fc_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_medium_skin_tone", R.drawable.emj_1f486_1f3fd_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_tone1", R.drawable.emj_1f486_1f3fb_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_tone2", R.drawable.emj_1f486_1f3fc_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_tone3", R.drawable.emj_1f486_1f3fd_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_tone4", R.drawable.emj_1f486_1f3fe_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_getting_face_massage_tone5", R.drawable.emj_1f486_1f3ff_200d_2642_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_getting_haircut", R.drawable.emj_1f487_200d_2642_fe0f, "\ud83d\udc87\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_dark_skin_tone", R.drawable.emj_1f487_1f3ff_200d_2642_fe0f, "\ud83d\udc87\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_light_skin_tone", R.drawable.emj_1f487_1f3fb_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_medium_dark_skin_tone", R.drawable.emj_1f487_1f3fe_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_medium_light_skin_tone", R.drawable.emj_1f487_1f3fc_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_medium_skin_tone", R.drawable.emj_1f487_1f3fd_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_tone1", R.drawable.emj_1f487_1f3fb_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_tone2", R.drawable.emj_1f487_1f3fc_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_tone3", R.drawable.emj_1f487_1f3fd_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_tone4", R.drawable.emj_1f487_1f3fe_200d_2642_fe0f, "\ud83d\udc87\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_getting_haircut_tone5", R.drawable.emj_1f487_1f3ff_200d_2642_fe0f, "\ud83d\udc87\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_getting_massage", R.drawable.emj_1f486_200d_2642_fe0f, "\ud83d\udc86\u200d\u2642\ufe0f" );
		name( "man_getting_massage_dark_skin_tone", R.drawable.emj_1f486_1f3ff_200d_2642_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_getting_massage_light_skin_tone", R.drawable.emj_1f486_1f3fb_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_getting_massage_medium_dark_skin_tone", R.drawable.emj_1f486_1f3fe_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_getting_massage_medium_light_skin_tone", R.drawable.emj_1f486_1f3fc_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_getting_massage_medium_skin_tone", R.drawable.emj_1f486_1f3fd_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_getting_massage_tone1", R.drawable.emj_1f486_1f3fb_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_getting_massage_tone2", R.drawable.emj_1f486_1f3fc_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_getting_massage_tone3", R.drawable.emj_1f486_1f3fd_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_getting_massage_tone4", R.drawable.emj_1f486_1f3fe_200d_2642_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_getting_massage_tone5", R.drawable.emj_1f486_1f3ff_200d_2642_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_girl", R.drawable.emj_1f468_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc67" );
		name( "man_girl_boy", R.drawable.emj_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "man_girl_girl", R.drawable.emj_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "man_golfing", R.drawable.emj_1f3cc_fe0f_200d_2642_fe0f, "\ud83c\udfcc\ufe0f\u200d\u2642\ufe0f" );
		name( "man_golfing_dark_skin_tone", R.drawable.emj_1f3cc_1f3ff_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_golfing_light_skin_tone", R.drawable.emj_1f3cc_1f3fb_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_golfing_medium_dark_skin_tone", R.drawable.emj_1f3cc_1f3fe_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_golfing_medium_light_skin_tone", R.drawable.emj_1f3cc_1f3fc_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_golfing_medium_skin_tone", R.drawable.emj_1f3cc_1f3fd_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_golfing_tone1", R.drawable.emj_1f3cc_1f3fb_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_golfing_tone2", R.drawable.emj_1f3cc_1f3fc_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_golfing_tone3", R.drawable.emj_1f3cc_1f3fd_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_golfing_tone4", R.drawable.emj_1f3cc_1f3fe_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_golfing_tone5", R.drawable.emj_1f3cc_1f3ff_200d_2642_fe0f, "\ud83c\udfcc\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_guard", R.drawable.emj_1f482_200d_2642_fe0f, "\ud83d\udc82\u200d\u2642\ufe0f" );
		name( "man_guard_dark_skin_tone", R.drawable.emj_1f482_1f3ff_200d_2642_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_guard_light_skin_tone", R.drawable.emj_1f482_1f3fb_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_guard_medium_dark_skin_tone", R.drawable.emj_1f482_1f3fe_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_guard_medium_light_skin_tone", R.drawable.emj_1f482_1f3fc_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_guard_medium_skin_tone", R.drawable.emj_1f482_1f3fd_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_guard_tone1", R.drawable.emj_1f482_1f3fb_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_guard_tone2", R.drawable.emj_1f482_1f3fc_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_guard_tone3", R.drawable.emj_1f482_1f3fd_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_guard_tone4", R.drawable.emj_1f482_1f3fe_200d_2642_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_guard_tone5", R.drawable.emj_1f482_1f3ff_200d_2642_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_health_worker", R.drawable.emj_1f468_200d_2695_fe0f, "\ud83d\udc68\u200d\u2695\ufe0f" );
		name( "man_health_worker_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_2695_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "man_health_worker_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "man_health_worker_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "man_health_worker_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2695\ufe0f" );
		name( "man_health_worker_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "man_health_worker_tone1", R.drawable.emj_1f468_1f3fb_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "man_health_worker_tone2", R.drawable.emj_1f468_1f3fc_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2695\ufe0f" );
		name( "man_health_worker_tone3", R.drawable.emj_1f468_1f3fd_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "man_health_worker_tone4", R.drawable.emj_1f468_1f3fe_200d_2695_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "man_health_worker_tone5", R.drawable.emj_1f468_1f3ff_200d_2695_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "man_heart_man", R.drawable.emj_1f468_200d_2764_fe0f_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc68" );
		name( "man_in_business_suit_levitating", R.drawable.emj_1f574_fe0f, "\ud83d\udd74\ufe0f" );
		name( "man_in_business_suit_levitating_dark_skin_tone", R.drawable.emj_1f574_1f3ff, "\ud83d\udd74\ud83c\udfff" );
		name( "man_in_business_suit_levitating_light_skin_tone", R.drawable.emj_1f574_1f3fb, "\ud83d\udd74\ud83c\udffb" );
		name( "man_in_business_suit_levitating_medium_dark_skin_tone", R.drawable.emj_1f574_1f3fe, "\ud83d\udd74\ud83c\udffe" );
		name( "man_in_business_suit_levitating_medium_light_skin_tone", R.drawable.emj_1f574_1f3fc, "\ud83d\udd74\ud83c\udffc" );
		name( "man_in_business_suit_levitating_medium_skin_tone", R.drawable.emj_1f574_1f3fd, "\ud83d\udd74\ud83c\udffd" );
		name( "man_in_business_suit_levitating_tone1", R.drawable.emj_1f574_1f3fb, "\ud83d\udd74\ud83c\udffb" );
		name( "man_in_business_suit_levitating_tone2", R.drawable.emj_1f574_1f3fc, "\ud83d\udd74\ud83c\udffc" );
		name( "man_in_business_suit_levitating_tone3", R.drawable.emj_1f574_1f3fd, "\ud83d\udd74\ud83c\udffd" );
		name( "man_in_business_suit_levitating_tone4", R.drawable.emj_1f574_1f3fe, "\ud83d\udd74\ud83c\udffe" );
		name( "man_in_business_suit_levitating_tone5", R.drawable.emj_1f574_1f3ff, "\ud83d\udd74\ud83c\udfff" );
		name( "man_in_lotus_position", R.drawable.emj_1f9d8_200d_2642_fe0f, "\ud83e\uddd8\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_dark_skin_tone", R.drawable.emj_1f9d8_1f3ff_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_light_skin_tone", R.drawable.emj_1f9d8_1f3fb_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_medium_dark_skin_tone", R.drawable.emj_1f9d8_1f3fe_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_medium_light_skin_tone", R.drawable.emj_1f9d8_1f3fc_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_medium_skin_tone", R.drawable.emj_1f9d8_1f3fd_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_tone1", R.drawable.emj_1f9d8_1f3fb_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_tone2", R.drawable.emj_1f9d8_1f3fc_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_tone3", R.drawable.emj_1f9d8_1f3fd_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_tone4", R.drawable.emj_1f9d8_1f3fe_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_in_lotus_position_tone5", R.drawable.emj_1f9d8_1f3ff_200d_2642_fe0f, "\ud83e\uddd8\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room", R.drawable.emj_1f9d6_200d_2642_fe0f, "\ud83e\uddd6\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_dark_skin_tone", R.drawable.emj_1f9d6_1f3ff_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_light_skin_tone", R.drawable.emj_1f9d6_1f3fb_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_medium_dark_skin_tone", R.drawable.emj_1f9d6_1f3fe_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_medium_light_skin_tone", R.drawable.emj_1f9d6_1f3fc_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_medium_skin_tone", R.drawable.emj_1f9d6_1f3fd_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffd\u200d\u2642\ufe0f" );
	}
	private static void init81(){
		name( "man_in_steamy_room_tone1", R.drawable.emj_1f9d6_1f3fb_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_tone2", R.drawable.emj_1f9d6_1f3fc_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_tone3", R.drawable.emj_1f9d6_1f3fd_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_tone4", R.drawable.emj_1f9d6_1f3fe_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_in_steamy_room_tone5", R.drawable.emj_1f9d6_1f3ff_200d_2642_fe0f, "\ud83e\uddd6\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_in_tuxedo", R.drawable.emj_1f935, "\ud83e\udd35" );
		name( "man_in_tuxedo_dark_skin_tone", R.drawable.emj_1f935_1f3ff, "\ud83e\udd35\ud83c\udfff" );
		name( "man_in_tuxedo_light_skin_tone", R.drawable.emj_1f935_1f3fb, "\ud83e\udd35\ud83c\udffb" );
		name( "man_in_tuxedo_medium_dark_skin_tone", R.drawable.emj_1f935_1f3fe, "\ud83e\udd35\ud83c\udffe" );
		name( "man_in_tuxedo_medium_light_skin_tone", R.drawable.emj_1f935_1f3fc, "\ud83e\udd35\ud83c\udffc" );
		name( "man_in_tuxedo_medium_skin_tone", R.drawable.emj_1f935_1f3fd, "\ud83e\udd35\ud83c\udffd" );
		name( "man_in_tuxedo_tone1", R.drawable.emj_1f935_1f3fb, "\ud83e\udd35\ud83c\udffb" );
		name( "man_in_tuxedo_tone2", R.drawable.emj_1f935_1f3fc, "\ud83e\udd35\ud83c\udffc" );
		name( "man_in_tuxedo_tone3", R.drawable.emj_1f935_1f3fd, "\ud83e\udd35\ud83c\udffd" );
		name( "man_in_tuxedo_tone4", R.drawable.emj_1f935_1f3fe, "\ud83e\udd35\ud83c\udffe" );
		name( "man_in_tuxedo_tone5", R.drawable.emj_1f935_1f3ff, "\ud83e\udd35\ud83c\udfff" );
		name( "man_judge", R.drawable.emj_1f468_200d_2696_fe0f, "\ud83d\udc68\u200d\u2696\ufe0f" );
		name( "man_judge_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_2696_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "man_judge_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "man_judge_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "man_judge_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "man_judge_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "man_judge_tone1", R.drawable.emj_1f468_1f3fb_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "man_judge_tone2", R.drawable.emj_1f468_1f3fc_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "man_judge_tone3", R.drawable.emj_1f468_1f3fd_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "man_judge_tone4", R.drawable.emj_1f468_1f3fe_200d_2696_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "man_judge_tone5", R.drawable.emj_1f468_1f3ff_200d_2696_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "man_juggling", R.drawable.emj_1f939_200d_2642_fe0f, "\ud83e\udd39\u200d\u2642\ufe0f" );
		name( "man_juggling_dark_skin_tone", R.drawable.emj_1f939_1f3ff_200d_2642_fe0f, "\ud83e\udd39\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_juggling_light_skin_tone", R.drawable.emj_1f939_1f3fb_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_juggling_medium_dark_skin_tone", R.drawable.emj_1f939_1f3fe_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_juggling_medium_light_skin_tone", R.drawable.emj_1f939_1f3fc_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_juggling_medium_skin_tone", R.drawable.emj_1f939_1f3fd_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_juggling_tone1", R.drawable.emj_1f939_1f3fb_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_juggling_tone2", R.drawable.emj_1f939_1f3fc_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_juggling_tone3", R.drawable.emj_1f939_1f3fd_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_juggling_tone4", R.drawable.emj_1f939_1f3fe_200d_2642_fe0f, "\ud83e\udd39\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_juggling_tone5", R.drawable.emj_1f939_1f3ff_200d_2642_fe0f, "\ud83e\udd39\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_kiss_man", R.drawable.emj_1f468_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc68\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68" );
		name( "man_lifting_weights", R.drawable.emj_1f3cb_fe0f_200d_2642_fe0f, "\ud83c\udfcb\ufe0f\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_dark_skin_tone", R.drawable.emj_1f3cb_1f3ff_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_light_skin_tone", R.drawable.emj_1f3cb_1f3fb_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_medium_dark_skin_tone", R.drawable.emj_1f3cb_1f3fe_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_medium_light_skin_tone", R.drawable.emj_1f3cb_1f3fc_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_medium_skin_tone", R.drawable.emj_1f3cb_1f3fd_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_tone1", R.drawable.emj_1f3cb_1f3fb_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_tone2", R.drawable.emj_1f3cb_1f3fc_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_tone3", R.drawable.emj_1f3cb_1f3fd_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_tone4", R.drawable.emj_1f3cb_1f3fe_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_lifting_weights_tone5", R.drawable.emj_1f3cb_1f3ff_200d_2642_fe0f, "\ud83c\udfcb\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_light_skin_tone", R.drawable.emj_1f468_1f3fb, "\ud83d\udc68\ud83c\udffb" );
		name( "man_man_boy", R.drawable.emj_1f468_200d_1f468_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc66" );
		name( "man_man_boy_boy", R.drawable.emj_1f468_200d_1f468_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "man_man_girl", R.drawable.emj_1f468_200d_1f468_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67" );
		name( "man_man_girl_boy", R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "man_man_girl_girl", R.drawable.emj_1f468_200d_1f468_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc68\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "man_mechanic", R.drawable.emj_1f468_200d_1f527, "\ud83d\udc68\u200d\ud83d\udd27" );
		name( "man_mechanic_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f527, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd27" );
		name( "man_mechanic_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f527, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd27" );
		name( "man_mechanic_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f527, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd27" );
		name( "man_mechanic_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f527, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd27" );
		name( "man_mechanic_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f527, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd27" );
		name( "man_mechanic_tone1", R.drawable.emj_1f468_1f3fb_200d_1f527, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd27" );
		name( "man_mechanic_tone2", R.drawable.emj_1f468_1f3fc_200d_1f527, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd27" );
		name( "man_mechanic_tone3", R.drawable.emj_1f468_1f3fd_200d_1f527, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd27" );
		name( "man_mechanic_tone4", R.drawable.emj_1f468_1f3fe_200d_1f527, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd27" );
		name( "man_mechanic_tone5", R.drawable.emj_1f468_1f3ff_200d_1f527, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd27" );
		name( "man_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe, "\ud83d\udc68\ud83c\udffe" );
		name( "man_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc, "\ud83d\udc68\ud83c\udffc" );
		name( "man_medium_skin_tone", R.drawable.emj_1f468_1f3fd, "\ud83d\udc68\ud83c\udffd" );
		name( "man_mountain_biking", R.drawable.emj_1f6b5_200d_2642_fe0f, "\ud83d\udeb5\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_dark_skin_tone", R.drawable.emj_1f6b5_1f3ff_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_light_skin_tone", R.drawable.emj_1f6b5_1f3fb_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_medium_dark_skin_tone", R.drawable.emj_1f6b5_1f3fe_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_medium_light_skin_tone", R.drawable.emj_1f6b5_1f3fc_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_medium_skin_tone", R.drawable.emj_1f6b5_1f3fd_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_tone1", R.drawable.emj_1f6b5_1f3fb_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_tone2", R.drawable.emj_1f6b5_1f3fc_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_tone3", R.drawable.emj_1f6b5_1f3fd_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_tone4", R.drawable.emj_1f6b5_1f3fe_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_mountain_biking_tone5", R.drawable.emj_1f6b5_1f3ff_200d_2642_fe0f, "\ud83d\udeb5\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_office_worker", R.drawable.emj_1f468_200d_1f4bc, "\ud83d\udc68\u200d\ud83d\udcbc" );
		name( "man_office_worker_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f4bc, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "man_office_worker_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f4bc, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "man_office_worker_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f4bc, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "man_office_worker_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f4bc, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "man_office_worker_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f4bc, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "man_office_worker_tone1", R.drawable.emj_1f468_1f3fb_200d_1f4bc, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "man_office_worker_tone2", R.drawable.emj_1f468_1f3fc_200d_1f4bc, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "man_office_worker_tone3", R.drawable.emj_1f468_1f3fd_200d_1f4bc, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "man_office_worker_tone4", R.drawable.emj_1f468_1f3fe_200d_1f4bc, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "man_office_worker_tone5", R.drawable.emj_1f468_1f3ff_200d_1f4bc, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "man_pilot", R.drawable.emj_1f468_200d_2708_fe0f, "\ud83d\udc68\u200d\u2708\ufe0f" );
		name( "man_pilot_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_2708_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "man_pilot_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "man_pilot_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "man_pilot_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "man_pilot_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2708\ufe0f" );
		name( "man_pilot_tone1", R.drawable.emj_1f468_1f3fb_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "man_pilot_tone2", R.drawable.emj_1f468_1f3fc_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "man_pilot_tone3", R.drawable.emj_1f468_1f3fd_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffd\u200d\u2708\ufe0f" );
	}
	private static void init82(){
		name( "man_pilot_tone4", R.drawable.emj_1f468_1f3fe_200d_2708_fe0f, "\ud83d\udc68\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "man_pilot_tone5", R.drawable.emj_1f468_1f3ff_200d_2708_fe0f, "\ud83d\udc68\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "man_playing_handball", R.drawable.emj_1f93e_200d_2642_fe0f, "\ud83e\udd3e\u200d\u2642\ufe0f" );
		name( "man_playing_handball_dark_skin_tone", R.drawable.emj_1f93e_1f3ff_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_playing_handball_light_skin_tone", R.drawable.emj_1f93e_1f3fb_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_playing_handball_medium_dark_skin_tone", R.drawable.emj_1f93e_1f3fe_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_playing_handball_medium_light_skin_tone", R.drawable.emj_1f93e_1f3fc_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_playing_handball_medium_skin_tone", R.drawable.emj_1f93e_1f3fd_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_playing_handball_tone1", R.drawable.emj_1f93e_1f3fb_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_playing_handball_tone2", R.drawable.emj_1f93e_1f3fc_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_playing_handball_tone3", R.drawable.emj_1f93e_1f3fd_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_playing_handball_tone4", R.drawable.emj_1f93e_1f3fe_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_playing_handball_tone5", R.drawable.emj_1f93e_1f3ff_200d_2642_fe0f, "\ud83e\udd3e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo", R.drawable.emj_1f93d_200d_2642_fe0f, "\ud83e\udd3d\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_dark_skin_tone", R.drawable.emj_1f93d_1f3ff_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_light_skin_tone", R.drawable.emj_1f93d_1f3fb_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_medium_dark_skin_tone", R.drawable.emj_1f93d_1f3fe_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_medium_light_skin_tone", R.drawable.emj_1f93d_1f3fc_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_medium_skin_tone", R.drawable.emj_1f93d_1f3fd_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_tone1", R.drawable.emj_1f93d_1f3fb_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_tone2", R.drawable.emj_1f93d_1f3fc_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_tone3", R.drawable.emj_1f93d_1f3fd_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_tone4", R.drawable.emj_1f93d_1f3fe_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_playing_water_polo_tone5", R.drawable.emj_1f93d_1f3ff_200d_2642_fe0f, "\ud83e\udd3d\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_police_officer", R.drawable.emj_1f46e_200d_2642_fe0f, "\ud83d\udc6e\u200d\u2642\ufe0f" );
		name( "man_police_officer_dark_skin_tone", R.drawable.emj_1f46e_1f3ff_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_police_officer_light_skin_tone", R.drawable.emj_1f46e_1f3fb_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_police_officer_medium_dark_skin_tone", R.drawable.emj_1f46e_1f3fe_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_police_officer_medium_light_skin_tone", R.drawable.emj_1f46e_1f3fc_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_police_officer_medium_skin_tone", R.drawable.emj_1f46e_1f3fd_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_police_officer_tone1", R.drawable.emj_1f46e_1f3fb_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_police_officer_tone2", R.drawable.emj_1f46e_1f3fc_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_police_officer_tone3", R.drawable.emj_1f46e_1f3fd_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_police_officer_tone4", R.drawable.emj_1f46e_1f3fe_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_police_officer_tone5", R.drawable.emj_1f46e_1f3ff_200d_2642_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_pouting", R.drawable.emj_1f64e_200d_2642_fe0f, "\ud83d\ude4e\u200d\u2642\ufe0f" );
		name( "man_pouting_dark_skin_tone", R.drawable.emj_1f64e_1f3ff_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_pouting_light_skin_tone", R.drawable.emj_1f64e_1f3fb_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_pouting_medium_dark_skin_tone", R.drawable.emj_1f64e_1f3fe_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_pouting_medium_light_skin_tone", R.drawable.emj_1f64e_1f3fc_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_pouting_medium_skin_tone", R.drawable.emj_1f64e_1f3fd_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_pouting_tone1", R.drawable.emj_1f64e_1f3fb_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_pouting_tone2", R.drawable.emj_1f64e_1f3fc_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_pouting_tone3", R.drawable.emj_1f64e_1f3fd_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_pouting_tone4", R.drawable.emj_1f64e_1f3fe_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_pouting_tone5", R.drawable.emj_1f64e_1f3ff_200d_2642_fe0f, "\ud83d\ude4e\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_raising_hand", R.drawable.emj_1f64b_200d_2642_fe0f, "\ud83d\ude4b\u200d\u2642\ufe0f" );
		name( "man_raising_hand_dark_skin_tone", R.drawable.emj_1f64b_1f3ff_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_raising_hand_light_skin_tone", R.drawable.emj_1f64b_1f3fb_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_raising_hand_medium_dark_skin_tone", R.drawable.emj_1f64b_1f3fe_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_raising_hand_medium_light_skin_tone", R.drawable.emj_1f64b_1f3fc_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_raising_hand_medium_skin_tone", R.drawable.emj_1f64b_1f3fd_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_raising_hand_tone1", R.drawable.emj_1f64b_1f3fb_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_raising_hand_tone2", R.drawable.emj_1f64b_1f3fc_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_raising_hand_tone3", R.drawable.emj_1f64b_1f3fd_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_raising_hand_tone4", R.drawable.emj_1f64b_1f3fe_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_raising_hand_tone5", R.drawable.emj_1f64b_1f3ff_200d_2642_fe0f, "\ud83d\ude4b\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_rowing_boat", R.drawable.emj_1f6a3_200d_2642_fe0f, "\ud83d\udea3\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_dark_skin_tone", R.drawable.emj_1f6a3_1f3ff_200d_2642_fe0f, "\ud83d\udea3\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_light_skin_tone", R.drawable.emj_1f6a3_1f3fb_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_medium_dark_skin_tone", R.drawable.emj_1f6a3_1f3fe_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_medium_light_skin_tone", R.drawable.emj_1f6a3_1f3fc_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_medium_skin_tone", R.drawable.emj_1f6a3_1f3fd_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_tone1", R.drawable.emj_1f6a3_1f3fb_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_tone2", R.drawable.emj_1f6a3_1f3fc_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_tone3", R.drawable.emj_1f6a3_1f3fd_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_tone4", R.drawable.emj_1f6a3_1f3fe_200d_2642_fe0f, "\ud83d\udea3\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_rowing_boat_tone5", R.drawable.emj_1f6a3_1f3ff_200d_2642_fe0f, "\ud83d\udea3\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_running", R.drawable.emj_1f3c3_200d_2642_fe0f, "\ud83c\udfc3\u200d\u2642\ufe0f" );
		name( "man_running_dark_skin_tone", R.drawable.emj_1f3c3_1f3ff_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_running_light_skin_tone", R.drawable.emj_1f3c3_1f3fb_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_running_medium_dark_skin_tone", R.drawable.emj_1f3c3_1f3fe_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_running_medium_light_skin_tone", R.drawable.emj_1f3c3_1f3fc_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_running_medium_skin_tone", R.drawable.emj_1f3c3_1f3fd_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_running_tone1", R.drawable.emj_1f3c3_1f3fb_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_running_tone2", R.drawable.emj_1f3c3_1f3fc_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_running_tone3", R.drawable.emj_1f3c3_1f3fd_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_running_tone4", R.drawable.emj_1f3c3_1f3fe_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_running_tone5", R.drawable.emj_1f3c3_1f3ff_200d_2642_fe0f, "\ud83c\udfc3\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_scientist", R.drawable.emj_1f468_200d_1f52c, "\ud83d\udc68\u200d\ud83d\udd2c" );
		name( "man_scientist_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f52c, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "man_scientist_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f52c, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "man_scientist_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f52c, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "man_scientist_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f52c, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd2c" );
		name( "man_scientist_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f52c, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "man_scientist_tone1", R.drawable.emj_1f468_1f3fb_200d_1f52c, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "man_scientist_tone2", R.drawable.emj_1f468_1f3fc_200d_1f52c, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udd2c" );
		name( "man_scientist_tone3", R.drawable.emj_1f468_1f3fd_200d_1f52c, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "man_scientist_tone4", R.drawable.emj_1f468_1f3fe_200d_1f52c, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "man_scientist_tone5", R.drawable.emj_1f468_1f3ff_200d_1f52c, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "man_shrugging", R.drawable.emj_1f937_200d_2642_fe0f, "\ud83e\udd37\u200d\u2642\ufe0f" );
		name( "man_shrugging_dark_skin_tone", R.drawable.emj_1f937_1f3ff_200d_2642_fe0f, "\ud83e\udd37\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_shrugging_light_skin_tone", R.drawable.emj_1f937_1f3fb_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_shrugging_medium_dark_skin_tone", R.drawable.emj_1f937_1f3fe_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_shrugging_medium_light_skin_tone", R.drawable.emj_1f937_1f3fc_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_shrugging_medium_skin_tone", R.drawable.emj_1f937_1f3fd_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_shrugging_tone1", R.drawable.emj_1f937_1f3fb_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_shrugging_tone2", R.drawable.emj_1f937_1f3fc_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_shrugging_tone3", R.drawable.emj_1f937_1f3fd_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_shrugging_tone4", R.drawable.emj_1f937_1f3fe_200d_2642_fe0f, "\ud83e\udd37\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_shrugging_tone5", R.drawable.emj_1f937_1f3ff_200d_2642_fe0f, "\ud83e\udd37\ud83c\udfff\u200d\u2642\ufe0f" );
	}
	private static void init83(){
		name( "man_singer", R.drawable.emj_1f468_200d_1f3a4, "\ud83d\udc68\u200d\ud83c\udfa4" );
		name( "man_singer_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3a4, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "man_singer_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3a4, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "man_singer_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3a4, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "man_singer_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3a4, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "man_singer_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3a4, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "man_singer_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3a4, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "man_singer_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3a4, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "man_singer_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3a4, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "man_singer_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3a4, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "man_singer_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3a4, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "man_student", R.drawable.emj_1f468_200d_1f393, "\ud83d\udc68\u200d\ud83c\udf93" );
		name( "man_student_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f393, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf93" );
		name( "man_student_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f393, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf93" );
		name( "man_student_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f393, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf93" );
		name( "man_student_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f393, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf93" );
		name( "man_student_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f393, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf93" );
		name( "man_student_tone1", R.drawable.emj_1f468_1f3fb_200d_1f393, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udf93" );
		name( "man_student_tone2", R.drawable.emj_1f468_1f3fc_200d_1f393, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udf93" );
		name( "man_student_tone3", R.drawable.emj_1f468_1f3fd_200d_1f393, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udf93" );
		name( "man_student_tone4", R.drawable.emj_1f468_1f3fe_200d_1f393, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udf93" );
		name( "man_student_tone5", R.drawable.emj_1f468_1f3ff_200d_1f393, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udf93" );
		name( "man_surfing", R.drawable.emj_1f3c4_200d_2642_fe0f, "\ud83c\udfc4\u200d\u2642\ufe0f" );
		name( "man_surfing_dark_skin_tone", R.drawable.emj_1f3c4_1f3ff_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_surfing_light_skin_tone", R.drawable.emj_1f3c4_1f3fb_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_surfing_medium_dark_skin_tone", R.drawable.emj_1f3c4_1f3fe_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_surfing_medium_light_skin_tone", R.drawable.emj_1f3c4_1f3fc_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_surfing_medium_skin_tone", R.drawable.emj_1f3c4_1f3fd_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_surfing_tone1", R.drawable.emj_1f3c4_1f3fb_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_surfing_tone2", R.drawable.emj_1f3c4_1f3fc_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_surfing_tone3", R.drawable.emj_1f3c4_1f3fd_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_surfing_tone4", R.drawable.emj_1f3c4_1f3fe_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_surfing_tone5", R.drawable.emj_1f3c4_1f3ff_200d_2642_fe0f, "\ud83c\udfc4\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_swimming", R.drawable.emj_1f3ca_200d_2642_fe0f, "\ud83c\udfca\u200d\u2642\ufe0f" );
		name( "man_swimming_dark_skin_tone", R.drawable.emj_1f3ca_1f3ff_200d_2642_fe0f, "\ud83c\udfca\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_swimming_light_skin_tone", R.drawable.emj_1f3ca_1f3fb_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_swimming_medium_dark_skin_tone", R.drawable.emj_1f3ca_1f3fe_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_swimming_medium_light_skin_tone", R.drawable.emj_1f3ca_1f3fc_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_swimming_medium_skin_tone", R.drawable.emj_1f3ca_1f3fd_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_swimming_tone1", R.drawable.emj_1f3ca_1f3fb_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_swimming_tone2", R.drawable.emj_1f3ca_1f3fc_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_swimming_tone3", R.drawable.emj_1f3ca_1f3fd_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_swimming_tone4", R.drawable.emj_1f3ca_1f3fe_200d_2642_fe0f, "\ud83c\udfca\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_swimming_tone5", R.drawable.emj_1f3ca_1f3ff_200d_2642_fe0f, "\ud83c\udfca\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_teacher", R.drawable.emj_1f468_200d_1f3eb, "\ud83d\udc68\u200d\ud83c\udfeb" );
		name( "man_teacher_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f3eb, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "man_teacher_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f3eb, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "man_teacher_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f3eb, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "man_teacher_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f3eb, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "man_teacher_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f3eb, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "man_teacher_tone1", R.drawable.emj_1f468_1f3fb_200d_1f3eb, "\ud83d\udc68\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "man_teacher_tone2", R.drawable.emj_1f468_1f3fc_200d_1f3eb, "\ud83d\udc68\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "man_teacher_tone3", R.drawable.emj_1f468_1f3fd_200d_1f3eb, "\ud83d\udc68\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "man_teacher_tone4", R.drawable.emj_1f468_1f3fe_200d_1f3eb, "\ud83d\udc68\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "man_teacher_tone5", R.drawable.emj_1f468_1f3ff_200d_1f3eb, "\ud83d\udc68\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "man_technologist", R.drawable.emj_1f468_200d_1f4bb, "\ud83d\udc68\u200d\ud83d\udcbb" );
		name( "man_technologist_dark_skin_tone", R.drawable.emj_1f468_1f3ff_200d_1f4bb, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "man_technologist_light_skin_tone", R.drawable.emj_1f468_1f3fb_200d_1f4bb, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "man_technologist_medium_dark_skin_tone", R.drawable.emj_1f468_1f3fe_200d_1f4bb, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "man_technologist_medium_light_skin_tone", R.drawable.emj_1f468_1f3fc_200d_1f4bb, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "man_technologist_medium_skin_tone", R.drawable.emj_1f468_1f3fd_200d_1f4bb, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "man_technologist_tone1", R.drawable.emj_1f468_1f3fb_200d_1f4bb, "\ud83d\udc68\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "man_technologist_tone2", R.drawable.emj_1f468_1f3fc_200d_1f4bb, "\ud83d\udc68\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "man_technologist_tone3", R.drawable.emj_1f468_1f3fd_200d_1f4bb, "\ud83d\udc68\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "man_technologist_tone4", R.drawable.emj_1f468_1f3fe_200d_1f4bb, "\ud83d\udc68\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "man_technologist_tone5", R.drawable.emj_1f468_1f3ff_200d_1f4bb, "\ud83d\udc68\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "man_tipping_hand", R.drawable.emj_1f481_200d_2642_fe0f, "\ud83d\udc81\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_dark_skin_tone", R.drawable.emj_1f481_1f3ff_200d_2642_fe0f, "\ud83d\udc81\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_light_skin_tone", R.drawable.emj_1f481_1f3fb_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_medium_dark_skin_tone", R.drawable.emj_1f481_1f3fe_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_medium_light_skin_tone", R.drawable.emj_1f481_1f3fc_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_medium_skin_tone", R.drawable.emj_1f481_1f3fd_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_tone1", R.drawable.emj_1f481_1f3fb_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_tone2", R.drawable.emj_1f481_1f3fc_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_tone3", R.drawable.emj_1f481_1f3fd_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_tone4", R.drawable.emj_1f481_1f3fe_200d_2642_fe0f, "\ud83d\udc81\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_tipping_hand_tone5", R.drawable.emj_1f481_1f3ff_200d_2642_fe0f, "\ud83d\udc81\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_tone1", R.drawable.emj_1f468_1f3fb, "\ud83d\udc68\ud83c\udffb" );
		name( "man_tone2", R.drawable.emj_1f468_1f3fc, "\ud83d\udc68\ud83c\udffc" );
		name( "man_tone3", R.drawable.emj_1f468_1f3fd, "\ud83d\udc68\ud83c\udffd" );
		name( "man_tone4", R.drawable.emj_1f468_1f3fe, "\ud83d\udc68\ud83c\udffe" );
		name( "man_tone5", R.drawable.emj_1f468_1f3ff, "\ud83d\udc68\ud83c\udfff" );
		name( "man_walking", R.drawable.emj_1f6b6_200d_2642_fe0f, "\ud83d\udeb6\u200d\u2642\ufe0f" );
		name( "man_walking_dark_skin_tone", R.drawable.emj_1f6b6_1f3ff_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_walking_light_skin_tone", R.drawable.emj_1f6b6_1f3fb_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_walking_medium_dark_skin_tone", R.drawable.emj_1f6b6_1f3fe_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_walking_medium_light_skin_tone", R.drawable.emj_1f6b6_1f3fc_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_walking_medium_skin_tone", R.drawable.emj_1f6b6_1f3fd_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_walking_tone1", R.drawable.emj_1f6b6_1f3fb_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_walking_tone2", R.drawable.emj_1f6b6_1f3fc_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_walking_tone3", R.drawable.emj_1f6b6_1f3fd_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_walking_tone4", R.drawable.emj_1f6b6_1f3fe_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_walking_tone5", R.drawable.emj_1f6b6_1f3ff_200d_2642_fe0f, "\ud83d\udeb6\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_wearing_turban", R.drawable.emj_1f473_200d_2642_fe0f, "\ud83d\udc73\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_dark_skin_tone", R.drawable.emj_1f473_1f3ff_200d_2642_fe0f, "\ud83d\udc73\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_light_skin_tone", R.drawable.emj_1f473_1f3fb_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_medium_dark_skin_tone", R.drawable.emj_1f473_1f3fe_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_medium_light_skin_tone", R.drawable.emj_1f473_1f3fc_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_medium_skin_tone", R.drawable.emj_1f473_1f3fd_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_tone1", R.drawable.emj_1f473_1f3fb_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_tone2", R.drawable.emj_1f473_1f3fc_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffc\u200d\u2642\ufe0f" );
	}
	private static void init84(){
		name( "man_wearing_turban_tone3", R.drawable.emj_1f473_1f3fd_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_tone4", R.drawable.emj_1f473_1f3fe_200d_2642_fe0f, "\ud83d\udc73\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "man_wearing_turban_tone5", R.drawable.emj_1f473_1f3ff_200d_2642_fe0f, "\ud83d\udc73\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "man_with_bunny_ears_partying", R.drawable.emj_1f46f_200d_2642_fe0f, "\ud83d\udc6f\u200d\u2642\ufe0f" );
		name( "man_with_gua_pi_mao", R.drawable.emj_1f472, "\ud83d\udc72" );
		name( "man_with_gua_pi_mao_dark_skin_tone", R.drawable.emj_1f472_1f3ff, "\ud83d\udc72\ud83c\udfff" );
		name( "man_with_gua_pi_mao_light_skin_tone", R.drawable.emj_1f472_1f3fb, "\ud83d\udc72\ud83c\udffb" );
		name( "man_with_gua_pi_mao_medium_dark_skin_tone", R.drawable.emj_1f472_1f3fe, "\ud83d\udc72\ud83c\udffe" );
		name( "man_with_gua_pi_mao_medium_light_skin_tone", R.drawable.emj_1f472_1f3fc, "\ud83d\udc72\ud83c\udffc" );
		name( "man_with_gua_pi_mao_medium_skin_tone", R.drawable.emj_1f472_1f3fd, "\ud83d\udc72\ud83c\udffd" );
		name( "man_with_gua_pi_mao_tone1", R.drawable.emj_1f472_1f3fb, "\ud83d\udc72\ud83c\udffb" );
		name( "man_with_gua_pi_mao_tone2", R.drawable.emj_1f472_1f3fc, "\ud83d\udc72\ud83c\udffc" );
		name( "man_with_gua_pi_mao_tone3", R.drawable.emj_1f472_1f3fd, "\ud83d\udc72\ud83c\udffd" );
		name( "man_with_gua_pi_mao_tone4", R.drawable.emj_1f472_1f3fe, "\ud83d\udc72\ud83c\udffe" );
		name( "man_with_gua_pi_mao_tone5", R.drawable.emj_1f472_1f3ff, "\ud83d\udc72\ud83c\udfff" );
		name( "man_with_turban", R.drawable.emj_1f473, "\ud83d\udc73" );
		name( "man_with_turban_dark_skin_tone", R.drawable.emj_1f473_1f3ff, "\ud83d\udc73\ud83c\udfff" );
		name( "man_with_turban_light_skin_tone", R.drawable.emj_1f473_1f3fb, "\ud83d\udc73\ud83c\udffb" );
		name( "man_with_turban_medium_dark_skin_tone", R.drawable.emj_1f473_1f3fe, "\ud83d\udc73\ud83c\udffe" );
		name( "man_with_turban_medium_light_skin_tone", R.drawable.emj_1f473_1f3fc, "\ud83d\udc73\ud83c\udffc" );
		name( "man_with_turban_medium_skin_tone", R.drawable.emj_1f473_1f3fd, "\ud83d\udc73\ud83c\udffd" );
		name( "man_with_turban_tone1", R.drawable.emj_1f473_1f3fb, "\ud83d\udc73\ud83c\udffb" );
		name( "man_with_turban_tone2", R.drawable.emj_1f473_1f3fc, "\ud83d\udc73\ud83c\udffc" );
		name( "man_with_turban_tone3", R.drawable.emj_1f473_1f3fd, "\ud83d\udc73\ud83c\udffd" );
		name( "man_with_turban_tone4", R.drawable.emj_1f473_1f3fe, "\ud83d\udc73\ud83c\udffe" );
		name( "man_with_turban_tone5", R.drawable.emj_1f473_1f3ff, "\ud83d\udc73\ud83c\udfff" );
		name( "man_woman_boy", R.drawable.emj_1f468_200d_1f469_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc66" );
		name( "man_woman_boy_boy", R.drawable.emj_1f468_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "man_woman_girl", R.drawable.emj_1f468_200d_1f469_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67" );
		name( "man_woman_girl_boy", R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "man_woman_girl_girl", R.drawable.emj_1f468_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc68\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "man_wrestling", R.drawable.emj_1f93c_200d_2642_fe0f, "\ud83e\udd3c\u200d\u2642\ufe0f" );
		name( "mans_shoe", R.drawable.emj_1f45e, "\ud83d\udc5e" );
		name( "mantelpiece_clock", R.drawable.emj_1f570_fe0f, "\ud83d\udd70\ufe0f" );
		name( "mantlepiece_clock", R.drawable.emj_1f570_fe0f, "\ud83d\udd70\ufe0f" );
		name( "map", R.drawable.emj_1f5fa_fe0f, "\ud83d\uddfa\ufe0f" );
		name( "maple_leaf", R.drawable.emj_1f341, "\ud83c\udf41" );
		name( "martial_arts_uniform", R.drawable.emj_1f94b, "\ud83e\udd4b" );
		name( "mask", R.drawable.emj_1f637, "\ud83d\ude37" );
		name( "massage", R.drawable.emj_1f486, "\ud83d\udc86" );
		name( "massage_dark_skin_tone", R.drawable.emj_1f486_1f3ff, "\ud83d\udc86\ud83c\udfff" );
		name( "massage_light_skin_tone", R.drawable.emj_1f486_1f3fb, "\ud83d\udc86\ud83c\udffb" );
		name( "massage_medium_dark_skin_tone", R.drawable.emj_1f486_1f3fe, "\ud83d\udc86\ud83c\udffe" );
		name( "massage_medium_light_skin_tone", R.drawable.emj_1f486_1f3fc, "\ud83d\udc86\ud83c\udffc" );
		name( "massage_medium_skin_tone", R.drawable.emj_1f486_1f3fd, "\ud83d\udc86\ud83c\udffd" );
		name( "massage_tone1", R.drawable.emj_1f486_1f3fb, "\ud83d\udc86\ud83c\udffb" );
		name( "massage_tone2", R.drawable.emj_1f486_1f3fc, "\ud83d\udc86\ud83c\udffc" );
		name( "massage_tone3", R.drawable.emj_1f486_1f3fd, "\ud83d\udc86\ud83c\udffd" );
		name( "massage_tone4", R.drawable.emj_1f486_1f3fe, "\ud83d\udc86\ud83c\udffe" );
		name( "massage_tone5", R.drawable.emj_1f486_1f3ff, "\ud83d\udc86\ud83c\udfff" );
		name( "mc", R.drawable.emj_1f1f2_1f1e8, "\ud83c\uddf2\ud83c\udde8" );
		name( "md", R.drawable.emj_1f1f2_1f1e9, "\ud83c\uddf2\ud83c\udde9" );
		name( "me", R.drawable.emj_1f1f2_1f1ea, "\ud83c\uddf2\ud83c\uddea" );
		name( "meat_on_bone", R.drawable.emj_1f356, "\ud83c\udf56" );
		name( "medal", R.drawable.emj_1f3c5, "\ud83c\udfc5" );
		name( "medical_symbol", R.drawable.emj_2695_fe0f, "\u2695\ufe0f" );
		name( "mega", R.drawable.emj_1f4e3, "\ud83d\udce3" );
		name( "melon", R.drawable.emj_1f348, "\ud83c\udf48" );
		name( "memo", R.drawable.emj_1f4dd, "\ud83d\udcdd" );
		name( "men_with_bunny_ears_partying", R.drawable.emj_1f46f_200d_2642_fe0f, "\ud83d\udc6f\u200d\u2642\ufe0f" );
		name( "men_wrestling", R.drawable.emj_1f93c_200d_2642_fe0f, "\ud83e\udd3c\u200d\u2642\ufe0f" );
		name( "menorah", R.drawable.emj_1f54e, "\ud83d\udd4e" );
		name( "menorah_with_nine_branches", R.drawable.emj_1f54e, "\ud83d\udd4e" );
		name( "mens", R.drawable.emj_1f6b9, "\ud83d\udeb9" );
		name( "mermaid", R.drawable.emj_1f9dc_200d_2640_fe0f, "\ud83e\udddc\u200d\u2640\ufe0f" );
		name( "mermaid_dark_skin_tone", R.drawable.emj_1f9dc_1f3ff_200d_2640_fe0f, "\ud83e\udddc\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "mermaid_light_skin_tone", R.drawable.emj_1f9dc_1f3fb_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "mermaid_medium_dark_skin_tone", R.drawable.emj_1f9dc_1f3fe_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "mermaid_medium_light_skin_tone", R.drawable.emj_1f9dc_1f3fc_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "mermaid_medium_skin_tone", R.drawable.emj_1f9dc_1f3fd_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "mermaid_tone1", R.drawable.emj_1f9dc_1f3fb_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "mermaid_tone2", R.drawable.emj_1f9dc_1f3fc_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "mermaid_tone3", R.drawable.emj_1f9dc_1f3fd_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "mermaid_tone4", R.drawable.emj_1f9dc_1f3fe_200d_2640_fe0f, "\ud83e\udddc\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "mermaid_tone5", R.drawable.emj_1f9dc_1f3ff_200d_2640_fe0f, "\ud83e\udddc\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "merman", R.drawable.emj_1f9dc_200d_2642_fe0f, "\ud83e\udddc\u200d\u2642\ufe0f" );
		name( "merman_dark_skin_tone", R.drawable.emj_1f9dc_1f3ff_200d_2642_fe0f, "\ud83e\udddc\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "merman_light_skin_tone", R.drawable.emj_1f9dc_1f3fb_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "merman_medium_dark_skin_tone", R.drawable.emj_1f9dc_1f3fe_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "merman_medium_light_skin_tone", R.drawable.emj_1f9dc_1f3fc_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "merman_medium_skin_tone", R.drawable.emj_1f9dc_1f3fd_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "merman_tone1", R.drawable.emj_1f9dc_1f3fb_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffb\u200d\u2642\ufe0f" );
		name( "merman_tone2", R.drawable.emj_1f9dc_1f3fc_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffc\u200d\u2642\ufe0f" );
		name( "merman_tone3", R.drawable.emj_1f9dc_1f3fd_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffd\u200d\u2642\ufe0f" );
		name( "merman_tone4", R.drawable.emj_1f9dc_1f3fe_200d_2642_fe0f, "\ud83e\udddc\ud83c\udffe\u200d\u2642\ufe0f" );
		name( "merman_tone5", R.drawable.emj_1f9dc_1f3ff_200d_2642_fe0f, "\ud83e\udddc\ud83c\udfff\u200d\u2642\ufe0f" );
		name( "merperson", R.drawable.emj_1f9dc, "\ud83e\udddc" );
		name( "merperson_dark_skin_tone", R.drawable.emj_1f9dc_1f3ff, "\ud83e\udddc\ud83c\udfff" );
		name( "merperson_light_skin_tone", R.drawable.emj_1f9dc_1f3fb, "\ud83e\udddc\ud83c\udffb" );
		name( "merperson_medium_dark_skin_tone", R.drawable.emj_1f9dc_1f3fe, "\ud83e\udddc\ud83c\udffe" );
		name( "merperson_medium_light_skin_tone", R.drawable.emj_1f9dc_1f3fc, "\ud83e\udddc\ud83c\udffc" );
		name( "merperson_medium_skin_tone", R.drawable.emj_1f9dc_1f3fd, "\ud83e\udddc\ud83c\udffd" );
		name( "merperson_tone1", R.drawable.emj_1f9dc_1f3fb, "\ud83e\udddc\ud83c\udffb" );
		name( "merperson_tone2", R.drawable.emj_1f9dc_1f3fc, "\ud83e\udddc\ud83c\udffc" );
		name( "merperson_tone3", R.drawable.emj_1f9dc_1f3fd, "\ud83e\udddc\ud83c\udffd" );
		name( "merperson_tone4", R.drawable.emj_1f9dc_1f3fe, "\ud83e\udddc\ud83c\udffe" );
		name( "merperson_tone5", R.drawable.emj_1f9dc_1f3ff, "\ud83e\udddc\ud83c\udfff" );
		name( "metal", R.drawable.emj_1f918, "\ud83e\udd18" );
		name( "metal_tone1", R.drawable.emj_1f918_1f3fb, "\ud83e\udd18\ud83c\udffb" );
		name( "metal_tone2", R.drawable.emj_1f918_1f3fc, "\ud83e\udd18\ud83c\udffc" );
		name( "metal_tone3", R.drawable.emj_1f918_1f3fd, "\ud83e\udd18\ud83c\udffd" );
	}
	private static void init85(){
		name( "metal_tone4", R.drawable.emj_1f918_1f3fe, "\ud83e\udd18\ud83c\udffe" );
		name( "metal_tone5", R.drawable.emj_1f918_1f3ff, "\ud83e\udd18\ud83c\udfff" );
		name( "metro", R.drawable.emj_1f687, "\ud83d\ude87" );
		name( "mf", R.drawable.emj_1f1f2_1f1eb, "\ud83c\uddf2\ud83c\uddeb" );
		name( "mg", R.drawable.emj_1f1f2_1f1ec, "\ud83c\uddf2\ud83c\uddec" );
		name( "mh", R.drawable.emj_1f1f2_1f1ed, "\ud83c\uddf2\ud83c\udded" );
		name( "microphone", R.drawable.emj_1f3a4, "\ud83c\udfa4" );
		name( "microphone2", R.drawable.emj_1f399_fe0f, "\ud83c\udf99\ufe0f" );
		name( "microscope", R.drawable.emj_1f52c, "\ud83d\udd2c" );
		name( "middle_finger", R.drawable.emj_1f595, "\ud83d\udd95" );
		name( "middle_finger_dark_skin_tone", R.drawable.emj_1f595_1f3ff, "\ud83d\udd95\ud83c\udfff" );
		name( "middle_finger_light_skin_tone", R.drawable.emj_1f595_1f3fb, "\ud83d\udd95\ud83c\udffb" );
		name( "middle_finger_medium_dark_skin_tone", R.drawable.emj_1f595_1f3fe, "\ud83d\udd95\ud83c\udffe" );
		name( "middle_finger_medium_light_skin_tone", R.drawable.emj_1f595_1f3fc, "\ud83d\udd95\ud83c\udffc" );
		name( "middle_finger_medium_skin_tone", R.drawable.emj_1f595_1f3fd, "\ud83d\udd95\ud83c\udffd" );
		name( "middle_finger_tone1", R.drawable.emj_1f595_1f3fb, "\ud83d\udd95\ud83c\udffb" );
		name( "middle_finger_tone2", R.drawable.emj_1f595_1f3fc, "\ud83d\udd95\ud83c\udffc" );
		name( "middle_finger_tone3", R.drawable.emj_1f595_1f3fd, "\ud83d\udd95\ud83c\udffd" );
		name( "middle_finger_tone4", R.drawable.emj_1f595_1f3fe, "\ud83d\udd95\ud83c\udffe" );
		name( "middle_finger_tone5", R.drawable.emj_1f595_1f3ff, "\ud83d\udd95\ud83c\udfff" );
		name( "military_medal", R.drawable.emj_1f396_fe0f, "\ud83c\udf96\ufe0f" );
		name( "milk", R.drawable.emj_1f95b, "\ud83e\udd5b" );
		name( "milky_way", R.drawable.emj_1f30c, "\ud83c\udf0c" );
		name( "minibus", R.drawable.emj_1f690, "\ud83d\ude90" );
		name( "minidisc", R.drawable.emj_1f4bd, "\ud83d\udcbd" );
		name( "mk", R.drawable.emj_1f1f2_1f1f0, "\ud83c\uddf2\ud83c\uddf0" );
		name( "ml", R.drawable.emj_1f1f2_1f1f1, "\ud83c\uddf2\ud83c\uddf1" );
		name( "mm", R.drawable.emj_1f1f2_1f1f2, "\ud83c\uddf2\ud83c\uddf2" );
		name( "mn", R.drawable.emj_1f1f2_1f1f3, "\ud83c\uddf2\ud83c\uddf3" );
		name( "mo", R.drawable.emj_1f1f2_1f1f4, "\ud83c\uddf2\ud83c\uddf4" );
		name( "mobile_phone_off", R.drawable.emj_1f4f4, "\ud83d\udcf4" );
		name( "money_mouth", R.drawable.emj_1f911, "\ud83e\udd11" );
		name( "money_mouth_face", R.drawable.emj_1f911, "\ud83e\udd11" );
		name( "money_with_wings", R.drawable.emj_1f4b8, "\ud83d\udcb8" );
		name( "moneybag", R.drawable.emj_1f4b0, "\ud83d\udcb0" );
		name( "monkey", R.drawable.emj_1f412, "\ud83d\udc12" );
		name( "monkey_face", R.drawable.emj_1f435, "\ud83d\udc35" );
		name( "monorail", R.drawable.emj_1f69d, "\ud83d\ude9d" );
		name( "moon", R.drawable.emj_1f314, "\ud83c\udf14" );
		name( "mortar_board", R.drawable.emj_1f393, "\ud83c\udf93" );
		name( "mosque", R.drawable.emj_1f54c, "\ud83d\udd4c" );
		name( "mostly_sunny", R.drawable.emj_1f324_fe0f, "\ud83c\udf24\ufe0f" );
		name( "mother_christmas", R.drawable.emj_1f936, "\ud83e\udd36" );
		name( "mother_christmas_dark_skin_tone", R.drawable.emj_1f936_1f3ff, "\ud83e\udd36\ud83c\udfff" );
		name( "mother_christmas_light_skin_tone", R.drawable.emj_1f936_1f3fb, "\ud83e\udd36\ud83c\udffb" );
		name( "mother_christmas_medium_dark_skin_tone", R.drawable.emj_1f936_1f3fe, "\ud83e\udd36\ud83c\udffe" );
		name( "mother_christmas_medium_light_skin_tone", R.drawable.emj_1f936_1f3fc, "\ud83e\udd36\ud83c\udffc" );
		name( "mother_christmas_medium_skin_tone", R.drawable.emj_1f936_1f3fd, "\ud83e\udd36\ud83c\udffd" );
		name( "mother_christmas_tone1", R.drawable.emj_1f936_1f3fb, "\ud83e\udd36\ud83c\udffb" );
		name( "mother_christmas_tone2", R.drawable.emj_1f936_1f3fc, "\ud83e\udd36\ud83c\udffc" );
		name( "mother_christmas_tone3", R.drawable.emj_1f936_1f3fd, "\ud83e\udd36\ud83c\udffd" );
		name( "mother_christmas_tone4", R.drawable.emj_1f936_1f3fe, "\ud83e\udd36\ud83c\udffe" );
		name( "mother_christmas_tone5", R.drawable.emj_1f936_1f3ff, "\ud83e\udd36\ud83c\udfff" );
		name( "motor_boat", R.drawable.emj_1f6e5_fe0f, "\ud83d\udee5\ufe0f" );
		name( "motor_scooter", R.drawable.emj_1f6f5, "\ud83d\udef5" );
		name( "motorbike", R.drawable.emj_1f6f5, "\ud83d\udef5" );
		name( "motorboat", R.drawable.emj_1f6e5_fe0f, "\ud83d\udee5\ufe0f" );
		name( "motorcycle", R.drawable.emj_1f3cd_fe0f, "\ud83c\udfcd\ufe0f" );
		name( "motorway", R.drawable.emj_1f6e3_fe0f, "\ud83d\udee3\ufe0f" );
		name( "mount_fuji", R.drawable.emj_1f5fb, "\ud83d\uddfb" );
		name( "mountain", R.drawable.emj_26f0_fe0f, "\u26f0\ufe0f" );
		name( "mountain_bicyclist", R.drawable.emj_1f6b5, "\ud83d\udeb5" );
		name( "mountain_bicyclist_dark_skin_tone", R.drawable.emj_1f6b5_1f3ff, "\ud83d\udeb5\ud83c\udfff" );
		name( "mountain_bicyclist_light_skin_tone", R.drawable.emj_1f6b5_1f3fb, "\ud83d\udeb5\ud83c\udffb" );
		name( "mountain_bicyclist_medium_dark_skin_tone", R.drawable.emj_1f6b5_1f3fe, "\ud83d\udeb5\ud83c\udffe" );
		name( "mountain_bicyclist_medium_light_skin_tone", R.drawable.emj_1f6b5_1f3fc, "\ud83d\udeb5\ud83c\udffc" );
		name( "mountain_bicyclist_medium_skin_tone", R.drawable.emj_1f6b5_1f3fd, "\ud83d\udeb5\ud83c\udffd" );
		name( "mountain_bicyclist_tone1", R.drawable.emj_1f6b5_1f3fb, "\ud83d\udeb5\ud83c\udffb" );
		name( "mountain_bicyclist_tone2", R.drawable.emj_1f6b5_1f3fc, "\ud83d\udeb5\ud83c\udffc" );
		name( "mountain_bicyclist_tone3", R.drawable.emj_1f6b5_1f3fd, "\ud83d\udeb5\ud83c\udffd" );
		name( "mountain_bicyclist_tone4", R.drawable.emj_1f6b5_1f3fe, "\ud83d\udeb5\ud83c\udffe" );
		name( "mountain_bicyclist_tone5", R.drawable.emj_1f6b5_1f3ff, "\ud83d\udeb5\ud83c\udfff" );
		name( "mountain_cableway", R.drawable.emj_1f6a0, "\ud83d\udea0" );
		name( "mountain_railway", R.drawable.emj_1f69e, "\ud83d\ude9e" );
		name( "mountain_snow", R.drawable.emj_1f3d4_fe0f, "\ud83c\udfd4\ufe0f" );
		name( "mouse", R.drawable.emj_1f42d, "\ud83d\udc2d" );
		name( "mouse2", R.drawable.emj_1f401, "\ud83d\udc01" );
		name( "mouse_three_button", R.drawable.emj_1f5b1_fe0f, "\ud83d\uddb1\ufe0f" );
		name( "movie_camera", R.drawable.emj_1f3a5, "\ud83c\udfa5" );
		name( "moyai", R.drawable.emj_1f5ff, "\ud83d\uddff" );
		name( "mp", R.drawable.emj_1f1f2_1f1f5, "\ud83c\uddf2\ud83c\uddf5" );
		name( "mq", R.drawable.emj_1f1f2_1f1f6, "\ud83c\uddf2\ud83c\uddf6" );
		name( "mr", R.drawable.emj_1f1f2_1f1f7, "\ud83c\uddf2\ud83c\uddf7" );
		name( "mrs_claus", R.drawable.emj_1f936, "\ud83e\udd36" );
		name( "mrs_claus_dark_skin_tone", R.drawable.emj_1f936_1f3ff, "\ud83e\udd36\ud83c\udfff" );
		name( "mrs_claus_light_skin_tone", R.drawable.emj_1f936_1f3fb, "\ud83e\udd36\ud83c\udffb" );
		name( "mrs_claus_medium_dark_skin_tone", R.drawable.emj_1f936_1f3fe, "\ud83e\udd36\ud83c\udffe" );
		name( "mrs_claus_medium_light_skin_tone", R.drawable.emj_1f936_1f3fc, "\ud83e\udd36\ud83c\udffc" );
		name( "mrs_claus_medium_skin_tone", R.drawable.emj_1f936_1f3fd, "\ud83e\udd36\ud83c\udffd" );
		name( "mrs_claus_tone1", R.drawable.emj_1f936_1f3fb, "\ud83e\udd36\ud83c\udffb" );
		name( "mrs_claus_tone2", R.drawable.emj_1f936_1f3fc, "\ud83e\udd36\ud83c\udffc" );
		name( "mrs_claus_tone3", R.drawable.emj_1f936_1f3fd, "\ud83e\udd36\ud83c\udffd" );
		name( "mrs_claus_tone4", R.drawable.emj_1f936_1f3fe, "\ud83e\udd36\ud83c\udffe" );
		name( "mrs_claus_tone5", R.drawable.emj_1f936_1f3ff, "\ud83e\udd36\ud83c\udfff" );
		name( "ms", R.drawable.emj_1f1f2_1f1f8, "\ud83c\uddf2\ud83c\uddf8" );
		name( "mt", R.drawable.emj_1f1f2_1f1f9, "\ud83c\uddf2\ud83c\uddf9" );
		name( "mu", R.drawable.emj_1f1f2_1f1fa, "\ud83c\uddf2\ud83c\uddfa" );
		name( "muscle", R.drawable.emj_1f4aa, "\ud83d\udcaa" );
		name( "muscle_dark_skin_tone", R.drawable.emj_1f4aa_1f3ff, "\ud83d\udcaa\ud83c\udfff" );
		name( "muscle_light_skin_tone", R.drawable.emj_1f4aa_1f3fb, "\ud83d\udcaa\ud83c\udffb" );
		name( "muscle_medium_dark_skin_tone", R.drawable.emj_1f4aa_1f3fe, "\ud83d\udcaa\ud83c\udffe" );
	}
	private static void init86(){
		name( "muscle_medium_light_skin_tone", R.drawable.emj_1f4aa_1f3fc, "\ud83d\udcaa\ud83c\udffc" );
		name( "muscle_medium_skin_tone", R.drawable.emj_1f4aa_1f3fd, "\ud83d\udcaa\ud83c\udffd" );
		name( "muscle_tone1", R.drawable.emj_1f4aa_1f3fb, "\ud83d\udcaa\ud83c\udffb" );
		name( "muscle_tone2", R.drawable.emj_1f4aa_1f3fc, "\ud83d\udcaa\ud83c\udffc" );
		name( "muscle_tone3", R.drawable.emj_1f4aa_1f3fd, "\ud83d\udcaa\ud83c\udffd" );
		name( "muscle_tone4", R.drawable.emj_1f4aa_1f3fe, "\ud83d\udcaa\ud83c\udffe" );
		name( "muscle_tone5", R.drawable.emj_1f4aa_1f3ff, "\ud83d\udcaa\ud83c\udfff" );
		name( "mushroom", R.drawable.emj_1f344, "\ud83c\udf44" );
		name( "musical_keyboard", R.drawable.emj_1f3b9, "\ud83c\udfb9" );
		name( "musical_note", R.drawable.emj_1f3b5, "\ud83c\udfb5" );
		name( "musical_score", R.drawable.emj_1f3bc, "\ud83c\udfbc" );
		name( "mute", R.drawable.emj_1f507, "\ud83d\udd07" );
		name( "mv", R.drawable.emj_1f1f2_1f1fb, "\ud83c\uddf2\ud83c\uddfb" );
		name( "mw", R.drawable.emj_1f1f2_1f1fc, "\ud83c\uddf2\ud83c\uddfc" );
		name( "mx", R.drawable.emj_1f1f2_1f1fd, "\ud83c\uddf2\ud83c\uddfd" );
		name( "my", R.drawable.emj_1f1f2_1f1fe, "\ud83c\uddf2\ud83c\uddfe" );
		name( "mz", R.drawable.emj_1f1f2_1f1ff, "\ud83c\uddf2\ud83c\uddff" );
		name( "na", R.drawable.emj_1f1f3_1f1e6, "\ud83c\uddf3\ud83c\udde6" );
		name( "nail_care", R.drawable.emj_1f485, "\ud83d\udc85" );
		name( "nail_care_dark_skin_tone", R.drawable.emj_1f485_1f3ff, "\ud83d\udc85\ud83c\udfff" );
		name( "nail_care_light_skin_tone", R.drawable.emj_1f485_1f3fb, "\ud83d\udc85\ud83c\udffb" );
		name( "nail_care_medium_dark_skin_tone", R.drawable.emj_1f485_1f3fe, "\ud83d\udc85\ud83c\udffe" );
		name( "nail_care_medium_light_skin_tone", R.drawable.emj_1f485_1f3fc, "\ud83d\udc85\ud83c\udffc" );
		name( "nail_care_medium_skin_tone", R.drawable.emj_1f485_1f3fd, "\ud83d\udc85\ud83c\udffd" );
		name( "nail_care_tone1", R.drawable.emj_1f485_1f3fb, "\ud83d\udc85\ud83c\udffb" );
		name( "nail_care_tone2", R.drawable.emj_1f485_1f3fc, "\ud83d\udc85\ud83c\udffc" );
		name( "nail_care_tone3", R.drawable.emj_1f485_1f3fd, "\ud83d\udc85\ud83c\udffd" );
		name( "nail_care_tone4", R.drawable.emj_1f485_1f3fe, "\ud83d\udc85\ud83c\udffe" );
		name( "nail_care_tone5", R.drawable.emj_1f485_1f3ff, "\ud83d\udc85\ud83c\udfff" );
		name( "name_badge", R.drawable.emj_1f4db, "\ud83d\udcdb" );
		name( "national_park", R.drawable.emj_1f3de_fe0f, "\ud83c\udfde\ufe0f" );
		name( "nauseated_face", R.drawable.emj_1f922, "\ud83e\udd22" );
		name( "nc", R.drawable.emj_1f1f3_1f1e8, "\ud83c\uddf3\ud83c\udde8" );
		name( "ne", R.drawable.emj_1f1f3_1f1ea, "\ud83c\uddf3\ud83c\uddea" );
		name( "necktie", R.drawable.emj_1f454, "\ud83d\udc54" );
		name( "negative_squared_cross_mark", R.drawable.emj_274e, "\u274e" );
		name( "nerd", R.drawable.emj_1f913, "\ud83e\udd13" );
		name( "nerd_face", R.drawable.emj_1f913, "\ud83e\udd13" );
		name( "neutral_face", R.drawable.emj_1f610, "\ud83d\ude10" );
		name( "new", R.drawable.emj_1f195, "\ud83c\udd95" );
		name( "new_moon", R.drawable.emj_1f311, "\ud83c\udf11" );
		name( "new_moon_with_face", R.drawable.emj_1f31a, "\ud83c\udf1a" );
		name( "newspaper", R.drawable.emj_1f4f0, "\ud83d\udcf0" );
		name( "newspaper2", R.drawable.emj_1f5de_fe0f, "\ud83d\uddde\ufe0f" );
		name( "next_track", R.drawable.emj_23ed_fe0f, "\u23ed\ufe0f" );
		name( "nf", R.drawable.emj_1f1f3_1f1eb, "\ud83c\uddf3\ud83c\uddeb" );
		name( "ng", R.drawable.emj_1f196, "\ud83c\udd96" );
		name( "ni", R.drawable.emj_1f1f3_1f1ee, "\ud83c\uddf3\ud83c\uddee" );
		name( "nigeria", R.drawable.emj_1f1f3_1f1ec, "\ud83c\uddf3\ud83c\uddec" );
		name( "night_with_stars", R.drawable.emj_1f303, "\ud83c\udf03" );
		name( "nine", R.drawable.emj_0039_fe0f_20e3, "\u0039\ufe0f\u20e3" );
		name( "nl", R.drawable.emj_1f1f3_1f1f1, "\ud83c\uddf3\ud83c\uddf1" );
		name( "no", R.drawable.emj_1f1f3_1f1f4, "\ud83c\uddf3\ud83c\uddf4" );
		name( "no_bell", R.drawable.emj_1f515, "\ud83d\udd15" );
		name( "no_bicycles", R.drawable.emj_1f6b3, "\ud83d\udeb3" );
		name( "no_entry", R.drawable.emj_26d4, "\u26d4" );
		name( "no_entry_sign", R.drawable.emj_1f6ab, "\ud83d\udeab" );
		name( "no_good", R.drawable.emj_1f645, "\ud83d\ude45" );
		name( "no_good_dark_skin_tone", R.drawable.emj_1f645_1f3ff, "\ud83d\ude45\ud83c\udfff" );
		name( "no_good_light_skin_tone", R.drawable.emj_1f645_1f3fb, "\ud83d\ude45\ud83c\udffb" );
		name( "no_good_medium_dark_skin_tone", R.drawable.emj_1f645_1f3fe, "\ud83d\ude45\ud83c\udffe" );
		name( "no_good_medium_light_skin_tone", R.drawable.emj_1f645_1f3fc, "\ud83d\ude45\ud83c\udffc" );
		name( "no_good_medium_skin_tone", R.drawable.emj_1f645_1f3fd, "\ud83d\ude45\ud83c\udffd" );
		name( "no_good_tone1", R.drawable.emj_1f645_1f3fb, "\ud83d\ude45\ud83c\udffb" );
		name( "no_good_tone2", R.drawable.emj_1f645_1f3fc, "\ud83d\ude45\ud83c\udffc" );
		name( "no_good_tone3", R.drawable.emj_1f645_1f3fd, "\ud83d\ude45\ud83c\udffd" );
		name( "no_good_tone4", R.drawable.emj_1f645_1f3fe, "\ud83d\ude45\ud83c\udffe" );
		name( "no_good_tone5", R.drawable.emj_1f645_1f3ff, "\ud83d\ude45\ud83c\udfff" );
		name( "no_mobile_phones", R.drawable.emj_1f4f5, "\ud83d\udcf5" );
		name( "no_mouth", R.drawable.emj_1f636, "\ud83d\ude36" );
		name( "no_pedestrians", R.drawable.emj_1f6b7, "\ud83d\udeb7" );
		name( "no_smoking", R.drawable.emj_1f6ad, "\ud83d\udead" );
		name( "non_potable_water", R.drawable.emj_1f6b1, "\ud83d\udeb1" );
		name( "nose", R.drawable.emj_1f443, "\ud83d\udc43" );
		name( "nose_dark_skin_tone", R.drawable.emj_1f443_1f3ff, "\ud83d\udc43\ud83c\udfff" );
		name( "nose_light_skin_tone", R.drawable.emj_1f443_1f3fb, "\ud83d\udc43\ud83c\udffb" );
		name( "nose_medium_dark_skin_tone", R.drawable.emj_1f443_1f3fe, "\ud83d\udc43\ud83c\udffe" );
		name( "nose_medium_light_skin_tone", R.drawable.emj_1f443_1f3fc, "\ud83d\udc43\ud83c\udffc" );
		name( "nose_medium_skin_tone", R.drawable.emj_1f443_1f3fd, "\ud83d\udc43\ud83c\udffd" );
		name( "nose_tone1", R.drawable.emj_1f443_1f3fb, "\ud83d\udc43\ud83c\udffb" );
		name( "nose_tone2", R.drawable.emj_1f443_1f3fc, "\ud83d\udc43\ud83c\udffc" );
		name( "nose_tone3", R.drawable.emj_1f443_1f3fd, "\ud83d\udc43\ud83c\udffd" );
		name( "nose_tone4", R.drawable.emj_1f443_1f3fe, "\ud83d\udc43\ud83c\udffe" );
		name( "nose_tone5", R.drawable.emj_1f443_1f3ff, "\ud83d\udc43\ud83c\udfff" );
		name( "notebook", R.drawable.emj_1f4d3, "\ud83d\udcd3" );
		name( "notebook_with_decorative_cover", R.drawable.emj_1f4d4, "\ud83d\udcd4" );
		name( "notepad_spiral", R.drawable.emj_1f5d2_fe0f, "\ud83d\uddd2\ufe0f" );
		name( "notes", R.drawable.emj_1f3b6, "\ud83c\udfb6" );
		name( "np", R.drawable.emj_1f1f3_1f1f5, "\ud83c\uddf3\ud83c\uddf5" );
		name( "nr", R.drawable.emj_1f1f3_1f1f7, "\ud83c\uddf3\ud83c\uddf7" );
		name( "nu", R.drawable.emj_1f1f3_1f1fa, "\ud83c\uddf3\ud83c\uddfa" );
		name( "nut_and_bolt", R.drawable.emj_1f529, "\ud83d\udd29" );
		name( "nz", R.drawable.emj_1f1f3_1f1ff, "\ud83c\uddf3\ud83c\uddff" );
		name( "o", R.drawable.emj_2b55, "\u2b55" );
		name( "o2", R.drawable.emj_1f17e_fe0f, "\ud83c\udd7e\ufe0f" );
		name( "ocean", R.drawable.emj_1f30a, "\ud83c\udf0a" );
		name( "octagonal_sign", R.drawable.emj_1f6d1, "\ud83d\uded1" );
		name( "octopus", R.drawable.emj_1f419, "\ud83d\udc19" );
		name( "oden", R.drawable.emj_1f362, "\ud83c\udf62" );
		name( "office", R.drawable.emj_1f3e2, "\ud83c\udfe2" );
		name( "oil", R.drawable.emj_1f6e2_fe0f, "\ud83d\udee2\ufe0f" );
	}
	private static void init87(){
		name( "oil_drum", R.drawable.emj_1f6e2_fe0f, "\ud83d\udee2\ufe0f" );
		name( "ok", R.drawable.emj_1f197, "\ud83c\udd97" );
		name( "ok_hand", R.drawable.emj_1f44c, "\ud83d\udc4c" );
		name( "ok_hand_dark_skin_tone", R.drawable.emj_1f44c_1f3ff, "\ud83d\udc4c\ud83c\udfff" );
		name( "ok_hand_light_skin_tone", R.drawable.emj_1f44c_1f3fb, "\ud83d\udc4c\ud83c\udffb" );
		name( "ok_hand_medium_dark_skin_tone", R.drawable.emj_1f44c_1f3fe, "\ud83d\udc4c\ud83c\udffe" );
		name( "ok_hand_medium_light_skin_tone", R.drawable.emj_1f44c_1f3fc, "\ud83d\udc4c\ud83c\udffc" );
		name( "ok_hand_medium_skin_tone", R.drawable.emj_1f44c_1f3fd, "\ud83d\udc4c\ud83c\udffd" );
		name( "ok_hand_tone1", R.drawable.emj_1f44c_1f3fb, "\ud83d\udc4c\ud83c\udffb" );
		name( "ok_hand_tone2", R.drawable.emj_1f44c_1f3fc, "\ud83d\udc4c\ud83c\udffc" );
		name( "ok_hand_tone3", R.drawable.emj_1f44c_1f3fd, "\ud83d\udc4c\ud83c\udffd" );
		name( "ok_hand_tone4", R.drawable.emj_1f44c_1f3fe, "\ud83d\udc4c\ud83c\udffe" );
		name( "ok_hand_tone5", R.drawable.emj_1f44c_1f3ff, "\ud83d\udc4c\ud83c\udfff" );
		name( "ok_woman", R.drawable.emj_1f646, "\ud83d\ude46" );
		name( "ok_woman_dark_skin_tone", R.drawable.emj_1f646_1f3ff, "\ud83d\ude46\ud83c\udfff" );
		name( "ok_woman_light_skin_tone", R.drawable.emj_1f646_1f3fb, "\ud83d\ude46\ud83c\udffb" );
		name( "ok_woman_medium_dark_skin_tone", R.drawable.emj_1f646_1f3fe, "\ud83d\ude46\ud83c\udffe" );
		name( "ok_woman_medium_light_skin_tone", R.drawable.emj_1f646_1f3fc, "\ud83d\ude46\ud83c\udffc" );
		name( "ok_woman_medium_skin_tone", R.drawable.emj_1f646_1f3fd, "\ud83d\ude46\ud83c\udffd" );
		name( "ok_woman_tone1", R.drawable.emj_1f646_1f3fb, "\ud83d\ude46\ud83c\udffb" );
		name( "ok_woman_tone2", R.drawable.emj_1f646_1f3fc, "\ud83d\ude46\ud83c\udffc" );
		name( "ok_woman_tone3", R.drawable.emj_1f646_1f3fd, "\ud83d\ude46\ud83c\udffd" );
		name( "ok_woman_tone4", R.drawable.emj_1f646_1f3fe, "\ud83d\ude46\ud83c\udffe" );
		name( "ok_woman_tone5", R.drawable.emj_1f646_1f3ff, "\ud83d\ude46\ud83c\udfff" );
		name( "old_key", R.drawable.emj_1f5dd_fe0f, "\ud83d\udddd\ufe0f" );
		name( "older_adult", R.drawable.emj_1f9d3, "\ud83e\uddd3" );
		name( "older_adult_dark_skin_tone", R.drawable.emj_1f9d3_1f3ff, "\ud83e\uddd3\ud83c\udfff" );
		name( "older_adult_light_skin_tone", R.drawable.emj_1f9d3_1f3fb, "\ud83e\uddd3\ud83c\udffb" );
		name( "older_adult_medium_dark_skin_tone", R.drawable.emj_1f9d3_1f3fe, "\ud83e\uddd3\ud83c\udffe" );
		name( "older_adult_medium_light_skin_tone", R.drawable.emj_1f9d3_1f3fc, "\ud83e\uddd3\ud83c\udffc" );
		name( "older_adult_medium_skin_tone", R.drawable.emj_1f9d3_1f3fd, "\ud83e\uddd3\ud83c\udffd" );
		name( "older_adult_tone1", R.drawable.emj_1f9d3_1f3fb, "\ud83e\uddd3\ud83c\udffb" );
		name( "older_adult_tone2", R.drawable.emj_1f9d3_1f3fc, "\ud83e\uddd3\ud83c\udffc" );
		name( "older_adult_tone3", R.drawable.emj_1f9d3_1f3fd, "\ud83e\uddd3\ud83c\udffd" );
		name( "older_adult_tone4", R.drawable.emj_1f9d3_1f3fe, "\ud83e\uddd3\ud83c\udffe" );
		name( "older_adult_tone5", R.drawable.emj_1f9d3_1f3ff, "\ud83e\uddd3\ud83c\udfff" );
		name( "older_man", R.drawable.emj_1f474, "\ud83d\udc74" );
		name( "older_man_dark_skin_tone", R.drawable.emj_1f474_1f3ff, "\ud83d\udc74\ud83c\udfff" );
		name( "older_man_light_skin_tone", R.drawable.emj_1f474_1f3fb, "\ud83d\udc74\ud83c\udffb" );
		name( "older_man_medium_dark_skin_tone", R.drawable.emj_1f474_1f3fe, "\ud83d\udc74\ud83c\udffe" );
		name( "older_man_medium_light_skin_tone", R.drawable.emj_1f474_1f3fc, "\ud83d\udc74\ud83c\udffc" );
		name( "older_man_medium_skin_tone", R.drawable.emj_1f474_1f3fd, "\ud83d\udc74\ud83c\udffd" );
		name( "older_man_tone1", R.drawable.emj_1f474_1f3fb, "\ud83d\udc74\ud83c\udffb" );
		name( "older_man_tone2", R.drawable.emj_1f474_1f3fc, "\ud83d\udc74\ud83c\udffc" );
		name( "older_man_tone3", R.drawable.emj_1f474_1f3fd, "\ud83d\udc74\ud83c\udffd" );
		name( "older_man_tone4", R.drawable.emj_1f474_1f3fe, "\ud83d\udc74\ud83c\udffe" );
		name( "older_man_tone5", R.drawable.emj_1f474_1f3ff, "\ud83d\udc74\ud83c\udfff" );
		name( "older_woman", R.drawable.emj_1f475, "\ud83d\udc75" );
		name( "older_woman_dark_skin_tone", R.drawable.emj_1f475_1f3ff, "\ud83d\udc75\ud83c\udfff" );
		name( "older_woman_light_skin_tone", R.drawable.emj_1f475_1f3fb, "\ud83d\udc75\ud83c\udffb" );
		name( "older_woman_medium_dark_skin_tone", R.drawable.emj_1f475_1f3fe, "\ud83d\udc75\ud83c\udffe" );
		name( "older_woman_medium_light_skin_tone", R.drawable.emj_1f475_1f3fc, "\ud83d\udc75\ud83c\udffc" );
		name( "older_woman_medium_skin_tone", R.drawable.emj_1f475_1f3fd, "\ud83d\udc75\ud83c\udffd" );
		name( "older_woman_tone1", R.drawable.emj_1f475_1f3fb, "\ud83d\udc75\ud83c\udffb" );
		name( "older_woman_tone2", R.drawable.emj_1f475_1f3fc, "\ud83d\udc75\ud83c\udffc" );
		name( "older_woman_tone3", R.drawable.emj_1f475_1f3fd, "\ud83d\udc75\ud83c\udffd" );
		name( "older_woman_tone4", R.drawable.emj_1f475_1f3fe, "\ud83d\udc75\ud83c\udffe" );
		name( "older_woman_tone5", R.drawable.emj_1f475_1f3ff, "\ud83d\udc75\ud83c\udfff" );
		name( "om", R.drawable.emj_1f1f4_1f1f2, "\ud83c\uddf4\ud83c\uddf2" );
		name( "om_symbol", R.drawable.emj_1f549_fe0f, "\ud83d\udd49\ufe0f" );
		name( "on", R.drawable.emj_1f51b, "\ud83d\udd1b" );
		name( "oncoming_automobile", R.drawable.emj_1f698, "\ud83d\ude98" );
		name( "oncoming_bus", R.drawable.emj_1f68d, "\ud83d\ude8d" );
		name( "oncoming_police_car", R.drawable.emj_1f694, "\ud83d\ude94" );
		name( "oncoming_taxi", R.drawable.emj_1f696, "\ud83d\ude96" );
		name( "one", R.drawable.emj_0031_fe0f_20e3, "\u0031\ufe0f\u20e3" );
		name( "open_book", R.drawable.emj_1f4d6, "\ud83d\udcd6" );
		name( "open_file_folder", R.drawable.emj_1f4c2, "\ud83d\udcc2" );
		name( "open_hands", R.drawable.emj_1f450, "\ud83d\udc50" );
		name( "open_hands_dark_skin_tone", R.drawable.emj_1f450_1f3ff, "\ud83d\udc50\ud83c\udfff" );
		name( "open_hands_light_skin_tone", R.drawable.emj_1f450_1f3fb, "\ud83d\udc50\ud83c\udffb" );
		name( "open_hands_medium_dark_skin_tone", R.drawable.emj_1f450_1f3fe, "\ud83d\udc50\ud83c\udffe" );
		name( "open_hands_medium_light_skin_tone", R.drawable.emj_1f450_1f3fc, "\ud83d\udc50\ud83c\udffc" );
		name( "open_hands_medium_skin_tone", R.drawable.emj_1f450_1f3fd, "\ud83d\udc50\ud83c\udffd" );
		name( "open_hands_tone1", R.drawable.emj_1f450_1f3fb, "\ud83d\udc50\ud83c\udffb" );
		name( "open_hands_tone2", R.drawable.emj_1f450_1f3fc, "\ud83d\udc50\ud83c\udffc" );
		name( "open_hands_tone3", R.drawable.emj_1f450_1f3fd, "\ud83d\udc50\ud83c\udffd" );
		name( "open_hands_tone4", R.drawable.emj_1f450_1f3fe, "\ud83d\udc50\ud83c\udffe" );
		name( "open_hands_tone5", R.drawable.emj_1f450_1f3ff, "\ud83d\udc50\ud83c\udfff" );
		name( "open_mouth", R.drawable.emj_1f62e, "\ud83d\ude2e" );
		name( "ophiuchus", R.drawable.emj_26ce, "\u26ce" );
		name( "orange_book", R.drawable.emj_1f4d9, "\ud83d\udcd9" );
		name( "orange_heart", R.drawable.emj_1f9e1, "\ud83e\udde1" );
		name( "orthodox_cross", R.drawable.emj_2626_fe0f, "\u2626\ufe0f" );
		name( "outbox_tray", R.drawable.emj_1f4e4, "\ud83d\udce4" );
		name( "owl", R.drawable.emj_1f989, "\ud83e\udd89" );
		name( "ox", R.drawable.emj_1f402, "\ud83d\udc02" );
		name( "pa", R.drawable.emj_1f1f5_1f1e6, "\ud83c\uddf5\ud83c\udde6" );
		name( "package", R.drawable.emj_1f4e6, "\ud83d\udce6" );
		name( "paella", R.drawable.emj_1f958, "\ud83e\udd58" );
		name( "page_facing_up", R.drawable.emj_1f4c4, "\ud83d\udcc4" );
		name( "page_with_curl", R.drawable.emj_1f4c3, "\ud83d\udcc3" );
		name( "pager", R.drawable.emj_1f4df, "\ud83d\udcdf" );
		name( "paintbrush", R.drawable.emj_1f58c_fe0f, "\ud83d\udd8c\ufe0f" );
		name( "palm_tree", R.drawable.emj_1f334, "\ud83c\udf34" );
		name( "palms_up_together", R.drawable.emj_1f932, "\ud83e\udd32" );
		name( "palms_up_together_dark_skin_tone", R.drawable.emj_1f932_1f3ff, "\ud83e\udd32\ud83c\udfff" );
		name( "palms_up_together_light_skin_tone", R.drawable.emj_1f932_1f3fb, "\ud83e\udd32\ud83c\udffb" );
		name( "palms_up_together_medium_dark_skin_tone", R.drawable.emj_1f932_1f3fe, "\ud83e\udd32\ud83c\udffe" );
		name( "palms_up_together_medium_light_skin_tone", R.drawable.emj_1f932_1f3fc, "\ud83e\udd32\ud83c\udffc" );
		name( "palms_up_together_medium_skin_tone", R.drawable.emj_1f932_1f3fd, "\ud83e\udd32\ud83c\udffd" );
	}
	private static void init88(){
		name( "palms_up_together_tone1", R.drawable.emj_1f932_1f3fb, "\ud83e\udd32\ud83c\udffb" );
		name( "palms_up_together_tone2", R.drawable.emj_1f932_1f3fc, "\ud83e\udd32\ud83c\udffc" );
		name( "palms_up_together_tone3", R.drawable.emj_1f932_1f3fd, "\ud83e\udd32\ud83c\udffd" );
		name( "palms_up_together_tone4", R.drawable.emj_1f932_1f3fe, "\ud83e\udd32\ud83c\udffe" );
		name( "palms_up_together_tone5", R.drawable.emj_1f932_1f3ff, "\ud83e\udd32\ud83c\udfff" );
		name( "pancakes", R.drawable.emj_1f95e, "\ud83e\udd5e" );
		name( "panda_face", R.drawable.emj_1f43c, "\ud83d\udc3c" );
		name( "paperclip", R.drawable.emj_1f4ce, "\ud83d\udcce" );
		name( "paperclips", R.drawable.emj_1f587_fe0f, "\ud83d\udd87\ufe0f" );
		name( "park", R.drawable.emj_1f3de_fe0f, "\ud83c\udfde\ufe0f" );
		name( "parking", R.drawable.emj_1f17f_fe0f, "\ud83c\udd7f\ufe0f" );
		name( "part_alternation_mark", R.drawable.emj_303d_fe0f, "\u303d\ufe0f" );
		name( "partly_sunny", R.drawable.emj_26c5, "\u26c5" );
		name( "partly_sunny_rain", R.drawable.emj_1f326_fe0f, "\ud83c\udf26\ufe0f" );
		name( "passenger_ship", R.drawable.emj_1f6f3_fe0f, "\ud83d\udef3\ufe0f" );
		name( "passport_control", R.drawable.emj_1f6c2, "\ud83d\udec2" );
		name( "pause_button", R.drawable.emj_23f8_fe0f, "\u23f8\ufe0f" );
		name( "paw_prints", R.drawable.emj_1f43e, "\ud83d\udc3e" );
		name( "pe", R.drawable.emj_1f1f5_1f1ea, "\ud83c\uddf5\ud83c\uddea" );
		name( "peace", R.drawable.emj_262e_fe0f, "\u262e\ufe0f" );
		name( "peace_symbol", R.drawable.emj_262e_fe0f, "\u262e\ufe0f" );
		name( "peach", R.drawable.emj_1f351, "\ud83c\udf51" );
		name( "peanuts", R.drawable.emj_1f95c, "\ud83e\udd5c" );
		name( "pear", R.drawable.emj_1f350, "\ud83c\udf50" );
		name( "pen_ballpoint", R.drawable.emj_1f58a_fe0f, "\ud83d\udd8a\ufe0f" );
		name( "pen_fountain", R.drawable.emj_1f58b_fe0f, "\ud83d\udd8b\ufe0f" );
		name( "pencil", R.drawable.emj_1f4dd, "\ud83d\udcdd" );
		name( "pencil2", R.drawable.emj_270f_fe0f, "\u270f\ufe0f" );
		name( "penguin", R.drawable.emj_1f427, "\ud83d\udc27" );
		name( "pensive", R.drawable.emj_1f614, "\ud83d\ude14" );
		name( "performing_arts", R.drawable.emj_1f3ad, "\ud83c\udfad" );
		name( "persevere", R.drawable.emj_1f623, "\ud83d\ude23" );
		name( "person_climbing", R.drawable.emj_1f9d7, "\ud83e\uddd7" );
		name( "person_climbing_dark_skin_tone", R.drawable.emj_1f9d7_1f3ff, "\ud83e\uddd7\ud83c\udfff" );
		name( "person_climbing_light_skin_tone", R.drawable.emj_1f9d7_1f3fb, "\ud83e\uddd7\ud83c\udffb" );
		name( "person_climbing_medium_dark_skin_tone", R.drawable.emj_1f9d7_1f3fe, "\ud83e\uddd7\ud83c\udffe" );
		name( "person_climbing_medium_light_skin_tone", R.drawable.emj_1f9d7_1f3fc, "\ud83e\uddd7\ud83c\udffc" );
		name( "person_climbing_medium_skin_tone", R.drawable.emj_1f9d7_1f3fd, "\ud83e\uddd7\ud83c\udffd" );
		name( "person_climbing_tone1", R.drawable.emj_1f9d7_1f3fb, "\ud83e\uddd7\ud83c\udffb" );
		name( "person_climbing_tone2", R.drawable.emj_1f9d7_1f3fc, "\ud83e\uddd7\ud83c\udffc" );
		name( "person_climbing_tone3", R.drawable.emj_1f9d7_1f3fd, "\ud83e\uddd7\ud83c\udffd" );
		name( "person_climbing_tone4", R.drawable.emj_1f9d7_1f3fe, "\ud83e\uddd7\ud83c\udffe" );
		name( "person_climbing_tone5", R.drawable.emj_1f9d7_1f3ff, "\ud83e\uddd7\ud83c\udfff" );
		name( "person_doing_cartwheel", R.drawable.emj_1f938, "\ud83e\udd38" );
		name( "person_doing_cartwheel_dark_skin_tone", R.drawable.emj_1f938_1f3ff, "\ud83e\udd38\ud83c\udfff" );
		name( "person_doing_cartwheel_light_skin_tone", R.drawable.emj_1f938_1f3fb, "\ud83e\udd38\ud83c\udffb" );
		name( "person_doing_cartwheel_medium_dark_skin_tone", R.drawable.emj_1f938_1f3fe, "\ud83e\udd38\ud83c\udffe" );
		name( "person_doing_cartwheel_medium_light_skin_tone", R.drawable.emj_1f938_1f3fc, "\ud83e\udd38\ud83c\udffc" );
		name( "person_doing_cartwheel_medium_skin_tone", R.drawable.emj_1f938_1f3fd, "\ud83e\udd38\ud83c\udffd" );
		name( "person_doing_cartwheel_tone1", R.drawable.emj_1f938_1f3fb, "\ud83e\udd38\ud83c\udffb" );
		name( "person_doing_cartwheel_tone2", R.drawable.emj_1f938_1f3fc, "\ud83e\udd38\ud83c\udffc" );
		name( "person_doing_cartwheel_tone3", R.drawable.emj_1f938_1f3fd, "\ud83e\udd38\ud83c\udffd" );
		name( "person_doing_cartwheel_tone4", R.drawable.emj_1f938_1f3fe, "\ud83e\udd38\ud83c\udffe" );
		name( "person_doing_cartwheel_tone5", R.drawable.emj_1f938_1f3ff, "\ud83e\udd38\ud83c\udfff" );
		name( "person_frowning", R.drawable.emj_1f64d, "\ud83d\ude4d" );
		name( "person_frowning_dark_skin_tone", R.drawable.emj_1f64d_1f3ff, "\ud83d\ude4d\ud83c\udfff" );
		name( "person_frowning_light_skin_tone", R.drawable.emj_1f64d_1f3fb, "\ud83d\ude4d\ud83c\udffb" );
		name( "person_frowning_medium_dark_skin_tone", R.drawable.emj_1f64d_1f3fe, "\ud83d\ude4d\ud83c\udffe" );
		name( "person_frowning_medium_light_skin_tone", R.drawable.emj_1f64d_1f3fc, "\ud83d\ude4d\ud83c\udffc" );
		name( "person_frowning_medium_skin_tone", R.drawable.emj_1f64d_1f3fd, "\ud83d\ude4d\ud83c\udffd" );
		name( "person_frowning_tone1", R.drawable.emj_1f64d_1f3fb, "\ud83d\ude4d\ud83c\udffb" );
		name( "person_frowning_tone2", R.drawable.emj_1f64d_1f3fc, "\ud83d\ude4d\ud83c\udffc" );
		name( "person_frowning_tone3", R.drawable.emj_1f64d_1f3fd, "\ud83d\ude4d\ud83c\udffd" );
		name( "person_frowning_tone4", R.drawable.emj_1f64d_1f3fe, "\ud83d\ude4d\ud83c\udffe" );
		name( "person_frowning_tone5", R.drawable.emj_1f64d_1f3ff, "\ud83d\ude4d\ud83c\udfff" );
		name( "person_golfing_dark_skin_tone", R.drawable.emj_1f3cc_1f3ff, "\ud83c\udfcc\ud83c\udfff" );
		name( "person_golfing_light_skin_tone", R.drawable.emj_1f3cc_1f3fb, "\ud83c\udfcc\ud83c\udffb" );
		name( "person_golfing_medium_dark_skin_tone", R.drawable.emj_1f3cc_1f3fe, "\ud83c\udfcc\ud83c\udffe" );
		name( "person_golfing_medium_light_skin_tone", R.drawable.emj_1f3cc_1f3fc, "\ud83c\udfcc\ud83c\udffc" );
		name( "person_golfing_medium_skin_tone", R.drawable.emj_1f3cc_1f3fd, "\ud83c\udfcc\ud83c\udffd" );
		name( "person_golfing_tone1", R.drawable.emj_1f3cc_1f3fb, "\ud83c\udfcc\ud83c\udffb" );
		name( "person_golfing_tone2", R.drawable.emj_1f3cc_1f3fc, "\ud83c\udfcc\ud83c\udffc" );
		name( "person_golfing_tone3", R.drawable.emj_1f3cc_1f3fd, "\ud83c\udfcc\ud83c\udffd" );
		name( "person_golfing_tone4", R.drawable.emj_1f3cc_1f3fe, "\ud83c\udfcc\ud83c\udffe" );
		name( "person_golfing_tone5", R.drawable.emj_1f3cc_1f3ff, "\ud83c\udfcc\ud83c\udfff" );
		name( "person_in_bed_dark_skin_tone", R.drawable.emj_1f6cc_1f3ff, "\ud83d\udecc\ud83c\udfff" );
		name( "person_in_bed_light_skin_tone", R.drawable.emj_1f6cc_1f3fb, "\ud83d\udecc\ud83c\udffb" );
		name( "person_in_bed_medium_dark_skin_tone", R.drawable.emj_1f6cc_1f3fe, "\ud83d\udecc\ud83c\udffe" );
		name( "person_in_bed_medium_light_skin_tone", R.drawable.emj_1f6cc_1f3fc, "\ud83d\udecc\ud83c\udffc" );
		name( "person_in_bed_medium_skin_tone", R.drawable.emj_1f6cc_1f3fd, "\ud83d\udecc\ud83c\udffd" );
		name( "person_in_bed_tone1", R.drawable.emj_1f6cc_1f3fb, "\ud83d\udecc\ud83c\udffb" );
		name( "person_in_bed_tone2", R.drawable.emj_1f6cc_1f3fc, "\ud83d\udecc\ud83c\udffc" );
		name( "person_in_bed_tone3", R.drawable.emj_1f6cc_1f3fd, "\ud83d\udecc\ud83c\udffd" );
		name( "person_in_bed_tone4", R.drawable.emj_1f6cc_1f3fe, "\ud83d\udecc\ud83c\udffe" );
		name( "person_in_bed_tone5", R.drawable.emj_1f6cc_1f3ff, "\ud83d\udecc\ud83c\udfff" );
		name( "person_in_lotus_position", R.drawable.emj_1f9d8, "\ud83e\uddd8" );
		name( "person_in_lotus_position_dark_skin_tone", R.drawable.emj_1f9d8_1f3ff, "\ud83e\uddd8\ud83c\udfff" );
		name( "person_in_lotus_position_light_skin_tone", R.drawable.emj_1f9d8_1f3fb, "\ud83e\uddd8\ud83c\udffb" );
		name( "person_in_lotus_position_medium_dark_skin_tone", R.drawable.emj_1f9d8_1f3fe, "\ud83e\uddd8\ud83c\udffe" );
		name( "person_in_lotus_position_medium_light_skin_tone", R.drawable.emj_1f9d8_1f3fc, "\ud83e\uddd8\ud83c\udffc" );
		name( "person_in_lotus_position_medium_skin_tone", R.drawable.emj_1f9d8_1f3fd, "\ud83e\uddd8\ud83c\udffd" );
		name( "person_in_lotus_position_tone1", R.drawable.emj_1f9d8_1f3fb, "\ud83e\uddd8\ud83c\udffb" );
		name( "person_in_lotus_position_tone2", R.drawable.emj_1f9d8_1f3fc, "\ud83e\uddd8\ud83c\udffc" );
		name( "person_in_lotus_position_tone3", R.drawable.emj_1f9d8_1f3fd, "\ud83e\uddd8\ud83c\udffd" );
		name( "person_in_lotus_position_tone4", R.drawable.emj_1f9d8_1f3fe, "\ud83e\uddd8\ud83c\udffe" );
		name( "person_in_lotus_position_tone5", R.drawable.emj_1f9d8_1f3ff, "\ud83e\uddd8\ud83c\udfff" );
		name( "person_in_steamy_room", R.drawable.emj_1f9d6, "\ud83e\uddd6" );
		name( "person_in_steamy_room_dark_skin_tone", R.drawable.emj_1f9d6_1f3ff, "\ud83e\uddd6\ud83c\udfff" );
		name( "person_in_steamy_room_light_skin_tone", R.drawable.emj_1f9d6_1f3fb, "\ud83e\uddd6\ud83c\udffb" );
		name( "person_in_steamy_room_medium_dark_skin_tone", R.drawable.emj_1f9d6_1f3fe, "\ud83e\uddd6\ud83c\udffe" );
		name( "person_in_steamy_room_medium_light_skin_tone", R.drawable.emj_1f9d6_1f3fc, "\ud83e\uddd6\ud83c\udffc" );
	}
	private static void init89(){
		name( "person_in_steamy_room_medium_skin_tone", R.drawable.emj_1f9d6_1f3fd, "\ud83e\uddd6\ud83c\udffd" );
		name( "person_in_steamy_room_tone1", R.drawable.emj_1f9d6_1f3fb, "\ud83e\uddd6\ud83c\udffb" );
		name( "person_in_steamy_room_tone2", R.drawable.emj_1f9d6_1f3fc, "\ud83e\uddd6\ud83c\udffc" );
		name( "person_in_steamy_room_tone3", R.drawable.emj_1f9d6_1f3fd, "\ud83e\uddd6\ud83c\udffd" );
		name( "person_in_steamy_room_tone4", R.drawable.emj_1f9d6_1f3fe, "\ud83e\uddd6\ud83c\udffe" );
		name( "person_in_steamy_room_tone5", R.drawable.emj_1f9d6_1f3ff, "\ud83e\uddd6\ud83c\udfff" );
		name( "person_with_ball", R.drawable.emj_26f9_fe0f, "\u26f9\ufe0f" );
		name( "person_with_ball_dark_skin_tone", R.drawable.emj_26f9_1f3ff, "\u26f9\ud83c\udfff" );
		name( "person_with_ball_light_skin_tone", R.drawable.emj_26f9_1f3fb, "\u26f9\ud83c\udffb" );
		name( "person_with_ball_medium_dark_skin_tone", R.drawable.emj_26f9_1f3fe, "\u26f9\ud83c\udffe" );
		name( "person_with_ball_medium_light_skin_tone", R.drawable.emj_26f9_1f3fc, "\u26f9\ud83c\udffc" );
		name( "person_with_ball_medium_skin_tone", R.drawable.emj_26f9_1f3fd, "\u26f9\ud83c\udffd" );
		name( "person_with_ball_tone1", R.drawable.emj_26f9_1f3fb, "\u26f9\ud83c\udffb" );
		name( "person_with_ball_tone2", R.drawable.emj_26f9_1f3fc, "\u26f9\ud83c\udffc" );
		name( "person_with_ball_tone3", R.drawable.emj_26f9_1f3fd, "\u26f9\ud83c\udffd" );
		name( "person_with_ball_tone4", R.drawable.emj_26f9_1f3fe, "\u26f9\ud83c\udffe" );
		name( "person_with_ball_tone5", R.drawable.emj_26f9_1f3ff, "\u26f9\ud83c\udfff" );
		name( "person_with_blond_hair", R.drawable.emj_1f471, "\ud83d\udc71" );
		name( "person_with_blond_hair_dark_skin_tone", R.drawable.emj_1f471_1f3ff, "\ud83d\udc71\ud83c\udfff" );
		name( "person_with_blond_hair_light_skin_tone", R.drawable.emj_1f471_1f3fb, "\ud83d\udc71\ud83c\udffb" );
		name( "person_with_blond_hair_medium_dark_skin_tone", R.drawable.emj_1f471_1f3fe, "\ud83d\udc71\ud83c\udffe" );
		name( "person_with_blond_hair_medium_light_skin_tone", R.drawable.emj_1f471_1f3fc, "\ud83d\udc71\ud83c\udffc" );
		name( "person_with_blond_hair_medium_skin_tone", R.drawable.emj_1f471_1f3fd, "\ud83d\udc71\ud83c\udffd" );
		name( "person_with_blond_hair_tone1", R.drawable.emj_1f471_1f3fb, "\ud83d\udc71\ud83c\udffb" );
		name( "person_with_blond_hair_tone2", R.drawable.emj_1f471_1f3fc, "\ud83d\udc71\ud83c\udffc" );
		name( "person_with_blond_hair_tone3", R.drawable.emj_1f471_1f3fd, "\ud83d\udc71\ud83c\udffd" );
		name( "person_with_blond_hair_tone4", R.drawable.emj_1f471_1f3fe, "\ud83d\udc71\ud83c\udffe" );
		name( "person_with_blond_hair_tone5", R.drawable.emj_1f471_1f3ff, "\ud83d\udc71\ud83c\udfff" );
		name( "person_with_headscarf", R.drawable.emj_1f9d5, "\ud83e\uddd5" );
		name( "person_with_headscarf_dark_skin_tone", R.drawable.emj_1f9d5_1f3ff, "\ud83e\uddd5\ud83c\udfff" );
		name( "person_with_headscarf_light_skin_tone", R.drawable.emj_1f9d5_1f3fb, "\ud83e\uddd5\ud83c\udffb" );
		name( "person_with_headscarf_medium_dark_skin_tone", R.drawable.emj_1f9d5_1f3fe, "\ud83e\uddd5\ud83c\udffe" );
		name( "person_with_headscarf_medium_light_skin_tone", R.drawable.emj_1f9d5_1f3fc, "\ud83e\uddd5\ud83c\udffc" );
		name( "person_with_headscarf_medium_skin_tone", R.drawable.emj_1f9d5_1f3fd, "\ud83e\uddd5\ud83c\udffd" );
		name( "person_with_headscarf_tone1", R.drawable.emj_1f9d5_1f3fb, "\ud83e\uddd5\ud83c\udffb" );
		name( "person_with_headscarf_tone2", R.drawable.emj_1f9d5_1f3fc, "\ud83e\uddd5\ud83c\udffc" );
		name( "person_with_headscarf_tone3", R.drawable.emj_1f9d5_1f3fd, "\ud83e\uddd5\ud83c\udffd" );
		name( "person_with_headscarf_tone4", R.drawable.emj_1f9d5_1f3fe, "\ud83e\uddd5\ud83c\udffe" );
		name( "person_with_headscarf_tone5", R.drawable.emj_1f9d5_1f3ff, "\ud83e\uddd5\ud83c\udfff" );
		name( "person_with_pouting_face", R.drawable.emj_1f64e, "\ud83d\ude4e" );
		name( "person_with_pouting_face_dark_skin_tone", R.drawable.emj_1f64e_1f3ff, "\ud83d\ude4e\ud83c\udfff" );
		name( "person_with_pouting_face_light_skin_tone", R.drawable.emj_1f64e_1f3fb, "\ud83d\ude4e\ud83c\udffb" );
		name( "person_with_pouting_face_medium_dark_skin_tone", R.drawable.emj_1f64e_1f3fe, "\ud83d\ude4e\ud83c\udffe" );
		name( "person_with_pouting_face_medium_light_skin_tone", R.drawable.emj_1f64e_1f3fc, "\ud83d\ude4e\ud83c\udffc" );
		name( "person_with_pouting_face_medium_skin_tone", R.drawable.emj_1f64e_1f3fd, "\ud83d\ude4e\ud83c\udffd" );
		name( "person_with_pouting_face_tone1", R.drawable.emj_1f64e_1f3fb, "\ud83d\ude4e\ud83c\udffb" );
		name( "person_with_pouting_face_tone2", R.drawable.emj_1f64e_1f3fc, "\ud83d\ude4e\ud83c\udffc" );
		name( "person_with_pouting_face_tone3", R.drawable.emj_1f64e_1f3fd, "\ud83d\ude4e\ud83c\udffd" );
		name( "person_with_pouting_face_tone4", R.drawable.emj_1f64e_1f3fe, "\ud83d\ude4e\ud83c\udffe" );
		name( "person_with_pouting_face_tone5", R.drawable.emj_1f64e_1f3ff, "\ud83d\ude4e\ud83c\udfff" );
		name( "pf", R.drawable.emj_1f1f5_1f1eb, "\ud83c\uddf5\ud83c\uddeb" );
		name( "pg", R.drawable.emj_1f1f5_1f1ec, "\ud83c\uddf5\ud83c\uddec" );
		name( "ph", R.drawable.emj_1f1f5_1f1ed, "\ud83c\uddf5\ud83c\udded" );
		name( "phone", R.drawable.emj_260e_fe0f, "\u260e\ufe0f" );
		name( "pick", R.drawable.emj_26cf_fe0f, "\u26cf\ufe0f" );
		name( "pie", R.drawable.emj_1f967, "\ud83e\udd67" );
		name( "pig", R.drawable.emj_1f437, "\ud83d\udc37" );
		name( "pig2", R.drawable.emj_1f416, "\ud83d\udc16" );
		name( "pig_nose", R.drawable.emj_1f43d, "\ud83d\udc3d" );
		name( "pill", R.drawable.emj_1f48a, "\ud83d\udc8a" );
		name( "pineapple", R.drawable.emj_1f34d, "\ud83c\udf4d" );
		name( "ping_pong", R.drawable.emj_1f3d3, "\ud83c\udfd3" );
		name( "pisces", R.drawable.emj_2653, "\u2653" );
		name( "pizza", R.drawable.emj_1f355, "\ud83c\udf55" );
		name( "pk", R.drawable.emj_1f1f5_1f1f0, "\ud83c\uddf5\ud83c\uddf0" );
		name( "pl", R.drawable.emj_1f1f5_1f1f1, "\ud83c\uddf5\ud83c\uddf1" );
		name( "place_of_worship", R.drawable.emj_1f6d0, "\ud83d\uded0" );
		name( "play_pause", R.drawable.emj_23ef_fe0f, "\u23ef\ufe0f" );
		name( "pm", R.drawable.emj_1f1f5_1f1f2, "\ud83c\uddf5\ud83c\uddf2" );
		name( "pn", R.drawable.emj_1f1f5_1f1f3, "\ud83c\uddf5\ud83c\uddf3" );
		name( "point_down", R.drawable.emj_1f447, "\ud83d\udc47" );
		name( "point_down_dark_skin_tone", R.drawable.emj_1f447_1f3ff, "\ud83d\udc47\ud83c\udfff" );
		name( "point_down_light_skin_tone", R.drawable.emj_1f447_1f3fb, "\ud83d\udc47\ud83c\udffb" );
		name( "point_down_medium_dark_skin_tone", R.drawable.emj_1f447_1f3fe, "\ud83d\udc47\ud83c\udffe" );
		name( "point_down_medium_light_skin_tone", R.drawable.emj_1f447_1f3fc, "\ud83d\udc47\ud83c\udffc" );
		name( "point_down_medium_skin_tone", R.drawable.emj_1f447_1f3fd, "\ud83d\udc47\ud83c\udffd" );
		name( "point_down_tone1", R.drawable.emj_1f447_1f3fb, "\ud83d\udc47\ud83c\udffb" );
		name( "point_down_tone2", R.drawable.emj_1f447_1f3fc, "\ud83d\udc47\ud83c\udffc" );
		name( "point_down_tone3", R.drawable.emj_1f447_1f3fd, "\ud83d\udc47\ud83c\udffd" );
		name( "point_down_tone4", R.drawable.emj_1f447_1f3fe, "\ud83d\udc47\ud83c\udffe" );
		name( "point_down_tone5", R.drawable.emj_1f447_1f3ff, "\ud83d\udc47\ud83c\udfff" );
		name( "point_left", R.drawable.emj_1f448, "\ud83d\udc48" );
		name( "point_left_dark_skin_tone", R.drawable.emj_1f448_1f3ff, "\ud83d\udc48\ud83c\udfff" );
		name( "point_left_light_skin_tone", R.drawable.emj_1f448_1f3fb, "\ud83d\udc48\ud83c\udffb" );
		name( "point_left_medium_dark_skin_tone", R.drawable.emj_1f448_1f3fe, "\ud83d\udc48\ud83c\udffe" );
		name( "point_left_medium_light_skin_tone", R.drawable.emj_1f448_1f3fc, "\ud83d\udc48\ud83c\udffc" );
		name( "point_left_medium_skin_tone", R.drawable.emj_1f448_1f3fd, "\ud83d\udc48\ud83c\udffd" );
		name( "point_left_tone1", R.drawable.emj_1f448_1f3fb, "\ud83d\udc48\ud83c\udffb" );
		name( "point_left_tone2", R.drawable.emj_1f448_1f3fc, "\ud83d\udc48\ud83c\udffc" );
		name( "point_left_tone3", R.drawable.emj_1f448_1f3fd, "\ud83d\udc48\ud83c\udffd" );
		name( "point_left_tone4", R.drawable.emj_1f448_1f3fe, "\ud83d\udc48\ud83c\udffe" );
		name( "point_left_tone5", R.drawable.emj_1f448_1f3ff, "\ud83d\udc48\ud83c\udfff" );
		name( "point_right", R.drawable.emj_1f449, "\ud83d\udc49" );
		name( "point_right_dark_skin_tone", R.drawable.emj_1f449_1f3ff, "\ud83d\udc49\ud83c\udfff" );
		name( "point_right_light_skin_tone", R.drawable.emj_1f449_1f3fb, "\ud83d\udc49\ud83c\udffb" );
		name( "point_right_medium_dark_skin_tone", R.drawable.emj_1f449_1f3fe, "\ud83d\udc49\ud83c\udffe" );
		name( "point_right_medium_light_skin_tone", R.drawable.emj_1f449_1f3fc, "\ud83d\udc49\ud83c\udffc" );
		name( "point_right_medium_skin_tone", R.drawable.emj_1f449_1f3fd, "\ud83d\udc49\ud83c\udffd" );
		name( "point_right_tone1", R.drawable.emj_1f449_1f3fb, "\ud83d\udc49\ud83c\udffb" );
		name( "point_right_tone2", R.drawable.emj_1f449_1f3fc, "\ud83d\udc49\ud83c\udffc" );
		name( "point_right_tone3", R.drawable.emj_1f449_1f3fd, "\ud83d\udc49\ud83c\udffd" );
	}
	private static void init90(){
		name( "point_right_tone4", R.drawable.emj_1f449_1f3fe, "\ud83d\udc49\ud83c\udffe" );
		name( "point_right_tone5", R.drawable.emj_1f449_1f3ff, "\ud83d\udc49\ud83c\udfff" );
		name( "point_up", R.drawable.emj_261d_fe0f, "\u261d\ufe0f" );
		name( "point_up_2", R.drawable.emj_1f446, "\ud83d\udc46" );
		name( "point_up_2_dark_skin_tone", R.drawable.emj_1f446_1f3ff, "\ud83d\udc46\ud83c\udfff" );
		name( "point_up_2_light_skin_tone", R.drawable.emj_1f446_1f3fb, "\ud83d\udc46\ud83c\udffb" );
		name( "point_up_2_medium_dark_skin_tone", R.drawable.emj_1f446_1f3fe, "\ud83d\udc46\ud83c\udffe" );
		name( "point_up_2_medium_light_skin_tone", R.drawable.emj_1f446_1f3fc, "\ud83d\udc46\ud83c\udffc" );
		name( "point_up_2_medium_skin_tone", R.drawable.emj_1f446_1f3fd, "\ud83d\udc46\ud83c\udffd" );
		name( "point_up_2_tone1", R.drawable.emj_1f446_1f3fb, "\ud83d\udc46\ud83c\udffb" );
		name( "point_up_2_tone2", R.drawable.emj_1f446_1f3fc, "\ud83d\udc46\ud83c\udffc" );
		name( "point_up_2_tone3", R.drawable.emj_1f446_1f3fd, "\ud83d\udc46\ud83c\udffd" );
		name( "point_up_2_tone4", R.drawable.emj_1f446_1f3fe, "\ud83d\udc46\ud83c\udffe" );
		name( "point_up_2_tone5", R.drawable.emj_1f446_1f3ff, "\ud83d\udc46\ud83c\udfff" );
		name( "point_up_dark_skin_tone", R.drawable.emj_261d_1f3ff, "\u261d\ud83c\udfff" );
		name( "point_up_light_skin_tone", R.drawable.emj_261d_1f3fb, "\u261d\ud83c\udffb" );
		name( "point_up_medium_dark_skin_tone", R.drawable.emj_261d_1f3fe, "\u261d\ud83c\udffe" );
		name( "point_up_medium_light_skin_tone", R.drawable.emj_261d_1f3fc, "\u261d\ud83c\udffc" );
		name( "point_up_medium_skin_tone", R.drawable.emj_261d_1f3fd, "\u261d\ud83c\udffd" );
		name( "point_up_tone1", R.drawable.emj_261d_1f3fb, "\u261d\ud83c\udffb" );
		name( "point_up_tone2", R.drawable.emj_261d_1f3fc, "\u261d\ud83c\udffc" );
		name( "point_up_tone3", R.drawable.emj_261d_1f3fd, "\u261d\ud83c\udffd" );
		name( "point_up_tone4", R.drawable.emj_261d_1f3fe, "\u261d\ud83c\udffe" );
		name( "point_up_tone5", R.drawable.emj_261d_1f3ff, "\u261d\ud83c\udfff" );
		name( "police_car", R.drawable.emj_1f693, "\ud83d\ude93" );
		name( "poo", R.drawable.emj_1f4a9, "\ud83d\udca9" );
		name( "poodle", R.drawable.emj_1f429, "\ud83d\udc29" );
		name( "poop", R.drawable.emj_1f4a9, "\ud83d\udca9" );
		name( "popcorn", R.drawable.emj_1f37f, "\ud83c\udf7f" );
		name( "post_office", R.drawable.emj_1f3e3, "\ud83c\udfe3" );
		name( "postal_horn", R.drawable.emj_1f4ef, "\ud83d\udcef" );
		name( "postbox", R.drawable.emj_1f4ee, "\ud83d\udcee" );
		name( "potable_water", R.drawable.emj_1f6b0, "\ud83d\udeb0" );
		name( "potato", R.drawable.emj_1f954, "\ud83e\udd54" );
		name( "pouch", R.drawable.emj_1f45d, "\ud83d\udc5d" );
		name( "poultry_leg", R.drawable.emj_1f357, "\ud83c\udf57" );
		name( "pound", R.drawable.emj_1f4b7, "\ud83d\udcb7" );
		name( "pound_symbol", R.drawable.emj_0023, "\u0023" );
		name( "pouting_cat", R.drawable.emj_1f63e, "\ud83d\ude3e" );
		name( "pr", R.drawable.emj_1f1f5_1f1f7, "\ud83c\uddf5\ud83c\uddf7" );
		name( "pray", R.drawable.emj_1f64f, "\ud83d\ude4f" );
		name( "pray_dark_skin_tone", R.drawable.emj_1f64f_1f3ff, "\ud83d\ude4f\ud83c\udfff" );
		name( "pray_light_skin_tone", R.drawable.emj_1f64f_1f3fb, "\ud83d\ude4f\ud83c\udffb" );
		name( "pray_medium_dark_skin_tone", R.drawable.emj_1f64f_1f3fe, "\ud83d\ude4f\ud83c\udffe" );
		name( "pray_medium_light_skin_tone", R.drawable.emj_1f64f_1f3fc, "\ud83d\ude4f\ud83c\udffc" );
		name( "pray_medium_skin_tone", R.drawable.emj_1f64f_1f3fd, "\ud83d\ude4f\ud83c\udffd" );
		name( "pray_tone1", R.drawable.emj_1f64f_1f3fb, "\ud83d\ude4f\ud83c\udffb" );
		name( "pray_tone2", R.drawable.emj_1f64f_1f3fc, "\ud83d\ude4f\ud83c\udffc" );
		name( "pray_tone3", R.drawable.emj_1f64f_1f3fd, "\ud83d\ude4f\ud83c\udffd" );
		name( "pray_tone4", R.drawable.emj_1f64f_1f3fe, "\ud83d\ude4f\ud83c\udffe" );
		name( "pray_tone5", R.drawable.emj_1f64f_1f3ff, "\ud83d\ude4f\ud83c\udfff" );
		name( "prayer_beads", R.drawable.emj_1f4ff, "\ud83d\udcff" );
		name( "pregnant_woman", R.drawable.emj_1f930, "\ud83e\udd30" );
		name( "pregnant_woman_dark_skin_tone", R.drawable.emj_1f930_1f3ff, "\ud83e\udd30\ud83c\udfff" );
		name( "pregnant_woman_light_skin_tone", R.drawable.emj_1f930_1f3fb, "\ud83e\udd30\ud83c\udffb" );
		name( "pregnant_woman_medium_dark_skin_tone", R.drawable.emj_1f930_1f3fe, "\ud83e\udd30\ud83c\udffe" );
		name( "pregnant_woman_medium_light_skin_tone", R.drawable.emj_1f930_1f3fc, "\ud83e\udd30\ud83c\udffc" );
		name( "pregnant_woman_medium_skin_tone", R.drawable.emj_1f930_1f3fd, "\ud83e\udd30\ud83c\udffd" );
		name( "pregnant_woman_tone1", R.drawable.emj_1f930_1f3fb, "\ud83e\udd30\ud83c\udffb" );
		name( "pregnant_woman_tone2", R.drawable.emj_1f930_1f3fc, "\ud83e\udd30\ud83c\udffc" );
		name( "pregnant_woman_tone3", R.drawable.emj_1f930_1f3fd, "\ud83e\udd30\ud83c\udffd" );
		name( "pregnant_woman_tone4", R.drawable.emj_1f930_1f3fe, "\ud83e\udd30\ud83c\udffe" );
		name( "pregnant_woman_tone5", R.drawable.emj_1f930_1f3ff, "\ud83e\udd30\ud83c\udfff" );
		name( "pretzel", R.drawable.emj_1f968, "\ud83e\udd68" );
		name( "previous_track", R.drawable.emj_23ee_fe0f, "\u23ee\ufe0f" );
		name( "prince", R.drawable.emj_1f934, "\ud83e\udd34" );
		name( "prince_dark_skin_tone", R.drawable.emj_1f934_1f3ff, "\ud83e\udd34\ud83c\udfff" );
		name( "prince_light_skin_tone", R.drawable.emj_1f934_1f3fb, "\ud83e\udd34\ud83c\udffb" );
		name( "prince_medium_dark_skin_tone", R.drawable.emj_1f934_1f3fe, "\ud83e\udd34\ud83c\udffe" );
		name( "prince_medium_light_skin_tone", R.drawable.emj_1f934_1f3fc, "\ud83e\udd34\ud83c\udffc" );
		name( "prince_medium_skin_tone", R.drawable.emj_1f934_1f3fd, "\ud83e\udd34\ud83c\udffd" );
		name( "prince_tone1", R.drawable.emj_1f934_1f3fb, "\ud83e\udd34\ud83c\udffb" );
		name( "prince_tone2", R.drawable.emj_1f934_1f3fc, "\ud83e\udd34\ud83c\udffc" );
		name( "prince_tone3", R.drawable.emj_1f934_1f3fd, "\ud83e\udd34\ud83c\udffd" );
		name( "prince_tone4", R.drawable.emj_1f934_1f3fe, "\ud83e\udd34\ud83c\udffe" );
		name( "prince_tone5", R.drawable.emj_1f934_1f3ff, "\ud83e\udd34\ud83c\udfff" );
		name( "princess", R.drawable.emj_1f478, "\ud83d\udc78" );
		name( "princess_dark_skin_tone", R.drawable.emj_1f478_1f3ff, "\ud83d\udc78\ud83c\udfff" );
		name( "princess_light_skin_tone", R.drawable.emj_1f478_1f3fb, "\ud83d\udc78\ud83c\udffb" );
		name( "princess_medium_dark_skin_tone", R.drawable.emj_1f478_1f3fe, "\ud83d\udc78\ud83c\udffe" );
		name( "princess_medium_light_skin_tone", R.drawable.emj_1f478_1f3fc, "\ud83d\udc78\ud83c\udffc" );
		name( "princess_medium_skin_tone", R.drawable.emj_1f478_1f3fd, "\ud83d\udc78\ud83c\udffd" );
		name( "princess_tone1", R.drawable.emj_1f478_1f3fb, "\ud83d\udc78\ud83c\udffb" );
		name( "princess_tone2", R.drawable.emj_1f478_1f3fc, "\ud83d\udc78\ud83c\udffc" );
		name( "princess_tone3", R.drawable.emj_1f478_1f3fd, "\ud83d\udc78\ud83c\udffd" );
		name( "princess_tone4", R.drawable.emj_1f478_1f3fe, "\ud83d\udc78\ud83c\udffe" );
		name( "princess_tone5", R.drawable.emj_1f478_1f3ff, "\ud83d\udc78\ud83c\udfff" );
		name( "printer", R.drawable.emj_1f5a8_fe0f, "\ud83d\udda8\ufe0f" );
		name( "projector", R.drawable.emj_1f4fd_fe0f, "\ud83d\udcfd\ufe0f" );
		name( "ps", R.drawable.emj_1f1f5_1f1f8, "\ud83c\uddf5\ud83c\uddf8" );
		name( "pt", R.drawable.emj_1f1f5_1f1f9, "\ud83c\uddf5\ud83c\uddf9" );
		name( "pudding", R.drawable.emj_1f36e, "\ud83c\udf6e" );
		name( "punch", R.drawable.emj_1f44a, "\ud83d\udc4a" );
		name( "punch_dark_skin_tone", R.drawable.emj_1f44a_1f3ff, "\ud83d\udc4a\ud83c\udfff" );
		name( "punch_light_skin_tone", R.drawable.emj_1f44a_1f3fb, "\ud83d\udc4a\ud83c\udffb" );
		name( "punch_medium_dark_skin_tone", R.drawable.emj_1f44a_1f3fe, "\ud83d\udc4a\ud83c\udffe" );
		name( "punch_medium_light_skin_tone", R.drawable.emj_1f44a_1f3fc, "\ud83d\udc4a\ud83c\udffc" );
		name( "punch_medium_skin_tone", R.drawable.emj_1f44a_1f3fd, "\ud83d\udc4a\ud83c\udffd" );
		name( "punch_tone1", R.drawable.emj_1f44a_1f3fb, "\ud83d\udc4a\ud83c\udffb" );
		name( "punch_tone2", R.drawable.emj_1f44a_1f3fc, "\ud83d\udc4a\ud83c\udffc" );
		name( "punch_tone3", R.drawable.emj_1f44a_1f3fd, "\ud83d\udc4a\ud83c\udffd" );
	}
	private static void init91(){
		name( "punch_tone4", R.drawable.emj_1f44a_1f3fe, "\ud83d\udc4a\ud83c\udffe" );
		name( "punch_tone5", R.drawable.emj_1f44a_1f3ff, "\ud83d\udc4a\ud83c\udfff" );
		name( "purple_heart", R.drawable.emj_1f49c, "\ud83d\udc9c" );
		name( "purse", R.drawable.emj_1f45b, "\ud83d\udc5b" );
		name( "pushpin", R.drawable.emj_1f4cc, "\ud83d\udccc" );
		name( "put_litter_in_its_place", R.drawable.emj_1f6ae, "\ud83d\udeae" );
		name( "pw", R.drawable.emj_1f1f5_1f1fc, "\ud83c\uddf5\ud83c\uddfc" );
		name( "py", R.drawable.emj_1f1f5_1f1fe, "\ud83c\uddf5\ud83c\uddfe" );
		name( "qa", R.drawable.emj_1f1f6_1f1e6, "\ud83c\uddf6\ud83c\udde6" );
		name( "question", R.drawable.emj_2753, "\u2753" );
		name( "rabbit", R.drawable.emj_1f430, "\ud83d\udc30" );
		name( "rabbit2", R.drawable.emj_1f407, "\ud83d\udc07" );
		name( "race_car", R.drawable.emj_1f3ce_fe0f, "\ud83c\udfce\ufe0f" );
		name( "racehorse", R.drawable.emj_1f40e, "\ud83d\udc0e" );
		name( "racing_car", R.drawable.emj_1f3ce_fe0f, "\ud83c\udfce\ufe0f" );
		name( "racing_motorcycle", R.drawable.emj_1f3cd_fe0f, "\ud83c\udfcd\ufe0f" );
		name( "radio", R.drawable.emj_1f4fb, "\ud83d\udcfb" );
		name( "radio_button", R.drawable.emj_1f518, "\ud83d\udd18" );
		name( "radioactive", R.drawable.emj_2622_fe0f, "\u2622\ufe0f" );
		name( "radioactive_sign", R.drawable.emj_2622_fe0f, "\u2622\ufe0f" );
		name( "rage", R.drawable.emj_1f621, "\ud83d\ude21" );
		name( "railroad_track", R.drawable.emj_1f6e4_fe0f, "\ud83d\udee4\ufe0f" );
		name( "railway_car", R.drawable.emj_1f683, "\ud83d\ude83" );
		name( "railway_track", R.drawable.emj_1f6e4_fe0f, "\ud83d\udee4\ufe0f" );
		name( "rain_cloud", R.drawable.emj_1f327_fe0f, "\ud83c\udf27\ufe0f" );
		name( "rainbow", R.drawable.emj_1f308, "\ud83c\udf08" );
		name( "rainbow_flag", R.drawable.emj_1f3f3_fe0f_200d_1f308, "\ud83c\udff3\ufe0f\u200d\ud83c\udf08" );
		name( "raised_back_of_hand", R.drawable.emj_1f91a, "\ud83e\udd1a" );
		name( "raised_back_of_hand_dark_skin_tone", R.drawable.emj_1f91a_1f3ff, "\ud83e\udd1a\ud83c\udfff" );
		name( "raised_back_of_hand_light_skin_tone", R.drawable.emj_1f91a_1f3fb, "\ud83e\udd1a\ud83c\udffb" );
		name( "raised_back_of_hand_medium_dark_skin_tone", R.drawable.emj_1f91a_1f3fe, "\ud83e\udd1a\ud83c\udffe" );
		name( "raised_back_of_hand_medium_light_skin_tone", R.drawable.emj_1f91a_1f3fc, "\ud83e\udd1a\ud83c\udffc" );
		name( "raised_back_of_hand_medium_skin_tone", R.drawable.emj_1f91a_1f3fd, "\ud83e\udd1a\ud83c\udffd" );
		name( "raised_back_of_hand_tone1", R.drawable.emj_1f91a_1f3fb, "\ud83e\udd1a\ud83c\udffb" );
		name( "raised_back_of_hand_tone2", R.drawable.emj_1f91a_1f3fc, "\ud83e\udd1a\ud83c\udffc" );
		name( "raised_back_of_hand_tone3", R.drawable.emj_1f91a_1f3fd, "\ud83e\udd1a\ud83c\udffd" );
		name( "raised_back_of_hand_tone4", R.drawable.emj_1f91a_1f3fe, "\ud83e\udd1a\ud83c\udffe" );
		name( "raised_back_of_hand_tone5", R.drawable.emj_1f91a_1f3ff, "\ud83e\udd1a\ud83c\udfff" );
		name( "raised_hand", R.drawable.emj_270b, "\u270b" );
		name( "raised_hand_dark_skin_tone", R.drawable.emj_270b_1f3ff, "\u270b\ud83c\udfff" );
		name( "raised_hand_light_skin_tone", R.drawable.emj_270b_1f3fb, "\u270b\ud83c\udffb" );
		name( "raised_hand_medium_dark_skin_tone", R.drawable.emj_270b_1f3fe, "\u270b\ud83c\udffe" );
		name( "raised_hand_medium_light_skin_tone", R.drawable.emj_270b_1f3fc, "\u270b\ud83c\udffc" );
		name( "raised_hand_medium_skin_tone", R.drawable.emj_270b_1f3fd, "\u270b\ud83c\udffd" );
		name( "raised_hand_tone1", R.drawable.emj_270b_1f3fb, "\u270b\ud83c\udffb" );
		name( "raised_hand_tone2", R.drawable.emj_270b_1f3fc, "\u270b\ud83c\udffc" );
		name( "raised_hand_tone3", R.drawable.emj_270b_1f3fd, "\u270b\ud83c\udffd" );
		name( "raised_hand_tone4", R.drawable.emj_270b_1f3fe, "\u270b\ud83c\udffe" );
		name( "raised_hand_tone5", R.drawable.emj_270b_1f3ff, "\u270b\ud83c\udfff" );
		name( "raised_hand_with_fingers_splayed", R.drawable.emj_1f590_fe0f, "\ud83d\udd90\ufe0f" );
		name( "raised_hand_with_fingers_splayed_dark_skin_tone", R.drawable.emj_1f590_1f3ff, "\ud83d\udd90\ud83c\udfff" );
		name( "raised_hand_with_fingers_splayed_light_skin_tone", R.drawable.emj_1f590_1f3fb, "\ud83d\udd90\ud83c\udffb" );
		name( "raised_hand_with_fingers_splayed_medium_dark_skin_tone", R.drawable.emj_1f590_1f3fe, "\ud83d\udd90\ud83c\udffe" );
		name( "raised_hand_with_fingers_splayed_medium_light_skin_tone", R.drawable.emj_1f590_1f3fc, "\ud83d\udd90\ud83c\udffc" );
		name( "raised_hand_with_fingers_splayed_medium_skin_tone", R.drawable.emj_1f590_1f3fd, "\ud83d\udd90\ud83c\udffd" );
		name( "raised_hand_with_fingers_splayed_tone1", R.drawable.emj_1f590_1f3fb, "\ud83d\udd90\ud83c\udffb" );
		name( "raised_hand_with_fingers_splayed_tone2", R.drawable.emj_1f590_1f3fc, "\ud83d\udd90\ud83c\udffc" );
		name( "raised_hand_with_fingers_splayed_tone3", R.drawable.emj_1f590_1f3fd, "\ud83d\udd90\ud83c\udffd" );
		name( "raised_hand_with_fingers_splayed_tone4", R.drawable.emj_1f590_1f3fe, "\ud83d\udd90\ud83c\udffe" );
		name( "raised_hand_with_fingers_splayed_tone5", R.drawable.emj_1f590_1f3ff, "\ud83d\udd90\ud83c\udfff" );
		name( "raised_hand_with_part_between_middle_and_ring_fingers", R.drawable.emj_1f596, "\ud83d\udd96" );
		name( "raised_hand_with_part_between_middle_and_ring_fingers_tone1", R.drawable.emj_1f596_1f3fb, "\ud83d\udd96\ud83c\udffb" );
		name( "raised_hand_with_part_between_middle_and_ring_fingers_tone2", R.drawable.emj_1f596_1f3fc, "\ud83d\udd96\ud83c\udffc" );
		name( "raised_hand_with_part_between_middle_and_ring_fingers_tone3", R.drawable.emj_1f596_1f3fd, "\ud83d\udd96\ud83c\udffd" );
		name( "raised_hand_with_part_between_middle_and_ring_fingers_tone4", R.drawable.emj_1f596_1f3fe, "\ud83d\udd96\ud83c\udffe" );
		name( "raised_hand_with_part_between_middle_and_ring_fingers_tone5", R.drawable.emj_1f596_1f3ff, "\ud83d\udd96\ud83c\udfff" );
		name( "raised_hands", R.drawable.emj_1f64c, "\ud83d\ude4c" );
		name( "raised_hands_dark_skin_tone", R.drawable.emj_1f64c_1f3ff, "\ud83d\ude4c\ud83c\udfff" );
		name( "raised_hands_light_skin_tone", R.drawable.emj_1f64c_1f3fb, "\ud83d\ude4c\ud83c\udffb" );
		name( "raised_hands_medium_dark_skin_tone", R.drawable.emj_1f64c_1f3fe, "\ud83d\ude4c\ud83c\udffe" );
		name( "raised_hands_medium_light_skin_tone", R.drawable.emj_1f64c_1f3fc, "\ud83d\ude4c\ud83c\udffc" );
		name( "raised_hands_medium_skin_tone", R.drawable.emj_1f64c_1f3fd, "\ud83d\ude4c\ud83c\udffd" );
		name( "raised_hands_tone1", R.drawable.emj_1f64c_1f3fb, "\ud83d\ude4c\ud83c\udffb" );
		name( "raised_hands_tone2", R.drawable.emj_1f64c_1f3fc, "\ud83d\ude4c\ud83c\udffc" );
		name( "raised_hands_tone3", R.drawable.emj_1f64c_1f3fd, "\ud83d\ude4c\ud83c\udffd" );
		name( "raised_hands_tone4", R.drawable.emj_1f64c_1f3fe, "\ud83d\ude4c\ud83c\udffe" );
		name( "raised_hands_tone5", R.drawable.emj_1f64c_1f3ff, "\ud83d\ude4c\ud83c\udfff" );
		name( "raising_hand", R.drawable.emj_1f64b, "\ud83d\ude4b" );
		name( "raising_hand_dark_skin_tone", R.drawable.emj_1f64b_1f3ff, "\ud83d\ude4b\ud83c\udfff" );
		name( "raising_hand_light_skin_tone", R.drawable.emj_1f64b_1f3fb, "\ud83d\ude4b\ud83c\udffb" );
		name( "raising_hand_medium_dark_skin_tone", R.drawable.emj_1f64b_1f3fe, "\ud83d\ude4b\ud83c\udffe" );
		name( "raising_hand_medium_light_skin_tone", R.drawable.emj_1f64b_1f3fc, "\ud83d\ude4b\ud83c\udffc" );
		name( "raising_hand_medium_skin_tone", R.drawable.emj_1f64b_1f3fd, "\ud83d\ude4b\ud83c\udffd" );
		name( "raising_hand_tone1", R.drawable.emj_1f64b_1f3fb, "\ud83d\ude4b\ud83c\udffb" );
		name( "raising_hand_tone2", R.drawable.emj_1f64b_1f3fc, "\ud83d\ude4b\ud83c\udffc" );
		name( "raising_hand_tone3", R.drawable.emj_1f64b_1f3fd, "\ud83d\ude4b\ud83c\udffd" );
		name( "raising_hand_tone4", R.drawable.emj_1f64b_1f3fe, "\ud83d\ude4b\ud83c\udffe" );
		name( "raising_hand_tone5", R.drawable.emj_1f64b_1f3ff, "\ud83d\ude4b\ud83c\udfff" );
		name( "ram", R.drawable.emj_1f40f, "\ud83d\udc0f" );
		name( "ramen", R.drawable.emj_1f35c, "\ud83c\udf5c" );
		name( "rat", R.drawable.emj_1f400, "\ud83d\udc00" );
		name( "re", R.drawable.emj_1f1f7_1f1ea, "\ud83c\uddf7\ud83c\uddea" );
		name( "record_button", R.drawable.emj_23fa_fe0f, "\u23fa\ufe0f" );
		name( "recycle", R.drawable.emj_267b_fe0f, "\u267b\ufe0f" );
		name( "red_car", R.drawable.emj_1f697, "\ud83d\ude97" );
		name( "red_circle", R.drawable.emj_1f534, "\ud83d\udd34" );
		name( "regional_indicator_a", R.drawable.emj_1f1e6, "\ud83c\udde6" );
		name( "regional_indicator_b", R.drawable.emj_1f1e7, "\ud83c\udde7" );
		name( "regional_indicator_c", R.drawable.emj_1f1e8, "\ud83c\udde8" );
		name( "regional_indicator_d", R.drawable.emj_1f1e9, "\ud83c\udde9" );
		name( "regional_indicator_e", R.drawable.emj_1f1ea, "\ud83c\uddea" );
	}
	private static void init92(){
		name( "regional_indicator_f", R.drawable.emj_1f1eb, "\ud83c\uddeb" );
		name( "regional_indicator_g", R.drawable.emj_1f1ec, "\ud83c\uddec" );
		name( "regional_indicator_h", R.drawable.emj_1f1ed, "\ud83c\udded" );
		name( "regional_indicator_i", R.drawable.emj_1f1ee, "\ud83c\uddee" );
		name( "regional_indicator_j", R.drawable.emj_1f1ef, "\ud83c\uddef" );
		name( "regional_indicator_k", R.drawable.emj_1f1f0, "\ud83c\uddf0" );
		name( "regional_indicator_l", R.drawable.emj_1f1f1, "\ud83c\uddf1" );
		name( "regional_indicator_m", R.drawable.emj_1f1f2, "\ud83c\uddf2" );
		name( "regional_indicator_n", R.drawable.emj_1f1f3, "\ud83c\uddf3" );
		name( "regional_indicator_o", R.drawable.emj_1f1f4, "\ud83c\uddf4" );
		name( "regional_indicator_p", R.drawable.emj_1f1f5, "\ud83c\uddf5" );
		name( "regional_indicator_q", R.drawable.emj_1f1f6, "\ud83c\uddf6" );
		name( "regional_indicator_r", R.drawable.emj_1f1f7, "\ud83c\uddf7" );
		name( "regional_indicator_s", R.drawable.emj_1f1f8, "\ud83c\uddf8" );
		name( "regional_indicator_t", R.drawable.emj_1f1f9, "\ud83c\uddf9" );
		name( "regional_indicator_u", R.drawable.emj_1f1fa, "\ud83c\uddfa" );
		name( "regional_indicator_v", R.drawable.emj_1f1fb, "\ud83c\uddfb" );
		name( "regional_indicator_w", R.drawable.emj_1f1fc, "\ud83c\uddfc" );
		name( "regional_indicator_x", R.drawable.emj_1f1fd, "\ud83c\uddfd" );
		name( "regional_indicator_y", R.drawable.emj_1f1fe, "\ud83c\uddfe" );
		name( "regional_indicator_z", R.drawable.emj_1f1ff, "\ud83c\uddff" );
		name( "registered", R.drawable.emj_00ae_fe0f, "\u00ae\ufe0f" );
		name( "relaxed", R.drawable.emj_263a_fe0f, "\u263a\ufe0f" );
		name( "relieved", R.drawable.emj_1f60c, "\ud83d\ude0c" );
		name( "reminder_ribbon", R.drawable.emj_1f397_fe0f, "\ud83c\udf97\ufe0f" );
		name( "repeat", R.drawable.emj_1f501, "\ud83d\udd01" );
		name( "repeat_one", R.drawable.emj_1f502, "\ud83d\udd02" );
		name( "restroom", R.drawable.emj_1f6bb, "\ud83d\udebb" );
		name( "reversed_hand_with_middle_finger_extended", R.drawable.emj_1f595, "\ud83d\udd95" );
		name( "reversed_hand_with_middle_finger_extended_dark_skin_tone", R.drawable.emj_1f595_1f3ff, "\ud83d\udd95\ud83c\udfff" );
		name( "reversed_hand_with_middle_finger_extended_light_skin_tone", R.drawable.emj_1f595_1f3fb, "\ud83d\udd95\ud83c\udffb" );
		name( "reversed_hand_with_middle_finger_extended_medium_dark_skin_tone", R.drawable.emj_1f595_1f3fe, "\ud83d\udd95\ud83c\udffe" );
		name( "reversed_hand_with_middle_finger_extended_medium_light_skin_tone", R.drawable.emj_1f595_1f3fc, "\ud83d\udd95\ud83c\udffc" );
		name( "reversed_hand_with_middle_finger_extended_medium_skin_tone", R.drawable.emj_1f595_1f3fd, "\ud83d\udd95\ud83c\udffd" );
		name( "reversed_hand_with_middle_finger_extended_tone1", R.drawable.emj_1f595_1f3fb, "\ud83d\udd95\ud83c\udffb" );
		name( "reversed_hand_with_middle_finger_extended_tone2", R.drawable.emj_1f595_1f3fc, "\ud83d\udd95\ud83c\udffc" );
		name( "reversed_hand_with_middle_finger_extended_tone3", R.drawable.emj_1f595_1f3fd, "\ud83d\udd95\ud83c\udffd" );
		name( "reversed_hand_with_middle_finger_extended_tone4", R.drawable.emj_1f595_1f3fe, "\ud83d\udd95\ud83c\udffe" );
		name( "reversed_hand_with_middle_finger_extended_tone5", R.drawable.emj_1f595_1f3ff, "\ud83d\udd95\ud83c\udfff" );
		name( "revolving_hearts", R.drawable.emj_1f49e, "\ud83d\udc9e" );
		name( "rewind", R.drawable.emj_23ea, "\u23ea" );
		name( "rhino", R.drawable.emj_1f98f, "\ud83e\udd8f" );
		name( "rhinoceros", R.drawable.emj_1f98f, "\ud83e\udd8f" );
		name( "ribbon", R.drawable.emj_1f380, "\ud83c\udf80" );
		name( "rice", R.drawable.emj_1f35a, "\ud83c\udf5a" );
		name( "rice_ball", R.drawable.emj_1f359, "\ud83c\udf59" );
		name( "rice_cracker", R.drawable.emj_1f358, "\ud83c\udf58" );
		name( "rice_scene", R.drawable.emj_1f391, "\ud83c\udf91" );
		name( "right_anger_bubble", R.drawable.emj_1f5ef_fe0f, "\ud83d\uddef\ufe0f" );
		name( "right_facing_fist", R.drawable.emj_1f91c, "\ud83e\udd1c" );
		name( "right_facing_fist_dark_skin_tone", R.drawable.emj_1f91c_1f3ff, "\ud83e\udd1c\ud83c\udfff" );
		name( "right_facing_fist_light_skin_tone", R.drawable.emj_1f91c_1f3fb, "\ud83e\udd1c\ud83c\udffb" );
		name( "right_facing_fist_medium_dark_skin_tone", R.drawable.emj_1f91c_1f3fe, "\ud83e\udd1c\ud83c\udffe" );
		name( "right_facing_fist_medium_light_skin_tone", R.drawable.emj_1f91c_1f3fc, "\ud83e\udd1c\ud83c\udffc" );
		name( "right_facing_fist_medium_skin_tone", R.drawable.emj_1f91c_1f3fd, "\ud83e\udd1c\ud83c\udffd" );
		name( "right_facing_fist_tone1", R.drawable.emj_1f91c_1f3fb, "\ud83e\udd1c\ud83c\udffb" );
		name( "right_facing_fist_tone2", R.drawable.emj_1f91c_1f3fc, "\ud83e\udd1c\ud83c\udffc" );
		name( "right_facing_fist_tone3", R.drawable.emj_1f91c_1f3fd, "\ud83e\udd1c\ud83c\udffd" );
		name( "right_facing_fist_tone4", R.drawable.emj_1f91c_1f3fe, "\ud83e\udd1c\ud83c\udffe" );
		name( "right_facing_fist_tone5", R.drawable.emj_1f91c_1f3ff, "\ud83e\udd1c\ud83c\udfff" );
		name( "right_fist", R.drawable.emj_1f91c, "\ud83e\udd1c" );
		name( "right_fist_tone1", R.drawable.emj_1f91c_1f3fb, "\ud83e\udd1c\ud83c\udffb" );
		name( "right_fist_tone2", R.drawable.emj_1f91c_1f3fc, "\ud83e\udd1c\ud83c\udffc" );
		name( "right_fist_tone3", R.drawable.emj_1f91c_1f3fd, "\ud83e\udd1c\ud83c\udffd" );
		name( "right_fist_tone4", R.drawable.emj_1f91c_1f3fe, "\ud83e\udd1c\ud83c\udffe" );
		name( "right_fist_tone5", R.drawable.emj_1f91c_1f3ff, "\ud83e\udd1c\ud83c\udfff" );
		name( "ring", R.drawable.emj_1f48d, "\ud83d\udc8d" );
		name( "ro", R.drawable.emj_1f1f7_1f1f4, "\ud83c\uddf7\ud83c\uddf4" );
		name( "robot", R.drawable.emj_1f916, "\ud83e\udd16" );
		name( "robot_face", R.drawable.emj_1f916, "\ud83e\udd16" );
		name( "rocket", R.drawable.emj_1f680, "\ud83d\ude80" );
		name( "rofl", R.drawable.emj_1f923, "\ud83e\udd23" );
		name( "rolled_up_newspaper", R.drawable.emj_1f5de_fe0f, "\ud83d\uddde\ufe0f" );
		name( "roller_coaster", R.drawable.emj_1f3a2, "\ud83c\udfa2" );
		name( "rolling_eyes", R.drawable.emj_1f644, "\ud83d\ude44" );
		name( "rolling_on_the_floor_laughing", R.drawable.emj_1f923, "\ud83e\udd23" );
		name( "rooster", R.drawable.emj_1f413, "\ud83d\udc13" );
		name( "rose", R.drawable.emj_1f339, "\ud83c\udf39" );
		name( "rosette", R.drawable.emj_1f3f5_fe0f, "\ud83c\udff5\ufe0f" );
		name( "rotating_light", R.drawable.emj_1f6a8, "\ud83d\udea8" );
		name( "round_pushpin", R.drawable.emj_1f4cd, "\ud83d\udccd" );
		name( "rowboat", R.drawable.emj_1f6a3, "\ud83d\udea3" );
		name( "rowboat_dark_skin_tone", R.drawable.emj_1f6a3_1f3ff, "\ud83d\udea3\ud83c\udfff" );
		name( "rowboat_light_skin_tone", R.drawable.emj_1f6a3_1f3fb, "\ud83d\udea3\ud83c\udffb" );
		name( "rowboat_medium_dark_skin_tone", R.drawable.emj_1f6a3_1f3fe, "\ud83d\udea3\ud83c\udffe" );
		name( "rowboat_medium_light_skin_tone", R.drawable.emj_1f6a3_1f3fc, "\ud83d\udea3\ud83c\udffc" );
		name( "rowboat_medium_skin_tone", R.drawable.emj_1f6a3_1f3fd, "\ud83d\udea3\ud83c\udffd" );
		name( "rowboat_tone1", R.drawable.emj_1f6a3_1f3fb, "\ud83d\udea3\ud83c\udffb" );
		name( "rowboat_tone2", R.drawable.emj_1f6a3_1f3fc, "\ud83d\udea3\ud83c\udffc" );
		name( "rowboat_tone3", R.drawable.emj_1f6a3_1f3fd, "\ud83d\udea3\ud83c\udffd" );
		name( "rowboat_tone4", R.drawable.emj_1f6a3_1f3fe, "\ud83d\udea3\ud83c\udffe" );
		name( "rowboat_tone5", R.drawable.emj_1f6a3_1f3ff, "\ud83d\udea3\ud83c\udfff" );
		name( "rs", R.drawable.emj_1f1f7_1f1f8, "\ud83c\uddf7\ud83c\uddf8" );
		name( "ru", R.drawable.emj_1f1f7_1f1fa, "\ud83c\uddf7\ud83c\uddfa" );
		name( "rugby_football", R.drawable.emj_1f3c9, "\ud83c\udfc9" );
		name( "runner", R.drawable.emj_1f3c3, "\ud83c\udfc3" );
		name( "runner_dark_skin_tone", R.drawable.emj_1f3c3_1f3ff, "\ud83c\udfc3\ud83c\udfff" );
		name( "runner_light_skin_tone", R.drawable.emj_1f3c3_1f3fb, "\ud83c\udfc3\ud83c\udffb" );
		name( "runner_medium_dark_skin_tone", R.drawable.emj_1f3c3_1f3fe, "\ud83c\udfc3\ud83c\udffe" );
		name( "runner_medium_light_skin_tone", R.drawable.emj_1f3c3_1f3fc, "\ud83c\udfc3\ud83c\udffc" );
		name( "runner_medium_skin_tone", R.drawable.emj_1f3c3_1f3fd, "\ud83c\udfc3\ud83c\udffd" );
	}
	private static void init93(){
		name( "runner_tone1", R.drawable.emj_1f3c3_1f3fb, "\ud83c\udfc3\ud83c\udffb" );
		name( "runner_tone2", R.drawable.emj_1f3c3_1f3fc, "\ud83c\udfc3\ud83c\udffc" );
		name( "runner_tone3", R.drawable.emj_1f3c3_1f3fd, "\ud83c\udfc3\ud83c\udffd" );
		name( "runner_tone4", R.drawable.emj_1f3c3_1f3fe, "\ud83c\udfc3\ud83c\udffe" );
		name( "runner_tone5", R.drawable.emj_1f3c3_1f3ff, "\ud83c\udfc3\ud83c\udfff" );
		name( "running", R.drawable.emj_1f3c3, "\ud83c\udfc3" );
		name( "running_dark_skin_tone", R.drawable.emj_1f3c3_1f3ff, "\ud83c\udfc3\ud83c\udfff" );
		name( "running_light_skin_tone", R.drawable.emj_1f3c3_1f3fb, "\ud83c\udfc3\ud83c\udffb" );
		name( "running_medium_dark_skin_tone", R.drawable.emj_1f3c3_1f3fe, "\ud83c\udfc3\ud83c\udffe" );
		name( "running_medium_light_skin_tone", R.drawable.emj_1f3c3_1f3fc, "\ud83c\udfc3\ud83c\udffc" );
		name( "running_medium_skin_tone", R.drawable.emj_1f3c3_1f3fd, "\ud83c\udfc3\ud83c\udffd" );
		name( "running_shirt_with_sash", R.drawable.emj_1f3bd, "\ud83c\udfbd" );
		name( "running_tone1", R.drawable.emj_1f3c3_1f3fb, "\ud83c\udfc3\ud83c\udffb" );
		name( "running_tone2", R.drawable.emj_1f3c3_1f3fc, "\ud83c\udfc3\ud83c\udffc" );
		name( "running_tone3", R.drawable.emj_1f3c3_1f3fd, "\ud83c\udfc3\ud83c\udffd" );
		name( "running_tone4", R.drawable.emj_1f3c3_1f3fe, "\ud83c\udfc3\ud83c\udffe" );
		name( "running_tone5", R.drawable.emj_1f3c3_1f3ff, "\ud83c\udfc3\ud83c\udfff" );
		name( "rw", R.drawable.emj_1f1f7_1f1fc, "\ud83c\uddf7\ud83c\uddfc" );
		name( "sa", R.drawable.emj_1f202_fe0f, "\ud83c\ude02\ufe0f" );
		name( "sagittarius", R.drawable.emj_2650, "\u2650" );
		name( "sailboat", R.drawable.emj_26f5, "\u26f5" );
		name( "sake", R.drawable.emj_1f376, "\ud83c\udf76" );
		name( "salad", R.drawable.emj_1f957, "\ud83e\udd57" );
		name( "sandal", R.drawable.emj_1f461, "\ud83d\udc61" );
		name( "sandwich", R.drawable.emj_1f96a, "\ud83e\udd6a" );
		name( "santa", R.drawable.emj_1f385, "\ud83c\udf85" );
		name( "santa_dark_skin_tone", R.drawable.emj_1f385_1f3ff, "\ud83c\udf85\ud83c\udfff" );
		name( "santa_light_skin_tone", R.drawable.emj_1f385_1f3fb, "\ud83c\udf85\ud83c\udffb" );
		name( "santa_medium_dark_skin_tone", R.drawable.emj_1f385_1f3fe, "\ud83c\udf85\ud83c\udffe" );
		name( "santa_medium_light_skin_tone", R.drawable.emj_1f385_1f3fc, "\ud83c\udf85\ud83c\udffc" );
		name( "santa_medium_skin_tone", R.drawable.emj_1f385_1f3fd, "\ud83c\udf85\ud83c\udffd" );
		name( "santa_tone1", R.drawable.emj_1f385_1f3fb, "\ud83c\udf85\ud83c\udffb" );
		name( "santa_tone2", R.drawable.emj_1f385_1f3fc, "\ud83c\udf85\ud83c\udffc" );
		name( "santa_tone3", R.drawable.emj_1f385_1f3fd, "\ud83c\udf85\ud83c\udffd" );
		name( "santa_tone4", R.drawable.emj_1f385_1f3fe, "\ud83c\udf85\ud83c\udffe" );
		name( "santa_tone5", R.drawable.emj_1f385_1f3ff, "\ud83c\udf85\ud83c\udfff" );
		name( "satellite", R.drawable.emj_1f6f0_fe0f, "\ud83d\udef0\ufe0f" );
		name( "satellite_antenna", R.drawable.emj_1f4e1, "\ud83d\udce1" );
		name( "satellite_orbital", R.drawable.emj_1f6f0_fe0f, "\ud83d\udef0\ufe0f" );
		name( "satisfied", R.drawable.emj_1f606, "\ud83d\ude06" );
		name( "saudi", R.drawable.emj_1f1f8_1f1e6, "\ud83c\uddf8\ud83c\udde6" );
		name( "saudiarabia", R.drawable.emj_1f1f8_1f1e6, "\ud83c\uddf8\ud83c\udde6" );
		name( "sauropod", R.drawable.emj_1f995, "\ud83e\udd95" );
		name( "saxophone", R.drawable.emj_1f3b7, "\ud83c\udfb7" );
		name( "sb", R.drawable.emj_1f1f8_1f1e7, "\ud83c\uddf8\ud83c\udde7" );
		name( "sc", R.drawable.emj_1f1f8_1f1e8, "\ud83c\uddf8\ud83c\udde8" );
		name( "scales", R.drawable.emj_2696_fe0f, "\u2696\ufe0f" );
		name( "scarf", R.drawable.emj_1f9e3, "\ud83e\udde3" );
		name( "school", R.drawable.emj_1f3eb, "\ud83c\udfeb" );
		name( "school_satchel", R.drawable.emj_1f392, "\ud83c\udf92" );
		name( "scissors", R.drawable.emj_2702_fe0f, "\u2702\ufe0f" );
		name( "scooter", R.drawable.emj_1f6f4, "\ud83d\udef4" );
		name( "scorpion", R.drawable.emj_1f982, "\ud83e\udd82" );
		name( "scorpius", R.drawable.emj_264f, "\u264f" );
		name( "scream", R.drawable.emj_1f631, "\ud83d\ude31" );
		name( "scream_cat", R.drawable.emj_1f640, "\ud83d\ude40" );
		name( "scroll", R.drawable.emj_1f4dc, "\ud83d\udcdc" );
		name( "sd", R.drawable.emj_1f1f8_1f1e9, "\ud83c\uddf8\ud83c\udde9" );
		name( "se", R.drawable.emj_1f1f8_1f1ea, "\ud83c\uddf8\ud83c\uddea" );
		name( "seat", R.drawable.emj_1f4ba, "\ud83d\udcba" );
		name( "second_place", R.drawable.emj_1f948, "\ud83e\udd48" );
		name( "second_place_medal", R.drawable.emj_1f948, "\ud83e\udd48" );
		name( "secret", R.drawable.emj_3299_fe0f, "\u3299\ufe0f" );
		name( "see_no_evil", R.drawable.emj_1f648, "\ud83d\ude48" );
		name( "seedling", R.drawable.emj_1f331, "\ud83c\udf31" );
		name( "selfie", R.drawable.emj_1f933, "\ud83e\udd33" );
		name( "selfie_dark_skin_tone", R.drawable.emj_1f933_1f3ff, "\ud83e\udd33\ud83c\udfff" );
		name( "selfie_light_skin_tone", R.drawable.emj_1f933_1f3fb, "\ud83e\udd33\ud83c\udffb" );
		name( "selfie_medium_dark_skin_tone", R.drawable.emj_1f933_1f3fe, "\ud83e\udd33\ud83c\udffe" );
		name( "selfie_medium_light_skin_tone", R.drawable.emj_1f933_1f3fc, "\ud83e\udd33\ud83c\udffc" );
		name( "selfie_medium_skin_tone", R.drawable.emj_1f933_1f3fd, "\ud83e\udd33\ud83c\udffd" );
		name( "selfie_tone1", R.drawable.emj_1f933_1f3fb, "\ud83e\udd33\ud83c\udffb" );
		name( "selfie_tone2", R.drawable.emj_1f933_1f3fc, "\ud83e\udd33\ud83c\udffc" );
		name( "selfie_tone3", R.drawable.emj_1f933_1f3fd, "\ud83e\udd33\ud83c\udffd" );
		name( "selfie_tone4", R.drawable.emj_1f933_1f3fe, "\ud83e\udd33\ud83c\udffe" );
		name( "selfie_tone5", R.drawable.emj_1f933_1f3ff, "\ud83e\udd33\ud83c\udfff" );
		name( "serious_face_with_symbols_covering_mouth", R.drawable.emj_1f92c, "\ud83e\udd2c" );
		name( "seven", R.drawable.emj_0037_fe0f_20e3, "\u0037\ufe0f\u20e3" );
		name( "sg", R.drawable.emj_1f1f8_1f1ec, "\ud83c\uddf8\ud83c\uddec" );
		name( "sh", R.drawable.emj_1f1f8_1f1ed, "\ud83c\uddf8\ud83c\udded" );
		name( "shaking_hands", R.drawable.emj_1f91d, "\ud83e\udd1d" );
		name( "shallow_pan_of_food", R.drawable.emj_1f958, "\ud83e\udd58" );
		name( "shamrock", R.drawable.emj_2618_fe0f, "\u2618\ufe0f" );
		name( "shark", R.drawable.emj_1f988, "\ud83e\udd88" );
		name( "shaved_ice", R.drawable.emj_1f367, "\ud83c\udf67" );
		name( "sheep", R.drawable.emj_1f411, "\ud83d\udc11" );
		name( "shell", R.drawable.emj_1f41a, "\ud83d\udc1a" );
		name( "shelled_peanut", R.drawable.emj_1f95c, "\ud83e\udd5c" );
		name( "shield", R.drawable.emj_1f6e1_fe0f, "\ud83d\udee1\ufe0f" );
		name( "shinto_shrine", R.drawable.emj_26e9_fe0f, "\u26e9\ufe0f" );
		name( "ship", R.drawable.emj_1f6a2, "\ud83d\udea2" );
		name( "shirt", R.drawable.emj_1f455, "\ud83d\udc55" );
		name( "shit", R.drawable.emj_1f4a9, "\ud83d\udca9" );
		name( "shocked_face_with_exploding_head", R.drawable.emj_1f92f, "\ud83e\udd2f" );
		name( "shoe", R.drawable.emj_1f45e, "\ud83d\udc5e" );
		name( "shopping_bags", R.drawable.emj_1f6cd_fe0f, "\ud83d\udecd\ufe0f" );
		name( "shopping_cart", R.drawable.emj_1f6d2, "\ud83d\uded2" );
		name( "shopping_trolley", R.drawable.emj_1f6d2, "\ud83d\uded2" );
		name( "shower", R.drawable.emj_1f6bf, "\ud83d\udebf" );
		name( "shrimp", R.drawable.emj_1f990, "\ud83e\udd90" );
		name( "shrug", R.drawable.emj_1f937, "\ud83e\udd37" );
	}
	private static void init94(){
		name( "shrug_dark_skin_tone", R.drawable.emj_1f937_1f3ff, "\ud83e\udd37\ud83c\udfff" );
		name( "shrug_light_skin_tone", R.drawable.emj_1f937_1f3fb, "\ud83e\udd37\ud83c\udffb" );
		name( "shrug_medium_dark_skin_tone", R.drawable.emj_1f937_1f3fe, "\ud83e\udd37\ud83c\udffe" );
		name( "shrug_medium_light_skin_tone", R.drawable.emj_1f937_1f3fc, "\ud83e\udd37\ud83c\udffc" );
		name( "shrug_medium_skin_tone", R.drawable.emj_1f937_1f3fd, "\ud83e\udd37\ud83c\udffd" );
		name( "shrug_tone1", R.drawable.emj_1f937_1f3fb, "\ud83e\udd37\ud83c\udffb" );
		name( "shrug_tone2", R.drawable.emj_1f937_1f3fc, "\ud83e\udd37\ud83c\udffc" );
		name( "shrug_tone3", R.drawable.emj_1f937_1f3fd, "\ud83e\udd37\ud83c\udffd" );
		name( "shrug_tone4", R.drawable.emj_1f937_1f3fe, "\ud83e\udd37\ud83c\udffe" );
		name( "shrug_tone5", R.drawable.emj_1f937_1f3ff, "\ud83e\udd37\ud83c\udfff" );
		name( "shushing_face", R.drawable.emj_1f92b, "\ud83e\udd2b" );
		name( "si", R.drawable.emj_1f1f8_1f1ee, "\ud83c\uddf8\ud83c\uddee" );
		name( "sick", R.drawable.emj_1f922, "\ud83e\udd22" );
		name( "sign_of_the_horns", R.drawable.emj_1f918, "\ud83e\udd18" );
		name( "sign_of_the_horns_dark_skin_tone", R.drawable.emj_1f918_1f3ff, "\ud83e\udd18\ud83c\udfff" );
		name( "sign_of_the_horns_light_skin_tone", R.drawable.emj_1f918_1f3fb, "\ud83e\udd18\ud83c\udffb" );
		name( "sign_of_the_horns_medium_dark_skin_tone", R.drawable.emj_1f918_1f3fe, "\ud83e\udd18\ud83c\udffe" );
		name( "sign_of_the_horns_medium_light_skin_tone", R.drawable.emj_1f918_1f3fc, "\ud83e\udd18\ud83c\udffc" );
		name( "sign_of_the_horns_medium_skin_tone", R.drawable.emj_1f918_1f3fd, "\ud83e\udd18\ud83c\udffd" );
		name( "sign_of_the_horns_tone1", R.drawable.emj_1f918_1f3fb, "\ud83e\udd18\ud83c\udffb" );
		name( "sign_of_the_horns_tone2", R.drawable.emj_1f918_1f3fc, "\ud83e\udd18\ud83c\udffc" );
		name( "sign_of_the_horns_tone3", R.drawable.emj_1f918_1f3fd, "\ud83e\udd18\ud83c\udffd" );
		name( "sign_of_the_horns_tone4", R.drawable.emj_1f918_1f3fe, "\ud83e\udd18\ud83c\udffe" );
		name( "sign_of_the_horns_tone5", R.drawable.emj_1f918_1f3ff, "\ud83e\udd18\ud83c\udfff" );
		name( "signal_strength", R.drawable.emj_1f4f6, "\ud83d\udcf6" );
		name( "six", R.drawable.emj_0036_fe0f_20e3, "\u0036\ufe0f\u20e3" );
		name( "six_pointed_star", R.drawable.emj_1f52f, "\ud83d\udd2f" );
		name( "sj", R.drawable.emj_1f1f8_1f1ef, "\ud83c\uddf8\ud83c\uddef" );
		name( "sk", R.drawable.emj_1f1f8_1f1f0, "\ud83c\uddf8\ud83c\uddf0" );
		name( "skeleton", R.drawable.emj_1f480, "\ud83d\udc80" );
		name( "ski", R.drawable.emj_1f3bf, "\ud83c\udfbf" );
		name( "skier", R.drawable.emj_26f7_fe0f, "\u26f7\ufe0f" );
		name( "skin_tone_2", R.drawable.emj_1f3fb, "\ud83c\udffb" );
		name( "skin_tone_3", R.drawable.emj_1f3fc, "\ud83c\udffc" );
		name( "skin_tone_4", R.drawable.emj_1f3fd, "\ud83c\udffd" );
		name( "skin_tone_5", R.drawable.emj_1f3fe, "\ud83c\udffe" );
		name( "skin_tone_6", R.drawable.emj_1f3ff, "\ud83c\udfff" );
		name( "skull", R.drawable.emj_1f480, "\ud83d\udc80" );
		name( "skull_and_crossbones", R.drawable.emj_2620_fe0f, "\u2620\ufe0f" );
		name( "skull_crossbones", R.drawable.emj_2620_fe0f, "\u2620\ufe0f" );
		name( "sl", R.drawable.emj_1f1f8_1f1f1, "\ud83c\uddf8\ud83c\uddf1" );
		name( "sled", R.drawable.emj_1f6f7, "\ud83d\udef7" );
		name( "sleeping", R.drawable.emj_1f634, "\ud83d\ude34" );
		name( "sleeping_accommodation", R.drawable.emj_1f6cc, "\ud83d\udecc" );
		name( "sleeping_accommodation_dark_skin_tone", R.drawable.emj_1f6cc_1f3ff, "\ud83d\udecc\ud83c\udfff" );
		name( "sleeping_accommodation_light_skin_tone", R.drawable.emj_1f6cc_1f3fb, "\ud83d\udecc\ud83c\udffb" );
		name( "sleeping_accommodation_medium_dark_skin_tone", R.drawable.emj_1f6cc_1f3fe, "\ud83d\udecc\ud83c\udffe" );
		name( "sleeping_accommodation_medium_light_skin_tone", R.drawable.emj_1f6cc_1f3fc, "\ud83d\udecc\ud83c\udffc" );
		name( "sleeping_accommodation_medium_skin_tone", R.drawable.emj_1f6cc_1f3fd, "\ud83d\udecc\ud83c\udffd" );
		name( "sleeping_accommodation_tone1", R.drawable.emj_1f6cc_1f3fb, "\ud83d\udecc\ud83c\udffb" );
		name( "sleeping_accommodation_tone2", R.drawable.emj_1f6cc_1f3fc, "\ud83d\udecc\ud83c\udffc" );
		name( "sleeping_accommodation_tone3", R.drawable.emj_1f6cc_1f3fd, "\ud83d\udecc\ud83c\udffd" );
		name( "sleeping_accommodation_tone4", R.drawable.emj_1f6cc_1f3fe, "\ud83d\udecc\ud83c\udffe" );
		name( "sleeping_accommodation_tone5", R.drawable.emj_1f6cc_1f3ff, "\ud83d\udecc\ud83c\udfff" );
		name( "sleepy", R.drawable.emj_1f62a, "\ud83d\ude2a" );
		name( "sleuth_or_spy", R.drawable.emj_1f575_fe0f, "\ud83d\udd75\ufe0f" );
		name( "sleuth_or_spy_dark_skin_tone", R.drawable.emj_1f575_1f3ff, "\ud83d\udd75\ud83c\udfff" );
		name( "sleuth_or_spy_light_skin_tone", R.drawable.emj_1f575_1f3fb, "\ud83d\udd75\ud83c\udffb" );
		name( "sleuth_or_spy_medium_dark_skin_tone", R.drawable.emj_1f575_1f3fe, "\ud83d\udd75\ud83c\udffe" );
		name( "sleuth_or_spy_medium_light_skin_tone", R.drawable.emj_1f575_1f3fc, "\ud83d\udd75\ud83c\udffc" );
		name( "sleuth_or_spy_medium_skin_tone", R.drawable.emj_1f575_1f3fd, "\ud83d\udd75\ud83c\udffd" );
		name( "sleuth_or_spy_tone1", R.drawable.emj_1f575_1f3fb, "\ud83d\udd75\ud83c\udffb" );
		name( "sleuth_or_spy_tone2", R.drawable.emj_1f575_1f3fc, "\ud83d\udd75\ud83c\udffc" );
		name( "sleuth_or_spy_tone3", R.drawable.emj_1f575_1f3fd, "\ud83d\udd75\ud83c\udffd" );
		name( "sleuth_or_spy_tone4", R.drawable.emj_1f575_1f3fe, "\ud83d\udd75\ud83c\udffe" );
		name( "sleuth_or_spy_tone5", R.drawable.emj_1f575_1f3ff, "\ud83d\udd75\ud83c\udfff" );
		name( "slight_frown", R.drawable.emj_1f641, "\ud83d\ude41" );
		name( "slight_smile", R.drawable.emj_1f642, "\ud83d\ude42" );
		name( "slightly_frowning_face", R.drawable.emj_1f641, "\ud83d\ude41" );
		name( "slightly_smiling_face", R.drawable.emj_1f642, "\ud83d\ude42" );
		name( "slot_machine", R.drawable.emj_1f3b0, "\ud83c\udfb0" );
		name( "sm", R.drawable.emj_1f1f8_1f1f2, "\ud83c\uddf8\ud83c\uddf2" );
		name( "small_airplane", R.drawable.emj_1f6e9_fe0f, "\ud83d\udee9\ufe0f" );
		name( "small_blue_diamond", R.drawable.emj_1f539, "\ud83d\udd39" );
		name( "small_orange_diamond", R.drawable.emj_1f538, "\ud83d\udd38" );
		name( "small_red_triangle", R.drawable.emj_1f53a, "\ud83d\udd3a" );
		name( "small_red_triangle_down", R.drawable.emj_1f53b, "\ud83d\udd3b" );
		name( "smile", R.drawable.emj_1f604, "\ud83d\ude04" );
		name( "smile_cat", R.drawable.emj_1f638, "\ud83d\ude38" );
		name( "smiley", R.drawable.emj_1f603, "\ud83d\ude03" );
		name( "smiley_cat", R.drawable.emj_1f63a, "\ud83d\ude3a" );
		name( "smiling_face_with_smiling_eyes_and_hand_covering_mouth", R.drawable.emj_1f92d, "\ud83e\udd2d" );
		name( "smiling_imp", R.drawable.emj_1f608, "\ud83d\ude08" );
		name( "smirk", R.drawable.emj_1f60f, "\ud83d\ude0f" );
		name( "smirk_cat", R.drawable.emj_1f63c, "\ud83d\ude3c" );
		name( "smoking", R.drawable.emj_1f6ac, "\ud83d\udeac" );
		name( "sn", R.drawable.emj_1f1f8_1f1f3, "\ud83c\uddf8\ud83c\uddf3" );
		name( "snail", R.drawable.emj_1f40c, "\ud83d\udc0c" );
		name( "snake", R.drawable.emj_1f40d, "\ud83d\udc0d" );
		name( "sneeze", R.drawable.emj_1f927, "\ud83e\udd27" );
		name( "sneezing_face", R.drawable.emj_1f927, "\ud83e\udd27" );
		name( "snow_capped_mountain", R.drawable.emj_1f3d4_fe0f, "\ud83c\udfd4\ufe0f" );
		name( "snow_cloud", R.drawable.emj_1f328_fe0f, "\ud83c\udf28\ufe0f" );
		name( "snowboarder", R.drawable.emj_1f3c2, "\ud83c\udfc2" );
		name( "snowboarder_dark_skin_tone", R.drawable.emj_1f3c2_1f3ff, "\ud83c\udfc2\ud83c\udfff" );
		name( "snowboarder_light_skin_tone", R.drawable.emj_1f3c2_1f3fb, "\ud83c\udfc2\ud83c\udffb" );
		name( "snowboarder_medium_dark_skin_tone", R.drawable.emj_1f3c2_1f3fe, "\ud83c\udfc2\ud83c\udffe" );
		name( "snowboarder_medium_light_skin_tone", R.drawable.emj_1f3c2_1f3fc, "\ud83c\udfc2\ud83c\udffc" );
		name( "snowboarder_medium_skin_tone", R.drawable.emj_1f3c2_1f3fd, "\ud83c\udfc2\ud83c\udffd" );
		name( "snowboarder_tone1", R.drawable.emj_1f3c2_1f3fb, "\ud83c\udfc2\ud83c\udffb" );
		name( "snowboarder_tone2", R.drawable.emj_1f3c2_1f3fc, "\ud83c\udfc2\ud83c\udffc" );
	}
	private static void init95(){
		name( "snowboarder_tone3", R.drawable.emj_1f3c2_1f3fd, "\ud83c\udfc2\ud83c\udffd" );
		name( "snowboarder_tone4", R.drawable.emj_1f3c2_1f3fe, "\ud83c\udfc2\ud83c\udffe" );
		name( "snowboarder_tone5", R.drawable.emj_1f3c2_1f3ff, "\ud83c\udfc2\ud83c\udfff" );
		name( "snowflake", R.drawable.emj_2744_fe0f, "\u2744\ufe0f" );
		name( "snowman", R.drawable.emj_26c4, "\u26c4" );
		name( "snowman2", R.drawable.emj_2603_fe0f, "\u2603\ufe0f" );
		name( "snowman_without_snow", R.drawable.emj_26c4, "\u26c4" );
		name( "so", R.drawable.emj_1f1f8_1f1f4, "\ud83c\uddf8\ud83c\uddf4" );
		name( "sob", R.drawable.emj_1f62d, "\ud83d\ude2d" );
		name( "soccer", R.drawable.emj_26bd, "\u26bd" );
		name( "socks", R.drawable.emj_1f9e6, "\ud83e\udde6" );
		name( "soon", R.drawable.emj_1f51c, "\ud83d\udd1c" );
		name( "sos", R.drawable.emj_1f198, "\ud83c\udd98" );
		name( "sound", R.drawable.emj_1f509, "\ud83d\udd09" );
		name( "space_invader", R.drawable.emj_1f47e, "\ud83d\udc7e" );
		name( "spades", R.drawable.emj_2660_fe0f, "\u2660\ufe0f" );
		name( "spaghetti", R.drawable.emj_1f35d, "\ud83c\udf5d" );
		name( "sparkle", R.drawable.emj_2747_fe0f, "\u2747\ufe0f" );
		name( "sparkler", R.drawable.emj_1f387, "\ud83c\udf87" );
		name( "sparkles", R.drawable.emj_2728, "\u2728" );
		name( "sparkling_heart", R.drawable.emj_1f496, "\ud83d\udc96" );
		name( "speak_no_evil", R.drawable.emj_1f64a, "\ud83d\ude4a" );
		name( "speaker", R.drawable.emj_1f508, "\ud83d\udd08" );
		name( "speaking_head", R.drawable.emj_1f5e3_fe0f, "\ud83d\udde3\ufe0f" );
		name( "speaking_head_in_silhouette", R.drawable.emj_1f5e3_fe0f, "\ud83d\udde3\ufe0f" );
		name( "speech_balloon", R.drawable.emj_1f4ac, "\ud83d\udcac" );
		name( "speech_left", R.drawable.emj_1f5e8_fe0f, "\ud83d\udde8\ufe0f" );
		name( "speedboat", R.drawable.emj_1f6a4, "\ud83d\udea4" );
		name( "spider", R.drawable.emj_1f577_fe0f, "\ud83d\udd77\ufe0f" );
		name( "spider_web", R.drawable.emj_1f578_fe0f, "\ud83d\udd78\ufe0f" );
		name( "spiral_calendar_pad", R.drawable.emj_1f5d3_fe0f, "\ud83d\uddd3\ufe0f" );
		name( "spiral_note_pad", R.drawable.emj_1f5d2_fe0f, "\ud83d\uddd2\ufe0f" );
		name( "spock_hand", R.drawable.emj_1f596, "\ud83d\udd96" );
		name( "spock_hand_dark_skin_tone", R.drawable.emj_1f596_1f3ff, "\ud83d\udd96\ud83c\udfff" );
		name( "spock_hand_light_skin_tone", R.drawable.emj_1f596_1f3fb, "\ud83d\udd96\ud83c\udffb" );
		name( "spock_hand_medium_dark_skin_tone", R.drawable.emj_1f596_1f3fe, "\ud83d\udd96\ud83c\udffe" );
		name( "spock_hand_medium_light_skin_tone", R.drawable.emj_1f596_1f3fc, "\ud83d\udd96\ud83c\udffc" );
		name( "spock_hand_medium_skin_tone", R.drawable.emj_1f596_1f3fd, "\ud83d\udd96\ud83c\udffd" );
		name( "spock_hand_tone1", R.drawable.emj_1f596_1f3fb, "\ud83d\udd96\ud83c\udffb" );
		name( "spock_hand_tone2", R.drawable.emj_1f596_1f3fc, "\ud83d\udd96\ud83c\udffc" );
		name( "spock_hand_tone3", R.drawable.emj_1f596_1f3fd, "\ud83d\udd96\ud83c\udffd" );
		name( "spock_hand_tone4", R.drawable.emj_1f596_1f3fe, "\ud83d\udd96\ud83c\udffe" );
		name( "spock_hand_tone5", R.drawable.emj_1f596_1f3ff, "\ud83d\udd96\ud83c\udfff" );
		name( "spoon", R.drawable.emj_1f944, "\ud83e\udd44" );
		name( "sports_medal", R.drawable.emj_1f3c5, "\ud83c\udfc5" );
		name( "spy", R.drawable.emj_1f575_fe0f, "\ud83d\udd75\ufe0f" );
		name( "spy_tone1", R.drawable.emj_1f575_1f3fb, "\ud83d\udd75\ud83c\udffb" );
		name( "spy_tone2", R.drawable.emj_1f575_1f3fc, "\ud83d\udd75\ud83c\udffc" );
		name( "spy_tone3", R.drawable.emj_1f575_1f3fd, "\ud83d\udd75\ud83c\udffd" );
		name( "spy_tone4", R.drawable.emj_1f575_1f3fe, "\ud83d\udd75\ud83c\udffe" );
		name( "spy_tone5", R.drawable.emj_1f575_1f3ff, "\ud83d\udd75\ud83c\udfff" );
		name( "squid", R.drawable.emj_1f991, "\ud83e\udd91" );
		name( "sr", R.drawable.emj_1f1f8_1f1f7, "\ud83c\uddf8\ud83c\uddf7" );
		name( "ss", R.drawable.emj_1f1f8_1f1f8, "\ud83c\uddf8\ud83c\uddf8" );
		name( "st", R.drawable.emj_1f1f8_1f1f9, "\ud83c\uddf8\ud83c\uddf9" );
		name( "stadium", R.drawable.emj_1f3df_fe0f, "\ud83c\udfdf\ufe0f" );
		name( "staff_of_aesculapius", R.drawable.emj_2695_fe0f, "\u2695\ufe0f" );
		name( "star", R.drawable.emj_2b50, "\u2b50" );
		name( "star2", R.drawable.emj_1f31f, "\ud83c\udf1f" );
		name( "star_and_crescent", R.drawable.emj_262a_fe0f, "\u262a\ufe0f" );
		name( "star_of_david", R.drawable.emj_2721_fe0f, "\u2721\ufe0f" );
		name( "star_struck", R.drawable.emj_1f929, "\ud83e\udd29" );
		name( "stars", R.drawable.emj_1f320, "\ud83c\udf20" );
		name( "station", R.drawable.emj_1f689, "\ud83d\ude89" );
		name( "statue_of_liberty", R.drawable.emj_1f5fd, "\ud83d\uddfd" );
		name( "steam_locomotive", R.drawable.emj_1f682, "\ud83d\ude82" );
		name( "stew", R.drawable.emj_1f372, "\ud83c\udf72" );
		name( "stop_button", R.drawable.emj_23f9_fe0f, "\u23f9\ufe0f" );
		name( "stop_sign", R.drawable.emj_1f6d1, "\ud83d\uded1" );
		name( "stopwatch", R.drawable.emj_23f1_fe0f, "\u23f1\ufe0f" );
		name( "straight_ruler", R.drawable.emj_1f4cf, "\ud83d\udccf" );
		name( "strawberry", R.drawable.emj_1f353, "\ud83c\udf53" );
		name( "stuck_out_tongue", R.drawable.emj_1f61b, "\ud83d\ude1b" );
		name( "stuck_out_tongue_closed_eyes", R.drawable.emj_1f61d, "\ud83d\ude1d" );
		name( "stuck_out_tongue_winking_eye", R.drawable.emj_1f61c, "\ud83d\ude1c" );
		name( "studio_microphone", R.drawable.emj_1f399_fe0f, "\ud83c\udf99\ufe0f" );
		name( "stuffed_flatbread", R.drawable.emj_1f959, "\ud83e\udd59" );
		name( "stuffed_pita", R.drawable.emj_1f959, "\ud83e\udd59" );
		name( "sun_behind_cloud", R.drawable.emj_1f325_fe0f, "\ud83c\udf25\ufe0f" );
		name( "sun_behind_rain_cloud", R.drawable.emj_1f326_fe0f, "\ud83c\udf26\ufe0f" );
		name( "sun_small_cloud", R.drawable.emj_1f324_fe0f, "\ud83c\udf24\ufe0f" );
		name( "sun_with_face", R.drawable.emj_1f31e, "\ud83c\udf1e" );
		name( "sunflower", R.drawable.emj_1f33b, "\ud83c\udf3b" );
		name( "sunglasses", R.drawable.emj_1f60e, "\ud83d\ude0e" );
		name( "sunny", R.drawable.emj_2600_fe0f, "\u2600\ufe0f" );
		name( "sunrise", R.drawable.emj_1f305, "\ud83c\udf05" );
		name( "sunrise_over_mountains", R.drawable.emj_1f304, "\ud83c\udf04" );
		name( "surfer", R.drawable.emj_1f3c4, "\ud83c\udfc4" );
		name( "surfer_dark_skin_tone", R.drawable.emj_1f3c4_1f3ff, "\ud83c\udfc4\ud83c\udfff" );
		name( "surfer_light_skin_tone", R.drawable.emj_1f3c4_1f3fb, "\ud83c\udfc4\ud83c\udffb" );
		name( "surfer_medium_dark_skin_tone", R.drawable.emj_1f3c4_1f3fe, "\ud83c\udfc4\ud83c\udffe" );
		name( "surfer_medium_light_skin_tone", R.drawable.emj_1f3c4_1f3fc, "\ud83c\udfc4\ud83c\udffc" );
		name( "surfer_medium_skin_tone", R.drawable.emj_1f3c4_1f3fd, "\ud83c\udfc4\ud83c\udffd" );
		name( "surfer_tone1", R.drawable.emj_1f3c4_1f3fb, "\ud83c\udfc4\ud83c\udffb" );
		name( "surfer_tone2", R.drawable.emj_1f3c4_1f3fc, "\ud83c\udfc4\ud83c\udffc" );
		name( "surfer_tone3", R.drawable.emj_1f3c4_1f3fd, "\ud83c\udfc4\ud83c\udffd" );
		name( "surfer_tone4", R.drawable.emj_1f3c4_1f3fe, "\ud83c\udfc4\ud83c\udffe" );
		name( "surfer_tone5", R.drawable.emj_1f3c4_1f3ff, "\ud83c\udfc4\ud83c\udfff" );
		name( "sushi", R.drawable.emj_1f363, "\ud83c\udf63" );
		name( "suspension_railway", R.drawable.emj_1f69f, "\ud83d\ude9f" );
		name( "sv", R.drawable.emj_1f1f8_1f1fb, "\ud83c\uddf8\ud83c\uddfb" );
	}
	private static void init96(){
		name( "sweat", R.drawable.emj_1f613, "\ud83d\ude13" );
		name( "sweat_drops", R.drawable.emj_1f4a6, "\ud83d\udca6" );
		name( "sweat_smile", R.drawable.emj_1f605, "\ud83d\ude05" );
		name( "sweet_potato", R.drawable.emj_1f360, "\ud83c\udf60" );
		name( "swimmer", R.drawable.emj_1f3ca, "\ud83c\udfca" );
		name( "swimmer_dark_skin_tone", R.drawable.emj_1f3ca_1f3ff, "\ud83c\udfca\ud83c\udfff" );
		name( "swimmer_light_skin_tone", R.drawable.emj_1f3ca_1f3fb, "\ud83c\udfca\ud83c\udffb" );
		name( "swimmer_medium_dark_skin_tone", R.drawable.emj_1f3ca_1f3fe, "\ud83c\udfca\ud83c\udffe" );
		name( "swimmer_medium_light_skin_tone", R.drawable.emj_1f3ca_1f3fc, "\ud83c\udfca\ud83c\udffc" );
		name( "swimmer_medium_skin_tone", R.drawable.emj_1f3ca_1f3fd, "\ud83c\udfca\ud83c\udffd" );
		name( "swimmer_tone1", R.drawable.emj_1f3ca_1f3fb, "\ud83c\udfca\ud83c\udffb" );
		name( "swimmer_tone2", R.drawable.emj_1f3ca_1f3fc, "\ud83c\udfca\ud83c\udffc" );
		name( "swimmer_tone3", R.drawable.emj_1f3ca_1f3fd, "\ud83c\udfca\ud83c\udffd" );
		name( "swimmer_tone4", R.drawable.emj_1f3ca_1f3fe, "\ud83c\udfca\ud83c\udffe" );
		name( "swimmer_tone5", R.drawable.emj_1f3ca_1f3ff, "\ud83c\udfca\ud83c\udfff" );
		name( "sx", R.drawable.emj_1f1f8_1f1fd, "\ud83c\uddf8\ud83c\uddfd" );
		name( "sy", R.drawable.emj_1f1f8_1f1fe, "\ud83c\uddf8\ud83c\uddfe" );
		name( "symbols", R.drawable.emj_1f523, "\ud83d\udd23" );
		name( "synagogue", R.drawable.emj_1f54d, "\ud83d\udd4d" );
		name( "syringe", R.drawable.emj_1f489, "\ud83d\udc89" );
		name( "sz", R.drawable.emj_1f1f8_1f1ff, "\ud83c\uddf8\ud83c\uddff" );
		name( "t_rex", R.drawable.emj_1f996, "\ud83e\udd96" );
		name( "ta", R.drawable.emj_1f1f9_1f1e6, "\ud83c\uddf9\ud83c\udde6" );
		name( "table_tennis", R.drawable.emj_1f3d3, "\ud83c\udfd3" );
		name( "table_tennis_paddle_and_ball", R.drawable.emj_1f3d3, "\ud83c\udfd3" );
		name( "taco", R.drawable.emj_1f32e, "\ud83c\udf2e" );
		name( "tada", R.drawable.emj_1f389, "\ud83c\udf89" );
		name( "takeout_box", R.drawable.emj_1f961, "\ud83e\udd61" );
		name( "tanabata_tree", R.drawable.emj_1f38b, "\ud83c\udf8b" );
		name( "tangerine", R.drawable.emj_1f34a, "\ud83c\udf4a" );
		name( "taurus", R.drawable.emj_2649, "\u2649" );
		name( "taxi", R.drawable.emj_1f695, "\ud83d\ude95" );
		name( "tc", R.drawable.emj_1f1f9_1f1e8, "\ud83c\uddf9\ud83c\udde8" );
		name( "td", R.drawable.emj_1f1f9_1f1e9, "\ud83c\uddf9\ud83c\udde9" );
		name( "tea", R.drawable.emj_1f375, "\ud83c\udf75" );
		name( "telephone", R.drawable.emj_260e_fe0f, "\u260e\ufe0f" );
		name( "telephone_receiver", R.drawable.emj_1f4de, "\ud83d\udcde" );
		name( "telescope", R.drawable.emj_1f52d, "\ud83d\udd2d" );
		name( "tennis", R.drawable.emj_1f3be, "\ud83c\udfbe" );
		name( "tent", R.drawable.emj_26fa, "\u26fa" );
		name( "tf", R.drawable.emj_1f1f9_1f1eb, "\ud83c\uddf9\ud83c\uddeb" );
		name( "tg", R.drawable.emj_1f1f9_1f1ec, "\ud83c\uddf9\ud83c\uddec" );
		name( "th", R.drawable.emj_1f1f9_1f1ed, "\ud83c\uddf9\ud83c\udded" );
		name( "the_horns", R.drawable.emj_1f918, "\ud83e\udd18" );
		name( "the_horns_dark_skin_tone", R.drawable.emj_1f918_1f3ff, "\ud83e\udd18\ud83c\udfff" );
		name( "the_horns_light_skin_tone", R.drawable.emj_1f918_1f3fb, "\ud83e\udd18\ud83c\udffb" );
		name( "the_horns_medium_dark_skin_tone", R.drawable.emj_1f918_1f3fe, "\ud83e\udd18\ud83c\udffe" );
		name( "the_horns_medium_light_skin_tone", R.drawable.emj_1f918_1f3fc, "\ud83e\udd18\ud83c\udffc" );
		name( "the_horns_medium_skin_tone", R.drawable.emj_1f918_1f3fd, "\ud83e\udd18\ud83c\udffd" );
		name( "the_horns_tone1", R.drawable.emj_1f918_1f3fb, "\ud83e\udd18\ud83c\udffb" );
		name( "the_horns_tone2", R.drawable.emj_1f918_1f3fc, "\ud83e\udd18\ud83c\udffc" );
		name( "the_horns_tone3", R.drawable.emj_1f918_1f3fd, "\ud83e\udd18\ud83c\udffd" );
		name( "the_horns_tone4", R.drawable.emj_1f918_1f3fe, "\ud83e\udd18\ud83c\udffe" );
		name( "the_horns_tone5", R.drawable.emj_1f918_1f3ff, "\ud83e\udd18\ud83c\udfff" );
		name( "thermometer", R.drawable.emj_1f321_fe0f, "\ud83c\udf21\ufe0f" );
		name( "thermometer_face", R.drawable.emj_1f912, "\ud83e\udd12" );
		name( "thinking", R.drawable.emj_1f914, "\ud83e\udd14" );
		name( "thinking_face", R.drawable.emj_1f914, "\ud83e\udd14" );
		name( "third_place", R.drawable.emj_1f949, "\ud83e\udd49" );
		name( "third_place_medal", R.drawable.emj_1f949, "\ud83e\udd49" );
		name( "thought_balloon", R.drawable.emj_1f4ad, "\ud83d\udcad" );
		name( "three", R.drawable.emj_0033_fe0f_20e3, "\u0033\ufe0f\u20e3" );
		name( "three_button_mouse", R.drawable.emj_1f5b1_fe0f, "\ud83d\uddb1\ufe0f" );
		name( "thumbdown", R.drawable.emj_1f44e, "\ud83d\udc4e" );
		name( "thumbdown_tone1", R.drawable.emj_1f44e_1f3fb, "\ud83d\udc4e\ud83c\udffb" );
		name( "thumbdown_tone2", R.drawable.emj_1f44e_1f3fc, "\ud83d\udc4e\ud83c\udffc" );
		name( "thumbdown_tone3", R.drawable.emj_1f44e_1f3fd, "\ud83d\udc4e\ud83c\udffd" );
		name( "thumbdown_tone4", R.drawable.emj_1f44e_1f3fe, "\ud83d\udc4e\ud83c\udffe" );
		name( "thumbdown_tone5", R.drawable.emj_1f44e_1f3ff, "\ud83d\udc4e\ud83c\udfff" );
		name( "thumbsdown", R.drawable.emj_1f44e, "\ud83d\udc4e" );
		name( "thumbsdown_dark_skin_tone", R.drawable.emj_1f44e_1f3ff, "\ud83d\udc4e\ud83c\udfff" );
		name( "thumbsdown_light_skin_tone", R.drawable.emj_1f44e_1f3fb, "\ud83d\udc4e\ud83c\udffb" );
		name( "thumbsdown_medium_dark_skin_tone", R.drawable.emj_1f44e_1f3fe, "\ud83d\udc4e\ud83c\udffe" );
		name( "thumbsdown_medium_light_skin_tone", R.drawable.emj_1f44e_1f3fc, "\ud83d\udc4e\ud83c\udffc" );
		name( "thumbsdown_medium_skin_tone", R.drawable.emj_1f44e_1f3fd, "\ud83d\udc4e\ud83c\udffd" );
		name( "thumbsdown_tone1", R.drawable.emj_1f44e_1f3fb, "\ud83d\udc4e\ud83c\udffb" );
		name( "thumbsdown_tone2", R.drawable.emj_1f44e_1f3fc, "\ud83d\udc4e\ud83c\udffc" );
		name( "thumbsdown_tone3", R.drawable.emj_1f44e_1f3fd, "\ud83d\udc4e\ud83c\udffd" );
		name( "thumbsdown_tone4", R.drawable.emj_1f44e_1f3fe, "\ud83d\udc4e\ud83c\udffe" );
		name( "thumbsdown_tone5", R.drawable.emj_1f44e_1f3ff, "\ud83d\udc4e\ud83c\udfff" );
		name( "thumbsup", R.drawable.emj_1f44d, "\ud83d\udc4d" );
		name( "thumbsup_dark_skin_tone", R.drawable.emj_1f44d_1f3ff, "\ud83d\udc4d\ud83c\udfff" );
		name( "thumbsup_light_skin_tone", R.drawable.emj_1f44d_1f3fb, "\ud83d\udc4d\ud83c\udffb" );
		name( "thumbsup_medium_dark_skin_tone", R.drawable.emj_1f44d_1f3fe, "\ud83d\udc4d\ud83c\udffe" );
		name( "thumbsup_medium_light_skin_tone", R.drawable.emj_1f44d_1f3fc, "\ud83d\udc4d\ud83c\udffc" );
		name( "thumbsup_medium_skin_tone", R.drawable.emj_1f44d_1f3fd, "\ud83d\udc4d\ud83c\udffd" );
		name( "thumbsup_tone1", R.drawable.emj_1f44d_1f3fb, "\ud83d\udc4d\ud83c\udffb" );
		name( "thumbsup_tone2", R.drawable.emj_1f44d_1f3fc, "\ud83d\udc4d\ud83c\udffc" );
		name( "thumbsup_tone3", R.drawable.emj_1f44d_1f3fd, "\ud83d\udc4d\ud83c\udffd" );
		name( "thumbsup_tone4", R.drawable.emj_1f44d_1f3fe, "\ud83d\udc4d\ud83c\udffe" );
		name( "thumbsup_tone5", R.drawable.emj_1f44d_1f3ff, "\ud83d\udc4d\ud83c\udfff" );
		name( "thumbup", R.drawable.emj_1f44d, "\ud83d\udc4d" );
		name( "thumbup_tone1", R.drawable.emj_1f44d_1f3fb, "\ud83d\udc4d\ud83c\udffb" );
		name( "thumbup_tone2", R.drawable.emj_1f44d_1f3fc, "\ud83d\udc4d\ud83c\udffc" );
		name( "thumbup_tone3", R.drawable.emj_1f44d_1f3fd, "\ud83d\udc4d\ud83c\udffd" );
		name( "thumbup_tone4", R.drawable.emj_1f44d_1f3fe, "\ud83d\udc4d\ud83c\udffe" );
		name( "thumbup_tone5", R.drawable.emj_1f44d_1f3ff, "\ud83d\udc4d\ud83c\udfff" );
		name( "thunder_cloud_and_rain", R.drawable.emj_26c8_fe0f, "\u26c8\ufe0f" );
		name( "thunder_cloud_rain", R.drawable.emj_26c8_fe0f, "\u26c8\ufe0f" );
		name( "ticket", R.drawable.emj_1f3ab, "\ud83c\udfab" );
		name( "tickets", R.drawable.emj_1f39f_fe0f, "\ud83c\udf9f\ufe0f" );
	}
	private static void init97(){
		name( "tiger", R.drawable.emj_1f42f, "\ud83d\udc2f" );
		name( "tiger2", R.drawable.emj_1f405, "\ud83d\udc05" );
		name( "timer", R.drawable.emj_23f2_fe0f, "\u23f2\ufe0f" );
		name( "timer_clock", R.drawable.emj_23f2_fe0f, "\u23f2\ufe0f" );
		name( "tired_face", R.drawable.emj_1f62b, "\ud83d\ude2b" );
		name( "tj", R.drawable.emj_1f1f9_1f1ef, "\ud83c\uddf9\ud83c\uddef" );
		name( "tk", R.drawable.emj_1f1f9_1f1f0, "\ud83c\uddf9\ud83c\uddf0" );
		name( "tl", R.drawable.emj_1f1f9_1f1f1, "\ud83c\uddf9\ud83c\uddf1" );
		name( "tm", R.drawable.emj_2122_fe0f, "\u2122\ufe0f" );
		name( "tn", R.drawable.emj_1f1f9_1f1f3, "\ud83c\uddf9\ud83c\uddf3" );
		name( "to", R.drawable.emj_1f1f9_1f1f4, "\ud83c\uddf9\ud83c\uddf4" );
		name( "toilet", R.drawable.emj_1f6bd, "\ud83d\udebd" );
		name( "tokyo_tower", R.drawable.emj_1f5fc, "\ud83d\uddfc" );
		name( "tomato", R.drawable.emj_1f345, "\ud83c\udf45" );
		name( "tone1", R.drawable.emj_1f3fb, "\ud83c\udffb" );
		name( "tone2", R.drawable.emj_1f3fc, "\ud83c\udffc" );
		name( "tone3", R.drawable.emj_1f3fd, "\ud83c\udffd" );
		name( "tone4", R.drawable.emj_1f3fe, "\ud83c\udffe" );
		name( "tone5", R.drawable.emj_1f3ff, "\ud83c\udfff" );
		name( "tongue", R.drawable.emj_1f445, "\ud83d\udc45" );
		name( "tools", R.drawable.emj_1f6e0_fe0f, "\ud83d\udee0\ufe0f" );
		name( "top", R.drawable.emj_1f51d, "\ud83d\udd1d" );
		name( "tophat", R.drawable.emj_1f3a9, "\ud83c\udfa9" );
		name( "tornado", R.drawable.emj_1f32a_fe0f, "\ud83c\udf2a\ufe0f" );
		name( "tornado_cloud", R.drawable.emj_1f32a_fe0f, "\ud83c\udf2a\ufe0f" );
		name( "tr", R.drawable.emj_1f1f9_1f1f7, "\ud83c\uddf9\ud83c\uddf7" );
		name( "track_next", R.drawable.emj_23ed_fe0f, "\u23ed\ufe0f" );
		name( "track_previous", R.drawable.emj_23ee_fe0f, "\u23ee\ufe0f" );
		name( "trackball", R.drawable.emj_1f5b2_fe0f, "\ud83d\uddb2\ufe0f" );
		name( "tractor", R.drawable.emj_1f69c, "\ud83d\ude9c" );
		name( "traffic_light", R.drawable.emj_1f6a5, "\ud83d\udea5" );
		name( "train", R.drawable.emj_1f68b, "\ud83d\ude8b" );
		name( "train2", R.drawable.emj_1f686, "\ud83d\ude86" );
		name( "tram", R.drawable.emj_1f68a, "\ud83d\ude8a" );
		name( "triangular_flag_on_post", R.drawable.emj_1f6a9, "\ud83d\udea9" );
		name( "triangular_ruler", R.drawable.emj_1f4d0, "\ud83d\udcd0" );
		name( "trident", R.drawable.emj_1f531, "\ud83d\udd31" );
		name( "triumph", R.drawable.emj_1f624, "\ud83d\ude24" );
		name( "trolleybus", R.drawable.emj_1f68e, "\ud83d\ude8e" );
		name( "trophy", R.drawable.emj_1f3c6, "\ud83c\udfc6" );
		name( "tropical_drink", R.drawable.emj_1f379, "\ud83c\udf79" );
		name( "tropical_fish", R.drawable.emj_1f420, "\ud83d\udc20" );
		name( "truck", R.drawable.emj_1f69a, "\ud83d\ude9a" );
		name( "trumpet", R.drawable.emj_1f3ba, "\ud83c\udfba" );
		name( "tshirt", R.drawable.emj_1f455, "\ud83d\udc55" );
		name( "tt", R.drawable.emj_1f1f9_1f1f9, "\ud83c\uddf9\ud83c\uddf9" );
		name( "tulip", R.drawable.emj_1f337, "\ud83c\udf37" );
		name( "tumbler_glass", R.drawable.emj_1f943, "\ud83e\udd43" );
		name( "turkey", R.drawable.emj_1f983, "\ud83e\udd83" );
		name( "turkmenistan", R.drawable.emj_1f1f9_1f1f2, "\ud83c\uddf9\ud83c\uddf2" );
		name( "turtle", R.drawable.emj_1f422, "\ud83d\udc22" );
		name( "tuvalu", R.drawable.emj_1f1f9_1f1fb, "\ud83c\uddf9\ud83c\uddfb" );
		name( "tuxedo_tone1", R.drawable.emj_1f935_1f3fb, "\ud83e\udd35\ud83c\udffb" );
		name( "tuxedo_tone2", R.drawable.emj_1f935_1f3fc, "\ud83e\udd35\ud83c\udffc" );
		name( "tuxedo_tone3", R.drawable.emj_1f935_1f3fd, "\ud83e\udd35\ud83c\udffd" );
		name( "tuxedo_tone4", R.drawable.emj_1f935_1f3fe, "\ud83e\udd35\ud83c\udffe" );
		name( "tuxedo_tone5", R.drawable.emj_1f935_1f3ff, "\ud83e\udd35\ud83c\udfff" );
		name( "tv", R.drawable.emj_1f4fa, "\ud83d\udcfa" );
		name( "tw", R.drawable.emj_1f1f9_1f1fc, "\ud83c\uddf9\ud83c\uddfc" );
		name( "twisted_rightwards_arrows", R.drawable.emj_1f500, "\ud83d\udd00" );
		name( "two", R.drawable.emj_0032_fe0f_20e3, "\u0032\ufe0f\u20e3" );
		name( "two_hearts", R.drawable.emj_1f495, "\ud83d\udc95" );
		name( "two_men_holding_hands", R.drawable.emj_1f46c, "\ud83d\udc6c" );
		name( "two_women_holding_hands", R.drawable.emj_1f46d, "\ud83d\udc6d" );
		name( "tz", R.drawable.emj_1f1f9_1f1ff, "\ud83c\uddf9\ud83c\uddff" );
		name( "u5272", R.drawable.emj_1f239, "\ud83c\ude39" );
		name( "u5408", R.drawable.emj_1f234, "\ud83c\ude34" );
		name( "u55b6", R.drawable.emj_1f23a, "\ud83c\ude3a" );
		name( "u6307", R.drawable.emj_1f22f, "\ud83c\ude2f" );
		name( "u6708", R.drawable.emj_1f237_fe0f, "\ud83c\ude37\ufe0f" );
		name( "u6709", R.drawable.emj_1f236, "\ud83c\ude36" );
		name( "u6e80", R.drawable.emj_1f235, "\ud83c\ude35" );
		name( "u7121", R.drawable.emj_1f21a, "\ud83c\ude1a" );
		name( "u7533", R.drawable.emj_1f238, "\ud83c\ude38" );
		name( "u7981", R.drawable.emj_1f232, "\ud83c\ude32" );
		name( "u7a7a", R.drawable.emj_1f233, "\ud83c\ude33" );
		name( "ua", R.drawable.emj_1f1fa_1f1e6, "\ud83c\uddfa\ud83c\udde6" );
		name( "ug", R.drawable.emj_1f1fa_1f1ec, "\ud83c\uddfa\ud83c\uddec" );
		name( "uk", R.drawable.emj_1f1ec_1f1e7, "\ud83c\uddec\ud83c\udde7" );
		name( "um", R.drawable.emj_1f1fa_1f1f2, "\ud83c\uddfa\ud83c\uddf2" );
		name( "umbrella", R.drawable.emj_2602_fe0f, "\u2602\ufe0f" );
		name( "umbrella2", R.drawable.emj_2602_fe0f, "\u2602\ufe0f" );
		name( "umbrella_on_ground", R.drawable.emj_26f1_fe0f, "\u26f1\ufe0f" );
		name( "umbrella_with_rain_drops", R.drawable.emj_2614, "\u2614" );
		name( "unamused", R.drawable.emj_1f612, "\ud83d\ude12" );
		name( "underage", R.drawable.emj_1f51e, "\ud83d\udd1e" );
		name( "unicorn", R.drawable.emj_1f984, "\ud83e\udd84" );
		name( "unicorn_face", R.drawable.emj_1f984, "\ud83e\udd84" );
		name( "united_nations", R.drawable.emj_1f1fa_1f1f3, "\ud83c\uddfa\ud83c\uddf3" );
		name( "unlock", R.drawable.emj_1f513, "\ud83d\udd13" );
		name( "up", R.drawable.emj_1f199, "\ud83c\udd99" );
		name( "upside_down", R.drawable.emj_1f643, "\ud83d\ude43" );
		name( "upside_down_face", R.drawable.emj_1f643, "\ud83d\ude43" );
		name( "urn", R.drawable.emj_26b1_fe0f, "\u26b1\ufe0f" );
		name( "us", R.drawable.emj_1f1fa_1f1f8, "\ud83c\uddfa\ud83c\uddf8" );
		name( "uy", R.drawable.emj_1f1fa_1f1fe, "\ud83c\uddfa\ud83c\uddfe" );
		name( "uz", R.drawable.emj_1f1fa_1f1ff, "\ud83c\uddfa\ud83c\uddff" );
		name( "v", R.drawable.emj_270c_fe0f, "\u270c\ufe0f" );
		name( "v_dark_skin_tone", R.drawable.emj_270c_1f3ff, "\u270c\ud83c\udfff" );
		name( "v_light_skin_tone", R.drawable.emj_270c_1f3fb, "\u270c\ud83c\udffb" );
		name( "v_medium_dark_skin_tone", R.drawable.emj_270c_1f3fe, "\u270c\ud83c\udffe" );
	}
	private static void init98(){
		name( "v_medium_light_skin_tone", R.drawable.emj_270c_1f3fc, "\u270c\ud83c\udffc" );
		name( "v_medium_skin_tone", R.drawable.emj_270c_1f3fd, "\u270c\ud83c\udffd" );
		name( "v_tone1", R.drawable.emj_270c_1f3fb, "\u270c\ud83c\udffb" );
		name( "v_tone2", R.drawable.emj_270c_1f3fc, "\u270c\ud83c\udffc" );
		name( "v_tone3", R.drawable.emj_270c_1f3fd, "\u270c\ud83c\udffd" );
		name( "v_tone4", R.drawable.emj_270c_1f3fe, "\u270c\ud83c\udffe" );
		name( "v_tone5", R.drawable.emj_270c_1f3ff, "\u270c\ud83c\udfff" );
		name( "va", R.drawable.emj_1f1fb_1f1e6, "\ud83c\uddfb\ud83c\udde6" );
		name( "vampire", R.drawable.emj_1f9db, "\ud83e\udddb" );
		name( "vampire_dark_skin_tone", R.drawable.emj_1f9db_1f3ff, "\ud83e\udddb\ud83c\udfff" );
		name( "vampire_light_skin_tone", R.drawable.emj_1f9db_1f3fb, "\ud83e\udddb\ud83c\udffb" );
		name( "vampire_medium_dark_skin_tone", R.drawable.emj_1f9db_1f3fe, "\ud83e\udddb\ud83c\udffe" );
		name( "vampire_medium_light_skin_tone", R.drawable.emj_1f9db_1f3fc, "\ud83e\udddb\ud83c\udffc" );
		name( "vampire_medium_skin_tone", R.drawable.emj_1f9db_1f3fd, "\ud83e\udddb\ud83c\udffd" );
		name( "vampire_tone1", R.drawable.emj_1f9db_1f3fb, "\ud83e\udddb\ud83c\udffb" );
		name( "vampire_tone2", R.drawable.emj_1f9db_1f3fc, "\ud83e\udddb\ud83c\udffc" );
		name( "vampire_tone3", R.drawable.emj_1f9db_1f3fd, "\ud83e\udddb\ud83c\udffd" );
		name( "vampire_tone4", R.drawable.emj_1f9db_1f3fe, "\ud83e\udddb\ud83c\udffe" );
		name( "vampire_tone5", R.drawable.emj_1f9db_1f3ff, "\ud83e\udddb\ud83c\udfff" );
		name( "vc", R.drawable.emj_1f1fb_1f1e8, "\ud83c\uddfb\ud83c\udde8" );
		name( "ve", R.drawable.emj_1f1fb_1f1ea, "\ud83c\uddfb\ud83c\uddea" );
		name( "vertical_traffic_light", R.drawable.emj_1f6a6, "\ud83d\udea6" );
		name( "vg", R.drawable.emj_1f1fb_1f1ec, "\ud83c\uddfb\ud83c\uddec" );
		name( "vhs", R.drawable.emj_1f4fc, "\ud83d\udcfc" );
		name( "vi", R.drawable.emj_1f1fb_1f1ee, "\ud83c\uddfb\ud83c\uddee" );
		name( "vibration_mode", R.drawable.emj_1f4f3, "\ud83d\udcf3" );
		name( "video_camera", R.drawable.emj_1f4f9, "\ud83d\udcf9" );
		name( "video_game", R.drawable.emj_1f3ae, "\ud83c\udfae" );
		name( "violin", R.drawable.emj_1f3bb, "\ud83c\udfbb" );
		name( "virgo", R.drawable.emj_264d, "\u264d" );
		name( "vn", R.drawable.emj_1f1fb_1f1f3, "\ud83c\uddfb\ud83c\uddf3" );
		name( "volcano", R.drawable.emj_1f30b, "\ud83c\udf0b" );
		name( "volleyball", R.drawable.emj_1f3d0, "\ud83c\udfd0" );
		name( "vs", R.drawable.emj_1f19a, "\ud83c\udd9a" );
		name( "vu", R.drawable.emj_1f1fb_1f1fa, "\ud83c\uddfb\ud83c\uddfa" );
		name( "vulcan", R.drawable.emj_1f596, "\ud83d\udd96" );
		name( "vulcan_tone1", R.drawable.emj_1f596_1f3fb, "\ud83d\udd96\ud83c\udffb" );
		name( "vulcan_tone2", R.drawable.emj_1f596_1f3fc, "\ud83d\udd96\ud83c\udffc" );
		name( "vulcan_tone3", R.drawable.emj_1f596_1f3fd, "\ud83d\udd96\ud83c\udffd" );
		name( "vulcan_tone4", R.drawable.emj_1f596_1f3fe, "\ud83d\udd96\ud83c\udffe" );
		name( "vulcan_tone5", R.drawable.emj_1f596_1f3ff, "\ud83d\udd96\ud83c\udfff" );
		name( "walking", R.drawable.emj_1f6b6, "\ud83d\udeb6" );
		name( "walking_dark_skin_tone", R.drawable.emj_1f6b6_1f3ff, "\ud83d\udeb6\ud83c\udfff" );
		name( "walking_light_skin_tone", R.drawable.emj_1f6b6_1f3fb, "\ud83d\udeb6\ud83c\udffb" );
		name( "walking_medium_dark_skin_tone", R.drawable.emj_1f6b6_1f3fe, "\ud83d\udeb6\ud83c\udffe" );
		name( "walking_medium_light_skin_tone", R.drawable.emj_1f6b6_1f3fc, "\ud83d\udeb6\ud83c\udffc" );
		name( "walking_medium_skin_tone", R.drawable.emj_1f6b6_1f3fd, "\ud83d\udeb6\ud83c\udffd" );
		name( "walking_tone1", R.drawable.emj_1f6b6_1f3fb, "\ud83d\udeb6\ud83c\udffb" );
		name( "walking_tone2", R.drawable.emj_1f6b6_1f3fc, "\ud83d\udeb6\ud83c\udffc" );
		name( "walking_tone3", R.drawable.emj_1f6b6_1f3fd, "\ud83d\udeb6\ud83c\udffd" );
		name( "walking_tone4", R.drawable.emj_1f6b6_1f3fe, "\ud83d\udeb6\ud83c\udffe" );
		name( "walking_tone5", R.drawable.emj_1f6b6_1f3ff, "\ud83d\udeb6\ud83c\udfff" );
		name( "waning_crescent_moon", R.drawable.emj_1f318, "\ud83c\udf18" );
		name( "waning_gibbous_moon", R.drawable.emj_1f316, "\ud83c\udf16" );
		name( "warning", R.drawable.emj_26a0_fe0f, "\u26a0\ufe0f" );
		name( "wastebasket", R.drawable.emj_1f5d1_fe0f, "\ud83d\uddd1\ufe0f" );
		name( "watch", R.drawable.emj_231a, "\u231a" );
		name( "water_buffalo", R.drawable.emj_1f403, "\ud83d\udc03" );
		name( "water_polo", R.drawable.emj_1f93d, "\ud83e\udd3d" );
		name( "water_polo_dark_skin_tone", R.drawable.emj_1f93d_1f3ff, "\ud83e\udd3d\ud83c\udfff" );
		name( "water_polo_light_skin_tone", R.drawable.emj_1f93d_1f3fb, "\ud83e\udd3d\ud83c\udffb" );
		name( "water_polo_medium_dark_skin_tone", R.drawable.emj_1f93d_1f3fe, "\ud83e\udd3d\ud83c\udffe" );
		name( "water_polo_medium_light_skin_tone", R.drawable.emj_1f93d_1f3fc, "\ud83e\udd3d\ud83c\udffc" );
		name( "water_polo_medium_skin_tone", R.drawable.emj_1f93d_1f3fd, "\ud83e\udd3d\ud83c\udffd" );
		name( "water_polo_tone1", R.drawable.emj_1f93d_1f3fb, "\ud83e\udd3d\ud83c\udffb" );
		name( "water_polo_tone2", R.drawable.emj_1f93d_1f3fc, "\ud83e\udd3d\ud83c\udffc" );
		name( "water_polo_tone3", R.drawable.emj_1f93d_1f3fd, "\ud83e\udd3d\ud83c\udffd" );
		name( "water_polo_tone4", R.drawable.emj_1f93d_1f3fe, "\ud83e\udd3d\ud83c\udffe" );
		name( "water_polo_tone5", R.drawable.emj_1f93d_1f3ff, "\ud83e\udd3d\ud83c\udfff" );
		name( "watermelon", R.drawable.emj_1f349, "\ud83c\udf49" );
		name( "wave", R.drawable.emj_1f44b, "\ud83d\udc4b" );
		name( "wave_dark_skin_tone", R.drawable.emj_1f44b_1f3ff, "\ud83d\udc4b\ud83c\udfff" );
		name( "wave_light_skin_tone", R.drawable.emj_1f44b_1f3fb, "\ud83d\udc4b\ud83c\udffb" );
		name( "wave_medium_dark_skin_tone", R.drawable.emj_1f44b_1f3fe, "\ud83d\udc4b\ud83c\udffe" );
		name( "wave_medium_light_skin_tone", R.drawable.emj_1f44b_1f3fc, "\ud83d\udc4b\ud83c\udffc" );
		name( "wave_medium_skin_tone", R.drawable.emj_1f44b_1f3fd, "\ud83d\udc4b\ud83c\udffd" );
		name( "wave_tone1", R.drawable.emj_1f44b_1f3fb, "\ud83d\udc4b\ud83c\udffb" );
		name( "wave_tone2", R.drawable.emj_1f44b_1f3fc, "\ud83d\udc4b\ud83c\udffc" );
		name( "wave_tone3", R.drawable.emj_1f44b_1f3fd, "\ud83d\udc4b\ud83c\udffd" );
		name( "wave_tone4", R.drawable.emj_1f44b_1f3fe, "\ud83d\udc4b\ud83c\udffe" );
		name( "wave_tone5", R.drawable.emj_1f44b_1f3ff, "\ud83d\udc4b\ud83c\udfff" );
		name( "waving_black_flag", R.drawable.emj_1f3f4, "\ud83c\udff4" );
		name( "waving_white_flag", R.drawable.emj_1f3f3_fe0f, "\ud83c\udff3\ufe0f" );
		name( "wavy_dash", R.drawable.emj_3030_fe0f, "\u3030\ufe0f" );
		name( "waxing_crescent_moon", R.drawable.emj_1f312, "\ud83c\udf12" );
		name( "waxing_gibbous_moon", R.drawable.emj_1f314, "\ud83c\udf14" );
		name( "wc", R.drawable.emj_1f6be, "\ud83d\udebe" );
		name( "weary", R.drawable.emj_1f629, "\ud83d\ude29" );
		name( "wedding", R.drawable.emj_1f492, "\ud83d\udc92" );
		name( "weight_lifter", R.drawable.emj_1f3cb_fe0f, "\ud83c\udfcb\ufe0f" );
		name( "weight_lifter_dark_skin_tone", R.drawable.emj_1f3cb_1f3ff, "\ud83c\udfcb\ud83c\udfff" );
		name( "weight_lifter_light_skin_tone", R.drawable.emj_1f3cb_1f3fb, "\ud83c\udfcb\ud83c\udffb" );
		name( "weight_lifter_medium_dark_skin_tone", R.drawable.emj_1f3cb_1f3fe, "\ud83c\udfcb\ud83c\udffe" );
		name( "weight_lifter_medium_light_skin_tone", R.drawable.emj_1f3cb_1f3fc, "\ud83c\udfcb\ud83c\udffc" );
		name( "weight_lifter_medium_skin_tone", R.drawable.emj_1f3cb_1f3fd, "\ud83c\udfcb\ud83c\udffd" );
		name( "weight_lifter_tone1", R.drawable.emj_1f3cb_1f3fb, "\ud83c\udfcb\ud83c\udffb" );
		name( "weight_lifter_tone2", R.drawable.emj_1f3cb_1f3fc, "\ud83c\udfcb\ud83c\udffc" );
		name( "weight_lifter_tone3", R.drawable.emj_1f3cb_1f3fd, "\ud83c\udfcb\ud83c\udffd" );
		name( "weight_lifter_tone4", R.drawable.emj_1f3cb_1f3fe, "\ud83c\udfcb\ud83c\udffe" );
		name( "weight_lifter_tone5", R.drawable.emj_1f3cb_1f3ff, "\ud83c\udfcb\ud83c\udfff" );
		name( "wf", R.drawable.emj_1f1fc_1f1eb, "\ud83c\uddfc\ud83c\uddeb" );
	}
	private static void init99(){
		name( "whale", R.drawable.emj_1f433, "\ud83d\udc33" );
		name( "whale2", R.drawable.emj_1f40b, "\ud83d\udc0b" );
		name( "wheel_of_dharma", R.drawable.emj_2638_fe0f, "\u2638\ufe0f" );
		name( "wheelchair", R.drawable.emj_267f, "\u267f" );
		name( "whisky", R.drawable.emj_1f943, "\ud83e\udd43" );
		name( "white_check_mark", R.drawable.emj_2705, "\u2705" );
		name( "white_circle", R.drawable.emj_26aa, "\u26aa" );
		name( "white_flower", R.drawable.emj_1f4ae, "\ud83d\udcae" );
		name( "white_frowning_face", R.drawable.emj_2639_fe0f, "\u2639\ufe0f" );
		name( "white_large_square", R.drawable.emj_2b1c, "\u2b1c" );
		name( "white_medium_small_square", R.drawable.emj_25fd, "\u25fd" );
		name( "white_medium_square", R.drawable.emj_25fb_fe0f, "\u25fb\ufe0f" );
		name( "white_small_square", R.drawable.emj_25ab_fe0f, "\u25ab\ufe0f" );
		name( "white_square_button", R.drawable.emj_1f533, "\ud83d\udd33" );
		name( "white_sun_behind_cloud", R.drawable.emj_1f325_fe0f, "\ud83c\udf25\ufe0f" );
		name( "white_sun_behind_cloud_with_rain", R.drawable.emj_1f326_fe0f, "\ud83c\udf26\ufe0f" );
		name( "white_sun_cloud", R.drawable.emj_1f325_fe0f, "\ud83c\udf25\ufe0f" );
		name( "white_sun_rain_cloud", R.drawable.emj_1f326_fe0f, "\ud83c\udf26\ufe0f" );
		name( "white_sun_small_cloud", R.drawable.emj_1f324_fe0f, "\ud83c\udf24\ufe0f" );
		name( "white_sun_with_small_cloud", R.drawable.emj_1f324_fe0f, "\ud83c\udf24\ufe0f" );
		name( "wilted_flower", R.drawable.emj_1f940, "\ud83e\udd40" );
		name( "wilted_rose", R.drawable.emj_1f940, "\ud83e\udd40" );
		name( "wind_blowing_face", R.drawable.emj_1f32c_fe0f, "\ud83c\udf2c\ufe0f" );
		name( "wind_chime", R.drawable.emj_1f390, "\ud83c\udf90" );
		name( "wine_glass", R.drawable.emj_1f377, "\ud83c\udf77" );
		name( "wink", R.drawable.emj_1f609, "\ud83d\ude09" );
		name( "wolf", R.drawable.emj_1f43a, "\ud83d\udc3a" );
		name( "woman", R.drawable.emj_1f469, "\ud83d\udc69" );
		name( "woman_artist", R.drawable.emj_1f469_200d_1f3a8, "\ud83d\udc69\u200d\ud83c\udfa8" );
		name( "woman_artist_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3a8, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa8" );
		name( "woman_artist_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3a8, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "woman_artist_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3a8, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "woman_artist_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3a8, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "woman_artist_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3a8, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "woman_artist_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3a8, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa8" );
		name( "woman_artist_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3a8, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa8" );
		name( "woman_artist_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3a8, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa8" );
		name( "woman_artist_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3a8, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa8" );
		name( "woman_artist_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3a8, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa8" );
		name( "woman_astronaut", R.drawable.emj_1f469_200d_1f680, "\ud83d\udc69\u200d\ud83d\ude80" );
		name( "woman_astronaut_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f680, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude80" );
		name( "woman_astronaut_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f680, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude80" );
		name( "woman_astronaut_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f680, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude80" );
		name( "woman_astronaut_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f680, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude80" );
		name( "woman_astronaut_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f680, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude80" );
		name( "woman_astronaut_tone1", R.drawable.emj_1f469_1f3fb_200d_1f680, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude80" );
		name( "woman_astronaut_tone2", R.drawable.emj_1f469_1f3fc_200d_1f680, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude80" );
		name( "woman_astronaut_tone3", R.drawable.emj_1f469_1f3fd_200d_1f680, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude80" );
		name( "woman_astronaut_tone4", R.drawable.emj_1f469_1f3fe_200d_1f680, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude80" );
		name( "woman_astronaut_tone5", R.drawable.emj_1f469_1f3ff_200d_1f680, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude80" );
		name( "woman_biking", R.drawable.emj_1f6b4_200d_2640_fe0f, "\ud83d\udeb4\u200d\u2640\ufe0f" );
		name( "woman_biking_dark_skin_tone", R.drawable.emj_1f6b4_1f3ff_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_biking_light_skin_tone", R.drawable.emj_1f6b4_1f3fb_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_biking_medium_dark_skin_tone", R.drawable.emj_1f6b4_1f3fe_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_biking_medium_light_skin_tone", R.drawable.emj_1f6b4_1f3fc_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_biking_medium_skin_tone", R.drawable.emj_1f6b4_1f3fd_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_biking_tone1", R.drawable.emj_1f6b4_1f3fb_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_biking_tone2", R.drawable.emj_1f6b4_1f3fc_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_biking_tone3", R.drawable.emj_1f6b4_1f3fd_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_biking_tone4", R.drawable.emj_1f6b4_1f3fe_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_biking_tone5", R.drawable.emj_1f6b4_1f3ff_200d_2640_fe0f, "\ud83d\udeb4\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball", R.drawable.emj_26f9_fe0f_200d_2640_fe0f, "\u26f9\ufe0f\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_dark_skin_tone", R.drawable.emj_26f9_1f3ff_200d_2640_fe0f, "\u26f9\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_light_skin_tone", R.drawable.emj_26f9_1f3fb_200d_2640_fe0f, "\u26f9\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_medium_dark_skin_tone", R.drawable.emj_26f9_1f3fe_200d_2640_fe0f, "\u26f9\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_medium_light_skin_tone", R.drawable.emj_26f9_1f3fc_200d_2640_fe0f, "\u26f9\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_medium_skin_tone", R.drawable.emj_26f9_1f3fd_200d_2640_fe0f, "\u26f9\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_tone1", R.drawable.emj_26f9_1f3fb_200d_2640_fe0f, "\u26f9\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_tone2", R.drawable.emj_26f9_1f3fc_200d_2640_fe0f, "\u26f9\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_tone3", R.drawable.emj_26f9_1f3fd_200d_2640_fe0f, "\u26f9\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_tone4", R.drawable.emj_26f9_1f3fe_200d_2640_fe0f, "\u26f9\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_bouncing_ball_tone5", R.drawable.emj_26f9_1f3ff_200d_2640_fe0f, "\u26f9\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_bowing", R.drawable.emj_1f647_200d_2640_fe0f, "\ud83d\ude47\u200d\u2640\ufe0f" );
		name( "woman_bowing_dark_skin_tone", R.drawable.emj_1f647_1f3ff_200d_2640_fe0f, "\ud83d\ude47\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_bowing_light_skin_tone", R.drawable.emj_1f647_1f3fb_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_bowing_medium_dark_skin_tone", R.drawable.emj_1f647_1f3fe_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_bowing_medium_light_skin_tone", R.drawable.emj_1f647_1f3fc_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_bowing_medium_skin_tone", R.drawable.emj_1f647_1f3fd_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_bowing_tone1", R.drawable.emj_1f647_1f3fb_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_bowing_tone2", R.drawable.emj_1f647_1f3fc_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_bowing_tone3", R.drawable.emj_1f647_1f3fd_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_bowing_tone4", R.drawable.emj_1f647_1f3fe_200d_2640_fe0f, "\ud83d\ude47\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_bowing_tone5", R.drawable.emj_1f647_1f3ff_200d_2640_fe0f, "\ud83d\ude47\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_boy", R.drawable.emj_1f469_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc66" );
		name( "woman_boy_boy", R.drawable.emj_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "woman_cartwheeling", R.drawable.emj_1f938_200d_2640_fe0f, "\ud83e\udd38\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_dark_skin_tone", R.drawable.emj_1f938_1f3ff_200d_2640_fe0f, "\ud83e\udd38\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_light_skin_tone", R.drawable.emj_1f938_1f3fb_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_medium_dark_skin_tone", R.drawable.emj_1f938_1f3fe_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_medium_light_skin_tone", R.drawable.emj_1f938_1f3fc_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_medium_skin_tone", R.drawable.emj_1f938_1f3fd_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_tone1", R.drawable.emj_1f938_1f3fb_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_tone2", R.drawable.emj_1f938_1f3fc_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_tone3", R.drawable.emj_1f938_1f3fd_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_tone4", R.drawable.emj_1f938_1f3fe_200d_2640_fe0f, "\ud83e\udd38\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_cartwheeling_tone5", R.drawable.emj_1f938_1f3ff_200d_2640_fe0f, "\ud83e\udd38\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_climbing", R.drawable.emj_1f9d7_200d_2640_fe0f, "\ud83e\uddd7\u200d\u2640\ufe0f" );
		name( "woman_climbing_dark_skin_tone", R.drawable.emj_1f9d7_1f3ff_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_climbing_light_skin_tone", R.drawable.emj_1f9d7_1f3fb_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_climbing_medium_dark_skin_tone", R.drawable.emj_1f9d7_1f3fe_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_climbing_medium_light_skin_tone", R.drawable.emj_1f9d7_1f3fc_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffc\u200d\u2640\ufe0f" );
	}
	private static void init100(){
		name( "woman_climbing_medium_skin_tone", R.drawable.emj_1f9d7_1f3fd_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_climbing_tone1", R.drawable.emj_1f9d7_1f3fb_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_climbing_tone2", R.drawable.emj_1f9d7_1f3fc_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_climbing_tone3", R.drawable.emj_1f9d7_1f3fd_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_climbing_tone4", R.drawable.emj_1f9d7_1f3fe_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_climbing_tone5", R.drawable.emj_1f9d7_1f3ff_200d_2640_fe0f, "\ud83e\uddd7\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_construction_worker", R.drawable.emj_1f477_200d_2640_fe0f, "\ud83d\udc77\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_dark_skin_tone", R.drawable.emj_1f477_1f3ff_200d_2640_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_light_skin_tone", R.drawable.emj_1f477_1f3fb_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_medium_dark_skin_tone", R.drawable.emj_1f477_1f3fe_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_medium_light_skin_tone", R.drawable.emj_1f477_1f3fc_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_medium_skin_tone", R.drawable.emj_1f477_1f3fd_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_tone1", R.drawable.emj_1f477_1f3fb_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_tone2", R.drawable.emj_1f477_1f3fc_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_tone3", R.drawable.emj_1f477_1f3fd_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_tone4", R.drawable.emj_1f477_1f3fe_200d_2640_fe0f, "\ud83d\udc77\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_construction_worker_tone5", R.drawable.emj_1f477_1f3ff_200d_2640_fe0f, "\ud83d\udc77\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_cook", R.drawable.emj_1f469_200d_1f373, "\ud83d\udc69\u200d\ud83c\udf73" );
		name( "woman_cook_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f373, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf73" );
		name( "woman_cook_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f373, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf73" );
		name( "woman_cook_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f373, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf73" );
		name( "woman_cook_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f373, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf73" );
		name( "woman_cook_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f373, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf73" );
		name( "woman_cook_tone1", R.drawable.emj_1f469_1f3fb_200d_1f373, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf73" );
		name( "woman_cook_tone2", R.drawable.emj_1f469_1f3fc_200d_1f373, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf73" );
		name( "woman_cook_tone3", R.drawable.emj_1f469_1f3fd_200d_1f373, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf73" );
		name( "woman_cook_tone4", R.drawable.emj_1f469_1f3fe_200d_1f373, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf73" );
		name( "woman_cook_tone5", R.drawable.emj_1f469_1f3ff_200d_1f373, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf73" );
		name( "woman_dark_skin_tone", R.drawable.emj_1f469_1f3ff, "\ud83d\udc69\ud83c\udfff" );
		name( "woman_detective", R.drawable.emj_1f575_fe0f_200d_2640_fe0f, "\ud83d\udd75\ufe0f\u200d\u2640\ufe0f" );
		name( "woman_detective_dark_skin_tone", R.drawable.emj_1f575_1f3ff_200d_2640_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_detective_light_skin_tone", R.drawable.emj_1f575_1f3fb_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_detective_medium_dark_skin_tone", R.drawable.emj_1f575_1f3fe_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_detective_medium_light_skin_tone", R.drawable.emj_1f575_1f3fc_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_detective_medium_skin_tone", R.drawable.emj_1f575_1f3fd_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_detective_tone1", R.drawable.emj_1f575_1f3fb_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_detective_tone2", R.drawable.emj_1f575_1f3fc_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_detective_tone3", R.drawable.emj_1f575_1f3fd_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_detective_tone4", R.drawable.emj_1f575_1f3fe_200d_2640_fe0f, "\ud83d\udd75\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_detective_tone5", R.drawable.emj_1f575_1f3ff_200d_2640_fe0f, "\ud83d\udd75\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_facepalming", R.drawable.emj_1f926_200d_2640_fe0f, "\ud83e\udd26\u200d\u2640\ufe0f" );
		name( "woman_facepalming_dark_skin_tone", R.drawable.emj_1f926_1f3ff_200d_2640_fe0f, "\ud83e\udd26\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_facepalming_light_skin_tone", R.drawable.emj_1f926_1f3fb_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_facepalming_medium_dark_skin_tone", R.drawable.emj_1f926_1f3fe_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_facepalming_medium_light_skin_tone", R.drawable.emj_1f926_1f3fc_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_facepalming_medium_skin_tone", R.drawable.emj_1f926_1f3fd_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_facepalming_tone1", R.drawable.emj_1f926_1f3fb_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_facepalming_tone2", R.drawable.emj_1f926_1f3fc_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_facepalming_tone3", R.drawable.emj_1f926_1f3fd_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_facepalming_tone4", R.drawable.emj_1f926_1f3fe_200d_2640_fe0f, "\ud83e\udd26\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_facepalming_tone5", R.drawable.emj_1f926_1f3ff_200d_2640_fe0f, "\ud83e\udd26\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_factory_worker", R.drawable.emj_1f469_200d_1f3ed, "\ud83d\udc69\u200d\ud83c\udfed" );
		name( "woman_factory_worker_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3ed, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfed" );
		name( "woman_factory_worker_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3ed, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfed" );
		name( "woman_factory_worker_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3ed, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfed" );
		name( "woman_factory_worker_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3ed, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfed" );
		name( "woman_factory_worker_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3ed, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfed" );
		name( "woman_factory_worker_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3ed, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfed" );
		name( "woman_factory_worker_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3ed, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfed" );
		name( "woman_factory_worker_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3ed, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfed" );
		name( "woman_factory_worker_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3ed, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfed" );
		name( "woman_factory_worker_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3ed, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfed" );
		name( "woman_farmer", R.drawable.emj_1f469_200d_1f33e, "\ud83d\udc69\u200d\ud83c\udf3e" );
		name( "woman_farmer_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f33e, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "woman_farmer_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f33e, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "woman_farmer_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f33e, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf3e" );
		name( "woman_farmer_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f33e, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "woman_farmer_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f33e, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "woman_farmer_tone1", R.drawable.emj_1f469_1f3fb_200d_1f33e, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf3e" );
		name( "woman_farmer_tone2", R.drawable.emj_1f469_1f3fc_200d_1f33e, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf3e" );
		name( "woman_farmer_tone3", R.drawable.emj_1f469_1f3fd_200d_1f33e, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf3e" );
		name( "woman_farmer_tone4", R.drawable.emj_1f469_1f3fe_200d_1f33e, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf3e" );
		name( "woman_farmer_tone5", R.drawable.emj_1f469_1f3ff_200d_1f33e, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf3e" );
		name( "woman_firefighter", R.drawable.emj_1f469_200d_1f692, "\ud83d\udc69\u200d\ud83d\ude92" );
		name( "woman_firefighter_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f692, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude92" );
		name( "woman_firefighter_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f692, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude92" );
		name( "woman_firefighter_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f692, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude92" );
		name( "woman_firefighter_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f692, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude92" );
		name( "woman_firefighter_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f692, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude92" );
		name( "woman_firefighter_tone1", R.drawable.emj_1f469_1f3fb_200d_1f692, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\ude92" );
		name( "woman_firefighter_tone2", R.drawable.emj_1f469_1f3fc_200d_1f692, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\ude92" );
		name( "woman_firefighter_tone3", R.drawable.emj_1f469_1f3fd_200d_1f692, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\ude92" );
		name( "woman_firefighter_tone4", R.drawable.emj_1f469_1f3fe_200d_1f692, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\ude92" );
		name( "woman_firefighter_tone5", R.drawable.emj_1f469_1f3ff_200d_1f692, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\ude92" );
		name( "woman_frowning", R.drawable.emj_1f64d_200d_2640_fe0f, "\ud83d\ude4d\u200d\u2640\ufe0f" );
		name( "woman_frowning_dark_skin_tone", R.drawable.emj_1f64d_1f3ff_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_frowning_light_skin_tone", R.drawable.emj_1f64d_1f3fb_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_frowning_medium_dark_skin_tone", R.drawable.emj_1f64d_1f3fe_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_frowning_medium_light_skin_tone", R.drawable.emj_1f64d_1f3fc_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_frowning_medium_skin_tone", R.drawable.emj_1f64d_1f3fd_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_frowning_tone1", R.drawable.emj_1f64d_1f3fb_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_frowning_tone2", R.drawable.emj_1f64d_1f3fc_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_frowning_tone3", R.drawable.emj_1f64d_1f3fd_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_frowning_tone4", R.drawable.emj_1f64d_1f3fe_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_frowning_tone5", R.drawable.emj_1f64d_1f3ff_200d_2640_fe0f, "\ud83d\ude4d\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no", R.drawable.emj_1f645_200d_2640_fe0f, "\ud83d\ude45\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_dark_skin_tone", R.drawable.emj_1f645_1f3ff_200d_2640_fe0f, "\ud83d\ude45\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_light_skin_tone", R.drawable.emj_1f645_1f3fb_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_medium_dark_skin_tone", R.drawable.emj_1f645_1f3fe_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_medium_light_skin_tone", R.drawable.emj_1f645_1f3fc_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_medium_skin_tone", R.drawable.emj_1f645_1f3fd_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffd\u200d\u2640\ufe0f" );
	}
	private static void init101(){
		name( "woman_gesturing_no_tone1", R.drawable.emj_1f645_1f3fb_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_tone2", R.drawable.emj_1f645_1f3fc_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_tone3", R.drawable.emj_1f645_1f3fd_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_tone4", R.drawable.emj_1f645_1f3fe_200d_2640_fe0f, "\ud83d\ude45\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_gesturing_no_tone5", R.drawable.emj_1f645_1f3ff_200d_2640_fe0f, "\ud83d\ude45\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok", R.drawable.emj_1f646_200d_2640_fe0f, "\ud83d\ude46\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_dark_skin_tone", R.drawable.emj_1f646_1f3ff_200d_2640_fe0f, "\ud83d\ude46\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_light_skin_tone", R.drawable.emj_1f646_1f3fb_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_medium_dark_skin_tone", R.drawable.emj_1f646_1f3fe_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_medium_light_skin_tone", R.drawable.emj_1f646_1f3fc_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_medium_skin_tone", R.drawable.emj_1f646_1f3fd_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_tone1", R.drawable.emj_1f646_1f3fb_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_tone2", R.drawable.emj_1f646_1f3fc_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_tone3", R.drawable.emj_1f646_1f3fd_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_tone4", R.drawable.emj_1f646_1f3fe_200d_2640_fe0f, "\ud83d\ude46\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_gesturing_ok_tone5", R.drawable.emj_1f646_1f3ff_200d_2640_fe0f, "\ud83d\ude46\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage", R.drawable.emj_1f486_200d_2640_fe0f, "\ud83d\udc86\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_dark_skin_tone", R.drawable.emj_1f486_1f3ff_200d_2640_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_light_skin_tone", R.drawable.emj_1f486_1f3fb_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_medium_dark_skin_tone", R.drawable.emj_1f486_1f3fe_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_medium_light_skin_tone", R.drawable.emj_1f486_1f3fc_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_medium_skin_tone", R.drawable.emj_1f486_1f3fd_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_tone1", R.drawable.emj_1f486_1f3fb_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_tone2", R.drawable.emj_1f486_1f3fc_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_tone3", R.drawable.emj_1f486_1f3fd_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_tone4", R.drawable.emj_1f486_1f3fe_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_getting_face_massage_tone5", R.drawable.emj_1f486_1f3ff_200d_2640_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut", R.drawable.emj_1f487_200d_2640_fe0f, "\ud83d\udc87\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_dark_skin_tone", R.drawable.emj_1f487_1f3ff_200d_2640_fe0f, "\ud83d\udc87\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_light_skin_tone", R.drawable.emj_1f487_1f3fb_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_medium_dark_skin_tone", R.drawable.emj_1f487_1f3fe_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_medium_light_skin_tone", R.drawable.emj_1f487_1f3fc_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_medium_skin_tone", R.drawable.emj_1f487_1f3fd_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_tone1", R.drawable.emj_1f487_1f3fb_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_tone2", R.drawable.emj_1f487_1f3fc_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_tone3", R.drawable.emj_1f487_1f3fd_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_tone4", R.drawable.emj_1f487_1f3fe_200d_2640_fe0f, "\ud83d\udc87\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_getting_haircut_tone5", R.drawable.emj_1f487_1f3ff_200d_2640_fe0f, "\ud83d\udc87\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_getting_massage", R.drawable.emj_1f486_200d_2640_fe0f, "\ud83d\udc86\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_dark_skin_tone", R.drawable.emj_1f486_1f3ff_200d_2640_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_light_skin_tone", R.drawable.emj_1f486_1f3fb_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_medium_dark_skin_tone", R.drawable.emj_1f486_1f3fe_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_medium_light_skin_tone", R.drawable.emj_1f486_1f3fc_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_medium_skin_tone", R.drawable.emj_1f486_1f3fd_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_tone1", R.drawable.emj_1f486_1f3fb_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_tone2", R.drawable.emj_1f486_1f3fc_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_tone3", R.drawable.emj_1f486_1f3fd_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_tone4", R.drawable.emj_1f486_1f3fe_200d_2640_fe0f, "\ud83d\udc86\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_getting_massage_tone5", R.drawable.emj_1f486_1f3ff_200d_2640_fe0f, "\ud83d\udc86\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_girl", R.drawable.emj_1f469_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc67" );
		name( "woman_girl_boy", R.drawable.emj_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "woman_girl_girl", R.drawable.emj_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "woman_golfing", R.drawable.emj_1f3cc_fe0f_200d_2640_fe0f, "\ud83c\udfcc\ufe0f\u200d\u2640\ufe0f" );
		name( "woman_golfing_dark_skin_tone", R.drawable.emj_1f3cc_1f3ff_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_golfing_light_skin_tone", R.drawable.emj_1f3cc_1f3fb_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_golfing_medium_dark_skin_tone", R.drawable.emj_1f3cc_1f3fe_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_golfing_medium_light_skin_tone", R.drawable.emj_1f3cc_1f3fc_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_golfing_medium_skin_tone", R.drawable.emj_1f3cc_1f3fd_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_golfing_tone1", R.drawable.emj_1f3cc_1f3fb_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_golfing_tone2", R.drawable.emj_1f3cc_1f3fc_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_golfing_tone3", R.drawable.emj_1f3cc_1f3fd_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_golfing_tone4", R.drawable.emj_1f3cc_1f3fe_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_golfing_tone5", R.drawable.emj_1f3cc_1f3ff_200d_2640_fe0f, "\ud83c\udfcc\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_guard", R.drawable.emj_1f482_200d_2640_fe0f, "\ud83d\udc82\u200d\u2640\ufe0f" );
		name( "woman_guard_dark_skin_tone", R.drawable.emj_1f482_1f3ff_200d_2640_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_guard_light_skin_tone", R.drawable.emj_1f482_1f3fb_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_guard_medium_dark_skin_tone", R.drawable.emj_1f482_1f3fe_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_guard_medium_light_skin_tone", R.drawable.emj_1f482_1f3fc_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_guard_medium_skin_tone", R.drawable.emj_1f482_1f3fd_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_guard_tone1", R.drawable.emj_1f482_1f3fb_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_guard_tone2", R.drawable.emj_1f482_1f3fc_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_guard_tone3", R.drawable.emj_1f482_1f3fd_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_guard_tone4", R.drawable.emj_1f482_1f3fe_200d_2640_fe0f, "\ud83d\udc82\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_guard_tone5", R.drawable.emj_1f482_1f3ff_200d_2640_fe0f, "\ud83d\udc82\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_health_worker", R.drawable.emj_1f469_200d_2695_fe0f, "\ud83d\udc69\u200d\u2695\ufe0f" );
		name( "woman_health_worker_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_2695_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "woman_health_worker_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "woman_health_worker_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "woman_health_worker_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2695\ufe0f" );
		name( "woman_health_worker_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "woman_health_worker_tone1", R.drawable.emj_1f469_1f3fb_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2695\ufe0f" );
		name( "woman_health_worker_tone2", R.drawable.emj_1f469_1f3fc_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2695\ufe0f" );
		name( "woman_health_worker_tone3", R.drawable.emj_1f469_1f3fd_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2695\ufe0f" );
		name( "woman_health_worker_tone4", R.drawable.emj_1f469_1f3fe_200d_2695_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2695\ufe0f" );
		name( "woman_health_worker_tone5", R.drawable.emj_1f469_1f3ff_200d_2695_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2695\ufe0f" );
		name( "woman_heart_man", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f468, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc68" );
		name( "woman_heart_woman", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc69" );
		name( "woman_in_lotus_position", R.drawable.emj_1f9d8_200d_2640_fe0f, "\ud83e\uddd8\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_dark_skin_tone", R.drawable.emj_1f9d8_1f3ff_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_light_skin_tone", R.drawable.emj_1f9d8_1f3fb_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_medium_dark_skin_tone", R.drawable.emj_1f9d8_1f3fe_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_medium_light_skin_tone", R.drawable.emj_1f9d8_1f3fc_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_medium_skin_tone", R.drawable.emj_1f9d8_1f3fd_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_tone1", R.drawable.emj_1f9d8_1f3fb_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_tone2", R.drawable.emj_1f9d8_1f3fc_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_tone3", R.drawable.emj_1f9d8_1f3fd_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_tone4", R.drawable.emj_1f9d8_1f3fe_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_in_lotus_position_tone5", R.drawable.emj_1f9d8_1f3ff_200d_2640_fe0f, "\ud83e\uddd8\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room", R.drawable.emj_1f9d6_200d_2640_fe0f, "\ud83e\uddd6\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_dark_skin_tone", R.drawable.emj_1f9d6_1f3ff_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_light_skin_tone", R.drawable.emj_1f9d6_1f3fb_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffb\u200d\u2640\ufe0f" );
	}
	private static void init102(){
		name( "woman_in_steamy_room_medium_dark_skin_tone", R.drawable.emj_1f9d6_1f3fe_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_medium_light_skin_tone", R.drawable.emj_1f9d6_1f3fc_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_medium_skin_tone", R.drawable.emj_1f9d6_1f3fd_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_tone1", R.drawable.emj_1f9d6_1f3fb_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_tone2", R.drawable.emj_1f9d6_1f3fc_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_tone3", R.drawable.emj_1f9d6_1f3fd_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_tone4", R.drawable.emj_1f9d6_1f3fe_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_in_steamy_room_tone5", R.drawable.emj_1f9d6_1f3ff_200d_2640_fe0f, "\ud83e\uddd6\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_judge", R.drawable.emj_1f469_200d_2696_fe0f, "\ud83d\udc69\u200d\u2696\ufe0f" );
		name( "woman_judge_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_2696_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "woman_judge_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "woman_judge_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "woman_judge_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "woman_judge_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "woman_judge_tone1", R.drawable.emj_1f469_1f3fb_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2696\ufe0f" );
		name( "woman_judge_tone2", R.drawable.emj_1f469_1f3fc_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2696\ufe0f" );
		name( "woman_judge_tone3", R.drawable.emj_1f469_1f3fd_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2696\ufe0f" );
		name( "woman_judge_tone4", R.drawable.emj_1f469_1f3fe_200d_2696_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2696\ufe0f" );
		name( "woman_judge_tone5", R.drawable.emj_1f469_1f3ff_200d_2696_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2696\ufe0f" );
		name( "woman_juggling", R.drawable.emj_1f939_200d_2640_fe0f, "\ud83e\udd39\u200d\u2640\ufe0f" );
		name( "woman_juggling_dark_skin_tone", R.drawable.emj_1f939_1f3ff_200d_2640_fe0f, "\ud83e\udd39\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_juggling_light_skin_tone", R.drawable.emj_1f939_1f3fb_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_juggling_medium_dark_skin_tone", R.drawable.emj_1f939_1f3fe_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_juggling_medium_light_skin_tone", R.drawable.emj_1f939_1f3fc_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_juggling_medium_skin_tone", R.drawable.emj_1f939_1f3fd_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_juggling_tone1", R.drawable.emj_1f939_1f3fb_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_juggling_tone2", R.drawable.emj_1f939_1f3fc_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_juggling_tone3", R.drawable.emj_1f939_1f3fd_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_juggling_tone4", R.drawable.emj_1f939_1f3fe_200d_2640_fe0f, "\ud83e\udd39\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_juggling_tone5", R.drawable.emj_1f939_1f3ff_200d_2640_fe0f, "\ud83e\udd39\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_kiss_man", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f468, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc68" );
		name( "woman_kiss_woman", R.drawable.emj_1f469_200d_2764_fe0f_200d_1f48b_200d_1f469, "\ud83d\udc69\u200d\u2764\ufe0f\u200d\ud83d\udc8b\u200d\ud83d\udc69" );
		name( "woman_lifting_weights", R.drawable.emj_1f3cb_fe0f_200d_2640_fe0f, "\ud83c\udfcb\ufe0f\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_dark_skin_tone", R.drawable.emj_1f3cb_1f3ff_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_light_skin_tone", R.drawable.emj_1f3cb_1f3fb_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_medium_dark_skin_tone", R.drawable.emj_1f3cb_1f3fe_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_medium_light_skin_tone", R.drawable.emj_1f3cb_1f3fc_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_medium_skin_tone", R.drawable.emj_1f3cb_1f3fd_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_tone1", R.drawable.emj_1f3cb_1f3fb_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_tone2", R.drawable.emj_1f3cb_1f3fc_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_tone3", R.drawable.emj_1f3cb_1f3fd_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_tone4", R.drawable.emj_1f3cb_1f3fe_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_lifting_weights_tone5", R.drawable.emj_1f3cb_1f3ff_200d_2640_fe0f, "\ud83c\udfcb\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_light_skin_tone", R.drawable.emj_1f469_1f3fb, "\ud83d\udc69\ud83c\udffb" );
		name( "woman_mechanic", R.drawable.emj_1f469_200d_1f527, "\ud83d\udc69\u200d\ud83d\udd27" );
		name( "woman_mechanic_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f527, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd27" );
		name( "woman_mechanic_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f527, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd27" );
		name( "woman_mechanic_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f527, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd27" );
		name( "woman_mechanic_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f527, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd27" );
		name( "woman_mechanic_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f527, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd27" );
		name( "woman_mechanic_tone1", R.drawable.emj_1f469_1f3fb_200d_1f527, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd27" );
		name( "woman_mechanic_tone2", R.drawable.emj_1f469_1f3fc_200d_1f527, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd27" );
		name( "woman_mechanic_tone3", R.drawable.emj_1f469_1f3fd_200d_1f527, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd27" );
		name( "woman_mechanic_tone4", R.drawable.emj_1f469_1f3fe_200d_1f527, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd27" );
		name( "woman_mechanic_tone5", R.drawable.emj_1f469_1f3ff_200d_1f527, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd27" );
		name( "woman_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe, "\ud83d\udc69\ud83c\udffe" );
		name( "woman_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc, "\ud83d\udc69\ud83c\udffc" );
		name( "woman_medium_skin_tone", R.drawable.emj_1f469_1f3fd, "\ud83d\udc69\ud83c\udffd" );
		name( "woman_mountain_biking", R.drawable.emj_1f6b5_200d_2640_fe0f, "\ud83d\udeb5\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_dark_skin_tone", R.drawable.emj_1f6b5_1f3ff_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_light_skin_tone", R.drawable.emj_1f6b5_1f3fb_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_medium_dark_skin_tone", R.drawable.emj_1f6b5_1f3fe_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_medium_light_skin_tone", R.drawable.emj_1f6b5_1f3fc_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_medium_skin_tone", R.drawable.emj_1f6b5_1f3fd_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_tone1", R.drawable.emj_1f6b5_1f3fb_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_tone2", R.drawable.emj_1f6b5_1f3fc_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_tone3", R.drawable.emj_1f6b5_1f3fd_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_tone4", R.drawable.emj_1f6b5_1f3fe_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_mountain_biking_tone5", R.drawable.emj_1f6b5_1f3ff_200d_2640_fe0f, "\ud83d\udeb5\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_office_worker", R.drawable.emj_1f469_200d_1f4bc, "\ud83d\udc69\u200d\ud83d\udcbc" );
		name( "woman_office_worker_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f4bc, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "woman_office_worker_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f4bc, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "woman_office_worker_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f4bc, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "woman_office_worker_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f4bc, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "woman_office_worker_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f4bc, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "woman_office_worker_tone1", R.drawable.emj_1f469_1f3fb_200d_1f4bc, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbc" );
		name( "woman_office_worker_tone2", R.drawable.emj_1f469_1f3fc_200d_1f4bc, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbc" );
		name( "woman_office_worker_tone3", R.drawable.emj_1f469_1f3fd_200d_1f4bc, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbc" );
		name( "woman_office_worker_tone4", R.drawable.emj_1f469_1f3fe_200d_1f4bc, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbc" );
		name( "woman_office_worker_tone5", R.drawable.emj_1f469_1f3ff_200d_1f4bc, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbc" );
		name( "woman_pilot", R.drawable.emj_1f469_200d_2708_fe0f, "\ud83d\udc69\u200d\u2708\ufe0f" );
		name( "woman_pilot_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_2708_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "woman_pilot_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "woman_pilot_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "woman_pilot_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "woman_pilot_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2708\ufe0f" );
		name( "woman_pilot_tone1", R.drawable.emj_1f469_1f3fb_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffb\u200d\u2708\ufe0f" );
		name( "woman_pilot_tone2", R.drawable.emj_1f469_1f3fc_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffc\u200d\u2708\ufe0f" );
		name( "woman_pilot_tone3", R.drawable.emj_1f469_1f3fd_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffd\u200d\u2708\ufe0f" );
		name( "woman_pilot_tone4", R.drawable.emj_1f469_1f3fe_200d_2708_fe0f, "\ud83d\udc69\ud83c\udffe\u200d\u2708\ufe0f" );
		name( "woman_pilot_tone5", R.drawable.emj_1f469_1f3ff_200d_2708_fe0f, "\ud83d\udc69\ud83c\udfff\u200d\u2708\ufe0f" );
		name( "woman_playing_handball", R.drawable.emj_1f93e_200d_2640_fe0f, "\ud83e\udd3e\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_dark_skin_tone", R.drawable.emj_1f93e_1f3ff_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_light_skin_tone", R.drawable.emj_1f93e_1f3fb_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_medium_dark_skin_tone", R.drawable.emj_1f93e_1f3fe_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_medium_light_skin_tone", R.drawable.emj_1f93e_1f3fc_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_medium_skin_tone", R.drawable.emj_1f93e_1f3fd_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_tone1", R.drawable.emj_1f93e_1f3fb_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_tone2", R.drawable.emj_1f93e_1f3fc_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_tone3", R.drawable.emj_1f93e_1f3fd_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_playing_handball_tone4", R.drawable.emj_1f93e_1f3fe_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udffe\u200d\u2640\ufe0f" );
	}
	private static void init103(){
		name( "woman_playing_handball_tone5", R.drawable.emj_1f93e_1f3ff_200d_2640_fe0f, "\ud83e\udd3e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo", R.drawable.emj_1f93d_200d_2640_fe0f, "\ud83e\udd3d\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_dark_skin_tone", R.drawable.emj_1f93d_1f3ff_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_light_skin_tone", R.drawable.emj_1f93d_1f3fb_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_medium_dark_skin_tone", R.drawable.emj_1f93d_1f3fe_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_medium_light_skin_tone", R.drawable.emj_1f93d_1f3fc_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_medium_skin_tone", R.drawable.emj_1f93d_1f3fd_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_tone1", R.drawable.emj_1f93d_1f3fb_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_tone2", R.drawable.emj_1f93d_1f3fc_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_tone3", R.drawable.emj_1f93d_1f3fd_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_tone4", R.drawable.emj_1f93d_1f3fe_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_playing_water_polo_tone5", R.drawable.emj_1f93d_1f3ff_200d_2640_fe0f, "\ud83e\udd3d\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_police_officer", R.drawable.emj_1f46e_200d_2640_fe0f, "\ud83d\udc6e\u200d\u2640\ufe0f" );
		name( "woman_police_officer_dark_skin_tone", R.drawable.emj_1f46e_1f3ff_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_police_officer_light_skin_tone", R.drawable.emj_1f46e_1f3fb_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_police_officer_medium_dark_skin_tone", R.drawable.emj_1f46e_1f3fe_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_police_officer_medium_light_skin_tone", R.drawable.emj_1f46e_1f3fc_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_police_officer_medium_skin_tone", R.drawable.emj_1f46e_1f3fd_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_police_officer_tone1", R.drawable.emj_1f46e_1f3fb_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_police_officer_tone2", R.drawable.emj_1f46e_1f3fc_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_police_officer_tone3", R.drawable.emj_1f46e_1f3fd_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_police_officer_tone4", R.drawable.emj_1f46e_1f3fe_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_police_officer_tone5", R.drawable.emj_1f46e_1f3ff_200d_2640_fe0f, "\ud83d\udc6e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_pouting", R.drawable.emj_1f64e_200d_2640_fe0f, "\ud83d\ude4e\u200d\u2640\ufe0f" );
		name( "woman_pouting_dark_skin_tone", R.drawable.emj_1f64e_1f3ff_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_pouting_light_skin_tone", R.drawable.emj_1f64e_1f3fb_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_pouting_medium_dark_skin_tone", R.drawable.emj_1f64e_1f3fe_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_pouting_medium_light_skin_tone", R.drawable.emj_1f64e_1f3fc_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_pouting_medium_skin_tone", R.drawable.emj_1f64e_1f3fd_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_pouting_tone1", R.drawable.emj_1f64e_1f3fb_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_pouting_tone2", R.drawable.emj_1f64e_1f3fc_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_pouting_tone3", R.drawable.emj_1f64e_1f3fd_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_pouting_tone4", R.drawable.emj_1f64e_1f3fe_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_pouting_tone5", R.drawable.emj_1f64e_1f3ff_200d_2640_fe0f, "\ud83d\ude4e\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_raising_hand", R.drawable.emj_1f64b_200d_2640_fe0f, "\ud83d\ude4b\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_dark_skin_tone", R.drawable.emj_1f64b_1f3ff_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_light_skin_tone", R.drawable.emj_1f64b_1f3fb_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_medium_dark_skin_tone", R.drawable.emj_1f64b_1f3fe_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_medium_light_skin_tone", R.drawable.emj_1f64b_1f3fc_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_medium_skin_tone", R.drawable.emj_1f64b_1f3fd_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_tone1", R.drawable.emj_1f64b_1f3fb_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_tone2", R.drawable.emj_1f64b_1f3fc_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_tone3", R.drawable.emj_1f64b_1f3fd_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_tone4", R.drawable.emj_1f64b_1f3fe_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_raising_hand_tone5", R.drawable.emj_1f64b_1f3ff_200d_2640_fe0f, "\ud83d\ude4b\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat", R.drawable.emj_1f6a3_200d_2640_fe0f, "\ud83d\udea3\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_dark_skin_tone", R.drawable.emj_1f6a3_1f3ff_200d_2640_fe0f, "\ud83d\udea3\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_light_skin_tone", R.drawable.emj_1f6a3_1f3fb_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_medium_dark_skin_tone", R.drawable.emj_1f6a3_1f3fe_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_medium_light_skin_tone", R.drawable.emj_1f6a3_1f3fc_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_medium_skin_tone", R.drawable.emj_1f6a3_1f3fd_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_tone1", R.drawable.emj_1f6a3_1f3fb_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_tone2", R.drawable.emj_1f6a3_1f3fc_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_tone3", R.drawable.emj_1f6a3_1f3fd_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_tone4", R.drawable.emj_1f6a3_1f3fe_200d_2640_fe0f, "\ud83d\udea3\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_rowing_boat_tone5", R.drawable.emj_1f6a3_1f3ff_200d_2640_fe0f, "\ud83d\udea3\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_running", R.drawable.emj_1f3c3_200d_2640_fe0f, "\ud83c\udfc3\u200d\u2640\ufe0f" );
		name( "woman_running_dark_skin_tone", R.drawable.emj_1f3c3_1f3ff_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_running_light_skin_tone", R.drawable.emj_1f3c3_1f3fb_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_running_medium_dark_skin_tone", R.drawable.emj_1f3c3_1f3fe_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_running_medium_light_skin_tone", R.drawable.emj_1f3c3_1f3fc_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_running_medium_skin_tone", R.drawable.emj_1f3c3_1f3fd_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_running_tone1", R.drawable.emj_1f3c3_1f3fb_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_running_tone2", R.drawable.emj_1f3c3_1f3fc_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_running_tone3", R.drawable.emj_1f3c3_1f3fd_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_running_tone4", R.drawable.emj_1f3c3_1f3fe_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_running_tone5", R.drawable.emj_1f3c3_1f3ff_200d_2640_fe0f, "\ud83c\udfc3\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_scientist", R.drawable.emj_1f469_200d_1f52c, "\ud83d\udc69\u200d\ud83d\udd2c" );
		name( "woman_scientist_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f52c, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "woman_scientist_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f52c, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "woman_scientist_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f52c, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "woman_scientist_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f52c, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd2c" );
		name( "woman_scientist_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f52c, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "woman_scientist_tone1", R.drawable.emj_1f469_1f3fb_200d_1f52c, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udd2c" );
		name( "woman_scientist_tone2", R.drawable.emj_1f469_1f3fc_200d_1f52c, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udd2c" );
		name( "woman_scientist_tone3", R.drawable.emj_1f469_1f3fd_200d_1f52c, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udd2c" );
		name( "woman_scientist_tone4", R.drawable.emj_1f469_1f3fe_200d_1f52c, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udd2c" );
		name( "woman_scientist_tone5", R.drawable.emj_1f469_1f3ff_200d_1f52c, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udd2c" );
		name( "woman_shrugging", R.drawable.emj_1f937_200d_2640_fe0f, "\ud83e\udd37\u200d\u2640\ufe0f" );
		name( "woman_shrugging_dark_skin_tone", R.drawable.emj_1f937_1f3ff_200d_2640_fe0f, "\ud83e\udd37\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_shrugging_light_skin_tone", R.drawable.emj_1f937_1f3fb_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_shrugging_medium_dark_skin_tone", R.drawable.emj_1f937_1f3fe_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_shrugging_medium_light_skin_tone", R.drawable.emj_1f937_1f3fc_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_shrugging_medium_skin_tone", R.drawable.emj_1f937_1f3fd_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_shrugging_tone1", R.drawable.emj_1f937_1f3fb_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_shrugging_tone2", R.drawable.emj_1f937_1f3fc_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_shrugging_tone3", R.drawable.emj_1f937_1f3fd_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_shrugging_tone4", R.drawable.emj_1f937_1f3fe_200d_2640_fe0f, "\ud83e\udd37\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_shrugging_tone5", R.drawable.emj_1f937_1f3ff_200d_2640_fe0f, "\ud83e\udd37\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_singer", R.drawable.emj_1f469_200d_1f3a4, "\ud83d\udc69\u200d\ud83c\udfa4" );
		name( "woman_singer_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3a4, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "woman_singer_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3a4, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "woman_singer_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3a4, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "woman_singer_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3a4, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "woman_singer_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3a4, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "woman_singer_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3a4, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfa4" );
		name( "woman_singer_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3a4, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfa4" );
		name( "woman_singer_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3a4, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfa4" );
		name( "woman_singer_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3a4, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfa4" );
		name( "woman_singer_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3a4, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfa4" );
		name( "woman_student", R.drawable.emj_1f469_200d_1f393, "\ud83d\udc69\u200d\ud83c\udf93" );
	}
	private static void init104(){
		name( "woman_student_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f393, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf93" );
		name( "woman_student_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f393, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf93" );
		name( "woman_student_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f393, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf93" );
		name( "woman_student_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f393, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf93" );
		name( "woman_student_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f393, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf93" );
		name( "woman_student_tone1", R.drawable.emj_1f469_1f3fb_200d_1f393, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udf93" );
		name( "woman_student_tone2", R.drawable.emj_1f469_1f3fc_200d_1f393, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udf93" );
		name( "woman_student_tone3", R.drawable.emj_1f469_1f3fd_200d_1f393, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udf93" );
		name( "woman_student_tone4", R.drawable.emj_1f469_1f3fe_200d_1f393, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udf93" );
		name( "woman_student_tone5", R.drawable.emj_1f469_1f3ff_200d_1f393, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udf93" );
		name( "woman_surfing", R.drawable.emj_1f3c4_200d_2640_fe0f, "\ud83c\udfc4\u200d\u2640\ufe0f" );
		name( "woman_surfing_dark_skin_tone", R.drawable.emj_1f3c4_1f3ff_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_surfing_light_skin_tone", R.drawable.emj_1f3c4_1f3fb_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_surfing_medium_dark_skin_tone", R.drawable.emj_1f3c4_1f3fe_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_surfing_medium_light_skin_tone", R.drawable.emj_1f3c4_1f3fc_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_surfing_medium_skin_tone", R.drawable.emj_1f3c4_1f3fd_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_surfing_tone1", R.drawable.emj_1f3c4_1f3fb_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_surfing_tone2", R.drawable.emj_1f3c4_1f3fc_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_surfing_tone3", R.drawable.emj_1f3c4_1f3fd_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_surfing_tone4", R.drawable.emj_1f3c4_1f3fe_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_surfing_tone5", R.drawable.emj_1f3c4_1f3ff_200d_2640_fe0f, "\ud83c\udfc4\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_swimming", R.drawable.emj_1f3ca_200d_2640_fe0f, "\ud83c\udfca\u200d\u2640\ufe0f" );
		name( "woman_swimming_dark_skin_tone", R.drawable.emj_1f3ca_1f3ff_200d_2640_fe0f, "\ud83c\udfca\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_swimming_light_skin_tone", R.drawable.emj_1f3ca_1f3fb_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_swimming_medium_dark_skin_tone", R.drawable.emj_1f3ca_1f3fe_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_swimming_medium_light_skin_tone", R.drawable.emj_1f3ca_1f3fc_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_swimming_medium_skin_tone", R.drawable.emj_1f3ca_1f3fd_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_swimming_tone1", R.drawable.emj_1f3ca_1f3fb_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_swimming_tone2", R.drawable.emj_1f3ca_1f3fc_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_swimming_tone3", R.drawable.emj_1f3ca_1f3fd_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_swimming_tone4", R.drawable.emj_1f3ca_1f3fe_200d_2640_fe0f, "\ud83c\udfca\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_swimming_tone5", R.drawable.emj_1f3ca_1f3ff_200d_2640_fe0f, "\ud83c\udfca\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_teacher", R.drawable.emj_1f469_200d_1f3eb, "\ud83d\udc69\u200d\ud83c\udfeb" );
		name( "woman_teacher_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f3eb, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "woman_teacher_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f3eb, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "woman_teacher_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f3eb, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "woman_teacher_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f3eb, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "woman_teacher_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f3eb, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "woman_teacher_tone1", R.drawable.emj_1f469_1f3fb_200d_1f3eb, "\ud83d\udc69\ud83c\udffb\u200d\ud83c\udfeb" );
		name( "woman_teacher_tone2", R.drawable.emj_1f469_1f3fc_200d_1f3eb, "\ud83d\udc69\ud83c\udffc\u200d\ud83c\udfeb" );
		name( "woman_teacher_tone3", R.drawable.emj_1f469_1f3fd_200d_1f3eb, "\ud83d\udc69\ud83c\udffd\u200d\ud83c\udfeb" );
		name( "woman_teacher_tone4", R.drawable.emj_1f469_1f3fe_200d_1f3eb, "\ud83d\udc69\ud83c\udffe\u200d\ud83c\udfeb" );
		name( "woman_teacher_tone5", R.drawable.emj_1f469_1f3ff_200d_1f3eb, "\ud83d\udc69\ud83c\udfff\u200d\ud83c\udfeb" );
		name( "woman_technologist", R.drawable.emj_1f469_200d_1f4bb, "\ud83d\udc69\u200d\ud83d\udcbb" );
		name( "woman_technologist_dark_skin_tone", R.drawable.emj_1f469_1f3ff_200d_1f4bb, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "woman_technologist_light_skin_tone", R.drawable.emj_1f469_1f3fb_200d_1f4bb, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "woman_technologist_medium_dark_skin_tone", R.drawable.emj_1f469_1f3fe_200d_1f4bb, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "woman_technologist_medium_light_skin_tone", R.drawable.emj_1f469_1f3fc_200d_1f4bb, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "woman_technologist_medium_skin_tone", R.drawable.emj_1f469_1f3fd_200d_1f4bb, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "woman_technologist_tone1", R.drawable.emj_1f469_1f3fb_200d_1f4bb, "\ud83d\udc69\ud83c\udffb\u200d\ud83d\udcbb" );
		name( "woman_technologist_tone2", R.drawable.emj_1f469_1f3fc_200d_1f4bb, "\ud83d\udc69\ud83c\udffc\u200d\ud83d\udcbb" );
		name( "woman_technologist_tone3", R.drawable.emj_1f469_1f3fd_200d_1f4bb, "\ud83d\udc69\ud83c\udffd\u200d\ud83d\udcbb" );
		name( "woman_technologist_tone4", R.drawable.emj_1f469_1f3fe_200d_1f4bb, "\ud83d\udc69\ud83c\udffe\u200d\ud83d\udcbb" );
		name( "woman_technologist_tone5", R.drawable.emj_1f469_1f3ff_200d_1f4bb, "\ud83d\udc69\ud83c\udfff\u200d\ud83d\udcbb" );
		name( "woman_tipping_hand", R.drawable.emj_1f481_200d_2640_fe0f, "\ud83d\udc81\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_dark_skin_tone", R.drawable.emj_1f481_1f3ff_200d_2640_fe0f, "\ud83d\udc81\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_light_skin_tone", R.drawable.emj_1f481_1f3fb_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_medium_dark_skin_tone", R.drawable.emj_1f481_1f3fe_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_medium_light_skin_tone", R.drawable.emj_1f481_1f3fc_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_medium_skin_tone", R.drawable.emj_1f481_1f3fd_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_tone1", R.drawable.emj_1f481_1f3fb_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_tone2", R.drawable.emj_1f481_1f3fc_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_tone3", R.drawable.emj_1f481_1f3fd_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_tone4", R.drawable.emj_1f481_1f3fe_200d_2640_fe0f, "\ud83d\udc81\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_tipping_hand_tone5", R.drawable.emj_1f481_1f3ff_200d_2640_fe0f, "\ud83d\udc81\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_tone1", R.drawable.emj_1f469_1f3fb, "\ud83d\udc69\ud83c\udffb" );
		name( "woman_tone2", R.drawable.emj_1f469_1f3fc, "\ud83d\udc69\ud83c\udffc" );
		name( "woman_tone3", R.drawable.emj_1f469_1f3fd, "\ud83d\udc69\ud83c\udffd" );
		name( "woman_tone4", R.drawable.emj_1f469_1f3fe, "\ud83d\udc69\ud83c\udffe" );
		name( "woman_tone5", R.drawable.emj_1f469_1f3ff, "\ud83d\udc69\ud83c\udfff" );
		name( "woman_walking", R.drawable.emj_1f6b6_200d_2640_fe0f, "\ud83d\udeb6\u200d\u2640\ufe0f" );
		name( "woman_walking_dark_skin_tone", R.drawable.emj_1f6b6_1f3ff_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_walking_light_skin_tone", R.drawable.emj_1f6b6_1f3fb_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_walking_medium_dark_skin_tone", R.drawable.emj_1f6b6_1f3fe_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_walking_medium_light_skin_tone", R.drawable.emj_1f6b6_1f3fc_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_walking_medium_skin_tone", R.drawable.emj_1f6b6_1f3fd_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_walking_tone1", R.drawable.emj_1f6b6_1f3fb_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_walking_tone2", R.drawable.emj_1f6b6_1f3fc_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_walking_tone3", R.drawable.emj_1f6b6_1f3fd_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_walking_tone4", R.drawable.emj_1f6b6_1f3fe_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_walking_tone5", R.drawable.emj_1f6b6_1f3ff_200d_2640_fe0f, "\ud83d\udeb6\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban", R.drawable.emj_1f473_200d_2640_fe0f, "\ud83d\udc73\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_dark_skin_tone", R.drawable.emj_1f473_1f3ff_200d_2640_fe0f, "\ud83d\udc73\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_light_skin_tone", R.drawable.emj_1f473_1f3fb_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_medium_dark_skin_tone", R.drawable.emj_1f473_1f3fe_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_medium_light_skin_tone", R.drawable.emj_1f473_1f3fc_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_medium_skin_tone", R.drawable.emj_1f473_1f3fd_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_tone1", R.drawable.emj_1f473_1f3fb_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffb\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_tone2", R.drawable.emj_1f473_1f3fc_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffc\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_tone3", R.drawable.emj_1f473_1f3fd_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffd\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_tone4", R.drawable.emj_1f473_1f3fe_200d_2640_fe0f, "\ud83d\udc73\ud83c\udffe\u200d\u2640\ufe0f" );
		name( "woman_wearing_turban_tone5", R.drawable.emj_1f473_1f3ff_200d_2640_fe0f, "\ud83d\udc73\ud83c\udfff\u200d\u2640\ufe0f" );
		name( "woman_with_bunny_ears_partying", R.drawable.emj_1f46f_200d_2640_fe0f, "\ud83d\udc6f\u200d\u2640\ufe0f" );
		name( "woman_woman_boy", R.drawable.emj_1f469_200d_1f469_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc66" );
		name( "woman_woman_boy_boy", R.drawable.emj_1f469_200d_1f469_200d_1f466_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc66\u200d\ud83d\udc66" );
		name( "woman_woman_girl", R.drawable.emj_1f469_200d_1f469_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67" );
		name( "woman_woman_girl_boy", R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f466, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc66" );
		name( "woman_woman_girl_girl", R.drawable.emj_1f469_200d_1f469_200d_1f467_200d_1f467, "\ud83d\udc69\u200d\ud83d\udc69\u200d\ud83d\udc67\u200d\ud83d\udc67" );
		name( "woman_wrestling", R.drawable.emj_1f93c_200d_2640_fe0f, "\ud83e\udd3c\u200d\u2640\ufe0f" );
		name( "womans_clothes", R.drawable.emj_1f45a, "\ud83d\udc5a" );
		name( "womans_hat", R.drawable.emj_1f452, "\ud83d\udc52" );
	}
	private static void init105(){
		name( "women_with_bunny_ears_partying", R.drawable.emj_1f46f_200d_2640_fe0f, "\ud83d\udc6f\u200d\u2640\ufe0f" );
		name( "women_wrestling", R.drawable.emj_1f93c_200d_2640_fe0f, "\ud83e\udd3c\u200d\u2640\ufe0f" );
		name( "womens", R.drawable.emj_1f6ba, "\ud83d\udeba" );
		name( "world_map", R.drawable.emj_1f5fa_fe0f, "\ud83d\uddfa\ufe0f" );
		name( "worried", R.drawable.emj_1f61f, "\ud83d\ude1f" );
		name( "worship_symbol", R.drawable.emj_1f6d0, "\ud83d\uded0" );
		name( "wrench", R.drawable.emj_1f527, "\ud83d\udd27" );
		name( "wrestlers", R.drawable.emj_1f93c, "\ud83e\udd3c" );
		name( "wrestling", R.drawable.emj_1f93c, "\ud83e\udd3c" );
		name( "writing_hand", R.drawable.emj_270d_fe0f, "\u270d\ufe0f" );
		name( "writing_hand_dark_skin_tone", R.drawable.emj_270d_1f3ff, "\u270d\ud83c\udfff" );
		name( "writing_hand_light_skin_tone", R.drawable.emj_270d_1f3fb, "\u270d\ud83c\udffb" );
		name( "writing_hand_medium_dark_skin_tone", R.drawable.emj_270d_1f3fe, "\u270d\ud83c\udffe" );
		name( "writing_hand_medium_light_skin_tone", R.drawable.emj_270d_1f3fc, "\u270d\ud83c\udffc" );
		name( "writing_hand_medium_skin_tone", R.drawable.emj_270d_1f3fd, "\u270d\ud83c\udffd" );
		name( "writing_hand_tone1", R.drawable.emj_270d_1f3fb, "\u270d\ud83c\udffb" );
		name( "writing_hand_tone2", R.drawable.emj_270d_1f3fc, "\u270d\ud83c\udffc" );
		name( "writing_hand_tone3", R.drawable.emj_270d_1f3fd, "\u270d\ud83c\udffd" );
		name( "writing_hand_tone4", R.drawable.emj_270d_1f3fe, "\u270d\ud83c\udffe" );
		name( "writing_hand_tone5", R.drawable.emj_270d_1f3ff, "\u270d\ud83c\udfff" );
		name( "ws", R.drawable.emj_1f1fc_1f1f8, "\ud83c\uddfc\ud83c\uddf8" );
		name( "x", R.drawable.emj_274c, "\u274c" );
		name( "xk", R.drawable.emj_1f1fd_1f1f0, "\ud83c\uddfd\ud83c\uddf0" );
		name( "ye", R.drawable.emj_1f1fe_1f1ea, "\ud83c\uddfe\ud83c\uddea" );
		name( "yellow_heart", R.drawable.emj_1f49b, "\ud83d\udc9b" );
		name( "yen", R.drawable.emj_1f4b4, "\ud83d\udcb4" );
		name( "yin_yang", R.drawable.emj_262f_fe0f, "\u262f\ufe0f" );
		name( "yt", R.drawable.emj_1f1fe_1f1f9, "\ud83c\uddfe\ud83c\uddf9" );
		name( "yum", R.drawable.emj_1f60b, "\ud83d\ude0b" );
		name( "za", R.drawable.emj_1f1ff_1f1e6, "\ud83c\uddff\ud83c\udde6" );
		name( "zany_face", R.drawable.emj_1f92a, "\ud83e\udd2a" );
		name( "zap", R.drawable.emj_26a1, "\u26a1" );
		name( "zebra_face", R.drawable.emj_1f993, "\ud83e\udd93" );
		name( "zero", R.drawable.emj_0030_fe0f_20e3, "\u0030\ufe0f\u20e3" );
		name( "zipper_mouth", R.drawable.emj_1f910, "\ud83e\udd10" );
		name( "zipper_mouth_face", R.drawable.emj_1f910, "\ud83e\udd10" );
		name( "zm", R.drawable.emj_1f1ff_1f1f2, "\ud83c\uddff\ud83c\uddf2" );
		name( "zombie", R.drawable.emj_1f9df, "\ud83e\udddf" );
		name( "zw", R.drawable.emj_1f1ff_1f1fc, "\ud83c\uddff\ud83c\uddfc" );
		name( "zzz", R.drawable.emj_1f4a4, "\ud83d\udca4" );
		category(CATEGORY_PEOPLE, "grinning");
		category(CATEGORY_PEOPLE, "grin");
		category(CATEGORY_PEOPLE, "joy");
		category(CATEGORY_PEOPLE, "rolling_on_the_floor_laughing");
		category(CATEGORY_PEOPLE, "smiley");
		category(CATEGORY_PEOPLE, "smile");
		category(CATEGORY_PEOPLE, "sweat_smile");
		category(CATEGORY_PEOPLE, "laughing");
		category(CATEGORY_PEOPLE, "wink");
		category(CATEGORY_PEOPLE, "blush");
		category(CATEGORY_PEOPLE, "yum");
		category(CATEGORY_PEOPLE, "sunglasses");
		category(CATEGORY_PEOPLE, "heart_eyes");
		category(CATEGORY_PEOPLE, "kissing_heart");
		category(CATEGORY_PEOPLE, "kissing");
		category(CATEGORY_PEOPLE, "kissing_smiling_eyes");
		category(CATEGORY_PEOPLE, "kissing_closed_eyes");
		category(CATEGORY_PEOPLE, "relaxed");
		category(CATEGORY_PEOPLE, "slightly_smiling_face");
		category(CATEGORY_PEOPLE, "hugging_face");
		category(CATEGORY_PEOPLE, "star_struck");
		category(CATEGORY_PEOPLE, "thinking_face");
		category(CATEGORY_PEOPLE, "face_with_raised_eyebrow");
		category(CATEGORY_PEOPLE, "neutral_face");
		category(CATEGORY_PEOPLE, "expressionless");
		category(CATEGORY_PEOPLE, "no_mouth");
		category(CATEGORY_PEOPLE, "face_with_rolling_eyes");
		category(CATEGORY_PEOPLE, "smirk");
		category(CATEGORY_PEOPLE, "persevere");
		category(CATEGORY_PEOPLE, "disappointed_relieved");
		category(CATEGORY_PEOPLE, "open_mouth");
		category(CATEGORY_PEOPLE, "zipper_mouth_face");
		category(CATEGORY_PEOPLE, "hushed");
		category(CATEGORY_PEOPLE, "sleepy");
		category(CATEGORY_PEOPLE, "tired_face");
		category(CATEGORY_PEOPLE, "sleeping");
		category(CATEGORY_PEOPLE, "relieved");
		category(CATEGORY_PEOPLE, "stuck_out_tongue");
		category(CATEGORY_PEOPLE, "stuck_out_tongue_winking_eye");
		category(CATEGORY_PEOPLE, "stuck_out_tongue_closed_eyes");
		category(CATEGORY_PEOPLE, "drooling_face");
		category(CATEGORY_PEOPLE, "unamused");
		category(CATEGORY_PEOPLE, "sweat");
		category(CATEGORY_PEOPLE, "pensive");
		category(CATEGORY_PEOPLE, "confused");
		category(CATEGORY_PEOPLE, "upside_down_face");
		category(CATEGORY_PEOPLE, "money_mouth_face");
		category(CATEGORY_PEOPLE, "astonished");
		category(CATEGORY_PEOPLE, "white_frowning_face");
		category(CATEGORY_PEOPLE, "slightly_frowning_face");
		category(CATEGORY_PEOPLE, "confounded");
		category(CATEGORY_PEOPLE, "disappointed");
		category(CATEGORY_PEOPLE, "worried");
		category(CATEGORY_PEOPLE, "triumph");
		category(CATEGORY_PEOPLE, "cry");
		category(CATEGORY_PEOPLE, "sob");
		category(CATEGORY_PEOPLE, "frowning");
		category(CATEGORY_PEOPLE, "anguished");
		category(CATEGORY_PEOPLE, "fearful");
		category(CATEGORY_PEOPLE, "weary");
		category(CATEGORY_PEOPLE, "exploding_head");
	}
	private static void init106(){
		category(CATEGORY_PEOPLE, "grimacing");
		category(CATEGORY_PEOPLE, "cold_sweat");
		category(CATEGORY_PEOPLE, "scream");
		category(CATEGORY_PEOPLE, "flushed");
		category(CATEGORY_PEOPLE, "zany_face");
		category(CATEGORY_PEOPLE, "dizzy_face");
		category(CATEGORY_PEOPLE, "rage");
		category(CATEGORY_PEOPLE, "angry");
		category(CATEGORY_PEOPLE, "face_with_symbols_on_mouth");
		category(CATEGORY_PEOPLE, "mask");
		category(CATEGORY_PEOPLE, "face_with_thermometer");
		category(CATEGORY_PEOPLE, "face_with_head_bandage");
		category(CATEGORY_PEOPLE, "nauseated_face");
		category(CATEGORY_PEOPLE, "face_vomiting");
		category(CATEGORY_PEOPLE, "sneezing_face");
		category(CATEGORY_PEOPLE, "innocent");
		category(CATEGORY_PEOPLE, "face_with_cowboy_hat");
		category(CATEGORY_PEOPLE, "clown_face");
		category(CATEGORY_PEOPLE, "lying_face");
		category(CATEGORY_PEOPLE, "shushing_face");
		category(CATEGORY_PEOPLE, "face_with_hand_over_mouth");
		category(CATEGORY_PEOPLE, "face_with_monocle");
		category(CATEGORY_PEOPLE, "nerd_face");
		category(CATEGORY_PEOPLE, "smiling_imp");
		category(CATEGORY_PEOPLE, "imp");
		category(CATEGORY_PEOPLE, "japanese_ogre");
		category(CATEGORY_PEOPLE, "japanese_goblin");
		category(CATEGORY_PEOPLE, "skull");
		category(CATEGORY_PEOPLE, "skull_and_crossbones");
		category(CATEGORY_PEOPLE, "ghost");
		category(CATEGORY_PEOPLE, "alien");
		category(CATEGORY_PEOPLE, "space_invader");
		category(CATEGORY_PEOPLE, "robot_face");
		category(CATEGORY_PEOPLE, "hankey");
		category(CATEGORY_PEOPLE, "smiley_cat");
		category(CATEGORY_PEOPLE, "smile_cat");
		category(CATEGORY_PEOPLE, "joy_cat");
		category(CATEGORY_PEOPLE, "heart_eyes_cat");
		category(CATEGORY_PEOPLE, "smirk_cat");
		category(CATEGORY_PEOPLE, "kissing_cat");
		category(CATEGORY_PEOPLE, "scream_cat");
		category(CATEGORY_PEOPLE, "crying_cat_face");
		category(CATEGORY_PEOPLE, "pouting_cat");
		category(CATEGORY_PEOPLE, "see_no_evil");
		category(CATEGORY_PEOPLE, "hear_no_evil");
		category(CATEGORY_PEOPLE, "speak_no_evil");
		category(CATEGORY_PEOPLE, "baby");
		category(CATEGORY_PEOPLE, "child");
		category(CATEGORY_PEOPLE, "boy");
		category(CATEGORY_PEOPLE, "girl");
		category(CATEGORY_PEOPLE, "adult");
		category(CATEGORY_PEOPLE, "man");
		category(CATEGORY_PEOPLE, "woman");
		category(CATEGORY_PEOPLE, "older_adult");
		category(CATEGORY_PEOPLE, "older_man");
		category(CATEGORY_PEOPLE, "older_woman");
		category(CATEGORY_PEOPLE, "male_doctor");
		category(CATEGORY_PEOPLE, "female_doctor");
		category(CATEGORY_PEOPLE, "male_student");
		category(CATEGORY_PEOPLE, "female_student");
		category(CATEGORY_PEOPLE, "male_teacher");
		category(CATEGORY_PEOPLE, "female_teacher");
		category(CATEGORY_PEOPLE, "male_judge");
		category(CATEGORY_PEOPLE, "female_judge");
		category(CATEGORY_PEOPLE, "male_farmer");
		category(CATEGORY_PEOPLE, "female_farmer");
		category(CATEGORY_PEOPLE, "male_cook");
		category(CATEGORY_PEOPLE, "female_cook");
		category(CATEGORY_PEOPLE, "male_mechanic");
		category(CATEGORY_PEOPLE, "female_mechanic");
		category(CATEGORY_PEOPLE, "male_factory_worker");
		category(CATEGORY_PEOPLE, "female_factory_worker");
		category(CATEGORY_PEOPLE, "male_office_worker");
		category(CATEGORY_PEOPLE, "female_office_worker");
		category(CATEGORY_PEOPLE, "male_scientist");
		category(CATEGORY_PEOPLE, "female_scientist");
		category(CATEGORY_PEOPLE, "male_technologist");
		category(CATEGORY_PEOPLE, "female_technologist");
		category(CATEGORY_PEOPLE, "male_singer");
		category(CATEGORY_PEOPLE, "female_singer");
		category(CATEGORY_PEOPLE, "male_artist");
		category(CATEGORY_PEOPLE, "female_artist");
		category(CATEGORY_PEOPLE, "male_pilot");
		category(CATEGORY_PEOPLE, "female_pilot");
		category(CATEGORY_PEOPLE, "male_astronaut");
		category(CATEGORY_PEOPLE, "female_astronaut");
		category(CATEGORY_PEOPLE, "male_firefighter");
		category(CATEGORY_PEOPLE, "female_firefighter");
		category(CATEGORY_PEOPLE, "cop");
		category(CATEGORY_PEOPLE, "male_police_officer");
		category(CATEGORY_PEOPLE, "female_police_officer");
		category(CATEGORY_PEOPLE, "sleuth_or_spy");
		category(CATEGORY_PEOPLE, "male_detective");
		category(CATEGORY_PEOPLE, "female_detective");
		category(CATEGORY_PEOPLE, "guardsman");
		category(CATEGORY_PEOPLE, "male_guard");
		category(CATEGORY_PEOPLE, "female_guard");
		category(CATEGORY_PEOPLE, "construction_worker");
		category(CATEGORY_PEOPLE, "male_construction_worker");
		category(CATEGORY_PEOPLE, "female_construction_worker");
		category(CATEGORY_PEOPLE, "prince");
	}
	private static void init107(){
		category(CATEGORY_PEOPLE, "princess");
		category(CATEGORY_PEOPLE, "man_with_turban");
		category(CATEGORY_PEOPLE, "man_wearing_turban");
		category(CATEGORY_PEOPLE, "woman_wearing_turban");
		category(CATEGORY_PEOPLE, "man_with_gua_pi_mao");
		category(CATEGORY_PEOPLE, "person_with_headscarf");
		category(CATEGORY_PEOPLE, "bearded_person");
		category(CATEGORY_PEOPLE, "person_with_blond_hair");
		category(CATEGORY_PEOPLE, "blond_haired_man");
		category(CATEGORY_PEOPLE, "blond_haired_woman");
		category(CATEGORY_PEOPLE, "man_in_tuxedo");
		category(CATEGORY_PEOPLE, "bride_with_veil");
		category(CATEGORY_PEOPLE, "pregnant_woman");
		category(CATEGORY_PEOPLE, "breast_feeding");
		category(CATEGORY_PEOPLE, "angel");
		category(CATEGORY_PEOPLE, "santa");
		category(CATEGORY_PEOPLE, "mrs_claus");
		category(CATEGORY_PEOPLE, "mage");
		category(CATEGORY_PEOPLE, "female_mage");
		category(CATEGORY_PEOPLE, "male_mage");
		category(CATEGORY_PEOPLE, "fairy");
		category(CATEGORY_PEOPLE, "female_fairy");
		category(CATEGORY_PEOPLE, "male_fairy");
		category(CATEGORY_PEOPLE, "vampire");
		category(CATEGORY_PEOPLE, "female_vampire");
		category(CATEGORY_PEOPLE, "male_vampire");
		category(CATEGORY_PEOPLE, "merperson");
		category(CATEGORY_PEOPLE, "mermaid");
		category(CATEGORY_PEOPLE, "merman");
		category(CATEGORY_PEOPLE, "elf");
		category(CATEGORY_PEOPLE, "female_elf");
		category(CATEGORY_PEOPLE, "male_elf");
		category(CATEGORY_PEOPLE, "genie");
		category(CATEGORY_PEOPLE, "female_genie");
		category(CATEGORY_PEOPLE, "male_genie");
		category(CATEGORY_PEOPLE, "zombie");
		category(CATEGORY_PEOPLE, "female_zombie");
		category(CATEGORY_PEOPLE, "male_zombie");
		category(CATEGORY_PEOPLE, "person_frowning");
		category(CATEGORY_PEOPLE, "man_frowning");
		category(CATEGORY_PEOPLE, "woman_frowning");
		category(CATEGORY_PEOPLE, "person_with_pouting_face");
		category(CATEGORY_PEOPLE, "man_pouting");
		category(CATEGORY_PEOPLE, "woman_pouting");
		category(CATEGORY_PEOPLE, "no_good");
		category(CATEGORY_PEOPLE, "man_gesturing_no");
		category(CATEGORY_PEOPLE, "woman_gesturing_no");
		category(CATEGORY_PEOPLE, "ok_woman");
		category(CATEGORY_PEOPLE, "man_gesturing_ok");
		category(CATEGORY_PEOPLE, "woman_gesturing_ok");
		category(CATEGORY_PEOPLE, "information_desk_person");
		category(CATEGORY_PEOPLE, "man_tipping_hand");
		category(CATEGORY_PEOPLE, "woman_tipping_hand");
		category(CATEGORY_PEOPLE, "raising_hand");
		category(CATEGORY_PEOPLE, "man_raising_hand");
		category(CATEGORY_PEOPLE, "woman_raising_hand");
		category(CATEGORY_PEOPLE, "bow");
		category(CATEGORY_PEOPLE, "man_bowing");
		category(CATEGORY_PEOPLE, "woman_bowing");
		category(CATEGORY_PEOPLE, "face_palm");
		category(CATEGORY_PEOPLE, "man_facepalming");
		category(CATEGORY_PEOPLE, "woman_facepalming");
		category(CATEGORY_PEOPLE, "shrug");
		category(CATEGORY_PEOPLE, "man_shrugging");
		category(CATEGORY_PEOPLE, "woman_shrugging");
		category(CATEGORY_PEOPLE, "massage");
		category(CATEGORY_PEOPLE, "man_getting_massage");
		category(CATEGORY_PEOPLE, "woman_getting_massage");
		category(CATEGORY_PEOPLE, "haircut");
		category(CATEGORY_PEOPLE, "man_getting_haircut");
		category(CATEGORY_PEOPLE, "woman_getting_haircut");
		category(CATEGORY_PEOPLE, "walking");
		category(CATEGORY_PEOPLE, "man_walking");
		category(CATEGORY_PEOPLE, "woman_walking");
		category(CATEGORY_PEOPLE, "runner");
		category(CATEGORY_PEOPLE, "man_running");
		category(CATEGORY_PEOPLE, "woman_running");
		category(CATEGORY_PEOPLE, "dancer");
		category(CATEGORY_PEOPLE, "man_dancing");
		category(CATEGORY_PEOPLE, "dancers");
		category(CATEGORY_PEOPLE, "man_with_bunny_ears_partying");
		category(CATEGORY_PEOPLE, "woman_with_bunny_ears_partying");
		category(CATEGORY_PEOPLE, "person_in_steamy_room");
		category(CATEGORY_PEOPLE, "woman_in_steamy_room");
		category(CATEGORY_PEOPLE, "man_in_steamy_room");
		category(CATEGORY_PEOPLE, "person_climbing");
		category(CATEGORY_PEOPLE, "woman_climbing");
		category(CATEGORY_PEOPLE, "man_climbing");
		category(CATEGORY_PEOPLE, "person_in_lotus_position");
		category(CATEGORY_PEOPLE, "woman_in_lotus_position");
		category(CATEGORY_PEOPLE, "man_in_lotus_position");
		category(CATEGORY_PEOPLE, "bath");
		category(CATEGORY_PEOPLE, "sleeping_accommodation");
		category(CATEGORY_PEOPLE, "man_in_business_suit_levitating");
		category(CATEGORY_PEOPLE, "speaking_head_in_silhouette");
		category(CATEGORY_PEOPLE, "bust_in_silhouette");
		category(CATEGORY_PEOPLE, "busts_in_silhouette");
		category(CATEGORY_PEOPLE, "fencer");
		category(CATEGORY_PEOPLE, "horse_racing");
		category(CATEGORY_PEOPLE, "skier");
		category(CATEGORY_PEOPLE, "snowboarder");
	}
	private static void init108(){
		category(CATEGORY_PEOPLE, "golfer");
		category(CATEGORY_PEOPLE, "man_golfing");
		category(CATEGORY_PEOPLE, "woman_golfing");
		category(CATEGORY_PEOPLE, "surfer");
		category(CATEGORY_PEOPLE, "man_surfing");
		category(CATEGORY_PEOPLE, "woman_surfing");
		category(CATEGORY_PEOPLE, "rowboat");
		category(CATEGORY_PEOPLE, "man_rowing_boat");
		category(CATEGORY_PEOPLE, "woman_rowing_boat");
		category(CATEGORY_PEOPLE, "swimmer");
		category(CATEGORY_PEOPLE, "man_swimming");
		category(CATEGORY_PEOPLE, "woman_swimming");
		category(CATEGORY_PEOPLE, "person_with_ball");
		category(CATEGORY_PEOPLE, "man_bouncing_ball");
		category(CATEGORY_PEOPLE, "woman_bouncing_ball");
		category(CATEGORY_PEOPLE, "weight_lifter");
		category(CATEGORY_PEOPLE, "man_lifting_weights");
		category(CATEGORY_PEOPLE, "woman_lifting_weights");
		category(CATEGORY_PEOPLE, "bicyclist");
		category(CATEGORY_PEOPLE, "man_biking");
		category(CATEGORY_PEOPLE, "woman_biking");
		category(CATEGORY_PEOPLE, "mountain_bicyclist");
		category(CATEGORY_PEOPLE, "man_mountain_biking");
		category(CATEGORY_PEOPLE, "woman_mountain_biking");
		category(CATEGORY_PEOPLE, "racing_car");
		category(CATEGORY_PEOPLE, "racing_motorcycle");
		category(CATEGORY_PEOPLE, "person_doing_cartwheel");
		category(CATEGORY_PEOPLE, "man_cartwheeling");
		category(CATEGORY_PEOPLE, "woman_cartwheeling");
		category(CATEGORY_PEOPLE, "wrestlers");
		category(CATEGORY_PEOPLE, "man_wrestling");
		category(CATEGORY_PEOPLE, "woman_wrestling");
		category(CATEGORY_PEOPLE, "water_polo");
		category(CATEGORY_PEOPLE, "man_playing_water_polo");
		category(CATEGORY_PEOPLE, "woman_playing_water_polo");
		category(CATEGORY_PEOPLE, "handball");
		category(CATEGORY_PEOPLE, "man_playing_handball");
		category(CATEGORY_PEOPLE, "woman_playing_handball");
		category(CATEGORY_PEOPLE, "juggling");
		category(CATEGORY_PEOPLE, "man_juggling");
		category(CATEGORY_PEOPLE, "woman_juggling");
		category(CATEGORY_PEOPLE, "couple");
		category(CATEGORY_PEOPLE, "two_men_holding_hands");
		category(CATEGORY_PEOPLE, "two_women_holding_hands");
		category(CATEGORY_PEOPLE, "couplekiss");
		category(CATEGORY_PEOPLE, "woman_kiss_man");
		category(CATEGORY_PEOPLE, "man_kiss_man");
		category(CATEGORY_PEOPLE, "woman_kiss_woman");
		category(CATEGORY_PEOPLE, "couple_with_heart");
		category(CATEGORY_PEOPLE, "woman_heart_man");
		category(CATEGORY_PEOPLE, "man_heart_man");
		category(CATEGORY_PEOPLE, "woman_heart_woman");
		category(CATEGORY_PEOPLE, "family");
		category(CATEGORY_PEOPLE, "man_woman_boy");
		category(CATEGORY_PEOPLE, "man_woman_girl");
		category(CATEGORY_PEOPLE, "man_woman_girl_boy");
		category(CATEGORY_PEOPLE, "man_woman_boy_boy");
		category(CATEGORY_PEOPLE, "man_woman_girl_girl");
		category(CATEGORY_PEOPLE, "man_man_boy");
		category(CATEGORY_PEOPLE, "man_man_girl");
		category(CATEGORY_PEOPLE, "man_man_girl_boy");
		category(CATEGORY_PEOPLE, "man_man_boy_boy");
		category(CATEGORY_PEOPLE, "man_man_girl_girl");
		category(CATEGORY_PEOPLE, "woman_woman_boy");
		category(CATEGORY_PEOPLE, "woman_woman_girl");
		category(CATEGORY_PEOPLE, "woman_woman_girl_boy");
		category(CATEGORY_PEOPLE, "woman_woman_boy_boy");
		category(CATEGORY_PEOPLE, "woman_woman_girl_girl");
		category(CATEGORY_PEOPLE, "man_boy");
		category(CATEGORY_PEOPLE, "man_boy_boy");
		category(CATEGORY_PEOPLE, "man_girl");
		category(CATEGORY_PEOPLE, "man_girl_boy");
		category(CATEGORY_PEOPLE, "man_girl_girl");
		category(CATEGORY_PEOPLE, "woman_boy");
		category(CATEGORY_PEOPLE, "woman_boy_boy");
		category(CATEGORY_PEOPLE, "woman_girl");
		category(CATEGORY_PEOPLE, "woman_girl_boy");
		category(CATEGORY_PEOPLE, "woman_girl_girl");
		category(CATEGORY_PEOPLE, "selfie");
		category(CATEGORY_PEOPLE, "muscle");
		category(CATEGORY_PEOPLE, "point_left");
		category(CATEGORY_PEOPLE, "point_right");
		category(CATEGORY_PEOPLE, "point_up");
		category(CATEGORY_PEOPLE, "point_up_2");
		category(CATEGORY_PEOPLE, "middle_finger");
		category(CATEGORY_PEOPLE, "point_down");
		category(CATEGORY_PEOPLE, "v");
		category(CATEGORY_PEOPLE, "crossed_fingers");
		category(CATEGORY_PEOPLE, "spock_hand");
		category(CATEGORY_PEOPLE, "the_horns");
		category(CATEGORY_PEOPLE, "call_me_hand");
		category(CATEGORY_PEOPLE, "raised_hand_with_fingers_splayed");
		category(CATEGORY_PEOPLE, "hand");
		category(CATEGORY_PEOPLE, "ok_hand");
		category(CATEGORY_PEOPLE, "+1");
		category(CATEGORY_PEOPLE, "_1");
		category(CATEGORY_PEOPLE, "fist");
		category(CATEGORY_PEOPLE, "facepunch");
		category(CATEGORY_PEOPLE, "left_facing_fist");
		category(CATEGORY_PEOPLE, "right_facing_fist");
		category(CATEGORY_PEOPLE, "raised_back_of_hand");
	}
	private static void init109(){
		category(CATEGORY_PEOPLE, "wave");
		category(CATEGORY_PEOPLE, "i_love_you_hand_sign");
		category(CATEGORY_PEOPLE, "writing_hand");
		category(CATEGORY_PEOPLE, "clap");
		category(CATEGORY_PEOPLE, "open_hands");
		category(CATEGORY_PEOPLE, "raised_hands");
		category(CATEGORY_PEOPLE, "palms_up_together");
		category(CATEGORY_PEOPLE, "pray");
		category(CATEGORY_PEOPLE, "handshake");
		category(CATEGORY_PEOPLE, "nail_care");
		category(CATEGORY_PEOPLE, "ear");
		category(CATEGORY_PEOPLE, "nose");
		category(CATEGORY_PEOPLE, "footprints");
		category(CATEGORY_PEOPLE, "eyes");
		category(CATEGORY_PEOPLE, "eye");
		category(CATEGORY_PEOPLE, "eye_in_speech_bubble");
		category(CATEGORY_PEOPLE, "brain");
		category(CATEGORY_PEOPLE, "tongue");
		category(CATEGORY_PEOPLE, "lips");
		category(CATEGORY_PEOPLE, "kiss");
		category(CATEGORY_PEOPLE, "cupid");
		category(CATEGORY_PEOPLE, "heart");
		category(CATEGORY_PEOPLE, "heartbeat");
		category(CATEGORY_PEOPLE, "broken_heart");
		category(CATEGORY_PEOPLE, "two_hearts");
		category(CATEGORY_PEOPLE, "sparkling_heart");
		category(CATEGORY_PEOPLE, "heartpulse");
		category(CATEGORY_PEOPLE, "blue_heart");
		category(CATEGORY_PEOPLE, "green_heart");
		category(CATEGORY_PEOPLE, "yellow_heart");
		category(CATEGORY_PEOPLE, "orange_heart");
		category(CATEGORY_PEOPLE, "purple_heart");
		category(CATEGORY_PEOPLE, "black_heart");
		category(CATEGORY_PEOPLE, "gift_heart");
		category(CATEGORY_PEOPLE, "revolving_hearts");
		category(CATEGORY_PEOPLE, "heart_decoration");
		category(CATEGORY_PEOPLE, "heavy_heart_exclamation_mark_ornament");
		category(CATEGORY_PEOPLE, "love_letter");
		category(CATEGORY_PEOPLE, "zzz");
		category(CATEGORY_PEOPLE, "anger");
		category(CATEGORY_PEOPLE, "bomb");
		category(CATEGORY_PEOPLE, "boom");
		category(CATEGORY_PEOPLE, "sweat_drops");
		category(CATEGORY_PEOPLE, "dash");
		category(CATEGORY_PEOPLE, "dizzy");
		category(CATEGORY_PEOPLE, "speech_balloon");
		category(CATEGORY_PEOPLE, "left_speech_bubble");
		category(CATEGORY_PEOPLE, "right_anger_bubble");
		category(CATEGORY_PEOPLE, "thought_balloon");
		category(CATEGORY_PEOPLE, "hole");
		category(CATEGORY_PEOPLE, "eyeglasses");
		category(CATEGORY_PEOPLE, "dark_sunglasses");
		category(CATEGORY_PEOPLE, "necktie");
		category(CATEGORY_PEOPLE, "shirt");
		category(CATEGORY_PEOPLE, "jeans");
		category(CATEGORY_PEOPLE, "scarf");
		category(CATEGORY_PEOPLE, "gloves");
		category(CATEGORY_PEOPLE, "coat");
		category(CATEGORY_PEOPLE, "socks");
		category(CATEGORY_PEOPLE, "dress");
		category(CATEGORY_PEOPLE, "kimono");
		category(CATEGORY_PEOPLE, "bikini");
		category(CATEGORY_PEOPLE, "womans_clothes");
		category(CATEGORY_PEOPLE, "purse");
		category(CATEGORY_PEOPLE, "handbag");
		category(CATEGORY_PEOPLE, "pouch");
		category(CATEGORY_PEOPLE, "shopping_bags");
		category(CATEGORY_PEOPLE, "school_satchel");
		category(CATEGORY_PEOPLE, "mans_shoe");
		category(CATEGORY_PEOPLE, "athletic_shoe");
		category(CATEGORY_PEOPLE, "high_heel");
		category(CATEGORY_PEOPLE, "sandal");
		category(CATEGORY_PEOPLE, "boot");
		category(CATEGORY_PEOPLE, "crown");
		category(CATEGORY_PEOPLE, "womans_hat");
		category(CATEGORY_PEOPLE, "tophat");
		category(CATEGORY_PEOPLE, "mortar_board");
		category(CATEGORY_PEOPLE, "billed_cap");
		category(CATEGORY_PEOPLE, "helmet_with_white_cross");
		category(CATEGORY_PEOPLE, "prayer_beads");
		category(CATEGORY_PEOPLE, "lipstick");
		category(CATEGORY_PEOPLE, "ring");
		category(CATEGORY_PEOPLE, "gem");
		category(CATEGORY_NATURE, "monkey_face");
		category(CATEGORY_NATURE, "monkey");
		category(CATEGORY_NATURE, "gorilla");
		category(CATEGORY_NATURE, "dog");
		category(CATEGORY_NATURE, "dog2");
		category(CATEGORY_NATURE, "poodle");
		category(CATEGORY_NATURE, "wolf");
		category(CATEGORY_NATURE, "fox_face");
		category(CATEGORY_NATURE, "cat");
		category(CATEGORY_NATURE, "cat2");
		category(CATEGORY_NATURE, "lion_face");
		category(CATEGORY_NATURE, "tiger");
		category(CATEGORY_NATURE, "tiger2");
		category(CATEGORY_NATURE, "leopard");
		category(CATEGORY_NATURE, "horse");
		category(CATEGORY_NATURE, "racehorse");
		category(CATEGORY_NATURE, "unicorn_face");
		category(CATEGORY_NATURE, "zebra_face");
	}
	private static void init110(){
		category(CATEGORY_NATURE, "deer");
		category(CATEGORY_NATURE, "cow");
		category(CATEGORY_NATURE, "ox");
		category(CATEGORY_NATURE, "water_buffalo");
		category(CATEGORY_NATURE, "cow2");
		category(CATEGORY_NATURE, "pig");
		category(CATEGORY_NATURE, "pig2");
		category(CATEGORY_NATURE, "boar");
		category(CATEGORY_NATURE, "pig_nose");
		category(CATEGORY_NATURE, "ram");
		category(CATEGORY_NATURE, "sheep");
		category(CATEGORY_NATURE, "goat");
		category(CATEGORY_NATURE, "dromedary_camel");
		category(CATEGORY_NATURE, "camel");
		category(CATEGORY_NATURE, "giraffe_face");
		category(CATEGORY_NATURE, "elephant");
		category(CATEGORY_NATURE, "rhinoceros");
		category(CATEGORY_NATURE, "mouse");
		category(CATEGORY_NATURE, "mouse2");
		category(CATEGORY_NATURE, "rat");
		category(CATEGORY_NATURE, "hamster");
		category(CATEGORY_NATURE, "rabbit");
		category(CATEGORY_NATURE, "rabbit2");
		category(CATEGORY_NATURE, "chipmunk");
		category(CATEGORY_NATURE, "hedgehog");
		category(CATEGORY_NATURE, "bat");
		category(CATEGORY_NATURE, "bear");
		category(CATEGORY_NATURE, "koala");
		category(CATEGORY_NATURE, "panda_face");
		category(CATEGORY_NATURE, "feet");
		category(CATEGORY_NATURE, "turkey");
		category(CATEGORY_NATURE, "chicken");
		category(CATEGORY_NATURE, "rooster");
		category(CATEGORY_NATURE, "hatching_chick");
		category(CATEGORY_NATURE, "baby_chick");
		category(CATEGORY_NATURE, "hatched_chick");
		category(CATEGORY_NATURE, "bird");
		category(CATEGORY_NATURE, "penguin");
		category(CATEGORY_NATURE, "dove_of_peace");
		category(CATEGORY_NATURE, "eagle");
		category(CATEGORY_NATURE, "duck");
		category(CATEGORY_NATURE, "owl");
		category(CATEGORY_NATURE, "frog");
		category(CATEGORY_NATURE, "crocodile");
		category(CATEGORY_NATURE, "turtle");
		category(CATEGORY_NATURE, "lizard");
		category(CATEGORY_NATURE, "snake");
		category(CATEGORY_NATURE, "dragon_face");
		category(CATEGORY_NATURE, "dragon");
		category(CATEGORY_NATURE, "sauropod");
		category(CATEGORY_NATURE, "t_rex");
		category(CATEGORY_NATURE, "whale");
		category(CATEGORY_NATURE, "whale2");
		category(CATEGORY_NATURE, "dolphin");
		category(CATEGORY_NATURE, "fish");
		category(CATEGORY_NATURE, "tropical_fish");
		category(CATEGORY_NATURE, "blowfish");
		category(CATEGORY_NATURE, "shark");
		category(CATEGORY_NATURE, "octopus");
		category(CATEGORY_NATURE, "shell");
		category(CATEGORY_NATURE, "crab");
		category(CATEGORY_NATURE, "shrimp");
		category(CATEGORY_NATURE, "squid");
		category(CATEGORY_NATURE, "snail");
		category(CATEGORY_NATURE, "butterfly");
		category(CATEGORY_NATURE, "bug");
		category(CATEGORY_NATURE, "ant");
		category(CATEGORY_NATURE, "bee");
		category(CATEGORY_NATURE, "beetle");
		category(CATEGORY_NATURE, "cricket");
		category(CATEGORY_NATURE, "spider");
		category(CATEGORY_NATURE, "spider_web");
		category(CATEGORY_NATURE, "scorpion");
		category(CATEGORY_NATURE, "bouquet");
		category(CATEGORY_NATURE, "cherry_blossom");
		category(CATEGORY_NATURE, "white_flower");
		category(CATEGORY_NATURE, "rosette");
		category(CATEGORY_NATURE, "rose");
		category(CATEGORY_NATURE, "wilted_flower");
		category(CATEGORY_NATURE, "hibiscus");
		category(CATEGORY_NATURE, "sunflower");
		category(CATEGORY_NATURE, "blossom");
		category(CATEGORY_NATURE, "tulip");
		category(CATEGORY_NATURE, "seedling");
		category(CATEGORY_NATURE, "evergreen_tree");
		category(CATEGORY_NATURE, "deciduous_tree");
		category(CATEGORY_NATURE, "palm_tree");
		category(CATEGORY_NATURE, "cactus");
		category(CATEGORY_NATURE, "ear_of_rice");
		category(CATEGORY_NATURE, "herb");
		category(CATEGORY_NATURE, "shamrock");
		category(CATEGORY_NATURE, "four_leaf_clover");
		category(CATEGORY_NATURE, "maple_leaf");
		category(CATEGORY_NATURE, "fallen_leaf");
		category(CATEGORY_NATURE, "leaves");
		category(CATEGORY_FOODS, "grapes");
		category(CATEGORY_FOODS, "melon");
		category(CATEGORY_FOODS, "watermelon");
		category(CATEGORY_FOODS, "tangerine");
		category(CATEGORY_FOODS, "lemon");
		category(CATEGORY_FOODS, "banana");
	}
	private static void init111(){
		category(CATEGORY_FOODS, "pineapple");
		category(CATEGORY_FOODS, "apple");
		category(CATEGORY_FOODS, "green_apple");
		category(CATEGORY_FOODS, "pear");
		category(CATEGORY_FOODS, "peach");
		category(CATEGORY_FOODS, "cherries");
		category(CATEGORY_FOODS, "strawberry");
		category(CATEGORY_FOODS, "kiwifruit");
		category(CATEGORY_FOODS, "tomato");
		category(CATEGORY_FOODS, "coconut");
		category(CATEGORY_FOODS, "avocado");
		category(CATEGORY_FOODS, "eggplant");
		category(CATEGORY_FOODS, "potato");
		category(CATEGORY_FOODS, "carrot");
		category(CATEGORY_FOODS, "corn");
		category(CATEGORY_FOODS, "hot_pepper");
		category(CATEGORY_FOODS, "cucumber");
		category(CATEGORY_FOODS, "broccoli");
		category(CATEGORY_FOODS, "mushroom");
		category(CATEGORY_FOODS, "peanuts");
		category(CATEGORY_FOODS, "chestnut");
		category(CATEGORY_FOODS, "bread");
		category(CATEGORY_FOODS, "croissant");
		category(CATEGORY_FOODS, "baguette_bread");
		category(CATEGORY_FOODS, "pretzel");
		category(CATEGORY_FOODS, "pancakes");
		category(CATEGORY_FOODS, "cheese_wedge");
		category(CATEGORY_FOODS, "meat_on_bone");
		category(CATEGORY_FOODS, "poultry_leg");
		category(CATEGORY_FOODS, "cut_of_meat");
		category(CATEGORY_FOODS, "bacon");
		category(CATEGORY_FOODS, "hamburger");
		category(CATEGORY_FOODS, "fries");
		category(CATEGORY_FOODS, "pizza");
		category(CATEGORY_FOODS, "hotdog");
		category(CATEGORY_FOODS, "sandwich");
		category(CATEGORY_FOODS, "taco");
		category(CATEGORY_FOODS, "burrito");
		category(CATEGORY_FOODS, "stuffed_flatbread");
		category(CATEGORY_FOODS, "egg");
		category(CATEGORY_FOODS, "fried_egg");
		category(CATEGORY_FOODS, "shallow_pan_of_food");
		category(CATEGORY_FOODS, "stew");
		category(CATEGORY_FOODS, "bowl_with_spoon");
		category(CATEGORY_FOODS, "green_salad");
		category(CATEGORY_FOODS, "popcorn");
		category(CATEGORY_FOODS, "canned_food");
		category(CATEGORY_FOODS, "bento");
		category(CATEGORY_FOODS, "rice_cracker");
		category(CATEGORY_FOODS, "rice_ball");
		category(CATEGORY_FOODS, "rice");
		category(CATEGORY_FOODS, "curry");
		category(CATEGORY_FOODS, "ramen");
		category(CATEGORY_FOODS, "spaghetti");
		category(CATEGORY_FOODS, "sweet_potato");
		category(CATEGORY_FOODS, "oden");
		category(CATEGORY_FOODS, "sushi");
		category(CATEGORY_FOODS, "fried_shrimp");
		category(CATEGORY_FOODS, "fish_cake");
		category(CATEGORY_FOODS, "dango");
		category(CATEGORY_FOODS, "dumpling");
		category(CATEGORY_FOODS, "fortune_cookie");
		category(CATEGORY_FOODS, "takeout_box");
		category(CATEGORY_FOODS, "icecream");
		category(CATEGORY_FOODS, "shaved_ice");
		category(CATEGORY_FOODS, "ice_cream");
		category(CATEGORY_FOODS, "doughnut");
		category(CATEGORY_FOODS, "cookie");
		category(CATEGORY_FOODS, "birthday");
		category(CATEGORY_FOODS, "cake");
		category(CATEGORY_FOODS, "pie");
		category(CATEGORY_FOODS, "chocolate_bar");
		category(CATEGORY_FOODS, "candy");
		category(CATEGORY_FOODS, "lollipop");
		category(CATEGORY_FOODS, "custard");
		category(CATEGORY_FOODS, "honey_pot");
		category(CATEGORY_FOODS, "baby_bottle");
		category(CATEGORY_FOODS, "glass_of_milk");
		category(CATEGORY_FOODS, "coffee");
		category(CATEGORY_FOODS, "tea");
		category(CATEGORY_FOODS, "sake");
		category(CATEGORY_FOODS, "champagne");
		category(CATEGORY_FOODS, "wine_glass");
		category(CATEGORY_FOODS, "cocktail");
		category(CATEGORY_FOODS, "tropical_drink");
		category(CATEGORY_FOODS, "beer");
		category(CATEGORY_FOODS, "beers");
		category(CATEGORY_FOODS, "clinking_glasses");
		category(CATEGORY_FOODS, "tumbler_glass");
		category(CATEGORY_FOODS, "cup_with_straw");
		category(CATEGORY_FOODS, "chopsticks");
		category(CATEGORY_FOODS, "knife_fork_plate");
		category(CATEGORY_FOODS, "fork_and_knife");
		category(CATEGORY_FOODS, "spoon");
		category(CATEGORY_FOODS, "hocho");
		category(CATEGORY_FOODS, "amphora");
		category(CATEGORY_ACTIVITY, "jack_o_lantern");
		category(CATEGORY_ACTIVITY, "christmas_tree");
		category(CATEGORY_ACTIVITY, "fireworks");
		category(CATEGORY_ACTIVITY, "sparkler");
		category(CATEGORY_ACTIVITY, "sparkles");
	}
	private static void init112(){
		category(CATEGORY_ACTIVITY, "balloon");
		category(CATEGORY_ACTIVITY, "tada");
		category(CATEGORY_ACTIVITY, "confetti_ball");
		category(CATEGORY_ACTIVITY, "tanabata_tree");
		category(CATEGORY_ACTIVITY, "bamboo");
		category(CATEGORY_ACTIVITY, "dolls");
		category(CATEGORY_ACTIVITY, "flags");
		category(CATEGORY_ACTIVITY, "wind_chime");
		category(CATEGORY_ACTIVITY, "rice_scene");
		category(CATEGORY_ACTIVITY, "ribbon");
		category(CATEGORY_ACTIVITY, "gift");
		category(CATEGORY_ACTIVITY, "reminder_ribbon");
		category(CATEGORY_ACTIVITY, "admission_tickets");
		category(CATEGORY_ACTIVITY, "ticket");
		category(CATEGORY_ACTIVITY, "medal");
		category(CATEGORY_ACTIVITY, "trophy");
		category(CATEGORY_ACTIVITY, "sports_medal");
		category(CATEGORY_ACTIVITY, "first_place_medal");
		category(CATEGORY_ACTIVITY, "second_place_medal");
		category(CATEGORY_ACTIVITY, "third_place_medal");
		category(CATEGORY_ACTIVITY, "soccer");
		category(CATEGORY_ACTIVITY, "baseball");
		category(CATEGORY_ACTIVITY, "basketball");
		category(CATEGORY_ACTIVITY, "volleyball");
		category(CATEGORY_ACTIVITY, "football");
		category(CATEGORY_ACTIVITY, "rugby_football");
		category(CATEGORY_ACTIVITY, "tennis");
		category(CATEGORY_ACTIVITY, "8ball");
		category(CATEGORY_ACTIVITY, "bowling");
		category(CATEGORY_ACTIVITY, "cricket_bat_and_ball");
		category(CATEGORY_ACTIVITY, "field_hockey_stick_and_ball");
		category(CATEGORY_ACTIVITY, "ice_hockey_stick_and_puck");
		category(CATEGORY_ACTIVITY, "table_tennis_paddle_and_ball");
		category(CATEGORY_ACTIVITY, "badminton_racquet_and_shuttlecock");
		category(CATEGORY_ACTIVITY, "boxing_glove");
		category(CATEGORY_ACTIVITY, "martial_arts_uniform");
		category(CATEGORY_ACTIVITY, "goal_net");
		category(CATEGORY_ACTIVITY, "dart");
		category(CATEGORY_ACTIVITY, "golf");
		category(CATEGORY_ACTIVITY, "ice_skate");
		category(CATEGORY_ACTIVITY, "fishing_pole_and_fish");
		category(CATEGORY_ACTIVITY, "running_shirt_with_sash");
		category(CATEGORY_ACTIVITY, "ski");
		category(CATEGORY_ACTIVITY, "sled");
		category(CATEGORY_ACTIVITY, "curling_stone");
		category(CATEGORY_ACTIVITY, "video_game");
		category(CATEGORY_ACTIVITY, "joystick");
		category(CATEGORY_ACTIVITY, "game_die");
		category(CATEGORY_ACTIVITY, "spades");
		category(CATEGORY_ACTIVITY, "hearts");
		category(CATEGORY_ACTIVITY, "diamonds");
		category(CATEGORY_ACTIVITY, "clubs");
		category(CATEGORY_ACTIVITY, "black_joker");
		category(CATEGORY_ACTIVITY, "mahjong");
		category(CATEGORY_ACTIVITY, "flower_playing_cards");
		category(CATEGORY_PLACES, "earth_africa");
		category(CATEGORY_PLACES, "earth_americas");
		category(CATEGORY_PLACES, "earth_asia");
		category(CATEGORY_PLACES, "globe_with_meridians");
		category(CATEGORY_PLACES, "world_map");
		category(CATEGORY_PLACES, "japan");
		category(CATEGORY_PLACES, "snow_capped_mountain");
		category(CATEGORY_PLACES, "mountain");
		category(CATEGORY_PLACES, "volcano");
		category(CATEGORY_PLACES, "mount_fuji");
		category(CATEGORY_PLACES, "camping");
		category(CATEGORY_PLACES, "beach_with_umbrella");
		category(CATEGORY_PLACES, "desert");
		category(CATEGORY_PLACES, "desert_island");
		category(CATEGORY_PLACES, "national_park");
		category(CATEGORY_PLACES, "stadium");
		category(CATEGORY_PLACES, "classical_building");
		category(CATEGORY_PLACES, "building_construction");
		category(CATEGORY_PLACES, "house_buildings");
		category(CATEGORY_PLACES, "cityscape");
		category(CATEGORY_PLACES, "derelict_house_building");
		category(CATEGORY_PLACES, "house");
		category(CATEGORY_PLACES, "house_with_garden");
		category(CATEGORY_PLACES, "office");
		category(CATEGORY_PLACES, "post_office");
		category(CATEGORY_PLACES, "european_post_office");
		category(CATEGORY_PLACES, "hospital");
		category(CATEGORY_PLACES, "bank");
		category(CATEGORY_PLACES, "hotel");
		category(CATEGORY_PLACES, "love_hotel");
		category(CATEGORY_PLACES, "convenience_store");
		category(CATEGORY_PLACES, "school");
		category(CATEGORY_PLACES, "department_store");
		category(CATEGORY_PLACES, "factory");
		category(CATEGORY_PLACES, "japanese_castle");
		category(CATEGORY_PLACES, "european_castle");
		category(CATEGORY_PLACES, "wedding");
		category(CATEGORY_PLACES, "tokyo_tower");
		category(CATEGORY_PLACES, "statue_of_liberty");
		category(CATEGORY_PLACES, "church");
		category(CATEGORY_PLACES, "mosque");
		category(CATEGORY_PLACES, "synagogue");
		category(CATEGORY_PLACES, "shinto_shrine");
		category(CATEGORY_PLACES, "kaaba");
		category(CATEGORY_PLACES, "fountain");
		category(CATEGORY_PLACES, "tent");
	}
	private static void init113(){
		category(CATEGORY_PLACES, "foggy");
		category(CATEGORY_PLACES, "night_with_stars");
		category(CATEGORY_PLACES, "sunrise_over_mountains");
		category(CATEGORY_PLACES, "sunrise");
		category(CATEGORY_PLACES, "city_sunset");
		category(CATEGORY_PLACES, "city_sunrise");
		category(CATEGORY_PLACES, "bridge_at_night");
		category(CATEGORY_PLACES, "hotsprings");
		category(CATEGORY_PLACES, "milky_way");
		category(CATEGORY_PLACES, "carousel_horse");
		category(CATEGORY_PLACES, "ferris_wheel");
		category(CATEGORY_PLACES, "roller_coaster");
		category(CATEGORY_PLACES, "barber");
		category(CATEGORY_PLACES, "circus_tent");
		category(CATEGORY_PLACES, "performing_arts");
		category(CATEGORY_PLACES, "frame_with_picture");
		category(CATEGORY_PLACES, "art");
		category(CATEGORY_PLACES, "slot_machine");
		category(CATEGORY_PLACES, "steam_locomotive");
		category(CATEGORY_PLACES, "railway_car");
		category(CATEGORY_PLACES, "bullettrain_side");
		category(CATEGORY_PLACES, "bullettrain_front");
		category(CATEGORY_PLACES, "train2");
		category(CATEGORY_PLACES, "metro");
		category(CATEGORY_PLACES, "light_rail");
		category(CATEGORY_PLACES, "station");
		category(CATEGORY_PLACES, "tram");
		category(CATEGORY_PLACES, "monorail");
		category(CATEGORY_PLACES, "mountain_railway");
		category(CATEGORY_PLACES, "train");
		category(CATEGORY_PLACES, "bus");
		category(CATEGORY_PLACES, "oncoming_bus");
		category(CATEGORY_PLACES, "trolleybus");
		category(CATEGORY_PLACES, "minibus");
		category(CATEGORY_PLACES, "ambulance");
		category(CATEGORY_PLACES, "fire_engine");
		category(CATEGORY_PLACES, "police_car");
		category(CATEGORY_PLACES, "oncoming_police_car");
		category(CATEGORY_PLACES, "taxi");
		category(CATEGORY_PLACES, "oncoming_taxi");
		category(CATEGORY_PLACES, "car");
		category(CATEGORY_PLACES, "oncoming_automobile");
		category(CATEGORY_PLACES, "blue_car");
		category(CATEGORY_PLACES, "truck");
		category(CATEGORY_PLACES, "articulated_lorry");
		category(CATEGORY_PLACES, "tractor");
		category(CATEGORY_PLACES, "bike");
		category(CATEGORY_PLACES, "scooter");
		category(CATEGORY_PLACES, "motor_scooter");
		category(CATEGORY_PLACES, "busstop");
		category(CATEGORY_PLACES, "motorway");
		category(CATEGORY_PLACES, "railway_track");
		category(CATEGORY_PLACES, "fuelpump");
		category(CATEGORY_PLACES, "rotating_light");
		category(CATEGORY_PLACES, "traffic_light");
		category(CATEGORY_PLACES, "vertical_traffic_light");
		category(CATEGORY_PLACES, "construction");
		category(CATEGORY_PLACES, "octagonal_sign");
		category(CATEGORY_PLACES, "anchor");
		category(CATEGORY_PLACES, "boat");
		category(CATEGORY_PLACES, "canoe");
		category(CATEGORY_PLACES, "speedboat");
		category(CATEGORY_PLACES, "passenger_ship");
		category(CATEGORY_PLACES, "ferry");
		category(CATEGORY_PLACES, "motor_boat");
		category(CATEGORY_PLACES, "ship");
		category(CATEGORY_PLACES, "airplane");
		category(CATEGORY_PLACES, "small_airplane");
		category(CATEGORY_PLACES, "airplane_departure");
		category(CATEGORY_PLACES, "airplane_arriving");
		category(CATEGORY_PLACES, "seat");
		category(CATEGORY_PLACES, "helicopter");
		category(CATEGORY_PLACES, "suspension_railway");
		category(CATEGORY_PLACES, "mountain_cableway");
		category(CATEGORY_PLACES, "aerial_tramway");
		category(CATEGORY_PLACES, "satellite");
		category(CATEGORY_PLACES, "rocket");
		category(CATEGORY_PLACES, "flying_saucer");
		category(CATEGORY_PLACES, "bellhop_bell");
		category(CATEGORY_PLACES, "door");
		category(CATEGORY_PLACES, "bed");
		category(CATEGORY_PLACES, "couch_and_lamp");
		category(CATEGORY_PLACES, "toilet");
		category(CATEGORY_PLACES, "shower");
		category(CATEGORY_PLACES, "bathtub");
		category(CATEGORY_PLACES, "hourglass");
		category(CATEGORY_PLACES, "hourglass_flowing_sand");
		category(CATEGORY_PLACES, "watch");
		category(CATEGORY_PLACES, "alarm_clock");
		category(CATEGORY_PLACES, "stopwatch");
		category(CATEGORY_PLACES, "timer_clock");
		category(CATEGORY_PLACES, "mantelpiece_clock");
		category(CATEGORY_PLACES, "clock12");
		category(CATEGORY_PLACES, "clock1230");
		category(CATEGORY_PLACES, "clock1");
		category(CATEGORY_PLACES, "clock130");
		category(CATEGORY_PLACES, "clock2");
		category(CATEGORY_PLACES, "clock230");
		category(CATEGORY_PLACES, "clock3");
		category(CATEGORY_PLACES, "clock330");
		category(CATEGORY_PLACES, "clock4");
	}
	private static void init114(){
		category(CATEGORY_PLACES, "clock430");
		category(CATEGORY_PLACES, "clock5");
		category(CATEGORY_PLACES, "clock530");
		category(CATEGORY_PLACES, "clock6");
		category(CATEGORY_PLACES, "clock630");
		category(CATEGORY_PLACES, "clock7");
		category(CATEGORY_PLACES, "clock730");
		category(CATEGORY_PLACES, "clock8");
		category(CATEGORY_PLACES, "clock830");
		category(CATEGORY_PLACES, "clock9");
		category(CATEGORY_PLACES, "clock930");
		category(CATEGORY_PLACES, "clock10");
		category(CATEGORY_PLACES, "clock1030");
		category(CATEGORY_PLACES, "clock11");
		category(CATEGORY_PLACES, "clock1130");
		category(CATEGORY_PLACES, "new_moon");
		category(CATEGORY_PLACES, "waxing_crescent_moon");
		category(CATEGORY_PLACES, "first_quarter_moon");
		category(CATEGORY_PLACES, "moon");
		category(CATEGORY_PLACES, "full_moon");
		category(CATEGORY_PLACES, "waning_gibbous_moon");
		category(CATEGORY_PLACES, "last_quarter_moon");
		category(CATEGORY_PLACES, "waning_crescent_moon");
		category(CATEGORY_PLACES, "crescent_moon");
		category(CATEGORY_PLACES, "new_moon_with_face");
		category(CATEGORY_PLACES, "first_quarter_moon_with_face");
		category(CATEGORY_PLACES, "last_quarter_moon_with_face");
		category(CATEGORY_PLACES, "thermometer");
		category(CATEGORY_PLACES, "sunny");
		category(CATEGORY_PLACES, "full_moon_with_face");
		category(CATEGORY_PLACES, "sun_with_face");
		category(CATEGORY_PLACES, "star");
		category(CATEGORY_PLACES, "star2");
		category(CATEGORY_PLACES, "stars");
		category(CATEGORY_PLACES, "cloud");
		category(CATEGORY_PLACES, "partly_sunny");
		category(CATEGORY_PLACES, "thunder_cloud_and_rain");
		category(CATEGORY_PLACES, "mostly_sunny");
		category(CATEGORY_PLACES, "barely_sunny");
		category(CATEGORY_PLACES, "partly_sunny_rain");
		category(CATEGORY_PLACES, "rain_cloud");
		category(CATEGORY_PLACES, "snow_cloud");
		category(CATEGORY_PLACES, "lightning");
		category(CATEGORY_PLACES, "tornado");
		category(CATEGORY_PLACES, "fog");
		category(CATEGORY_PLACES, "wind_blowing_face");
		category(CATEGORY_PLACES, "cyclone");
		category(CATEGORY_PLACES, "rainbow");
		category(CATEGORY_PLACES, "closed_umbrella");
		category(CATEGORY_PLACES, "umbrella");
		category(CATEGORY_PLACES, "umbrella_with_rain_drops");
		category(CATEGORY_PLACES, "umbrella_on_ground");
		category(CATEGORY_PLACES, "zap");
		category(CATEGORY_PLACES, "snowflake");
		category(CATEGORY_PLACES, "snowman");
		category(CATEGORY_PLACES, "snowman_without_snow");
		category(CATEGORY_PLACES, "comet");
		category(CATEGORY_PLACES, "fire");
		category(CATEGORY_PLACES, "droplet");
		category(CATEGORY_PLACES, "ocean");
		category(CATEGORY_OBJECTS, "mute");
		category(CATEGORY_OBJECTS, "speaker");
		category(CATEGORY_OBJECTS, "sound");
		category(CATEGORY_OBJECTS, "loud_sound");
		category(CATEGORY_OBJECTS, "loudspeaker");
		category(CATEGORY_OBJECTS, "mega");
		category(CATEGORY_OBJECTS, "postal_horn");
		category(CATEGORY_OBJECTS, "bell");
		category(CATEGORY_OBJECTS, "no_bell");
		category(CATEGORY_OBJECTS, "musical_score");
		category(CATEGORY_OBJECTS, "musical_note");
		category(CATEGORY_OBJECTS, "notes");
		category(CATEGORY_OBJECTS, "studio_microphone");
		category(CATEGORY_OBJECTS, "level_slider");
		category(CATEGORY_OBJECTS, "control_knobs");
		category(CATEGORY_OBJECTS, "microphone");
		category(CATEGORY_OBJECTS, "headphones");
		category(CATEGORY_OBJECTS, "radio");
		category(CATEGORY_OBJECTS, "saxophone");
		category(CATEGORY_OBJECTS, "guitar");
		category(CATEGORY_OBJECTS, "musical_keyboard");
		category(CATEGORY_OBJECTS, "trumpet");
		category(CATEGORY_OBJECTS, "violin");
		category(CATEGORY_OBJECTS, "drum_with_drumsticks");
		category(CATEGORY_OBJECTS, "iphone");
		category(CATEGORY_OBJECTS, "calling");
		category(CATEGORY_OBJECTS, "phone");
		category(CATEGORY_OBJECTS, "telephone_receiver");
		category(CATEGORY_OBJECTS, "pager");
		category(CATEGORY_OBJECTS, "fax");
		category(CATEGORY_OBJECTS, "battery");
		category(CATEGORY_OBJECTS, "electric_plug");
		category(CATEGORY_OBJECTS, "computer");
		category(CATEGORY_OBJECTS, "desktop_computer");
		category(CATEGORY_OBJECTS, "printer");
		category(CATEGORY_OBJECTS, "keyboard");
		category(CATEGORY_OBJECTS, "three_button_mouse");
		category(CATEGORY_OBJECTS, "trackball");
		category(CATEGORY_OBJECTS, "minidisc");
		category(CATEGORY_OBJECTS, "floppy_disk");
		category(CATEGORY_OBJECTS, "cd");
	}
	private static void init115(){
		category(CATEGORY_OBJECTS, "dvd");
		category(CATEGORY_OBJECTS, "movie_camera");
		category(CATEGORY_OBJECTS, "film_frames");
		category(CATEGORY_OBJECTS, "film_projector");
		category(CATEGORY_OBJECTS, "clapper");
		category(CATEGORY_OBJECTS, "tv");
		category(CATEGORY_OBJECTS, "camera");
		category(CATEGORY_OBJECTS, "camera_with_flash");
		category(CATEGORY_OBJECTS, "video_camera");
		category(CATEGORY_OBJECTS, "vhs");
		category(CATEGORY_OBJECTS, "mag");
		category(CATEGORY_OBJECTS, "mag_right");
		category(CATEGORY_OBJECTS, "microscope");
		category(CATEGORY_OBJECTS, "telescope");
		category(CATEGORY_OBJECTS, "satellite_antenna");
		category(CATEGORY_OBJECTS, "candle");
		category(CATEGORY_OBJECTS, "bulb");
		category(CATEGORY_OBJECTS, "flashlight");
		category(CATEGORY_OBJECTS, "izakaya_lantern");
		category(CATEGORY_OBJECTS, "notebook_with_decorative_cover");
		category(CATEGORY_OBJECTS, "closed_book");
		category(CATEGORY_OBJECTS, "book");
		category(CATEGORY_OBJECTS, "green_book");
		category(CATEGORY_OBJECTS, "blue_book");
		category(CATEGORY_OBJECTS, "orange_book");
		category(CATEGORY_OBJECTS, "books");
		category(CATEGORY_OBJECTS, "notebook");
		category(CATEGORY_OBJECTS, "ledger");
		category(CATEGORY_OBJECTS, "page_with_curl");
		category(CATEGORY_OBJECTS, "scroll");
		category(CATEGORY_OBJECTS, "page_facing_up");
		category(CATEGORY_OBJECTS, "newspaper");
		category(CATEGORY_OBJECTS, "rolled_up_newspaper");
		category(CATEGORY_OBJECTS, "bookmark_tabs");
		category(CATEGORY_OBJECTS, "bookmark");
		category(CATEGORY_OBJECTS, "label");
		category(CATEGORY_OBJECTS, "moneybag");
		category(CATEGORY_OBJECTS, "yen");
		category(CATEGORY_OBJECTS, "dollar");
		category(CATEGORY_OBJECTS, "euro");
		category(CATEGORY_OBJECTS, "pound");
		category(CATEGORY_OBJECTS, "money_with_wings");
		category(CATEGORY_OBJECTS, "credit_card");
		category(CATEGORY_OBJECTS, "chart");
		category(CATEGORY_OBJECTS, "currency_exchange");
		category(CATEGORY_OBJECTS, "heavy_dollar_sign");
		category(CATEGORY_OBJECTS, "email");
		category(CATEGORY_OBJECTS, "e_mail");
		category(CATEGORY_OBJECTS, "incoming_envelope");
		category(CATEGORY_OBJECTS, "envelope_with_arrow");
		category(CATEGORY_OBJECTS, "outbox_tray");
		category(CATEGORY_OBJECTS, "inbox_tray");
		category(CATEGORY_OBJECTS, "package");
		category(CATEGORY_OBJECTS, "mailbox");
		category(CATEGORY_OBJECTS, "mailbox_closed");
		category(CATEGORY_OBJECTS, "mailbox_with_mail");
		category(CATEGORY_OBJECTS, "mailbox_with_no_mail");
		category(CATEGORY_OBJECTS, "postbox");
		category(CATEGORY_OBJECTS, "ballot_box_with_ballot");
		category(CATEGORY_OBJECTS, "pencil2");
		category(CATEGORY_OBJECTS, "black_nib");
		category(CATEGORY_OBJECTS, "lower_left_fountain_pen");
		category(CATEGORY_OBJECTS, "lower_left_ballpoint_pen");
		category(CATEGORY_OBJECTS, "lower_left_paintbrush");
		category(CATEGORY_OBJECTS, "lower_left_crayon");
		category(CATEGORY_OBJECTS, "memo");
		category(CATEGORY_OBJECTS, "briefcase");
		category(CATEGORY_OBJECTS, "file_folder");
		category(CATEGORY_OBJECTS, "open_file_folder");
		category(CATEGORY_OBJECTS, "card_index_dividers");
		category(CATEGORY_OBJECTS, "date");
		category(CATEGORY_OBJECTS, "calendar");
		category(CATEGORY_OBJECTS, "spiral_note_pad");
		category(CATEGORY_OBJECTS, "spiral_calendar_pad");
		category(CATEGORY_OBJECTS, "card_index");
		category(CATEGORY_OBJECTS, "chart_with_upwards_trend");
		category(CATEGORY_OBJECTS, "chart_with_downwards_trend");
		category(CATEGORY_OBJECTS, "bar_chart");
		category(CATEGORY_OBJECTS, "clipboard");
		category(CATEGORY_OBJECTS, "pushpin");
		category(CATEGORY_OBJECTS, "round_pushpin");
		category(CATEGORY_OBJECTS, "paperclip");
		category(CATEGORY_OBJECTS, "linked_paperclips");
		category(CATEGORY_OBJECTS, "straight_ruler");
		category(CATEGORY_OBJECTS, "triangular_ruler");
		category(CATEGORY_OBJECTS, "scissors");
		category(CATEGORY_OBJECTS, "card_file_box");
		category(CATEGORY_OBJECTS, "file_cabinet");
		category(CATEGORY_OBJECTS, "wastebasket");
		category(CATEGORY_OBJECTS, "lock");
		category(CATEGORY_OBJECTS, "unlock");
		category(CATEGORY_OBJECTS, "lock_with_ink_pen");
		category(CATEGORY_OBJECTS, "closed_lock_with_key");
		category(CATEGORY_OBJECTS, "key");
		category(CATEGORY_OBJECTS, "old_key");
		category(CATEGORY_OBJECTS, "hammer");
		category(CATEGORY_OBJECTS, "pick");
		category(CATEGORY_OBJECTS, "hammer_and_pick");
		category(CATEGORY_OBJECTS, "hammer_and_wrench");
		category(CATEGORY_OBJECTS, "dagger_knife");
		category(CATEGORY_OBJECTS, "crossed_swords");
	}
	private static void init116(){
		category(CATEGORY_OBJECTS, "gun");
		category(CATEGORY_OBJECTS, "bow_and_arrow");
		category(CATEGORY_OBJECTS, "shield");
		category(CATEGORY_OBJECTS, "wrench");
		category(CATEGORY_OBJECTS, "nut_and_bolt");
		category(CATEGORY_OBJECTS, "gear");
		category(CATEGORY_OBJECTS, "compression");
		category(CATEGORY_OBJECTS, "alembic");
		category(CATEGORY_OBJECTS, "scales");
		category(CATEGORY_OBJECTS, "link");
		category(CATEGORY_OBJECTS, "chains");
		category(CATEGORY_OBJECTS, "syringe");
		category(CATEGORY_OBJECTS, "pill");
		category(CATEGORY_OBJECTS, "smoking");
		category(CATEGORY_OBJECTS, "coffin");
		category(CATEGORY_OBJECTS, "funeral_urn");
		category(CATEGORY_OBJECTS, "moyai");
		category(CATEGORY_OBJECTS, "oil_drum");
		category(CATEGORY_OBJECTS, "crystal_ball");
		category(CATEGORY_OBJECTS, "shopping_trolley");
		category(CATEGORY_SYMBOLS, "atm");
		category(CATEGORY_SYMBOLS, "put_litter_in_its_place");
		category(CATEGORY_SYMBOLS, "potable_water");
		category(CATEGORY_SYMBOLS, "wheelchair");
		category(CATEGORY_SYMBOLS, "mens");
		category(CATEGORY_SYMBOLS, "womens");
		category(CATEGORY_SYMBOLS, "restroom");
		category(CATEGORY_SYMBOLS, "baby_symbol");
		category(CATEGORY_SYMBOLS, "wc");
		category(CATEGORY_SYMBOLS, "passport_control");
		category(CATEGORY_SYMBOLS, "customs");
		category(CATEGORY_SYMBOLS, "baggage_claim");
		category(CATEGORY_SYMBOLS, "left_luggage");
		category(CATEGORY_SYMBOLS, "warning");
		category(CATEGORY_SYMBOLS, "children_crossing");
		category(CATEGORY_SYMBOLS, "no_entry");
		category(CATEGORY_SYMBOLS, "no_entry_sign");
		category(CATEGORY_SYMBOLS, "no_bicycles");
		category(CATEGORY_SYMBOLS, "no_smoking");
		category(CATEGORY_SYMBOLS, "do_not_litter");
		category(CATEGORY_SYMBOLS, "non_potable_water");
		category(CATEGORY_SYMBOLS, "no_pedestrians");
		category(CATEGORY_SYMBOLS, "no_mobile_phones");
		category(CATEGORY_SYMBOLS, "underage");
		category(CATEGORY_SYMBOLS, "radioactive_sign");
		category(CATEGORY_SYMBOLS, "biohazard_sign");
		category(CATEGORY_SYMBOLS, "arrow_up");
		category(CATEGORY_SYMBOLS, "arrow_upper_right");
		category(CATEGORY_SYMBOLS, "arrow_right");
		category(CATEGORY_SYMBOLS, "arrow_lower_right");
		category(CATEGORY_SYMBOLS, "arrow_down");
		category(CATEGORY_SYMBOLS, "arrow_lower_left");
		category(CATEGORY_SYMBOLS, "arrow_left");
		category(CATEGORY_SYMBOLS, "arrow_upper_left");
		category(CATEGORY_SYMBOLS, "arrow_up_down");
		category(CATEGORY_SYMBOLS, "left_right_arrow");
		category(CATEGORY_SYMBOLS, "leftwards_arrow_with_hook");
		category(CATEGORY_SYMBOLS, "arrow_right_hook");
		category(CATEGORY_SYMBOLS, "arrow_heading_up");
		category(CATEGORY_SYMBOLS, "arrow_heading_down");
		category(CATEGORY_SYMBOLS, "arrows_clockwise");
		category(CATEGORY_SYMBOLS, "arrows_counterclockwise");
		category(CATEGORY_SYMBOLS, "back");
		category(CATEGORY_SYMBOLS, "end");
		category(CATEGORY_SYMBOLS, "on");
		category(CATEGORY_SYMBOLS, "soon");
		category(CATEGORY_SYMBOLS, "top");
		category(CATEGORY_SYMBOLS, "place_of_worship");
		category(CATEGORY_SYMBOLS, "atom_symbol");
		category(CATEGORY_SYMBOLS, "om_symbol");
		category(CATEGORY_SYMBOLS, "star_of_david");
		category(CATEGORY_SYMBOLS, "wheel_of_dharma");
		category(CATEGORY_SYMBOLS, "yin_yang");
		category(CATEGORY_SYMBOLS, "latin_cross");
		category(CATEGORY_SYMBOLS, "orthodox_cross");
		category(CATEGORY_SYMBOLS, "star_and_crescent");
		category(CATEGORY_SYMBOLS, "peace_symbol");
		category(CATEGORY_SYMBOLS, "menorah_with_nine_branches");
		category(CATEGORY_SYMBOLS, "six_pointed_star");
		category(CATEGORY_SYMBOLS, "aries");
		category(CATEGORY_SYMBOLS, "taurus");
		category(CATEGORY_SYMBOLS, "gemini");
		category(CATEGORY_SYMBOLS, "cancer");
		category(CATEGORY_SYMBOLS, "leo");
		category(CATEGORY_SYMBOLS, "virgo");
		category(CATEGORY_SYMBOLS, "libra");
		category(CATEGORY_SYMBOLS, "scorpius");
		category(CATEGORY_SYMBOLS, "sagittarius");
		category(CATEGORY_SYMBOLS, "capricorn");
		category(CATEGORY_SYMBOLS, "aquarius");
		category(CATEGORY_SYMBOLS, "pisces");
		category(CATEGORY_SYMBOLS, "ophiuchus");
		category(CATEGORY_SYMBOLS, "twisted_rightwards_arrows");
		category(CATEGORY_SYMBOLS, "repeat");
		category(CATEGORY_SYMBOLS, "repeat_one");
		category(CATEGORY_SYMBOLS, "arrow_forward");
		category(CATEGORY_SYMBOLS, "fast_forward");
		category(CATEGORY_SYMBOLS, "black_right_pointing_double_triangle_with_vertical_bar");
		category(CATEGORY_SYMBOLS, "black_right_pointing_triangle_with_double_vertical_bar");
		category(CATEGORY_SYMBOLS, "arrow_backward");
		category(CATEGORY_SYMBOLS, "rewind");
	}
	private static void init117(){
		category(CATEGORY_SYMBOLS, "black_left_pointing_double_triangle_with_vertical_bar");
		category(CATEGORY_SYMBOLS, "arrow_up_small");
		category(CATEGORY_SYMBOLS, "arrow_double_up");
		category(CATEGORY_SYMBOLS, "arrow_down_small");
		category(CATEGORY_SYMBOLS, "arrow_double_down");
		category(CATEGORY_SYMBOLS, "double_vertical_bar");
		category(CATEGORY_SYMBOLS, "black_square_for_stop");
		category(CATEGORY_SYMBOLS, "black_circle_for_record");
		category(CATEGORY_SYMBOLS, "eject");
		category(CATEGORY_SYMBOLS, "cinema");
		category(CATEGORY_SYMBOLS, "low_brightness");
		category(CATEGORY_SYMBOLS, "high_brightness");
		category(CATEGORY_SYMBOLS, "signal_strength");
		category(CATEGORY_SYMBOLS, "vibration_mode");
		category(CATEGORY_SYMBOLS, "mobile_phone_off");
		category(CATEGORY_SYMBOLS, "female_sign");
		category(CATEGORY_SYMBOLS, "male_sign");
		category(CATEGORY_SYMBOLS, "medical_symbol");
		category(CATEGORY_SYMBOLS, "recycle");
		category(CATEGORY_SYMBOLS, "fleur_de_lis");
		category(CATEGORY_SYMBOLS, "trident");
		category(CATEGORY_SYMBOLS, "name_badge");
		category(CATEGORY_SYMBOLS, "beginner");
		category(CATEGORY_SYMBOLS, "o");
		category(CATEGORY_SYMBOLS, "white_check_mark");
		category(CATEGORY_SYMBOLS, "ballot_box_with_check");
		category(CATEGORY_SYMBOLS, "heavy_check_mark");
		category(CATEGORY_SYMBOLS, "heavy_multiplication_x");
		category(CATEGORY_SYMBOLS, "x");
		category(CATEGORY_SYMBOLS, "negative_squared_cross_mark");
		category(CATEGORY_SYMBOLS, "heavy_plus_sign");
		category(CATEGORY_SYMBOLS, "heavy_minus_sign");
		category(CATEGORY_SYMBOLS, "heavy_division_sign");
		category(CATEGORY_SYMBOLS, "curly_loop");
		category(CATEGORY_SYMBOLS, "loop");
		category(CATEGORY_SYMBOLS, "part_alternation_mark");
		category(CATEGORY_SYMBOLS, "eight_spoked_asterisk");
		category(CATEGORY_SYMBOLS, "eight_pointed_black_star");
		category(CATEGORY_SYMBOLS, "sparkle");
		category(CATEGORY_SYMBOLS, "bangbang");
		category(CATEGORY_SYMBOLS, "interrobang");
		category(CATEGORY_SYMBOLS, "question");
		category(CATEGORY_SYMBOLS, "grey_question");
		category(CATEGORY_SYMBOLS, "grey_exclamation");
		category(CATEGORY_SYMBOLS, "exclamation");
		category(CATEGORY_SYMBOLS, "wavy_dash");
		category(CATEGORY_SYMBOLS, "copyright");
		category(CATEGORY_SYMBOLS, "registered");
		category(CATEGORY_SYMBOLS, "tm");
		category(CATEGORY_SYMBOLS, "hash");
		category(CATEGORY_SYMBOLS, "keycap_star");
		category(CATEGORY_SYMBOLS, "zero");
		category(CATEGORY_SYMBOLS, "one");
		category(CATEGORY_SYMBOLS, "two");
		category(CATEGORY_SYMBOLS, "three");
		category(CATEGORY_SYMBOLS, "four");
		category(CATEGORY_SYMBOLS, "five");
		category(CATEGORY_SYMBOLS, "six");
		category(CATEGORY_SYMBOLS, "seven");
		category(CATEGORY_SYMBOLS, "eight");
		category(CATEGORY_SYMBOLS, "nine");
		category(CATEGORY_SYMBOLS, "keycap_ten");
		category(CATEGORY_SYMBOLS, "100");
		category(CATEGORY_SYMBOLS, "capital_abcd");
		category(CATEGORY_SYMBOLS, "abcd");
		category(CATEGORY_SYMBOLS, "1234");
		category(CATEGORY_SYMBOLS, "symbols");
		category(CATEGORY_SYMBOLS, "abc");
		category(CATEGORY_SYMBOLS, "a");
		category(CATEGORY_SYMBOLS, "ab");
		category(CATEGORY_SYMBOLS, "b");
		category(CATEGORY_SYMBOLS, "cl");
		category(CATEGORY_SYMBOLS, "cool");
		category(CATEGORY_SYMBOLS, "free");
		category(CATEGORY_SYMBOLS, "information_source");
		category(CATEGORY_SYMBOLS, "id");
		category(CATEGORY_SYMBOLS, "m");
		category(CATEGORY_SYMBOLS, "new");
		category(CATEGORY_SYMBOLS, "ng");
		category(CATEGORY_SYMBOLS, "o2");
		category(CATEGORY_SYMBOLS, "ok");
		category(CATEGORY_SYMBOLS, "parking");
		category(CATEGORY_SYMBOLS, "sos");
		category(CATEGORY_SYMBOLS, "up");
		category(CATEGORY_SYMBOLS, "vs");
		category(CATEGORY_SYMBOLS, "koko");
		category(CATEGORY_SYMBOLS, "sa");
		category(CATEGORY_SYMBOLS, "u6708");
		category(CATEGORY_SYMBOLS, "u6709");
		category(CATEGORY_SYMBOLS, "u6307");
		category(CATEGORY_SYMBOLS, "ideograph_advantage");
		category(CATEGORY_SYMBOLS, "u5272");
		category(CATEGORY_SYMBOLS, "u7121");
		category(CATEGORY_SYMBOLS, "u7981");
		category(CATEGORY_SYMBOLS, "accept");
		category(CATEGORY_SYMBOLS, "u7533");
		category(CATEGORY_SYMBOLS, "u5408");
		category(CATEGORY_SYMBOLS, "u7a7a");
		category(CATEGORY_SYMBOLS, "congratulations");
		category(CATEGORY_SYMBOLS, "secret");
		category(CATEGORY_SYMBOLS, "u55b6");
	}
	private static void init118(){
		category(CATEGORY_SYMBOLS, "u6e80");
		category(CATEGORY_SYMBOLS, "black_small_square");
		category(CATEGORY_SYMBOLS, "white_small_square");
		category(CATEGORY_SYMBOLS, "white_medium_square");
		category(CATEGORY_SYMBOLS, "black_medium_square");
		category(CATEGORY_SYMBOLS, "white_medium_small_square");
		category(CATEGORY_SYMBOLS, "black_medium_small_square");
		category(CATEGORY_SYMBOLS, "black_large_square");
		category(CATEGORY_SYMBOLS, "white_large_square");
		category(CATEGORY_SYMBOLS, "large_orange_diamond");
		category(CATEGORY_SYMBOLS, "large_blue_diamond");
		category(CATEGORY_SYMBOLS, "small_orange_diamond");
		category(CATEGORY_SYMBOLS, "small_blue_diamond");
		category(CATEGORY_SYMBOLS, "small_red_triangle");
		category(CATEGORY_SYMBOLS, "small_red_triangle_down");
		category(CATEGORY_SYMBOLS, "diamond_shape_with_a_dot_inside");
		category(CATEGORY_SYMBOLS, "radio_button");
		category(CATEGORY_SYMBOLS, "black_square_button");
		category(CATEGORY_SYMBOLS, "white_square_button");
		category(CATEGORY_SYMBOLS, "white_circle");
		category(CATEGORY_SYMBOLS, "black_circle");
		category(CATEGORY_SYMBOLS, "red_circle");
		category(CATEGORY_SYMBOLS, "large_blue_circle");
		category(CATEGORY_FLAGS, "checkered_flag");
		category(CATEGORY_FLAGS, "cn");
		category(CATEGORY_FLAGS, "crossed_flags");
		category(CATEGORY_FLAGS, "de");
		category(CATEGORY_FLAGS, "es");
		category(CATEGORY_FLAGS, "flag_ac");
		category(CATEGORY_FLAGS, "flag_ad");
		category(CATEGORY_FLAGS, "flag_ae");
		category(CATEGORY_FLAGS, "flag_af");
		category(CATEGORY_FLAGS, "flag_ag");
		category(CATEGORY_FLAGS, "flag_ai");
		category(CATEGORY_FLAGS, "flag_al");
		category(CATEGORY_FLAGS, "flag_am");
		category(CATEGORY_FLAGS, "flag_ao");
		category(CATEGORY_FLAGS, "flag_aq");
		category(CATEGORY_FLAGS, "flag_ar");
		category(CATEGORY_FLAGS, "flag_as");
		category(CATEGORY_FLAGS, "flag_at");
		category(CATEGORY_FLAGS, "flag_au");
		category(CATEGORY_FLAGS, "flag_aw");
		category(CATEGORY_FLAGS, "flag_ax");
		category(CATEGORY_FLAGS, "flag_az");
		category(CATEGORY_FLAGS, "flag_ba");
		category(CATEGORY_FLAGS, "flag_bb");
		category(CATEGORY_FLAGS, "flag_bd");
		category(CATEGORY_FLAGS, "flag_be");
		category(CATEGORY_FLAGS, "flag_bf");
		category(CATEGORY_FLAGS, "flag_bg");
		category(CATEGORY_FLAGS, "flag_bh");
		category(CATEGORY_FLAGS, "flag_bi");
		category(CATEGORY_FLAGS, "flag_bj");
		category(CATEGORY_FLAGS, "flag_bl");
		category(CATEGORY_FLAGS, "flag_bm");
		category(CATEGORY_FLAGS, "flag_bn");
		category(CATEGORY_FLAGS, "flag_bo");
		category(CATEGORY_FLAGS, "flag_bq");
		category(CATEGORY_FLAGS, "flag_br");
		category(CATEGORY_FLAGS, "flag_bs");
		category(CATEGORY_FLAGS, "flag_bt");
		category(CATEGORY_FLAGS, "flag_bv");
		category(CATEGORY_FLAGS, "flag_bw");
		category(CATEGORY_FLAGS, "flag_by");
		category(CATEGORY_FLAGS, "flag_bz");
		category(CATEGORY_FLAGS, "flag_ca");
		category(CATEGORY_FLAGS, "flag_cc");
		category(CATEGORY_FLAGS, "flag_cd");
		category(CATEGORY_FLAGS, "flag_cf");
		category(CATEGORY_FLAGS, "flag_cg");
		category(CATEGORY_FLAGS, "flag_ch");
		category(CATEGORY_FLAGS, "flag_ci");
		category(CATEGORY_FLAGS, "flag_ck");
		category(CATEGORY_FLAGS, "flag_cl");
		category(CATEGORY_FLAGS, "flag_cm");
		category(CATEGORY_FLAGS, "flag_co");
		category(CATEGORY_FLAGS, "flag_cp");
		category(CATEGORY_FLAGS, "flag_cr");
		category(CATEGORY_FLAGS, "flag_cu");
		category(CATEGORY_FLAGS, "flag_cv");
		category(CATEGORY_FLAGS, "flag_cw");
		category(CATEGORY_FLAGS, "flag_cx");
		category(CATEGORY_FLAGS, "flag_cy");
		category(CATEGORY_FLAGS, "flag_cz");
		category(CATEGORY_FLAGS, "flag_dg");
		category(CATEGORY_FLAGS, "flag_dj");
		category(CATEGORY_FLAGS, "flag_dk");
		category(CATEGORY_FLAGS, "flag_dm");
		category(CATEGORY_FLAGS, "flag_do");
		category(CATEGORY_FLAGS, "flag_dz");
		category(CATEGORY_FLAGS, "flag_ea");
		category(CATEGORY_FLAGS, "flag_ec");
		category(CATEGORY_FLAGS, "flag_ee");
		category(CATEGORY_FLAGS, "flag_eg");
		category(CATEGORY_FLAGS, "flag_eh");
		category(CATEGORY_FLAGS, "flag_england");
		category(CATEGORY_FLAGS, "flag_er");
		category(CATEGORY_FLAGS, "flag_et");
		category(CATEGORY_FLAGS, "flag_eu");
		category(CATEGORY_FLAGS, "flag_fi");
	}
	private static void init119(){
		category(CATEGORY_FLAGS, "flag_fj");
		category(CATEGORY_FLAGS, "flag_fk");
		category(CATEGORY_FLAGS, "flag_fm");
		category(CATEGORY_FLAGS, "flag_fo");
		category(CATEGORY_FLAGS, "flag_ga");
		category(CATEGORY_FLAGS, "flag_gd");
		category(CATEGORY_FLAGS, "flag_ge");
		category(CATEGORY_FLAGS, "flag_gf");
		category(CATEGORY_FLAGS, "flag_gg");
		category(CATEGORY_FLAGS, "flag_gh");
		category(CATEGORY_FLAGS, "flag_gi");
		category(CATEGORY_FLAGS, "flag_gl");
		category(CATEGORY_FLAGS, "flag_gm");
		category(CATEGORY_FLAGS, "flag_gn");
		category(CATEGORY_FLAGS, "flag_gp");
		category(CATEGORY_FLAGS, "flag_gq");
		category(CATEGORY_FLAGS, "flag_gr");
		category(CATEGORY_FLAGS, "flag_gs");
		category(CATEGORY_FLAGS, "flag_gt");
		category(CATEGORY_FLAGS, "flag_gu");
		category(CATEGORY_FLAGS, "flag_gw");
		category(CATEGORY_FLAGS, "flag_gy");
		category(CATEGORY_FLAGS, "flag_hk");
		category(CATEGORY_FLAGS, "flag_hm");
		category(CATEGORY_FLAGS, "flag_hn");
		category(CATEGORY_FLAGS, "flag_hr");
		category(CATEGORY_FLAGS, "flag_ht");
		category(CATEGORY_FLAGS, "flag_hu");
		category(CATEGORY_FLAGS, "flag_ic");
		category(CATEGORY_FLAGS, "flag_id");
		category(CATEGORY_FLAGS, "flag_ie");
		category(CATEGORY_FLAGS, "flag_il");
		category(CATEGORY_FLAGS, "flag_im");
		category(CATEGORY_FLAGS, "flag_in");
		category(CATEGORY_FLAGS, "flag_io");
		category(CATEGORY_FLAGS, "flag_iq");
		category(CATEGORY_FLAGS, "flag_ir");
		category(CATEGORY_FLAGS, "flag_is");
		category(CATEGORY_FLAGS, "flag_je");
		category(CATEGORY_FLAGS, "flag_jm");
		category(CATEGORY_FLAGS, "flag_jo");
		category(CATEGORY_FLAGS, "flag_ke");
		category(CATEGORY_FLAGS, "flag_kg");
		category(CATEGORY_FLAGS, "flag_kh");
		category(CATEGORY_FLAGS, "flag_ki");
		category(CATEGORY_FLAGS, "flag_km");
		category(CATEGORY_FLAGS, "flag_kn");
		category(CATEGORY_FLAGS, "flag_kp");
		category(CATEGORY_FLAGS, "flag_kw");
		category(CATEGORY_FLAGS, "flag_ky");
		category(CATEGORY_FLAGS, "flag_kz");
		category(CATEGORY_FLAGS, "flag_la");
		category(CATEGORY_FLAGS, "flag_lb");
		category(CATEGORY_FLAGS, "flag_lc");
		category(CATEGORY_FLAGS, "flag_li");
		category(CATEGORY_FLAGS, "flag_lk");
		category(CATEGORY_FLAGS, "flag_lr");
		category(CATEGORY_FLAGS, "flag_ls");
		category(CATEGORY_FLAGS, "flag_lt");
		category(CATEGORY_FLAGS, "flag_lu");
		category(CATEGORY_FLAGS, "flag_lv");
		category(CATEGORY_FLAGS, "flag_ly");
		category(CATEGORY_FLAGS, "flag_ma");
		category(CATEGORY_FLAGS, "flag_mc");
		category(CATEGORY_FLAGS, "flag_md");
		category(CATEGORY_FLAGS, "flag_me");
		category(CATEGORY_FLAGS, "flag_mf");
		category(CATEGORY_FLAGS, "flag_mg");
		category(CATEGORY_FLAGS, "flag_mh");
		category(CATEGORY_FLAGS, "flag_mk");
		category(CATEGORY_FLAGS, "flag_ml");
		category(CATEGORY_FLAGS, "flag_mm");
		category(CATEGORY_FLAGS, "flag_mn");
		category(CATEGORY_FLAGS, "flag_mo");
		category(CATEGORY_FLAGS, "flag_mp");
		category(CATEGORY_FLAGS, "flag_mq");
		category(CATEGORY_FLAGS, "flag_mr");
		category(CATEGORY_FLAGS, "flag_ms");
		category(CATEGORY_FLAGS, "flag_mt");
		category(CATEGORY_FLAGS, "flag_mu");
		category(CATEGORY_FLAGS, "flag_mv");
		category(CATEGORY_FLAGS, "flag_mw");
		category(CATEGORY_FLAGS, "flag_mx");
		category(CATEGORY_FLAGS, "flag_my");
		category(CATEGORY_FLAGS, "flag_mz");
		category(CATEGORY_FLAGS, "flag_na");
		category(CATEGORY_FLAGS, "flag_nc");
		category(CATEGORY_FLAGS, "flag_ne");
		category(CATEGORY_FLAGS, "flag_nf");
		category(CATEGORY_FLAGS, "flag_ng");
		category(CATEGORY_FLAGS, "flag_ni");
		category(CATEGORY_FLAGS, "flag_nl");
		category(CATEGORY_FLAGS, "flag_no");
		category(CATEGORY_FLAGS, "flag_np");
		category(CATEGORY_FLAGS, "flag_nr");
		category(CATEGORY_FLAGS, "flag_nu");
		category(CATEGORY_FLAGS, "flag_nz");
		category(CATEGORY_FLAGS, "flag_om");
		category(CATEGORY_FLAGS, "flag_pa");
		category(CATEGORY_FLAGS, "flag_pe");
		category(CATEGORY_FLAGS, "flag_pf");
	}
	private static void init120(){
		category(CATEGORY_FLAGS, "flag_pg");
		category(CATEGORY_FLAGS, "flag_ph");
		category(CATEGORY_FLAGS, "flag_pk");
		category(CATEGORY_FLAGS, "flag_pl");
		category(CATEGORY_FLAGS, "flag_pm");
		category(CATEGORY_FLAGS, "flag_pn");
		category(CATEGORY_FLAGS, "flag_pr");
		category(CATEGORY_FLAGS, "flag_ps");
		category(CATEGORY_FLAGS, "flag_pt");
		category(CATEGORY_FLAGS, "flag_pw");
		category(CATEGORY_FLAGS, "flag_py");
		category(CATEGORY_FLAGS, "flag_qa");
		category(CATEGORY_FLAGS, "flag_re");
		category(CATEGORY_FLAGS, "flag_ro");
		category(CATEGORY_FLAGS, "flag_rs");
		category(CATEGORY_FLAGS, "flag_rw");
		category(CATEGORY_FLAGS, "flag_sa");
		category(CATEGORY_FLAGS, "flag_sb");
		category(CATEGORY_FLAGS, "flag_sc");
		category(CATEGORY_FLAGS, "flag_scotland");
		category(CATEGORY_FLAGS, "flag_sd");
		category(CATEGORY_FLAGS, "flag_se");
		category(CATEGORY_FLAGS, "flag_sg");
		category(CATEGORY_FLAGS, "flag_sh");
		category(CATEGORY_FLAGS, "flag_si");
		category(CATEGORY_FLAGS, "flag_sj");
		category(CATEGORY_FLAGS, "flag_sk");
		category(CATEGORY_FLAGS, "flag_sl");
		category(CATEGORY_FLAGS, "flag_sm");
		category(CATEGORY_FLAGS, "flag_sn");
		category(CATEGORY_FLAGS, "flag_so");
		category(CATEGORY_FLAGS, "flag_sr");
		category(CATEGORY_FLAGS, "flag_ss");
		category(CATEGORY_FLAGS, "flag_st");
		category(CATEGORY_FLAGS, "flag_sv");
		category(CATEGORY_FLAGS, "flag_sx");
		category(CATEGORY_FLAGS, "flag_sy");
		category(CATEGORY_FLAGS, "flag_sz");
		category(CATEGORY_FLAGS, "flag_ta");
		category(CATEGORY_FLAGS, "flag_tc");
		category(CATEGORY_FLAGS, "flag_td");
		category(CATEGORY_FLAGS, "flag_tf");
		category(CATEGORY_FLAGS, "flag_tg");
		category(CATEGORY_FLAGS, "flag_th");
		category(CATEGORY_FLAGS, "flag_tj");
		category(CATEGORY_FLAGS, "flag_tk");
		category(CATEGORY_FLAGS, "flag_tl");
		category(CATEGORY_FLAGS, "flag_tm");
		category(CATEGORY_FLAGS, "flag_tn");
		category(CATEGORY_FLAGS, "flag_to");
		category(CATEGORY_FLAGS, "flag_tr");
		category(CATEGORY_FLAGS, "flag_tt");
		category(CATEGORY_FLAGS, "flag_tv");
		category(CATEGORY_FLAGS, "flag_tw");
		category(CATEGORY_FLAGS, "flag_tz");
		category(CATEGORY_FLAGS, "flag_ua");
		category(CATEGORY_FLAGS, "flag_ug");
		category(CATEGORY_FLAGS, "flag_um");
		category(CATEGORY_FLAGS, "flag_uy");
		category(CATEGORY_FLAGS, "flag_uz");
		category(CATEGORY_FLAGS, "flag_va");
		category(CATEGORY_FLAGS, "flag_vc");
		category(CATEGORY_FLAGS, "flag_ve");
		category(CATEGORY_FLAGS, "flag_vg");
		category(CATEGORY_FLAGS, "flag_vi");
		category(CATEGORY_FLAGS, "flag_vn");
		category(CATEGORY_FLAGS, "flag_vu");
		category(CATEGORY_FLAGS, "flag_wales");
		category(CATEGORY_FLAGS, "flag_wf");
		category(CATEGORY_FLAGS, "flag_ws");
		category(CATEGORY_FLAGS, "flag_xk");
		category(CATEGORY_FLAGS, "flag_ye");
		category(CATEGORY_FLAGS, "flag_yt");
		category(CATEGORY_FLAGS, "flag_za");
		category(CATEGORY_FLAGS, "flag_zm");
		category(CATEGORY_FLAGS, "flag_zw");
		category(CATEGORY_FLAGS, "fr");
		category(CATEGORY_FLAGS, "gb");
		category(CATEGORY_FLAGS, "it");
		category(CATEGORY_FLAGS, "jp");
		category(CATEGORY_FLAGS, "kr");
		category(CATEGORY_FLAGS, "rainbow_flag");
		category(CATEGORY_FLAGS, "ru");
		category(CATEGORY_FLAGS, "triangular_flag_on_post");
		category(CATEGORY_FLAGS, "us");
		category(CATEGORY_FLAGS, "waving_black_flag");
		category(CATEGORY_FLAGS, "waving_white_flag");
	}
	public static final int utf16_max_length=14;
	
	static void initAll(){
		init1();
		init2();
		init3();
		init4();
		init5();
		init6();
		init7();
		init8();
		init9();
		init10();
		init11();
		init12();
		init13();
		init14();
		init15();
		init16();
		init17();
		init18();
		init19();
		init20();
		init21();
		init22();
		init23();
		init24();
		init25();
		init26();
		init27();
		init28();
		init29();
		init30();
		init31();
		init32();
		init33();
		init34();
		init35();
		init36();
		init37();
		init38();
		init39();
		init40();
		init41();
		init42();
		init43();
		init44();
		init45();
		init46();
		init47();
		init48();
		init49();
		init50();
		init51();
		init52();
		init53();
		init54();
		init55();
		init56();
		init57();
		init58();
		init59();
		init60();
		init61();
		init62();
		init63();
		init64();
		init65();
		init66();
		init67();
		init68();
		init69();
		init70();
		init71();
		init72();
		init73();
		init74();
		init75();
		init76();
		init77();
		init78();
		init79();
		init80();
		init81();
		init82();
		init83();
		init84();
		init85();
		init86();
		init87();
		init88();
		init89();
		init90();
		init91();
		init92();
		init93();
		init94();
		init95();
		init96();
		init97();
		init98();
		init99();
		init100();
		init101();
		init102();
		init103();
		init104();
		init105();
		init106();
		init107();
		init108();
		init109();
		init110();
		init111();
		init112();
		init113();
		init114();
		init115();
		init116();
		init117();
		init118();
		init119();
		init120();
	}

}